/*
 LaTeXLine.java - November 13, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.latex;
 
import jpicedt.format.input.pstricks.PsRPutExpression;
import jpicedt.format.input.util.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;

/**
 * Parses LaTeX lines, with or without arrows, i.e. \\line or \\vector. This is to be used as a child expression of either
 * {@link #LaTeXPutExpression LaTeXPutExpression} or 
 * {@link #jpicedt.format.input.pstricks.PsRPutExpression jpicedt.format.input.pstricks.PsRPutExpression}.
 * <ul>
 * <li> \\line(x,y){hor-len} -> line starting at (x,y), and having hor-len as its horizonal extent (except
 *  for vertical lines, where hor-len specifies the vertical extent.</li>
 * <li> \\vector(x,y){hor-len} -> same thing, yet with an arrow</li>
 * </ul>
 * @author Sylvain Reynal
 * @since jpicedt 1.4pre2
 * @version $Id: LaTeXLine.java,v 1.3.2.1 2007/09/02 11:56:01 reynal Exp $
 */
public class LaTeXLine extends SequenceExpression implements PicObjectConstants, ExpressionConstants {

	private double x,y; // used to compute the line slope
	private Pool pool;

	public LaTeXLine(Pool pl) {

		super(true);// throw IncompleteExpression
		this.pool = pl;

		// \line -> set first polygon point
		this.add(new AlternateExpression(
		             new LaTeXInstanciationExpression("\\line", new PicMultiCurve(new PicPoint(),new PicPoint()), pool),
		             new LaTeXInstanciationExpression("\\vector", new PicMultiCurve(new PicPoint(),new PicPoint()), pool){
			             public void action(ParserEvent e) {
				             super.action(e);
				             pool.currentObj.setAttribute(RIGHT_ARROW, Arrow.ARROW_HEAD);
			             }}));
		this.add(WHITE_SPACES_OR_EOL);
		// (x,y) -> set slope
		this.add(
		    new PicPointExpression("(", ",", ")"){
			    public void action(ParserEvent e) {
				    if (DEBUG) System.out.println(e);
				    PicPoint pt = (PicPoint)e.getValue();
				    x = pt.x ; y = pt.y;
			    }
		    }
		);
		this.add(WHITE_SPACES_OR_EOL);
		// {
		this.add(new LiteralExpression("{"));
		// len} -> must be .ge. 0
		this.add(
		    new NumericalExpression(DOUBLE, POSITIVE, "}", true) {
			    public void action(ParserEvent e) {
				    if (DEBUG) System.out.println(e);
				    // horizontal extent (except for vertical lines) :
				    double len = ((Double)e.getValue()).doubleValue() * pool.getDouble(LaTeXParser.KEY_UNIT_LENGTH);
				    PicMultiCurve pp = ((PicMultiCurve)(pool.currentObj));
					// set put point:
					PicPoint putPoint = (PicPoint)pool.get(PsRPutExpression.KEY_RPUT_POINT);
					if (putPoint==null) putPoint = (PicPoint)pool.get(LaTeXPutExpression.KEY_PUT_POINT);
					if (putPoint==null) putPoint = new PicPoint();
				    // vertical line :
				    PicPoint pt1 = new PicPoint(putPoint);
				    if (x==0){
					    if (y>=0) pt1.translate(0,len);
					    else pt1.translate(0,-len);
				    }
				    // other lines :
				    else {
					    if (x>=0) pt1.x += len;
					    else pt1.x -= len;
					    if (y>=0) pt1.y += len * Math.abs(y/x);
					    else pt1.y -= len * Math.abs(y/x);
				    }
				    pp.setPoint(0,putPoint,pp.FREELY); // first end-point (and drag corresponding control-point accordingly)
				    pp.setPoint(3,pt1,pp.FREELY); // second end-point (ibid.)
			    }
		    }
		);
	}
}


