/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.latex;

import jpicedt.format.input.util.RootExpression;
import jpicedt.format.input.util.Pool;
import jpicedt.format.input.util.ExpressionConstants;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.*;

import java.awt.Color;
import java.io.*;
import java.util.*;

/**
 * Grammar rules for the LaTeX's picture environment parser
 */
public class LaTeXParser extends RootExpression implements ExpressionConstants {

	/** key associated with unit-length parameter (double) */
	public static final Pool.Key KEY_UNIT_LENGTH = new Pool.Key("latex.unit-length");

	/** key associated with the shared attribute set (PicAttributeSet) */
	public static final Pool.Key KEY_ATTRIBUTES = new Pool.Key("latex.attributes");

	private Pool pool;
	
	/**
	 * Creates a new LaTeXParser, and build grammar rules
	 * @param pool a hashmap used to share variables across the grammar tree
	 */
	public LaTeXParser(Pool pool) {
		this.pool = pool;

		// 0°) unitlength and begin picture
		add(new UnitLengthExpression(pool)); // \\unitlength
		add(new BeginPictureExpression()); // not used (but otherwise, it's considered a non-parsable exp)
		
		// 1°) parameters
		add(new LineThicknessExpression(pool)); 	// \\linethickness

		// 2°) groups
		add(new PicGroupExpression(pool));       // %Begin|End group

		// 3°) lines & polygons
		add(new PicLineExpression(pool));		// %PicLine
		add(new PicPolygonExpression(pool));	// %PicPolygon

		// 4°) frames
		add(new PicRectangleExpression(pool));	// %PicRectangle

		// 5°) ellipses and arcs
		add(new PicEllipseExpression(pool));	// %PicEllipse

		// 6°) splines 
		add(new PicBezierExpression(pool));	// %PicBezier (quad), \\qbezier or \\bezier{N}

		// 7°) text and misc.
		add(new LaTeXPutExpression(pool));	// \\put(x,y){...} -> PicText, PicPut (also support LaTeX's "\\circle" command w/o PicEdt comment line)
		
		// 8°) comment
		add(new jpicedt.format.input.util.CommentExpression("%"));

		// [todo] multiput
		
		// throw REParserException.EndOfPicture to signals that a \end{picture} was found
		// and that the parsing process ends up here (otherwise, we stop at EOF) 
		add(new EndPictureExpression());
	}


	/**
	 * reinit LaTeX related parameters
	 */
	public void reinit(){

		PicAttributeSet ltxSet = new PicAttributeSet();
		double unitLength = 1; // 1mm
		pool.put(KEY_UNIT_LENGTH, new Double(unitLength));
		pool.put(KEY_ATTRIBUTES, ltxSet);
	}
}
