/*
 LaTeXPutExpression.java - 1999 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://trashx.ensea.fr/jpicedt/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.latex;

import jpicedt.format.input.util.*;
import jpicedt.format.input.pstricks.PsBox;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.*;

/**
 *  \\put(x,y){argument} 
 * <p>
 * Depending on the argument found, instanciates :
 * <ul> 
 * <li> a PicText : \\makebox, \\framebox, \\dashbox or LR-argument </li>
 * <li> a PicMultiCurve : \\line, \\vector.</li>
 * <li>a PicEllipse : \\circle</li>
 * </ul>
 * <p>
 * The parsing code sets the following Pool's key/value pair (as set by the enclosing XXXPutExpression beforehands):
 * <ul>
 * <li> KEY_PUT_POINT: \\put's (x,y) parameter </li>
 * </ul>
 * This key/value pairs may then be used by client expression, e.g. PsBox, LaTeXCircle,...
 * @author Sylvain Reynal
 * @since jpicedt 1.3
 * @version $Id: LaTeXPutExpression.java,v 1.8.2.1 2007/09/02 11:56:02 reynal Exp $
 *
 */
public class LaTeXPutExpression extends SequenceExpression implements ExpressionConstants,PicObjectConstants {

	/** key entry in Pool containing the (x,y) parameter */ 
	public static final Pool.Key KEY_PUT_POINT = new Pool.Key("put-point");
	
	private Pool pool;
	
	public LaTeXPutExpression(Pool pl) {

		super(true);
		pool = pl;

		add(new LiteralExpression("\\put"));
		add(WHITE_SPACES_OR_EOL);
		// "(x,y)" => set current "putPoint":
		add(
		    new PicPointExpression("(", ",", ")") {
			    public void action(ParserEvent e) {
				    if (DEBUG) System.out.println(e);
				    PicPoint putPoint = ((PicPoint) e.getValue()).toMm(pool.getDouble(LaTeXParser.KEY_UNIT_LENGTH));// global var
					pool.put(KEY_PUT_POINT, putPoint);
			    }
		    });
		add(WHITE_SPACES_OR_EOL);

		// {stuff} :
		AlternateExpression putArgs = new AlternateExpression();
		putArgs.add(new LaTeXBox(pool));// \\makebox, \\framebox, \\frame, \\dashbox
		putArgs.add(new LaTeXCircle(pool));// \\circle{d} and \\circle*{d}
		putArgs.add(new LaTeXOval(pool));// \\oval(w,h)
		putArgs.add(new LaTeXLine(pool));// \\line(x,y){len} and \\vector
		// PsTricks commands wrapped in a \\put command: (need input.pstricks.PsBox !)
		putArgs.add(new PsBox(pool, PsBox.RECTANGLE_BOX));// \\psframebox
		putArgs.add(new PsBox(pool, PsBox.CIRCLE_BOX));// \\pscirclebox
		putArgs.add(new PsBox(pool, PsBox.OVAL_BOX));// \\psovalbox
		
		putArgs.add(new LRArgument()); // pure HR argument (must always be added last)
		SequenceExpression putArgsEOL = new SequenceExpression(WHITE_SPACES_OR_EOL, putArgs, false);
		add(new EnclosingExpression("{", putArgsEOL, "}"));

	}

	/** called when this SequenceExpression was successfully parsed -> reinit locals and Pool's keys for next time */
	public void action(ParserEvent e){
		if (DEBUG) System.out.println(e);
		pool.put(KEY_PUT_POINT, null);
	}

	public String toString(){
		return "[LaTeXPutExpression]";
	}




	/**
	 * handles content of "{stuff}" when no other expression matches, by instanciating a PicText
	 * with "stuff" as the PicText string
	 */
	class LRArgument extends RepeatExpression {

		StringBuffer lrArgument = new StringBuffer(); 
		
		LRArgument(){
			super(null,0,AT_LEAST);
			setPattern(new LRArgumentSwallower());
		}

		public void action(ParserEvent e){
			if (DEBUG) System.out.println(e);
			String s = Context.removeRedundantWhiteSpaces(lrArgument.toString());
			// set put point :
			PicPoint putPoint = (PicPoint)pool.get(KEY_PUT_POINT);
			if (putPoint==null) putPoint = new PicPoint(); // security, default to (0,0)
			pool.currentObj = new PicText(putPoint, s, pool.getAttributeSet(LaTeXParser.KEY_ATTRIBUTES));
			pool.currentGroup.addChild(pool.currentObj);
			lrArgument = new StringBuffer(); // reset for next time
		}

		/**
		 * swallow as many chars as possible and push them in "lrArgument" buffer, replacing
		 * CR by whitespaces, as TeX does.
		 */
		class LRArgumentSwallower extends WildCharExpression {

			public LRArgumentSwallower(){
				super(ANY_CHAR_EOL);
			}
			public void action(ParserEvent e){
				if (DEBUG) System.out.println(e);
				Character cc = getCharacter();
				if (cc.charValue() == '\n') lrArgument.append(' ');
				else lrArgument.append(cc);
			}
		}
	}

}

