/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.latex;

import jpicedt.format.input.util.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PEToolKit;

/**
 * Parser expression for :
 * <p>
 * %Bezier 0|1 0|1 (x1,y1)(xC,yC)(x2,y2) dash=value<br>
 * Any string<br>
 * %End Bezier<br>
 * where dash is optional. The whole expression MUST hold on a single line (as opposed to LaTeX or eepic commands)<br>
 * <p>
 * Or : <br>
 * \\qbezier(x1,y1)(xc,yc)(x2,y2)<br>
 * \\bezier{n}(x1,y1)(xc,yc)(x2,y2)<br>
 * @author Sylvain Reynal
 * @since jpicedt 1.3
 * @version $Id: PicBezierExpression.java,v 1.6.2.1 2007/09/02 11:56:02 reynal Exp $
 *
 */
public class PicBezierExpression extends AlternateExpression  implements ExpressionConstants {

	Pool pool;
	PicPoint p1 = new PicPoint(); // buffer for quad => cubic conversion
	PicPoint pCtrl = new PicPoint();
	PicPoint p2 = new PicPoint();
	
    public PicBezierExpression(Pool pl){

		this.pool=pl;
		
		SequenceExpression picBezierExpression = new SequenceExpression(true); // throw IncompleteSequence Exception 
		picBezierExpression.add(new LaTeXInstanciationExpression("%Bezier", new PicMultiCurve(new PicPoint()),pool));// curve reduced to a single point
		picBezierExpression.add(WHITE_SPACES);
		picBezierExpression.add(new PicArrowTypeExpression(pool));
		picBezierExpression.add(new BufferPicPointExpression(p1));
		picBezierExpression.add(WHITE_SPACES);
		picBezierExpression.add(new BufferPicPointExpression(pCtrl));
		picBezierExpression.add(WHITE_SPACES);
		picBezierExpression.add(new ConvertQuadToCubicExpression(p2));
		picBezierExpression.add(WHITE_SPACES);
		picBezierExpression.add(new OptionalExpression(new PicDashStatement(pool)));
		picBezierExpression.add(new PicEndExpression("%End Bezier"));
		this.add(picBezierExpression);
		
		SequenceExpression latexBezierExpression = new SequenceExpression(true);
		latexBezierExpression.add(new AlternateExpression(
			new LaTeXInstanciationExpression("\\qbezier", new PicMultiCurve(new PicPoint()),pool),
			new SequenceExpression(
				new LaTeXInstanciationExpression("\\bezier{", new PicMultiCurve(new PicPoint()),pool),
				new NumericalExpression(INTEGER, POSITIVE, "}", true),
				true
		)));
		latexBezierExpression.add(WHITE_SPACES_OR_EOL);
		latexBezierExpression.add(new BufferPicPointExpression(p1));
		latexBezierExpression.add(WHITE_SPACES);
		latexBezierExpression.add(new BufferPicPointExpression(pCtrl));
		latexBezierExpression.add(WHITE_SPACES);
		latexBezierExpression.add(new ConvertQuadToCubicExpression(p2));
		this.add(latexBezierExpression);
    }
	
	public String toString(){
		return "[PicBezierExpression]";
	}
	
	class BufferPicPointExpression extends PicPointExpression {
		PicPoint ptBuf;
		public BufferPicPointExpression(PicPoint pt){
			super("(",",",")");
			this.ptBuf = pt;
		}
		
		public void action(ParserEvent e){
			if (DEBUG) System.out.println(e);
			PicPoint pt = getPicPoint();
			double unitLength = pool.getDouble(LaTeXParser.KEY_UNIT_LENGTH);
			ptBuf.setCoordinates(pt.toMm(unitLength));
		}
	}
	
	class ConvertQuadToCubicExpression extends BufferPicPointExpression {
		public ConvertQuadToCubicExpression(PicPoint pt){
			super(pt);
		}
		
		public void action(ParserEvent e){
			super.action(e); // store p2
			PicPoint[] cubicPts = PEToolKit.convertQuadBezierToCubic(p1,pCtrl,p2);
			pool.currentObj.setPoint(0,cubicPts[0],null);
			((PicMultiCurve)pool.currentObj).curveTo(cubicPts[1], cubicPts[2], cubicPts[3]);
		}
	}

} // PicBezierExpression
