/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.model.*;
import java.awt.*;

/**
 * PsTricks graphics parameters, e.g. "linewidth=5pt, fillcolor=blue,..." (no leading, nor trailing brackets,
 * these should be added to an enclosing expression)<p>  
 * Currently supported parameters :<br><ul>
 * <li>xunit, yunit, runit, unit
 * <li>linestyle=xxx
 * <li>linewidth=xxx (with unit) + dash'n dot parameters
 * <li>linecolor=xxx (may be a native PsTrick colour, see predefined colours in PsTricksFormater, OR a user-defined colours)
 * <li>fillstyle=xxx 
 * <li>fillcolor=xxx (same note as for linecolor)
 * <li>shadow parameters
 * <li>hatch parameters.
 * </ul>
 * <p>
 * Depending on the value of "attributeSetKey" given to the constructor, 
 * these parameters may act on the attribute set of the current Element (in the Pool), or on another
 * attribute set (for instance, PsTricks's registers).
 * @author Sylvain Reynal
 * @version $Id: PSTParametersExpression.java,v 1.5.2.1 2007/09/02 11:56:12 reynal Exp $ 
 * <p>
 */
public class PSTParametersExpression extends RepeatExpression implements PicObjectConstants {


	/**
	 * @param pool parser's pool
	 * @param attributeSetKey used to fetch the attribute set in which parsed parameters are stored. 
	 */
	public PSTParametersExpression(Pool pool, Pool.Key attributeSetKey){

		super(null, 0, AT_LEAST); // doesn't throw IncompleteSequenceException (which means that there might be nothing b/w [ and ])

		// an alternate list of every possible parameters that gets repeated until there are no more parameters to be parsed !
		AlternateExpression altParms = new AlternateExpression();

		// ******************************************************************
		// **************** UNIT LENGTH *******************************
		// ******************************************************************
		altParms.add(new PsUnitLengthExpression(pool,PsUnitLengthExpression.XUNIT)); // always push in pool !
		altParms.add(new PsUnitLengthExpression(pool,PsUnitLengthExpression.YUNIT)); // always push in pool !
		altParms.add(new PsUnitLengthExpression(pool,PsUnitLengthExpression.RUNIT)); // always push in pool !
		altParms.add(new PsUnitLengthExpression(pool,PsUnitLengthExpression.UNIT)); // always push in pool !
		
		// ******************************************************************
		// **************** STROKE properties *******************************
		// ******************************************************************

		// =============== LINEWIDTH ===========================================
		// ex : linewidth=13mm or linewidth=5.6 (default to current
		altParms.add(new PSTLengthParameter(pool, "linewidth", LINE_WIDTH, attributeSetKey));

		// ============= LINECOLOR ==============================================
		// ex : linecolor=green (native) or linecolor=mygray2 (user-defined)
		altParms.add(new PSTColorExpression(pool, "linecolor", LINE_COLOR, attributeSetKey));

		// ============ LINESTYLE ===================================================
		altParms.add(new PSTLineStyleExpression(pool, attributeSetKey));

		// ============= DASH ==============================================
		altParms.add(new PSTDashExpression(pool, attributeSetKey));

		// =============== DOTSEP ===========================================
		altParms.add(new PSTLengthParameter(pool, "dotsep", DOT_SEP, attributeSetKey));

		// ============= DOUBLELINE (true/false) ==============================================
		altParms.add(new PSTBooleanExpression(pool, "doubleline", DOUBLE_LINE, attributeSetKey));

		// =============== DOUBLESEP ===========================================
		altParms.add(new PSTLengthParameter(pool, "doublesep", DOUBLE_SEP, attributeSetKey));

		// ============= DOUBLECOLOR ==============================================
		altParms.add(new PSTColorExpression(pool, "doublecolor", DOUBLE_COLOR, attributeSetKey));

		// ******************************************************************
		// **************** FILL properties *******************************
		// ******************************************************************

		// ============ FILLSTYLE ===================================================
		altParms.add(new PSTFillStyleExpression(pool, attributeSetKey));

		// ============= FILLCOLOR ==============================================
		altParms.add(new PSTColorExpression(pool, "fillcolor", FILL_COLOR, attributeSetKey));

		// =============== HATCHWIDTH ===========================================
		altParms.add(new PSTLengthParameter(pool, "hatchwidth", HATCH_WIDTH, attributeSetKey));

		// =============== HATCHSEP ===========================================
		altParms.add(new PSTLengthParameter(pool, "hatchsep", HATCH_SEP, attributeSetKey));

		// ============= HATCHCOLOR ==============================================
		altParms.add(new PSTColorExpression(pool, "hatchcolor", HATCH_COLOR, attributeSetKey));

		// ============= HATCHANGLE ==============================================
		altParms.add(new PSTAngleParameter(pool, "hatchangle", HATCH_ANGLE, attributeSetKey));


		// ******************************************************************
		// **************** SHADOW properties *******************************
		// ******************************************************************

		// ============= SHADOW (true/false) ==============================================
		altParms.add(new PSTBooleanExpression(pool, "shadow", SHADOW, attributeSetKey));

		// ============= SHADOWANGLE ==============================================
		altParms.add(new PSTAngleParameter(pool, "shadowangle", SHADOW_ANGLE, attributeSetKey));

		// =============== SHADOWSIZE ===========================================
		altParms.add(new PSTLengthParameter(pool, "shadowsize", SHADOW_SIZE, attributeSetKey));

		// ============= SHADOWCOLOR ==============================================
		altParms.add(new PSTColorExpression(pool, "shadowcolor", SHADOW_COLOR, attributeSetKey));

		// ******************************************************************
		// **************** DOTS properties *******************************
		// ******************************************************************
		
		// ============== DOTSTYLE ===============================================
		altParms.add(new PSTDotStyleExpression(pool, attributeSetKey));
		
		// ============== DOTANGLE ================================================
		altParms.add(new PSTAngleParameter(pool, "dotangle", POLYDOTS_ANGLE, attributeSetKey));
		
		// [SR:pending] dotsize, dotscale.
		
		
		// [pending] =========== NOT PARSABLE PARAMETERS (either not implemented yet, or not syntaxically correct)
		// Note : non-implemented parameters are, so far : "dimen", "border" and related, "origin" and "swapaxes"
		altParms.add(new WildCharExpression(ANY_CHAR_EOL)); 
		// this exp. will be successfully parsed as many times
		// as needed for the non-implemented parameter to be completely swallowed.
		// Besides, it may also parse separating "," and EOL/spaces (even where they aren't accepted by PsTricks)

		// ================================================================
		setPattern(altParms);
	}
	
	public String toString(){
		return "[PSTParametersExpression]";
	}
}
