/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.PicPoint;

/**
 * Rules for parsing cubic splines with a PsTricks syntax :
 * <ul>
 * <li>\\psbezier[parameters]{arrows}(x0,y0)(x1,y1)(x2,y2)(x3,y3)
 * <li>\\psbezier*[parameters]{arrows}(x0,y0)(x1,y1)(x2,y2)(x3,y3)
 * </ul>
 * or PsTricks's object previously defined by help of a \\newpsobject command.
 * @author Sylvain Reynal
 * @version $Id: PsBezierExpression.java,v 1.5.2.1 2007/09/02 11:56:12 reynal Exp $ 
 *
 */
public class PsBezierExpression extends SequenceExpression implements PicObjectConstants {

	private Pool pool;
	private double xQ0, yQ0;

	/**
	 * tag = \\psbezier and proto = new PicBezierCubic.
	 */
	public PsBezierExpression(Pool pl){
		this(pl, "\\psbezier");
	}

	/** 
	 * Constructor allowing to set a non-standard tag. Used e.g. by PsObjectExpression
	 */
	public PsBezierExpression(Pool pl, String tag){

		super(true); // throw IncompleteSequence Exception
		pool = pl;

		PicMultiCurve proto = new PicMultiCurve(new PicPoint(),new PicPoint(),new PicPoint(),new PicPoint());
		add(new PSTInstanciationExpression(tag, proto,pool));
		add(WHITE_SPACES_OR_EOL);

		add(new OptionalExpression(new StarExpression(pool))); // filled ?
		add(WHITE_SPACES_OR_EOL);

		// add optional param here
		add(new OptionalExpression(new EnclosingExpression("[",new PSTParametersExpression(pool,Pool.CURRENT_OBJ_ATTRIBUTES),"]"))); // push in object's attributeSet
		add(WHITE_SPACES_OR_EOL);

		// arrows :
		add(new OptionalExpression(new PSTArrowExpression(pool)));
		add(WHITE_SPACES_OR_EOL);

		// first point
		add(new PSTPicPointExpression(0,null,pool)); // null => set specification point
		add(WHITE_SPACES_OR_EOL);

		// first control point
		add(new PSTPicPointExpression(1,null,pool));
		add(WHITE_SPACES_OR_EOL);

		// second control point
		add(new PSTPicPointExpression(2,null,pool));
		add(WHITE_SPACES_OR_EOL);

		// last point
		add(new PSTPicPointExpression(3,null,pool));
	}

	public String toString(){
		return "[PsBezierExpression]";
	}

}
