/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;

/**
 * Used by PSTParamExpression to parse statement involving the 4 pstricks length registers, that is
 * "unit=1mm,xunit=1mm,yunit=1mm,runit=1mm" (or pt or cm), where setting "unit" sets "xunit", 
 * "yunit" and "runit" simultaneously. 
 * <p>
 * Parsed value are stored in Pool's hashtable using 
 * PstricksParser.KEY_X_UNIT and related keys.
 * <p>
 * Note that default PsTricks's unit is 1cm ! (whereas jPicEdt default unit is 1mm) hence no unit leads
 * to pstXunit = 0.1, etc...
 * <p>
 * Allowed syntaxs are : "0.11" (i.e. cm), "0.11mm", "0.11cm", "0.11pt" 
 * <p>
 * According to PsTricks's doc, whitespaces are allowed ONLY after the comma (see PsTricks doc. page 6),
 * though here we use a StatementExpression which swallows them, which makes it less stringent.
 * <p>
 */
public class PsUnitLengthExpression extends SequenceExpression  implements ExpressionConstants {

	private Pool pool;
	public static final String XUNIT="xunit";
	public static final String YUNIT="yunit";
	public static final String RUNIT="runit";
	public static final String UNIT="unit";
	private String type;

	/**
	 * unit actually sets the following three parameters : xunit, yunit, and runit.
	 * @param type XUNIT, YUNIT,...
	 */
	public PsUnitLengthExpression(Pool pl, String type){

		super(false);
		this.pool = pl;
		if (type!=XUNIT && type!=YUNIT && type!= RUNIT && type!=UNIT) throw new IllegalArgumentException(type);
		this.type = type;

		add(new StatementExpression(type,"=",null,DOUBLE,POSITIVE){ // no postfix, but possibly whitespaces before
			            public void action(ParserEvent e){
				            if(DEBUG) System.out.println(e);
							setUnit(10.0 * ((Number)e.getValue()).doubleValue());// suppose it's in cm (the default) and scale to mm
			            }});
		add(new OptionalExpression(
				new AlternateExpression(
					new LiteralExpression("cm"), // do nothing !
		            new LiteralExpression("mm"){
						public void action(ParserEvent e){
							if (DEBUG) System.out.println(e);
							scaleUnit(0.1);}},
		            new LiteralExpression("pt"){
			            public void action(ParserEvent e){
							if (DEBUG) System.out.println(e);
				            scaleUnit(2.56/72.0);}}))); // rescale from pt to mm
	}
	
	private void setUnit(double value){
		
		if (type==XUNIT) pool.put(PstricksParser.KEY_X_UNIT, value);
		else if (type==YUNIT) pool.put(PstricksParser.KEY_Y_UNIT, value);
		else if (type==RUNIT) pool.put(PstricksParser.KEY_R_UNIT, value);
		else if (type==UNIT) {
			pool.put(PstricksParser.KEY_X_UNIT, value);
			pool.put(PstricksParser.KEY_Y_UNIT, value);
			pool.put(PstricksParser.KEY_R_UNIT, value);
		}
	}

	private void scaleUnit(double scale){
		if (type==XUNIT) pool.put(PstricksParser.KEY_X_UNIT, scale*pool.getDouble(PstricksParser.KEY_X_UNIT));
		else if (type==YUNIT) pool.put(PstricksParser.KEY_Y_UNIT, scale*pool.getDouble(PstricksParser.KEY_Y_UNIT));
		else if (type==RUNIT) pool.put(PstricksParser.KEY_R_UNIT, scale*pool.getDouble(PstricksParser.KEY_R_UNIT));
		else if (type==UNIT) {
			pool.put(PstricksParser.KEY_X_UNIT, scale*pool.getDouble(PstricksParser.KEY_X_UNIT));
			pool.put(PstricksParser.KEY_Y_UNIT, scale*pool.getDouble(PstricksParser.KEY_Y_UNIT));
			pool.put(PstricksParser.KEY_R_UNIT, scale*pool.getDouble(PstricksParser.KEY_R_UNIT));
		}
	}

} // PsUnitLengthExpression
