/*
 PstricksParser.java - July 24, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.RootExpression;
import jpicedt.format.input.util.EnclosingExpression;
import jpicedt.format.input.util.Pool;
import jpicedt.format.input.util.ExpressionConstants;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.*;

import java.awt.Color;
import java.io.*;
import java.util.*;

/**
 * Root expression for the Pstricks parser, containing grammar rules for the pstricks format.
 * Pstricks support is pretty ok, though add-ons (like pst-nodes.sty) are still missing.
 * <p>
 * This class consists of grammar rules, wrapped in a RootExpression which can itself be directly added
 * to the main tree (for instance, using DefaultParser.addGrammar()).
 * <P> 
 * Note : the current parser used across the 
 * JPicEdt application can be retrieved using jpicedt.JPicEdt.getParser().
 * <p>
 * [TODO] : \psgrid, \psaxes, \pscustom, arrow size, dot style.
 * @author Sylvain Reynal
 * @version $Id: PstricksParser.java,v 1.7.2.1 2007/09/02 11:56:12 reynal Exp $ 
 */
public class PstricksParser extends RootExpression implements ExpressionConstants {

	/** key associated with \\psxunit register (double) */
	public static final Pool.Key KEY_X_UNIT = new Pool.Key("pst.x-unit");
	/** key associated with \\psyunit register (double) */
	public static final Pool.Key KEY_Y_UNIT = new Pool.Key("pst.y-unit");
	/** key associated with \\psrunit register (double) */
	public static final Pool.Key KEY_R_UNIT = new Pool.Key("pst.r-unit");
	/** key associated with the attribute set which stores PsTricks default graphical parameters (PicAttributeSet) */
	public static final Pool.Key KEY_ATTRIBUTES = new Pool.Key("pst.attributes");
	/** user-defined colours (a HashMap which gets filled by UserDefinedColorsExpression, see \\newgray, \\newrgbcolor,... in PsTricks documentation p.5) */
	public static final Pool.Key KEY_USER_COLOURS = new Pool.Key("pst.user-colours"); 
	/** \\newpsobject : key associated with the HashMap that associates pairs "macroName -> param_string",
	 * e.g. "\\myline" -> "linecolor=green,filltype=solid" (HashMap) */
	public static final Pool.Key KEY_NEWPSOBJECTS = new Pool.Key("pst.newpsobjects");
	
	private Pool pool;

	/**
	 * Creates a new PstricksParser, and build grammar rules
	 */
	public PstricksParser(Pool pool) {

		this.pool = pool;
		
		// 0°) begin picture
		add(new BeginPsPictureExpression()); // not used (but otherwise, it's considered a non-parsable exp)
		
		// 1°) parameters
		add(new EnclosingExpression("\\psset{", new PSTParametersExpression(pool, KEY_ATTRIBUTES), "}")); // \\psset{...} 
		add(new UserDefinedColorExpression(pool)); // \\newrgbcolor{...}() or similar commands

		// 2°) lines and polygons :
		add(new PsQLineExpression(pool)); // \\qline (PsTricks)
		add(new PsPolygonExpression(pool,PsPolygonExpression.POLYGON));	// \\pspolygon (closed) 
		add(new PsPolygonExpression(pool,PsPolygonExpression.LINE));	//\\psline (open)
		add(new PsPolygonExpression(pool,PsPolygonExpression.DOTS));	//\\psdots

		// 3°) frames
		add(new PsFrameExpression(pool));

		// 4°) ellipses and arcs
		add(new PsEllipseExpression(pool));	// \\psellipse
		add(new PsCircleExpression(pool));	// \\pscircle
		add(new PsArcExpression(pool,PsArcExpression.ARC));	// \\psarc
		add(new PsArcExpression(pool,PsArcExpression.WEDGE));	// \\pswedge
		add(new PsQDiskExpression(pool)); // \\qdisk

		// 5°) splines 
		add(new PsBezierExpression(pool));	// \\psbezier

		// 6°) text and misc.
		add(new PsRPutExpression(pool)); // \\rput...
		
		// 7°) comment
		add(new jpicedt.format.input.util.CommentExpression("%"));

		// 8°) \\newpsobject
		add(new PsObjectExpression(this,pool));
		
		// throw REParserException.EndOfPicture to signals that a \end{pspicture} was found
		// and that the parsing process ends up here (otherwise, we stop at EOF) 
		add(new EndPsPictureExpression());
		
		//System.out.println("PsTricks grammar : \n" + this.toString());
	}


	/**
	 * reinit shared parameters belonging to the Pool
	 */
	public void reinit(){
		//System.out.println("Reinit' PstricksParser...");
		double  pstXunit,pstYunit,pstRunit;
		pstXunit = pstYunit = pstRunit = 10.0; // PsTricks's default = 1cm
		pool.put(KEY_X_UNIT, new Double(pstXunit));
		pool.put(KEY_Y_UNIT, new Double(pstYunit));
		pool.put(KEY_R_UNIT, new Double(pstRunit));
		pool.put(KEY_ATTRIBUTES, new PicAttributeSet()); // PsTricks's registers
		pool.put(KEY_NEWPSOBJECTS, new HashMap());
		pool.put(KEY_USER_COLOURS, new HashMap());
	}
}
