/*
 *  jPicEdt, a picture editor for LaTeX.
 *  Copyright (C) 1999-2007  Sylvain Reynal
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Sylvain Reynal
 *  Département de Physique
 *  Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 *  6, avenue du Ponceau
 *  95014 CERGY CEDEX
 *  FRANCE
 *  Tel : 00 +33 130 736 245
 *  Fax : 00 +33 130 736 667
 *  e-mail : reynal@ensea.fr
 *  jPicEdt web page : http://www.jpicedt.org
 */
package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;

import java.awt.*;
import java.awt.color.*;
import java.util.*;

/**
 * Parse statement defining user-colours, that is :<ul> 
 * <li> \\newgray{user-def colour name}{num} (where num is a gray scale specification b/w 0 and 1)<br> 
 * <li> \\newrgbcolor{user-def-colour-name}{num1 num2 num3} (where num are numbers b/w 0 and 1) <br>
 * <li> \\newhsbcolor{user-def-colour-name}{num1 num2 num3} (where num are numbers b/w 0 and 1) <br>
 * <li> \\newcmykcolor{user-def-colour-name}{num1 num2 num3 num4} (where num are numbers b/w 0 and 1)<br>
 * </ul> 
 * Colour names may contain letters and numbers. Spaces are NOT allowed, except as separators b/w colour
 * numbers. New colours
 * are then added to the pool's hashtable associated with key = PstricksParser.KEY_USER_COLOURS.<br>
 * @author Sylvain Reynal
 * @version $Id: UserDefinedColorExpression.java,v 1.4.2.1 2007/09/02 11:56:13 reynal Exp $ 
 */
public class UserDefinedColorExpression extends AlternateExpression implements ExpressionConstants {

	private Pool pool;
	private String newColourName;
	private float num1, num2, num3, num4;


	/**
	 * Constructor for the UserDefinedColorExpression object
	 *
	 */
	public UserDefinedColorExpression(Pool pl) {

		this.pool = pl;
		// ================ \\newgray ===============================
		SequenceExpression newGray = new SequenceExpression(true);
		newGray.add(new LiteralExpression("\\newgray{"));
		newGray.add(
			new WordExpression("}", true) {// postfix="}", swallowed
				public void action(ParserEvent e) {
					if (DEBUG)
						System.out.println(e);
					// set new gray name
					newColourName = (String)(e.getValue());
				}
			});
		newGray.add(new LiteralExpression("{"));
		newGray.add(
			new NumericalExpression(DOUBLE, POSITIVE, "}", true) {// postfix="}", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG)
						System.out.println(e);
					// add new gray name with its value to the User-defined colours hashtable
					float val = ((Number)e.getValue()).floatValue();
					if (val > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
					HashMap userColoursMap = (HashMap)pool.get(PstricksParser.KEY_USER_COLOURS); 
					userColoursMap.put(newColourName, new Color(val, val, val));// this is a gray-level !
				}
			});
		add(newGray);

		// ================ \\newrgbcolor ===============================
		SequenceExpression newRgb = new SequenceExpression(true);
		newRgb.add(new LiteralExpression("\\newrgbcolor{"));
		newRgb.add(
			new WordExpression("}", true) {// postfix="}", swallowed
				public void action(ParserEvent e) {
					if (DEBUG) System.out.println(e);
					newColourName = (String)(e.getValue());
				}
			});
		newRgb.add(new LiteralExpression("{"));
		// num1 (red)
		newRgb.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num1 = ((Number)e.getValue()).floatValue();
					if (num1 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num2 (green)
		newRgb.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num2 = ((Number)e.getValue()).floatValue();
					if (num2 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num3 (blue)
		newRgb.add(
			new NumericalExpression(DOUBLE, POSITIVE, "}", true) {// postfix="}", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num3 = ((Number)e.getValue()).floatValue();
					if (num3 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
					HashMap userColoursMap = (HashMap)pool.get(PstricksParser.KEY_USER_COLOURS); 
					userColoursMap.put(newColourName, new Color(num1, num2, num3));
				}
			});
		add(newRgb);

		// ================ \\newhsbcolor ===============================
		SequenceExpression newHsb = new SequenceExpression(true);
		newHsb.add(new LiteralExpression("\\newhsbcolor{"));
		newHsb.add(
			new WordExpression("}", true) {// postfix="}", swallowed
				public void action(ParserEvent e) {
					if (DEBUG)
						System.out.println(e);
					newColourName = (String)(e.getValue());
				}
			});
		newHsb.add(new LiteralExpression("{"));
		// num1 (hue)
		newHsb.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) {
					if (DEBUG) System.out.println(e);
					num1 = ((Number)e.getValue()).floatValue();// any positive number
				}
			});
		// num2 (saturation)
		newHsb.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num2 = ((Number)e.getValue()).floatValue();
					if (num2 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num3 (brightness)
		newHsb.add(
			new NumericalExpression(DOUBLE, POSITIVE, "}", true) {// postfix="}", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num3 = ((Number)e.getValue()).floatValue();
					if (num3 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
					System.out.println("HSB->RGB = " + Color.getHSBColor(num1, num2, num3));
					HashMap userColoursMap = (HashMap)pool.get(PstricksParser.KEY_USER_COLOURS); 
					userColoursMap.put(newColourName, Color.getHSBColor(num1, num2, num3));
				}
			});
		add(newHsb);

		// ================ \\newcmykcolor ===============================
		// [pending] BUGGY !!! On Linux platforms, raises an IllegalArgumentException : Unknown color space ... and I don't know why...
		SequenceExpression newCmyk = new SequenceExpression(true);
		newCmyk.add(new LiteralExpression("\\newcmykcolor{"));
		newCmyk.add(
			new WordExpression("}", true) {// postfix="}", swallowed
				public void action(ParserEvent e) {
					if (DEBUG) System.out.println(e);
					newColourName = (String)(e.getValue());
				}
			});
		newCmyk.add(new LiteralExpression("{"));
		// num1 (Cyan)
		newCmyk.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num1 = ((Number)e.getValue()).floatValue();
					if (num1 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num2 (Magenta)
		newCmyk.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num2 = ((Number)e.getValue()).floatValue();
					if (num2 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num3 (Yellow)
		newCmyk.add(
			new NumericalExpression(DOUBLE, POSITIVE, " ", true) {// postfix=" ", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num3 = ((Number)e.getValue()).floatValue();
					if (num3 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
				}
			});
		// num4 (Black)
		newCmyk.add(
			new NumericalExpression(DOUBLE, POSITIVE, "}", true) {// postfix="}", swallows it
				public void action(ParserEvent e) throws REParserException {
					if (DEBUG) System.out.println(e);
					num4 = ((Number)e.getValue()).floatValue();
					if (num4 > 1.0) throw new REParserException.NumberFormat(e.getContext(), this);
					float[] components = new float[4];
					components[0] = num1;
					components[1] = num2;
					components[2] = num3;
					components[3] = num4;
					HashMap userColoursMap = (HashMap)pool.get(PstricksParser.KEY_USER_COLOURS); 
					userColoursMap.put(newColourName, new Color(ColorSpace.getInstance(ColorSpace.TYPE_CMYK), components, 1.0f));
				}
			});
		add(newCmyk);
	}
	
	public String toString(){
		return "[UserDefinedColorExpression]";
	}
	
}

