/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

import jpicedt.graphic.PicPoint;

import java.util.*;

/**
 * An expression for 2D-Point parsing e.g. "(12.3, 34.5)" or "[12.1;-16]"
 * If a coordinate conversion is necessary, it must be computed in the body of the action() method
 * @since jpicedt 1.3
 * @version $Id: PicPointExpression.java,v 1.4.2.1 2007/09/02 11:56:14 reynal Exp $
 * @author Sylvain Reynal
 *
 */
public class PicPointExpression extends AbstractRegularExpression {

	private String prefix, delimiter, postfix;
	private Double x,y;
    private AbstractRegularExpression exp1, exp2, exp3, exp4;

	/**
	 * @param prefix the opening bracket e.g. "("
	 * @param delimiter the string that delimits the border between the two numbers e.g. ","
	 * @param postfix the closing bracket e.g. ")"
	 * A typical use would be : PicPointExpression("(", ",", ")")
	 */
	public PicPointExpression(String prefix, String delimiter, String postfix){

		this.prefix=prefix;
		this.delimiter=delimiter;
		this.postfix=postfix;
		exp1 = new LiteralExpression(prefix);
		exp2 = new NumericalExpression(DOUBLE, ANY_SIGN, delimiter, true){
			                         public void action(ParserEvent e){ x = (Double)e.getValue();}};
		exp3 = new NumericalExpression(DOUBLE, ANY_SIGN, postfix, true){
			   						 public void action(ParserEvent e){ y = (Double)e.getValue();}};
							 
	}

	/**
	 * @return TRUE if expr has been found 
	 * send a ParserEvent with value = parsed PicPoint
	 *
	 * @throws IncompleteSequence exception if there's nothing after the prefix...
	 */
	public boolean interpret(Context c) throws REParserException {

		// prefix:
		if (!exp1.interpret(c)) return false;

		// first number (x) :
		if (!exp2.interpret(c)) throw new REParserException.IncompleteSequence(c,this);

		// second number (y) :
		if (!exp3.interpret(c)) throw new REParserException.IncompleteSequence(c,this);

		// build PicPoint
		action(new ParserEvent(this, c, true, new PicPoint(x, y)));
		return true;
	}
	
	/**
	 * Returns the parsed PicPoint
	 */
	public PicPoint getPicPoint(){
		return new PicPoint(x,y);
	}

	/**
	 *
	 */
	public String toString(){

		return "[PicPointExpression : pre=\"" + prefix + "\" del=\"" + delimiter + "\" post=\"" + postfix + "\"]";
	}
}
