/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

import java.util.*;

/**
 * an expression that represents a pattern repeating a given number of times
 * @since jpicedt 1.3
 * @author Sylvain Reynal
 * @version $Id: RepeatExpression.java,v 1.4.2.1 2007/09/02 11:56:14 reynal Exp $
 *
 */
public class RepeatExpression extends AbstractRegularExpression {


	private AbstractRegularExpression expr;
	private int repeat;
	private int mode;

	/**
	 * @param repeat number of times this expression has to be found sequentially ; 
	 * @param mode EXACTLY, AT_LEAST or AT_MOST
	 */
	public RepeatExpression(AbstractRegularExpression expr, int repeat, int mode){

		this.expr=expr;
		this.repeat = repeat;
		this.mode=mode;
	}

	/**
	 * set the base-pattern to the given expression
	 */
	public void setPattern(AbstractRegularExpression expr){ this.expr=expr;}


	/**
	 * @return TRUE if and only if the given "expr" has been found as many times as specified by
	 *         the "mode" given in the constructor.
	 * If TRUE, calls action with : key="*", value=new Integer(number of effective repeat found)
	 */
	public boolean interpret(Context c) throws REParserException {

		int i=0;
		switch(mode){
		case AT_LEAST:
			for(i=0; i<repeat; i++){
				if (expr.interpret(c) == false) {
					action(new ParserEvent(this, c, false, new Integer(i)));
					return false;
				}
			}
			while(expr.interpret(c)){i++;} // <-- "swallow" remaining expr occurences
			action(new ParserEvent(this, c, true,  new Integer(i)));
			return true;
		case AT_MOST:
			for(i=0; i<repeat; i++){
				if (expr.interpret(c) == false) {
					action(new ParserEvent(this, c, true, new Integer(i)));
					return true; // <-- break as soon as no more occurence of expr
				}
			}
			if (expr.interpret(c) == true) {
				i++;
				action(new ParserEvent(this, c, false, new Integer(i)));
				return false; // <-- now, repeat count has been exceeded
			}
			action(new ParserEvent(this, c, true,  new Integer(i)));
			return true;
		case EXACTLY:
		default: // EXACTLY
			for(i=0; i<repeat; i++){
				if (expr.interpret(c) == false) {
					action(new ParserEvent(this, c, false, new Integer(i)));
					return false;
				}
			}
			action(new ParserEvent(this, c, true,  new Integer(i)));
			return true;
		}
	}

	/**
	 *
	 */
	public String toString(){

		String modeStr;
		switch(mode){
		case AT_LEAST :
			modeStr="AT_LEAST";
			break;
		case AT_MOST :
			modeStr="AT_MOST";
			break;
		default:
			modeStr="EXACT";
			break;
		}
		return "[Repeat:" + modeStr + " " + repeat + " " + expr + "]";
	}
}
