/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

import jpicedt.graphic.PicPoint;

import java.util.*;

/** 
 * An expression for "statement"-parsing, i.e. a name followed by an assignment sign followed by a numerical value
 *  e.g. "dash=4.0" or "thickness->8"
 * @since jpicedt 1.3
 * @author Sylvain Reynal
 * @version $Id: StatementExpression.java,v 1.4.2.1 2007/09/02 11:56:14 reynal Exp $
 *
 */
public class StatementExpression extends AbstractRegularExpression {

    private Number value;
    private String lhs, as, pf; // leftHandSide, equalSymbol, postfix (used in toString() method)
    private AbstractRegularExpression exp1, exp2, exp3;

    /**
     * @param leftHandSide aka name of a variable
     * @param equalSymbol e.g. "=" or "->" etc...
     * @param postfix marks the end of the numerical expression (can be null, see NumericalExpression comments)
     * @param type type of numerical value (DOUBLE or INTEGER)
     * @param sign constraints laid on the sign of numerical values (ANY_SIGN, POSITIVE, ...)
     *
     * A typical use would be : StatementExpression("dash", "=", null, "pt_1", DOUBLE, POSITIVE)
     */
    public StatementExpression(String leftHandSide, String equalSymbol, String postfix, int type, int sign){

		lhs = leftHandSide;
		as = equalSymbol;
		pf = postfix;

		exp1 = new LiteralExpression(leftHandSide+equalSymbol);
		exp2 = new WhiteSpacesOrEOL();
		exp3 = new NumericalExpression(type, sign, postfix, true){
			public void action(ParserEvent e){
				if (DEBUG) System.out.println(e);
		    	value = (Number)e.getValue();}};
    	}

	/**
     * @return TRUE if and only if every expression contained in this expression has been found, in the proper order.
     * @throws REParserException.IncompleteSequence if any expression but the first one wasn't found
     *
     * Whether TRUE or FALSE, calls action with key="&" and value=value of number found (of type Number)
     */
    public boolean interpret(Context c) throws REParserException {

		if (!exp1.interpret(c)) return false;
		exp2.interpret(c); // always return true (repeat_at_least_0)
		if (!exp3.interpret(c)) throw new REParserException.IncompleteSequence(c,this);
		action(new ParserEvent(this, c, true, value));
		return true;
    }

	/**
	 * Return the value of the RHS, wrapped in a Number (either Integer or Double)
	 */
	public Number getValue(){
		return value;
	}
	
    /**
     * @return a String representation of this expression
     */
    public String toString(){

		return "[StatementExpression : " + lhs + ", " + as + ", " + pf + "]";
    }
}
