/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

/**
 * A RegExp that parses a word, that is, a string :
 * <ul> 
 * <li> either composed of letters only, or letters and digits only (see java.lang.Character.isLetter() for details),</li> 
 * <li> or terminated by the specified end-delimiter (in which case it may contain chars not restricted to letters)</li>
 * </ul>
 * @since jpicedt 1.3
 * @version $Id: WordExpression.java,v 1.4.2.1 2007/09/02 11:56:15 reynal Exp $
 * @author Sylvain Reynal
 */
public class WordExpression extends AbstractRegularExpression {

	private String postfix;
	private boolean swallowPostFix, allowDigit, noLineFeed;
	private String value;

	/**
	 * @param postfix a String that signals the end of the word (if NULL, swallows as many letters as possible)
	 * @param swallowPostFix if TRUE, move cursor JUST BEHIND the postfix; else move it TO THE POSTFIX POSITION
	 * @param allowsDigit if TRUE, and postfix=NULL, swallows as many "letter-or-digit" as possible (i.e. alphanumeric expression)
	 * @param noLineFeed if true, expression must fit on a single line. Otherwise, it may span several lines.
	 */
	public WordExpression(String postfix, boolean swallowPostFix, boolean allowDigit, boolean noLineFeed){

		this.postfix=postfix;
		this.swallowPostFix = swallowPostFix;
		this.allowDigit = allowDigit;
		this.noLineFeed = noLineFeed;
	}

	/**
	 * Allow line-feed's, i.e. expression may span more than one line. 
	 * @param postfix a String that signals the end of the word (if NULL, swallows as many letters as possible)
	 * @param swallowPostFix if TRUE, move cursor JUST BEHIND the postfix; else move it TO THE POSTFIX POSITION
	 * @param allowsDigit if TRUE, and postfix=NULL, swallows as many "letter-or-digit" as possible (i.e. alphanumeric expression)
	 */
	public WordExpression(String postfix, boolean swallowPostFix, boolean allowDigit){
		this(postfix, swallowPostFix, allowDigit, false);
	}

	/**
	 * Constructor with allowsDigit=false (i.e. only alphabetic character are swallowed if postfix = null),
	 * and which allows linefeeds.
	 * @param postfix a String that signals the end of the word (if NULL, swallows as many letters as possible)
	 * @param swallowPostFix if TRUE, move cursor JUST BEHIND the postfix; else move it TO THE POSTFIX POSITION
	 */
	public WordExpression(String postfix, boolean swallowPostFix){

		this(postfix, swallowPostFix, false, false);
	}

	/**
	 * Parse an expression containing letters only, and stops when a non-letter char is found. Allows
	 * line-feed's, which can be removed later on using Context.removeLineFeeds (static method).
	 */
	public WordExpression(){
		this(null, false, false, false);
	}

	/**
	 * If parsing was successfull, sends a ParserEvent with value=string found (this may be an empty string)
	 * <p>
	 * You may then use <code>Context.removeLineFeeds(String)</code> to remove CR from "value".
	 * @return TRUE if expr has been found
	 */
	public boolean interpret(Context context) throws REParserException {

		// if there's a postfix specified, let's fetch its position in the current substring :
		if (postfix != null){
			int postFixIndex;
			if (noLineFeed) postFixIndex = context.indexOfBeforeEOL(postfix);
			else postFixIndex = context.indexOf(postfix); // return -1 if not found
			if (postFixIndex < 0) return false; // no matching postfix was found
			value = context.readTo(postFixIndex);
			if (swallowPostFix) context.read(); // read postfix but don't use it
			// send a ParserEvent with value=string found
			action(new ParserEvent(this, context, true, value));
			return true;
		}
		
		// no postfix specified => swallows as many letters as possible (swallowPostFix N/A here) 
		else {
			StringBuffer buf = new StringBuffer(); // temp. storage for chars
			while(true){
				Character C=null;
				try {C = context.read();}
				catch (REParserException.EOF eofEx){ // EOF
					value = buf.toString();
					action(new ParserEvent(this, context, true, value));
					throw new REParserException.EOF(); // rethrows after action !!! (otherwise, the receiver never gets the event)
				}
				if (C == null) break; // end-of-block
				char c = C.charValue();
				if (c=='\n') {
					if (noLineFeed) {context.pushBack(); break;}
					else buf.append(c); 
				}
				else if (allowDigit){
					if (Character.isLetterOrDigit(c)) buf.append(c);
					else {context.pushBack(); break;}
				}
				else if (Character.isLetter(c)) buf.append(c);
				else {context.pushBack(); break;}
			}
			// send a ParserEvent with value=string found
			value = buf.toString();
			action(new ParserEvent(this, context, true, value));
			return true;
		}
	}

	/**
	 * Return the parsed content of this expression after a successfull parsing.
	 */
	public String getValue(){
		return value;
	}
	
	public String toString(){
		return "[WordExpression:post=" + postfix + " swallow=" + swallowPostFix + "]";
	}
}
