/*
 AbstractCurveFormatter.java - August 29, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package jpicedt.format.output.eepic;

import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.PicPoint;

import java.awt.*;

/**
 * Format an AbstractCurve, or any subclass of it, to the eepic format.
 * Basically, a single \\path macro is used if the curve is a polygon, so that filling works properly. Otherwise we make use
 * of \\qbezier, \\path and/or \\dashline macros.
 * @author Sylvain Reynal
 * @since jpicedt 1.4
 * @version $Id: AbstractCurveFormatter.java,v 1.4.2.1 2007/09/02 11:55:12 reynal Exp $
 * <p>
 *
 */
public class AbstractCurveFormatter implements Formatter, PicObjectConstants, EepicConstants {

	/** the Element this formater acts upon */
	protected AbstractCurve curve;
	protected EepicFormatter factory;

	/** */
	public AbstractCurveFormatter(AbstractCurve curve, EepicFormatter factory){
		this.curve = curve;
		this.factory=factory;
	}

	/**
	 * Returns a string representing this AbstractCurve using \\line and \\qbezier LaTeX macros only.
	 * Dashed style is allowed if all segments are straight, since \\qbezier does not support this style.
	 */
	public String format(){
		StringBuffer buf = new StringBuffer(100); // 100 char as initial capacity seems a good guess
		if (curve.isPolygon()) return formatPolygon(buf);
		else return formatCurve(buf);
	}
	
	/** 
	 * formats curves containing at least one non-straight segment 
	 */
	protected String formatCurve(StringBuffer buf){

		
		PicAttributeSet set = curve.getAttributeSet();

		factory.appendThicknessString(buf,curve);
		final double dash=0; // qbezier are not dashable
		
		// for each segment :
		// - straight : append \\path
		// - curved : append \\qbezier command, after converting cubic to two quads (see PEToolKit)
		// If curve is open, first and last segment deserve special processing as for arrows
		for (int segIdx=0; segIdx < curve.getNumberOfSegments(); segIdx++){
			PicPoint p1 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SUBDIVISION_POINT),null);
			PicPoint p2 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx+1, curve.SUBDIVISION_POINT),null);
			if (curve.isStraight(segIdx)){
				Arrow arrowL=Arrow.NONE;
				Arrow arrowR=Arrow.NONE;
				if (!curve.isClosed()){
					if (segIdx==0) arrowL = StyleConstants.getLeftArrow(set);
					if (segIdx==curve.getNumberOfSegments()-1) arrowR = StyleConstants.getRightArrow(set);
				}
				if (segIdx > 0 && !curve.isStraight(segIdx-1)) {
					buf.append(CR_LF);
					buf.append("\\path"); 
					buf.append(p1);
				}
				buf.append(p2);
			}
			else {
				PicPoint pCtrl1 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.FIRST_SEGMENT_CONTROL_POINT),null);
				PicPoint pCtrl2 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SECOND_SEGMENT_CONTROL_POINT),null);
				PicPoint[] quadsPts = PEToolKit.convertCubicBezierToQuad(p1,pCtrl1,pCtrl2,p2); // five points = two joined quads
				buf.append(CR_LF);
				buf.append("\\qbezier"); // first quad
				buf.append(quadsPts[0]);buf.append(quadsPts[1]);buf.append(quadsPts[2]);
				buf.append(CR_LF);
				buf.append("\\qbezier"); // second quad
				buf.append(quadsPts[2]);buf.append(quadsPts[3]);buf.append(quadsPts[4]);
				buf.append(CR_LF);
				if (!curve.isClosed()){
					if (segIdx==0 && StyleConstants.getLeftArrow(set) != Arrow.NONE){
						PicVector dir = new PicVector(pCtrl1,p1).normalize();
						buf.append(factory.arrowToLatexString(p1, dir));
					}
					if (segIdx==curve.getNumberOfSegments()-1 && StyleConstants.getRightArrow(set) != Arrow.NONE){
						PicVector dir = new PicVector(pCtrl2,p2).normalize();
						buf.append(factory.arrowToLatexString(p2, dir));
					} 
				}
			}
		}
		buf.append(CR_LF);
		return buf.toString();
	}

	/**
	  * Formats curves containing straight segments only
	  */
	protected String formatPolygon(StringBuffer buf){

		factory.appendThicknessString(buf,curve);
		
		PicAttributeSet set = curve.getAttributeSet();

		// there are only straight segments, hence we can afford dashed lines :
		double dash=(StyleConstants.getDashOpaque(set)+StyleConstants.getDashTransparent(set))/2.0;
		if (StyleConstants.getLineStyle(set)!=DASHED) dash=0;// [pending] add support for black/white dash + dotted

		// eepic commands
		int lastPt = curve.getLastPointIndex();
		if (StyleConstants.getFillStyle(set)==SOLID){
			Color fillColor = StyleConstants.getFillColor(set);
			if (fillColor.equals(Color.WHITE)) buf.append("\\whiten");
			else if (fillColor.equals(Color.BLACK)) buf.append("\\blacken");
			else buf.append("\\shade"); // SHADE
			buf.append("\\path"); // create a closed path so that it can be filled correctly
			for (int segIdx=0; segIdx<=curve.getNumberOfSegments(); segIdx++){
				PicPoint pt = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SUBDIVISION_POINT),null);
				buf.append(pt);
			}
			//buf.append(curve.getSpecificationPoint(0,null)); // close path
			buf.append(CR_LF);
		}
		else {	 // not filled
			if (dash <= 0){
				buf.append("\\path");
				for (int segIdx=0; segIdx<=curve.getNumberOfSegments(); segIdx++){
					PicPoint pt = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SUBDIVISION_POINT),null);
					buf.append(pt);
				}
				//if (curve.isClosed()) buf.append(curve.getSpecificationPoint(0,null)); // close path
				buf.append(CR_LF);
			}
			else {
				String dashStr = PEToolKit.doubleToString(dash);
				for (int segIdx=0; segIdx<curve.getNumberOfSegments(); segIdx++){
					buf.append("\\dashline{");
					buf.append(dashStr);
					buf.append("}");
					PicPoint p1 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SUBDIVISION_POINT),null);
					PicPoint p2 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx+1, curve.SUBDIVISION_POINT),null);
					buf.append(p1);
					buf.append(p2);
					buf.append(CR_LF);
				}
			}
		}
		// add arrows if not closed
		if (!curve.isClosed()){
			// first arrow
			if (StyleConstants.getLeftArrow(set)!=Arrow.NONE){
				PicPoint p0 = curve.getSpecificationPoint(curve.segmentToPointIndex(0, curve.SUBDIVISION_POINT),null);
				PicPoint p1 = curve.getSpecificationPoint(curve.segmentToPointIndex(1, curve.SUBDIVISION_POINT),null);
				PicVector dir = new PicVector(p1,p0).normalize(); 
				buf.append(factory.arrowToLatexString(p0, dir));
				buf.append(CR_LF);
			}
			// second arrow
			if (StyleConstants.getRightArrow(set)!=Arrow.NONE){
				int lastSegIdx = curve.getNumberOfSegments()-1;
				PicPoint p1 = curve.getSpecificationPoint(curve.segmentToPointIndex(lastSegIdx, curve.SUBDIVISION_POINT),null);
				PicPoint p0 = curve.getSpecificationPoint(curve.segmentToPointIndex(lastSegIdx+1, curve.SUBDIVISION_POINT),null);
				PicVector dir = new PicVector(p1,p0).normalize(); 
				buf.append(factory.arrowToLatexString(p0, dir));
				buf.append(CR_LF);
			}
		}
		buf.append(CR_LF);
		return buf.toString();
	}
	
}

 
