/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://trashx.ensea.fr/jpicedt/
*/

package jpicedt.format.output.eepic;

import jpicedt.format.output.util.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.format.output.latex.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.Properties;

/**
 * Formatter for the eepic package.
 */
public class EepicFormatter extends LatexFormatter  implements EepicConstants  {

	private static double thinLinesMaxWidth = DEFAULT_THIN_LINES_MAX_WIDTH;
	private static double thickLinesMaxWidth = 	DEFAULT_THICK_LINES_MAX_WIDTH;
	
	/**
	 * Constructor using default properties values
	 */
	public EepicFormatter(){
		super();
		fileWrapperProlog = EepicConstants.DEFAULT_FILE_WRAPPER_PROLOG;
		fileWrapperEpilog = EepicConstants.DEFAULT_FILE_WRAPPER_EPILOG;
	}

	/**
	 * Configure static fields using the given Properties object
	 * @param properties used to read shared parameters
	 *        If null, default values are used.
	 */
	public static void configure(Properties preferences){
	    thinLinesMaxWidth = Double.parseDouble(preferences.getProperty(EepicConstants.KEY_THIN_LINES_MAXWIDTH,PEToolKit.doubleToString(EepicConstants.DEFAULT_THIN_LINES_MAX_WIDTH)));
	    thickLinesMaxWidth = Double.parseDouble(preferences.getProperty(EepicConstants.KEY_THICK_LINES_MAXWIDTH,PEToolKit.doubleToString(EepicConstants.DEFAULT_THICK_LINES_MAX_WIDTH)));
		fileWrapperProlog = preferences.getProperty(EepicConstants.KEY_FILE_WRAPPER_PROLOG,EepicConstants.DEFAULT_FILE_WRAPPER_PROLOG);
		fileWrapperEpilog = preferences.getProperty(EepicConstants.KEY_FILE_WRAPPER_EPILOG,EepicConstants.DEFAULT_FILE_WRAPPER_EPILOG);
	}	

	/** 
	 * @return a Formatter able to format the given Element according to the format of this factory
	 */
	public Formatter createFormatter(Element e){
		if (e instanceof PicEllipse) return new PicEllipseFormatter((PicEllipse)e,this);
		// [SR:en_cours] if (e instanceof PicCircleFrom3Points) return new PicEllipseFormatter(((PicCircleFrom3Points)e).toEllipse(),this);//[pending: to be improved: add comments for the 3 control points]
		if (e instanceof PicParallelogram) return new PicParallelogramFormatter((PicParallelogram)e,this);
		if (e instanceof AbstractCurve) return new AbstractCurveFormatter((AbstractCurve)e,this);
		return super.createFormatter(e);
	}

	/** 
	 * @return a Formatter able to format the given Drawing according to the format of this factory ;
	 *         this may reliy on calls to <code>createFormatter(Element e)</code> on the elements
	 *         of the drawing, plus creating auxiliary
	 * @param outputConstraint constraint used by the factory to create a specific Formatter on-the-fly
	 */
	public Formatter createFormatter(Drawing d, Object outputConstraint){
		return new DrawingFormatter(d, outputConstraint);
	}


	////////////////////////////////////////////////////////////////////////////////////
	//// TOOLKIT
	////////////////////////////////////////////////////////////////////////////////////

	/**
	 * <p>Create a string representation of the thickness command for the given PicObjet in the eepic format,
	 *  and append it to the given StringBuffer.</p>
	 * <p>Such a command should preceed every object command. This string is CR-terminated.</p>
	 * @since jpicedt 1.3.2
	 */
	public void appendThicknessString(StringBuffer buf, Element obj){

		double thickness = ((Double)obj.getAttribute(LINE_WIDTH)).doubleValue();
		if (thickness < thinLinesMaxWidth) buf.append("\\thinlines"+CR_LF); 
		else if (thickness < thickLinesMaxWidth) buf.append("\\thicklines"+CR_LF);
		else buf.append("\\Thicklines"+CR_LF);
	}

	/**
	 * Append a \\dashline with the given points and dash value to the given buffer
	 */
	protected void appendDashLine(StringBuffer buf, PicPoint p1, PicPoint p2, double dash){
		buf.append("\\dashline{");
		buf.append(PEToolKit.doubleToString(dash));
		buf.append("}");
		buf.append(p1);
		buf.append(p2);
		buf.append(CR_LF);
	}
		

	////////////////////////////////////////////////////////////////////////////////////
	//// DRAWING
	////////////////////////////////////////////////////////////////////////////////////
	/**
	 *
	 */
	public class DrawingFormatter extends LatexFormatter.DrawingFormatter {

		/**
		 * init this formater for the given drawing
		 */
		public DrawingFormatter(Drawing drawing, Object outputConstraint){
			super(drawing, outputConstraint);
		}

		/**
		 * @return the content-type specific header
		 */
		protected String createHeader(){
			return "%Eepic content-type (epic.sty and eepic.sty packages needed)"+CR_LF+
			"%Add \\usepackage{epic,eepic} in the preambule of your LaTeX file"+CR_LF+
			"%You can rescale the whole picture (to 80% for instance) by using the command \\def"+
				   LatexConstants.RESCALING_TEX_FUNCTION+"{0.8}"+CR_LF;

		}
	}
}

