/*
 EepicViewFactory.java - February 11, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://trashx.ensea.fr/jpicedt/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.output.eepic;

import jpicedt.graphic.model.*; // all elements + drawing
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.view.*;

import java.awt.*;
import java.awt.geom.*;

/**
 * A factory to create a View for a given Element when the content type for the model is Eepic.
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: EepicViewFactory.java,v 1.12.2.1 2007/09/02 11:55:13 reynal Exp $
 * <p>
 *
 */
public class EepicViewFactory extends DefaultViewFactory {

    protected static final Color EEPIC_STROKE = Color.BLACK;
    protected static final Color EEPIC_WHITEN = Color.WHITE;
    protected static final Color EEPIC_SHADE = Color.GRAY;
    protected static final Color EEPIC_BLACKEN = Color.BLACK;
    protected static final Color EEPIC_COLOR = Color.GRAY; // eepic doesn't support colour, hence it's a good idea to paint in grayscale

	/**
	 * @return a View for the given Element
	 * This returns a specific Epic/Eepic view for PicEllipse, AbstractCurve and PicText.
	 * Calls superclass otherwise.
	 */
	public View createView(Element element){
		// curves : polygons only can be filled
		if (element instanceof AbstractCurve) return new EepicViewFactory.AbstractCurveView((AbstractCurve)element);
		// ellipses : only non-rotated ellipses can be filled
		else if (element instanceof PicEllipse) return new EepicViewFactory.EllipseView((PicEllipse)element);
		// text : rectangular boxes only
		else if (element instanceof PicText) return new EepicViewFactory.TextView((PicText)element);
		// parallelo can be filled (by using \\path macros) => super class
		else return super.createView(element);
	}

	/**
	 * @return a Stroke built from the given attributes ; null if LINE_STYLE=NONE. 
	 *         Only SOLID and DASHED (with DASH_OPAQUE only) are supported by LaTeX/Eepic
	 */
	protected Stroke createStroke(PicAttributeSet set){
		// [pending] work out the pb with ThickLine, Thinline, etc...
		if(set.getAttribute(LINE_STYLE)==DASHED){
			// make sure dashes > 0
			float dash; // = mean value of (opaque,transparent)
			dash = (float)StyleConstants.getDashOpaque(set);
			dash += (float)StyleConstants.getDashTransparent(set);
			dash /= 2.0f;
			if (dash>0){
				float[] dashArray = {dash,dash};
				return new BasicStroke((float)StyleConstants.getLineWidth(set),
				                      BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, dashArray, 0);
			}
			// otherwise, paint as SOLID :
		}
		return new BasicStroke((float)StyleConstants.getLineWidth(set));
	}

	/**
	 * @return a Paint from the given attributes, suited for painting outlines.
	 * Eepic only support painting in black.
	 */
	protected Paint createPaintForOutline(PicAttributeSet set){
		return EEPIC_STROKE;
	}

	/**
	 * @return a Paint from the given attributes, suited for painting interior.
	 * Eepic support filling with black, shade (i.e. gray) and white ; return null
	 * if FILL_STYLE is NONE.
	 */
	protected Paint createPaintForInterior(PicAttributeSet set){
		if (set.getAttribute(FILL_STYLE)!=NONE){
			if(set.getAttribute(FILL_COLOR)==EEPIC_WHITEN) return EEPIC_WHITEN;
			else if (set.getAttribute(FILL_COLOR)==EEPIC_BLACKEN) return EEPIC_BLACKEN;
			else if (set.getAttribute(FILL_COLOR)==EEPIC_SHADE) return EEPIC_SHADE;
			else return EEPIC_COLOR;
		}
		else return null;
	}

	/**
	 * paint hatches (vlines,...) depending on the FILL_STYLE attributes of the given set.
	 * Overriden to do nothing, since (so far) hatches aren't supported by eepic or LaTeX.
	 */
	protected void paintHatches(Graphics2D g, PicAttributeSet set, Shape shape){
	}
	
	/** paint shadow ; overriden to do nothing, since eepic and LaTeX don't support shadowing */
	protected void paintShadow(Graphics2D g, PicAttributeSet set, Shape shape){
	}
	
	/**
	 * paint overstrike ; overriden to do nothing 
	 */
	protected void paintOverStrike(Graphics2D g, PicAttributeSet set, Shape shape){
	}
	
	/**
	 * @return an Arrow from the given attribute set ; Eepic supports only Arrow.ARROW_HEAD
	 *         null if Arrow.NONE
	 */
	protected Arrow createArrow(PicAttributeSet set, PicAttributeName direction){
		Arrow a = (Arrow)set.getAttribute(direction);
		if (a != Arrow.NONE) {  // if there's an arrow, we draw it as a ARROW_HEAD
			a = Arrow.ARROW_HEAD; // width default values => don't synchronize !
			return a;
		}
		else return null; // consistent with superclass contract
	}
		
	//////////////////////////////////////////////
	/// ABSTRACT CURVE VIEW
	/////////////////////////////////////////////
	/**
	 * a View for rendering any concrete implementation of jpicedt.graphic.model.AbstractCurve
	 * eepic allows filling polygon only (using \\path macro).
	 */
	public class AbstractCurveView extends DefaultViewFactory.AbstractCurveView {
	    
		/**
		 * construct a new View for the given curve
		 */
		public AbstractCurveView(AbstractCurve curve){
			super(curve);
		}

		/**
		 * Inherits from superclass, except for filling behaviour
		 */
		protected void syncAttributes(){
			super.syncAttributes();
			AbstractCurve curve = (AbstractCurve)getElement();
			if (!curve.isPolygon()) interiorPaint = null;
		}
		
		/**
		 * Computes the Shape (centered at 0,0) used for drawing polygon's dots, built from the given attributes ;
		 * this implementation sets it to null, as dotting is not supported.
		 */
		 protected void syncDotShape(){
			 dotShape=null;
		 }
	}


	/////////////////////////////////////////////////////////////////////////////
	
	/**
	 * view for ellipses specific to the Epic/Eepic content type.
	 */
	public class EllipseView extends DefaultViewFactory.EllipseView {
		
		/**
		 * Construct a View for the given PicEllipse.<p>
		 * Following rules apply for the Epic/Eepic content type :
		 * <ul>
		 * <li> Arcs (circular or elliptic) and ellipse can be rendered -> inherited from DefaultViewFactory.
		 *      [pending] bug : only OPEN arc ! No PIE !
		 * <li> Ellipse and circulars arcs can be filled only if surrounding parallelogram is a rectangle parallel to X and Y axes
		 * <li> Elliptic arcs can't be filled anyway.
		 * <li> Supported fill-colours are BLACK, GRAY and WHITE.
		 * <li> Only solid stroke is supported [pending] implementation in formatter
		 * </ul>
		 */
		public EllipseView(PicEllipse ellipse){
			super(ellipse);
		}

		/**
		 * Give notification from the model that a change occured for an element this view is responsible
		 * for rendering.<p>
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
			//if geometry changed from ellipse to circle or vice-versa, we need to *force*
			//update of interior paint attribute ; otherwise, if attribute changed, 
			//no need to worry, for this is done automatically by superclass :
			if (eventType!=DrawingEvent.EventType.ATTRIBUTE_CHANGE) syncAttributes(); 
			super.changedUpdate(eventType); // super calls syncAttribute if ATTRIBUTE_CHANGE, it's ok.
		}
		
		/**
		 * Inherits from superclass, except for elliptic arcs, which are never filled, as well
		 * as rotated ellipses. Conversely, circles, circular arcs and non-rotated ellipses are always filled.
		 */
		protected void syncAttributes(){
			super.syncAttributes(); // that's ok for ellipses and circular arcs
			outlineStroke = new BasicStroke((float)lineWidth); // dont' dash or dot
			PicEllipse ellipse = (PicEllipse)element;
			// filled : isCircular() || (isPlain() && non-rotated)
			if (!ellipse.isCircular() && (ellipse.getRotationAngle()!=0 || !ellipse.isPlain())) { // elliptic arcs or rotated ellipses
				interiorPaint = null; // forbides filling
			}
			//never paint arrows :
			//arrowStart = arrowEnd = null; // now rendered by formatter
		}
	}

	//////////////////////////////////////////////////////////////////////////
	//// TEXTEDITABLE VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering Text's : only rectangular frame boxes (through \\makebox and rel.) are supported,
	 * + no filling + only black stroke.
	 * Alignment :
	 * - no frame : left, right or center ; top, bottom or center
	 * - frame : left + baseline. (i.e. \\framebox{text} doesn't allow alignment specification if the
	 *           box size is to be computed by LaTeX itself)
	 */
	public class TextView extends DefaultViewFactory.TextView {

		// inherited : strx, stry, textLayout, shape (=frame box), tl2ModelTr (affine transform updated from syncShape)
		/**
		 * construct a new View for the given PicRectangle
		 */
		public TextView(PicText te){
			super(te);
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init> (done!)");
		}

		/**
		 * Returns the text rotation in radians : we return 0, since eepic doesn't support rotating text
		 * This is used by syncShape to compute the current tl2ModelTr AffineTransform
		 */
		protected double getRotation(){
			return 0;
		}
		
		/** update frame shape : all frames are drawn as a rectangle */
		protected void syncFrame(){
			// frame : [SR:pending] oval and circle boxes ! (todo)
			PicText te = (PicText)element;
			if (te.getFrameType()!=TEXT_BOX_NO_FRAME){
				Rectangle2D tb = textLayout.getBounds();
				shape = tl2ModelTr.createTransformedShape(tb); // at (0,0)
			}
			else shape=null;
		}
			
		/** update strx */
		protected void syncStringLocationX(){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("syncStringLocationX");
			PicText te = (PicText)element;
			double textWidth = textLayout.getBounds().getWidth();
			if (te.getFrameType()!=TEXT_BOX_NO_FRAME){ // framed -> left alignment
				strx = 0;
			}
			else {
				if (te.getHorAlign() == TEXT_HALIGN_LEFT) strx = 0;
				else if (te.getHorAlign() == TEXT_HALIGN_RIGHT) strx =  - textWidth;
				else {//if (te.getHorAlign() == TEXT_HALIGN_CENTER) 
					strx =  - textWidth/2;
				}
			}
		}
			
		/** update stry */
		protected void syncStringLocationY(){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("syncStringLocationY");
			PicText te = (PicText)element;
			//double ascent = textLayout.getAscent();
			// e.g. return : y = -7 ; h = 8 (i.e. upside-down)
			double ascent = -textLayout.getBounds().getMinY();
			//double descent = textLayout.getDescent();
			if (te.getFrameType()!=TEXT_BOX_NO_FRAME){ // framed -> baseline
				stry = 0;
			}
			else {
				if (te.getVertAlign()== TEXT_VALIGN_TOP) {
					stry = textLayout.getBounds().getMinY();
				}
				else if (te.getVertAlign()== TEXT_VALIGN_BOTTOM || te.getVertAlign()== TEXT_VALIGN_BASELINE) {
					stry = 0;
				}
				else { // if (te.getVertAlign()== TEXT_VALIGN_CENTER) {
					Rectangle2D tlb = textLayout.getBounds();
					double d = tlb.getHeight()*0.5 - tlb.getMaxY(); // e.g. 8 * 0.5 - 1 = 3 
					stry =  - d;
				}
			}
		}
	}
}

