/*
 LaTeXViewFactory.java - February 9, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.output.latex;

import jpicedt.format.output.eepic.EepicViewFactory;
import jpicedt.graphic.model.*; // all elements + drawing
import jpicedt.graphic.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.view.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

/**
 * Creates a View for a given Element when the underlying content type is the LaTeX-picture environment.
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: LatexViewFactory.java,v 1.10.2.1 2007/09/02 11:55:13 reynal Exp $
 * <p>
 *
 */
public class LatexViewFactory extends EepicViewFactory implements LatexConstants {

	/** color for outlines */	
    protected static final Color LATEX_STROKE = Color.black;
	/** color for interior */
    protected static final Color LATEX_FILL = Color.black;
	
	static double maxLatexDiskDiameter = DEFAULT_MAX_DISK_DIAMETER; 
	

	/**
	 * Configure static fields using the given Properties object
	 * @param properties used to read shared parameters
	 *        If null, default values are used.
	 */
	public static void configure(Properties preferences){
	    maxLatexDiskDiameter = Double.parseDouble(preferences.getProperty(KEY_MAX_DISK_DIAMETER,PEToolKit.doubleToString(DEFAULT_MAX_DISK_DIAMETER)));
	}
	
	/**
	 * @return a View for the given Element
	 */
	public View createView(Element element){
		// curve : not filled at all, whether it is a polygon or not...
		if (element instanceof AbstractCurve) return new LatexViewFactory.AbstractCurveView((AbstractCurve)element);
		// ellipses : only small circles can be filled
		else if (element instanceof PicEllipse) return new LatexViewFactory.EllipseView((PicEllipse)element);
		// parallelo : only rectangle parallel to X and Y axes can be filled
		else if (element instanceof PicParallelogram) return new LatexViewFactory.ParallelogramView((PicParallelogram)element);
		return super.createView(element); // delegates to EepicViewFactory
	}

	/**
	 * Returns LATEX_FILL, ie black filling only is allowed.
	 * @return a Paint from the given attributes, suited for painting "interior"
	 */
	protected Paint createPaintForInterior(PicAttributeSet set){
		if (set.getAttribute(FILL_STYLE)!=NONE) return LATEX_FILL;
		else return null;
		
	}

	// stroke : same as eepic (no dot)
	
	// hatches : same as eepic (no hatch)
	
	
	
	//////////////////////////////////////////////////////////////////////////
	//// PARALLELOGRAM VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering parallelograms.
	 * Can be filled only if this is rectangle parallel to the horizontal axis
	 */
	public class ParallelogramView extends DefaultViewFactory.ParallelogramView {

		/**
		 * construct a new View for the given PicParallelogram
		 */
		public ParallelogramView(PicParallelogram para){
			super(para);
		}

		/**
		 * Inherits from superclass, except for filling behaviour
		 */
		protected void syncAttributes(){
			super.syncAttributes();
			PicParallelogram para = (PicParallelogram)getElement();
			// test whether it's a rectangle and then if sides are parallel to X and Y axes
			if (!para.isRectangle()) interiorPaint = null;
			else if (para.getPointY(para.P_BL)!=para.getPointY(para.P_BR)) interiorPaint = null;
		}

	} // class

		
	/////////////////////////////////////////////////////////////////////////////
	
	/**
	 * view for ellipse specific to the <i>LaTeX picture environment</i>.
	 * Only small circles can be filled.
	 * [pending] bug : only OPEN arcs !
	 */
	public class EllipseView extends DefaultViewFactory.EllipseView {
		
		public EllipseView(PicEllipse ell){
			super(ell);
		}
		
		/**
		 * Give notification from the model that a change occured for an element this view is responsible
		 * for rendering.<p>
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
			//if geometry changed from ellipse to circle or vice-versa, and if radius becomes less or
			//equal to maxLatexDiskDiameter,  we need to *force*
			//update of interior paint attribute ; otherwise, if attribute changed, 
			//no need to worry, for this is done automatically by superclass :
			if (eventType!=DrawingEvent.EventType.ATTRIBUTE_CHANGE) syncAttributes(); 
			super.changedUpdate(eventType); // super calls syncAttribute if ATTRIBUTE_CHANGE, it's ok.
		}
		
		/**
		 * Inherits from superclass, except for elliptic arcs, which are never filled.
		 */
		protected void syncAttributes(){
			super.syncAttributes(); // init cached values from default EllipseView
			outlineStroke = new BasicStroke((float)lineWidth); // dont' dash or dot
			//small disks :
			//we call getAngleExtent() instead of using variable "angleExtent" because changedUpdate()
			//(which updates 'angleExtent') is called after syncAttributes !
			PicEllipse ellipse = (PicEllipse)element;
			if (ellipse.isPlain() && ellipse.isCircular() && ellipse.getGreatAxisLength() <= maxLatexDiskDiameter) return;
			else interiorPaint = null; // forbids filling anything else
			//never paint arrows :
			// arrowStart = arrowEnd = null; // this is handled now by formatter as of 1.4
		}
	}

	
	//////////////////////////////////////////////
	/// ABSTRACT CURVE VIEW
	/////////////////////////////////////////////
	/**
	 * a View for rendering any concrete implementation of jpicedt.graphic.model.AbstractCurve
	 * LaTeX-picture environment does not allow filling such shapes.
	 */
	public class AbstractCurveView extends DefaultViewFactory.AbstractCurveView {
	    
		/**
		 * construct a new View for the given curve
		 */
		public AbstractCurveView(AbstractCurve curve){
			super(curve);
		}

		/**
		 * Inherits from superclass, except for filling behaviour
		 */
		protected void syncAttributes(){
			super.syncAttributes();
			interiorPaint = null;
		}
		/**
		 * Computes the Shape (centered at 0,0) used for drawing polygon's dots, built from the given attributes ;
		 * this implementation sets it to null, as dotting is not supported.
		 */
		 protected void syncDotShape(){
			 dotShape=null;
		 }
		
	}
}

