/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org/
*/

package jpicedt.format.output.pstricks;

import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;

import java.awt.*;

/**
 * A formatter for PicText objects in the PsTricks format :<br>
 * \\rput[refpoint]{rotation}(x,y){stuff} <br>
 * where : <br>
 * <ul>
 * <li>stuff can be any of the following :<br>
 * \\psframebox[param]{text}}<br>
 * \\pscirclebox[param]{text}}<br>
 * \\psovalbox[param]{text}}<br>
 * <li>refpoint = B|b|t for vertical and l|r for horizontal (B = baseline)
 * <li>rotation = any angle in degree or U,L,D,R,N,W,S or E.
 * </ul>
 * <p>
 * The following are built from "frame", "oval" or "circle", but are allegedly more efficient :<br>
 *	\\psdblframebox[param]{stuff} % same as psframebox with doubleline=true<br>
 *	\\psshadowbox[param]{stuff} % same as psframebox with shadow=true<br>
 *	\\cput[shadow=true](x,y){a cput with a shadow} % same as pscirclebox, yet more efficient<br>
 * We don't use them so far.
 * @author Sylvain Reynal
 * @version $Id: PicTextFormatter.java,v 1.6.2.1 2007/09/02 11:55:14 reynal Exp $ 
 */
public class PicTextFormatter implements Formatter, PicObjectConstants, PstricksConstants {

	/** the Element this formatter acts upon */
	private PicText obj;
	private PstricksFormatter factory;

	/**
	 *
	 */
	public PicTextFormatter(PicText obj, PstricksFormatter factory){
		this.obj = obj;
		this.factory=factory;
	}
	
	/**
	 * @return a String representing this Element in the PsTricks 
	 */
	public String format(){
		
		StringBuffer buf = new StringBuffer(100);

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(obj);
		if (paramStr.isDefinedColourString()) buf.append(paramStr.getUserDefinedColourBuffer());

		if (obj.isFramed()){
		// line-thickness correction
  		       //we need to \rput this further to take care of the linewidth because pstricks does align the frame on its outer boundary,
		       // and not on the middle of the thickness of the frame
		// solution: without rotation, and aligned on bottom-left point, that would be \rput(-0.5*linethickness,-0.5*linethickness){ }
		// if alignment is on the right or top, signs should be adapted, and reset to 0 if aligned on center or baseline. 
		// in case of rotation, we also have to rotate the vector (-0.5*linethickness,-0.5*linethickness).
		// Note that the following solution does not work:  \rput[l]{angle}(x,y){\rput(-0.5*linethickness,-0.5*linethickness){\psframebox{T}}}
		// because the inner rput sets the box size to zero and thus changes the refpoint position

		       double sX=0; double sY=0; // for center or baseline
		       if (obj.getVertAlign()== TEXT_VALIGN_TOP) sY=1;
		       else if (obj.getVertAlign()== TEXT_VALIGN_BOTTOM) sY=-1;
		       if (obj.getHorAlign()== TEXT_HALIGN_LEFT) sX=-1;
		       else if (obj.getHorAlign()== TEXT_HALIGN_RIGHT) sX=1;
		       double lw=0.5*((Double)(obj.getAttribute(LINE_WIDTH))).doubleValue();
		       // rotate
		       double angle = ((Double)obj.getAttribute(TEXT_ROTATION)).doubleValue(); // in degrees
		       double c=Math.cos(Math.toRadians(angle));
		       double s=Math.sin(Math.toRadians(angle));
		       double dx=lw*(c*sX-s*sY);
		       double dy=lw*(c*sY+s*sX);
		       buf.append("\\rput(" + PEToolKit.doubleToString(dx)
				  + ","+  PEToolKit.doubleToString(dy) +")");
		       buf.append("{");

		       //end of linethickness correction
		}

		// rput command : \rput[refpoint]{rotation}(x,y){stuff}
		
		// \\rput
		buf.append("\\rput");
		if (obj.getVertAlign()!= TEXT_VALIGN_CENTER || obj.getHorAlign()!= TEXT_HALIGN_CENTER){
			buf.append("[");
			// [refpoint]
			if (obj.getVertAlign()== TEXT_VALIGN_TOP) buf.append("t");
			else if (obj.getVertAlign()== TEXT_VALIGN_BOTTOM) buf.append("b");
			else if (obj.getVertAlign()== TEXT_VALIGN_BASELINE) buf.append("B");
			
			if (obj.getHorAlign()==TEXT_HALIGN_LEFT) buf.append("l");
			else if (obj.getHorAlign()==TEXT_HALIGN_RIGHT) buf.append("r");
		
			buf.append("]");
		}
		
		// {rotation} [pending] todo
		double angle = ((Double)obj.getAttribute(TEXT_ROTATION)).doubleValue(); // in degrees
		if (angle!=0) {
			buf.append('{');
			buf.append(PEToolKit.doubleToString(angle));
			buf.append('}');
		}
		
		// (x,y)
		buf.append(obj.getPoint(PicText.P_ANCHOR,null));
		
		// {stuff} 
		buf.append("{");
		if (obj.getFrameType()==TEXT_BOX_NO_FRAME) { // stuff = text 
			buf.append(obj.getText());
		}
		else { // stuff = psframebox or psovalbox or pscirclebox

			if (obj.getFrameType()==TEXT_BOX_CIRCLE) buf.append("\\pscirclebox");
			else if (obj.getFrameType()==TEXT_BOX_OVAL) buf.append("\\psovalbox");
			else buf.append("\\psframebox");
			// parameters
			buf.append("[");
			buf.append(paramStr.getParameterBuffer());
			buf.append("]");
			buf.append("{");
			buf.append(obj.getText());
			buf.append("}");
			// closing line-thickness correction:
			buf.append("}");
		}
		buf.append("}");buf.append(CR_LF);
		return buf.toString();
	}

} 
