/*
 PstricksUtilities.java - March 1, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
package jpicedt.format.output.pstricks;

import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.Arrow;
import jpicedt.graphic.model.PicObjectConstants;

import java.awt.Color;
import java.util.StringTokenizer;

/**
 * Collection of static methods for the PsTricks format
 * @author Sylvain Reynal
 * @version $Id: PstricksUtilities.java,v 1.6.2.1 2007/09/02 11:55:14 reynal Exp $ 
 * @since jPicEdt 1.3.1
 */
public class PstricksUtilities implements PstricksConstants, PicObjectConstants {
	
	/**
	 * Returns PsTricks's colour name from the given Color object; 
	 * a null String if none matches (in which case the caller might define a new colour by using a \\newrgbcolor or a \\newgray command...)
	 */
	public static String getPsTricksColorName(Color c){
		for(int i=0; i<PST_COLORS.length; i++){
			if (PST_COLORS[i].equals(c)) return PST_COLORNAMES[i];
		}
		return null;
	}

	/**
	 * Returns a Color object corresponding to the given PsTricks's colour name ; 
	 * a null Color if none matches (in which case the caller might look up a table of user-defined colours to see if one matches)
	 */
	public static Color getPsTricksColor(String name){
		for(int i=0; i<PST_COLORNAMES.length; i++){
			if (PST_COLORNAMES[i].equals(name)) return PST_COLORS[i];
		}
		return null;
	}

	
	/**
	 * Returns a String representing the given Arrow in the PsTricks syntax, e.g. ">" or "(" (or "" for NONE)
	 * @param rightArrow if TRUE, returns the String representation of this Arrow if it's a right-arrow ; 
	 *        otherwise, returns the left-arrow string.
	 */
	public static String toPstricksString(Arrow arrow, boolean rightArrow){
		// lookup arrow in the array of predefined arrows :
		int j=0; // default = Arrow.NONE
		for (int i=0; i<PST_ARROWS.length; i++){
			if (arrow == PST_ARROWS[i]){
				j = i;
				break;
			}
		}
		if (rightArrow) return PST_ARROWS_RIGHT[j]; else return PST_ARROWS_LEFT[j]; 
	}
	
	/**
	 * @return PsTricks's arrow string for the given Element (e.g. "{<->}") with parenthesis included.
	 */
	public static StringBuffer createPstricksStringFromArrows(Element obj){

		StringBuffer buf = new StringBuffer(5);
		Arrow leftArrow = (Arrow)obj.getAttribute(LEFT_ARROW);
		Arrow rightArrow = (Arrow)obj.getAttribute(RIGHT_ARROW);
		if (leftArrow == Arrow.NONE && rightArrow == Arrow.NONE) return buf; // no arrow => empty string !
		buf.append("{");
		buf.append(toPstricksString(leftArrow,false)); // e.g. "<" for ARROW_HEAD
		buf.append("-");
		buf.append(toPstricksString(rightArrow,true)); // e.g. ">" for ARROW_HEAD (this way, we really get "<->" instead of ">->" which was wrong !)
		buf.append("}");
		return buf;
	}


	/**
	 * @return an array containing two Arrows corresponding to the given PsTricks string
	 * (e.g. "-" or "<-" or ">>->" ...), left arrow first.
	 */
	public static Arrow[] createArrowsFromPstricksString(String str){

		Arrow[] tab = new Arrow[2];
		tab[0] = tab[1] = Arrow.NONE; // security
		StringTokenizer tokenizer = new StringTokenizer(str, "-");
		String tokenLeft="";
		String tokenRight="";
		tab[0] = tab[1] = Arrow.NONE; // security
		if (tokenizer.hasMoreTokens()){ // this is not a "-" string !
			if (str.startsWith("-")) { // e.g. "->"
				tokenLeft = "";
				tokenRight = tokenizer.nextToken();
			}
			else { // e.g. "<-" or "<->"
				tokenLeft = tokenizer.nextToken();
				if (tokenizer.hasMoreTokens()){ // e.g. "<->"
					tokenRight = tokenizer.nextToken();
				}
				else tokenRight = "";
			}
		}

		for(int i=0; i<PST_ARROWS.length; i++){ // [pending] "do-while" loop is faster here.
			if (tokenLeft.equals(PST_ARROWS_LEFT[i])) {
				tab[0]=PST_ARROWS[i];
				break;
			}
		}
		for(int i=0; i<PST_ARROWS.length; i++){ // [pending] "do-while" loop is faster here.
			if (tokenRight.equals(PST_ARROWS_RIGHT[i])) {
				tab[1]=PST_ARROWS[i];
				break;
			}
		}
		return tab;
	}


	/**
	 * Test
	 */
	public static void main(String arg[]){

		//	System.out.println("arg=" + arg[0]);
		Arrow[] arrows;
		String str = "<->";
		while(true){
			try {
				System.out.println("Input string ?");
				java.io.BufferedReader r = new java.io.BufferedReader(new java.io.InputStreamReader(System.in));
				str = r.readLine();
				System.out.println("OK !");
			}
			catch(Exception e){e.printStackTrace();}

			arrows = createArrowsFromPstricksString(str);
			System.out.println("Left Arrow = (" + arrows[0] + ")");
			System.out.println("Right Arrow = (" + arrows[1] + ")");
		}
	}



	
	
}


