/*
 AbstractSelectionHandler.java - February 15, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic;

import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.Drawing;

import java.util.*;

/**
 * Provides some basic implementation of the SelectionHandler interface. 
 * This implementation merely targets methods that can access or modify the content
 * of the SelectionHandler via the {@link #asCollection() asCollection()} method, hence the latter
 * method is left unimplemented here.
 * <br>
 * Besides, no storage mechanism is implemented in this abstract class.
 * Concrete implementation of this mechanism should be based on mutable Java Collection's, otherwise
 * selection operation will fail.
 * <br>
 * Finally, the content of the selection-handler is supposed to reflect the z-ordering of the 
 * {@link jpicedt.graphic.model.Drawing Drawing} it is
 * associated with. This abstract implementation does not take care of this.
 * @since jpicedt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: AbstractSelectionHandler.java,v 1.5.2.1 2007/09/02 11:56:15 reynal Exp $
 */
public abstract class AbstractSelectionHandler implements SelectionHandler {
	
	/**
	 * Return the number of selected objects. Convenience call to {@link #asCollection() asCollection()}.
	 */
	public int size(){
		return asCollection().size();
	}

	/**
	 * Return an iterator over the selected Element's. Convenience call to {@link #asCollection() asCollection()}.
	 */
	public Iterator elements(){
		return asCollection().iterator();
	}
	
	/**
	 * Return the selected elements wrapped in a Collection
	 */
	public abstract Collection asCollection();
	
	/**
	 * Return the selected elements wrapped in an array. Convenience call to {@link #asCollection() asCollection()}.
	 */
	public Element[] asArray(){
		return (Element[]) asCollection().toArray(new Element[0]);
	}
	
	/**
	 * Return whether the given Element is selected or not. Convenience call to {@link #asCollection() asCollection()}.
	 * @param includeAncestors if true, return true if this Element or one of its ancestors is selected. 
	 */
	public boolean isSelected(Element e, boolean includeAncestors){
		if (asCollection().contains(e)) return true;
		if (includeAncestors && e.getParent()!=null) return isSelected(e.getParent(), true);
		else return false;
	}
	
	/**
	 * Select all Element's in the given drawing.
	 */
	public void selectAll(Drawing d){
		unSelectAll(); // avoid duplicate
		for (Iterator it=d.elements(); it.hasNext();){
			Element e = (Element)it.next();
			addToSelection(e);
		}
	}
	
	/**
	 * Unselect all Element's in the given drawing. Convenience call to {@link #asCollection() asCollection()}.
	 */
	public void unSelectAll(){
		asCollection().clear();
	}
	
	/**
	 * Add the given Element to the current selection. Convenience call to {@link #asCollection() asCollection()}.
	 */
	public void addToSelection(Element e){
		asCollection().add(e);
	}
	
	/**
	 * Replace the current selection with the given element. 
	 */
	public void replaceSelection(Element e){
		unSelectAll();
		addToSelection(e);
	}

	/**
	 * Unselect the given Element, i.e. remove it from the current selection.
	 * Convenience call to {@link #asCollection() asCollection()}.
	 */
	public void unSelect(Element e){
		asCollection().remove(e);
	}
	
	/**
	 * Delete all selected Element's from the given Drawing, and remove the reference to them
	 * from this SelectionHandler.
	 */
	public void delete(Drawing d){
		for (Iterator it = elements(); it.hasNext();){
			Element e = (Element)it.next();
			d.removeElement(e);
		}
		unSelectAll();
	}
}
