/*
 PEMouseEvent.java - January 3, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.event;

import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;

import java.awt.event.MouseEvent;
import java.awt.event.InputEvent;
import java.util.EventObject;


/**
 * This is an enriched implementation of MouseEvent that return mouse coordinates in model-coordinates in addition
 * to the usual screen-coordinates.
 * @author Sylvain Reynal
 * @version $Id: PEMouseEvent.java,v 1.8.2.1 2007/09/02 11:56:18 reynal Exp $
 * @since jpicedt 1.3.2
 */
public class PEMouseEvent extends EventObject {

	private MouseEvent awtEvent;
	private PicPoint pt;
	
	/**
	 * @param awtEvent the "standard" MouseEvent, as generated by Component's processMouseXXXEvent method.
	 * @param source This constrains PEMouseEvent's to be generated by PEMouseListener's attached to 
	 *        PECanvas's only. This indeed is the only event source that makes sense with this kind of 
	 *        events (notably because we need the zoom factor at some point).
	 * @param coordinates of click point in model-coordinate system
	 */
	public PEMouseEvent(MouseEvent awtEvent, PECanvas source, PicPoint pt){
			
			super(source);
			this.awtEvent = awtEvent;
			this.pt = pt;
	}
	
	/**
	 * Return the "standard" MouseEvent wrapped in this PEMouseEvent ; this may be used
	 *         to retrieve the MouseEvent's fields not accessible through PEMouseEvent's methods.
	 */
	public MouseEvent getAwtMouseEvent(){
		return awtEvent;
	}
	
	/**
	 * Return the PECanvas on which this mouse event occured 
	 */
	public PECanvas getCanvas(){
		return (PECanvas)awtEvent.getSource(); // guaranteed not to raise a ClassCastException !
	}
		
	/**
	 * Return the x-coordinate of the mouse click point in model-coordinate.
	 */
	public double getX(){
		return pt.x;
	}
	
	/**
	 * Return the y-coordinate of the mouse click point in model-coordinate.
	 */
	public double getY(){
		return pt.y;
	}
	
	/**
	 * Return the coordinates of the mouse click point in model-coordinate ; returned value is NOT
	 *         guaranteed to remain unchaged over time ! Make a copy if you need an immutable value
	 *         (e.g. using new PicPoint(pt)).
	 */
	public PicPoint getPicPoint(){
		return pt;
	}
	
	
	/**
	 * Return the number of mouse clicks associated with this event
	 */
	public int getClickCount(){
		return awtEvent.getClickCount();
	}
	
	/**
	 * Return  the modifiers flag for this mouse event
	 * @see java.awt.event.InputEvent
	 */
	public int getModifiers(){
		return awtEvent.getModifiers();
	}
	
	/**
	 * Return  the modifiers flag for this mouse event
	 * @see java.awt.event.InputEvent
	 */
	public int getModifiersEx(){ // since JDK1.4
		return awtEvent.getModifiersEx();
	}

	/**
	 * Convenience method to test the state of all modal keys and mouse buttons just after the mouse event occured.
	 * Mask are made of InputEvent's mask OR'ed together
	 * @param onMask keys/buttons which must be on, e.g. SHIFT_DOWN_MASK | BUTTON1_DOWN_MASK
	 * @param offMask keys/buttons which must be off, e.g. CTRL_DOWN_MASK | ALT_DOWN_MASK
	 * @see java.awt.event.InputEvent (extended modifiers masks have been introduced starting from JDK1.4)
	 */
	public boolean checkModifiersEx(int onMask, int offMask){
		if ((getModifiersEx() & (onMask|offMask)) == onMask) return true;
		return false;
	}
	
	/**
	 * Return whether or not the ALT modifier is down on this mouse event 
	 */
	public boolean isAltDown(){
		return awtEvent.isAltDown();
	}
	
	/**
	 * Return whether or not the "AltGr" modifier is down on this mouse event
	 */
	public boolean isAltGraphDown(){
		return awtEvent.isAltGraphDown();
	}
	
	/**
	 * Return whether or not the CTRL modifier is down on this mouse event
	 */
	public boolean isControlDown(){
		return awtEvent.isControlDown();
	}
	
	/**
	 * Return whether or not the META (aka ALT) modifier is down on this mouse event
	 */
	public boolean isMetaDown(){
		return awtEvent.isMetaDown();
	}
	
	/**
	 * Return whether or not the SHIFT modifier is down on this mouse event
	 */
	public boolean isShiftDown(){
		return awtEvent.isShiftDown();
	}

	/**
	 * Return whether or not the Left button was pressed
	 */
	public boolean isLeftButton(){
		return (awtEvent.getModifiers() & InputEvent.BUTTON1_MASK) != 0;
	}
	
	/**
	 * Return whether or not the Middle button was pressed ([pending] what happens on MacOS ?)
	 */
	public boolean isMiddleButton(){
		return (awtEvent.getModifiers() & InputEvent.BUTTON2_MASK) != 0;
	}

	/**
	 * Return whether or not the Right button was pressed
	 */
	public boolean isRightButton(){
		return (awtEvent.getModifiers() & InputEvent.BUTTON3_MASK) != 0;
	}
	
	/**
	 * Return whether or not this mouse event is the popup-menu trigger event
	 */
	public boolean isPopupTrigger(){
		//return awtEvent.isPopupTrigger();
		// terrible hack : isPopupTrigger work weird on windows (triggered by a mouse released !)
		//return (awtEvent.getModifiers()==InputEvent.BUTTON2_MASK || awtEvent.getModifiers()==InputEvent.BUTTON3_MASK);
		return isRightButton();
	}

	public String toString(){
		return "PEMouseEvent:"
		//+ "src=" + getBoard()
		+ " x=" + pt.x
		+ " y=" + pt.y
		+ " Buttons=" 
		+ (isLeftButton() ? "^" : "-") 
		+ (isMiddleButton() ? "^" : "-") 
		+ (isRightButton() ? "^" : "-")
		+ " Modifiers=" 
		+ (isShiftDown() ? "Shift " : "")
		+ (isAltDown() ? "Alt " : "")
		+ (isAltGraphDown() ? "AltGr " : "")
		+ (isControlDown() ? "Ctrl " : "")
		+ (isMetaDown() ? "Meta " : "")
		+ (isPopupTrigger() ? "Popup " : "")
		+ " Clicks="
		+ getClickCount();
	}
	
	
} 

