/*
 XmlAttributeSet.java - January 9, 2006 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.io.formatter;
 
import java.util.HashMap;
import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import java.awt.Color;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PEToolKit;

import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.Arrow;
import jpicedt.graphic.model.PicAttributeSet;
import jpicedt.graphic.model.PicCircleFrom3Points;
import jpicedt.graphic.model.PicEllipse;
import jpicedt.graphic.model.PicGroup;
import jpicedt.graphic.model.PicMultiCurve;
import jpicedt.graphic.model.PicObjectConstants;
import jpicedt.graphic.model.PicParallelogram;
import jpicedt.graphic.model.PicSmoothPolygon;
import jpicedt.graphic.model.PicPsCurve;
import jpicedt.graphic.model.PicText;
import jpicedt.graphic.model.StyleConstants;
 
/**
 * A map that contains XML-like name=value pairs, to be ultimately formatted as XML.
 * With respect to jpicedt.graphic.element.AttributeSet, this is more XML-oriented, and
 * not all attributes contained in AttributeSet are included in this map so as to reduce 
 * the size of the ultimate StringBuffer.
 * In comparison with the older scheme, where we would directly append 
 * succesive XML-formatted name=value pairs to the main StringBuffer, this offers the benefit
 * of avoiding redundant definition of XML parameters, which XML strictl forbids. 
 * @since jpicedt 1.4.pre5
 * @author Sylvain Reynal 
 */
public class XmlAttributeSet extends HashMap implements PicObjectConstants, JPICConstants {
	
	private final PicAttributeSet defaultAttributes = new PicAttributeSet(); // for comparison purpose with default values

	/**
	 * Constructs an empty map
	 */
	public XmlAttributeSet(){
		super();
	}
	 
	/**
	 * Returns a buffer constructed by iterating over the entries in this
	 *  and appending a name="value" string to the buffer for each attribute.
	 */
	protected StringBuffer toXML(){
		StringBuffer buf = new StringBuffer();
		Set entries = entrySet();
		String str;
		for (Iterator it = entries.iterator(); it.hasNext();){
			Map.Entry entry = (Map.Entry)it.next();
			buf.append(' ');
			buf.append(entry.getKey().toString()); // param
			buf.append('=');
			buf.append(' ');
			str = (String)entry.getValue().toString();
			if (!str.startsWith("\"")){
				buf.append('\"');
				buf.append((String)entry.getValue().toString()); // value
				buf.append('\"');
			}
			else
				buf.append((String)entry.getValue().toString()); // value
			buf.append(CR_LF_TAB);
		}
		return buf;
	}
	 
	/**
	 * Fills the given map with XML-like entries (name->value will translate into name="value" once text formatted) 
	 * for common attributes of the given Element, that is,
	 * attributes specific to a given subclass of Element are not included here, but should
	 * be handled directly by the appropriate formatter.<p>
	 * Attributes handled in this method include stroke-, fill-, shadow- and double-line attributes, as well as arrows,
	 * and are appended only if they differ from the default value in PicAttributeSet.
	 * See documentation on top of this class for details on the JPIC-XML syntax.
	 * <p>
	 * Not supported yet : "border" and "bordercolor"
	 */
	protected void putCommonAttributes(Element obj){

		PicAttributeSet attributes = obj.getAttributeSet();
		Color col;
		double x;
		
		// linewidth [SR:pending] add only if not the default value, this will save some space
		putNameValuePair(LINE_WIDTH,StyleConstants.getLineWidth(attributes),StyleConstants.getLineWidth(defaultAttributes));

		// linecolor
		putNameValuePair(LINE_COLOR,StyleConstants.getLineColor(attributes),StyleConstants.getLineColor(defaultAttributes));

		// linestyle
		if (StyleConstants.getLineStyle(attributes)==DASHED){
			putNameValuePair(LINE_STYLE,DASHED); // it's too complicated to distinguish b/w default and non-default values
			double[] dashArray = {StyleConstants.getDashOpaque(attributes),StyleConstants.getDashTransparent(attributes)};
			putNameValuePair(JPICFormatter.STROKE_DASHARRAY,dashArray);
		}
		else if (StyleConstants.getLineStyle(attributes)==DOTTED){
			putNameValuePair(LINE_STYLE, DOTTED);
			putNameValuePair(DOT_SEP,StyleConstants.getDotSep(attributes),StyleConstants.getDotSep(defaultAttributes));
			
		}
		else if (StyleConstants.getLineStyle(attributes)==NONE){
			putNameValuePair(LINE_STYLE, NONE);
		}
		//else buf.append(" stroke-style=\"solid\""); // default !

		// double line
		if (StyleConstants.isDoubleLine(attributes)) {
			putNameValuePair(DOUBLE_LINE,true);
			putNameValuePair(DOUBLE_SEP,StyleConstants.getDoubleLineSep(attributes),StyleConstants.getDoubleLineSep(defaultAttributes));
			putNameValuePair(DOUBLE_COLOR,StyleConstants.getDoubleLineColor(attributes),StyleConstants.getDoubleLineColor(defaultAttributes));
		}

		// fill style and fill colour
		String fillStyle = StyleConstants.getFillStyle(attributes);
		if (fillStyle==NONE){
			putNameValuePair(FILL_STYLE,NONE);
		}
		else { // solid or hatches -> add fill colour (even if it's redundant for hatches that are not "starred"=
			putNameValuePair(FILL_COLOR,StyleConstants.getFillColor(attributes),StyleConstants.getFillColor(defaultAttributes));
			if (fillStyle==SOLID) putNameValuePair(FILL_STYLE,SOLID);
			else { // hatches
				if (fillStyle==VLINES) putNameValuePair(FILL_STYLE,VLINES);
				else if (fillStyle==VLINES_FILLED) putNameValuePair(FILL_STYLE,VLINES_FILLED);
				else if (fillStyle==HLINES) putNameValuePair(FILL_STYLE,HLINES);
				else if (fillStyle==HLINES_FILLED) putNameValuePair(FILL_STYLE,HLINES_FILLED);
				else if (fillStyle==CROSSHATCH) putNameValuePair(FILL_STYLE,CROSSHATCH);
				else if (fillStyle==CROSSHATCH_FILLED) putNameValuePair(FILL_STYLE,CROSSHATCH_FILLED);

				putNameValuePair(HATCH_WIDTH,StyleConstants.getHatchWidth(attributes),StyleConstants.getHatchWidth(defaultAttributes));
				putNameValuePair(HATCH_SEP,StyleConstants.getHatchSep(attributes),StyleConstants.getHatchSep(defaultAttributes));
				putNameValuePair(HATCH_ANGLE,StyleConstants.getHatchAngle(attributes),StyleConstants.getHatchAngle(defaultAttributes));
				putNameValuePair(HATCH_COLOR,StyleConstants.getHatchColor(attributes),StyleConstants.getHatchColor(defaultAttributes));
			}
		}
		// shadow
		if (StyleConstants.isShadow(attributes)){
			putNameValuePair(SHADOW,true);
			putNameValuePair(SHADOW_SIZE,StyleConstants.getShadowSize(attributes),StyleConstants.getShadowSize(defaultAttributes));
			putNameValuePair(SHADOW_ANGLE,StyleConstants.getShadowAngle(attributes),StyleConstants.getShadowAngle(defaultAttributes));
			putNameValuePair(SHADOW_COLOR,StyleConstants.getShadowColor(attributes),StyleConstants.getShadowColor(defaultAttributes));
		}

		// arrows
		Arrow leftArrow = (Arrow)obj.getAttribute(LEFT_ARROW);
		Arrow rightArrow = (Arrow)obj.getAttribute(RIGHT_ARROW);
		if (leftArrow != Arrow.NONE) putNameValuePair(LEFT_ARROW,leftArrow.getName());
		if (rightArrow != Arrow.NONE) putNameValuePair(RIGHT_ARROW,rightArrow.getName());
		if (leftArrow != Arrow.NONE || rightArrow != Arrow.NONE){
			putNameValuePair(ARROW_GLOBAL_SCALE_WIDTH,StyleConstants.getArrowGlobalScaleWidth(attributes),StyleConstants.getArrowGlobalScaleWidth(defaultAttributes));
			putNameValuePair(ARROW_GLOBAL_SCALE_LENGTH,StyleConstants.getArrowGlobalScaleLength(attributes),StyleConstants.getArrowGlobalScaleLength(defaultAttributes));
			putNameValuePair(ARROW_WIDTH_MINIMUM_MM,StyleConstants.getArrowWidthMinimumMm(attributes),StyleConstants.getArrowWidthMinimumMm(defaultAttributes));
			putNameValuePair(ARROW_WIDTH_LINEWIDTH_SCALE,StyleConstants.getArrowWidthLineWidthScale(attributes),StyleConstants.getArrowWidthLineWidthScale(defaultAttributes));
			putNameValuePair(ARROW_LENGTH_SCALE,StyleConstants.getArrowLengthScale(attributes),StyleConstants.getArrowLengthScale(defaultAttributes));
			putNameValuePair(ARROW_INSET_SCALE,StyleConstants.getArrowInsetScale(attributes),StyleConstants.getArrowInsetScale(defaultAttributes));
			putNameValuePair(TBAR_WIDTH_MINIMUM_MM,StyleConstants.getTBarWidthMinimumMm(attributes),StyleConstants.getTBarWidthMinimumMm(defaultAttributes));
			putNameValuePair(TBAR_WIDTH_LINEWIDTH_SCALE,StyleConstants.getTBarWidthLineWidthScale(attributes),StyleConstants.getTBarWidthLineWidthScale(defaultAttributes));
			putNameValuePair(BRACKET_LENGTH_SCALE,StyleConstants.getBracketLengthScale(attributes),StyleConstants.getBracketLengthScale(defaultAttributes));
			putNameValuePair(RBRACKET_LENGTH_SCALE,StyleConstants.getRoundedBracketLengthScale(attributes),StyleConstants.getRoundedBracketLengthScale(defaultAttributes));
			putNameValuePair(POLYDOTS_SIZE_MINIMUM_MM,StyleConstants.getPolydotsSizeMinimumMm(attributes),StyleConstants.getPolydotsSizeMinimumMm(defaultAttributes));
			putNameValuePair(POLYDOTS_SIZE_LINEWIDTH_SCALE,StyleConstants.getPolydotsSizeLineWidthScale(attributes),StyleConstants.getPolydotsSizeLineWidthScale(defaultAttributes));
		}
		
		// overstrike :
		if (StyleConstants.isOverStrike(attributes)){
			putNameValuePair( OVER_STRIKE, true);
			putNameValuePair( OVER_STRIKE_WIDTH, StyleConstants.getOverStrikeWidth(attributes), StyleConstants.getOverStrikeWidth(defaultAttributes));
			putNameValuePair( OVER_STRIKE_COLOR, StyleConstants.getOverStrikeColor(attributes),StyleConstants.getOverStrikeColor(defaultAttributes));
		}

		// custom pstricks parameters
		if (StyleConstants.getPstCustom(attributes).length()>0){
			putNameValuePair(PST_CUSTOM,StyleConstants.getPstCustom(attributes));
		}
		
		// misc params
		if (StyleConstants.getDimen(attributes)==INNER) putNameValuePair(DIMEN,INNER);
		else if (StyleConstants.getDimen(attributes)==MIDDLE) putNameValuePair(DIMEN,MIDDLE);
		// else default OUTER
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map.
	 */
	protected void putNameValuePair(Object name, String value){
		put(name,value);
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map if value differs from the given default value.
	 */
	protected void putNameValuePair(Object name, Color col, Color def){
		if (def!=null && def.equals(col)) return;
		putNameValuePair( name, colorToHex(col));
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map.
	 */
	protected void putNameValuePair(Object name, boolean bol){
		putNameValuePair( name, (bol==true ? "true" : "false"));
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map
	 */
	protected void putNameValuePair(Object name, PicPoint pt){
		putNameValuePair( name, pt.toString());
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map.
	 * double are formatted with two-digit precision.
	 */
	protected void putNameValuePair(Object name, double value){
		putNameValuePair( name, PEToolKit.doubleToString(value));
	}

	/**
	 * Given attribute name and value, append a name/value entry to the given map if value differs from the given default value.
	 * double are formatted with two-digit precision.
	 */
	protected void putNameValuePair(Object name, double value, double def){
		if (def==value) return;
		putNameValuePair( name, PEToolKit.doubleToString(value));
	}

	/**
	 * Given attribute name and an array of doubles, append a name/value entry to the given map.
	 * doubles are formatted with two-digit precision, and are separated by semicolons.
	 */
	protected void putNameValuePair(Object name, double[] values){
		StringBuffer buf =new StringBuffer(); // string of comma-separated-values
		for (int i=0; i<values.length-1; i++){
			buf.append(PEToolKit.doubleToString(values[i]));
			buf.append(";");
			if (i%18==17) buf.append(CR_LF_TAB);
		}
		buf.append(PEToolKit.doubleToString(values[values.length-1]));
		put(name, buf.toString());
	}

	/**
	 * Given attribute name and an array of PicPoint's, append a name/value entry to the given map.
	 * Points are formatted by means of their toString() method, and are separated by semicolons.
	 */
	protected void putNameValuePair(Object name, PicPoint[] values){
		StringBuffer buf =new StringBuffer(); // string of comma-separated-values
		for (int i = 0; i < values.length-1; i++) {
			buf.append(values[i]);
			buf.append(";");
			if (i%6==5) buf.append(CR_LF_TAB);
		}
		buf.append(values[values.length-1]);
		put(name, buf.toString());
	}

	/**
	 * Convert the given Color to a XML-like RGB string, each colour being expressed in the hexadecimal radix 
	 */
	protected String colorToHex(Color c){
		String red,green,blue;
		if (c.getRed()==0) red="00"; else red = Integer.toHexString(c.getRed());
		if (c.getGreen()==0) green="00"; else green = Integer.toHexString(c.getGreen());
		if (c.getBlue()==0) blue="00"; else blue = Integer.toHexString(c.getBlue());
		StringBuffer buf = new StringBuffer("#");
		buf.append(red);
		buf.append(green);
		buf.append(blue);
		return buf.toString();
	}
	 
}
 
