/*
 AbstractCurve.java - Decembre 25, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PicVector;
import jpicedt.graphic.PEToolKit;

import jpicedt.graphic.toolkit.ActionFactory;
import jpicedt.graphic.toolkit.ActionLocalizer;
import jpicedt.graphic.toolkit.ActionDispatcher;
import jpicedt.graphic.toolkit.AbstractCustomizer;
import jpicedt.graphic.toolkit.PEAction;
import jpicedt.graphic.view.HitInfo;

import jpicedt.graphic.event.DrawingEvent;

import jpicedt.widgets.*;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;

import jpicedt.Log;//debug

/**
 * A curve made up of a concatenation of cubic bezier curves and straight lines,
 * hereafter globally denoted as "segments". Straight lines are actually just bezier curves
 * with control-points being identical with end-points.
 * This curve may be either closed or not, and has a
 * variable number of points. Segments may be added/inserted/deleted.<p>
 *
 * There are two sort of points in this curve:
 * <ul>
 *   <li> <b>subdivision</b> points, that is, endpoints of Bezier segments (indicated by SP(i), where i=segment_index, on
 *        picture above)
 *   <li> <b>control-points</b> of these Bezier segments (indicated by CP_first(i) and CP_second(i) for the first and
 *        the second control-point respectively of the ith-segment, on picture above) 
 * </ul>
 * They are globally denoted as "Bezier points" and act as specification-points.
 * <p>
 *
 * <h1>Specification points</h1>
 * Protected <code>pts</code> ArrayList inherited from 
 * <a href="DefaultLeafElement.html">DefaultLeafElement</a>, which (according to the contract in the
 * latter class) store specification points, contain the coordinates
 * of all Bezier points with the following indexing scheme :
 * <ul>
 * <li> subdivision-points are indexed 0,3,6,9,12,... 
 * <li> control points are indexed 1,2 ; 4,5 ; 7,8 ; 10,11 ; ... 
 * </ul>
 * These indices are globally denoted as "Bezier indices" throughout the code documentation.
 * <p>
 * Arrays of specification-point coordinates are managed differently if the curve is open or close :
 * <ul>
 * <li>If the curve is <b>open</b>, the last element of the array is a subdivision point, 
 *     namely the second curve's end-point.  To sum up, pts[3i,3i+1,3i+2,3i+3] are the
 *     1st endpoint, 2 control-points and last endpoint of the corresponding
 *     elementary Bezier curve resp., for i=[0..number of segments-1].
 * <li>If the curve is <b>closed</b>, these arrays end up at the last control-point (e.g. 11 for a curve having 4
 *     subdivision-points), hence for the last segment, pts[3i+3] is not defined. 
 *     Yet in this case, getPBCBezierIndex() arranges for every index to be a valid index
 *     using periodic-boundary-conditions (PBC). This rule in turns applies to every method involving a point index.
 * </ul><p>
 *
 * <h1>Subdivision-point vs. segment numbering scheme</h1>
 * When used throughout the code documentation, the former expression
 * means that 0,1,2 ... indices relate to the first, second, third,... subdivision point.
 * This is indicated by SP(i) on picture above.
 * Conversely, each segment is made up of three points (with the exception of the last curve's end-point if curve is open, which
 * comprises a single segment), amongst which there is one subdivision point.
 * Hence segment numbering ranges from 0 to getNumberOfSegments()-1 following the same numbering scheme as subdivision-points. 
 * Methods pointToSegmentIndex() and
 * segmentToPointIndex() allow to translate between segment and bezier-point numbering schemes.<p>
 *
 * <h1>Adding/removing points</h1>
 * Methods splitSegment(), removeSubdivisionPoint(), lineTo() and curveTo() provide means to add/remove points to this curve.
 * Abstract methods addPoint(),removePoint() and single-argument splitSegment(), whose implementation depends on the
 * geometric "meaning" of user-controlled points,  is left to subclassers.
 *
 * <h1>DrawingEvent's dispatching</h1>
 * As a rule of thumb, any concrete subclass should post DrawingEvent's itself,
 * where it thinks it has to, since none of the methods implemented here does it (this is aimed at reducing the burden
 * for registered listeners which otherwise might -now and often- receive events twice).
 *
 * <p><img src="doc-files/abstract_curve.png"></p>
 *
 * [SR:TODO] split curve into several pieces (aka old PicPolygon.convertToLines() method) ; <br> 
 * [SR:TODO] join curves ;  <br>
 * [SR:TODO] add resample() method allowing to delete m point out of n (useful for polygons generated from
 *           data files)
 * [SR:BUG] setSmooth()/setSymmetric() => bug when neighbouring control-points are identical.
 * @author    Vincent Guirardel, Sylvain Reynal.
 * @since     jpicedt 1.3.3
 * @version $Id: AbstractCurve.java,v 1.24.2.1 2007/09/02 11:56:20 reynal Exp $
 */
public abstract class AbstractCurve extends DefaultLeafElement implements ActionFactory {

	/** constant field for getPointType()  */
	public static final int INVALID_POINT_INDEX = 0;
	
	/** constant field for getPointType() */
	public static final int FIRST_CURVE_END_POINT = 1;
	
	/** constant field for getPointType() */
	public static final int SUBDIVISION_POINT = 2;
	
	/** constant field for getPointType() */
	public static final int FIRST_SEGMENT_CONTROL_POINT = 3;
	
	/** constant field for getPointType() */
	public static final int SECOND_SEGMENT_CONTROL_POINT = 4;
	
	/** constant field for getPointType() */
	public static final int LAST_CURVE_END_POINT = 5;

	//////////////////////////
	//// PROTECTED FIELDS
	//////////////////////////

	/**
	 * tells whether this curve is closed of not
	 */
	protected boolean isClosed = false;
	
	
	////////////// 
	/// PRIVATE
	//////////////
	private PicVector ptBuffer1 = new PicVector(); // used by some utility methods to reduce object creation
	private PicVector ptBuffer2 = new PicVector();

	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Creates a new empty open Abstract curve
	 */
	public AbstractCurve() {
		this(false);
	}

	/**
	 * Creates a new empty Abstract curve
	 * @param closed whether the generated multi-curve will be closed or not
	 */
	public AbstractCurve(boolean closed) {
		super(0);
		isClosed = closed;
		//isStraightSeg.add(Boolean.FALSE); // add at least one segment
	}


	/**
	 * Creates a new empty Abstract curve with the given set of attributes
	 * @param closed  whether the generated multi-curve will be closed or not
	 * @param set     attribute set to be bound to this curve
	 */
	public AbstractCurve(boolean closed, PicAttributeSet set) {
		this(closed);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * Creates a new Abstract curve and allocates as many points as needed by
	 * the given number of segments. 
	 * Each segment comprises three Bezier points (i.e. one end-point, and
	 * two control-points), yet if the curve is open, there is one more Bezier point, namely the last
	 * curve end-point.
	 * @param nbSegments nb of elementary lines or cubic Bezier curves ; if 0, this curve is reduced to a single point
	 * @param closed whether this curve is closed or not
	 * @throws IllegalArgumentException if nbSegments is negative
	 */
	public AbstractCurve(int nbSegments, boolean closed) {
		this(closed);
		if (nbSegments<0) throw new IllegalArgumentException("nbSegments must be non-negative");
		else if (nbSegments>0){
			int nPts = 3*nbSegments;
			if (!closed) nPts += 1;
			for (int i=0; i<nPts; i++) pts.add(new PicPoint());
			//for (int i=0; i<nbSegments; i++) isStraightSeg.add(Boolean.FALSE);
		}
		else if (nbSegments==0){
			pts.add(new PicPoint());
		}
			
	}


	/**
	 * Creates a new Abstract curve with the given number of segments, and attaches the given attribute set to it.
	 *
	 * @param nbSegments nb of elementary lines or cubic Bezier curves
	 * @param set     attribute set to be bound to this element
	 */
	public AbstractCurve(int nbSegments, boolean closed, PicAttributeSet set) {
		this(nbSegments, closed);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * "cloning" constructor (to be used by clone())
	 */
	public AbstractCurve(AbstractCurve curve) {
		super(curve); // reminder : deep copy of pts and attribute set
		//this.isStraightSeg = (ArrayList)curve.isStraightSeg.clone(); // shallow copy, yet this obviously doesn't matter since we copy refs to Boolean.FALSE/TRUE
		this.isClosed = curve.isClosed;
	}


	///////////////////////////////////////////
	/// OPERATIONS on point and segment indices
	///////////////////////////////////////////

	/**
	 * Returns whether the given point index relates to a curve's end-point, 
	 * a subdivision point, a control point (in which case
	 * this method returns the rank of the control point inside the segment to which it belongs), or (if the curve
	 * is open) is an invalid point index.<br>
	 * Note that if this curve is CLOSED, this method never returns FIRST_CURVE_END_POINT nor LAST_CURVE_END_POINT for that matter.
	 * @return a constant field, i.e. either SUBDIVISION_POINT, FIRST_SEGMENT_CONTROL_POINT,
	 *         SECOND_SEGMENT_CONTROL_POINT, FIRST_CURVE_END_POINT, LAST_CURVE_END_POINT or INVALID_POINT_INDEX.
	 * @author Sylvain Reynal [18/01/2003]
	 */
	public int getPointType(int index){
		index = getPBCBezierIndex(index);
		if (index<0 || index>=getNumberOfSpecificationPoints()) return INVALID_POINT_INDEX;
		else if (index==0  && (!isClosed())) return FIRST_CURVE_END_POINT;
		else if (index==getNumberOfSpecificationPoints()-1 && (!isClosed())) return LAST_CURVE_END_POINT;
		else switch (index%3){
			case 0 : return SUBDIVISION_POINT;
			case 1 : return FIRST_SEGMENT_CONTROL_POINT;
			case 2 : return SECOND_SEGMENT_CONTROL_POINT;
			default : return INVALID_POINT_INDEX; // unreachable (unless JVM is buggy, huu huu)
			}
	}

	/**
	 * Convenience method for Periodic Boundary Condition (PBC) management when curve is CLOSED.<p>
	 * If curve is CLOSED, returns the given index modulo the number of bezier points, i.e. an index
	 * guaranteed to lie b/w 0 and the number of bezier points minus one.<br>
	 * Otherwise leaves unchanged.<p>
	 * @param bezierPtIdx any integer, positive or not, greater than the number of bezier points or not, etc...
	 * @return a bezier-point index guaranteed to lie between 0 and the number of bezier points minus one.
	 */
	public int getPBCBezierIndex(int bezierPtIdx) {
		if (isClosed()) {
			bezierPtIdx = bezierPtIdx % getNumberOfSpecificationPoints(); // may be negative ...
			if (bezierPtIdx < 0) bezierPtIdx += getNumberOfSpecificationPoints();
			return bezierPtIdx;
		}
		else return bezierPtIdx;
	}
	
	/**
	 * Convenience method for Periodic Boundary Condition (PBC) management when curve is CLOSED.<p>
	 * If curve is CLOSED, returns the given index modulo the number of segment, i.e. an index
	 * guaranteed to lie b/w 0 and the number of segments minus one.<br>
	 * Otherwise leaves unchanged.<p>
	 * @param segmentIdx any integer, positive or not, greater than the number of bezier points or not, etc...
	 * @return a segment index guaranteed to lie between 0 and the number of segments minus one.
	 */
	public int getPBCSegmentIndex(int segmentIdx) {
		return getPBCBezierIndex(segmentIdx*3)/3;
	}

	/**
	 * Returns the bezier-index of the nearest subdivision-point of the given control-point. The policy is to
	 * return the same index if the given point is NOT a control-point.
	 */
	public int getNearestSubdivisionPoint(int controlPtIdx){
		if (!isControlPoint(controlPtIdx)) return controlPtIdx;
		int type = getPointType(controlPtIdx);
		switch (type){
			case FIRST_SEGMENT_CONTROL_POINT: return getPBCBezierIndex(controlPtIdx-1);
			case SECOND_SEGMENT_CONTROL_POINT: return getPBCBezierIndex(controlPtIdx+1);
			default: return controlPtIdx;
		}
	}
			
	/**
	 * Returns the bezier-index of the alternate control-point of the given control-point. The policy is to
	 * return the same index if the given point is NOT a control-point. Also note that the returned index is
	 * not guaranteed to be a valid index (there's no convincing way to test it here).
	 */
	public int getAlternateControlPoint(int controlPtIdx){
		if (!isControlPoint(controlPtIdx)) return controlPtIdx;
		int type = getPointType(controlPtIdx);
		switch (type){
			case FIRST_SEGMENT_CONTROL_POINT: return getPBCBezierIndex(controlPtIdx-2);
			case SECOND_SEGMENT_CONTROL_POINT: return getPBCBezierIndex(controlPtIdx+2);
			default: return controlPtIdx;
		}
	}
			
	/**
	 * Return whether the bezier-point with the given index is a control-point.
	 */
	public boolean isControlPoint(int index){
		return getPointType(index)==FIRST_SEGMENT_CONTROL_POINT || getPointType(index)==SECOND_SEGMENT_CONTROL_POINT;
	}
	
	/**
	 * Returns the index of the segment the given point belongs to.
	 */
	public int pointToSegmentIndex(int pointIndex){
		return getPBCBezierIndex(pointIndex)/3;
	}
	
	/**
	 * Returns the index of the bezier-point belonging to the given segment, and having the given
	 * pointType.
	 * @param pointType one of SUBDIVISION_POINT, FIRST_SEGMENT_CONTROL_POINT or SECOND_SEGMENT_CONTROL_POINT.
	 *        Other qualifiers are meaningless here, and act as the SUBDIVISION_POINT qualifier.
	 */
	public int segmentToPointIndex(int segmentIndex, int pointType){
		segmentIndex = getPBCSegmentIndex(segmentIndex); // security
		switch (pointType){
			case FIRST_SEGMENT_CONTROL_POINT : return getPBCBezierIndex(segmentIndex * 3 + 1);
			case SECOND_SEGMENT_CONTROL_POINT : return getPBCBezierIndex(segmentIndex * 3 + 2);
			default : return getPBCBezierIndex(segmentIndex * 3);
		}
		
	}
	
	/**
	 * Returns true iff the given bezier point index is a valid point index. 
	 * @see #getPointType(int)
	 */
	public boolean isValidBezierIndex(int pointIndex) {
		return getPointType(pointIndex)!=INVALID_POINT_INDEX;
	}

	/**
	 * Returns true iff the given segment index is a valid segment index. This is always true for closed curves. 
	 */
	public boolean isValidSegmentIndex(int segIdx) {
		if (isClosed()) return true;
		return (segIdx >= 0) && (segIdx < getNumberOfSegments());
	}
	
	/**
	 * Return whether the current nb of specification points is valid, i.e. there is no incomplete segment.
	 * This may be used in conjunction with the "assert" keyword (as of JDK1.4) for debugging purpose.
	 */
	protected final boolean hasValidSize(){
		return pts.isEmpty() || pts.size()==1 || (isClosed() && (pts.size())%3 == 0) || (!isClosed() && (pts.size()-1)%3 == 0);
	}
	
	//////////////////////////////////
	/// OPERATIONS on Bezier POINTS
	//////////////////////////////////

	/**
	 * Returns a reference on the specification point with the given index.
	 * This method is overriden from superclass so that, if the curve is closed, any index
	 * is a valid index using Periodic Boundary Conditions.
	 */
	protected PicPoint getSpecificationPoint(int index){
		return super.getSpecificationPoint(getPBCBezierIndex(index));
	}

	
	/**
	 * Set the coordinates of the Bezier point with the given index to the given location, 
	 * but doesn't fire any DrawingEvent. If the curve is CLOSED, this methods arranges for any
	 * index to be a valid index by means of PBC's.
	 * @param c not used here
	 */
	public void setPoint(int index, PicPoint pt, EditPointConstraint c) {
		if (!isValidBezierIndex(index)) throw new IndexOutOfBoundsException(new Integer(index).toString());
		else setSpecificationPoint(getPBCBezierIndex(index),pt);
	}

	/**
	 * Adds the given point to the end of this curve.
	 */
	public abstract void addPoint(PicPoint pt);
	
	/**
	 * Split the segment having the given index.
	 * @return the index of the user-controlled point which got inserted, according to the indexing scheme of set/getPoint().
	 *         This may for instance allow a receiver to control the "new" segment shape by calling setPoint() 
	 *         with this index as a parameter w/o the burden of computing an exact point index (a thing that may depend
	 *         on the particular implementation of this method).
	 */
	public abstract int splitSegment(int index, PicPoint pt);
	
	/**
	 * Remove the point with the given index from this curve
	 */
	public abstract void removePoint(int index);
	
	/**
	 * If this curve if OPEN and NON-EMPTY, 
	 * adds the given points (2 control points and an endpoint) to the end of the curve.<br>
	 * If the curve is closed,  use <code>splitSegment</code> instead.
	 * @param ptCtrl1  first control point of the new Bezier segment
	 * @param ptCtrl2  second control point of the new Bezier segment
	 * @param ptEnd second end-point of the new Bezier segment
	 */
	public void curveTo(PicPoint ptCtrl1, PicPoint ptCtrl2, PicPoint ptEnd) {
		if (isClosed() || getNumberOfSpecificationPoints()==0) return;
		pts.add(ptCtrl1.clone()); // safe copy
		pts.add(ptCtrl2.clone());
		pts.add(ptEnd.clone());
		//isStraightSeg.add(Boolean.FALSE); // new segment is curved
	}


	/**
	 * Adds a STRAIGHT segment to the end of this curve, i.e. control-points and subdivision-points are identical
	 * at each segment's end. This is a convenient call to {@link #curveTo curveTo}.
	 * <p>
	 * Note that this method does NOT fire any DrawingEvent.<br>
	 * This method does nothing if the curve is EMPTY or CLOSED.
	 * @param pt The second end-point of the line to be added
	 */
	public void lineTo(PicPoint pt) {
		if (isClosed() || getNumberOfSpecificationPoints()==0) return; // do nothing if curve is either closed or empty.

		// int oldNbPts = getNumberOfSpecificationPoints(); // length of original curve
		// PicVector v = new PicVector(getSpecificationPoint(oldNbPts-1), pt);
		// set first bezier ctrl point to 1/3 :
		// PicPoint C1 = getSpecificationPoint(oldNbPts-1,null); // safe copy
		// C1.translate(v,1/3.0);
		// set second bezier ctrl point to 2/3 :
		// PicPoint C2 = getSpecificationPoint(oldNbPts-1,null);
		// C2.translate(v,2/3.0);

		PicPoint ctrl1 = getSpecificationPoint(getNumberOfSpecificationPoints()-1); // last curve end-point
		curveTo(ctrl1,pt,pt);
		
		//isStraightSeg.set(isStraightSeg.size()-1, Boolean.TRUE); // make correction against curveTo() behaviour
	}


	/**
	 * Split a segment (either straight or curved) at a given point using two additionnal control points given
	 * as parameters.<p>
	 * Implementation works as follow : parameters ptLeft and ptRight act as new control points, 
	 * the original segment - labelled as [a,b,c,d] - being split up into two new segments, i.e. 
	 * [a,b,ptleft,pt] and [pt,ptright,c,d]. These two segments inherit their
	 * straightness attribute from the original segment. 
	 * This works for closed as well as open curves.<p> 
	 * Note that this method does NOT fire any DrawingEvent.
	 *
	 * @param seg      index of segment to be split, i.e. 0,1,2,... for the 1st, 2nd, 3rd,... segment.
	 * @param ptleft   first new control-point
	 * @param pt       the point at which segment must be split (= new subdivision point)
	 * @param ptright  second new control-point
	 * @return the index of the subdivision-point which got inserted (see splitSegment(int,PicPoint) for details)
	 */
	public int splitSegment(int seg, PicPoint ptleft, PicPoint pt, PicPoint ptright) {

		if (seg < 0 || seg >= getNumberOfSegments()) throw new IndexOutOfBoundsException(new Integer(seg).toString());

		int oldNbPts = getNumberOfSpecificationPoints(); //length of the old curve
		int oldNbSeg = getNumberOfSegments(); //old number of segments
		int ileft = segmentToPointIndex(seg, SECOND_SEGMENT_CONTROL_POINT); // index at which ptleft has to be inserted in bezierPtX
		
		// PicPoint arguments must be SAFELY copied, since we can't take for granted caller will never alter them afterwards
		pts.add(ileft,ptright.clone()); // ArrayList.add(i,object) shifts any element at index >= i to the "right"
		pts.add(ileft,pt.clone());      // hence we indeed end up with "ptleft ; pt ; ptright" in the right order ;-)
		pts.add(ileft,ptleft.clone());
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "seg="+seg+", ptL="+ptleft+", pt="+pt+", ptR="+ptright);
		return getPBCBezierIndex(ileft+1);
	}


	/**
	 * Remove a subdivision point from this curve, together with the two neighbouring control points.
	 * This shifts any ensuing points to the "left", i.e. reduces indices by one.<br>
	 * Nothing is done if this curve has only one point. If both neighbouring segments
	 * had the same straightness value, it is inherited by the new segment. Otherwise, we set it to false.<p>
	 * This method does NOT fire any DrawingEvent. 
	 * @param index index of the subdivision point to be removed with respect to SUBDIVISION point numbering scheme, 
	 *        e.g. 0 for the first subdivision point (= first curve end-point), 1 for the second one (= bezier point
	 *        with index "3"), etc... This is similar to the SEGMENT numbering scheme incidentally.
	 * @throws an IllegalArgumentException if the given index is not a valid subdivision point index, i.e. 
	 *         is greater than the nb of segments.
	 */
	public void removeSubdivisionPoint(int subdivIndex) { 
		if (getNumberOfSegments() <= 1) return; // we won't remove the only remaining segment ! [meaning 1 subdivision point + 2 control points for a closed curve]

		int pos = segmentToPointIndex(subdivIndex, SUBDIVISION_POINT);
		int ptType = getPointType(pos);
		if (ptType == INVALID_POINT_INDEX) throw new IllegalArgumentException("Invalid subdivision-point index : " + new Integer(pos).toString());

		if (!isClosed()){
			if (ptType!=FIRST_CURVE_END_POINT && ptType!=LAST_CURVE_END_POINT) {
				pts.remove(pos-1); // remove control-point, and shift subsequent points to the left
				pts.remove(pos-1); // remove subdivision-point, and "                             "
				pts.remove(pos-1); // remove other control-point
			}
			else if (ptType==FIRST_CURVE_END_POINT) {// remove first segment
				pts.remove(0); // remove end-point
				pts.remove(0); // remove first control-point
				pts.remove(0); // remove second control-point
			}
			else if (ptType == LAST_CURVE_END_POINT) {// remove last segment
				pts.remove(pts.size()-1); // remove end-point
				pts.remove(pts.size()-1); // remove 2nd control-point
				pts.remove(pts.size()-1); // remove 2nd contorl-point
			}
		}
		else  {// remove first Bezier point (index 0) and its neigbouring points: index 1 and oldNbPts-1.
			if (pos==0){
				pts.set(pts.size()-1, pts.get(2)); pts.remove(2); // move 2nd control-point to end of point list
				pts.remove(1); // remove 1st control point (index 1)
				pts.remove(0); // remove end-point (index 0)
			}
			else {
				pts.remove(pos-1);pts.remove(pos-1);pts.remove(pos-1);
			}
		}
	}

	/**
	 * Removes the last subdivision point. This is a convenience call to <code>removeSubdivisionPoint</code>.
	 */
	public void removeLastSubdivisionPoint() {
		removeSubdivisionPoint(getNumberOfSubdivisionPoints()-1);
	}

	////////////////////////////
	//// OTHER FIELDS ACCESSORS
	////////////////////////////

	/**
	 * Returns the number of segments that make up this curve.  
	 */
	public int getNumberOfSegments() {
		// return isStraightSeg.size();
		if (isClosed()) return getNumberOfSpecificationPoints()/3;
		else return (getNumberOfSpecificationPoints()-1)/3;
	}

	/**
	 * Returns the number of subdivision points (including endpoints if the curve is open)
	 */
	public int getNumberOfSubdivisionPoints() {
		if (isClosed) return getNumberOfSegments();
		else if (getNumberOfSpecificationPoints() == 0) return 0;// special case if curve is EMPTY
		else return getNumberOfSegments() + 1;
	}


	/**
	 * Return a non-normalized vector tangent to the incoming segment (i.e., wrt the control-point that comes before the given subdivision point)
	 * @return a non-normalized vector, or null if the given subdivision point has no incoming tangent because it's the first point of an open curve
	 */
	public PicVector getIncomingTangent(int subdivIndex){
		if (!isClosed() && subdivIndex ==0) return null;
		int ptIndex = segmentToPointIndex(subdivIndex,SUBDIVISION_POINT);
		if (!isValidBezierIndex(ptIndex)) throw new IllegalArgumentException("Invalid subdiv-point index : " + new Integer(subdivIndex).toString());
		int previousControlPointIndex = getPBCBezierIndex(ptIndex-1);
		// check if both control- and subdiv- points are at the same location :
		if (getSpecificationPoint(ptIndex).equals(getSpecificationPoint(previousControlPointIndex))){
			// in this case, tangent is given by the next-to-previous control-point:
			previousControlPointIndex = getPBCBezierIndex(ptIndex-2);
		}
		PicVector v = new PicVector(getSpecificationPoint(ptIndex), getSpecificationPoint(previousControlPointIndex));
		return v;
	}
	
	/**
	 * Return a non-normalized vector tangent to the outcoming segment (i.e., wrt the control-point that comes after the given subdivision point)
	 * @return a non-normalized vector, or null if the given subdivision point has no outcoming tangent because it's the last point of an open curve
	 */
	public PicVector getOutgoingTangent(int subdivIndex){
		if (!isClosed() && subdivIndex ==getNumberOfSubdivisionPoints()-1) return null;
		int ptIndex = segmentToPointIndex(subdivIndex,SUBDIVISION_POINT);
		if (!isValidBezierIndex(ptIndex)) throw new IllegalArgumentException("Invalid subdiv-point index : " + new Integer(subdivIndex).toString());
		int nextControlPointIndex = getPBCBezierIndex(ptIndex+1);
		// check if both control- and subdiv- points are at the same location :
		if (getSpecificationPoint(ptIndex).equals(getSpecificationPoint(nextControlPointIndex))){
			// in this case, tangent is given by the next-to-next control-point:
			nextControlPointIndex = getPBCBezierIndex(ptIndex+2);
		}
		PicVector v = new PicVector(getSpecificationPoint(ptIndex), getSpecificationPoint(nextControlPointIndex));
		return v;
	}

	/**
	 * Returns whether the given segment is straight. This is carried out by checking whether
	 * control-points and subdivision-points are identical at each end of the given segment.
	 * Return false if this segment is reduced to a point.
	 * @param segIndex segment index with respect to the segment numbering scheme, that is, 0 for the first segment, etc... 
	 */
	public boolean isStraight(int segIndex) {
		PicPoint ptLeft = getSpecificationPoint(segmentToPointIndex(segIndex, SUBDIVISION_POINT));
		PicPoint ptCtrl1 = getSpecificationPoint(segmentToPointIndex(segIndex, FIRST_SEGMENT_CONTROL_POINT));
		PicPoint ptCtrl2 = getSpecificationPoint(segmentToPointIndex(segIndex, SECOND_SEGMENT_CONTROL_POINT));
		PicPoint ptRight = getSpecificationPoint(segmentToPointIndex(segIndex+1, SUBDIVISION_POINT));
		if (ptLeft.equals(ptRight)) return false;
		return (ptLeft.equals(ptCtrl1) && ptRight.equals(ptCtrl2));
	}

	/**
	 * Return true if this curve is a polygon, ie has only straight segments
	 */
	public boolean isPolygon(){
		for (int segIdx=0; segIdx < getNumberOfSegments(); segIdx++){
			if (!isStraight(segIdx)) return false;
		}
		return true;
	}
	
	/**
	 * Check whether the two control points around the given subdivision point satisfy (up to 5%) 
	 * the 2nd order smoothness criterion. If this curve is open, and <code>subdivIndex</code> relates
	 * to a curve's end-point, return true. If one of the neighbouring segment is straight, end-points are
	 * used instead of control-points.
	 * @param subdivIndex subdivision-point index, with respect to subdivision-point (SP) numbering scheme, i.e. 
	 *        0,1,2,... for the 1st, 2nd, 3rd... subdivision point.
	 */
	public boolean isSmooth(int subdivIndex) { 
		// new code; takes care of special cases, e.g. control- and subdivision points at the location (=> quadratic curve)
		int ptIndex = segmentToPointIndex(subdivIndex,SUBDIVISION_POINT);
		if (!isValidBezierIndex(ptIndex)) throw new IllegalArgumentException("Invalid point index : " + new Integer(ptIndex).toString());
		
		// if curve is open and subdivIndex points to an end-point, THIS is smooth:
		if (!isValidBezierIndex(ptIndex - 1) || !isValidBezierIndex(ptIndex + 1)) return true;

		// normalized tangents:
		PicVector incomingTg = getIncomingTangent(subdivIndex);
		PicVector outgoingTg = getOutgoingTangent(subdivIndex);
		return Math.abs(incomingTg.det(outgoingTg)) < (.05) * Math.max(incomingTg.norm2(), outgoingTg.norm2());//Returns true if aligned at up to 5%
	}

	/**
	 * Check whether the two control points around the given subdivision point satisfy (up to 5%) 
	 * the 2nd order smoothness criterion. If this curve is open, and <code>subdivIndex</code> relates
	 * to a curve's end-point, return true. If one of the neighbouring segment is straight, end-points are
	 * used instead of control-points.
	 * @param subdivIndex subdivision-point index, with respect to subdivision-point (SP) numbering scheme, i.e. 
	 *        0,1,2,... for the 1st, 2nd, 3rd... subdivision point.
	 */
	 /*
	private boolean _isSmooth(int subdivIndex) { // was "checkSmoothness" [OLD CODE]
		int ptIndex = segmentToPointIndex(subdivIndex,SUBDIVISION_POINT);
		if (!isValidBezierIndex(ptIndex)) throw new IllegalArgumentException("Invalid point index : " + new Integer(ptIndex).toString());
		
		// if curve is open and subdivIndex points to an end-point, THIS is smooth:
		if (!isValidBezierIndex(ptIndex - 1) || !isValidBezierIndex(ptIndex + 1)) return true;

		// if one of neighb. segment is straight :
		int thisSeg = subdivIndex;
		int prevSeg = getPBCSegmentIndex(thisSeg-1);
		if (isStraight(prevSeg))
			ptBuffer1.setCoordinates(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex-3));
		else 
			ptBuffer1.setCoordinates(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex-1));

		if (isStraight(thisSeg))
			ptBuffer2.setCoordinates(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex+3));
		else 
			ptBuffer2.setCoordinates(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex+1));

		return Math.abs(ptBuffer1.det(ptBuffer2)) < (.05) * Math.max(ptBuffer1.norm2(), ptBuffer2.norm2());//Returns true if aligned at up to 5%
	} */

	
	/**
	 * Check whether the control points around the given subdivision point satisfy (up to 5%) the
	 * symmetry criterion. 
	 * @param subdivIndex subdivision-point index (with respect to subdivision-point numbering scheme), that is,
	 *        0,1,2,... for the 1st, 2nd, 3rd,... subdivision point.
	 */
	public boolean isSymmetric(int subdivIndex) { // was "checkSymmetry"
		int ptIndex = segmentToPointIndex(subdivIndex,SUBDIVISION_POINT);
		if (!isValidBezierIndex(ptIndex)) throw new IllegalArgumentException("Invalid point index : " + new Integer(ptIndex).toString());
		if (!isValidBezierIndex(ptIndex - 1) || !isValidBezierIndex(ptIndex + 1)) return true;
		
		int thisSeg = subdivIndex;
		int prevSeg = getPBCSegmentIndex(thisSeg-1);
		if (isStraight(thisSeg) || isStraight(prevSeg)) return true;
		
		// [SR: there was a bug as for sign ?] return Math.abs(vx * vx + vy * vy - wx * wx + wy * wy) < (.05) * (vx * vx + vy * vy);//Returns true if symmetric at up to 5%
		//[old code; didn't care for special cases where both subdiv- and control- points were similar]
		//PicVector v = new PicVector(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex-1));
		//PicVector w = new PicVector(getSpecificationPoint(ptIndex),getSpecificationPoint(ptIndex+1));
		PicVector v = getIncomingTangent(subdivIndex); // non-normalized
		PicVector w = getOutgoingTangent(subdivIndex);
		return Math.abs(v.norm2()-w.norm2()) < .05 * v.norm2();//Returns true if length are equal up to 5%
	}


	/**
	 * close or open this curve, either by opening the last segment if curve is closed,
	 * or by adding two control-points after the last segment if it is open.<br>
	 * This methods doesn't fire any DrawingEvent.
	 * @param state  The new closeness value
	 */
	public void setClosed(boolean state) {
		if (isClosed() == state) return; // no change
		else if (getNumberOfSpecificationPoints() <= 1) {
			isClosed = state;
		}
		else { // now there's a least one segment (3*i pts if closed, 3*i+1 if open)
			isClosed = state;
			
			// close curve -> the policy is to leave the existing points as is, and to add
			// two control-points after the last curve end-point :
			if (state){
				PicPoint pCtrl1 = getSpecificationPoint(pts.size()-2,null); // second control-point of last segment (clone)
				pts.add(pCtrl1.symmetry(getSpecificationPoint(pts.size()-1))); // sym. wrt curve last end-point
				PicPoint pCtrl2 = getSpecificationPoint(1,null); // first control-point of first segment (clone)
				pts.add(pCtrl2.symmetry(getSpecificationPoint(0))); // sym. wrt curve first end-point
			}
			
			// open the curve -> simply remove the last two control-points
			else {
				pts.remove(pts.size()-1);
				pts.remove(pts.size()-1);
				//pts.add(getSpecificationPoint(getNumberOfSpecificationPoints()-1).clone());
			}
		}
	}


	/**
	 * Returns true if this curve is closed.
	 */
	public boolean isClosed() {
		return isClosed;
	}


	/**
	 * Returns the the smallest rectangle enclosing the shape of this curve, as
	 * opposed to {@link #getBoundingBox() getBoundingBox} 
	 * which is defined as the smallest rectangle encompassing ALL 
	 * specification points.
	 * @param r if null, gets allocated and returned for convenience
	 * @see java.awt.Shape.getBounds2D()
	 */
	public Rectangle2D getShapeBounds2D(Rectangle2D r){ // was : getGeometricBoundingBox
		if (r==null) r = new Rectangle2D.Double();
		
		if (getNumberOfSpecificationPoints() == 0) {
			r.setFrameFromDiagonal(0,0,0,0);
			return r;
		}
		if (getNumberOfSegments()==0) {
			r.setFrameFromDiagonal(getSpecificationPoint(0),getSpecificationPoint(0));
			return r;
		}
		double minX,maxX,minY,maxY;
		double[] minMaxX, minMaxY ;
		// curved segment :
		if (! isStraight(0))  
			r = getSegmentShapeBounds2D(getSpecificationPoint(0),getSpecificationPoint(1),getSpecificationPoint(2),getSpecificationPoint(3),r);
		//straight segment :
		else r.setFrameFromDiagonal(getSpecificationPoint(0),getSpecificationPoint(3));

		Rectangle2D rectBuffer = new Rectangle2D.Double(); // [SR:pending] move "rectBuffer" to private class fields ? (i.e. is this method critical in terms of performances ?)
		for (int i=1; i< getNumberOfSegments(); i++){
			//curved segment :
			if (! isStraight(i))  
				rectBuffer = getSegmentShapeBounds2D(getSpecificationPoint(3*i),getSpecificationPoint(3*i+1),getSpecificationPoint(3*i+2),getSpecificationPoint(3*i+3),rectBuffer); 
			// straight segment :
			else rectBuffer.setFrameFromDiagonal(getSpecificationPoint(3*i),getSpecificationPoint(3*i+3));
			r.add(rectBuffer);
		}
		return r;
	}

	/**
	 * Computes a rectangle that bounds the PATH of a single bezier segment. This comes in strong contrast with 
	 * {@link java.awt.geom.GeneralPath.getBounds2D() java.awt.geom.GeneralPath.getBounds2D()}, 
	 * where the smallest rectangle including ALL end- and control-points gets returned, no matter
	 * where the curve's path actually snakes around.
	 * @param a start-point coordinate
	 * @param b first control-point coordinate
	 * @param c second control-point coordinate
	 * @param d end-point coordinate
	 * @param dest if null, gets allocated and returned for convenience
	 * @return the shape's bounding-box of the given segment wrapped in a Rectangle2D
	 */
	private Rectangle2D getSegmentShapeBounds2D(PicPoint a, PicPoint b, PicPoint c, PicPoint d, Rectangle2D dest){
		if (dest==null) dest = new Rectangle2D.Double();
		double[] minMaxX = getSegmentShapeBounds2D(a.x, b.x, c.x, d.x);
		double[] minMaxY = getSegmentShapeBounds2D(a.y, b.y, c.y, d.y);
		dest.setFrameFromDiagonal(minMaxX[0], minMaxY[0], minMaxX[1], minMaxY[1]);
		return dest;
	}
		

	/**
	 * Depending on input coordinates being given on the X- or Y- axis, 
	 * computes the X- or Y-coordinates of the 
	 * rectangle that bounds the PATH of a single bezier segment. This comes in strong contrast with 
	 * {@link java.awt.geom.GeneralPath.getBounds2D() java.awt.geom.GeneralPath.getBounds2D()}, 
	 * where the smallest rectangle including ALL end- and control-points gets returned, no matter
	 * where the curve's path actually snakes around.
	 * @param a start-point coordinate, along X- or Y- axis.
	 * @param b first control-point coordinate, along X- or Y- axis.
	 * @param c second control-point coordinate, along X- or Y- axis.
	 * @param d end-point coordinate, along X- or Y- axis.
	 * @return min and max coordinate of the shape's bounding-box along the axis corresponding to the input coordinates axis. 
	 *         These are stored in an array of double, min given first.
	 */
	private double[] getSegmentShapeBounds2D(double a,double b,double c,double d){
		double min,max;
		// the bezier function is p(t)= a*(1-t)^3+3*b*t*(1-t)^2+3*c*t^2*(1-t)+d*t^3
		// the derivative of the bezier coordinate is A*t^2+2*B*t+C (up to a factor 3)
		if(a<d){min=a; max=d; } else {min=d; max=a; }
		double A=-a+3*b-3*c+d;
		double B=a-2*b+c;
		double C=-a+b;
		double t,p;
		if ( A != 0 ) { //degree is really 2
			double delta=B*B-A*C;
			if (delta>=0) {
				delta=Math.sqrt(delta);
				t=(-B-delta)/A; // t=critical point
				if ( t>0 && t<1 ) {
					p=a*(1-t)*(1-t)*(1-t) + 3*b*(1-t)*(1-t)*t + 3*c*(1-t)*t*t + d*t*t*t ;
					// p= corresponding point (coordinate) on the bezier curve
					if (p<min) { min=p; }
					else if (p>max) { max=p; }
				}
				t=(-B+delta)/A;
				if ( t>0 && t<1 ) {
					p=a*(1-t)*(1-t)*(1-t) + 3*b*(1-t)*(1-t)*t + 3*c*(1-t)*t*t + d*t*t*t ;
					if (p<min) { min=p; }
					else if (p>max) { max=p; }
				}
			}
		}//degree 2
		else if (B != 0) { //degree is really 1
			t=-C/(2*B);
			if ( t>0 && t<1 ) {
				p=a*(1-t)*(1-t)*(1-t) + 3*b*(1-t)*(1-t)*t + 3*c*(1-t)*t*t + d*t*t*t ;
				if (p<min) { min=p; }
				else if (p>max) { max=p; }
			}
		}// degree 1
		//if degree 0, nothing to do
		double[] minMax = new double[2];
		minMax[0]=min;
		minMax[1]=max;
		return minMax;
	}

	/**
	 * Returns a string for debugging purpose.
	 */
	public String toString() {
		String s = super.toString();
		int i;
		s += "\n\tnbSegs=" + getNumberOfSegments();
		s += "\n\tisStraightSeg=";
		for (i = 0; i < getNumberOfSegments(); i++) s+= " " + isStraight(i);
		s += (isClosed ? " closed" : " open");
		return s;
	}

	////////////////////////////////
	//// Action's
	////////////////////////////////

	/**
	 * Creates an array of Action's related to this object
	 *
	 * @param actionDispatcher  dispatches events to the proper PECanvas
	 * @param localizer         i18n localizer for PEAction's
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {

		int N = getNumberOfSpecificationPoints();
		PEAction[] actionArray;

		// [reynal:17/01/2003] test if getNumberOfSpecificationPoints() is a good indicator of "closeability"
		if (N <= 2) {
			actionArray = new PEAction[0];
			return actionArray;
		}
		else {// N>2
			actionArray = new PEAction[1];
			if (isClosed())
				actionArray[0] = new CloseCurveAction(actionDispatcher, CloseCurveAction.OPEN, localizer);
			else
				actionArray[0] = new CloseCurveAction(actionDispatcher, CloseCurveAction.CLOSE, localizer);
			return actionArray;
		}
	}

	/**
	 * Closes the path of this curve
	 * @author    Sylvain Reynal
	 * @since     jpicedt 1.3.3
	 */
	class CloseCurveAction extends PEAction {
		final static String CLOSE = "action.editorkit.CloseCurve";
		final static String OPEN = "action.editorkit.OpenCurve";
		private String type;

		/**
		 * @param type CLOSE or OPEN
		 */
		public CloseCurveAction(ActionDispatcher actionDispatcher, String type, ActionLocalizer localizer) {
			super(actionDispatcher, type, localizer);
			this.type = type;
		}

		public void undoableActionPerformed(ActionEvent e) {
			setClosed(type==CLOSE);
		}
	}
	
	////////////////////////////////
	//// GUI
	////////////////////////////////

	/**
	 * Returns a Customizer for geometry editing
	 */
	public AbstractCustomizer createCustomizer() {
		return new Customizer();
	}


	/**
	 * geometry customizer
	 */
	class Customizer extends AbstractCustomizer implements ActionListener {

		private JCheckBox isClosedCB;
		private boolean isListenersAdded = false;// flag set to true after listener have been registered

		/**
		 * Initializes the GUI. 		 
		 */
		protected Box createPanel(){
			Box p = new Box(BoxLayout.Y_AXIS);
			p.add(isClosedCB = new JCheckBox(jpicedt.Localizer.currentLocalizer().get("action.editorkit.CloseCurve")));
			return p;
		}
		
		/**
		 * add action listeners to widgets to reflect changes immediately
		 */
		private void addActionListeners() {
			if (isListenersAdded)
				return;// already done
			isClosedCB.addActionListener(this);
		}


		/**
		 * load widgets with object's properties
		 */
		public void load() {
			isClosedCB.setSelected(isClosed());
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded
			addActionListeners();// done the first time load is called

		}

		/**
		 * update Element's properties
		 */
		public void store() {
			setClosed(isClosedCB.isSelected());
		}

		public void actionPerformed(ActionEvent e) {
			store();
		}
	}
	
}

