/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.graphic.model;

import jpicedt.graphic.*;


import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.swing.*;
import java.io.*; // for test

/**
 * A class implementing the arrow attribute. This class acts as a model and a view at the same time. A syncAttributes()
 * method allows for synchronizing the current attribute set with the arrow parameters used by the paint() method.
 * <p>
 * This class is also a container for predefined arrows (implemented as static inner classes).
 * @author Sylvain Reynal
 * @since jPicEdt 1.3.1
 * @version $Id: Arrow.java,v 1.9.2.2 2007/09/02 11:56:26 reynal Exp $
 */
public class Arrow implements PicObjectConstants {

	private String name;
	protected double globalScaleWidth, globalScaleLength;
	
	/**
	 * @param name the name (possibly explicit) of this Arrow
	 */
	protected Arrow(String name){
		this.name = name;
	}

	protected Arrow(Arrow src){
		this.name = src.name;
		this.globalScaleWidth = src.globalScaleWidth;
		this.globalScaleLength = src.globalScaleLength;
	}
	
	public Arrow copy(){
		return new Arrow(this);
	}
	
	/** 
	 * Return the name of this arrow.
	 */
	public String getName(){
		return this.name;
	}

	/**
	 * Configure the parameters of this arrow (size, scale, inset, ...) from the
	 * given attribute set, aka view/model synchronization. Note that (according to PsTricks doc), 
	 * attributes apply to both arrows at once for a given element; hence there's no
	 * need to separate left and right attributes.
	 * This implementation simply synchronizes global scale factors (ARROW_SCALE_WIDTH and
	 * ARROW_SCALE_LENGTH), and should be overriden in each concrete subclass.
	 * <p>
	 * Usage: this method should invoked by the view responsible for rendering the drawing, whenever
	 * synchronization is needed b/w the view and the model, i.e. whenever at least one attribute value changed.
	 */
	public void syncAttributes(PicAttributeSet set){
		this.globalScaleLength = StyleConstants.getArrowGlobalScaleLength(set);
		this.globalScaleWidth = StyleConstants.getArrowGlobalScaleWidth(set);
	}
	
	////////////////////////////////////////////////
	/// STATIC METHODS
	////////////////////////////////////////////////

	/**
	 * Returns an integer associated with the given arrow (essentially for JComboxBox's...)
	 * @param arrow one of the predefined Arrow'ss
	 */
	public static int getIndex(Arrow arrow){

		for(int i=0; i< arrowArray.length; i++){
			if (arrow == arrowArray[i]) return i;
		}
		return 0; // should never happen...
	}


	/**
	 * Returns the Arrow corresponding to the given indx
	 * Arrow.NONE if indx > number of arrows
	 */
	public static Arrow getArrow(int indx){

		if (indx > arrowArray.length-1) return Arrow.NONE;
		else return arrowArray[indx];
	}

	/**
	 * Returns an array of predefined arrows
	 */
	public static Arrow[] getPredefinedArrows(){
		return arrowArray;
	}

	/**
	 * build icons for e.g. ComboBoxlists from predefined arrows
	 * @param rightArrow false for left arrow, true for right arrow
	 */
	public static ImageIcon[] createArrowIcons(boolean rightArrow){

		PicAttributeSet set = new PicAttributeSet(); // used to feed arrows with default attribute values
		BufferedImage im;
		ImageIcon[] arrowIcons = new ImageIcon[arrowArray.length];
		double w = 50;
		double h = 32;
		double x = 0.9*w;
		double y = 0.5*h;
		for (int i=0; i<arrowArray.length; i++){
			im = new BufferedImage((int)w,(int)h,BufferedImage.TYPE_INT_RGB);
			Graphics2D g = im.createGraphics();
			g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON);
			if (!rightArrow) g.rotate(Math.PI,w/2,h/2); // mirror around central point
			// fill background with white
			g.setStroke(new BasicStroke(1.0f));
			g.setPaint(Color.white);
			g.fill(new Rectangle2D.Double(0,0,w,h));
			// paint a vertical red line to mark end-point (esp. useful where one's want to know the difference b/w FLUSHED and CENTERED arrows)
			g.setStroke(new BasicStroke(1.0f));
			g.setPaint(Color.red);
			g.draw(new Line2D.Double(x,0,x,h));
			// paint an horizontal black line to figure line's end
			g.setStroke(new BasicStroke((float)(0.8)));
			g.setPaint(Color.black);
			g.draw(new Line2D.Double(0,y,0.9*x,y));
			// paint arrow in black
			g.setPaint(Color.black);
			//	    PicPoint endPoint = new PicPoint(12,8);
			PicPoint endPoint = new PicPoint(x,y);
			PicPoint director = new PicPoint(1,0);
			// 	    g.translate(34.0,8.0);
			// 	    g.scale(2.0,2.0);
			//g.translate(-4.0,-4.0);
			arrowArray[i].syncAttributes(set);
			arrowArray[i].paint(g, 4.0, endPoint, director); // linewidth = 4.0
			// fill image buffer
			arrowIcons[i] = new ImageIcon(im);
		}
		return arrowIcons;
	}

	public static void main(String arg[]){
		
		JFrame f = new JFrame();
		f.getContentPane().setLayout(new BorderLayout(5,5));
		JComboBox cb = new JComboBox(createArrowIcons(true));
		Box box = new Box(BoxLayout.X_AXIS);
		box.add(cb);
		f.getContentPane().add(box);
		f.pack();
		f.setVisible(true);
	}
		


	
	///////////////////////////////////////////////////
	/// PREDEFINED ARROWS (currently based on PsTricks)
	///////////////////////////////////////////////////

	public final static Arrow NONE = new None();
	public final static Arrow ARROW_HEAD = new Head();
	public final static Arrow REVERSE_ARROW_HEAD = new ReverseHead();
	public final static Arrow DOUBLE_ARROW_HEAD = new DoubleHead();
	public final static Arrow DOUBLE_REVERSE_ARROW_HEAD = new DoubleReverseHead();
	public final static Arrow T_BAR_CENTERED = new TBarFlushed(); // centered on end-points
	public final static Arrow T_BAR_FLUSHED = new TBarCentered(); // flush to end-points
	public final static Arrow SQUARE_BRACKET = new SquareBracket();
	public final static Arrow ROUNDED_BRACKET = new RoundedBracket();
	public final static Arrow CIRCLE_FLUSHED = new CircleFlushed();
	public final static Arrow CIRCLE_CENTERED = new CircleCentered();
	public final static Arrow DISK_FLUSHED = new DiskFlushed();
	public final static Arrow DISK_CENTERED = new DiskCentered();
	//     public final static Arrow ROUNDED_END_FLUSH [todo]
	//     public final static Arrow ROUNDED_END_EXTENDED [todo]
	//     public final static Arrow SQUARE_END [todo]

	private final static Arrow[] arrowArray = {
	            NONE,
	            ARROW_HEAD,
	            REVERSE_ARROW_HEAD,
	            DOUBLE_ARROW_HEAD,
	            DOUBLE_REVERSE_ARROW_HEAD,
	            T_BAR_CENTERED,
	            T_BAR_FLUSHED,
	            SQUARE_BRACKET,
	            ROUNDED_BRACKET,
	            CIRCLE_FLUSHED,
	            CIRCLE_CENTERED,
	            DISK_FLUSHED,
	            DISK_CENTERED};




	//////////////////////////////////////////////////
	//// PAINT and related
	//////////////////////////////////////////////////

	// following are PsTricks parameters (see PST doc. page 30 for default values) in "pt" (not in "mm" !)
	// We use PicObjectConstants.PS_POINT to convert in mm ! 
	// so far, they're only used in paint()
	// [pending] should be user tweakable !
	////[deprecated] private static final double SCALE = 1.0*PS_POINT;
	// 1°) ARROW_HEAD and related :
	//protected static final double[] ARROW_SIZE = {2.0*SCALE, 3}; // actual width = ARROW_SIZE[0] + ARROW_SIZE[1] * linewidth
	////[deprecated] protected static final double[] ARROW_SIZE = {1.2*SCALE, 2.1}; // -> because default values as given by the doc don't seem to fit  
	////[deprecated] protected static final double ARROW_LENGTH = 1.4; // actual length (or height) = actual width * ARROW_LENGTH
	////[deprecated] protected static final double ARROW_INSET = 0.4; // actual inset = ARROW_INSET * actual length
	// 2°) TBAR, SQUARE and ROUNDED BRACKETS
	////[deprecated] protected static final double[] TBAR_SIZE = {2.0*SCALE, 5.0}; // actual width = TBAR_SIZE[0] + TBAR_SIZE[1] * linewidth
	////[deprecated] protected static final double BRACKET_LENGTH = 0.15; // actual length = actual width * BRACKET_LENGTH
	////[deprecated] protected static final double R_BRACKET_LENGTH = 0.15; // same but for rounded bracket
	// 3°) CIRCLE and DISK (flushed or not)
	////[deprecated] protected static final double[] DOT_SIZE = {0.5*SCALE, 2.5}; // actual diameter = DOT_SIZE[0] + DOT_SIZE[1] * linewidth
	// 4°) ALL
	////[deprecated] protected static final double[] ARROW_SCALE = {1.0, 1.0}; // scale width by ARROW_SCALE[0] and length by ARROW_SCALE[1]



	/**
	 * Paints this arrow. Convenient call to #paint(Graphics2D, double, double, double, PicPoint).
	 * @param g graphic context
	 * @param linewidth current lineWidth of the object this arrow is attached to (in mm)
	 * @param loc arrow reference point on the screen (that is, line end-point)
	 * @param dir a vector of length 1 indicating the direction of the arrow (see PEToolKit.getDirector...)
	 */
	public void paint(Graphics2D g, double linewidth, PicPoint loc, PicPoint dir){
		paint(g,linewidth,loc.x,loc.y,dir);
	}

	/**
	 * Paints this arrow (this implementation does nothing)
	 * @param g graphic context
	 * @param linewidth current lineWidth of the object this arrow is attached to (in mm)
	 * @param locX X-coord of arrow reference point on the screen (that is, line end-point)
	 * @param locY Y-coord of arrow reference point on the screen (that is, line end-point)
	 * @param dir a vector of length 1 indicating the direction of the arrow (see PEToolKit.getDirector...)
	 */
	public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
	}













	/////////////////////////////////////////////////////
	//// FORMATERS/PARSERS
	/////////////////////////////////////////////////////

	/**
	 * Returns a string representation of this Arrow for debugging purpose
	 */
	public String toString(){
		return name;
	}



	//////////////////////////////////////////////////////////////////////////
	static class None extends Arrow {

		None(){
			super("none"); 
		}
		
		None(None src){
			super(src);
		}
		
		public Arrow copy(){
			return new None(this);
		}
		
		/**
		 * paint this arrow
		 * @param g graphic context
		 * @param linewidth current lineWidth of the object this arrow is attached to (in mm)
		 * @param locX X-coord of arrow reference point on the screen (that is, line end-point)
		 * @param locY Y-coord of arrow reference point on the screen (that is, line end-point)
		 * @param dir a vector of length 1 indicating the direction of the arrow (see PEToolKit.getDirector...)
		 */
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			return;
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class Head extends Arrow {

		private double width_MinimumMm, width_LineWidthScale; // arrow-width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale, insetScale;
		private GeneralPath path = new GeneralPath();
		
		Head(){
			super("head");
			
		}
		
		Head(Head src){
			super(src);
			this.width_MinimumMm = src.width_MinimumMm;
			this.width_LineWidthScale = src.width_LineWidthScale; 
			this.lengthScale = src.lengthScale;
			this.insetScale = src.insetScale;
		}
		
		public Arrow copy(){
			return new Head(this);
		}
			
		
		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			width_MinimumMm = StyleConstants.getArrowWidthMinimumMm(set);
			width_LineWidthScale = StyleConstants.getArrowWidthLineWidthScale(set);
			lengthScale = StyleConstants.getArrowLengthScale(set);
			insetScale = StyleConstants.getArrowInsetScale(set);
		}
		
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			path.reset();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = width_MinimumMm + width_LineWidthScale * linewidth;
			double length = width * lengthScale;
			double inset = length * insetScale;
			width *= globalScaleWidth;
			length *= globalScaleLength;
			inset *= globalScaleLength;
			path.moveTo((float)locX, (float)locY);
			path.lineTo((float)(locX - length*dir.x + width*0.5f*ort.x), (float)(locY - length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(locX - (length-inset)*dir.x), (float)(locY - (length-inset)*dir.y));
			path.lineTo((float)(locX - length*dir.x - width*0.5f*ort.x), (float)(locY - length*dir.y - width*0.5f*ort.y));
			path.closePath();
			g.fill(path);
		}
	}
	
	//////////////////////////////////////////////////////////////////////////
	static class ReverseHead extends Arrow {
		private double width_MinimumMm, width_LineWidthScale; // arrow-width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale, insetScale;
		
		ReverseHead(){
			super("reverse-head");
		}
		
		ReverseHead(ReverseHead src){
			super(src);
			this.width_MinimumMm = src.width_MinimumMm;
			this.width_LineWidthScale = src.width_LineWidthScale; 
			this.lengthScale = src.lengthScale;
			this.insetScale = src.insetScale;
		}
			
		public Arrow copy(){
			return new ReverseHead(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			width_MinimumMm = StyleConstants.getArrowWidthMinimumMm(set);
			width_LineWidthScale = StyleConstants.getArrowWidthLineWidthScale(set);
			lengthScale = StyleConstants.getArrowLengthScale(set);
			insetScale = StyleConstants.getArrowInsetScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = width_MinimumMm + width_LineWidthScale * linewidth;
			double length = width * lengthScale;
			double inset = length * insetScale;
			width *= globalScaleWidth;
			length *= globalScaleLength;
			inset *= globalScaleLength;
			PicPoint loc1 = new PicPoint(locX - length*dir.x, locY - length*dir.y);
			path.moveTo((float)loc1.x, (float)loc1.y);
			path.lineTo((float)(loc1.x + length*dir.x + width*0.5f*ort.x), (float)(loc1.y + length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(loc1.x + (length-inset)*dir.x), (float)(loc1.y + (length-inset)*dir.y));
			path.lineTo((float)(loc1.x + length*dir.x - width*0.5f*ort.x), (float)(loc1.y + length*dir.y - width*0.5f*ort.y));
			path.closePath();
			g.fill(path);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class DoubleHead extends Arrow {
		private double width_MinimumMm, width_LineWidthScale; // arrow-width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale, insetScale;
		
		DoubleHead(){
			super("double-head");
			
		}
		DoubleHead(DoubleHead src){
			super(src);
			this.width_MinimumMm = src.width_MinimumMm;
			this.width_LineWidthScale = src.width_LineWidthScale; 
			this.lengthScale = src.lengthScale;
			this.insetScale = src.insetScale;
		}
			
		public Arrow copy(){
			return new DoubleHead(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			width_MinimumMm = StyleConstants.getArrowWidthMinimumMm(set);
			width_LineWidthScale = StyleConstants.getArrowWidthLineWidthScale(set);
			lengthScale = StyleConstants.getArrowLengthScale(set);
			insetScale = StyleConstants.getArrowInsetScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = width_MinimumMm + width_LineWidthScale * linewidth;
			double length = width * lengthScale;
			double inset = length * insetScale;
			width *= globalScaleWidth;
			length *= globalScaleLength;
			inset *= globalScaleLength;
			PicPoint loc1 = new PicPoint(locX - length*dir.x, locY-length*dir.y);
			path.moveTo((float)locX, (float)locY);
			path.lineTo((float)(locX - length*dir.x + width*0.5f*ort.x), (float)(locY - length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(locX - (length-inset)*dir.x), (float)(locY - (length-inset)*dir.y));
			path.lineTo((float)(locX - length*dir.x - width*0.5f*ort.x), (float)(locY - length*dir.y - width*0.5f*ort.y));
			path.moveTo((float)loc1.x, (float)loc1.y);
			path.lineTo((float)(loc1.x - length*dir.x + width*0.5f*ort.x), (float)(loc1.y - length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(loc1.x - (length-inset)*dir.x), (float)(loc1.y - (length-inset)*dir.y));
			path.lineTo((float)(loc1.x - length*dir.x - width*0.5f*ort.x), (float)(loc1.y - length*dir.y - width*0.5f*ort.y));
			path.closePath();
			g.fill(path);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class DoubleReverseHead extends Arrow {
		private double width_MinimumMm, width_LineWidthScale; // arrow-width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale, insetScale;
		
		DoubleReverseHead(){
			super("double-reverse-head");
			
		}
		DoubleReverseHead(DoubleReverseHead src){
			super(src);
			this.width_MinimumMm = src.width_MinimumMm;
			this.width_LineWidthScale = src.width_LineWidthScale; 
			this.lengthScale = src.lengthScale;
			this.insetScale = src.insetScale;
		}
			
		public Arrow copy(){
			return new DoubleReverseHead(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			width_MinimumMm = StyleConstants.getArrowWidthMinimumMm(set);
			width_LineWidthScale = StyleConstants.getArrowWidthLineWidthScale(set);
			lengthScale = StyleConstants.getArrowLengthScale(set);
			insetScale = StyleConstants.getArrowInsetScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = width_MinimumMm + width_LineWidthScale * linewidth;
			double length = width * lengthScale;
			double inset = length * insetScale;
			width *= globalScaleWidth;
			length *= globalScaleLength;
			inset *= globalScaleLength;
			PicPoint loc1 = new PicPoint(locX - length*dir.x, locY - length*dir.y);
			PicPoint loc2 = new PicPoint(locX - 2.0*length*dir.x, locY - 2.0*length*dir.y);
			path.moveTo((float)loc1.x, (float)loc1.y);
			path.lineTo((float)(loc1.x + length*dir.x + width*0.5f*ort.x), (float)(loc1.y + length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(loc1.x + (length-inset)*dir.x), (float)(loc1.y + (length-inset)*dir.y));
			path.lineTo((float)(loc1.x + length*dir.x - width*0.5f*ort.x), (float)(loc1.y + length*dir.y - width*0.5f*ort.y));
			path.moveTo((float)loc2.x, (float)loc2.y);
			path.lineTo((float)(loc2.x + length*dir.x + width*0.5f*ort.x), (float)(loc2.y + length*dir.y + width*0.5f*ort.y));
			path.lineTo((float)(loc2.x + (length-inset)*dir.x), (float)(loc2.y + (length-inset)*dir.y));
			path.lineTo((float)(loc2.x + length*dir.x - width*0.5f*ort.x), (float)(loc2.y + length*dir.y - width*0.5f*ort.y));
			path.closePath();
			g.fill(path);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class TBarCentered extends Arrow {

		private double width_MinimumMm, width_LineWidthScale; // width = width_MinimumMm + width_LineWidthScale * lineWidth 
		
		TBarCentered(){
			super("t-bar-centered");
		}
		
		TBarCentered(TBarCentered src){
			super(src);
			this.width_MinimumMm=src.width_MinimumMm;
			this.width_LineWidthScale=src.width_LineWidthScale;
		}
		
		public Arrow copy(){
			return new TBarCentered(this);
		}
		
		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			this.width_MinimumMm = StyleConstants.getTBarWidthMinimumMm(set);
			this.width_LineWidthScale = StyleConstants.getTBarWidthLineWidthScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = (width_MinimumMm + width_LineWidthScale * linewidth) * globalScaleWidth;
			g.draw(new Line2D.Double(locX - 0.5*width*ort.x, locY - 0.5*width*ort.y,locX + 0.5*width*ort.x, locY + 0.5*width*ort.y));
			
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class TBarFlushed extends Arrow {

		private double width_MinimumMm, width_LineWidthScale; // width = width_MinimumMm + width_LineWidthScale * lineWidth 

		TBarFlushed(){
			super("t-bar-flushed");
			
		}
		TBarFlushed(TBarFlushed src){
			super(src);
			this.width_MinimumMm=src.width_MinimumMm;
			this.width_LineWidthScale=src.width_LineWidthScale;
		}
		
		public Arrow copy(){
			return new TBarFlushed(this);
		}
		
		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			this.width_MinimumMm = StyleConstants.getTBarWidthMinimumMm(set);
			this.width_LineWidthScale = StyleConstants.getTBarWidthLineWidthScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = (width_MinimumMm + width_LineWidthScale * linewidth) * globalScaleWidth;
			g.draw(new Line2D.Double(locX - 0.5*width*ort.x - 0.5*linewidth*dir.x, locY - 0.5*width*ort.y- 0.5*linewidth*dir.y,
			                         locX + 0.5*width*ort.x- 0.5*linewidth*dir.x, locY + 0.5*width*ort.y- 0.5*linewidth*dir.y));
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class SquareBracket extends Arrow {

		private double width_MinimumMm, width_LineWidthScale; // width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale;
		
		SquareBracket(){
			super("square-bracket");
			
		}
		SquareBracket(SquareBracket src){
			super(src);
			this.width_MinimumMm=src.width_MinimumMm;
			this.width_LineWidthScale=src.width_LineWidthScale;
			this.lengthScale = src.lengthScale;
		}
		
		public Arrow copy(){
			return new SquareBracket(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			this.width_MinimumMm = StyleConstants.getTBarWidthMinimumMm(set);
			this.width_LineWidthScale = StyleConstants.getTBarWidthLineWidthScale(set);
			this.lengthScale = StyleConstants.getBracketLengthScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = (width_MinimumMm + width_LineWidthScale * linewidth) * globalScaleWidth;
			double length = width * lengthScale * globalScaleLength;
			path.moveTo((float)(locX - 0.5f*width*ort.x - (length+linewidth)*dir.x), (float)(locY - 0.5*width*ort.y - (length+linewidth)*dir.y)); //3
			path.lineTo((float)(locX - 0.5f*width*ort.x - 0.5*linewidth*dir.x), (float)(locY - 0.5*width*ort.y- 0.5*linewidth*dir.y)); // 2
			path.lineTo((float)(locX + 0.5f*width*ort.x - 0.5*linewidth*dir.x), (float)(locY + 0.5*width*ort.y- 0.5*linewidth*dir.y));//9
			path.lineTo((float)(locX + 0.5f*width*ort.x - (length+linewidth)*dir.x), (float)(locY + 0.5*width*ort.y - (length+linewidth)*dir.y));//8
			g.draw(path);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class RoundedBracket extends Arrow {

		private double width_MinimumMm, width_LineWidthScale; // width = width_MinimumMm + width_LineWidthScale * lineWidth 
		private double lengthScale;
		
		RoundedBracket(){
			super("rounded-bracket");
			
		}
		RoundedBracket(RoundedBracket src){
			super(src);
			this.width_MinimumMm=src.width_MinimumMm;
			this.width_LineWidthScale=src.width_LineWidthScale;
			this.lengthScale = src.lengthScale;
		}
		
		public Arrow copy(){
			return new RoundedBracket(this);
		}
		
		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			this.width_MinimumMm = StyleConstants.getTBarWidthMinimumMm(set);
			this.width_LineWidthScale = StyleConstants.getTBarWidthLineWidthScale(set);
			this.lengthScale = StyleConstants.getRoundedBracketLengthScale(set);
		}
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			double width = (width_MinimumMm + width_LineWidthScale * linewidth) * globalScaleWidth;
			double length = width * lengthScale * globalScaleLength;
			g.draw(new QuadCurve2D.Double(locX - (length+linewidth)*dir.x - 0.5*width*ort.x,locY - (length+linewidth)*dir.y - 0.5*width*ort.y,
			                              locX,locY,
			                              locX - (length+linewidth)*dir.x + 0.5*width*ort.x,locY - (length+linewidth)*dir.y + 0.5*width*ort.y));
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class CircleFlushed extends Arrow {

		private double dotSize_LineWidthScale, dotSize_MinimumMm; // diam = lineWidth * dotSize_LineWidthScale + dotSize_MinimumMm

		CircleFlushed(){
			super("circle-flushed");
			
		}
		
		CircleFlushed(CircleFlushed src){
			super(src);
			this.dotSize_LineWidthScale = src.dotSize_LineWidthScale; 
			this.dotSize_MinimumMm = src.dotSize_MinimumMm;
		}

		public Arrow copy(){
			return new CircleFlushed(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			dotSize_LineWidthScale = StyleConstants.getPolydotsSizeLineWidthScale(set);
			dotSize_MinimumMm = StyleConstants.getPolydotsSizeMinimumMm(set);
		}
		
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			//double diameter = DOT_SIZE[0] + DOT_SIZE[1] * linewidth;
			double diameter = dotSize_MinimumMm + dotSize_LineWidthScale * linewidth;
			g.draw(new Ellipse2D.Double(locX - 0.5*diameter*(dir.x+1.0), locY - 0.5*diameter*(dir.y+1.0),
			                            diameter,diameter));
		}
	}

	//////////////////////////////////////////////////////////////////////////
	static class CircleCentered extends Arrow {

		private double dotSize_LineWidthScale, dotSize_MinimumMm; // diam = lineWidth * dotSize_LineWidthScale + dotSize_MinimumMm
		
		CircleCentered(){
			super("circle-centered");
			
		}
		
		CircleCentered(CircleCentered src){
			super(src);
			this.dotSize_LineWidthScale = src.dotSize_LineWidthScale; 
			this.dotSize_MinimumMm = src.dotSize_MinimumMm;
		}

		public Arrow copy(){
			return new CircleCentered(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			dotSize_LineWidthScale = StyleConstants.getPolydotsSizeLineWidthScale(set);
			dotSize_MinimumMm = StyleConstants.getPolydotsSizeMinimumMm(set);
		}
		
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			//double diameter = DOT_SIZE[0] + DOT_SIZE[1] * linewidth;
			double diameter = dotSize_MinimumMm + dotSize_LineWidthScale * linewidth;
			g.draw(new Ellipse2D.Double(locX - diameter*0.5, locY - diameter*0.5,diameter,diameter));
		}
	}
	//////////////////////////////////////////////////////////////////////////
	static class DiskFlushed extends Arrow {
		private double dotSize_LineWidthScale, dotSize_MinimumMm; // diam = lineWidth * dotSize_LineWidthScale + dotSize_MinimumMm

		DiskFlushed(){
			super("disk-flushed");
			
		}
		
		DiskFlushed(DiskFlushed src){
			super(src);
			this.dotSize_LineWidthScale = src.dotSize_LineWidthScale; 
			this.dotSize_MinimumMm = src.dotSize_MinimumMm;
		}

		public Arrow copy(){
			return new DiskFlushed(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			dotSize_LineWidthScale = StyleConstants.getPolydotsSizeLineWidthScale(set);
			dotSize_MinimumMm = StyleConstants.getPolydotsSizeMinimumMm(set);
		}
		
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			//double diameter = DOT_SIZE[0] + DOT_SIZE[1] * linewidth;
			double diameter = dotSize_MinimumMm + dotSize_LineWidthScale * linewidth;
			g.fill(new Ellipse2D.Double(locX - 0.5*diameter*(dir.x+1.0), locY - 0.5*diameter*(dir.y+1.0),
			                            diameter,diameter));
		}
	}
	//////////////////////////////////////////////////////////////////////////
	static class DiskCentered extends Arrow {
		private double dotSize_LineWidthScale, dotSize_MinimumMm; // diam = lineWidth * dotSize_LineWidthScale + dotSize_MinimumMm

		DiskCentered(){
			super("disk-centered");
			
		}
		
		DiskCentered(DiskCentered src){
			super(src);
			this.dotSize_LineWidthScale = src.dotSize_LineWidthScale; 
			this.dotSize_MinimumMm = src.dotSize_MinimumMm;
		}

		public Arrow copy(){
			return new DiskCentered(this);
		}

		public void syncAttributes(PicAttributeSet set){
			super.syncAttributes(set);
			dotSize_LineWidthScale = StyleConstants.getPolydotsSizeLineWidthScale(set);
			dotSize_MinimumMm = StyleConstants.getPolydotsSizeMinimumMm(set);
		}
		
		
		public void paint(Graphics2D g, double linewidth, double locX, double locY, PicPoint dir){
			PicPoint ort = new PicPoint(-dir.y, dir.x); // orthogonal unitary vector
			// init path for filled shape
			GeneralPath path = new GeneralPath();
			g.setStroke(new BasicStroke((float)(linewidth)));
			//double diameter = DOT_SIZE[0] + DOT_SIZE[1] * linewidth;
			double diameter = dotSize_MinimumMm + dotSize_LineWidthScale * linewidth;
			g.fill(new Ellipse2D.Double(locX - diameter*0.5, locY - diameter*0.5,diameter,diameter));
		}
	}
	
}
