/*
 BranchElement.java - February 10, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.view.ViewFactory;
import jpicedt.graphic.event.DrawingEvent;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * An abstract class for Element's that allow children. The geometry specification mostly delegates to
 * the children of this Element, except for the bounding-box, which is backed by an array of coordinates.
 * <p>
 * The implementation of the Collection interface is done through calls to 
 * <code>addChild</code>, <code>removeChild</code>, <code>children</code>, 
 * <code>removeAllChildren</code>, except for the <code>addAll</code> method which is implemented from
 * scratch so as to fire a changed-event only once all elements have been added.
 * <br> 
 * As a result, subclasses only need to override these methods
 * to change the behaviour of content-modifying methods.
 * <p>
 * This element has only one control point, namely, the bottom-left corner of the bounding box.
 * Moving this point result in a global translation of the element (however no rescale capability is available here
 * using the <code>setPoint()</code> method, use the PicGroup class instead). 
 *
 * @since jPicEdt
 * @author Sylvain Reynal
 */
public abstract class BranchElement extends AbstractElement implements Collection {

	/** the array that contains children */
	protected ArrayList children = new ArrayList();
	
	/** a semaphor that signals a change of state is underway in this BranchElement and
	 * it shouldn't process events coming from its children before the change is completed 
	 * (hence this is used inside {@link #forwardChangedUpdate forwardChangedUpdate()});
	 * set it to true each time you start modifying children <b>in batch mode</b> and you don't want
	 * events to be forwarded to the root of the hierarchy before everything is completed (e.g. to
	 * get rid of side-effects, or to reduce the burden for the repaint manager).<p>
	 * Example of use : whenever we call <code>translate()</code> on this BranchElement,
	 * <ul>
	 * <li> this first raises the changeLock semaphor, 
	 * <li> then calls translate() on every children :
	 * this in turn make children forward a DrawingEvent to their parent (= this BranchElement), yet this event is trapped here, 
	 * hence not propagated upward ; note however that children update their own view whatever the value of the changeLock semaphor ;  
	 * <li> then lower the changeLock semaphor, and fire a change-update event (which in turn will propagate the event upward
	 *      along the tree).
	 * </ul>
	 */
	protected boolean changeLock;
	
	private static final int P_ANCHOR = 0; // index of the only control point
	
	/** an array of X-coordinates representing the two opposite corner of the bounding box that contains all children */ 
	protected double[] ptsX;
	/** an array of Y-coordinates representing the two opposite corner of the bounding box that contains all children */ 
	protected double[] ptsY;
	
	/** buffer used by updateBoundingBox */
	private Rectangle2D bbBuffer = new Rectangle2D.Double();

	
	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * construct a BranchElement with no parent and a default PicAttributeSet
	 */
	public BranchElement(){
		super(); 
		ptsX = new double[2];
		ptsY = new double[2];
	}

	/**
	 * construct a BranchElement with no parent and the given PicAttributeSet
	 */
	public BranchElement(PicAttributeSet attributeSet){
		this();
		this.attributeSet = new PicAttributeSet(attributeSet);
	}

	/**
	 * cloning constructor (though with no parent and no view) : 
	 * 1°) attribute set is deeply copied.
	 * 2°) source's children are cloned, then the copy is added to this BranchElement
	 */
	public BranchElement(BranchElement src){
		super(src);
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("cloning " + src);
		ptsX = new double[2];
		ptsY = new double[2];
		for(Iterator i = src.children(); i.hasNext(); ){
			Element obj = (Element)i.next();
			this.addChild((Element)obj.clone());
		}
		// ensure end-points are in the same order as the source (addChild might've modified it)
		ptsX[0] = src.ptsX[0];
		ptsY[0] = src.ptsY[0];
		ptsX[1] = src.ptsX[1];
		ptsY[1] = src.ptsY[1];
	}

	/**
	 * create a new BranchElement from the content of the given Collection of Element's.
	 * Children are cloned as well (i.e. this isa deep copy).
	 */
	public BranchElement(Collection c){
		this();
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("from collection " + c);
		for(Iterator i = c.iterator(); i.hasNext(); ){
			Element obj = (Element)i.next();
			this.addChild((Element)obj.clone());
		}
	}

	/**
	 * @return a clone of this element
	 */
	public abstract Object clone();

	



	/////////////////////////////////////////////////////////
	//// OPERATIONS ON CONTENT
	/////////////////////////////////////////////////////////

	/**
	 * Returns true if the receiver allows children. This implementation returns true.
	 */
	public boolean getAllowsChildren(){
		return true;
	}
	
	/**
	 * @return the child at the given index, or null if no children
	 */
	public Element getChildAt(int childIndex){
		return (Element)children.get(childIndex);
	}

    /**
     * @return the number of children for the element.
     */
	public int getChildCount(){
		return children.size();
	}


	/**
	 * @return an Iterator over children.
	 */
	public Iterator children(){
		return children.iterator();
	}

	/**
	 * Returns the index of the given Element in the receivers children.
	 * If the receiver does not contain the given Element, -1 will be returned.
	 */
	public int getChildIndex(Element node) {
		return children.indexOf(node);
	}
	
	/**
	 * @return whether this BranchElement contains the given child
	 */
	public boolean contains(Element o){
		return children.contains(o);
	}

	/**
	 * add the given child to this BranchElement, setting its parent to this.
	 * Since a child can't have two parents, this also remove the child from its former parent, if any.
	 * <br>
	 * If view is non-null, use the associated ViewFactory to create View's for the new child.
	 */
	public void addChild(Element child){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("addChild : " + child);
		Element oldParent = child.getParent();
		if (oldParent != null && oldParent instanceof BranchElement) ((BranchElement)oldParent).removeChild(child);
		child.setParent(this);
		children.add(child);
		// create View for child
		if (view != null){
			ViewFactory f = view.getViewFactory();
			child.setViewFromFactory(f);
		}
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.INSERT); 
	}

	/**
	 * insert the given child in this BranchElement at the given position, setting its parent to this,
	 * and create a view for the child using the ViewFactory that produced the View for this
	 * BranchElement.
	 */
	public void addChild(int position, Element child){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("addChild at pos " + position + " : " + child);
		Element oldParent = child.getParent();
		if (oldParent != null && oldParent instanceof BranchElement) ((BranchElement)oldParent).removeChild(child);
		child.setParent(this);
		children.add(position,child);
		// create View for child
		if (view != null){
			ViewFactory f = view.getViewFactory();
			child.setViewFromFactory(f);
		}
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.INSERT); 
	}

	/**
	 * replace the child at the given position by the given child
	 */
	public void replaceChild(int position, Element newChild){
		Element oldChild = getChildAt(position);
		oldChild.removeView(); // remove view for old child if any
		oldChild.setParent(null); // make eligible for gc
		Element oldParent = newChild.getParent();
		if (oldParent != null && oldParent instanceof BranchElement) ((BranchElement)oldParent).removeChild(newChild);
		newChild.setParent(this);
		children.set(position,newChild);
		// create View for child
		if (view != null){
			ViewFactory f = view.getViewFactory();
			newChild.setViewFromFactory(f);
		}
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.REPLACE); 
	}

	/**
	 * remove the given child from this BranchElement, setting its parent to null so that
	 * the given child become (possibly) eligible for garbage collection if there are no other
	 * reference to it.
	 */
	public void removeChild(Element child){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeChild : " + child);
		child.removeView();// remove view for child if any
		child.setParent(null);
		children.remove(child);
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.REMOVE); 
	}

	/**
	 * remove all children from this BranchElement, set each child's parent to null.
	 */
	public void removeAllChildren(){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeAllChildren");
		for(Iterator it=children(); it.hasNext(); ){
			Element removed = (Element)it.next();
			removed.removeView(); // remove view for child if any
			removed.setParent(null);
		}
		children.clear();
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.REMOVE); 
	}

	
	
	
	
	////////////////////////////////////////////////////////////////
	//// GEOMETRY
	////////////////////////////////////////////////////////////////
	
	/**
	 * @return the index of the first point that can be retrieved by getPoint
	 * This returns PT_ANCHOR.
	 */
	public int getFirstPointIndex(){
		return P_ANCHOR;
	}
	
	/**
	 * @return the index of the last point that can be retrieved by getPoint
	 * This returns PT_ANCHOR.
	 */
	public int getLastPointIndex(){
		return P_ANCHOR;
	}

	/**
	 * @return the X-coord of the point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointX(int numPoint){
		if (numPoint == P_ANCHOR) return Math.min(ptsX[0],ptsX[1]);
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());

	}

	/**
	 * This default implementation returns <code>ptsY[numPoint]</code>.This might be a valid implementation
	 * as long as subclasses don't have other control points.
	 * @return the Y-coord of the point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointY(int numPoint){
		if (numPoint == P_ANCHOR) return Math.min(ptsY[0],ptsY[1]);
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}
	
	/**
	 * Return the <b>user-controlled point</b> having the given index. The general contract in Element
	 * is to return an IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method, 
	 * hence this implementation simply calls getPointX() and getPointY().
	 * @return the point indexed by <code>numPoint</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public PicPoint getPoint(int numPoint, PicPoint src){
		if (src==null) src = new PicPoint();
		src.x = getPointX(numPoint);
		src.y = getPointY(numPoint);
		return src;
	}

	/**
	 * Set the user-controlled point with the given index to the given value.
	 * This default implementation simply call setPoint with a null constraint, then fires a change-update of
	 * type GEOMETRY_CHANGE
	 * @deprecated use setPoint(int, PicPoint, EditPointConstraint) instead.
	 */
	public void setPoint(int index, PicPoint pt){
		setPoint(index,pt,null);
	}
	
	/**
	 * Set the point indexed by "numPoint" to the given value.
	 */
	public void setPoint(int numPoint, PicPoint pt, EditPointConstraint c){
		if (numPoint == P_ANCHOR) translate(pt.x-getPointX(P_ANCHOR),pt.y-getPointY(P_ANCHOR)); // fire change event
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}

	/**
	 * Translate children by the given vector
	 * @param dx The X coordinate of translation vector
	 * @param dy The Y coordinate of translation vector
	 * @since jPicEdt
	 */
	public void translate(double dx, double dy){
		// smart bounding-box update :
		ptsX[0] += dx;
		ptsY[0] += dy;
		ptsX[1] += dx;
		ptsY[1] += dy;

		changeLock = true; // filter out event from children
		for(Iterator i = children.iterator(); i.hasNext();){
			Element o = (Element)i.next();
			o.translate(dx,dy); // children call forwardChangeUpdate via their fireChangedUpdate
			// but this has no effect since changeLock = true
		}
		changeLock = false; 
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); 
	}

	/**
	 * Scale children by (sx,sy) using ptOrg as origin ; sx and sy can be negative.
	 */
	public void scale(PicPoint ptOrg, double sx, double sy){
		scale(ptOrg.x, ptOrg.y, sx, sy);
	}
	
	/**
	 * Scale children by (sx,sy) using (ptOrgX,ptOrgY) as origin ; sx and sy can be negative.
	 */
	public void scale(double ptOrgX, double ptOrgY, double sx, double sy){
		// smart bounding-box update :
		ptsX[0] = ptOrgX + sx * (ptsX[0] - ptOrgX);
		ptsY[0] = ptOrgY + sy * (ptsY[0] - ptOrgY);
		ptsX[1] = ptOrgX + sx * (ptsX[1] - ptOrgX);
		ptsY[1] = ptOrgY + sy * (ptsY[1] - ptOrgY);
		// scale children :
		changeLock = true; // block event from children
		for(Iterator i = children.iterator(); i.hasNext();){
			Element o = (Element)i.next();
			o.scale(ptOrgX,ptOrgY,sx,sy); 
		}
		changeLock = false; // unblock event from children
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		// rotate children :
		changeLock = true; // block incoming event from children
		for(Iterator i = children.iterator(); i.hasNext();){
			Element o = (Element)i.next();
			o.rotate(ptOrg,angle); 
		}
		changeLock = false; // unblock event from children
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		// shear children :
		changeLock = true; // block incoming event from children
		for(Iterator i = children.iterator(); i.hasNext();){
			Element o = (Element)i.next();
			o.shear(ptOrg,shx,shy); 
		}
		changeLock = false; // unblock event from children
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
		
	
	/////////////////////////////////////
	/// EVENTS
	////////////////////////////////////

	/**
	 * Called by a child of this BranchElement to inform its parent of some change that occured to it or one of its children.
	 * This gives a chance to the receiver to update its layout, then to propagate the change-event upward.
	 * <br>
	 * This implementation update the bounding-box, then fire a changedUpdate event ONLY if changeLock is false.
	 * Otherwise, does nothing (this is the exact semantic behind "changeLock").
	 * @param eventType the event type
	 * @param child 
	 */
	public void forwardChangedUpdate(Element child,DrawingEvent.EventType eventType){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType+", child="+child.toString());
		if (changeLock) return;  // if a change is underway here, and it's driven by THIS element,
		// we block incoming events until the change is completed ; the method which led the change
		// should obviously fire an event itself afterwards.
		updateBoundingBox();
		fireChangedUpdate(eventType); // update view then forward (see superclass)
	}

	
	/////////////////////////////////
	/// PAINT
	/////////////////////////////////

	/**
	 * set the view for this Element from the given view factory, then set the view for children.
	 */
	public void setViewFromFactory(ViewFactory f){
		super.setViewFromFactory(f);
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("factory="+f);
		// set views for children
		for (Iterator it=children(); it.hasNext(); ){
			Element o = (Element)it.next();
			o.setViewFromFactory(f);
		}
	}

	/**
	 * remove the view that render this element and propagate to children ; 
	 * this may be used to remove any reference to the view,
	 * and render it eligible for garbage collection ; if no View, does nothing.
	 */
	public void removeView(){
		super.removeView();
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeView");
		// remove views for children
		for (Iterator it=children(); it.hasNext(); ){
			Element o = (Element)it.next();
			o.removeView();
		}
	}

	
	//////////////////////////////////////////////////
	//// CHILD DISPOSITION
	//////////////////////////////////////////////////

	/**
	 * Move the given child to back (i.e. following z-ordering policy), 
	 * i.e. removes it from its current position and insert it at position 0
	 * <br>
	 * Does nothing if the given child can't be found in this Drawing or is already to back.
	 */
	public void bringToBack(Element obj){

		int index=getChildIndex(obj);
		if (index!=-1 && index>0){ // if not already at back
			children.remove(index);
			children.add(0,obj); // insert at bottom
		}
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); 
	}


	/**
	 * Move the given child to front, i.e. removes it from its current position and appends to the drawing.<br>
	 * Does nothing if the given obj can't be found in this Drawing or if it's already to front.
	 */
	public void bringToFront(Element obj){

		int index=getChildIndex(obj);
		if (index!=-1 && index<getChildCount()-1){ // if foudn and not already at front
			children.remove(index);
			children.add(obj); // add at top
		}
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); 
	}

	/**
	 * Move the given child one position backward, i.e. removes it from its current position and insert it one position backward.<br>
	 * Does nothing if the given obj can't be found in this Drawing, or if it's already to back.
	 */
	public void bringBackward(Element obj){

		int index=getChildIndex(obj);
		if (index>0){ // if not already at back
			children.remove(index);
			children.add(index-1,obj);
		}
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); // repaint !
	}

	/**
	 * Move the given child one position forward, i.e. removes it from its current position and insert it one position forward.<br>
	 * Does nothing if the given obj can't be found in this Drawing, or if it's already to front.
	 */
	public void bringForward(Element obj){

		int index=getChildIndex(obj);
		if (index!=-1 && index<getChildCount()-1){ // if found and not already at front
			children.remove(index);
			children.add(index+1,obj);
		}
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); 
	}

	/**
	 * @return true if the given object has the lowest z-value of all objects in this drawing.
	 */
	public boolean isToBack(Element obj){
		return getChildIndex(obj)==0;
	}

	/**
	 * @return true if the given object has the highest z-value of all objects in this drawing.
	 */
	public boolean isToFront(Element obj){
		return getChildIndex(obj)==getChildCount()-1;
	}
	









	///////////////////////
	/// BOUNDING BOX
	///////////////////////

	/**
	 * Returns the bounding box (i.e. the surrounding rectangle) in double precision
	 * Used e.g. to determine the arguments of the \\begin{picture} command.<p>
	 * Straigthforwardly computed from the diagonal
	 * @return null if this BranchElement has no children
	 * @since PicEdt 1.0
	 */
	public Rectangle2D getBoundingBox(Rectangle2D r){
		if (isEmpty()) return null; // [SR:underway] test pour pb CompositeView
		if (r==null) r = new Rectangle2D.Double();
		r.setFrameFromDiagonal(ptsX[0],ptsY[0],ptsX[1],ptsY[1]);
		return r;
	}
	
	/**
	 * Update the bounding box by "unioning" the children's bounding box (actually computes the
	 * two specification points defining the diagonal of the box). 
	 * Should be called each time the content of this element gets modified : this may include 
	 * addition or deletion of a child, and modification of a child by the child itself.
	 */
	protected void updateBoundingBox(){
		// this method use "bbBuffer"
		if (children.isEmpty()) {
			ptsX[0]=ptsY[0]=ptsX[1]=ptsY[1]=0;
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("empty");
		}
		else {
			Rectangle2D bb;
			double minX, minY, maxX, maxY;
			// initializes "bound" from first object, else we catch a (0,0) inside it !!!
			bb = ((Element)children.get(0)).getBoundingBox(bbBuffer); // use private class buffer 
			minX=bb.getMinX();
			minY=bb.getMinY();
			maxX=bb.getMaxX();
			maxY=bb.getMaxY();
			for(int i=1; i<children.size(); i++){
				Element e = (Element)children.get(i);
				e.getBoundingBox(bb);
				maxX=Math.max(bb.getMaxX(),maxX);
				minX=Math.min(bb.getMinX(),minX);
				maxY=Math.max(bb.getMaxY(),maxY);
				minY=Math.min(bb.getMinY(),minY);
			}
			ptsX[0] = minX;
			ptsY[0] = minY;
			ptsX[1] = maxX;
			ptsY[1] = maxY;
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug(getBoundingBox(null).toString()+"\n\tthis="+this);
		}
	}
	







	////////////////////////////
	//// OTHER FIELDS ACCESSORS
	////////////////////////////

	/**
	 * set AttributeSet for this BranchElement and propagate to children
	 */
	public void setAttributeSet(PicAttributeSet set){
		changeLock = true;
		for(Iterator i = children.iterator(); i.hasNext();){
			((Element)i.next()).setAttributeSet(set);
		}
		changeLock = false;
		super.setAttributeSet(set); // fire changed event
	}

	/**
	 * set an Attribute for this BranchElement and propagate to children
	 */
	public void setAttribute(PicAttributeName name, Object value){
		changeLock = true;
		for(Iterator i = children.iterator(); i.hasNext();){
			((Element)i.next()).setAttribute(name,value);
		}
		changeLock = false;
		super.setAttribute(name,value); // fire event
	}


	/**
	 * Returns a String representing the group for debugging use only.
	 */
	public String toString(){

		String s = super.toString() + "\n\tchildren={";
		int j=0;
		for(Iterator i = children.iterator(); i.hasNext(); j++){
			StringBuffer buf = new StringBuffer(((Element)i.next()).toString());
			int pos=0;
			while(true){
				pos = buf.toString().indexOf('\n',pos);
				if (pos == -1) break;
				buf.insert(pos+1,'\t');
				pos += 2;
			}
			s += "\n\t" + j + ":" + buf.toString();
		}
		return s + "\n\t}]";
	}
	

	///////////////////////////////////////////////
	//// Implementation of the Collection interface
	///////////////////////////////////////////////
	
	/**
	 * add the given object as a new child of this BranchElement, but only 
	 * if it's an instance of Element, and return true. Return false otherwise
	 */
	public boolean add(Object o){
		if (o instanceof Element){
			addChild((Element)o);
			return true;
		}
		return false;
	}
	
	/**
	 * adds all the elements in the given collection that are instance of Element 
	 * as children of this BranchElement.
	 */
	public boolean addAll(Collection c){
		boolean changed = false;
		for(Iterator it = c.iterator(); it.hasNext();){
			Object o = it.next();
			if (o instanceof Element){
				// same as addChild((Element)o), but fire only at the end :
				Element child = (Element)o;
				Element oldParent = child.getParent();
				if (oldParent != null && oldParent instanceof BranchElement) ((BranchElement)oldParent).removeChild(child);
				child.setParent(this);
				children.add(child);
				// create View for child
				if (view != null){
					ViewFactory f = view.getViewFactory();
					child.setViewFromFactory(f);
				}
				changed = true;
			}
		}
		if (changed) {
			updateBoundingBox();
			fireChangedUpdate(DrawingEvent.EventType.INSERT);
		}
		return changed;
	}
	
	/**
	 * remove all the children
	 */
	public void clear(){
		removeAllChildren();
	}

    /**
     * @return same as getChildCount
     */
    public int size(){
		return children.size();
	}

    /**
     * @return <tt>true</tt> if this BranchElement contains no children
     */
    public boolean isEmpty(){
		return children.isEmpty();
	}

    /**
     * Returns <tt>true</tt> if this collection contains the specified
     * element. 
     */
    public boolean contains(Object o){
		return children.contains(o);
	}

    /**
     * Returns an iterator over children.
     */
    public Iterator iterator(){
		return children();
	}

    /**
     * Returns an array containing all of the elements in this collection.       */
    public Object[] toArray(){
		return children.toArray();
	}

    /**
     * Returns an array containing all of the elements in this collection
     * whose runtime type is that of the specified array.  
     */
    public Object[] toArray(Object a[]){
		return children.toArray(a);
	}

    /**
     * Removes a single instance of the specified Element from this
     * BranchElement, if it is present AND is instance of Element.
     * @param o element to be removed from this BranchElement, if present.
     * @return <tt>true</tt> if this BranchElement changed as a result of the call
     */
    public boolean remove(Object o){
		if (o instanceof Element){
			Element e = (Element)o;
			if (contains(e)){
				removeChild(e);
				return true;
			}
		}
		return false;
	}

    /**
     * Returns <tt>true</tt> if this BranchElement contains all of the elements
     * in the specified collection.
     */
    public boolean containsAll(Collection c){
		Iterator e = c.iterator();
		while (e.hasNext())
	    	if(!contains(e.next()))
				return false;
		return true;
	}

    /**
     * Removes all this collection's elements that are also contained in the
     * specified collection.  After this call returns,
     * this collection will contain no elements in common with the specified
     * collection.
     * @param c elements to be removed from this collection.
     * @return <tt>true</tt> if this collection changed as a result of the call
     */
    public boolean removeAll(Collection c){
		boolean modified = false;
		Iterator it = iterator();
		while (it.hasNext()) {
	    	if(c.contains(it.next())) {
				it.remove(); // call remove, which in turns call removeChild and fire changed event.
				modified = true;
	    	}
		}
		return modified;
	}
		

    /**
     * Retains only the elements in this BranchElement that are contained in the specified collection
     * @param c elements to be retained in this collection.
     * @return true if this collection changed as a result of the call
     */
    public boolean retainAll(Collection c){
		boolean modified = false;
		Iterator it = iterator();
		while (it.hasNext()) {
	    	if(!c.contains(it.next())) {
			it.remove(); // call remove, which in turns call removeChild and fire changed event. 
			modified = true;
	    	}
		}
		return modified;
	}

    /**
     * @return <tt>true</tt> if the specified object is equal to this collection
     */
    public boolean equals(Object o){
		return (o == this);
	}
	
	////////////////////////////////////////////////////////////////
	//// MISC
	////////////////////////////////////////////////////////////////
	
	/**
	 * Returns a list containing children of this BranchElement that are of the same type or
	 * inherit the given clazz.
	 * @since jpicedt 1.4pre5
	 */
	public ArrayList createFilteredCollection(Class clazz){
		ArrayList l = new ArrayList();
		for (Iterator it = children(); it.hasNext();){
			Object o = it.next();
			if (clazz.isAssignableFrom(o.getClass())) l.add(o);
		}
		return l;
	}
	
	/**
	 * Returns whether this BranchElement contains objects that are of the same type of inherit from the
	 * given clazz.
	 * @since jpicedt 1.4pre5
	 */
	public boolean containsClass(Class clazz){
		for (Iterator it = children(); it.hasNext();){
			Object o = it.next();
			if (clazz.isAssignableFrom(o.getClass())) return true;
		}
		return false;
	}
	
}
