/*
 DefaultAttributeSet.java - March 30, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import java.awt.*;
import java.util.*;

/**
 * This an immutable attribute set that stores default attributes value, and can thus serve as
 * the default resolving parent for mutable PicAttributeSet's. The <code>setAttribute()</code> method
 * is overriden so as to do nothing. This set can't have any resolving parent.<p>
 * This is how to add a new default attribute pair :
 * <ul>
 * <li>In <code>PicObjectConstants</code>, define a new static final PicAttributeName 
 *     with an explicit name (possibly following XML usual syntax, that is with hyphen-separated words)</li>
 * <li>Possibly define corresponding attribute values, if this makes sense (e.g. predefinite String's)</li>
 * <li>Initialize the pair in the constructor just below</li>
 * <p>
 * Refactored 30/03/2002 (hashtable)
 *
 */
/* package private */ class DefaultAttributeSet extends PicAttributeSet implements PicObjectConstants {

	/**
	 * Create a new PicAttributeSet with default values conforming to PsTricks default values, i.e. :
	 * <ul>
	 * <li>Line : style = solid, colours = black, width = Options.lineThickness</li>
	 * <li>Dash (transparent and opaque) = predefinite PicObjectConstants.DASH1</li>
	 * <li>Dot sep = predefinite PicObjectConstants.DOT_SEP1</li>
	 * <li>Double line : none, sep = 1.25*Options.lineThickness, colour = white</li>
	 * <li>Shadow : none, size=3pt, angle=-45, colour=dark gray</li>
	 * <li>Dimen = outer</li>
	 * <li>Fill : style = none, colour = black</li>
	 * <li>Hatch : width = 0.8pt, sep=4pt, colour=black, angle=45</li>
	 * <li>Arrows : left = no, right = no</li>
	 * </ul>
	 *
	 * @see jpicedt.graphic.model.PicObjectConstants
	 */
	/* package private */ DefaultAttributeSet(){

		super();
		parent=null;
		
		// Note that all values are given in "mm" when applicable ; besides, PS_POINT (i.e. 1pt) is defined in PicObjectConstants.
		// These default values mimick those of PsTricks as much as possible, except that some rounding is applied now and often when the corresponding
		// value would have too many digits once expressed in mm.
		
		double lineThickness = 0.3; // ~ 0.8pt
		
		_setAttribute(LINE_STYLE, SOLID);
		_setAttribute(LINE_COLOR, Color.black);
		_setAttribute(LINE_WIDTH, new Double(lineThickness));
		
		_setAttribute(DASH_TRANSPARENT, DASH1);
		_setAttribute(DASH_OPAQUE, DASH1);
		_setAttribute(DOT_SEP,DOT_SEP1);
		
		_setAttribute(DOUBLE_LINE, Boolean.FALSE);
		_setAttribute(DOUBLE_SEP,new Double(0.4)); // 0.4mm
		_setAttribute(DOUBLE_COLOR, Color.white);
		
		_setAttribute(SHADOW,Boolean.FALSE);
		_setAttribute(SHADOW_SIZE,new Double(1.0)); // 1mm
		_setAttribute(SHADOW_ANGLE, new Double(-45.0));
		_setAttribute(SHADOW_COLOR, Color.darkGray);
		
		_setAttribute(DIMEN, OUTER);
		
		_setAttribute(FILL_STYLE,NONE);
		_setAttribute(FILL_COLOR,Color.black);
		
		_setAttribute(HATCH_WIDTH, new Double(0.3)); // .3mm
		_setAttribute(HATCH_SEP, new Double (1.5));// 1.5mm
		_setAttribute(HATCH_COLOR, Color.black);
		_setAttribute(HATCH_ANGLE, new Double(45));
		
		// see PsTricks's doc p.30 for arrow default values:
		_setAttribute(LEFT_ARROW, Arrow.NONE);
		_setAttribute(RIGHT_ARROW,Arrow.NONE);
		_setAttribute(ARROW_GLOBAL_SCALE_WIDTH, new Double(1.0));
		_setAttribute(ARROW_GLOBAL_SCALE_LENGTH, new Double(1.0));
		_setAttribute(ARROW_WIDTH_MINIMUM_MM, new Double(1)); 
		_setAttribute(ARROW_WIDTH_LINEWIDTH_SCALE, new Double(2));
		_setAttribute(ARROW_LENGTH_SCALE, new Double(1));
		_setAttribute(ARROW_INSET_SCALE, new Double(0.25));
		_setAttribute(TBAR_WIDTH_MINIMUM_MM, new Double(.7)); // ~ 2pt
		_setAttribute(TBAR_WIDTH_LINEWIDTH_SCALE, new Double(5.0));
		_setAttribute(BRACKET_LENGTH_SCALE, new Double(0.15));
		_setAttribute(RBRACKET_LENGTH_SCALE, new Double(0.15));
		// note also that dot parameters for arrows (i.e. circles and disks) are the same as for polydots; hence see below at POLYDOTS_SIZE_XXXX
		
		_setAttribute(POLYDOTS_STYLE,NONE);
		_setAttribute(POLYDOTS_SUPERIMPOSE,Boolean.FALSE);
		_setAttribute(POLYDOTS_SIZE_LINEWIDTH_SCALE,new Double(2.5)); // 2.5 * line-width + minimum ; see pstricks doc p.30
		_setAttribute(POLYDOTS_SIZE_MINIMUM_MM,new Double(0.7));  // 2pt ~ 0.7mm (and PsTricks doc contains a mistake for that matter...)
		_setAttribute(POLYDOTS_SCALE_H,new Double(1));
		_setAttribute(POLYDOTS_SCALE_V,new Double(1));
		_setAttribute(POLYDOTS_ANGLE,new Double(0));
		
		_setAttribute(OVER_STRIKE, Boolean.FALSE);
		_setAttribute(OVER_STRIKE_WIDTH, new Double(lineThickness));
		_setAttribute(OVER_STRIKE_COLOR, Color.WHITE);
		
		_setAttribute(PST_CUSTOM,"");
		
		_setAttribute(TEXT_VERT_ALIGN,TEXT_VALIGN_CENTER);
		_setAttribute(TEXT_HOR_ALIGN,TEXT_HALIGN_CENTER);
		_setAttribute(TEXT_FRAME,TEXT_BOX_NO_FRAME);
		_setAttribute(TEXT_ROTATION, new Double(0));
	}

	/**
	 * Overriden to do nothing since this is an immutable set.
	 */
	public void setAttribute(PicAttributeName name, Object value){
	}

	/**
	 * Convenience initialization method
	 * @param name attribute name
	 * @param value attribute value
	 */
	private void _setAttribute(PicAttributeName name, Object value){
		// check that value.getClass() is assignable from name's allowedValueClass, i.e. value has the expected runtime type :
		if (!name.getAllowedValueClass().isInstance(value)) throw new IllegalArgumentException("Illegal value class: "+value.getClass()+ " when "+name.getAllowedValueClass()+" was expected !");
		
		// check if name exist in this set :
		if (map.containsKey(name)) {
			// if name/value is not already defined here, replace old value with new one :
			if (!map.get(name).equals(value)) {
				map.put(name,value);
			}
			// otherwise leave unchanged
		}
		// otherwise add new name/value pair :
		else {
			map.put(name,value);
		}
	}
	
	/**
	 * @return the attribute with the given name, or null if the attribute name is not defined here. 
	 */
	public Object getAttribute(PicAttributeName name){
		if (map.containsKey(name)) return map.get(name);
		else return null;
	}

	/**
	 * set the resolving parent. Overriden to do nothing.
	 * @param parent if null, set the parent to DEFAULT_SET.
	 */
	public void setResolveParent(PicAttributeSet parent){
	}

	/**
	 * @return the resolving parent, i.e. null.
	 */
	public PicAttributeSet getResolveParent(){
		return null;
	}
	
	/**
	 * return this set (since it's immutable, a copy just makes no sense ; use new PicAttributeSet(this)
	 * if you really want to clone this set).
	 */
	public PicAttributeSet copyAttributes() {
		return this;
	}

	/** debug */
	public static void main(String[] args){
		DefaultAttributeSet set = new DefaultAttributeSet();
		set._setAttribute(LINE_STYLE, new Double(1.0));
		System.out.println(set);
		
	}

} // class
