/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.view.*;
import jpicedt.graphic.event.DrawingEvent;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;


/**
 * This interface specifies the general contract for models of graphic elements that may be added to a Drawing.<p>
 * A <b>MODEL</b> of graphic element basically comprises :
 * <ul>
 * <li> A set of (possibly virtual, see implementation hints below) user-controlled points 
 *      that helps the "user" alter the geometry
 *      of the element ; these points are made accessible by means of the various get/setPoint() methods,
 *      using an index ranging from <code>getFirstPointIndex()</code> to <code>getLastPointIndex()</code> inclusive. 
 *      These methods may be used e.g. by the UI machinery when processing mouse-events, or by parsers when 
 *      building a new Drawing from scratch.
 * <li> Amid this set, some points may serve as <b>anchor-points</b> for grid alignment (see <code>PointIterator</code>). 
 * <li> An attribute set that allows binding the element's geometry to a set of colour, stroke, etc... attributes.
 * <li> A parent and possibly children once the element has been added to a tree defining a Drawing's model. This allows
 *      us to let the root element only hold state about the tree while still maintaining the ability for
 *      a child to retrieve this state at any time through a (possibly recursive) call to getParent.
 * <li> Besides, each Element may have a View attached to it (but this is not necessary as long as it has not
 *      to be rendered), which is responsible for rendering the element to
 *      a graphic context. View's are usually created by a ViewFactory that's suited for the kind of content the
 *      model represents (e.g. PostScript, LaTeX, JPIC-XML, ...).
 * </ul>
 * <b>Implementation hints</b> : under the hood, the geometry may be backed e.g. by an array of points 
 * (in which case these may be the control-points as well),
 * by a <code>java.awt.Shape</code>, or by any other kind of mechanism the developper thinks is appropriate 
 * to the particular model she wants to implement.  
 * </ul>
 * <p>
 *
 * @author Sylvain Reynal
 * @since PicEdt 1.0
 */
public interface Element extends PicObjectConstants {


	/**
	 * Return a non-localised string representing this element's name. This may be used by a UI to display some
	 * information related to this element, or by a localizer to fetch a i18n'd string.
	 */
	String getName();


 	/**
 	 * Return a deep copy of this element.
 	 */
 	Object clone();

	
	///////////////////////////////////////
	//// tree structure
	///////////////////////////////////////
	
    /**
     * Retrieves the underlying drawing
     * @return the drawing ; null if this Element doesn't belong to any drawing yet.
     */
	Drawing getDrawing();

    /**
     * Gets the parent of the element.
     * @return the parent
     */
	Element getParent(); 

    /**
     * Sets the parent of the element.
     */
	void setParent(Element p);
	
	/**
	 * Returns true if the receiver allows children.
	 */
	boolean getAllowsChildren();

	/**
	 * Called by a child of this element to inform its parent of some change that occured to it or one of its children.
	 * This gives a chance to the receiver to update its layout, then to propagate the change-event upward.
	 * This method obviously makes sense only if this element allows children (else it will never be called anyway, 
	 * hence concrete implementation may do nothing here).
	 * @param eventType the event type
	 * @param child the child that sent the change-event. 
	 */
	void forwardChangedUpdate(Element child,DrawingEvent.EventType eventType);
		
	////////////////////////////////////
	//// GEOM.
	///////////////////////////////////

	/**
	 * Return the <b>user-controlled point</b> having the given index. The general contract is to return an
	 * IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method.
	 * @return the point indexed by <code>index</code> ; 
	 *         if <code>src</code> is null, concrete implementation of this method should
	 *         allocate a new PicPoint and return it, 
	 *         otherwise directly modify <code>src</code> and return it as well for convenience.
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 * [todo] change name to getControlPoint() or getUserPoint() or whatever seems more appropriate.
	 */
	PicPoint getPoint(int index, PicPoint src);

	/**
	 * Same as <code>getPoint()</code>, yet return the x-coordinate only.
	 * @return the X-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	double getPointX(int index);

	/**
	 * Same as <code>getPoint()</code>, yet return the y-coordinate only.
	 * @return the Y-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	double getPointY(int index);

	/**
	 * Return the index of the first user-controlled point that can be retrieved by getPoint()
	 */
	int getFirstPointIndex();
	
	/**
	 * Return the index of the last user-controlled point that can be retrieved by getPoint()
	 */
	int getLastPointIndex();
	
	/**
	 * Set the user-controlled point indexed by "index" to the given value.
	 * This should be a convenience call to setPoint(index,pt,null), i.e. using no particular geometrical constraint.
	 * @deprecated Use setPoint(int, PicPoint, EditPointConstraint)
	 */
	void setPoint(int index, PicPoint pt);

	/**
	 * Set the user-controlled point indexed by "index" to the given value, using the specified geometrical constraint.
	 * Constraints depend on the particular concrete implementation, and may involve restricting movement along
	 * a particular direction, moving several points at once to preserve parallelism,...
	 * @param constraint a geometry constraint, or null if no particular constraint is being imposed (aka default).
	 */
	void setPoint(int index, PicPoint pt, EditPointConstraint constraint);

	/**
	 * Return an Iterator over user-controlled point indexes that can serve as anchor points for grid alignment. 
	 * This is up to concrete implementations to decide what subset of user-controlled points
	 * is most appropriate for grid alignment.
	 */
	PointIndexIterator anchorPointsIterator();

	/**
	 * Translate this object by (dx,dy)
	 * @param dx The X coordinate of translation vector
	 * @param dy The Y coordinate of translation vector
	 *
	 * @since PicEdt 1.0
	 * [todo] add an "apply(Transform t)" method as replacement to "translate()" and "scale()", where Transform
	 *        may possibly inherit from java.awt.geom.AffineTransform [syd : 5/8/2002]
	 */
	void translate(double dx, double dy);

	/**
	 * Scale this object by (sx,sy) using ptOrg as origin
	 * sx and sy may be negative. This can be implemented as a convenience call to
	 * scale(double,double,double,double)
	 */
	void scale(PicPoint ptOrg, double sx, double sy);

	/**
	 * Scale this object by (sx,sy) using (ptOrgX,ptOrgY) as origin
	 * sx and sy can be negative.
	 */
	void scale(double ptOrgX, double ptOrgY, double sx, double sy);

	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	void rotate(PicPoint ptOrg, double angle);
	
	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	void shear(PicPoint ptOrg, double shx, double shy);
	
	/////////////////////////////////
	/// BOUNDING BOX
	/////////////////////////////////

	/**
	 * @return the bounding box (i.e. the surrounding rectangle) in double precision
	 * Used e.g. to determine the arguments of the \\begin{picture} command.
	 * If r is null, allocate a new rectangle and returns it. Otherwise the source rectangle is
	 * modified and returned for convenience.
	 * [todo:reynal] this really need to be improved : this method would probably better be moved to the 
	 *               attached view, since the latter knows exactly what the TRUE bounding box is.
	 * @since PicEdt 1.0
	 */
	Rectangle2D getBoundingBox(Rectangle2D r);
	
	/////////////////////////////////
	/// PAINT
	/////////////////////////////////

	/**
	 * @return the View that's responsible for rendering this Element
	 */
	View getView();
	
	/**
	 * set the view for this Element from the given view factory
	 */
	void setViewFromFactory(ViewFactory f);

	/**
	 * remove the view that render this element ; this may be used to remove any reference to the view,
	 * and render it eligible for garbage collection ; if no View, does nothing.
	 */
	void removeView();


	/////////////////////////////////
	/// VARIOUS FIELDS MANIPULATORS
	/////////////////////////////////

	/**
	 * @return AttributeSet for this Element
	 */
	PicAttributeSet getAttributeSet(); 

	/**
	 * @param attributeSet a new AttributeSet for this Element ; this actually make a deep copy of the
	 *        given attribute set beforehands.
	 */
	void setAttributeSet(PicAttributeSet attributeSet); 

	/**
	 * set the given attribute name to the given value for this Element
	 */
	void setAttribute(PicAttributeName name, Object value);

	/**
	 * @return the value for the given attribute name
	 */
	Object getAttribute(PicAttributeName name);



} // Element interface

