/*
 PicAttributeSet.java - March 30, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import java.awt.*;
import java.util.*;

/**
 * a class that stores non-default graphical attributes (e.g. "fill", "stroke", etc... ) in
 * a hashtable. If a key is not found in the map, we look up the resolving parent.
 */
public class PicAttributeSet implements PicObjectConstants {

	/** an immutable attribute set that serves as the resolving parent if no resolving parent
	    is explicitly set */
	public static final DefaultAttributeSet DEFAULT_SET = new DefaultAttributeSet();
	
	/** a map holding name/value pairs */
	protected HashMap map; 
	/** if non-null, used as the resolving parent when name is not found in current set */
	protected PicAttributeSet parent;

	/**
	 * Create a new PicAttributeSet with DEFAULT_SET as the resolving parent.
	 * @see jpicedt.graphic.model.PicObjectConstants
	 */
	public PicAttributeSet(){
		map = new HashMap();
		parent = DEFAULT_SET;
	}

	/**
	 * cloning constructor
	 * @param src the attribute set to be cloned ; set the resolving parent to the resolving parent
	 *        of the source.
	 */
	public PicAttributeSet(PicAttributeSet src){
		this();
		Iterator it = src.getAttributeNames();
		while(it.hasNext()){
			PicAttributeName name = (PicAttributeName)it.next();
			setAttribute(name, src.getAttribute(name));
		}
		setResolveParent(src.getResolveParent());
	}

	/**
	 * Returns the resolving parent, or DEFAULT_SET if it's null.
	 */
	public PicAttributeSet getResolveParent(){
		if (this.parent == null) return DEFAULT_SET;
		else return this.parent;
	}
	
	/**
	 * set the resolving parent.
	 * @param parent if null, set the parent to DEFAULT_SET.
	 */
	public void setResolveParent(PicAttributeSet parent){
		if (parent==null) this.parent=DEFAULT_SET;
		else this.parent = parent;
		removeRedundantAttributes(); // remove name/value pairs defined in the new resolving parent
	}
	
	/**
	 * return a copy of this PicAttributeSet with the same resolving parent as the source, 
	 * and which is guaranteed not to change over time.
	 */
	public PicAttributeSet copyAttributes() { 
		return new PicAttributeSet(this);
	}

	/**
	 * Returns the attribute with the given name, possibly resolving to parent if "name" is not
	 * defined in this set.
	 */
	public Object getAttribute(PicAttributeName name){
		if (map.containsKey(name)) return map.get(name);
		else return getResolveParent().getAttribute(name);
	}

	/**
	 * Remove name/value pair already defined in the resolving parent.
	 */
	private void removeRedundantAttributes(){
		Iterator it = getAttributeNames();
		while(it.hasNext()){
			PicAttributeName name = (PicAttributeName)it.next();
			if (parent.getAttribute(name)!=null){
				if (parent.getAttribute(name).equals(map.get(name)))
					map.remove(name);
			}
		}
	}

	/**
	 * Add a new name/value pair to the set if it's present nor in the set, nor in the resolving parent,
	 * otherwise modify the value for the given name. If the pair then exists in the resolving parent,
	 * remove it from this set.
	 * @param name attribute name
	 * @param value attribute value
	 * @throws IllegalArgumentException if value's runtime type is not an allowed class for the given attribute name
	 * @see jpicedt.graphic.model.PicObjectConstants
	 */
	public void setAttribute(PicAttributeName name, Object value){
		// check that value.getClass() is assignable from name's allowedValueClass, i.e. value has the expected runtime type :
		if (!name.getAllowedValueClass().isInstance(value)) throw new IllegalArgumentException("Illegal value class: "+value.getClass()+ " when "+name.getAllowedValueClass()+" was expected !");
		
		//jpicedt.Log.debug(name+"="+value);
		// check if name/value exist in resolving parent :
		if (getResolveParent().getAttribute(name)!=null && getResolveParent().getAttribute(name).equals(value)){
			// remove from hashmap, (hence we'll resolve to parent on next call to getAttribute):
			//jpicedt.Log.debug("map.remove("+name+")");
			map.remove(name);
		}
		// else check if name exist in this set :
		else if (map.containsKey(name)) {
			// if name/value is not already defined here, replace old value with new one :
			if (!map.get(name).equals(value)) {
				map.put(name,value);
				//jpicedt.Log.debug("map.put("+name+")");
			}
			// otherwise leave unchanged
		}
		// otherwise add new name/value pair :
		else {
			map.put(name,value);
			//jpicedt.Log.debug("map.put("+name+")");
		}
	}

	/**
	 * Sets several name/value pairs at a time from the given set. Key present in this set, yet which
	 * are not found in the given set, aren't overriden. Hence only non-default value are copied (in the
	 * sense that the resolving parent is the "default" set).
	 */
	public void setAttributes(PicAttributeSet src){
		Iterator it = src.getAttributeNames();
		while(it.hasNext()){
			PicAttributeName name = (PicAttributeName)it.next();
			Object value = src.getAttribute(name);
			setAttribute(name, value);
		}
	}
	
	/**
	 * Returns the number of attributes contained in this set ; this doesn't include the attributes
	 *         of the resolving parent.
	 */
	public int getAttributeCount(){
		return map.size();
	}
	
	/**
	 * Returns an iterator over the names of the attributes in this attribute set ; this
	 *         doesn't iterate over the keys of the resolving parent.
	 */
	public Iterator getAttributeNames(){
		Set keys = map.keySet();
		return keys.iterator();
	}

	/**
	 * Returns a text representation of this attribute set, for debugging purpose
	 */
	public String toString() {

		StringBuffer buf = new StringBuffer(100);
		buf.append("attrib_set={");
		// sort keys in ascending order :
		Comparator c = new Comparator(){
			public int compare(Object o1, Object o2){
				return o1.toString().compareTo(o2.toString());
			}
		};
		TreeSet sortedKeys = new TreeSet(c);
		sortedKeys.addAll(map.keySet());
		Iterator it = sortedKeys.iterator();
		while(it.hasNext()){
			PicAttributeName name = (PicAttributeName)it.next();
			buf.append("*");
			buf.append(name.toString());
			buf.append("=");
			buf.append(map.get(name));
		}
		buf.append("}");
		//buf.append("\n\tresolving_parent_set="); 
		//buf.append(getResolveParent().toString());
		//buf.append(getResolveParent());
		return buf.toString();
	}

} // class
