/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.widgets.*;
import jpicedt.graphic.view.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A concrete implementation of BranchElement that is suited for building groups of elements.
 * It allows easy global resize/translate transformation by providing 8 control points that
 * can be used by mouse operation.
 *
 * @since PicEdt 1.0
 * @author Sylvain Reynal
 */
public class PicGroup extends BranchElement implements CustomizerFactory, ActionFactory{

	////////////////////////////
	//// PUBLIC CONSTANT FIELDS
	////////////////////////////

	// warning ! these are mnemonics only : keep in mind that UL can becom LR after a call to setPoint...
	public static final int FIRST_PT = 0;
	public static final int UL = 0; // U = upper M = mid L = lower or left R = right
	public static final int UM = 1;
	public static final int UR = 2;
	public static final int ML = 3;
	public static final int MR = 4;
	public static final int LL = 5;
	public static final int LM = 6;
	public static final int LR = 7;
	public static final int LAST_PT = 7;




	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * construct a PicGroup with no parent and a default PicAttributeSet
	 */
	public PicGroup(){
		super();
	}

	/**
	 * construct a PicGroup with no parent and the given PicAttributeSet
	 */
	public PicGroup(PicAttributeSet attributeSet){
		super(attributeSet);
	}

	/**
	 * Cloning constructor.
	 * Simply call superclass's constructor.
	 */
	public PicGroup(BranchElement src){
		super(src);
	}

	/**
	 * Simply call superclass's constructor.
	 */
	public PicGroup(Collection c){
		super(c);
	}

	/**
	 * Return a deep copy of this PicGroup
	 * @since PicEdt 1.0
	 */
	public Object clone(){
		return new PicGroup(this);
	}

	/**
	 * @return a localised string that represents this object's name
	 */
	public String getName(){
		return jpicedt.Localizer.currentLocalizer().get("model.Group");
	}


	/////////////////////////////////////////////////////////
	//// OPERATIONS ON CONTENT (aka ArrayList implementation)
	/////////////////////////////////////////////////////////

	// inherited

	//////////////////////////////////
	//// EVENTS HANDLING
	//////////////////////////////////

	// inherited

	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////

	/**
	 * @return the index of the first point that can be retrieved by getPoint
	 * This default implementation returns 0.
	 */
	public int getFirstPointIndex(){
		return FIRST_PT;
	}

	/**
	 * @return the index of the last point that can be retrieved by getPoint
	 * This default implementation returns the greater index allowed in ptsX (or ptsY).
	 */
	public int getLastPointIndex(){
		return LAST_PT;
	}

	/**
	 * @return the X-coord of the point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointX(int numPoint){
		if (numPoint==UL || numPoint==ML || numPoint==LL) return ptsX[0];
		if (numPoint==UM || numPoint==LM) return 0.5*(ptsX[0]+ptsX[1]);
		if (numPoint==UR || numPoint==MR || numPoint==LR) return ptsX[1];
		throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}

	/**
	 * @return the Y-coord of the point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointY(int numPoint){
		if (numPoint==UL || numPoint==UM || numPoint==UR) return ptsY[1];
		if (numPoint==ML || numPoint==MR) return 0.5*(ptsY[0]+ptsY[1]);
		if (numPoint==LL || numPoint==LM || numPoint==LR) return ptsY[0];
		throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}

	/**
	 * Set the point numbered by "numPoint" to the "pt" value, and scales inward object accordingly
	 * @param constraint not used here, may be safely set to null
	 * @since PicEdt 1.0
	 */
	public void setPoint(int numPoint, PicPoint ptNew, EditPointConstraint constraint){

		double ptOrgX, ptOrgY;
		double ptOldX = getPointX(numPoint);
		double ptOldY = getPointY(numPoint);
		switch (numPoint){
		case UL :
			ptOrgX = getPointX(LR);ptOrgY = getPointY(LR);
			break;
		case UM :
			ptOrgX = getPointX(LL); ptOrgY = getPointY(LL); // LR would do as well
			ptOldX = ptNew.x; // sx=1
			break;
		case UR: //
			ptOrgX = getPointX(LL);ptOrgY = getPointY(LL);
			break;
		case ML:
			ptOrgX = getPointX(UR);ptOrgY = getPointY(UR); // LR would do as well
			ptOldY = ptNew.y; // sy=1
			break;
		case MR:
			ptOrgX = getPointX(UL);ptOrgY = getPointY(UL);// LL would do as well
			ptOldY = ptNew.y; // sy=1
			break;
		case LL:
			ptOrgX = getPointX(UR);ptOrgY = getPointY(UR);
			break;
		case LM:
			ptOrgX = getPointX(UL);ptOrgY = getPointY(UL); // UR would do as well
			ptOldX = ptNew.x; // sx=1
			break;
		case LR:
			ptOrgX = getPointX(UL);ptOrgY = getPointY(UL);
			break;
		default:
			throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
		}
		if (ptOldX == ptOrgX || ptOldY == ptOrgY) return; // division by zero
		if (ptNew.x == ptOrgX || ptNew.y == ptOrgY) return; // if we scale by 0, everything gets reduced to a point...
		scale(ptOrgX,ptOrgY,(ptNew.x - ptOrgX)/(ptOldX - ptOrgX), (ptNew.y - ptOrgY)/(ptOldY - ptOrgY)); // fire changed event
	}


	/**
	 * @return an Iterator over points that can serve as anchor points for grid alignment. 
	 *         This implementation returns a default PointIterator which simply iterates through
	 *         control points.
	 */
	public PointIndexIterator anchorPointsIterator(){
		return new GroupPointIterator();
	}

	/**
	 * An iterator over PicPoint's that can serve as anchors for grid alignment. 
	 * This implementation iterates over the 4 control points : UL, UR, LL and LR.
	 */
	public class GroupPointIterator implements PointIndexIterator {

		protected int counter;
		private int[] anchors = {UL, UR, LL, LR};

		/**
		 * create a new PointIterator
		 */
		public GroupPointIterator(){
			super();
			counter = 0;
		}

		public void reset(){
			counter=0;
		}
		/**
		 * @return true if the iteration has more elements
		 */
		public boolean hasNext(){
			if (counter > 3) return false;
			return true;
		}

		/**
		 * @return the next PicPoint in the iteration ; this doesn't create a new PicPoint, hence the
		 *         returned value might change over time.
		 */
		public int next(){
			counter++;
			return anchors[counter-1];
		}
	}

	///////////////////////////////
	//// STRING FORMATING
	///////////////////////////////

	/**
	 * Returns a String representing the group for debugging use only.
	 */
	public String toString(){

		String s = "[PicGroup@" + Integer.toHexString(hashCode()) + "{";
		int j=0;
		for(Iterator i = children(); i.hasNext(); j++){
			StringBuffer buf = new StringBuffer(((Element)i.next()).toString());
			int pos=0;
			while(true){
				pos = buf.toString().indexOf('\n',pos);
				if (pos == -1) break;
				buf.insert(pos+1,'\t');
				pos += 2;
			}
			s += "\n\t" + j + ":" + buf;
		}
		return s + "\n}]";
	}





	////////////////////////////////
	//// GUI
	////////////////////////////////

	/**
	 * Creates an array of Action's related to this object
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {
		
		PEAction toggleHighlightingModeAction = null, editSingleElementGeometryAction = null;
		
		if (view !=null && view.getHighlighter()!=null && view.getHighlighter() instanceof DefaultHighlighterFactory.CompositeHighlighter){
			toggleHighlightingModeAction = new ToggleHighlightingModeAction(actionDispatcher, localizer);
		}

		if (hi!=null){
			Element clickedChild = hi.getTarget();
			if (clickedChild == this){ // ie if HI.Composite
				if (hi instanceof HitInfo.Composite){ // assert true !
					HitInfo.Composite hic = (HitInfo.Composite)hi;
					clickedChild = hic.getClickedChild();
					if (clickedChild != null && clickedChild instanceof CustomizerFactory)
						editSingleElementGeometryAction = new EditSingleElementGeometryAction(actionDispatcher, localizer, (CustomizerFactory)clickedChild);
				}
			}
			else  { // hi = HitInfo.Point, Stroke, etc... for the group content
				clickedChild = hi.getTarget();
				if (clickedChild != null && clickedChild instanceof CustomizerFactory)
					editSingleElementGeometryAction = new EditSingleElementGeometryAction(actionDispatcher, localizer, (CustomizerFactory)clickedChild);
			}
		}
			

		int actionCount=0;
		if (toggleHighlightingModeAction != null) actionCount++;
		if (editSingleElementGeometryAction != null) actionCount++;
		PEAction[] actionArray = new PEAction[actionCount];
		
		int actionIndex=0;
		if (toggleHighlightingModeAction != null){
			actionArray[actionIndex++] = toggleHighlightingModeAction; 
		}
		if (editSingleElementGeometryAction != null){
			actionArray[actionIndex++] = editSingleElementGeometryAction; 
		}
		return actionArray;
	}

	/**
	 * Toggle the attached View's highlightingMode b/w LOCAL_MODE and GLOBAL_MODE
	 * @see jpicedt.graphic.view.DefaultHighlighterFactory
	 */
	class ToggleHighlightingModeAction extends PEAction {
		
		public static final String KEY = "action.editorkit.ToggleGroupHighlightingMode";

		public ToggleHighlightingModeAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer) {
			super(actionDispatcher, KEY, localizer);
		}

		public void actionPerformed(ActionEvent e) {
			Highlighter h = view.getHighlighter();
			if (h instanceof DefaultHighlighterFactory.CompositeHighlighter){
				DefaultHighlighterFactory.CompositeHighlighter hc = (DefaultHighlighterFactory.CompositeHighlighter)h;
				String mode = hc.getHighlightingMode();
				if (mode==hc.LOCAL_MODE) hc.setHighlightingMode(hc.GLOBAL_MODE);
				else  hc.setHighlightingMode(hc.LOCAL_MODE);
			}
		}
	}

	/**
	 * Open a dialog that allows the user to edit the geometry of the given element
	 */
	class EditSingleElementGeometryAction extends PEAction {
		
		public static final String KEY = "action.editorkit.EditSingleElementGeometry";
		private CustomizerFactory clickedChild;

		public EditSingleElementGeometryAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer, CustomizerFactory clickedChild) {
			super(actionDispatcher, KEY, localizer);
			this.clickedChild = clickedChild;
		}

		public void actionPerformed(ActionEvent e) {
			getCanvas().beginUndoableUpdate((String)getValue(NAME)); // i18n'd if there was a localizer provided in the contructor
			AbstractCustomizer pane = clickedChild.createCustomizer();
			int buttonsMask = CustomizerDialog.OK_CANCEL_BUTTONS | CustomizerDialog.RELOAD_BUTTON; 
			CustomizerDialog dlg = new CustomizerDialog(pane, buttonsMask);
			if (clickedChild instanceof TextEditable) dlg.setOkButtonClosesDialog(true);
			else dlg.setOkButtonClosesDialog(false);
			dlg.setVisible(true);
			boolean isCancelled = dlg.isCancelled();
			getCanvas().endUndoableUpdate();
			if (isCancelled) {
				try {
					getCanvas().undo();
				}
				catch (javax.swing.undo.CannotUndoException ex){
					//if (jpicedt.Log.DEBUG) 
					ex.printStackTrace();
				}
			}
		}
	}

	/**
	 * @return a Customizer for geometry editing
	 */
	public AbstractCustomizer createCustomizer(){
		return new Customizer();
	}

	
	/**
	 * geometry customizer
	 */
	class Customizer extends AbstractCustomizer implements ActionListener {

		private DecimalNumberField groupBLCornerXTF, groupBLCornerYTF, groupWidthTF, groupHeightTF;
		private boolean isListenersAdded = false; // flag

		public Customizer(){

			super();

			JPanel p = new JPanel(new GridLayout(3,3,5,5));

			// line 1: bottom left corner label
			p.add(PEToolKit.createJLabel("attributes.RectangleBLCorner"));
			// x-coordinate of ll-corner
			p.add(groupBLCornerXTF = new DecimalNumberField(4));
			// y-coordinate of ll-corner
			p.add(groupBLCornerYTF = new DecimalNumberField(4));

			// line 2: group width
			p.add(PEToolKit.createJLabel("attributes.RectangleWidth"));
			p.add(groupWidthTF = new DecimalNumberField(4));
			p.add(new JLabel());

			// line 3: group height
			p.add(PEToolKit.createJLabel("attributes.RectangleHeight"));
			p.add(groupHeightTF = new DecimalNumberField(4));
			p.add(new JLabel());

			// cyclic focus TAB
			groupHeightTF.setNextFocusableComponent(groupBLCornerXTF);
			add(p, BorderLayout.NORTH);
		}

		/** add action listeners to widgets to reflect changes immediately */ 
		private void addActionListeners(){
			if (isListenersAdded) return; // already done
			groupBLCornerXTF.addActionListener(this);
			groupBLCornerYTF.addActionListener(this);
			groupWidthTF.addActionListener(this);
			groupHeightTF.addActionListener(this);
			isListenersAdded = true;
		}


		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){
			groupBLCornerXTF.setValue(getPointX(PicGroup.LL));
			groupBLCornerYTF.setValue(getPointY(PicGroup.LL));
			groupWidthTF.setValue(getBoundingBox(null).getWidth());
			groupHeightTF.setValue(getBoundingBox(null).getHeight());
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded 
			addActionListeners(); // done the first time load is called
		}

		/**
		 * update Element's properties
		 */
		public void store(){
			double width = groupWidthTF.getValue();
			double height = groupHeightTF.getValue();
			double llCornerX = groupBLCornerXTF.getValue();
			double llCornerY = groupBLCornerYTF.getValue();
			PicPoint llPtFrame = new PicPoint(llCornerX, llCornerY);
			PicPoint urPtFrame = new PicPoint(llCornerX + width, llCornerY+height);
			setPoint(PicGroup.LL,llPtFrame,null);
			setPoint(PicGroup.UR,urPtFrame,null);
		}

		public void actionPerformed(ActionEvent e){
			store();
		}
		
		/**
		 * @return the panel title, used e.g. for Border or Tabpane title. 
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		public String getTitle(){ 
			return PicGroup.this.getName();
		}
		
	}
} // PicGroup
