/*
 PicSmoothPoly.java - December 25, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.view.View;
import jpicedt.graphic.view.HitInfo;
import jpicedt.widgets.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.util.*;

/**
 * <p>
 * A multicurve, either close or not, whose geometry is specified using a polygon, in such a way that the curve snakes 
 * "smoothly" along the polygon. For each polygon point, a coefficient helps altering the smoothness of the
 * the path. Like the class it inherits from, this element has a variable number of user-controlled points that can be
 * added/inserted/deleted by the user.<br> 
 * Each polygon point is also an anchor point (e.g. for grid alignment).
 * <p><b> Implementation notes : </b> inherited <code>pts</code> ArrayList acts as an array of specification
 * points for the multicurve (this behaviour being mostly inherited from superclass). Besides, there's an
 * ArrayList of PicPoint's, namely <code>polygonPts</code>, which backs the geometry
 * of the polygon, and acts as a list of user-controlled points. 
 * Moving a control-point thus updates specification points as appropriate.
 *
 * [SR:todo:11/01/2003] add documentation about curve equation. 
 *
 * @author    Vincent Guirardel
 * @since     jpicedt 1.3.3
 * @version   $Id: PicSmoothPolygon.java,v 1.18.2.1 2007/09/02 11:56:27 reynal Exp $
 *
 */
public class PicSmoothPolygon extends AbstractCurve implements ActionFactory, CustomizerFactory,PicMultiCurveConvertable {

	//////////////////////////
	//// PROTECTED FIELDS
	//////////////////////////


	/**
	 * An array of PicPoint's backing the geometry of the polygon. Aka caching mechanism holding a list of user-controlled points.
	 */
	protected ArrayList polygonPts = new ArrayList();

	// inherited : protected boolean isClosed;

	/**
	 * The following array contains one Double per polygon point, specifying how close to 
	 * polygon points the curve should go along. 
	 */
	protected ArrayList smoothCoeff = new ArrayList();

	/**
	 * The following array contains one boolean per polygon point, used in the Customizer inner class
	 * to know whether the JSlider component should adjust the corresponding coeff or not.
	 */
	private ArrayList adjust = new ArrayList();

	/**
	 * the default value for the smoothness coefficient on startup.
	 * [SR:pending] fetch from a Properties object in constructor
	 */
	protected static double DEFAULT_SMOOTH_COEFF = 0.7;

	/**
	 * Min value of smoothness coefficient accessible to JSlider in the geometry editor
	 */
	protected static int COEF_SLIDER_MIN = -50 ;
	
	/**
	 * Max value of smoothness coefficient accessible to Jslider in the geometry editor
	 */
	protected static int COEF_SLIDER_MAX = 200 ;



	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Create a new empty open smooth polygon, with a default attribute set.
	 */
	public PicSmoothPolygon() {
		this(false);
	}

	/**
	 * Create a new empty smooth polygon, open or closed, with a default attribute set bound to it.
	 * @param closed  whether the generated multi-curve will be closed or not
	 */
	public PicSmoothPolygon(boolean closed) {
		super(closed);
	}


	/**
	 * Create a new empty smooth polygon with the given attribute set.
	 * @param set     attribute set to be bound to this element
	 * @param closed  whether the generated multi-curve will be closed or not
	 */
	public PicSmoothPolygon(boolean closed, PicAttributeSet set) {
		this(closed);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * Create a new smooth polygon initialized from the given array of PicPoint's.
	 *
	 * @param polyPts array of PicPoint's specifying the polygon geometry.
	 * @param set     attribute set to be bound to this element
	 * @param closed  whether the generated multi-curve will be closed or not
	 * @param smoothCoeff the default smoothness value for each polygon point
	 * @since         PicEdt 1.3.3
	 */
	public PicSmoothPolygon(PicPoint[] polyPts, boolean closed, double smoothCoeff, PicAttributeSet set) {
		super(closed);
		this.attributeSet = new PicAttributeSet(set);
		for (int i = 0; i < polyPts.length; i++) {
			this.polygonPts.add(polyPts[i].clone());
			this.smoothCoeff.add(new Double(smoothCoeff));
			this.adjust.add(Boolean.TRUE);
		}
		this.allocateBezierPoints(); // at least, we this call we're clear !
		this.updateBezierPoints();
	}

	/**
	 * Create a new smooth polygon initialized from the given array of PicPoint's, using the
	 * given array of smoothness values. If array sizes do not match,
	 * number of polygon points gets trimmed to the smaller of both.
	 *
	 * @param polygonPts array of PicPoint's specifying the polygon geometry.
	 * @param set     attribute set to be bound to this element
	 * @param closed  whether the generated multi-curve will be closed or not
	 * @param smoothCoeff array of smoothness values for each polygon point
	 * @since         PicEdt 1.3.3
	 */
	public PicSmoothPolygon(PicPoint[] polygonPts, boolean closed, double[] smoothCoeffs, PicAttributeSet set) {
		super(closed);
		this.attributeSet = new PicAttributeSet(set);
		int size = (polygonPts.length < smoothCoeffs.length ? polygonPts.length : smoothCoeffs.length);
		for (int i = 0; i < size; i++) {
			this.polygonPts.add(polygonPts[i].clone());
			this.smoothCoeff.add(new Double(smoothCoeffs[i]));
			this.adjust.add(Boolean.TRUE);
		}
		this.allocateBezierPoints(); // at least, we this call we're clear !
		this.updateBezierPoints();
	}

	/**
	 * Create a new smooth polygon initialized from the given array of PicPoint's, using the
	 * default smoothness value.
	 *
	 * @param polygonPts array of PicPoint's specifying the polygon geometry.
	 * @param set     attribute set to be bound to this element
	 * @param closed  whether the generated multi-curve will be closed or not
	 * @since         PicEdt 1.3.3
	 */
	public PicSmoothPolygon(PicPoint[] polygonPts, boolean closed, PicAttributeSet set) {
		this(polygonPts, closed, DEFAULT_SMOOTH_COEFF, set);
	}

	/**
	 * "cloning" constructor (to be used by clone())
	 * @param poly  PicSmoothPolygon to be cloned
	 * @since       jpicedt 1.3.3
	 */
	public PicSmoothPolygon(PicSmoothPolygon poly) {
		super(poly);
		smoothCoeff = new ArrayList(poly.smoothCoeff); 
		adjust = new ArrayList(poly.adjust);
		for (Iterator it=poly.polygonPts.iterator(); it.hasNext();) this.polygonPts.add(new PicPoint((PicPoint)it.next()));
	}


	/**
	 * Overide Object.clone() method
	 *
	 * @return   Description of the Return Value
	 * @since    PicEdt 1.1
	 */
	public Object clone() {
		return new PicSmoothPolygon(this);
	}


	/**
	 * @return   a localised string that represents this object's name
	 */
	public String getName() {
		return jpicedt.Localizer.currentLocalizer().get("model.SmoothPoly");
	}


	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////
	/**
	 * Computes the number of Bezier points corresponding to the given number of
	 * polygon points and to the closeness of the curve.
	 */
	private int computeNumberOfBezierPoints(int nbPolygonPts, boolean closed) {
		int nbBezierPts;

		if (closed)
			switch (nbPolygonPts) {
			case 0: nbBezierPts = 0; break; // nbBezSeg=0
			case 1: nbBezierPts = 3; break; // nbBezSeg=1
			case 2: nbBezierPts = 3; break; // nbBezSeg=1
			default: nbBezierPts = 3*nbPolygonPts; break; // nbBezSeg=nbPolygonPts
			}
		else
			switch (nbPolygonPts) {
			case 0: nbBezierPts = 0; break; // nbBezSeg=inconsistant (-1/3) !!!
			case 1: nbBezierPts = 1; break; // nbBezSeg=0
			case 2: nbBezierPts = 4; break; // nbBezSeg=1
			default: nbBezierPts = 3*nbPolygonPts-5; break; // nbBezSeg=nbPolygonPts-2
			}
		return nbBezierPts;
	}

	/**
	 * allocates a number of Bezier points appropriate for the current number of polygon points.
	 */
	private void allocateBezierPoints() {
		int lb;
		int l = polygonPts.size();
		lb = computeNumberOfBezierPoints(l, isClosed);
		pts.clear();
		for (int i=0; i<lb; i++) pts.add(new PicPoint());
	}


	/**
	 * Update the location of Bezier points according to the current location of polygon points.
	 * This obviously relies on the value of the smoothness coefficient for each polygon point.
	 */
	private void updateBezierPoints() {
		int nbPolyPts = polygonPts.size();
		if (isClosed())
			switch (nbPolyPts) {
			case 0: break;
			case 1:
				setSpecificationPoint(0, getPolygonPoint(0));
				setSpecificationPoint(1, getPolygonPoint(0));
				setSpecificationPoint(2, getPolygonPoint(0));
				break;
			case 2:
				getSpecificationPoint(0).setCoordinates(getPolygonPoint(0)).middle(getPolygonPoint(1)); // mid of poly(0,1)
				
				setSpecificationPoint(1, getSpecificationPoint(0)); // mid of poly(0,1)
				getSpecificationPoint(1).translate(getPolygonPoint(0), getPolygonPoint(1), getSmoothCoefficient(1)/2.0); // translate by poly(0,1)*smooth/2
				
				setSpecificationPoint(2, getSpecificationPoint(0)); // mid of poly(0,1)
				getSpecificationPoint(2).translate(getPolygonPoint(1), getPolygonPoint(0), getSmoothCoefficient(0)/2.0);
				
				break;
			default:
				for (int i = 0; i < nbPolyPts; i++) {
					getSpecificationPoint(3*i).setCoordinates(getPolygonPoint(i)).middle(getPolygonPoint(i+1)); // mid of poly(i,i+1)
					
					setSpecificationPoint(3*i+1, getSpecificationPoint(3*i)); // mid of poly(i,i+1)
					getSpecificationPoint(3*i+1).translate(getPolygonPoint(i), getPolygonPoint(i+1), getSmoothCoefficient(i+1)/2.0); // translate by poly(0,1)*smooth/2
					
					getSpecificationPoint(3*i+2).setCoordinates(getPolygonPoint(i+1)).middle(getPolygonPoint(i+2)); // mid of poly(i+1,i+2)
					getSpecificationPoint(3*i+2).translate(getPolygonPoint(i+2), getPolygonPoint(i+1), getSmoothCoefficient(i+1)/2.0); // translate by poly(0,1)*smooth/2
				}
			}
		else { // open polygon
			switch (nbPolyPts) {
			case 0: break;
			case 1:
				setSpecificationPoint(0, getPolygonPoint(0));
				break;
			case 2:
				setSpecificationPoint(0, getPolygonPoint(0));
				setSpecificationPoint(3, getPolygonPoint(1));
				
				setSpecificationPoint(1, getPolygonPoint(0));
				getSpecificationPoint(1).translate(getPolygonPoint(0), getPolygonPoint(1), getSmoothCoefficient(0)/2.0); // translate by poly(0,1)*smooth/2
				
				setSpecificationPoint(2, getPolygonPoint(1));
				getSpecificationPoint(2).translate(getPolygonPoint(1), getPolygonPoint(0), getSmoothCoefficient(1)/2.0); // translate by poly(1,0)*smooth/2
				
				break;
			default:
				int nbBzPts = 3*nbPolyPts-5;

				setSpecificationPoint(0, getPolygonPoint(0));
				
				setSpecificationPoint(1, getPolygonPoint(0));
				getSpecificationPoint(1).translate(getPolygonPoint(0), getPolygonPoint(1), getSmoothCoefficient(0)); // translate by poly(0,1)*smooth

				setSpecificationPoint(nbBzPts-2, getPolygonPoint(nbPolyPts-1));
				getSpecificationPoint(nbBzPts-2).translate(getPolygonPoint(nbPolyPts-1), getPolygonPoint(nbPolyPts-2), getSmoothCoefficient(nbPolyPts-1)); // translate by poly(0,1)*smooth/2

				setSpecificationPoint(nbBzPts-1, getPolygonPoint(nbPolyPts-1)); // last end-point

				for (int i = 1; i < nbPolyPts-2; i++) {
					// Bezier endpoint at mid-seg
					getSpecificationPoint(3*i).setCoordinates(getPolygonPoint(i)).middle(getPolygonPoint(i+1)); // mid of poly(i,i+1)
					//control points
					setSpecificationPoint(3*i-1, getSpecificationPoint(3*i));
					getSpecificationPoint(3*i-1).translate(getSpecificationPoint(3*i), getPolygonPoint(i), getSmoothCoefficient(i)); // =polygonPtsX[i] when smoothCoeff==1

					setSpecificationPoint(3*i+1, getSpecificationPoint(3*i));
					getSpecificationPoint(3*i+1).translate(getSpecificationPoint(3*i), getPolygonPoint(i+1), getSmoothCoefficient(i+1)); // =polygonPtsX[i+1] when smoothCoeff==1
				}
				break;
			} // end of open case
		}
	}

	/**
	 * Returns the index of the first user-controlled point that can be retrieved by <code>getPoint()</code>.
	 * This implementation returns 0.
	 */
	public int getFirstPointIndex(){
		return 0;
	}

	/**
	 * Returns the index of the last user-controlled point that can be retrieved by <code>getPoint()</code>.
	 * This default implementation returns the number of polygon-points minus one.
	 */
	public int getLastPointIndex(){
		return polygonPts.size()-1;
	}

	/**
	 * Return the <b>user-controlled point</b> (ie a polygon point, not a bezier point) having the given index. 
	 * The general contract in Element
	 * is to return an IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method.<br>
	 * Overriden so as to return a "polygon" point instead of a bezier-point (which act as specification-point here).
	 * @return the point indexed by <code>numPoint</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 */
	public PicPoint getPoint(int numPoint, PicPoint src){
		if (src==null) src = new PicPoint();
		src.setCoordinates(getPolygonPoint(numPoint));
		return src;
	}
	
	/**
	 * Return a reference on the polygon point with the given index.
	 * This is equivalent to calling getPoint(int,PicPoint), except that a mere reference is returned.
	 */
	protected final PicPoint getPolygonPoint(int numPoint){
		return (PicPoint)polygonPts.get(numPoint % polygonPts.size());
	}
	
	/**
	 * This implementation returns the X-coord of the polygon-point with the same index. 
	 * @return the X-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since jpicedt 1.3.3
	 */
	public double getPointX(int index){
		return getPolygonPoint(index).x;
	}

	/**
	 * This implementation returns the Y-coord of the polygon-point with the same index. 
	 * @return the Y-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since jpicedt 1.3.3
	 */
	public double getPointY(int index){
		return getPolygonPoint(index).y;
	}

	/**
	 * Translate this Element by (dx,dy) ; this implementation translates the specification-points,
	 * then fires a changed-update event.
	 * @param dx The X coordinate of translation vector
	 * @param dy The Y coordinate of translation vector
	 * @since PicEdt 1.0
	 */
	public void translate(double dx, double dy){
		for (Iterator it = polygonPts.iterator(); it.hasNext(); ){
			PicPoint pt = (PicPoint)it.next();
			pt.translate(dx,dy);
		}
		super.translate(dx,dy); // translates pts, then fires a change-update event
	}

	/**
	 * Scale this object by (sx,sy) using (ptOrgX,ptOrgY) as the origin. This implementation
	 * simply apply a scaling transform to all specification-points.
	 * Note that <code>sx</code> and <code>sy</code> may be negative. 
	 * This method eventually fires a changed-update event.
	 */
	public void scale(double ptOrgX, double ptOrgY, double sx, double sy){
		for (Iterator it = polygonPts.iterator(); it.hasNext(); ){
			PicPoint pt = (PicPoint)it.next();
			pt.scale(ptOrgX, ptOrgY, sx, sy);
		}
		super.scale(ptOrgX, ptOrgY, sx, sy); // scales pts, then fires a change-update event
	}

	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		for (Iterator it = polygonPts.iterator(); it.hasNext();)
			((PicPoint)it.next()).rotate(ptOrg,angle);
		super.rotate(ptOrg, angle); // rotate pts, then fires a change-update event
	}
	
	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		for (Iterator it = polygonPts.iterator(); it.hasNext();)
			((PicPoint)it.next()).shear(ptOrg,shx,shy);
		super.shear(ptOrg,shx,shy); // rotate pts, then fires a change-update event
	}

	/**
	 * Set the user-controlled point with the given index to the given value.
	 * This implementation sets the value of the polygon-point having the same index,
	 * then update the associated Bezier curve, and fires a changed-update event.
	 * @param constraint not used so far
	 */
	public void setPoint(int index, PicPoint pt, EditPointConstraint constraint) {
		getPolygonPoint(index).setCoordinates(pt);
		updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * adds the given point at the end of this PicSmoothPolygon. 
	 */
	public void addPoint(PicPoint pt) {
		addPoint(pt, DEFAULT_SMOOTH_COEFF);
	}

	/**
	 * adds the given point with the given smoothness-coefficient at the end of this PicSmoothPolygon.
	 */
	public void addPoint(PicPoint pt, double smoothCoeff) {
		polygonPts.add(pt.clone());
		this.smoothCoeff.add(new Double(smoothCoeff));
		adjust.add(Boolean.TRUE);
		this.allocateBezierPoints();
		this.updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Adds the given polygon point at the given position, shifting ensuing point indices to the right.
	 * For instance, adding at point at position "1" is equivalent to splitting segment "0".
	 */
	public void addPoint(int ptIndex, PicPoint pt){
		addPoint(ptIndex, pt, DEFAULT_SMOOTH_COEFF);
	}
	
	/**
	 * Adds the given polygon point at the given position, shifting ensuing point indices to the right.
	 * For instance, adding at point at position "1" is equivalent to splitting segment "0".
	 */
	public void addPoint(int ptIndex, PicPoint pt, double smoothCoeff){
		splitSegment(ptIndex-1, pt, smoothCoeff);
	}

	/**
	 * Split the given polygon segment (starting from 0), 
	 * by inserting a new polygon point at the appropriate position.
	 * then fires a changed-update. 
	 * For instance, spliting segment "2" yields the following polygon points : 0, 1, 2, insertion pt, 3, 4, etc...
	 *
	 * @param segIdx index of the polygon segment that must be split.
	 * @param pt   the PicPoint to be added
	 */
	public int splitSegment(int segIdx, PicPoint pt) {
		return splitSegment(segIdx, pt, DEFAULT_SMOOTH_COEFF);
	}

	/**
	 * Split the given polygon segment (starting from 0), 
	 * by inserting a new polygon point at the appropriate position.
	 * then fires a changed-update. 
	 * For instance, spliting segment "2" yields the following polygon points : 0, 1, 2, insertion pt, 3, 4, etc...
	 *
	 * @param segIdx index of the polygon segment that must be split.
	 * @param smoothCoeff the smoothness coefficient to be used for the new polygon point.
	 * @param pt   the PicPoint to be added
	 */
	public int splitSegment(int segIdx, PicPoint pt, double smoothCoeff) {

		if (segIdx < 0 || segIdx > polygonPts.size()) throw new IndexOutOfBoundsException(new Integer(segIdx).toString());

		polygonPts.add(segIdx+1, pt.clone()); // ArrayList.add shifts ensuing indices to the right
		this.smoothCoeff.add(segIdx+1,new Double(smoothCoeff));
		adjust.add(segIdx+1, Boolean.TRUE);
		
		this.allocateBezierPoints();
		this.updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		return segIdx+1;
	}

	/**
	 * Remove a point at the given position from this PicSmoothPolygon, then fire a
	 * changed-update. This shifts any subsequent points to the left.<p>
	 *
	 * Nothing is done if this polygon has only one point.
	 *
	 * @param pos  index of point to be removed
	 */
	public void removePoint(int pos) {

		if (pos < 0 || pos >= polygonPts.size()) throw new IndexOutOfBoundsException(new Integer(pos).toString());
		if (polygonPts.size() == 1) return;// we won't remove the only point !
		
		polygonPts.remove(pos);
		smoothCoeff.remove(pos);
		adjust.remove(pos);
		
		this.allocateBezierPoints();
		this.updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * removes the last point of this PicSmoothPolygon. 
	 * This is a convenience call to <code>removePoint(polygonPts.size()-1)</code>
	 */
	public void removeLastPoint() {
		removePoint(polygonPts.size()-1);
	}


	/**
	 * converts this PicSmoothPolygon to a PicMultiCurve, by simply relying on the appropriate constructor
	 * in class {@link PicMultiCurve PicMultiCurve}.
	 */
	public PicMultiCurve convertToMultiCurve(){
		PicMultiCurve curve=new PicMultiCurve(this);
		return curve;
	}



	////////////////////////////
	//// OTHER FIELDS ACCESSORS
	////////////////////////////

	/**
	 * @param index  Description of the Parameter
	 * @return smoothness smoothCoeff of the given control point
	 */

	public double getSmoothCoefficient(int index) { // was : getCoeff
		return ((Double)smoothCoeff.get(index % polygonPts.size())).doubleValue();
	}


	/**
	 * @param index index of point whose smoothCoeff is to be adjusted
	 * @param c value of the smoothCoeff
	 */

	public void setSmoothCoefficient(int index, double c) { // was : setCoeff
		smoothCoeff.set(index, new Double(c));
		updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * set close path state, then fires a changed-update.
	 *
	 * @param state  The new closed value
	 */
	public void setClosed(boolean state) {
		super.setClosed(state);
		allocateBezierPoints();
		updateBezierPoints();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	//////////////////////////////////////////////////////////
	/// BOUNDING BOX
	//////////////////////////////////////////////////////////

	/**
	 * Returns the bounding box (ie the surrounding rectangle) in double precision
	 * This implementation compute the bb from the union of :
	 * <ul>
	 * <li> the bb as computed by super-class 
	 * <li> and the smallest rectangle that encompasses  all the polygon-points.
	 * </ul>
	 * @since jpicedt 1.3.3
	 */
	public Rectangle2D getBoundingBox(Rectangle2D r){
		r = super.getBoundingBox(r);
		for (int i=0; i<polygonPts.size(); i++) r.add((Point2D)polygonPts.get(i)); // then enlarge bb by adding subsequent points
		return r;
	}

	///////////////////////////////
	//// STRING FORMATING
	///////////////////////////////

	/**
	 * Used for debugging purpose.
	 */
	public String toString() {
		String s = super.toString();
		s += "\n\t";
		for (int i=0; i<polygonPts.size(); i++){
			s += "polyPts[" + i + "]=" + getPolygonPoint(i)  + " ";
		}
		s += "\n\t";
		for (int i=0; i<polygonPts.size(); i++){
			s += "coeff[" + i + "]=" + getSmoothCoefficient(i) + " ";
		}
		s += "\n\t";
		s += (isClosed ? " closed" : " open");
		return s;
	}

	////////////////////////////////
	//// Action's
	////////////////////////////////

	/**
	 * Create an array of Action's related to this object
	 *
	 * @param actionDispatcher  dispatches events to the proper PECanvas
	 * @param localizer         i18n localizer for PEAction's
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {
		//return super.createActions(actionDispatcher, localizer);
		// override if applicable by adding your own actions here.
		PEAction[] superClassActionArray = super.createActions(actionDispatcher, localizer,hi);
		PEAction[] actionArray = new PEAction[superClassActionArray.length + 1];
		System.arraycopy(superClassActionArray, 0, actionArray, 0, superClassActionArray.length);
		actionArray[actionArray.length-1] = new ConvertToCurveAction(actionDispatcher, localizer);
		return actionArray;
	}

	/**
	 * @return   a Customizer for geometry editing
	 */
	public AbstractCustomizer createCustomizer() {
		return new Customizer();
	}

	// ---- SmoothPoly to multiCurve ----
	
	/**
	 * Convert this SmoothPolygon to a multicurve, selecting it if applicable.
	 * @author    Sylvain Reynal
	 * @since     jpicedt 1.4
	 */
	class ConvertToCurveAction extends PEAction {

		public static final String KEY = "action.editorkit.ConvertSmoothPolyToCurve";
		
		public ConvertToCurveAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void undoableActionPerformed(ActionEvent e){
			PicMultiCurve curve = convertToMultiCurve();
			Drawing dr = getDrawing();
			if (dr != null){
				int index = dr.getRootElement().getChildIndex(PicSmoothPolygon.this);
				dr.replaceElement(index, curve);
				View view = curve.getView();
				if (view != null){
					PECanvas canvas = view.getContainer();
					if (canvas != null)   canvas.select(curve, true);
				}
			}
		}
	}


	////////////////////////////////////////////////////////////////////////////////////
	//// GUI
	////////////////////////////////////////////////////////////////////////////////////
	
	/**
	 * geometry customizer
	 *
	 * @author    reynal
	 * @created   January 11, 2003
	 */
	class Customizer extends AbstractCurve.Customizer {

		private DefaultCellEditor doubleEditor;
		private JSlider coefSlider;// to adjust coef
		private boolean isListenersAdded = false;// flag
		private	PolygonJTable table ;

		/**
		 * Constructor for the Customizer object
		 */
		public Customizer() {

			super();
			Box p = new Box(BoxLayout.Y_AXIS);
			// polygon points:
			table = new PolygonJTable();
			table.setPreferredScrollableViewportSize(new Dimension(200, 100));

			// wrap scrollpane around:
			JScrollPane scrollPane = new JScrollPane(table);
			p.add(scrollPane);
			p.add(coefSlider = new JSlider(COEF_SLIDER_MIN,COEF_SLIDER_MAX,70));
			coefSlider.setMajorTickSpacing(50);
			coefSlider.setMinorTickSpacing(10);
			coefSlider.setPaintTicks(true);
			coefSlider.setPaintLabels(true);
			add(p, BorderLayout.NORTH);
			add(super.createPanel(), BorderLayout.CENTER);
		}

		// a listener for the slider
		class SliderListener implements ChangeListener {
			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider)e.getSource();
				double coef = (double)source.getValue()/100;
				for (int i=0;i <= getLastPointIndex();i++){
					if (adjust.get(i)==Boolean.TRUE) smoothCoeff.set(i,new Double(coef));
				}
				if (!source.getValueIsAdjusting()) {table.repaint();}
				updateBezierPoints();
				fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
			}
		}

		/**
		 * add action listeners to widgets to reflect changes immediately
		 */
		private void addActionListeners() {
			if (isListenersAdded)
				return;// already done
			coefSlider.addChangeListener(new SliderListener());
			isListenersAdded = true;
		}

		/**
		 * load widgets with object's properties
		 */
		public void load() {
			super.load();
			// [pending] use JTable.resizeAndRepaint() ???
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded
			addActionListeners();// done the first time load is called

		}

		/**
		 * update Element's properties
		 */
		public void store() {
			super.store();
			for (int i=0;i <= getLastPointIndex();i++){
				smoothCoeff.set(i,new Double(((double)coefSlider.getValue())/100)); 
			}
			updateBezierPoints();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}

		// public void actionPerformed(ActionEvent e)  // inherited

		/**
		 * @return   the panel title, used e.g. for Border or Tabpane title.
		 */
		public String getTitle() {
			return PicSmoothPolygon.this.getName();
		}
	}

	///////////////////////////////////////////////////
	//// INNER CLASSES for Customizer
	///////////////////////////////////////////////////

	/**
	 * an implementation of JTable for PicSmoothPolygon's
	 */
	class PolygonJTable extends JTable {

		private final CoordinateCellEditor coordCellEditor = new CoordinateCellEditor(new DecimalNumberField(0, 5));
		//	        private final DefaultCellEditor BooleanEditor = new DefaultCellEditor(new Boolean(true));
		private final PolygonTableModel model;

		/**
		 * Constructor for the PolygonJTable object
		 */
		PolygonJTable() {
			super();
			setModel(model = new PolygonTableModel());
		}

		/**
		 * Returns an appropriate editor for the cell specified by row and column.
		 */
		public TableCellEditor getCellEditor(int row, int col) {
			if (col > 0 && col !=3)
				return coordCellEditor;
			//			else if (col >0 )
			//    return BooleanEditor;
			else
				return super.getCellEditor(row, col);
		}
	}

	/**
	 * PolygonJTable's cell editor is based on a DecimalNumberField
	 */
	class CoordinateCellEditor extends DefaultCellEditor {

		private DecimalNumberField dnf;


		/**
		 * construct a new cell editor from the given DecimalNumberField
		 */
		CoordinateCellEditor(DecimalNumberField dnf) {
			super(dnf);
			this.dnf = dnf;

		}


		/**
		 * Returns the value contained in the editor-overriden from
		 * DefaultCellEditor so as to return a Double instead of a String.
		 */
		public Object getCellEditorValue() {
			return new Double(dnf.getValue());
		}

	}

	/**
	 * a class that specifies the methods the JTable will use to interrogate a
	 * tabular data model.
	 */
	class PolygonTableModel extends AbstractTableModel {

		/**
		 * Returns the name of the column at columnIndex.
		 */
		public String getColumnName(int colIndex) {
			switch (colIndex) {
			case 0:
				return jpicedt.Localizer.currentLocalizer().get("misc.Point");// polygon point index
			case 1:
				return "X";// x-coordinate
			case 2:
				return "Y";// y-coordinate
			case 3:
				return jpicedt.Localizer.currentLocalizer().get("misc.Adjust");
			case 4:
				return jpicedt.Localizer.currentLocalizer().get("misc.Coefficient");
			default:
				return "";
			}
		}


		/**
		 * Returns the number of columns in the model.
		 *
		 * @return   The columnCount value
		 */
		public int getColumnCount() {
			return 5;// 4 columns : "pt", "X", "Y", and "adjust"
		}


		/**
		 * Returns the number of rows in the model.
		 *
		 * @return   The rowCount value
		 */
		public int getRowCount() {
			return polygonPts.size();// there are as many rows as points in the associated PicSmoothPolygon
		}


		/**
		 * Returns true if the cell at rowIndex and columnIndex is editable.
		 *
		 * @param rowIndex  Description of the Parameter
		 * @param colIndex  Description of the Parameter
		 * @return          The cellEditable value
		 */
		public boolean isCellEditable(int rowIndex, int colIndex) {
			if (colIndex == 0)
				return false;// first column is not editable since it containts point indexes
			else
				return true;
		}

		public Class getColumnClass(int c) {
			return getValueAt(0, c).getClass();
		}

		/**
		 * Called when the associated JTable wants to know what to display at
		 * columnIndex and rowIndex. Depending on colIndex, we return, either the
		 * point index, the x-coord, or the y-coord of the corresponding polygon's
		 * point.
		 *
		 * @param rowIndex  Description of the Parameter
		 * @param colIndex  Description of the Parameter
		 * @return          The valueAt value
		 */
		public Object getValueAt(int rowIndex, int colIndex) {

			switch (colIndex) {
			case 0:
				return new Integer(rowIndex);// polygon point index
			case 1:
				return new Double(PEToolKit.doubleToString(getPolygonPoint(rowIndex).x));// x-coord for the row-th point
			case 2:
				return new Double(PEToolKit.doubleToString(getPolygonPoint(rowIndex).y));// y-coord for the row-th point
			case 3:
				return (Boolean)adjust.get(rowIndex); // box is checked
			case 4:
				return new Double(PEToolKit.doubleToString(100*getSmoothCoefficient(rowIndex)));// coefficient
			default:
				return null;
			}
		}


		/**
		 * Called when a user entered a new value in the cell at columnIndex and
		 * rowIndex to value. We update PicSmoothPolygon geometry according to the cell
		 * value.<br>
		 * This method does nearly the same thing as actionPerformed in other
		 * xxxxCustomizer's.
		 *
		 * @param value     The new valueAt value
		 * @param rowIndex  The new valueAt value
		 * @param colIndex  The new valueAt value
		 */
		public void setValueAt(Object value, int rowIndex, int colIndex) {

			PicPoint pt = getPoint(rowIndex, null);
			switch (colIndex) {// do we modify x or y ?
			case 1:
				pt.x = ((Double)value).doubleValue();
				break;
			case 2:
				pt.y = ((Double)value).doubleValue();
				break;
			case 3:
				adjust.set(rowIndex,value); // Boolean
				break;
			case 4:
				smoothCoeff.set(rowIndex, new Double(((Double)value).doubleValue()/100));
				break;
			default:
				return;
			}
			setPoint(rowIndex, pt,null);
			updateBezierPoints();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}
	}// inner class
}

