/*
 PicText.java - 1999 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.widgets.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.*;

/**
 * Textual elements, possibly surrounded by a box (currently, only rectangular, oval and circular shapes
 * are supported). The box size is computed from the text size (ie aka LaTeX's \\framebox command with
 * no width/height argument).
 *
 * @author Sylvain Reynal
 * @since PicEdt 1.0
 * @version $Id: PicText.java,v 1.24.2.1 2007/09/02 11:56:27 reynal Exp $
 */

public class PicText extends TextEditable implements CustomizerFactory {  // exemple : \\put(15,20){\\makebox(0,0)[bl]{texte}}

	////////////////////////////
	//// PUBLIC CONSTANT FIELDS
	////////////////////////////


	//////////////////////////
	//// PRIVATE FIELDS
	//////////////////////////


	/** dimensions (in latex sense) of the latex box (this does not include frame) */
        protected double width; // width of the box
        protected double height; // the height above the baseline
        protected double depth; // the depth below the baseline (depth+height=total height of the box)


        protected Shape frame; //the frame surrounding the text
        

    // dimensions (in latex sense) of the frame
    private double frameWidth, frameHeight, frameDepth;  //  height= what's above the baseline
    //depth = what's below the baseline (depth+height=total height of the box)
    private double xOffset,yOffset; // coordinates of the vector BottomLeftOfText - BottomLeftOfFrame

        // TEMPORARY/ EXPERIMENTAL
        protected double frameSep=1; //to be set as an attribute. (contols amount of white space between text and frame).
                                    // exported in pstricks as framesep parameter
        protected double nodeSep=1; //to be set as an attribute. (contols amount of white space between frame and node-connections).
                     //exported as pstricks nodesep parameter

	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

				
	/**
	 * Create a new PicText object with the whole bunch of parameters !
	 *
	 * @param ptA Anchor point (see setHorAlign() and setVertAlign() for details)
	 * @param textString the text content
	 */
	public PicText(PicPoint ptA, String textString, PicAttributeSet set) {
		super(textString,ptA,set);
		width=0; height=0; depth=0;
		updateFrame();
	}

	/**
	 * Creates a PicText object with the following default values :<br>
	 * <ul>
	 * <li> empty string, 
	 * <li> frame and alignment according to the attribute set
	 * </ul>
	 */
	public PicText(PicPoint ptA, PicAttributeSet set) {
		super("",ptA,set);
		width=0; height=0; depth=0;
		updateFrame();
	}

	/**
	 * Creates a PicText object with the following default values :
	 * <ul> 
	 * <li> empty string, 
	 * <li> frame and alignment according to the attribute set
	 * </ul>
	 * and anchor point at (0,0)
	 */
	public PicText(PicAttributeSet set) {
		this(new PicPoint(), set);
		updateFrame();
	}

	/**
	 * Creates a PicText object with a default attribute set, and the following default values :
	 * <ul> 
	 * <li> empty string, 
	 * <li> no frame,
	 * <li> bottom-left alignment.
	 * </ul>
	 * and anchor point at (0,0)
	 */
	public PicText() {
		super("", new PicPoint());
		updateFrame();
	}

	/**
	 * cloning constructor
	 */
	public PicText(PicText text){
		super(text);
		this.width=text.width; this.height=text.height; this.depth=text.depth;
		this.frame=text.frame;
		this.frameWidth=text.frameWidth; this.frameHeight=text.frameHeight; this.frameDepth=text.frameDepth;
	}

	/**
	 * Override Object.clone() method
	 */
	public Object clone(){
		return new PicText(this);
	}

	/**
	 * Return a string that represents this object's name
	 */
	public String getName(){
		return jpicedt.Localizer.currentLocalizer().get("model.Text");
	}


	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////
    public void setPoint(int index, PicPoint pt, EditPointConstraint constraint){
	setSpecificationPoint(index,pt);
	updateFrame();
	fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
    }

	// rest is inherited

	////////////////////////////
	//// OTHER FIELDS ACCESSORS
	////////////////////////////

	/**
	 * Return horizontal alignment (see constant fields above)
	 */
	public String getHorAlign(){
		return (String)attributeSet.getAttribute(TEXT_HOR_ALIGN);
	}

	/**
	 * Set the horizontal alignment (see constant fields above)
	 */
	public void setHorAlign(String horAlign){
		if (horAlign==null) return;
		if (!horAlign.equals(TEXT_HALIGN_LEFT) && !horAlign.equals(TEXT_HALIGN_RIGHT) && !horAlign.equals(TEXT_HALIGN_CENTER)) 
			throw new IllegalArgumentException("Horizontal Alignment : LEFT, RIGHT or CENTER_H only !");
		attributeSet.setAttribute(TEXT_HOR_ALIGN, horAlign);
		fireChangedUpdate(DrawingEvent.EventType.ATTRIBUTE_CHANGE);
	}

	/**
	 * Return vertical alignement (see static constant above)
	 */
	public String getVertAlign(){
		return (String)attributeSet.getAttribute(TEXT_VERT_ALIGN);
	}

	/**
	 * set the vertical alignement (see constant fields above)
	 */
	public void setVertAlign(String vertAlign){
		if (vertAlign == null) return;
		if (!vertAlign.equals(TEXT_VALIGN_BOTTOM) && !vertAlign.equals(TEXT_VALIGN_BASELINE) && !vertAlign.equals(TEXT_VALIGN_CENTER)  && !vertAlign.equals(TEXT_VALIGN_TOP)) 
			throw new IllegalArgumentException("Vertical Alignment : BOTTOM, BASELINE, CENTER_V or TOP only !");
		attributeSet.setAttribute(TEXT_VERT_ALIGN, vertAlign);
		fireChangedUpdate(DrawingEvent.EventType.ATTRIBUTE_CHANGE);
	}

    ///////////////////////////////////////////////
    // Frame related methods
    //////////////////////////////////////////////

	/**
	 * Return TRUE if the box has to be drawn
	 */
	public boolean isFramed(){
		return (getFrameType()!= null && !getFrameType().equals(TEXT_BOX_NO_FRAME));
	}

	/**
	 * Set the type of frame to be put around the text.
	 */
	public void setFrameType(String frameType){
		if (frameType==null) return;
		attributeSet.setAttribute(TEXT_FRAME, frameType);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); // may change bounds if NO_FRAME -> RECT BOX
	}

	/**
	 * Set the frame type
	 */
	public String getFrameType(){
		return (String)attributeSet.getAttribute(TEXT_FRAME);
	}



	/**
	 * Returns the frame that surrounds the text, ignoring rotation (the view will rotate it)
	 */
    public Shape getShapeOfFrame(){
	return frame;
    }


	/**
	 * First computes frame dimensions (frameWidth, etc).
	 * Then computes the frame that surrounds the text, ignoring rotation (the view will rotate it)
	 */

    public void updateFrame(){
	//[pending] frameSep should be a parameter
	if (! isFramed()){
	    frameWidth=width; frameDepth=depth; frameHeight=height;
	    xOffset=0; yOffset=0;
	    return ;
	}
	double sep=frameSep+0.5*((Double)getAttribute(LINE_WIDTH)).doubleValue();
	if (getFrameType()==TEXT_BOX_RECTANGLE){
	    frameWidth=width+2*sep;
	    frameHeight=height+sep;
	    frameDepth=depth+sep;
	    xOffset=sep; yOffset=sep;
	    if (!(frame instanceof Rectangle2D.Double)) frame = new Rectangle2D.Double();
	    ((Rectangle2D)frame).setFrame(getFrameLeftX(),getFrameBottomY(),frameWidth,frameHeight+frameDepth);
	    return ;
	}
	else if (getFrameType()==TEXT_BOX_OVAL){//smallest area ellipse containing the box
	    // solution pour une ellipse donnee par x^2/a^2+y^2/b^2=1 passant par un point donne (x,y) : x^2/a^2=y^2/b^2=1/2  => a=sqrt(2)*x;
	    frameWidth=Math.sqrt(2)*width+2*sep;
	    xOffset=0.5*(frameWidth-width);
	    yOffset=Math.sqrt(0.5)*(height+depth)+sep-0.5*(height+depth);
	    frameHeight= height +yOffset ;
	    frameDepth= depth + yOffset ;
	    if (!(frame instanceof Ellipse2D.Double)) frame = new Ellipse2D.Double();
	    ((Ellipse2D)frame).setFrame(getFrameLeftX(),getFrameBottomY(),frameWidth,frameHeight+frameDepth);
	    return ;
	}
	else if (getFrameType()==TEXT_BOX_CIRCLE){
	    if (! (frame instanceof Ellipse2D.Double)) frame = new Ellipse2D.Double();
	    //Next formula means: increase both dimensions of the rectangle by 2*sep , and compute the distance from center to corner
	    double radius=0.5*Math.sqrt(width*width+(height+depth)*(height+depth))+sep;
	    xOffset=radius-0.5*width;
	    frameWidth=2*radius;
	    yOffset=radius-0.5*(height+depth);
	    frameHeight= height +yOffset ;
	    frameDepth= depth + yOffset ;
	    ((Ellipse2D)frame).setFrame(getFrameLeftX(),getFrameBottomY(),frameWidth,frameHeight+frameDepth);
	    return ;
	}
	return ; // should not occur
    }



    /////////////////////////////////////
    // text dimensions
    ////////////////////////////////////

    
        /**
         * Return the width of text only
         */
        public double getWidth(){
                return width;
        }

        /**
         * Return the height of text only
         */
        public double getHeight(){
                return height;
        }

        /**
         * Return the depth of text only
         */
        public double getDepth(){
                return depth;
        }

    /**
     * Return x-coordinate of left boundary of text only
     */
    
    public double getLeftX(){
	return getFrameLeftX()+xOffset;
    }


    /**
     * Return x-coordinate of right boundary of text only
     */
    
    public double getRightX(){
        return getLeftX()+width;
    }

    /**
     * Return x-coordinate of center of text only
     */
    
    public double getCenterX(){
        return getLeftX()+width/2;
    }




    /**
     * Return y-coordinate of bottom boundary of text only
     */
    
    public double getBottomY(){
        return getFrameBottomY()+yOffset;
    }


    /**
     * Return y-coordinate of top boundary of text only
     */
    
    public double getTopY(){
	return getBottomY()+depth+height;
    }

    /**
     * Return y-coordinate of baseline 
     */
    
    public double getBaseLineY(){
        return getBottomY()+depth;
    }


    /**
     * Return y-coordinate of center of text only
     */
    
    public double getCenterY(){
        return getBottomY()+(height+depth)/2;
    }




    /////////////////////////////////////
    // Frame dimensions
    /////////////////////////////////////


    /**
     * Return the width of the text including frame
     */
    public double getFrameWidth(){
	return frameWidth;
    }
    
    /**
     * Return the height of the text including frame
     */
    public double getFrameHeight(){
	return frameHeight;
    }
    
    /**
     * Return the depth of the text including frame
     */
	public double getFrameDepth(){
		return frameDepth;
	}



    /**
     * Return x-coordinate of left boundary, including frame
     */
    
    public double getFrameLeftX(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	if (getHorAlign() == TEXT_HALIGN_LEFT) return getPointX(P_ANCHOR);
	else if (getHorAlign() == TEXT_HALIGN_RIGHT) return getPointX(P_ANCHOR) - frameWidth ;
	else //centered
	    return getPointX(P_ANCHOR) - 0.5*frameWidth;
    }


    /**
     * Return x-coordinate of right boundary
     */
    
    public double getFrameRightX(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	return getFrameLeftX()+frameWidth;
    }

    /**
     * Return x-coordinate of center
     */
    
    public double getFrameCenterX(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	return getFrameLeftX()+0.5*frameWidth;
    }


    /**
     * Return y-coordinate of top boundary
     */
    
    public double getFrameTopY(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	if (getVertAlign() == TEXT_VALIGN_TOP) return getPointY(P_ANCHOR);
	else if (getVertAlign() == TEXT_VALIGN_BOTTOM) return getPointY(P_ANCHOR)+frameDepth+frameHeight;
	else if (getVertAlign() == TEXT_VALIGN_BASELINE) return getPointY(P_ANCHOR)+frameHeight;
	else return getPointY(P_ANCHOR)+(frameHeight+frameDepth)/2;
    }


    /**
     * Return y-coordinate of bottom boundary
     */
    
    public double getFrameBottomY(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	return getFrameTopY()-frameHeight-frameDepth;
    }

    /**
     * Return y-coordinate of baseline
     */
    
    public double getFrameBaseLineY(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	return getFrameTopY()-frameHeight;
    }


    /**
     * Return y-coordinate of center
     */
    
    public double getFrameCenterY(){
	// This method is not allowed to use frame for computing the width because it is used to compute the frame in updateFrame.
	// However, it can use frame dimensions (frameWidth etc) because they are computed before accessing this method.
	return getFrameTopY()-0.5*(frameHeight+frameDepth);
    }




	/**
	 * set the width, height, and depth (of the text box, without frame). Normally, are set by the view.
	 */
	public void setDimensions(double w, double h, double d){
		this.width = w;
		this.height = h;
		this.depth = d;
		updateFrame();
		System.out.println(frameDepth+" "+frameWidth+" "+frameHeight);//debug
		// ces dimensions seront en general mises a jour par la view, donc il ne faut pas faire de fireChangeUpdate
		//		fireChangedUpdate(jpicedt.graphic.event.DrawingEvent.EventType.ATTRIBUTE_CHANGE); 
	}



    /////////////////////////////////////////////////////////////////
    // NODE Stuff
    /////////////////////////////////////////////////////////////////
	
    /* 
     * Yes, Pictext can be made a node.
     */

    public boolean isNodeable(){
	return true;
    }

    /* returns the Picnode reference point. 
     * Used to compute connections: for instance, a straight connection between two nodes
     * should be a segment in the line connecting the reference points of the two nodes.
     */
    public PicPoint nodeReferencePoint(){
	if (! isFramed() || getFrameType()==TEXT_BOX_RECTANGLE)
	    return getSpecificationPoint(P_ANCHOR);
	else // otherwise, we always use the center
	    return new PicPoint(getFrameCenterX(),getFrameCenterY());
    }
    
	
    /* returns the Picnode reference point X coordinate 
     * does not create a new PicPoint.
     */
    public double nodeReferencePointX(){
	if (! isFramed() || getFrameType()==TEXT_BOX_RECTANGLE)
	    return getPointX(P_ANCHOR);
	else // otherwise, we always use the center
	    return getFrameCenterX();
    }
	
    /* returns the Picnode reference point X coordinate 
     * does not create a new PicPoint.
     */
    public double nodeReferencePointY(){
	if (! isFramed() || getFrameType()==TEXT_BOX_RECTANGLE)
	    return getPointY(P_ANCHOR);
	else // otherwise, we always use the center
	    return getFrameCenterY();
    }

    /* Given a non-zero vector of coordinates (dx, dy), computes the origin of a connection
     * that should start from that direction.
     * this takes into account the nodeSep parameter (separation between the frame and the origin of the node).
     * If nodeSep=0 for instance, this computes the intersection point of the semiline based at the reference point,
     * and directed by (dx,dy) with the frame.
     * About nodeSep: its effect mimics the effect in pstricks: we increase the size of the rectangle/ellipse/etc
     * by the nodeSep amount.
     */

    public PicPoint nodeConnectionOrigin(double dx, double dy){
	if (dx == 0 && dy==0){
	    // bad value of the vector. Can occur if we try to join 2 nodes having the sams ref point...
	    return nodeReferencePoint();
	}
	double refX=nodeReferencePointX();
	double refY=nodeReferencePointY();
	if (! isFramed() || getFrameType()==TEXT_BOX_RECTANGLE){
	    double lw; //linewidth, = 0 if no frame
	    if (!isFramed()) 
		lw=0; 
	    else 
		lw=((Double)(getAttribute(LINE_WIDTH))).doubleValue();

	    //orientation
	    // we work in coordinates where origin is at the reference point, and orientation of axes are given by signs of dx dy.
	    // the relative coordinates of our vector are dx*sgnX,dy*sgnY (=abs(dx),abs(dy))

	    double cornerX,cornerY; //relative coordinates of the corner of the rectangle. Should be non-negative
	    double sgnX,sgnY; // +1 or -1 according to orientation;
	    if (dx >= 0){
		sgnX=1;
		cornerX=getFrameRightX()-refX+0.5*lw+nodeSep;
	    }
	    else {
		sgnX=-1;
		cornerX=refX-getFrameLeftX()+0.5*lw+nodeSep;
	    }
	    if (dy >= 0){
		sgnY=1;
		cornerY=getFrameTopY()-refY+0.5*lw+nodeSep;
	    }
	    else {
		sgnY=-1;
		cornerY=refY-getFrameBottomY()+0.5*lw+nodeSep;
	    }

	    if (cornerX==0 || cornerY==0){return nodeReferencePoint();}
	    
	    double dx2=dx*sgnX; //relative coordinates of the vector
	    double dy2=dy*sgnY;

	    double x,y;// local coordinates of result
	    // does (dx,dy) point above or below the corner ? Find out with determinant
	    if (cornerX*dy2-cornerY*dx2 >=0) { // (dx2,dy2) is above.
		y=cornerY;
		x=cornerY*dx2/dy2; //dy2 is non-zero here because det >=0
	    }
	    else {
		x=cornerX;
		y=cornerX*dy2/dx2;  //dy is non-zero here because det <=0
	    }
	    //return the point in normal coordinates
	    return new PicPoint(refX+x*sgnX,refY+y*sgnY);
	}
	else if (getFrameType()==TEXT_BOX_OVAL){//smallest area ellipse containing the box
	    // solution pour une ellipse donnee par x^2/a^2+y^2/b^2=1 passant par un point donne (x,y) : x^2/a^2=y^2/b^2=1/2  => a=sqrt(2)*x;
	    double lw=frameSep+((Double)(getAttribute(LINE_WIDTH))).doubleValue();
	    double radiusX = 0.5*frameWidth+nodeSep+0.5*lw; 
	    double radiusY = 0.5*(frameHeight+frameDepth)+nodeSep+0.5*lw; 
	    if (radiusX==0 || radiusY==0) return nodeReferencePoint();
	    double t=Math.sqrt(dx*dx/(radiusX*radiusX) + dy*dy/(radiusY*radiusY));
	    return new PicPoint(refX+dx/t,refY+dy/t);
	}
	else if (getFrameType()==TEXT_BOX_CIRCLE){
	    double lw=frameSep+((Double)(getAttribute(LINE_WIDTH))).doubleValue();
	    double radius=0.5*frameWidth+nodeSep+0.5*lw;
	    // faut-il le sortir du sqrt(0.5)?
	    if (radius==0) return nodeReferencePoint();
	    double t=dx*dx/(radius*radius) + dy*dy/(radius*radius);
	    return new PicPoint(refX+dx/t,refY+dy/t);
	}
	return null; // should not occur
    }


	///////////////////////////////
	//// STRING FORMATING
	///////////////////////////////

	public String toString(){

		return "PicText@" + Integer.toHexString(hashCode()) + ", text=\""
		       + getText() + "\""
		    + " width = " + width + " frame width = " + frameWidth
		    + " height = " + height + " frame height = " + frameHeight
		    + " depth = " + depth + " frame depth = " + frameDepth
		    + " xOffset = " + xOffset + " yOffset = " + yOffset
		       + super.toString();
	}

	////////////////////////////////
	//// GUI
	////////////////////////////////

	/**	
	 * Return a Customizer for geometry editing
	 */
	public AbstractCustomizer createCustomizer(){
		return new Customizer();
	}

	/**
	 * text string and geometry customizer
	 */
	class Customizer extends AbstractCustomizer implements ActionListener {

		/* widgets */
		private JTextField latexTF;
		private DecimalNumberField textAnchorXTF, textAnchorYTF;
		private boolean isListenersAdded = false; // flag

		/** create a new Customizer for editing this PicText geometry and attributes */
		public Customizer(){

			super();
			JPanel p = new JPanel(new GridLayout(2,1,5,5));

			// line 1: latex string
			p.add(latexTF = new JTextField(15));

			// line 2: anchor point label
			JPanel box;
			box = new JPanel(new GridLayout(1,3,5,5));
			box.add(PEToolKit.createJLabel("attributes.TextAnchorPoint"));
			// x-coordinate of anchor point
			box.add(textAnchorXTF = new DecimalNumberField(4));
			// y-coordinate of anchor point
			box.add(textAnchorYTF = new DecimalNumberField(4));
			
			p.add(box);

			// cyclic focus TAB
			textAnchorYTF.setNextFocusableComponent(latexTF);
			add(p, BorderLayout.NORTH);
		}
		
		/** add action listeners to widgets to reflect changes immediately */ 
		private void addActionListeners(){
			if (isListenersAdded) return; // already done
			latexTF.addActionListener(this);
			textAnchorXTF.addActionListener(this);
			textAnchorYTF.addActionListener(this);
			isListenersAdded = true;
		}


		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){

			latexTF.setText(getText());
			//latexTF.requestFocus();
			
			textAnchorXTF.setValue(getPointX(P_ANCHOR));
			textAnchorYTF.setValue(getPointY(P_ANCHOR));
			
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded 
			addActionListeners(); // done the first time load is called
		}

		/**
		 * update Element's properties
		 */
		public void store(){

			setText(latexTF.getText()); 
			PicPoint point1Frame = new PicPoint(textAnchorXTF.getValue(),textAnchorYTF.getValue());
			setPoint(P_ANCHOR,point1Frame,null);
		}

		public void actionPerformed(ActionEvent e){
			store();
		}

		/** manage focus */
		protected void activated(){
			latexTF.requestFocus();
		}

		/**
		 * @return the panel title, used e.g. for Border or Tabpane title. 
		 */
		public String getTitle(){ 
			return PicText.this.getName();
		}
		
	}



} // PicText
