/*
 AbstractMouseTransform.java - February 23, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.PageFormat;
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.BranchElement;
import jpicedt.graphic.model.PicAttributeSet;
import jpicedt.graphic.model.PicAttributeName;
import jpicedt.graphic.model.Arrow;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.view.View;
import jpicedt.graphic.view.AbstractView;
import jpicedt.graphic.view.DefaultViewFactory;
import jpicedt.graphic.view.HitInfo;

import java.awt.geom.AffineTransform;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Paint;

/**
 * This implementation allows subclassers to modify a clone
 * of the target Element, then to update the original Element when the operation is completed.
 * The aim is to reduce the burden of repainting a (possibly) complex Element while transforming it,
 * by painting the Element using very lightweight View. This is done as follow :
 * <ul>
 * <li> first clone the given Element, and add it to a transient RootElement defined as an inner class ;
 * <li> then create a "lightweight" View for this Element using a dedicated ViewFactory defined as an inner class ; 
 * <li> apply the transformation to the clone while the mouse is being dragged/moved/pressed ... Note that the 
 *      implementation of MouseTransform.paint() arranges for properly painting the clone by asking its
 *      lightweight View.
 * <li> when the transformation has completed, update the original Element and delete the clone.
 * </ul>
 */
public abstract class AbstractMouseTransform implements MouseTransform {

	private Element target,clone;
	private PECanvas canvas;
	private double scale=1.0; // hold the current scale factor b/w model and screen coordinate ; used by paint()

	/**
 	 * @param target The target-element upon which this transform will act.
	 */
	protected AbstractMouseTransform(Element target){
		this.target = target;
	}
	
	/**
	 * Called by mousePressed.<br>
	 * Creates a clone of the target (e.g. a clicked Element) given as a parameter to the constructor,
	 * and sets its View using a dedicated ViewFactory (defined as an inner class) so that only the stroke
	 * of the element will be painted (using a default basic stroke to reduce the burden of repainting).
	 */
	public void start(PEMouseEvent e){
		canvas = e.getCanvas();
		PageFormat pf = e.getCanvas().getPageFormat();
		AffineTransform at = pf.getModel2ViewTransform(e.getCanvas().getZoomFactor());
		scale = at.getScaleX();
		if (clone == null){
			clone = (Element)target.clone(); // this will clone the View as well, but we want our own View
			//e.getCanvas().getDrawing().addElement(clone); // for debugging purpose
			clone.setParent(new RootElement(clone));
			clone.setViewFromFactory(new ViewFactory(scale)); // set our own view
		}
	}

	/**
	 * Called by mouseReleased.<br>
	 * This implementation set the clone's parent and view to null so as to make it elligible for gc.
	 * This should be called by subclassers at the end of the implementation of their "next" method.
	 */
	public boolean next(PEMouseEvent e){
		clone.setParent(null); // make clone
		clone.removeView();    // eligible for GC
		clone = null;
		return false; // this is the end !
	}
		
	/**
	 * Paints the clone. This is normally called by EditorKit upon request by the installed PECanvas.
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		if (clone==null) return;
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("paint");
		clone.getView().paint(g, allocation);
		// [reynal:21/01/2003] ok, let's try to paint the highlighter as well :
		g.setPaint(Color.blue);
		clone.getView().paintHighlighter(g, allocation, scale);
	}
	
	/**
	 * Return the element this transform acts upon
	 */
	protected Element getTarget(){
		return target;
	}
	
	/**
	 * Return the clone of the target element, which was init'd by <code>startTransform</code>
	 */
	protected Element getClone(){
		return clone;
	}

	///////////////////////////////////////////////////////////////////////
	
	/**
	 * a ViewFactory that produces view suited for MouseTransform's, ie with a minimal set of attributes
	 * so as to reduce processor burden during the transform operation (e.g. moving, resizing,...).<br>
	 * Note that adding new "shared" methods in  DefaultViewFactory (e.g. createGradient, like) IMPOSES
	 * us to override them here IF we want the corresponding feature not to be painted during
	 * the MouseTransform. This might be fairly cumbersome in the future, yet i don't see any other solution.
	 */
	private class ViewFactory extends DefaultViewFactory {
		
		double scale;
		
		/** 
		 * @param scale the current scaling factor b/w model and screen coordinate system ; this
		 *        usually depends on the current zoom factor in PECanvas, and may be retrieved using
		 *        the PageFormat associated with the canvas. Scale is needed here to construct a proper
		 *        BasicStroke object.
		 */
		ViewFactory(double scale){
			this.scale = scale;
		}
		
		/**
		 * Simply calls DefaultViewFactory, since this factory does't produce specialized View's by itself.  
		 */
		public View createView(Element e){
			return super.createView(e);
		}
		/**
		 * return a basic stroke instead of the (possibly) pretty complex stroke created by DefaultViewFactory.
		 */
		protected Stroke createStroke(PicAttributeSet set){
			return new BasicStroke((float)(0.5/scale));  
		}
	
		/**
		 * return the color "blue"  
		 */
		protected Paint createPaintForOutline(PicAttributeSet set){
			return Color.blue;
		}

		/**
		 * return null : no interior is painted during a mouse transform.
		 */
		protected Paint createPaintForInterior(PicAttributeSet set){
			return null;
		}

		/**
		 * would normally paint hatches (vlines,...) ; overriden to do nothing.
	 	 */
		protected void paintHatches(Graphics2D g, PicAttributeSet set, Shape shape){
		}
		
		/** overriden to do nothing (no shadow is rendered during a mouse transform) */
		protected void paintShadow(Graphics2D g, PicAttributeSet set, Shape shape){
		}
		
		/**
		 * Returns null (no arrow is rendered during a mouse transform operation)
		 */
		protected Arrow createArrow(PicAttributeSet set, PicAttributeName direction){
			return null;
		}
		/**
		 * Return null (no dotting is rendered during a mouse transform operation)
		 */
		 protected Shape createDotsForPolygons(PicAttributeSet set){
			 return null;
		 }
		
	}
			
	//////////////////////////////////////////////////////////
	/** RootElement hosts the target-Element being mouse-transformed during the operation */  
	private class RootElement extends BranchElement {
		View v;
		/** directly add the given "clone" to the branch at init time (this is much simpler than using addChild) */
		public RootElement(Element clone){
			super();
			v = new RootView(this);
		}
		
		public Object clone(){
			return null; // not used here
		}
		public View getView(){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("getView="+v);
			return v;
		}
		public String getName(){
			return "RootElement";
		}
	}
	
	///////////////////////////////////////////////////////////
	/** view for the RootElement only */
	private class RootView extends AbstractView {
		
		public RootView(RootElement e){
			super(e);
		}
		
		public void changedUpdate(DrawingEvent.EventType e){
		}
	
		/** return PECanvas : this is mandatory, as there's no other way for a View to fetch its container
		 *  than calling the root of the view-tree ! */
		public PECanvas getContainer() {
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("getContainer="+canvas);
			return  canvas;
		}
		/** does nothing */
		public HitInfo hitTest(PEMouseEvent e){return null;}
		/** does nothing */
		public void paint(Graphics2D g, Rectangle2D r){}
	}
	
} // MouseTransform
