/*
 AbstractMouseTransformFactory.java - August 23, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.event.PEMouseEvent;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;

/**
 * Provide a skeletal implementation of the MouseTransformFactory interface.
 * This include some generic MouseTransform's which may be easily extended by
 * concrete subclassers, like SelectAreaMouseTransform. 
 * @author Sylvain Reynal
 * @since jPicEdt 1.4
 * @version $Id: AbstractMouseTransformFactory.java,v 1.4.2.1 2007/09/02 11:56:27 reynal Exp $
 */
abstract public class AbstractMouseTransformFactory implements MouseTransformFactory {

	private EditorKit kit;
	private CursorFactory cursorFactory=new CursorFactory();

	/**
	 * @param kit the hosting editor-kit
	 */
	public AbstractMouseTransformFactory(EditorKit kit){
		this.kit = kit;
	}

	/**
	 * Return the hosting EditorKit for mouse-tools which work with this factory.
	 * This may be used to work with the selection-handler and use hitTest() methods.
	 */
	public final EditorKit getEditorKit(){
		return kit;
	}
	
	/**
	 * Allows the MouseTransformFactory to do specific graphic rendering when it's installed in a
	 * hosting SelectionTool. This default implementation does nothing.
	 * @since jpicedt 1.4
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
	}
	
	/** 
	 * called when the associated SelectionTool is being activated in the hosting EditorKit.
	 * Initialization work required before any mouse-event occurs should be done here.
	 * Other initialization work may be carried out in the MouseTransform's themselves.
	 * <p> This implementation does nothing.
	 */
	public void init(){}

	/** called when the associated SelectionTool is being deactivated in the hosting EditorKit.
	 * This provides a way for the factory to do some final clean-up, e.g. local buffers, graphic context,...
	 * <p> This implementation does nothing.
	 */
	public void flush(){}
	

	/////////////////////////////////////////////////////////////////////////////////
	//// INVALID MT
	/////////////////////////////////////////////////////////////////////////////////

	/**
	 * This mouse transform does nothing, it simply returns an invalid cursor, and can be used
	 * by MouseTransform factories to notify the user of an invalid UI action.
	 * @since jpicedt 1.3.2
	 */
	protected class InvalidMouseTransform implements MouseTransform {

		/**
		 * Called when the mouse is pressed. The transform should do the initialization work here.
		 * @since jpicedt 1.3.2
		 */
		public void start(PEMouseEvent e){}

		/**
		 * Called when the mouse is dragged. Element geometry update should occur here.
		* @since jpicedt 1.3.2
		 */
		public void process(PEMouseEvent e){}

		/**
		 * Called when the mouse is released
		 * @since jpicedt 1.3.2
		 */
		public boolean next(PEMouseEvent e){return false;}

		/**
		 * @return the cursor for this MouseTransform
		 * @since jpicedt 1.3.2
		 */
		public Cursor getCursor(){
			return new CursorFactory().getPECursor(CursorFactory.INVALID);
		}

		/**
		 * Allows the MouseTransform to do specific graphic rendering when it's operating.
		 * @since jpicedt 1.3.2
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		}

		/**
		 * @return a help-message for the UI, that makes sense with this transform.
		 */
		public String getHelpMessage(){
			return "help-message.InvalidAction";
		}

	} // InvalidMouseTransform

	
	/**
	 * This mouse transform does nothing, it is just a convenience to post a HelpMessage (regarding
	 * what the user COULD do) when no other mouse-transform makes sense with the current mouse-event.
	 * @since jpicedt 1.3.2
	 */
	protected class HelpMessageMouseTransform implements MouseTransform {

		String helpMsg;
		
		public HelpMessageMouseTransform(String helpMsg){
			this.helpMsg=helpMsg;
		}
		
		/**
		 * Called when the mouse is pressed. The transform should do the initialization work here.
		 * @since jpicedt 1.3.2
		 */
		public void start(PEMouseEvent e){}

		/**
		 * Called when the mouse is dragged. Element geometry update should occur here.
		* @since jpicedt 1.3.2
		 */
		public void process(PEMouseEvent e){}

		/**
		 * Called when the mouse is released
		 * @since jpicedt 1.3.2
		 */
		public boolean next(PEMouseEvent e){return false;}

		/**
		 * @return the cursor for this MouseTransform
		 * @since jpicedt 1.3.2
		 */
		public Cursor getCursor(){
			return new CursorFactory().getPECursor(CursorFactory.SELECT);
		}

		/**
		 * Allows the MouseTransform to do specific graphic rendering when it's operating.
		 * @since jpicedt 1.3.2
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		}

		/**
		 * @return a help-message for the UI, that makes sense with this transform.
		 */
		public String getHelpMessage(){
			return helpMsg;
		}

	} // HelpMessageMouseTransform


	/////////////////////////////////////////////////////////////////////////////////
	//// Select from rectangular area
	/////////////////////////////////////////////////////////////////////////////////

	/**
	 * a mouse-transform dedicated to select objects (Element's, control-point's,...) 
	 * inside a rectangle dragged by the user. Only a skeletal implementation is provided here.
	 * This implementation grows the selection rectangle,
	 * yet where *really* selecting things is concerned, subclassers may override methods by
	 * calling super.method_name before hand.
	 */
	protected class SelectAreaTransform implements MouseTransform {

		private Line2D.Double diag = new Line2D.Double();
		private double scale=0.0; 		// cache for scale factor
		private BasicStroke lineStroke = new BasicStroke(1.0f); 		// cache for BasicStroke

		/**
		 * Return the selection rectangle as built from the first mouse-pressed event and the
		 * current drag position.
		 */
		protected final Rectangle2D getSelectionRectangle(){
			return diag.getBounds2D();
		}
		
		/** 
		 * Invoked by mousePressed(). This implementation simply initializes the geometry
		 * of the selection rectangle. Override if you must carry out specific selection operations
		 * on mousePressed().
		 */
		public void start(PEMouseEvent e){
			diag.x1 = diag.x2 = e.getPicPoint().x;
			diag.y1 = diag.y2 = e.getPicPoint().y;
		}
		
		/** Called when the mouse is dragged. This simply grows the selection rectangle.
		 * Override if you must carry out specific selection operations while the mouse is being dragged.
		 */
		public void process(PEMouseEvent e){
			Rectangle2D oldclip = getClipRectangle();
			diag.x2 = e.getPicPoint().x;
			diag.y2 = e.getPicPoint().y;
			Rectangle2D clip = getClipRectangle();
			clip.add(oldclip);
			e.getCanvas().repaintFromModelRect(clip);
		}

		/** 
		 * Called when the mouse is released. This implementation simply repaint the selection
		 * rectangle. 
		 * Override if you must carry out specific selection operations while the mouse is being released.
		 */
		public boolean next(PEMouseEvent e){
			Rectangle2D clip = getClipRectangle();
			e.getCanvas().repaintFromModelRect(clip);
			return false;
		}

		/**
		 * render the selection rectangle
		 * @param scale Graphics2D's current scale factor b/w model and view
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			if (this.scale != scale){ // update stroke after zoom changed
				this.scale = scale;
				// scale thickness down so that it's displayed with the same thickness whatever value the current zoom factor may have
				lineStroke = new BasicStroke((float)(2.0/scale));
			}
			g.setPaint(Color.blue); // [pending] fetch color from Properties ?
			g.setStroke(lineStroke);
			g.draw(diag.getBounds2D());
		}

		/**
		 * Returns a rectangle defining the clip boundary for this mouse-transform, in model-coordinates 
		 */
		public Rectangle2D getClipRectangle() {

			Rectangle2D rect2D = diag.getBounds2D();
			rect2D.setFrame(
				rect2D.getX()-2*lineStroke.getLineWidth(),
			    rect2D.getY()-2*lineStroke.getLineWidth(),
			    rect2D.getWidth()+4*lineStroke.getLineWidth(),
			    rect2D.getHeight()+4*lineStroke.getLineWidth());
			return rect2D;
		}
		
		/**
		 * @return a help-message for the UI, that makes sense with this transform.
		 */
		public String getHelpMessage(){
			return "help-message.SelectArea";
		}

		/** @return a textual representation of this transform for debugging purpose */
		public String toString(){
			return "[SelectAreaTransform]";

		}

		/**
		 * @return a cursor adequate with this mouse-transform, delegating to CursorFactory.
		 */
		public Cursor getCursor(){
			return cursorFactory.getPECursor(CursorFactory.SELECT);
		}
	} // SelectAreaTransform




} // MouseTransformFactory

