/*
 CustomizerDialog.java - December 31, 2001 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.Localizer;

import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ContainerAdapter;
import java.awt.event.ContainerEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * A dialog box that lays out an array of AbstractCustomizer's inside a JPanel or
 * a JTabbedPane, depending on the number of customizers fed to the constructor.
 * It further provides four default actions, namely : ok, cancel, load default, reload.
 * @author Sylvain Reynal
 * @since jPicEdt 1.3.2
 * @version $Id: CustomizerDialog.java,v 1.9.2.1 2007/09/02 11:56:28 reynal Exp $
 */
public class CustomizerDialog extends JDialog {

	/** mask to display no button at all */	
	public static final int NO_BUTTON=0;
	/** mask to display the OK button */	
	public static final int OK_BUTTON=1;
	/** mask to display the CANCEL button */
	public static final int CANCEL_BUTTON=2;
	/** mask to display the DEFAULT button */
	public static final int DEFAULT_BUTTON=4;
	/** mask to display the RELOAD button */
	public static final int RELOAD_BUTTON=8;
	/** mask to display the OK and CANCEL buttons */
	public static final int OK_CANCEL_BUTTONS = 3;
	/** mask to display all buttons */
	public static final int ALL_BUTTONS = 15;

	// array of inner-panels
	private AbstractCustomizer[] customizers;

	// UI buttons
	private JButton buttonOk, buttonCancel, buttonDefault, buttonReload;

	// user pressed cancel ?
	private boolean cancelled = false;
	
	// keyListener for ESC and ENTER keys handling
	private KeyHandler keyHandler;
	
	private boolean okButtonClosesDialog = false;

	/**
	 * Creates a new dialog box from the given array of AbstractCustomizer's, laying
	 * them out in a JTabbedPane.<br>
	 * By default, clicking the OK button does not close this dialog box, this must be set
	 * separately by invoking {@link #setOkButtonClosesDialog setOkButtonClosesDialog}.
	 * @param title the JDialog title ; if null, we use the first customizer title.
	 * @param selected index of the selected customizer on start-up
	 */
	public CustomizerDialog(AbstractCustomizer[] customizers, String title, int selected) {
		this(customizers, title, selected, ALL_BUTTONS);
	}
	
	/**
	 * Creates a new dialog box from the given array of AbstractCustomizer's, laying
	 * them out in a JTabbedPane.<br>
	 * By default, clicking the OK button does not close this dialog box, this must be set
	 * separately by invoking {@link #setOkButtonClosesDialog setOkButtonClosesDialog}.
	 * @param title the JDialog title ; if null, we use the first customizer title.
	 * @param selected index of the selected customizer on start-up
	 * @param buttonsMask buttons to be displayed : a mask computed from predefinite masks OR'd together
	 */
	public CustomizerDialog(AbstractCustomizer[] customizers, String title, int selected, int buttonsMask) {
		this.customizers = customizers;
		if (title == null){
			setTitle(customizers[0].getTitle());
		}
		else setTitle(title);
		preInit(buttonsMask);

		/* root tabbed pane -> addTab(title,icon,component,tip)) */
		JTabbedPane tabbedPane = new JTabbedPane();
		tabbedPane.addChangeListener(new ChangeHandler());
		for (int i=0; i<customizers.length; i++){
			if (customizers[i]==null) jpicedt.Log.error("customizers["+i+"] is null !");
			else tabbedPane.addTab(customizers[i].getTitle(),customizers[i].getIcon(),customizers[i],customizers[i].getTooltip());
		}
		getContentPane().add(tabbedPane,BorderLayout.CENTER);
		for (int i=0; i<customizers.length; i++){
			try {
				customizers[i].load();
			}
			catch(NumberFormatException ex){
				if (jpicedt.Log.DEBUG) ex.printStackTrace();
				JOptionPane.showMessageDialog(null,
				                              Localizer.currentLocalizer().get("NumberFormatError") + " : " + ex.getMessage() + ", " + customizers[i].getTitle(),
				                              Localizer.currentLocalizer().get("Preferences"),JOptionPane.ERROR_MESSAGE);
			}
		}
		initGeometry();
		if (selected < 0) selected = 0;
		if (selected > customizers.length-1) selected = customizers.length-1;
		tabbedPane.setSelectedIndex(selected);
	}

	/**
	 * Builds a new dialog box from a single customizer.<br>
	 * By default, clicking the OK button does not close this dialog box, this must be set
	 * separately by invoking {@link #setOkButtonClosesDialog setOkButtonClosesDialog}.
	 */
	public CustomizerDialog(AbstractCustomizer customizer) {
		this(customizer, ALL_BUTTONS);
	}
	
	/**
	 * Builds a new dialog box from a single customizer.<br>
	 * By default, clicking the OK button does not close this dialog box, this must be set
	 * separately by invoking {@link #setOkButtonClosesDialog setOkButtonClosesDialog}.
	 * @param buttonsMask buttons to be displayed : a mask computed from predefinite masks OR'd together
	 */
	public CustomizerDialog(AbstractCustomizer customizer, int buttonsMask) {

		customizers = new AbstractCustomizer[1];
		customizers[0] = customizer;
		setTitle(customizer.getTitle());
		preInit(buttonsMask);
		getContentPane().add(customizer,BorderLayout.CENTER);
		customizer.load();
		initGeometry();
	}

	/**
	 * @return true if the user pressed the Cancel action
	 */
	public boolean isCancelled(){
		return cancelled;
	}

	/**
	 * If this dialog was constructed with more than one inner-pane, this
	 * methods is invoked when a new pane gets activated. 
	 */
	public void paneActivated(AbstractCustomizer c){
		c.activated();
	}

	/**
	 * Whether clicking the OK button, beside storing widgets content, also closes this dialog or not.
	 * It is valid to call this method at any time, e.g. to change behaviour once this dialog has
	 * been made visible.
	 */
	public void setOkButtonClosesDialog(boolean b){
		okButtonClosesDialog = b;
	}
	
	//////////////////////// PRIVATE MEMBERS /////////////////////////

	/**
	 * JDialog's attributes initialization.
	 */
	private void preInit(int buttonsMask){
		setModal(true);
		setResizable(true); // pb with pack() on some WindowMaker releases
		enableEvents(AWTEvent.WINDOW_EVENT_MASK);
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE); 
		getContentPane().setLayout(new BorderLayout(5,5)); //hgap, vgap
		getContentPane().add(createButtonPanel(buttonsMask),BorderLayout.SOUTH);
		addListeners();
	}

	/**
	 * JDialog's geometry initialization
	 */
	private void initGeometry(){
		//setSize(getPreferredSize()); // pb with pack() not working on some Linux/WindowMakers releases
		pack();
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension frameSize = getSize();
		if (frameSize.height > screenSize.height)
			frameSize.height = screenSize.height;
		if (frameSize.width > screenSize.width)
			frameSize.width = screenSize.width;
		setLocation((screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);
		//setVisible(true);
	}
	
	/**
	 * Register interested listeners
	 */
	private void addListeners(){
		addKeyListener(keyHandler=new KeyHandler());
		addWindowListener(new WindowHandler());
		((Container)getLayeredPane()).addContainerListener(new ContainerHandler()); // needs keyHandler !
		getContentPane().addContainerListener(new ContainerHandler());
	}

	/**
	 * Creates default buttons according to the given buttonsMask 
	 */
	private JPanel createButtonPanel(int buttonsMask){

		JPanel panelButtons = new JPanel(); //row, col, hgap, vgap
		panelButtons.setBorder(BorderFactory.createEtchedBorder());
		ActionListener actionHandler = new ActionHandler();
		if ((buttonsMask & OK_BUTTON)!=0){
			buttonOk = new JButton();
			buttonOk.setText(Localizer.currentLocalizer().get("button.OK"));
			buttonOk.addActionListener(actionHandler);
			panelButtons.add(buttonOk);
		}
		if ((buttonsMask & CANCEL_BUTTON)!=0){
			buttonCancel = new JButton();
			buttonCancel.setText(Localizer.currentLocalizer().get("button.Cancel"));
			buttonCancel.addActionListener(actionHandler);
			panelButtons.add(buttonCancel);
		}
		if ((buttonsMask & DEFAULT_BUTTON)!=0){
			buttonDefault = new JButton();
			buttonDefault.setText(Localizer.currentLocalizer().get("button.Default"));
			buttonDefault.addActionListener(actionHandler);
			panelButtons.add(buttonDefault);
		}
		if ((buttonsMask & RELOAD_BUTTON)!=0){
			buttonReload = new JButton();
			buttonReload.setText(Localizer.currentLocalizer().get("button.Reload"));
			buttonReload.addActionListener(actionHandler);
			panelButtons.add(buttonReload);
		}
		return panelButtons;
	}


	/**
	 * Helper for event handlers ; code to be executed after a click on the "ok" button occured,
	 * or whatever similar. Validate user entries and store them.
	 */
	private void ok(){
		for (int i=0; i<customizers.length; i++){
			customizers[i].store();
		}
		dispose();
	}

	/**
	 * Helper for event handlers ; code to be executed after a click on the "cancel" button occured,
	 * or whatever similar. Close dialog w/o saving
	 */
	private void cancel(){
		cancelled = true;
		dispose();
	}

	/**
	 * Helper for event handlers ; code to be executed after a click on the "load" button occured,
	 * or whatever similar. Load default values
	 */
	private void loadDefault(){
		for (int i=0; i<customizers.length; i++){
			customizers[i].loadDefault();
		}
	}

	/**
	 * Helper for event handlers ; code to be executed after a click on the "reload" button occured,
	 * or whatever similar. Reload widgets with previously stored preferences
	 */
	private void reload(){
		for (int i=0; i<customizers.length; i++){
			try {
				customizers[i].load();
			}
			catch(NumberFormatException ex){
				if (jpicedt.Log.DEBUG) ex.printStackTrace();
				JOptionPane.showMessageDialog(null,
				                              Localizer.currentLocalizer().get("NumberFormatError") + " : " + ex.getMessage()
				                              + ", " + customizers[i].getTitle(),
				                              Localizer.currentLocalizer().get("Preferences"),
				                              JOptionPane.ERROR_MESSAGE);
			}
		}
	}
	
	//////////////////////////////////////// EVENT HANDLERS ///////////////////

	/** windowListener */
	class WindowHandler extends WindowAdapter
	{
		public void windowClosing(WindowEvent evt){
			cancel();
		}
	}

	/** actionListener for JButton's */
	class ActionHandler implements ActionListener {

		/**
		 * Handles button related actions : store pref. to file, load default values,...
		 * This method is public as an implementation side-effet.
		 */
		public void actionPerformed(ActionEvent e) {

			if (e.getSource()==buttonOk) ok();
			else if (e.getSource()==buttonCancel) cancel();
			else if (e.getSource()==buttonDefault) loadDefault();
			else if (e.getSource()==buttonReload) reload();
		}
	}

	/** keyListener for ESC and ENTER keys ; adapted from jEdit */
	class KeyHandler extends KeyAdapter {

		public void keyPressed(KeyEvent evt){
			//jpicedt.Log.debug(evt.toString());
			if(evt.isConsumed()) return;
			
			if(evt.getKeyCode() == KeyEvent.VK_ENTER && okButtonClosesDialog){
				// crusty workaround
				Component comp = getFocusOwner();
				while(comp != null){
					//jpicedt.Log.debug("comp="+comp);
					if(comp instanceof JComboBox){
						JComboBox combo = (JComboBox)comp;
						if(combo.isEditable()){
							Object selected = combo.getEditor().getItem();
							if(selected != null)
								combo.setSelectedItem(selected);
						}
						break;
					}
					comp = comp.getParent();
				}

				ok();
				evt.consume();
			}
			else if(evt.getKeyCode() == KeyEvent.VK_ESCAPE){
				cancel();
				evt.consume();
			}
		}
	}

	/** stateChanged listener for JTabbedPane related event */
	class ChangeHandler implements ChangeListener {
	
		/** this method is public as an implementation side-effect */
		public void stateChanged(ChangeEvent e){
			JTabbedPane tabbedPane = (JTabbedPane)e.getSource();
			AbstractCustomizer c = (AbstractCustomizer)tabbedPane.getSelectedComponent();
			paneActivated(c);
		}
	}

	// Recursively adds our key listener to sub-components
	// code snippet adapted from jEdit
	class ContainerHandler extends ContainerAdapter
	{
		public void componentAdded(ContainerEvent evt){
			componentAdded(evt.getChild());
		}

		public void componentRemoved(ContainerEvent evt){
			componentRemoved(evt.getChild());
		}

		private void componentAdded(Component comp){
			comp.addKeyListener(keyHandler);
			if(comp instanceof Container){
				Container cont = (Container)comp;
				cont.addContainerListener(this);
				Component[] comps = cont.getComponents();
				for(int i = 0; i < comps.length; i++){
					componentAdded(comps[i]);
				}
			}
		}

		private void componentRemoved(Component comp){
			comp.removeKeyListener(keyHandler);
			if(comp instanceof Container){
				Container cont = (Container)comp;
				cont.removeContainerListener(this);
				Component[] comps = cont.getComponents();
				for(int i = 0; i < comps.length; i++){
					componentRemoved(comps[i]);
				}
			}
		}
	}
	
} 


