/*
 DefaultSelectionHandler.java - September 22, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.PECanvas;
import jpicedt.graphic.SelectionHandler;

import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.BranchElement;
import jpicedt.graphic.model.PicGroup;
import jpicedt.graphic.model.PicObjectConstants;

import jpicedt.graphic.event.DrawingListener;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.event.PEMouseEvent;

import jpicedt.graphic.view.CompositeView;
import jpicedt.graphic.view.DefaultHighlighterFactory;
import jpicedt.graphic.view.Highlighter;
import jpicedt.graphic.view.HitInfo;
import jpicedt.graphic.view.ViewConstants;

import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

/**
 * Stores references to selected Element's.<p>
 * This handler has the same capabilities as jpicedt.graphic.model.PicGroup, except that 
 * it does not belong to any jpicedt.graphic.model.Drawing.RootElement. 
 * Hence it is guaranteed that this handler will NEVER notify the parent document (=the hosting Drawing)
 * when content gets added to/removed from it, since it has no parent. This design choice is aimed at
 * avoiding redundant event generation, since (for instance) deleting a selected element would then trigger
 * the same DrawingEvent twice (one on behalf of this handler, the other one on behalf of the hosting Drawing). 
 * Content modifying method are overriden so as to reflect this.
 * @since jPicEdt 1.4
 * @author Sylvain Reynal
 * @version $Id: DefaultSelectionHandler.java,v 1.9.2.1 2007/09/02 11:56:28 reynal Exp $
 */
public class DefaultSelectionHandler extends PicGroup implements SelectionHandler, DrawingListener, ViewConstants, PicObjectConstants {

	/** 
	 * Highlighting mode corresponding to each Element belonging to the selection is highlighted 
	 */
	public static final String LOCAL_MODE = DefaultHighlighterFactory.CompositeHighlighter.LOCAL_MODE;

	/** 
	 * Highlighting mode corresponding to
	 * the selection-handler behaving as if it were a PicGroup, 
	 * hence its 8 control-points only are highlighted 
	 */
	public static final String GLOBAL_MODE = DefaultHighlighterFactory.CompositeHighlighter.GLOBAL_MODE;
	
	/** hosting EditorKit */
	private EditorKit kit;
	// inherited : 
	// protected boolean changeLock
	// protected ArrayList children

	/**
	 * construct a new selection handler for this editor kit, with SelectionHandlerView as the default view.
	 * Highlighting mode is GLOBAL by default.
	 */
	public DefaultSelectionHandler(EditorKit kit){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init>");
		this.kit=kit;
		view = new CompositeView(this);
		Highlighter h = kit.getHighlighterFactory().createHighlighter(this);
		view.setHighlighter(h);
		setHighlightingMode(LOCAL_MODE);
	}

	/**
	 * set the current highlighting mode ; 
	 * this influences the way the Highligther is painted, but also the result returned by HitTest.
	 * @param mode or LOCAL_MODE or GLOBAL_MODE
	 */
	public void setHighlightingMode(String mode){
		updateBoundingBox();
		Highlighter h = view.getHighlighter();
		if (h instanceof DefaultHighlighterFactory.CompositeHighlighter){
			DefaultHighlighterFactory.CompositeHighlighter hc = (DefaultHighlighterFactory.CompositeHighlighter)h;
			hc.setHighlightingMode(mode);
		}
	}

	/**
	 * Return the current highlighting mode
	 */
	public String getHighlightingMode(){
		Highlighter h = view.getHighlighter();
		if (h instanceof DefaultHighlighterFactory.CompositeHighlighter){
			DefaultHighlighterFactory.CompositeHighlighter hc = (DefaultHighlighterFactory.CompositeHighlighter)h;
			return hc.getHighlightingMode();
		}
		return null;
	}

	/**
	 * Toggle the current highlighting mode
	 */
	public void toggleHighlightingMode(){
		updateBoundingBox();
		Highlighter h = view.getHighlighter();
		if (h instanceof DefaultHighlighterFactory.CompositeHighlighter){
			DefaultHighlighterFactory.CompositeHighlighter hc = (DefaultHighlighterFactory.CompositeHighlighter)h;
			hc.toggleHighlightingMode();
		}
	}
		
	/** paint the selection handler highlighting */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		view.paintHighlighter(g,allocation,scale);
	}
	
	/** hitTest on the content of the selection only */
	public HitInfo hitTest(PEMouseEvent me){
		return view.hitTest(me,true); // highlighter.hitTest(), then possibly view.hitTest()
	}
		

	///////////// SelectionHandler interface methods /////////////////////

	/**
	 * @return an iterator over the selected Element's
	 */
	public Iterator elements(){
		return children();
	}

	/**
	 * @return the selected elements wrapped in a Collection
	 */
	public Collection asCollection(){
		return this;
	}

	/**
	 * @return the selected elements wrapped in an array (convenience call to asCollection)
	 */
	public Element[] asArray(){
		return (Element[])toArray(new Element[0]);
	}

	/**
	 * @return whether the given Element is selected or not
	 * @param includeAncestors if true, return true if this Element or one of its ancestors is selected. 
	 */
	public boolean isSelected(Element e, boolean includeAncestors){
		if (e == this) return false; // selection-handler is never selected !
		if (contains(e)) return true;
		if (includeAncestors && e.getParent()!=null) return isSelected(e.getParent(), true);
		else return false;
	}

	/**
	 * Select all Element's in the given drawing.
	 */
	public void selectAll(Drawing d){
		unSelectAll(); // avoid duplicate
		for (Iterator it=d.elements(); it.hasNext();){
			Element e = (Element)it.next();
			addToSelection(e);
		}
	}

	/**
	 * Unselect all Element's in the given drawing.
	 */
	public void unSelectAll(){
		removeAllChildren();
	}

	/**
	 * Add the given Element to the current selection
	 */
	public void addToSelection(Element e){
		addChild(e);
	}

	/**
	 * Replace the current selection with the given element. This is a convenience call to unSelectAll
	 * then addToSelection.
	 */
	public void replaceSelection(Element e){
		unSelectAll();
		addToSelection(e);
	}

	/**
	 * Unselect the given Element, ie remove it from the current selection.
	 */
	public void unSelect(Element e){
		removeChild(e);
	}

	/**
	 * Delete all selected Element's from the given Drawing, and remove the reference to them
	 * from the SelectionHandler.
	 */
	public void delete(Drawing dr){
		for (Iterator it = elements(); it.hasNext();){ // iterate over selected elements
			Element e = (Element)it.next();
			it.remove(); // first remove element from selection-handler to avoid concurent modification (see note below)
			dr.removeElement(e); // then remove it from drawing => fires REMOVE DrawingEvent => calls changedUpdate => *might* raise concurrent modification exception if not taken care of properly
		}
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.REMOVE); // update view
	}

	/////// overriden from PicGroup so as to leave the added/removed element's parent unchanged. ////////

	/**
	 * overriden so that the parent of the added element isn't set to this, and its view is left unchanged.
	 */
	public void addChild(Element child){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("child="+child);
		if (child == this) {
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Can't add selection-handler to itself !!!");
			return;
		}
		if (contains(child)) {
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Already selected!");
			return; // already selected
		}
		// [pending] rebuild buffer from scratch so as to preserve z-ordering in the drawing (use Collections.sort
		//           with an appropriate Comparator)
		children.add(child);
		sort();
		
		// leave parent and view unchanged
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.INSERT); // update view
	}

	/**
	 * overriden so as to leave child's parent and view unchanged.
	 * @param child if child is the selection-handler itself, call removeAllChildren().
	 */
	public void removeChild(Element child){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("child="+child);
		if (child == this) removeAllChildren(); // terrible hack to deselect all if click on a selection-handler's end-point
		else {
			if (!contains(child)) {
				if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Already unselected!");
				return; // already unselected
			}
			children.remove(child);
			
			updateBoundingBox();
			fireChangedUpdate(DrawingEvent.EventType.REMOVE); // update view
		}
	}

	/**
	 * overriden so as to leave children parent and view unchanged.
	 */
	public void removeAllChildren(){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeAllChlidren");
		if (isEmpty()){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("already empty!");
			return;
		}
		children.clear();
		updateBoundingBox();
		fireChangedUpdate(DrawingEvent.EventType.REMOVE); // update view
	}

    /**
     * Gets the parent of the element.
     * @return the current Drawing.RootElement attached to the hosting editor-kit ; used
	 * mainly by the attached view to fetch its container.
     */
	public Element getParent() {
		PECanvas c = kit.getCanvas(); 
	    if (c!=null) return c.getDrawing().getRootElement();
		else return null;
	}
	
	/**
	 * Sort selected elements according to their z-ordering in the given drawing
	 */
	public void sort(){
		PECanvas canvas = kit.getCanvas();
		if (canvas==null) return;
		Comparator comparator = new ZOrderingComparator(canvas.getDrawing());
		Collections.sort(children, comparator);
	}
	
	/**
	 *
	 */
	private class ZOrderingComparator implements Comparator {
	
		private Drawing dr;
		
		public ZOrderingComparator(Drawing dr){
			this.dr = dr;
		}
		
		/** 
		 * Returns a positive number whenever o1 > o2, etc...
		 */
		public int compare(Object o1, Object o2){
			BranchElement be = dr.getRootElement();
			int i1 = be.getChildIndex((Element)o1);
			int i2 = be.getChildIndex((Element)o2);
			return i1 - i2;
		}
	}
	
	/////////////////////////////////////
	/// EVENTS
	////////////////////////////////////
	
	/**
	 * Called each time this DefaultSelectionHandler changes. 
	 * This implementation is overriden from AbstractElement so as to update the associated View only, i.e. there is
	 * no event being forwarded to the parent (which here is the Drawing.RootElement, see e.g. getParent()).
	 * The point is that this method gets called whenever setPoint/scale/translate/... is called on the
	 * selection-handler, which indirectly triggers a DrawingEvent on behalf of the selected elements themselves.
	 * Hence it doesn't make sense to post these events twice.
	 * <br>
	 * If subclasser are willing to override this method, they should call <code>super.fireChangeUpdate</code>.
	 * @param type the event type
	 */
	protected void fireChangedUpdate(DrawingEvent.EventType eventType){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType+", this="+toString());
		if (view != null) view.changedUpdate(eventType);
		// getParent().forwardChangedUpdate(this,eventType); => commented out so as not to post the same event twice. 
	}
	
	
	//////////////////////////////////////////////
	
	/**
	 * Returns a String representing the group for debugging use only.
	 */
	public String toString(){

		String s = "[DefaultSelectionHandler@" + Integer.toHexString(hashCode()) + "{";
		int j=0;
		for(Iterator i = children(); i.hasNext(); j++){
			StringBuffer buf = new StringBuffer(((Element)i.next()).toString());
			int pos=0;
			while(true){
				pos = buf.toString().indexOf('\n',pos);
				if (pos == -1) break;
				buf.insert(pos+1,'\t');
				pos += 2;
			}
			s += "\n\t" + j + ":" + buf;
		}
		return s + "\n}]";
	}
	

	/////////////////////// DrawingListener interface ////////////
	public void _changedUpdate(DrawingEvent e){
		jpicedt.Log.debug(e.toString());
	}
	/**
	 * Implementation of DrawingListener interface aimed at keeping this selection-handler
	 * always synchronized with the Drawing, especially when the Drawing content gets modified
	 * directly throught the Drawing API (as opposed to using selection-related methods in PECanvas).<p>
	 * This method is invoked when an element changed in the Drawing.<p>
	 * If this is a REMOVE event type, and if the removed children (of the changed Element)
	 * were selected, we remove these children from the selection as well, ie we 
	 * always keep the content of the selection-handler
	 * synchronized with the associated Drawing.
	 */
	public void changedUpdate(DrawingEvent e){
		if (jpicedt.Log.DEBUG) 
			jpicedt.Log.debug("DrawingEvent="+e+"\nCHANGE_LOCK="+changeLock);
		if (e.getType()==DrawingEvent.EventType.REMOVE || e.getType()==DrawingEvent.EventType.REPLACE){
			// fetch the parent from which children were removed/replaced : (e.g. Drawing.RootElement if Drawing.removeElement/replaceElement triggered this event)
			// if its removed/replaced children are still in the selectionHandler, remove them from the selection as well
			Element changed = e.getElement();
			
			// if "changed" is NOT a Drawing.RootElement, e.g. an element got removed from a selected PicGroup, there is nothing
			// specific to be done, since this selectionHandler only keeps references to direct children of Drawing.RootElement,
			// and so long as these references do not get removed from the Drawing, synchronization b/w the Drawing and
			// this SelectionHandler is not lost.
			if (!(changed instanceof Drawing.RootElement)) return;
			
			Drawing.RootElement rootElem = (Drawing.RootElement)changed;
			if (jpicedt.Log.DEBUG) jpicedt.Log.debugAppend("selection="+this.toString());
			for (Iterator it=elements(); it.hasNext();){ // iterate over selected elements
				Element stillSelected = (Element)it.next();
				if (!rootElem.contains(stillSelected))  // if element was deleted from drawing, delete from selection as well
					it.remove(); // avoid concurrent modification exception
			}
			updateBoundingBox();
			view.changedUpdate(e.getType()); 
		}

		// the following event gets posted in two situations :
		// - either one of setPoint/scale/translate/... was invoked on the SelectionHandler itself, and as a result
		//   all currently selected elements posted a DrawingEvent (we shall then receive as many DrawingEvent's as there are
		//   selected elements) : inherited field BranchElement.changeLock=true in this case, but because this SelectionHandler was
		//   NOT set as the parent of those selected elements (they indeed keep their original parent, ie Drawing.RootElement),
		//   events forwarded by modified children can not be trapped by the forwardChangedUpdate() method,
		//   and they indeed propagate upward until they reach the hosting Drawing. We must filter out these events, since
		//   they are of no interest where the state of this selectionHandler is concerned.
		// - or setPoint/scale/translate/... was invoked directly on a selected element (either from a BSH macro, or a mouse-tool acting directly on an Element),
		//   and this element posted a DrawingEvent.
		//   BranchElement.changeLock=false in this case. This also means that the bounding-box is not up-to-date anymore,
		//   hence we call updateBoundingBox, then fire a change-event to update the view
		// - or finally, the disposition of selected elements was changed through a call to bringXXX() in Drawing.RootElement
		//   => we must reorder the selection.
		else if (e.getType() == DrawingEvent.EventType.GEOMETRY_CHANGE && changeLock==false ){ 
								// && getHighlightingMode()==GLOBAL_MODE){ => commented out
								// The point is that, even in LOCAL_MODE, the bounding-box needs to
								// be up-to-date, since CompositeHighlighter uses it to know if it has to paint the highlighter or not (i.e. clip must intersect allocation)
			if (isEmpty()) return;  // On the other hand, if the selection is empty, this event originated from 
								// the modification of the geometry of an
								// unselected Element, and we just don't care => this has no effect on the geometry of the selectionHandler
			
			Element changed = e.getElement(); // finally, check whether the changed element was selected or not:
			if (changed instanceof Drawing.RootElement){
				sort();
				return;
			}
			// Note that if "changed" is the SelectionHandler itself, isSelected() returns false => this is the way
			// we filter out useless events...
			if (isSelected(changed,false)){ // don't include ancestor [SR:pending] check !
				updateBoundingBox();
				view.changedUpdate(e.getType()); // [SR:pending] the only annoying point is that we eventually trigger two "repaint",
				// one on behalf of the Element that got edited (through a call to its view), and one here.
				// this may slow down a bit.
			}
		}
	}

} // DefaultSelectionHandler


