/*
 DrawToolFactory.java - February 25, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.AbstractCurve;
import jpicedt.graphic.model.EditPointConstraint;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.PicCircleFrom3Points;
import jpicedt.graphic.model.PicEllipse;
import jpicedt.graphic.model.PicMultiCurve;
import jpicedt.graphic.model.PicParallelogram;
import jpicedt.graphic.model.PicSmoothPolygon;
import jpicedt.graphic.model.PicPsCurve;
import jpicedt.graphic.model.PicText;
import jpicedt.graphic.event.PEMouseEvent;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Arc2D;
import java.awt.geom.Rectangle2D;
import java.awt.event.KeyEvent;
import java.util.Iterator;

import jpicedt.Log;//debug

/**
 * A factory that produces mouse-tools for drawing (ie adding) graphical element's to a canvas.
 * @author Sylvain Reynal
 * @since jpicedt 1.3.2
 * @version $Id: DrawToolFactory.java,v 1.29.2.1 2007/09/02 11:56:28 reynal Exp $
 */
public class DrawToolFactory {

	public static final String LINE = "action.draw.Line";
	public static final String RECTANGLE = "action.draw.Rectangle";
	public static final String PARALLELOGRAM = "action.draw.Parallelogram";
	public static final String ELLIPSE_FROM_PARALLELO = "action.draw.EllipseFromParallelo";
	public static final String ARC_CHORD_FROM_PARALLELO = "action.draw.ArcChordFromParallelo";
	public static final String ARC_PIE_FROM_PARALLELO = "action.draw.ArcPieFromParallelo";
	public static final String ARC_OPEN_FROM_PARALLELO = "action.draw.ArcOpenFromParallelo";
	public static final String ELLIPSE = "action.draw.Ellipse";
	public static final String ARC_CHORD = "action.draw.ArcChord";
	public static final String ARC_PIE = "action.draw.ArcPie";
	public static final String ARC_OPEN = "action.draw.ArcOpen";
	public static final String CIRCLE_FROM_3PTS = "action.draw.CircleFrom3Pts";
	public static final String ARC_CHORD_FROM_3PTS = "action.draw.ArcChordFrom3Pts";
	public static final String ARC_PIE_FROM_3PTS = "action.draw.ArcPieFrom3Pts";
	public static final String ARC_OPEN_FROM_3PTS = "action.draw.ArcOpenFrom3Pts";
	public static final String MULTI_CURVE = "action.draw.MultiCurve";
	public static final String CLOSED_MULTI_CURVE = "action.draw.ClosedMultiCurve";
	public static final String SMOOTH_POLY = "action.draw.SmoothPoly";
	public static final String CLOSED_SMOOTH_POLY = "action.draw.ClosedSmoothPoly";
	public static final String PS_CURVE = "action.draw.PsCurve";
	public static final String CLOSED_PS_CURVE = "action.draw.ClosedPsCurve";
	public static final String TEXT = "action.draw.Text";

	// add new tool name here :
	private static final String[][] ALL_TOOL_NAMES = new String[][]{
	            {TEXT},{LINE,RECTANGLE,PARALLELOGRAM,MULTI_CURVE,CLOSED_MULTI_CURVE},
	            {ELLIPSE,ARC_PIE,ARC_CHORD,ARC_OPEN},
	            {ELLIPSE_FROM_PARALLELO,ARC_PIE_FROM_PARALLELO,ARC_CHORD_FROM_PARALLELO,ARC_OPEN_FROM_PARALLELO},
	            {CIRCLE_FROM_3PTS,ARC_PIE_FROM_3PTS,ARC_CHORD_FROM_3PTS,ARC_OPEN_FROM_3PTS},
	            {SMOOTH_POLY,CLOSED_SMOOTH_POLY},{PS_CURVE,CLOSED_PS_CURVE}};

	private CursorFactory cursorFactory = new CursorFactory();
	private EditorKit kit;

	/**
	 * @param editorKit the EditorKit
	 */
	public DrawToolFactory(EditorKit editorKit){
		this.kit = editorKit;
	}

	/**
		 * Returns an array of arrays of String containing the name of each tool this factory can produce, 
	 * sorted by groups of tightly related tools.
	 * This may be useful to register tools with the hosting EditorKit, or to
	 * automate the building process for related PEAction's
	 */
	public static String[][] getAvailableToolNames(){
		return ALL_TOOL_NAMES;
	}

	/**
	 * Returns a MouseTool of the given type. 
	 * @param type one of LINE, RECTANGLE,... constant fields. 
	 * @param set input attribute set for the element to be drawn.
	 * @since jpicedt 1.3.2
	 */
	public MouseTool createDrawTool(String type){

		/* How to design a new XXX_ITERATOR ?
		First, it's to be understood that this scheme only works for Element's that have a finite number of
		control-points (ie PicPolygon and MultiCurve don't fit well here).
		The XXX_ITERATOR is an array of arrays of control-point indices that is fed to the GenericDrawTool
		contructor, so that the latter knows - at each cycle step - which points it's got to update when the user drags the mouse.
		Let's take QBEZIER_ITERATOR for example :

		int[][] QBEZIER_ITERATOR = {{PicBezierQuad.P_1,PicBezierQuad.P_2,PicBezierQuad.P_CTRL},{PicBezierQuad.P_2},{PicBezierQuad.P_CTRL}};

		This leads to the following cycle :
		- first mouse-press : subsequent calls to mouseDragged set PicBezierQuad.P_1, P_2 and P_CTRL (all at once) to the current
		                      mouse position. This is generally a good idea to do so, since it properly initializes the underlying
							  Element.
		- second mouse-press : mouseDragged only moves P_2 (while P_CTRL should still be equl to P_1)
		- third mouse-press : mouseDragged only moves P_CTRL (control the curve bending)
		*/ // [SR:pending] update above documentation snipet now that PicBezierQuad has been deprecated
		final int[][] LINE_ITERATOR = {{0,3},{3}}; // actually uses PicMultiCurve
		final int[][] PARALLELOGRAM_ITERATOR = {{PicParallelogram.P_BL,PicParallelogram.P_TR},{PicParallelogram.SIDE_R},{PicParallelogram.SIDE_T}};
		final int[][] RECTANGLE_ITERATOR = {{PicParallelogram.P_BL,PicParallelogram.P_TR},{PicParallelogram.P_TR}};
		final int[][] ARC_FROM_PARALLELO_ITERATOR = {{PicParallelogram.P_BL,PicParallelogram.P_TR},{PicParallelogram.SIDE_R},{PicParallelogram.SIDE_T},{PicEllipse.P_ANGLE_END},{PicEllipse.P_ANGLE_START}};
		final int[][] ARC_ITERATOR = {{PicParallelogram.P_BL,PicParallelogram.P_TR},{PicParallelogram.P_TR},{PicEllipse.P_ANGLE_END},{PicEllipse.P_ANGLE_START}};
		final int[][] _3PTS_ITERATOR = {{PicCircleFrom3Points.P_1,PicCircleFrom3Points.P_2,PicCircleFrom3Points.P_3},{PicCircleFrom3Points.P_3},{PicCircleFrom3Points.P_2}};

		if (type==RECTANGLE) return new GenericDrawTool(new PicParallelogram(),null, RECTANGLE_ITERATOR);
		if (type==PARALLELOGRAM) return new GenericDrawTool(new PicParallelogram(),null, PARALLELOGRAM_ITERATOR);

		if (type==ELLIPSE) return new GenericDrawTool(new PicEllipse(),null,RECTANGLE_ITERATOR);
		if (type==ARC_PIE) return new GenericDrawTool(new PicEllipse(Arc2D.PIE),null, ARC_ITERATOR);
		if (type==ARC_OPEN) return new GenericDrawTool(new PicEllipse(Arc2D.OPEN),null, ARC_ITERATOR);
		if (type==ARC_CHORD) return new GenericDrawTool(new PicEllipse(Arc2D.CHORD),null, ARC_ITERATOR);

		if (type==ELLIPSE_FROM_PARALLELO) return new GenericDrawTool(new PicEllipse(),null, PARALLELOGRAM_ITERATOR);
		if (type==ARC_PIE_FROM_PARALLELO) return new GenericDrawTool(new PicEllipse(Arc2D.PIE),null, ARC_FROM_PARALLELO_ITERATOR);
		if (type==ARC_CHORD_FROM_PARALLELO) return new GenericDrawTool(new PicEllipse(Arc2D.CHORD),null, ARC_FROM_PARALLELO_ITERATOR);
		if (type==ARC_OPEN_FROM_PARALLELO) return new GenericDrawTool(new PicEllipse(Arc2D.OPEN),null, ARC_FROM_PARALLELO_ITERATOR);

		if (type==CIRCLE_FROM_3PTS) return new GenericDrawTool(new PicCircleFrom3Points(),null, _3PTS_ITERATOR);
		if (type==ARC_OPEN_FROM_3PTS) return new GenericDrawTool(new PicCircleFrom3Points(false,Arc2D.OPEN),null, _3PTS_ITERATOR);
		if (type==ARC_PIE_FROM_3PTS) return new GenericDrawTool(new PicCircleFrom3Points(false,Arc2D.PIE),null, _3PTS_ITERATOR);
		if (type==ARC_CHORD_FROM_3PTS) return new GenericDrawTool(new PicCircleFrom3Points(false,Arc2D.CHORD),null, _3PTS_ITERATOR);

		if (type==LINE) return new GenericDrawTool(new PicMultiCurve(new PicPoint(),new PicPoint()),PicMultiCurve.SMOOTHNESS, LINE_ITERATOR);
		if (type==MULTI_CURVE) return new AbstractCurveDrawTool(new PicMultiCurve(false));
		if (type==CLOSED_MULTI_CURVE) return new AbstractCurveDrawTool(new PicMultiCurve(true));
		if (type==SMOOTH_POLY) return new AbstractCurveDrawTool(new PicSmoothPolygon(false));
		if (type==CLOSED_SMOOTH_POLY) return new AbstractCurveDrawTool(new PicSmoothPolygon(true));
		if (type==PS_CURVE) return new AbstractCurveDrawTool(new PicPsCurve(false));
		if (type==CLOSED_PS_CURVE) return new AbstractCurveDrawTool(new PicPsCurve(true));

		if (type==TEXT) return new TextDrawTool(new PicText());
		return null;
	}

	/////////////////////////////////////////////////////////////////////////////
	//// GenericDrawTool
	/////////////////////////////////////////////////////////////////////////////

	/**
	 * A MouseTool aimed at drawing new Elements. It works like an iterator,  
	 * by iterating (cyclically, creating a new Element at each cycle) 
	 * through a given array of point indexes, or (if the latter are not provided in the constructor) 
	 * by computing it from the available control points for the given element.<p>
	 * GenericDrawTool targets Element having a finite number of control points ONLY !
	 * @author Sylvain Reynal
	 */
	protected class GenericDrawTool extends MouseTool {

		private Element prototype, currentElement;
		private EditPointConstraint constraint, defaultConstraint;
		private int currentTaskIndex; // indexes the current draw-task, gets increased at each mouse-press event
		private int[][] drawPoints; // array of arrays of indices of points to be moved at each draw-task ;
		//  e.g. drawPoints[2] contains points to be moved during 3rd draw-task, ie after two mouse-clicks.
		/* buffer */
		private PicPoint ptBuffer = new PicPoint();
		private PicPoint ptBuffer1 = new PicPoint();
		// set to true as soon as all points have been set
		private boolean isCompleted;
		// help message to be posted by the hosting EditorKit (this may vary according to key modifiers)
		private String helpMsg;
		private String helpMsgDEFAULT="help-message.Draw";

		/**
		 * @param prototype the Element that will serve as the prototype (getting cloned, then
		 *        added to the drawing).
		 * @param constraint which constraint to use when moving control-points
		 * @param drawPoints an array of array of point indexes which the iterator will 
		 *        iterate through (in ascending order) to draw the element (at each iteration, the corresponding list of points
		 *        will be updated together). 
		 * @see jpicedt.graphic.model.Element#setPoint
		 */
		public GenericDrawTool(Element prototype, EditPointConstraint constraint, int[][] drawPoints){
			this.prototype = prototype;
			this.constraint = this.defaultConstraint = constraint; // because constraint can be modified during the invokation of mouseDragged()
			this.drawPoints = drawPoints;
			isCompleted = true; // mark has completed -> force (re)start on next mousePressed event
			helpMsg = helpMsgDEFAULT;
		}

		/**
		 * The <code>drawPoints</code> array gets computed
		 * from all available control-points for the given element.
		 * @param prototype the Element that will serve as the prototype (getting cloned, then
		 *        added to the drawing). 
		 * @param inputAttributes the input attribute set for the Element to be created.
		 */
		public GenericDrawTool(Element prototype,  EditPointConstraint constraint){
			this.prototype = prototype;
			this.constraint=constraint;
			// there are as many "drawPoints" indices as control-points for this element :
			drawPoints = new int[prototype.getLastPointIndex()-prototype.getFirstPointIndex()+1][];
			drawPoints[0] = new int[prototype.getLastPointIndex()-prototype.getFirstPointIndex()+1];
			int i,j;
			// 1) create a set of control-point indices to be dragged immediately after the first mousePressed occured :
			// (our policy is just to move ALL control-points at once)
			for (i=prototype.getFirstPointIndex(), j=0; i<prototype.getLastPointIndex();i++,j++){
				drawPoints[0][j] = i; // include ALL indices here
			}
			// 2) then, at each subsequent cycle step, we'll simply move a different control-point, starting from
			//    the second available one (the first control-point was set in the previous step)
			//    until all control-points have been properly dragged.
			//    Hence : a) the array is reduced to a single element b) this element
			//    is an int whose value is the very control-point index to be dragged.
			for (i=prototype.getFirstPointIndex()+1, j=0; i<prototype.getLastPointIndex();i++,j++){
				drawPoints[j]=new int[1]; // there's only one index
				drawPoints[j][0] = i; // and it's the very index of the control-point to be dragged
			}
			//??? je comprends pas a quoi ca sert ce drawpoints calcule, c'est le truc par defaut, c'est ca ? [SR] oui
			isCompleted = true; // mark has completed -> force (re)start on next mousePressed event
			helpMsg = helpMsgDEFAULT;
		}

		/**
		 * A popup-trigger mouse event (e.g. a right click on Unix/Windows) :<br>
		 * - switches back to SELECT_MODE if the task-iterator is completed<br>
		 * - or cancels the current task if the task-iterator is underway.<br>
		 * A left- or middle-button SINGLE click either :<br>
		 * - (re)start the task iterator if it was completed<br>
		 * - or select the next point if there are more points<br>
		 * - or terminate the drawing process if there are no more points.<br>
		 * @author Sylvain Reynal
		 * @since jPicEdt
		 */
		public void mousePressed(PEMouseEvent e) {
			super.mousePressed(e);
			// right button pressed
			if (e.isPopupTrigger()) {
				// either : back to SELECT_MODE
				if (isCompleted) kit.setCurrentMouseTool(EditorKit.SELECT); // calls back flush()
				// or : cancel last draw-operation
				else { // ARCS only !
					if (currentElement instanceof PicEllipse){
						PicEllipse ellipse = (PicEllipse)currentElement;
						if (!ellipse.isPlain()){ // it's an arc -> cycle through arc types
							int closure = ellipse.getArcType();
							switch (closure){
							case PicEllipse.OPEN:
								ellipse.setArcType(PicEllipse.PIE);
								break;
							case PicEllipse.PIE:
								ellipse.setArcType(PicEllipse.CHORD);
								break;
							case PicEllipse.CHORD:
								ellipse.setArcType(PicEllipse.OPEN);
								break;
							default:
							}
						}
					}
				}
			}
			// left button pressed :
			else if (e.isLeftButton()) {
				e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer);
				// restart tool :
				if (isCompleted) {
					isCompleted = false;
					e.getCanvas().selectAll(false); // unselect all
					// fill ptBuffer with nearest-neighbour on grid :
					e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer);
					// instanciates a new Element from the prototype :
					currentElement = (Element)prototype.clone();
					currentElement.setAttributeSet(kit.getInputAttributes());
					// init ALL object's points at once by giving them the value of the current clicked point :
					currentTaskIndex = 0;
					for (int i=0; i<(drawPoints[currentTaskIndex]).length;i++)
						currentElement.setPoint(drawPoints[currentTaskIndex][i],ptBuffer,this.constraint);

					e.getCanvas().beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("action.editorkit.Draw")
					                                  +" ("+prototype.getName()+")");
					e.getCanvas().getDrawing().addElement(currentElement);
				}
				// end ?
				if (currentTaskIndex >= drawPoints.length-1){
					setCurrentPoint(e);
					isCompleted = true;
					this.constraint = this.defaultConstraint; // in case constraint has been modified by mouseDragged()
					e.getCanvas().endUndoableUpdate();
					e.getCanvas().select(currentElement,false); // replace selection
					// [pending] fire undo event
				}
				// next task
				else {
					currentTaskIndex++;
					setCurrentPoint(e);
				}
			}
		}

		/** set current point */
		public void mouseDragged(PEMouseEvent e) {
			super.mouseDragged(e);
			if (e.isPopupTrigger())  return; // right button -> do nothing :
			setCurrentPoint(e);
		}

		/** set cursor for canvas, then call mouseDragged */
		public void mouseMoved(PEMouseEvent e){
			super.mouseMoved(e);
			e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
			updateHelpMessage(e);
			kit.postHelpMessage(helpMsg);
			mouseDragged(e);
		}

		/**
		 * Updates the value of the current Help-Message to be posted by the hosting EditorKit, according
		 * to current key modifiers. This aims at mimicking the behavior in EditElementMouseTransformFactory,
		 * where modifiers-dependent messages are posted in real-time.
		 */
		private void updateHelpMessage(PEMouseEvent e){
			helpMsg = helpMsgDEFAULT;
			if (currentElement instanceof PicEllipse){
				PicEllipse ellipse = (PicEllipse)currentElement;
				if (!ellipse.isPlain()){ // it's an arc -> cycle through arc types (this is valid only after the surrounding parallelogram has been set)
					helpMsg = "help-message.DrawArc";
					return;
				}
			}
			// --- parallelograms and ellipses ---
			if (currentElement instanceof PicParallelogram && !(currentElement instanceof PicCircleFrom3Points)){
				if (e.isControlDown() && e.isAltDown()) {
					if (e.isShiftDown()) {
						if (currentElement instanceof PicEllipse)  helpMsg = "help-message.EllipseCircle";
						else helpMsg = "help-message.ParalleloSquare";
					}
					else helpMsg = "help-message.MovePointCenterFixed";
				}
			}
		}
		
		/**
		 * move the points indexed in drawPoints[currentTaskIndex] to the current click-point
		 * (possibly after grid alignment).
		 */
		public void setCurrentPoint(PEMouseEvent e){
			if (isCompleted) return;
			e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer);
			if (ptBuffer.equals(currentElement.getPoint(drawPoints[currentTaskIndex][0],ptBuffer1))) return;
			if (currentElement instanceof PicParallelogram && !(currentElement instanceof PicCircleFrom3Points)){
				if (e.isControlDown() && e.isAltDown()) {
					if (e.isShiftDown()) this.constraint = PicParallelogram.SQUARE; // same as in EditElementMouseTransformFactory
					else this.constraint = PicParallelogram.CENTER_FIXED;
				}
				else this.constraint = PicParallelogram.DEFAULT;
			}
			for (int i=0; i<(drawPoints[currentTaskIndex]).length;i++)
				currentElement.setPoint(drawPoints[currentTaskIndex][i],ptBuffer,this.constraint);
		}

		/**
		 * called when this tool is being activated in the hosting editor kit
		 */
		public void init(){
			kit.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
		}

		/** called when this tool is being replaced by another mouse-tool in the hosting
		 * editor kit ; this is mainly for mousetools using more than one sequence, for it
		 * gives them a chance to clean themselves up for the next time */
		public void flush(){
			isCompleted = true;
			currentElement=null;
		}
		/**
		 * This method is called by the hosting EditorKit : this implementation paints
		 * the current element's highlighter.
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			if (currentElement==null || currentElement.getView()==null) return;
			g.setPaint(Color.blue);
			currentElement.getView().paintHighlighter(g, allocation, scale);
		}
	} // GenericDrawTool


	/**
	 * a MouseTool suited for drawing any element inheriting from AbstractCurve.
	 * Because of slight differences that appear in the drawing sequence, 
	 * this class currently specializes on PicMultiCurve, PicPsCurve and PicSmoothPolygon. This 
	 * should be improved in the near future by providing a common interface [SR:pending].<p>
	 * The drawing sequence is as follows : 
	 * OPEN curve : mpress add 0,1,2,3->drag 1->move 3->mpress add 4,5,6->drag 4->move 6->...
	 * CLOSED curve : mpress add 0,1,2->drag 1->mrel add 3,4,5->move 3->mpress->drag 4->mrel add 6,7,8->move
	 * and is terminated by a right-button click.
	 * @author Vincent Guirardel, Sylvain Reynal
	 */
	protected class AbstractCurveDrawTool extends MouseTool {

		private AbstractCurve currentElement, prototype;
		/* index of current "moved" or "dragged" control-points ; index is being updated during each call to mousePressed()/mouseReleased()
		 * according to the sequence indicated in documentation above */
		private int currentDrawPointIndex;
		/* buffers */
		private PicPoint ptBuffer1 = new PicPoint(), ptBuffer2 = new PicPoint();
		private boolean isCompleted;
		private boolean isLastControlPoint; // true as soon as a right-click occurs -> drag control-point associated with last end-point
		// [SR:pending] the following may be made CTRL+ALT+SHIFT dependent in the near future...
		private EditPointConstraint editPtConstraint;
		// help message to be posted by the hosting EditorKit (this may vary according to key modifiers)
		private String helpMsg;
		private String helpMsgDEFAULT="help-message.Draw";

		/**
		 * @param prototype Prototype which will be cloned, then added to the drawing.
		 */
		public AbstractCurveDrawTool(AbstractCurve prototype){
			if (Log.DEBUG) Log.debug("Abstract Curve constructor. Prototype : "+prototype.toString());
			isCompleted = true; // mark has completed -> force start to be called
			isLastControlPoint = false;
			this.prototype = prototype;
			if (prototype instanceof PicMultiCurve)  editPtConstraint= PicMultiCurve.SYMMETRY.and(PicMultiCurve.SMOOTHNESS);
			else editPtConstraint=null;
			helpMsg = helpMsgDEFAULT;
		}

		/**
		 * Update the current EditPointConstraint from the key modifiers associated with the given mouse event.
		 * As a by-product, the current help message (to be posted by the hosting EditorKit) get also updated.
		 */
		private void updateEditPointConstraint(PEMouseEvent me){
			this.helpMsg  = helpMsgDEFAULT;
			if (!(prototype instanceof PicMultiCurve)) this.editPtConstraint=null;
			if (me.isShiftDown() && me.isControlDown() && !me.isAltDown()){ // SHIFT+CTRL
				this.editPtConstraint = PicMultiCurve.SMOOTHNESS;
				this.helpMsg ="help-message.ControlSmooth";
			}
			else if (!me.isShiftDown() && me.isControlDown() && me.isAltDown()){ //CTRL+ALT
				this.editPtConstraint = PicMultiCurve.SYMMETRY;
				this.helpMsg="help-message.ControlSymmetric";
			}
			else if (me.isShiftDown() && me.isControlDown() && me.isAltDown()){ // CTRL+ALT+SHIFT
				this.editPtConstraint = PicMultiCurve.FREELY;
				this.helpMsg="help-message.ControlFreely";
			}
			else {
				this.editPtConstraint = PicMultiCurve.SMOOTHNESS.and(PicMultiCurve.SYMMETRY); // default (=CTRL in EditElementMouseTransformFactory)
				this.helpMsg = "help-message.ControlSmoothAndSymmetric";
			}
		}

		private void markCompleted(PEMouseEvent e){
			isCompleted = true;
			isLastControlPoint = false;
			e.getCanvas().endUndoableUpdate();
			e.getCanvas().select(currentElement,false); // select element, replacing old selection
			currentElement=null;
			if (prototype instanceof PicMultiCurve)  editPtConstraint= PicMultiCurve.SYMMETRY.and(PicMultiCurve.SMOOTHNESS);
			else editPtConstraint=null;
		}

		/**
			 * A popup-trigger mouse event (e.g. a right click on Unix/Windows) :<br>
		 * - switches back to SELECT_MODE if the iterator is completed<br>
		 * - or terminates the drawing process.
		 * A left-button SINGLE click either :<br>
		 * - (re)start the task iterator if completed ;<br>
		 * - or add a new point ;<br>
		 */
		public void mousePressed(PEMouseEvent e) {
			super.mousePressed(e);
			e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));

			// 		    if (Log.DEBUG) Log.debug("Current element2 : "+currentElement.toString());

			// right button pressed:
			if (e.isPopupTrigger()) {
				// either : back to SELECT_MODE
				if (isCompleted) kit.setCurrentMouseTool(EditorKit.SELECT);
				else { // or next move will drag last control-point
					if (currentElement instanceof PicSmoothPolygon){
						markCompleted(e);
						return;
					}
					if (currentElement instanceof PicPsCurve){
						if (currentElement.isClosed()){
							markCompleted(e); // no additional control point in a closed pscurve, so we are done
							return;
						}
						// yes, we're open !
						if (isLastControlPoint==false){
							isLastControlPoint=true; // next move drag control-point associated with last end-point
							currentDrawPointIndex= currentElement.getNumberOfSegments()+2; // index of last control point
							return;
						}
						else {
							markCompleted(e);
							return;
						}
					}
					// otherwise it's a multicurve
					int nbSeg = currentElement.getNumberOfSegments();
					boolean isStraightLastSegment = currentElement.isStraight(nbSeg-1);
					if (isStraightLastSegment==false && isLastControlPoint==false){
						isLastControlPoint=true; // next move drag control-point associated with last end-point
						currentDrawPointIndex=currentElement.getNumberOfSpecificationPoints()-2; // now index of last control-point
					}
					else { // mark process as completed
						markCompleted(e);
						return;
					}
				}
			}

			// left button pressed:
			else {
				e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer2);
				if (isLastControlPoint){ // previous move dragged last control-point => now complete drawing
					markCompleted(e);
					return;
				}
				else if (isCompleted) { // restart tool ; first mpress adds first segment
					e.getCanvas().selectAll(false); // unselect all
					isCompleted = false;
					isLastControlPoint = false;
					currentElement = (AbstractCurve)prototype.clone(); // instanciates a new Element from the prototype
					currentElement.setAttributeSet(kit.getInputAttributes());
					currentElement.addPoint(ptBuffer2); // add first point for OPEN curve, first three points for CLOSED curve
					currentDrawPointIndex = 0;
					e.getCanvas().beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("action.editorkit.Draw")
					                                  +" ("+prototype.getName()+")");
					e.getCanvas().getDrawing().addElement(currentElement);
				}
				currentDrawPointIndex ++; // move to next subdivision-point (multicurve) or next point (smoothpoly, closed pscurve)
				if (currentElement instanceof PicSmoothPolygon
				        || (currentElement instanceof PicMultiCurve && !currentElement.isClosed())
				        || (currentElement instanceof PicPsCurve && currentElement.isClosed()))
					currentElement.addPoint(ptBuffer2);
				// closed PicMultiCurve : this is being done in mouseReleased (see sequence in this class documentation)
				if (currentElement instanceof PicPsCurve && !currentElement.isClosed()) {
					if (currentDrawPointIndex==1){
						// we need 3 points for open pscurve: 2 control points + one non-control point
						currentElement.addPoint(ptBuffer2);// 1st and only non-control point
						currentElement.addPoint(ptBuffer2);// last control point
						currentDrawPointIndex=0 ; //dragging the mouse should change initial control point (index 0)
					}
					else {
						currentElement.addPoint(ptBuffer2);// adds a point. This is the final control point.
						currentElement.setPoint(currentDrawPointIndex,ptBuffer2);// sets current point (=former final control point)
					}
				}
			}
		}

		/** CLOSED curve only : add a new point by splitting last segment */
		public void mouseReleased(PEMouseEvent e){
			super.mouseReleased(e);
			if (e.isLeftButton() && !isCompleted && currentElement instanceof PicMultiCurve){
				if (currentElement.isClosed()){
					e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer2);
					currentElement.addPoint(ptBuffer2);
					currentDrawPointIndex +=2;
				}
				else {
					currentDrawPointIndex +=2; // addPoint() is being done is mousePressed()
				}
			}
			if (e.isLeftButton() && !isCompleted && currentElement instanceof PicPsCurve
			        && !currentElement.isClosed() && currentDrawPointIndex==0){
				currentDrawPointIndex =2; // next time, we'll go to next (non-control) point (index 2).
				currentElement.addPoint(ptBuffer2);//point with index 3
				currentElement.setPoint(2,ptBuffer2);

			}

		}

		/** set the current point */
		public void mouseDragged(PEMouseEvent e) {
			if (Log.DEBUG) Log.debug("Dragged. currentDrawPointIndex : "+currentDrawPointIndex);
			super.mouseDragged(e);
			if (e.isPopupTrigger() || isCompleted)  return; // right button -> do nothing :
			updateEditPointConstraint(e);
			e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer2);
			if (ptBuffer2.equals(currentElement.getPoint(currentDrawPointIndex,ptBuffer1))) return;
			currentElement.setPoint(currentDrawPointIndex,ptBuffer2,editPtConstraint); // set control-point
			if (currentElement instanceof PicPsCurve && !currentElement.isClosed() && currentDrawPointIndex !=0 && !isLastControlPoint)
				currentElement.setPoint(currentDrawPointIndex+1,ptBuffer2,editPtConstraint); // we also need to set the final control-point
		}

		/** set cursor for canvas, then call mouseDragged */
		public void mouseMoved(PEMouseEvent e){
			super.mouseMoved(e);
			e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
			mouseDragged(e); // also update helpMsg
			kit.postHelpMessage(helpMsg);
		}

		/**
		 * called when this tool is being activated in the hosting editor kit
		 */
		public void init(){
			kit.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
		}

		/** called when this tool is being replaced by another mouse-tool in the hosting
		 * editor kit ; this is mainly for mousetools using more than one sequence, for it
		 * gives them a chance to clean themselves up for the next time */
		public void flush(){
			isCompleted = true;
			isLastControlPoint = false;
			currentElement=null;
		}
		/**
		 * This method is called by the hosting EditorKit : this implementation paints
		 * the current element's highlighter.
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			if (currentElement==null || currentElement.getView()==null) return;
			g.setPaint(Color.blue);
			currentElement.getView().paintHighlighter(g, allocation, scale);
		}
	}


	/////////////////////////////////////////////////////////////////////////////
	//// TextDrawTool
	/////////////////////////////////////////////////////////////////////////////

	/**
	 * A tool used to place PicText's ; it opens a dialog box for editing text content. 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	protected class TextDrawTool extends MouseTool {

		private PicText prototype, currentElement;
		/* buffer */
		private PicPoint ptBuffer = new PicPoint();
		private PicPoint ptBuffer1 = new PicPoint();

		/**
		 * @param prototype the Element that will serve as the prototype (getting cloned, then added to the drawing).
		 */
		public TextDrawTool(PicText prototype){
			this.prototype = prototype;
		}

		/**
		 * A popup-trigger mouse event (e.g. a right click on Unix/Windows) :<br>
		 * - switches back to SELECT_MODE if the task-iterator is completed<br>
		 * - or cancels the current task if the task-iterator is underway.<br>
		 * A left- or middle-button SINGLE click either :<br>
		 * - (re)start the task iterator if it was completed<br>
		 * - or select the next point if there are more points<br>
		 * - or terminate the drawing process if there are no more points.<br>
		 * @author Sylvain Reynal
		 * @since jPicEdt
		 */
		public void mousePressed(PEMouseEvent e) {
			super.mousePressed(e);

			// right button pressed
			if (e.isPopupTrigger()) {
				kit.setCurrentMouseTool(EditorKit.SELECT); // calls back flush()
			}
			// left button pressed :
			else {
				e.getCanvas().getGrid().nearestNeighbour(e.getPicPoint(),ptBuffer);
				// instanciates a new Element from the prototype
				currentElement = (PicText)prototype.clone();
				currentElement.setAttributeSet(kit.getInputAttributes());
				currentElement.setPoint(PicText.P_ANCHOR,ptBuffer,null); // no constraint
				e.getCanvas().beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("action.editorkit.Draw")
				                                  +" ("+prototype.getName()+")");
				e.getCanvas().getDrawing().addElement(currentElement);
				// open dialog box for text editing :
				AbstractCustomizer pane = ((CustomizerFactory)currentElement).createCustomizer();
				CustomizerDialog dlg = new CustomizerDialog(pane,CustomizerDialog.OK_CANCEL_BUTTONS);
				dlg.setOkButtonClosesDialog(true);
				dlg.setVisible(true);
				e.getCanvas().endUndoableUpdate();
				e.getCanvas().select(currentElement,false); // replace selection
				// [pending] fire undo event
			}
		}

		/** do nothing */
		public void mouseDragged(PEMouseEvent e) {
			super.mouseDragged(e);
			if (e.isPopupTrigger())  return; // right button -> do nothing :
		}

		/** set cursor for canvas, then call mouseDragged */
		public void mouseMoved(PEMouseEvent e){
			super.mouseMoved(e);
			e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
			kit.postHelpMessage("help-message.Draw");
			mouseDragged(e);
		}

		/**
		 * called when this tool is being activated in the hosting editor kit
		 */
		public void init(){
			kit.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.DRAW));
		}

		/** called when this tool is being replaced by another mouse-tool in the hosting
		 * editor kit ; this is mainly for mousetools using more than one sequence, for it
		 * gives them a chance to clean themselves up for the next time */
		public void flush(){
		}
	} // TextDrawTool



} // class DrawToolFactory

