/*
 MouseTool.java - January 4, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.event.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;


/**
 * An implementation of PEMouseInputListener that factors common behaviour for EditorKit's mouse tools,
 * and implements KeyListener so that mouse-tools can also listen to key events (e.g. to update cursor,...) 
 * <br>
 * This implementation offers a convenient mechanism for generating MouseMoved events as a result of a KeyPressed/KeyReleased
 * event. If this behaviour is to be preserved, subclassers should either call <code>super.mouseXXX()</code> before all
 * in their own implementation of the corresponding methods, or alternately simply add the following line of code :
 * <code>lastMouseEvent=e</code>, where <code>e</code> denotes the mouse-event parameter.
 * <p>
 * Besides, this class provide a "paint" method which, like Element's, allows
*  the tool to draw shapes (on the associated PECanvas) that are specific to this tool.  
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: MouseTool.java,v 1.10.2.1 2007/09/02 11:56:28 reynal Exp $
 *
 */
public abstract class MouseTool extends KeyAdapter implements PEMouseInputListener {

	/**
	 * this flag is set to true/false whenever a mouseExited/Entered event is caught.
	 * <br>For instance, it may be used by subclasses to trigger an autoscrolling operation.
	 */
	protected boolean mouseOutside;

	/** this field remembers the last mouse-event produced by a call to one of the mouseXXX methods ; shared by every MouseTool */
	protected static PEMouseEvent lastMouseEvent; 
	
	// used by KeyListener to remember up- and down- keys modifiers
	private static int keyModifiers=0; 
	
	/** this default implementation simply stores the given mouse-event in lastMouseEvent field */
	public void mousePressed(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
	}
	
	/** this default implementation simply stores the given mouse-event in lastMouseEvent field */
	public void mouseReleased(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
	}

	/** this default implementation simply stores the given mouse-event in lastMouseEvent field */
	public void mouseClicked(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
	}

	/** this default implementation simply stores the given mouse-event in lastMouseEvent field */
	public void mouseMoved(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
	}

	/** this default implementation simply stores the given mouse-event in lastMouseEvent field */
	public void mouseDragged(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
	}
	
	/**
	 * Returns the last mouse-event as produced by the last invokation of a mouseXXX method
	 */
	public PEMouseEvent getLastMouseEvent(){
		return lastMouseEvent;
	}

	/**
	 * Sets the mouseOutside flag to true, stores the given mouse-event in lastMouseEvent field,
	 * and  request focus on the PECanvas that dispatched this mouse event.
	 */
	public void mouseEntered(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		mouseOutside = true;
		lastMouseEvent = e; // store for later use by processKeyEvent
		e.getCanvas().requestFocus();
	}

	/**
	 * Sets the mouseOutside flag to false, and stores the given mouse-event in lastMouseEvent field.
	 */
	public void mouseExited(PEMouseEvent e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("e="+e);
		lastMouseEvent = e; // store for later use by processKeyEvent
		mouseOutside = false;
	}

	/**
	 * Allow the tool to paint shapes that are specific to this tool. This method
	 * is called by the hosting EditorKit. 
	 * <br>
	 * For instance, this method may paint selection areas, markers, ...
	 * Current implementation does nothing.
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
	}
	
	/** called when this tool is being activated in the hosting
	 * editor kit ; this is mainly for mousetools that need to do some initialization work
	 * before any mouse-event occurs.
	 */
	public void init(){
	}

	/** called when this tool is being replaced by another mouse-tool in the hosting
	 * editor kit ; this is mainly for mousetools using more than one sequence, for it
	 * gives them a chance to clean themselves up for the next time */
	public void flush(){
	}

	/**
	 * Triggers a fake "mouseMoved" event when a key is pressed.
	 * Current implementation merely calls {@link processKeyEvent}.
	 */
	public void keyPressed(KeyEvent ke){
		processKeyEvent(ke,true); // pressed
	}

	/**
	 * Triggers a fake "mouseMoved" event when a key is released.
	 * Current implementation merely calls {@link processKeyEvent}.
	 */	
	public void keyReleased(KeyEvent ke){
		processKeyEvent(ke, false); // released
	}
			
	/**
	 * Triggers a fake "mouseMoved" event if a key is pressed or released, with the adequate key modifiers ; 
	 * this is a "lazy" handling of key-events
	 * which delegates to the mouse-handler. The goal is e.g. to allow the receiver to update 
	 * things as cursor's shape/help-message/... when a key is pressed.<br>
	 * @param pressed if TRUE, process a keyPressed event, else process a keyReleased
	 */
	protected void processKeyEvent(KeyEvent ke, boolean pressed){
		// if this is called before the mouse ever moved, default behaviour :
		if (lastMouseEvent==null) {
			keyModifiers=0;
			//((PECanvas)ke.getSource()).setCursor(cursorFactory.getPECursor(CursorFactory.SELECT));
			return;
		}
		else {
			// update this.keyModifiers from keycode
			int keycode = ke.getKeyCode();
			int mask = 0;
			switch (keycode){
				case KeyEvent.VK_SHIFT : 
					mask = InputEvent.SHIFT_DOWN_MASK; // [SR:pending] as of JDK1.4, use extended modifiers instead ? (look up Java tutorial) 
					break;
				case KeyEvent.VK_CONTROL : 
					mask = InputEvent.CTRL_DOWN_MASK; 
					break;
				case KeyEvent.VK_ALT : 
					mask = InputEvent.ALT_DOWN_MASK; 
					break;
				default : 
					return;
			}
			if (pressed) this.keyModifiers = this.keyModifiers | mask;
			else this.keyModifiers = this.keyModifiers & (~mask);
			if (jpicedt.Log.DEBUG) 
				jpicedt.Log.debug("keyModifiers="+keyModifiers);
			
			// create a new mouse-event including keyModifiers
			MouseEvent oldMe = lastMouseEvent.getAwtMouseEvent();
			MouseEvent newMe = new MouseEvent(oldMe.getComponent(),oldMe.getID(),oldMe.getWhen(),
				keyModifiers,oldMe.getX(),oldMe.getY(),oldMe.getClickCount(),oldMe.isPopupTrigger());
			PEMouseEvent withModifiersEvent = new PEMouseEvent(newMe, lastMouseEvent.getCanvas(), lastMouseEvent.getPicPoint());
			mouseMoved(withModifiersEvent); // lazily delegates to mouseMoved !
		}
	}
	
}  

