/*
 PEAction.java - January 2, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.*;
import jpicedt.graphic.model.*;

import javax.swing.*;
import java.awt.event.*;
import java.util.*;

/**
 * An abstract subclass of Swing's AbstractAction that can (but may as well not) 
 * be shared across multiple PECanvas. 
 * <p>
 * This implementation uses the same set of properties as AbstractAction.
 * <p>
 * The constructor gives a convenient way of automatically building localized properties 
 * associated with this PEAction from a simple String (the "actionName" parameter), using a 
 * user-provided ActionLocalizer.
 * <br>
 * Note : if an ActionLocalizer is provided to the constructor, and some of its methods return null values,
 * the corresponding properties are set to null. However, all jPicEdt components (e.g. PEMenu, PEToolBar,...), 
 * like Swing components, are guaranteed to handle this case properly w/o throwing an exception.
 * <br>
 * For instance, simply setting the NAME property to null and filling the SMALL_ICON property 
 * create a JButton without text under the component's Icon (this is the standard Swing behaviour).
 * <p>
 * <b>Example of use : </b> [todo]
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: PEAction.java,v 1.8.2.1 2007/09/02 11:56:28 reynal Exp $
 *
 */
public abstract class PEAction extends AbstractAction {

	/* the ActionDispatcher that provides a target (PECanvas) to this action */
	private ActionDispatcher actionDispatcher;
	

	/**
	 * Constructor to be used either :
	 * <ul>
	 * <li> when the target board is not known in advance (e.g. MDI), and one wants to
	 * provide an own ActionDispatcher implementation
	 * <li> or when the target board is known : in this case use a DefaultActionDispatcher 
	 * constructed with the given target board.
	 * </ul> 
	 * @param actionDispatcher the ActionDispatcher that provides a  
	 *        target when this Action gets invoked. If set to null, then subclasses may want to simply override actionPerformed().
	 * @param actionName the actionName of this Action that serve to build this Action's properties using
	 *        the given ActionLocalizer (if non-null) ; in any case, actionName is used to fill the
	 *        ACTION_COMMAND_KEY property.
	 * @param actionLocalizer the ActionLocalizer that feeds this Action with localized properties ;
	 *        if null, the Action's NAME property is simply initialized to actionName, but other 
	 *        properties are not set (this allow a subclass to set these properties manually).
	 */
	public PEAction(ActionDispatcher actionDispatcher, String actionName, ActionLocalizer localizer){
		super(actionName); // default, if no localizer is provided
		this.actionDispatcher = actionDispatcher;
		putValue(ACTION_COMMAND_KEY,actionName);
		if (localizer != null){
			putValue(NAME, localizer.getActionName(actionName));
			putValue(SHORT_DESCRIPTION, localizer.getActionTooltip(actionName));
			putValue(LONG_DESCRIPTION, localizer.getActionHelper(actionName));
			putValue(MNEMONIC_KEY, localizer.getActionMnemonic(actionName));
			putValue(ACCELERATOR_KEY, localizer.getActionAccelerator(actionName));
			putValue(SMALL_ICON, localizer.getActionIcon(actionName));
		}
	}
	
	/**
	 * This is a default implementation of the "actionPerformed" method suited for undoable actions :
	 * we first call "beginUndoableUpdate" with "actionName" as the presentation name,
	 * then delegate to "undoableActionPerformed", finally mark the undoable-update as ended.
	 * <p>
	 * Action that don't want to be undoable should simply override this method as usual.
	 */
	public void actionPerformed(ActionEvent e){
			PECanvas canvas = getCanvas();
			if (canvas==null) return;
			canvas.beginUndoableUpdate((String)getValue(NAME)); // i18n'd if there was a localizer provided in the contructor
			undoableActionPerformed(e);
			canvas.endUndoableUpdate();
	}
	
	/**
	 * called by "actionPerformed" ; default implementation does nothing. Action that want
	 * to be undoable override this method instead of "actionPerformed". 
	 * <p>
	 * Note : this is not an abstract method, since this would force all PEAction's, including
	 * those we don't want to be undoable, to implement this method.
	 */
	protected void undoableActionPerformed(ActionEvent e){
	}
	
	/**
	 * Returns the PECanvas this Action acts upon, as specified by the ActionDispatcher
	 *         provided in the constructor.
	 * @return a null reference if no suitable canvas is found by the ActionDispatcher.
	 */
	protected PECanvas getCanvas(){
		if (actionDispatcher==null) return null;
		return actionDispatcher.getTarget();
	}
	
	/**
	 * Returns the Drawing that serves as a model for the target PECanvas (this is a convenience
	 *         call to <code>getCanvas()</code>.
	 */
	protected Drawing getDrawing(){
		if (getCanvas()==null) return null;
		return getCanvas().getDrawing();
	}
	
	/**
	 * Returns the EditorKit associated with the target PECanvas (this is a convenience
	 *         call to <code>getCanvas()</code>.
	 */
	protected EditorKit getEditorKit(){
		if (getCanvas()==null) return null;
		return getCanvas().getEditorKit();
	}

	/**
	 * Returns the selected graphic element in the target board ;
	 * @return null if there's no selected graphic element, OR more than one.
	 */
	protected Element getSelectedObject(){
		if (getCanvas()==null) return null;
		int s = getCanvas().getSelectionSize();
		if (s == 1){ // if only ONE selected element
			Iterator it = getCanvas().selection();
			return (Element)it.next();
		}
		else return null;
	}
	
	/**
	 * Applies the given attribute to the drawing content.  If there is a selection, the attribute
	 * is applied to each element in the selection.  If there is no selection, the attribute 
	 * is applied to the input attribute set of the EditorKit (this specifies the attributes
	 * to be used for any new Element that gets added to the underlying Drawing).
	 * <p>
	 * If there's a selected element, we add an UndoableEdit (hence PEAction that explicitly call
	 * setAttribute() must not override <code>undoableActionPerformed</code>, but simply <code>actionPerformed</code>).
	 */
     protected final void setAttribute(PicAttributeName name, Object value) {
		 if (getCanvas()==null) return;
		 if (getCanvas().getSelectionSize()==0){
			 EditorKit kit = getCanvas().getEditorKit();
			 PicAttributeSet inputAttributes = kit.getInputAttributes();
			 inputAttributes.setAttribute(name,value);
		 }
		 else {
			getCanvas().beginUndoableUpdate((String)getValue(ACTION_COMMAND_KEY));
			 for(Iterator it = getCanvas().selection(); it.hasNext(); ){
				 Element e = (Element)it.next();
				 e.setAttribute(name, value);
			 }
			getCanvas().endUndoableUpdate();
		 }
		 getCanvas().repaint();
	}


} 

