/*
 TransferableGraphic.java - March 3, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
package jpicedt.graphic.toolkit;

import jpicedt.graphic.model.*;

import java.awt.datatransfer.*;

/**
 * Wrapper for Element's that can be transfered to/from a Clipboard. Contrary to a BranchElement 
 * implementation :
 * - this object stores a reference to a CLONE of the Element(s) of interest, not to the Element(s)
 *   itself (this allow the source Element to be further modified w/o affecting the content of this
 *   TransferableGraphic, which is necessary for proper ClipBoard behaviour)
 * - As a result, Element's added to this TransferableGraphic have no parent, no view,...
 */
public class TransferableGraphic implements Transferable, ClipboardOwner {
	
	/** the DataFlavor for Element's */
	public static final DataFlavor JPICEDT_DATA_FLAVOR = new DataFlavor(DataFlavor.javaJVMLocalObjectMimeType,"jPicEdt graphic content");
	
	private Element[] content;
	private String formatted;
	
	/**
	 * Construct a new TransferableGraphic from a clone of the given elements and the given String
	 * containing the formatted representation of the elements (this depends on the ContentType of
	 * of the caller).
	 * @param elements an array of elements to be set as the content of this Transferable
	 * @param formatted the text representation of the given elements (can be null)
	 * @since jPicEdt 1.3.2
	 */
	public TransferableGraphic(Element[] elements, String formatted){
		this.content = new Element[elements.length];
		for (int i=0; i<elements.length; i++){
			content[i] = (Element)elements[i].clone();
		}
		if (formatted == null) this.formatted = "";
		else this.formatted = formatted;
	}
	

	/**
	 * @return an Object which represents the data to be transferred.
	 *         Current implementation returns an array of elements if flavor is of type "JPICEDT_DATA_FLAVOR",
	 *         or a string representation of the elements (as given in the constructor) is
	 *         flavor is of type "DataFlavor.stringFlavor".
	 * @param flavor the requested flavor for the data
	 * @since jPicEdt 1.3.2
	 */
	public Object getTransferData(DataFlavor flavor){
		if (flavor.equals(JPICEDT_DATA_FLAVOR)){
			return content;
		}
		else if (flavor.equals(DataFlavor.stringFlavor)){
			return formatted;
		}
		else return null;
		
		
	}
	
	/**
	 * @return an array of DataFlavor objects indicating the flavors the data can be
	 *         provided in.
	 * @since jPicEdt 1.3.2
	 */
	public DataFlavor[] getTransferDataFlavors(){
		DataFlavor[] flavors = new DataFlavor[2];
		flavors[0] = JPICEDT_DATA_FLAVOR;
		flavors[1] = DataFlavor.stringFlavor;
		return flavors;
	}
	
	/**
	 * @return whether or not the specified data flavor is supported for this TransferableGraphic
	 *         Currently only "stringFlavor" and "javaJVMLocalObjectMimeType" are supported.
	 * @since jPicEdt 1.3.2
	 */
	public boolean isDataFlavorSupported(DataFlavor flavor){
		
		if (flavor.equals(DataFlavor.stringFlavor) || flavor.equals(JPICEDT_DATA_FLAVOR))
			return true;
		return false;
	}
	

	/**
	 * Notifies this object that it is no longer the owner of the content of the clipboard
	 * This implementation does nothing.
	 * @since jPicEdt 1.3.2
	 */
	 public void lostOwnership(Clipboard c, Transferable contents){
	 }
	
}

