/*
 PEDrawingBoard.java - January 12, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui;

import jpicedt.Localizer;
import jpicedt.MiscUtilities;
import jpicedt.graphic.*;
import jpicedt.graphic.grid.Grid;
import jpicedt.graphic.event.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.io.parser.*;
import jpicedt.ui.dialog.*;
import jpicedt.ui.util.*;
import jpicedt.widgets.*;


import javax.swing.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.beans.*;
import java.text.*;

/**
 * A drawing board based on jpicedt.graphic.PECanvas, with an embeded scrollpane, and
 * some I/O facilities (saving and loading) which PECanvas doesn't have.<br>
 * There are two accessors two retrieve the embedded canvas and scrollpane. 
 * @since jPicEdt
 * @author Sylvain Reynal
 */
public class PEDrawingBoard extends JPanel {


	private boolean isDirty; // true if it's been modified since the last save operation or creation
	private String title; // full file name with path if saved, otherwise the adequate localization of "untitled" + index
	private boolean isSaved; // true if this board's already been saved to disk
	private PECanvas canvas;
	private PEScrollPane scrollPane;
	//private EditorKitToolBar editorKitToolbar; // deprecated, we now use DockableEditorKitToolBar
	private PopupMenuFactory popupFactory;
	private StatusBar statusBar;
	private GridZoomToolBar gridZoomToolbar;
	private DefaultActionDispatcher actionDispatcher;

	/////////////////////////////////////////////////////
	//// CONSTRUCTORS
	/////////////////////////////////////////////////////

	/**
	 * create a new empty drawing board with a (localized) "untitled" title
	 * @param untitledIndex index appended to "untitled", e.g. "Untitled 2"
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public PEDrawingBoard(int untitledIndex,double zoom, PageFormat pageFormat, Grid grid, ContentType contentType){
		isDirty = false;
		isSaved = false;
		title = Localizer.currentLocalizer().get("misc.Noname") + untitledIndex;
		init(zoom, pageFormat, grid, contentType);
	}

	/**
	 * create a new drawing board pre-loaded from the given path ; set title from this path.
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public PEDrawingBoard(String path,double zoom, PageFormat pageFormat, Grid grid, ContentType contentType,Parser parser){
		isDirty = false;
		isSaved = true;
		title = path;
		init(zoom, pageFormat, grid, contentType);
		load(path,parser);
	}

	/**
	 * create a new empty drawing board with a (localized) "untitled" title
	 * Init zoom, pageFormat, contentType and undoableLimit from the given Properties, using :
	 * - key = PECanvas.KEY_ZOOM, value = a double
	 * - key = PageFormat.KEY_PAGE_FORMAT (see jpicedt.graphic.PageFormat for details)
	 * - key = PECanvas.KEY_CONTENT_TYPE, value = path of ContentType class (e.g. "jpicedt.format.latex.ContentType")
	 * - key = PECanvas.KEY_UNDOABLE_STEPS, value = nb of undoable events to remember
	 *   suitable for instanciation by Class.forName("xxx").newInstance().
	 * "preferences" is also used to init the ContentType, if non-null.
	 * @param untitledIndex index appended to "untitled", e.g. "Untitled 2"
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public PEDrawingBoard(int untitledIndex, Properties preferences){
		double zoom = MiscUtilities.parseProperty(preferences, PECanvas.KEY_ZOOM, PECanvas.ZOOM_DEFAULT);
		PageFormat pf = new PageFormat(preferences);
		Grid grid = new Grid(preferences);
		ContentType ct = MiscUtilities.getContentTypeFromClassName(preferences.getProperty(PECanvas.KEY_CONTENT_TYPE));
		if (ct != null) ct.configure(preferences);
		isDirty = false;
		isSaved = false;
		title = Localizer.currentLocalizer().get("misc.Noname") + untitledIndex;
		init(zoom, pf, grid, ct);
		int undoLimit = MiscUtilities.parseProperty(preferences, PECanvas.KEY_UNDOABLE_STEPS, PECanvas.MAX_UNDOABLE_STEPS_DEFAULT);
		if (undoLimit<1) undoLimit = PECanvas.MAX_UNDOABLE_STEPS_DEFAULT;
		canvas.setUndoLimit(undoLimit);
		canvas.getRenderingHints().add(MiscUtilities.parseRenderingHints(preferences));
	}

	/**
	 * create a new drawing board pre-loaded from the given path ; set title from this path.
	 * Init zoom, pageFormat and contentType from the given Properties, using :
	 * - key = PECanvas.KEY_ZOOM, value = a double
	 * - key = PageFormat.KEY_PAGE_FORMAT (see jpicedt.graphic.PageFormat for details)
	 * - key = PECanvas.KEY_CONTENT_TYPE, value = path of ContentType class (e.g. "jpicedt.format.latex.ContentType")
	 *   suitable for instanciation by Class.forName("xxx").newInstance().
	 * "preferences" is also used to init the ContentType, if non-null.
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public PEDrawingBoard(String path, Properties preferences,Parser parser){
		double zoom = MiscUtilities.parseProperty(preferences, PECanvas.KEY_ZOOM, PECanvas.ZOOM_DEFAULT);
		PageFormat pf = new PageFormat(preferences);
		Grid grid = new Grid(preferences);
		ContentType ct = MiscUtilities.getContentTypeFromClassName(preferences.getProperty(PECanvas.KEY_CONTENT_TYPE));
		if (ct != null) ct.configure(preferences);
		isDirty = false;
		isSaved = true;
		title = path;
		init(zoom, pf, grid, ct);
		int undoLimit = MiscUtilities.parseProperty(preferences, PECanvas.KEY_UNDOABLE_STEPS, PECanvas.MAX_UNDOABLE_STEPS_DEFAULT);
		if (undoLimit<1) undoLimit = PECanvas.MAX_UNDOABLE_STEPS_DEFAULT;
		canvas.setUndoLimit(undoLimit);
		canvas.getRenderingHints().add(MiscUtilities.parseRenderingHints(preferences));
		load(path,parser);
	}

	/*
	 * init component
	 */
	private void init(double zoom, PageFormat pageFormat, Grid grid, ContentType contentType){
		this.setLayout(new BorderLayout(5,5));

		canvas = new PECanvas(zoom, pageFormat, grid, contentType);

		scrollPane = new PEScrollPane(canvas);
		actionDispatcher = new DefaultActionDispatcher(canvas);
		gridZoomToolbar = new GridZoomToolBar(actionDispatcher,jpicedt.Localizer.currentLocalizer().getActionLocalizer());
		statusBar = new StatusBar();

		add(scrollPane,BorderLayout.CENTER);
		add(gridZoomToolbar, BorderLayout.NORTH);
		add(statusBar, BorderLayout.SOUTH);

		gridZoomToolbar.install(canvas); // init widgets state

		canvas.getEditorKit().addHelpMessageListener(statusBar); // listens to edit-mode-change
		canvas.addZoomListener(gridZoomToolbar); // update zoom combo box
		canvas.addPEMouseInputListener(statusBar); // display mouse coords in status bar
	}

	/**
	 * Set the PopupMenuFactory that produces JPopupMenu when a popup-trigger mouse event occurs
	 * on the board.
	 */
	public void setPopupMenuFactory(PopupMenuFactory factory){
		this.popupFactory = factory;
		canvas.getEditorKit().setPopupMenuFactory(popupFactory);
	}

	/**
	 * @return the PopupMenuFactory that produces JPopupMenu when a popup-trigger mouse event occurs
	 * on the board.
	 */
	public PopupMenuFactory getPopupMenuFactory(){
		return popupFactory;
	}


	//////////////////////////////////////////////////////////////
	//// SET/GET
	//////////////////////////////////////////////////////////////


	/**
	 * @author Sylvain Reynal
	 */
	public PECanvas getCanvas(){
		return canvas;
	}

	/**
	 * @author Sylvain Reynal
	 */
	public PEScrollPane getScrollPane(){
		return scrollPane;
	}

	/**
	 * @return whether this board has been already saved to disk (even if it's been modified since then)
	 * This is just a convenience to know if a file path is attached to this board or not.
	 * @author Sylvain Reynal
	 */
	public boolean isSaved(){
		return isSaved;
	}

	/**
	 * @return whether this board has been modified since the last time it was saved or created <p>
	 * This signal that this board need to be saved during "save", "save all",... actions.
	 * @author Sylvain Reynal
	 */
	public boolean isDirty(){
		return isDirty;
	}

	/** mark this board as being dirty */
	public void setDirty(boolean dirty){
		this.isDirty = dirty;
	}


	/**
	 * @return this board's title, e.g. used as the title of the hosting frame  
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public String getTitle(){
		return title;
	}


	////////////////////////////////////////////////////////////
	/// I/O MANAGEMENT
	///////////////////////////////////////////////////////////

	/**
	 * Save drawing content to disk.<p>
	 * if alwaysPrompt is false, save to disk using the current title ; prompt user if it's null,
	 * ie this board has never been saved.<br>
	 * if alwaysPrompt is true, prompt user for a file name (aka "save as...") in any case.
	 * @return false if operation was cancelled by user or an I/O error occured
	 */
	public boolean save(boolean alwaysPrompt){

		String tmpPath; // temp buffer for file name
		if (!isSaved() || alwaysPrompt){
			tmpPath = PEFileChooser.getFileName(PEFileChooser.SAVEFILE, this.canvas.getContentType().getClass().getName());
			if (tmpPath == null) return false; // cancel has occured
		}
		else tmpPath = title;
		canvas.selectAll(false);
		try{
			FileWriter fileWriter = new FileWriter(tmpPath);
			BufferedWriter bufferedWriter = new BufferedWriter(fileWriter);
			canvas.write(bufferedWriter, false); // no fragment
			bufferedWriter.close();
			fileWriter.close();
		}
		catch(IOException ioEx){
			JOptionPane.showMessageDialog(null,
			                              Localizer.currentLocalizer().get("exception.IOError") + " : " + ioEx.getMessage(),
			                              Localizer.currentLocalizer().get("action.ui.Save"),
			                              JOptionPane.ERROR_MESSAGE);
			return false;
		}
		title = tmpPath; // side-effect -> mark as saved to disk
		isSaved = true; // in case this was the first save operation
		isDirty = false; // mark as clean
		return true;
	}

	/**
	 * save the selection content to the "fragment directory" on disk, prompting user for a file name
	 * @return false if operation was cancelled by user or I/O error occured
	 */
	public boolean saveFragment(){

		String fragmentPath = PEFileChooser.getFileName(PEFileChooser.SAVEFRAGMENT, this.canvas.getContentType().getClass().getName());
		if (fragmentPath == null) return false; // CANCEL !
		try {
			FileWriter fileWriter = new FileWriter(fragmentPath);
			BufferedWriter bufferedWriter = new BufferedWriter(fileWriter);
			canvas.write(bufferedWriter, true); // save fragment only
			bufferedWriter.close();
			fileWriter.close();
		}
		catch(IOException ioEx){
			JOptionPane.showMessageDialog(null,
			                              Localizer.currentLocalizer().get("exception.IOError") + " : " + ioEx.getMessage(),
			                              Localizer.currentLocalizer().get("action.ui.Save"),
			                              JOptionPane.ERROR_MESSAGE);
			return false;
		}
		return true;
	}

	/**
	 * Save drawing or selection content the given file.
	 * @param fileName name of file to save content to
	 * @param selectionOnly whether to save the content of the selection or that of the whole the drawing
	 */
	public void save(String fileName, boolean selectionOnly) throws IOException {

		FileWriter fileWriter = new FileWriter(fileName);
		BufferedWriter bufferedWriter = new BufferedWriter(fileWriter);
		canvas.write(bufferedWriter, selectionOnly); 
		bufferedWriter.close();
		fileWriter.close();
	}
	
	/**
	 * Load this board's content from the given path
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void load(String path, Parser parser){

		/* implementation of a ProgressMonitor (underway) */
		// 	ProgressMonitorInputStream pmon = new ProgressMonitorInputStream(null,
		// 									 "Reading " + fullFileName,
		// 									 new FileInputStream(fullFileName));
		//pmon.getProgressMonitor().setMillisToDecideToPopup(10);
		//pmon.getProgressMonitor().setMillisToPopup(50);
		//	InputStreamReader reader = new InputStreamReader(pmon);

		try {
			FileReader fileReader = new FileReader(path);
			BufferedReader bufferedReader = new BufferedReader(fileReader);
			canvas.read(bufferedReader,parser);
			bufferedReader.close();
			fileReader.close();
		}
		catch (ParserException pEx){
			JOptionPane.showMessageDialog(
			    JOptionPane.getRootFrame(),
			    Localizer.currentLocalizer().get("exception.ParserWarning") + " :\n" + pEx,
			    Localizer.currentLocalizer().get("msg.Parsing"),
			    JOptionPane.ERROR_MESSAGE);
		}
		catch (IOException ioEx){
			JOptionPane.showMessageDialog(
			    JOptionPane.getRootFrame(),
			    Localizer.currentLocalizer().get("exception.IOError") + "\n" + ioEx.getMessage(),
			    Localizer.currentLocalizer().get("action.ui.Open"),
			    JOptionPane.ERROR_MESSAGE);
		}
		isDirty = false;
		isSaved = true;
		String notParsedCommand = canvas.getDrawing().getNotparsedCommands();
		if (notParsedCommand.length()>0){
			int res = JOptionPane.showConfirmDialog(
			    JOptionPane.getRootFrame(),
				"The following lines were not parsed : \n" // [pending] i18n
				+ notParsedCommand
				+ "\n Would you like to DISCARD them ?",
			    Localizer.currentLocalizer().get("action.ui.Open"),
			    JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE);
			if (res == JOptionPane.YES_OPTION) canvas.getDrawing().setNotparsedCommands("");
		}
			
	}

	/**
	 * Insert new content from the given path into this board's drawing
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void insert(String path, Parser parser){

		try {
			FileReader fileReader = new FileReader(path);
			BufferedReader bufferedReader = new BufferedReader(fileReader);
			canvas.insert(bufferedReader,parser);
			bufferedReader.close();
			fileReader.close();
		}
		catch (ParserException pEx){
			JOptionPane.showMessageDialog(
			    JOptionPane.getRootFrame(),
			    Localizer.currentLocalizer().get("exception.ParserWarning") + " :\n" + pEx,
			    Localizer.currentLocalizer().get("msg.Parsing"),
			    JOptionPane.ERROR_MESSAGE);
		}
		catch (IOException ioEx){
			JOptionPane.showMessageDialog(
			    JOptionPane.getRootFrame(),
			    Localizer.currentLocalizer().get("exception.IOError") + "\n" + ioEx.getMessage(),
			    Localizer.currentLocalizer().get("action.ui.InsertFragment"),
			    JOptionPane.ERROR_MESSAGE);
		}
		String notParsedCommand = canvas.getDrawing().getNotparsedCommands();
		if (notParsedCommand.length()>0){
			int res = JOptionPane.showConfirmDialog(
			    JOptionPane.getRootFrame(),
				"The following lines were not parsed : \n[" // [pending] i18n
				+ notParsedCommand
				+ "]\n Would you like to DISCARD them ?",
			    Localizer.currentLocalizer().get("action.ui.Open"),
			    JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE);
			if (res == JOptionPane.YES_OPTION) canvas.getDrawing().setNotparsedCommands("");
		}
		isDirty = true;
	}

	/**
	 * Reload the content of this board from its current file name if applicable ; do nothing otherwise.
	 * [pending] should raise an exception in the latter case
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void reload(Parser parser){
		if (!isSaved()) return;
		load(title,parser);
	}

	/**
	 * Return the status bar
	 */
	public StatusBar getStatusBar(){
		return statusBar;
	}

	/////////////////////////////////// Status Bar //////////////////////////////
	/**
	 * jpicedt's Status Bar
	 *
	 * @author 	Sylvain Reynal
	 * @since       PicEdt 1.0
	 */

	public class StatusBar extends JPanel implements PEMouseInputListener, HelpMessageListener {

		private JLabel mouseCoord; // displays mouse coordinates
		private JLabel toolTip; // displays a tooltip
		private PicPoint ptBuffer=new PicPoint(); // buffer
		private DecimalFormat nf = new DecimalFormat("#0.00", new DecimalFormatSymbols(Locale.US));

		// member methods

		/**
		 * mouseCoord | toolTip 
		 */
		public StatusBar() {

			nf.setMaximumFractionDigits(2); // mouse coords formatting

			this.setBorder(BorderFactory.createEtchedBorder());

			// init panel :
			GridBagLayout gridbag = new GridBagLayout();
			GridBagConstraints gridbagC = new GridBagConstraints();
			this.setLayout(gridbag);

			gridbagC.fill = GridBagConstraints.HORIZONTAL;

			gridbagC.gridwidth = 1;
			gridbagC.gridheight = 1;
			gridbagC.weightx=0;

			mouseCoord = new JLabel("(0.00,0.00)");
			mouseCoord.setBorder(BorderFactory.createLoweredBevelBorder());
			mouseCoord.setFont(new Font("SansSerif", Font.PLAIN, 12));
			mouseCoord.setForeground(Color.black);
			gridbag.setConstraints(mouseCoord, gridbagC);

			toolTip = new JLabel("     ");
			toolTip.setBorder(BorderFactory.createLoweredBevelBorder());
			toolTip.setFont(new Font("SansSerif", Font.PLAIN, 12));
			toolTip.setForeground(Color.black);
			gridbagC.weightx=2.0;
			gridbagC.gridwidth = GridBagConstraints.REMAINDER;
			gridbag.setConstraints(toolTip, gridbagC);

			this.add(mouseCoord);
			this.add(toolTip);
		}


		/**
		 * Displays a tooltip about current possible mouse-actions
		 * @since PicEdt 1.0
		 */
		public void showMessage(String str){
			StringBuffer buf = new StringBuffer(50);
			buf.append(" ");
			buf.append(str);
			toolTip.setText(buf.toString());
		}

		/**
		 * Displays current mouse coordinates
		 * @since PicEdt 1.0
		 */
		public void showCoord(PicPoint pt){
			StringBuffer buf = new StringBuffer(15);
			buf.append('(');
			buf.append(nf.format(pt.x));
			buf.append(',');
			buf.append(nf.format(pt.y));
			buf.append(')');
			mouseCoord.setText(buf.toString());
		}

		////////////////////////////////////////////////////
		/// EVENT
		////////////////////////////////////////////////////

		public void helpMessagePosted(HelpMessageEvent e){
			String s = e.getMessage();
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Board "+getTitle()+" received msg:"+s);
			if (s==null) return;
			toolTip.setText(jpicedt.Localizer.currentLocalizer().get(s));
		}

		/**
		 * display mouse coords
		 */
		public void mouseDragged(PEMouseEvent e){
			mouseMoved(e);
		}

		/**
			 * display mouse coords
		 */
		public void mouseMoved(PEMouseEvent e){
			PicPoint ptClick = e.getPicPoint();
			e.getCanvas().getGrid().nearestNeighbour(ptClick,ptBuffer);
			showCoord(ptBuffer);
		}

		public void mousePressed(PEMouseEvent e){}
		public void mouseReleased(PEMouseEvent e){}
		public void mouseClicked(PEMouseEvent e){}
		public void mouseEntered(PEMouseEvent e){}
		public void mouseExited(PEMouseEvent e){}

	} // StatusBar


	////////////////////////////////////// GRID & ZOOM ////////////////////////////////
	/**
	 * A toolbar that allows the user to change the grid and zoom properties
	 * @author Sylvain Reynal
	 * @since jPicEdt-beta9
	 */
	public class GridZoomToolBar extends jpicedt.graphic.toolkit.PEToolBar
		implements PicObjectConstants, ZoomListener {

		private ActionDispatcher actionDispatcher;

		// --------------- grid-on/off ------------------------
		private JToggleButton gridVisibleTB, snapOnOffTB;
		private PEToggleAction gridVisibleAction,snapOnOffAction;

		// --------------- grid- and snap-step ------------------------
		private JComboBox snapStepCB, gridStepCB;
		private GridStepAction gridStepAction;
		private SnapStepAction snapStepAction;
		// see Grid for PREDEFINED_SNAP_STEP_STRINGS and PREDEFINED_SNAP_STEPS

		// --------------- zoom ------------------------
		private JComboBox zoomCB;
		private ZoomAction zoomAction;
		// see PECanvas for PREDEFINED_ZOOM_STRINGS and PREDEFINED_ZOOMS

		
		// --- class methods ------------------------------------------------

		/**
		 * Constructor
		 */
		public GridZoomToolBar(ActionDispatcher actionDispatcher, ActionLocalizer localizer) {

			this.actionDispatcher = actionDispatcher;
			this.setOrientation(HORIZONTAL);
			this.setFloatable(true);
			this.setBorder(BorderFactory.createEtchedBorder());

			// grid ON/OFF:
			gridVisibleTB = add(gridVisibleAction=new EditorKit.ShowGridAction(actionDispatcher, localizer));
			Dimension dimButton = gridVisibleTB.getPreferredSize(); // used later to compute other widgets maximum size
			addSeparator();

			// grid steps:
			gridStepCB = new JComboBox(Grid.PREDEFINED_GRID_STEP_STRINGS);
			gridStepCB.setEditable(true);
			gridStepCB.setAlignmentY(CENTER_ALIGNMENT);
			gridStepCB.addActionListener(gridStepAction=new GridStepAction());
			gridStepCB.setMaximumSize(new Dimension(dimButton.width * 4,dimButton.height));
			this.add(gridStepCB);
			addSeparator();

			// snap ON/OFF:
			snapOnOffTB = add(snapOnOffAction=new EditorKit.ActivateSnapAction(actionDispatcher, localizer));
			addSeparator();

			// snap steps:
			snapStepCB = new JComboBox(Grid.PREDEFINED_SNAP_STEP_STRINGS);
			snapStepCB.setEditable(true);
			snapStepCB.setAlignmentY(CENTER_ALIGNMENT);
			snapStepCB.addActionListener(snapStepAction=new SnapStepAction());
			snapStepCB.setMaximumSize(new Dimension(dimButton.width * 4,dimButton.height));
			this.add(snapStepCB);
			addSeparator();

			// zoom factors:
			zoomCB = new JComboBox(PECanvas.PREDEFINED_ZOOM_STRINGS);
			zoomCB.setEditable(true);
			zoomCB.setAlignmentY(CENTER_ALIGNMENT);
			zoomCB.addActionListener(zoomAction=new ZoomAction());
			zoomCB.setMaximumSize(new Dimension(dimButton.width * 3,dimButton.height));
			this.add(zoomCB);
			addSeparator();
			
			// remaining gap:
			this.add(Box.createGlue());

			//this.setMargin(new Insets(0,0,0,0));
		}

		/**
		 * Called when this tool-bar is installed, and widgets values must reflect a given PECanvas state. 
		 * @param canvas the source from which this toolbar updates the value or state of its widgets
		 */
		public void install(PECanvas canvas){
			// grid visibility
			gridVisibleTB.setSelected(canvas.getGrid().isVisible());
			// snap on-off
			snapOnOffTB.setSelected(canvas.getGrid().isSnapOn());
			// snap-step
			int i = Grid.getSnapStepIndex(canvas.getGrid().getSnapStep());
			if (i != -1) snapStepCB.setSelectedIndex(i);
			else { // add non-predefined snap step to list
				snapStepCB.addItem(new Double(canvas.getGrid().getSnapStep()).toString());
				snapStepCB.setSelectedIndex(snapStepCB.getItemCount()-1);
			}
			// grid-step
			i = Grid.getGridStepIndex(canvas.getGrid().getGridStep());
			if (i != -1) gridStepCB.setSelectedIndex(i);
			else { // add non-predefined grid step to list
				gridStepCB.addItem(new Double(canvas.getGrid().getGridStep()).toString());
				gridStepCB.setSelectedIndex(gridStepCB.getItemCount()-1);
			}
			// zoom
			i = PECanvas.getZoomIndex(canvas.getZoomFactor());
			if (i != -1) zoomCB.setSelectedIndex(i);
			else {
				NumberFormat formatPercent = NumberFormat.getPercentInstance(Locale.US);
				formatPercent.setGroupingUsed(false);
				String s = formatPercent.format(canvas.getZoomFactor());
				zoomCB.addItem(s);
				zoomCB.setSelectedIndex(zoomCB.getItemCount()-1);
			}
		}

		/** ZoomListener interface ; called when the zoom changed in the sourcing canvas
		 *  following a ZoomTool operation -> update widget */ 
		public void zoomUpdate(ZoomEvent e){
			double zoom = e.getNewZoomValue();
			NumberFormat formatPercent = NumberFormat.getPercentInstance(Locale.US);
			formatPercent.setGroupingUsed(false);
			String s = formatPercent.format(zoom); // display in "%" format
			int i;
			boolean b;
			for (i=0,b=false; i<zoomCB.getItemCount(); i++){
				if (zoomCB.getItemAt(i).equals(s)) { // ComboBox was built from e.g. "200%" Strings
					zoomCB.setSelectedIndex(i); b=true;
					break;
				}
			}
			if (b==false) {
				zoomCB.addItem(s);
				zoomCB.setSelectedIndex(zoomCB.getItemCount()-1);
			}
		}

		///////////////////////////////////////////////////////////////////////////////////
		//// ACTIONS
		///////////////////////////////////////////////////////////////////////////////////

		/**
		 * Updates the grid-step in the current active frame as soon as 
		 * a new item gets selected by user in the list (aka JComboBox) of available grid steps
		 *
		 * @author Sylvain Reynal
		 * @since jPicEdt 1.4
		 */
		class GridStepAction extends PEAction {

			public GridStepAction(){
				super(actionDispatcher,"GridStepList",null);
			}

			public void actionPerformed(ActionEvent e){
				int i = gridStepCB.getSelectedIndex();
				this.getCanvas().getGrid().setGridStep(Double.parseDouble((String)gridStepCB.getSelectedItem()));
				this.getCanvas().repaint();
			}
		}

		/**
		 * Updates the snap-step in the current active frame as soon as 
		 * a new item gets selected by user in the list (aka JComboBox) of available grid steps
		 *
		 * @author Sylvain Reynal
		 * @since PicEdt 1.2.1
		 */
		class SnapStepAction extends PEAction {

			public SnapStepAction(){
				super(actionDispatcher,"SnapStepList",null);
			}

			public void actionPerformed(ActionEvent e){
				int i = snapStepCB.getSelectedIndex();
				this.getCanvas().getGrid().setSnapStep(Double.parseDouble((String)snapStepCB.getSelectedItem()));
				this.getCanvas().repaint();
			}
		}

		/**
		 * Updates zoom-factor in the current active frame as soon as 
		 * a new item gets selected by user in the list of available zoom factors (e.g. 100%, 200%, ...)
		 * @author Sylvain Reynal
		 * @since PicEdt 1.2.1
		 */
		class ZoomAction extends PEAction {

			public ZoomAction(){
				super(actionDispatcher,"ZoomList",null);
			}

			public void actionPerformed(ActionEvent e){
				int i = zoomCB.getSelectedIndex();
				try {
					double zoom = jpicedt.MiscUtilities.parseDouble((String)zoomCB.getSelectedItem());
					if (zoom == this.getCanvas().getZoomFactor()) return; // no change
					this.getCanvas().setZoomFactor(zoom);
				}
				catch(Exception ex){
					ex.printStackTrace();
					JOptionPane.showMessageDialog(null,
					                              jpicedt.Localizer.currentLocalizer().get("exception.NumberFormatError") + " : " + ex.getMessage(),
					                              jpicedt.Localizer.currentLocalizer().get("preferences.Zoom"),
					                              JOptionPane.ERROR_MESSAGE);
					return;
				}
				this.getCanvas().repaint();
			}
		}
	} // GridZoomToolBar

} // class

