/*
 PEActionLocalizer.java - January 2, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui.action;

import jpicedt.graphic.toolkit.ActionLocalizer;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.KeyStroke;


/**
 * A class that can feed PEAction's with localized properties (e.g. tooltip,...)
 * provided by a properties file through a ResourceBundle.<p> 
 * Each method takes a unique String argument, namely the non-localized Action name used in PEAction
 * constructor, and builds other Action's properties from this String.<p>
 * If there's no associated value for this String in the associated ResourceBundle, methods
 * catch the associated MissingResourceException and return a null value.
 * @since jPicEdt
 * @author Sylvain Reynal
 * @version $Id: PEActionLocalizer.java,v 1.7.2.1 2007/09/02 11:56:30 reynal Exp $
 *
 */
public class PEActionLocalizer implements ActionLocalizer {

	/** the ResourceBundle that feeds this localizer with localized Strings */
	protected ResourceBundle resBundle;
	
	/**
	 * Construct a new PEActionLocalizer from the given ResourceBundle.
	 */
	public PEActionLocalizer(ResourceBundle resBundle){
		this.resBundle = resBundle;
	}
	
	/**
	 * @return  the ResourceBundle that feeds this localizer with localized Strings
	 */
	public ResourceBundle getResourceBundle(){ 
		return resBundle;
	}
	
	/**
	 * @return a localized Action name (ie Swing Action.NAME property) for the given name<br>
	 * This in effect will return the value associated with "actionName" in the ResourceBundle 
	 * associated with the localizer.
	 */
	public String getActionName(String actionName){ 
		try { 
			return resBundle.getString(actionName);
		}
		catch (MissingResourceException mre){
			return actionName; //+"[missing]";
		}
	}
	
	/**
	 * @return a localized Action tooltip for the given name (ie Swing Action.SHORT_DESCRIPTION)<p>  
	 * This will return the value for the key build from : actionName + ".tooltip".
	 */
	public String getActionTooltip(String actionName){
		try { return resBundle.getString(actionName + ".tooltip");}
		catch (MissingResourceException mre){
			//jpicedt.Log.warning(actionName+".tooltip [missing]");
			return null;
		}
	}
	
	/**
	 * @return a localized Action helper (ie Swing Action.LONG_DESCRIPTION)<p> 
	 * This will return the value for the key build from : actionName + ".helper".
	 */
	public String getActionHelper(String actionName){
		try { return resBundle.getString(actionName + ".helper");}
		catch (MissingResourceException mre){return null;}
	}
	
	/**
	 * @return a localized Action mnemonic (ie Swing Action.MNEMONIC_KEY)<p> 
	 * This will return a Character initialized from the first character
	 * of the value associated with the key : actionName + ".mnemonic".
	 */
	public Integer getActionMnemonic(String actionName){
		try { 
			return new Integer(resBundle.getString(actionName + ".mnemonic").charAt(0));
		}
		catch (MissingResourceException mre){
			//jpicedt.Log.warning(actionName+".mnemonic [missing]");
			return null;
		}
	}
	
	/**
	 * @return a localized Action accelerator keystroke (ie Swing Action.ACCELERATOR_KEY)<p> 
	 * This will return a KeyStroke parsed from the following String: actionName + ".accelerator". 
	 * @see javax.swing.KeyStroke#getKeyStroke(String)
	 */
	public KeyStroke getActionAccelerator(String actionName){
		try { 
			return KeyStroke.getKeyStroke(resBundle.getString(actionName + ".accelerator"));
		}
		catch (MissingResourceException mre){return null;}
	}
	
	/**
	 * @return a localized Icon (ie Swing Action.SMALL_ICON)<p>
	 * This will return an ImageIcon initialized from a GIF file with path : "/jpicedt/images/"+actionName+".gif"
	 * <br>Return null if the GIF file doesn't exist.
	 */
	public Icon getActionIcon(String actionName){
		java.net.URL u = getClass().getResource("/jpicedt/images/" + actionName + ".gif");
		if (u!=null) return new ImageIcon(u);
		else {
			//jpicedt.Log.warning("/jpicedt/images/" + actionName + ".gif not found");
			return null;
		}
	}
	
} // class 

