/*
 DockableAttributesCustomizer.java - February 27, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui.dialog;

import jpicedt.Localizer;
import jpicedt.JPicEdt;
import jpicedt.ui.MDIManager;
import jpicedt.ui.PEDrawingBoard;
import jpicedt.widgets.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.event.*;

import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.beans.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.colorchooser.*;
import java.util.*;

/**
 * A dockable customizer for attributes (ie PicAttributesSet).
 * This is currently implemented as a floating JPanel.
 * <p>
 * @author Sylvain Reynal
 * @since jPicEdt 1.3.2
 * <p>
 */
public class DockableAttributesCustomizer extends JPanel 
	implements PicObjectConstants, SelectionListener, ChangeListener, PropertyChangeListener { 

	/** key for persistent storage */
	public static final String KEY = "dockable-panel.Attributes";

	//private Element obj; // current selection
	private PicAttributeSet inputAttributes=new PicAttributeSet(); // work-around for start-up
	private JLabel titleLbl;
	private HatchPropertiesPanel hatchPropertiesPanel; // enabled/disabled by FillPropertiesPanel
	private JTabbedPane tabbedPane;
	private Box box; // container for the tabbed pane
	private PECanvas canvas; // selectionUpdate takes care of keeping the customizer state synchronized with the selection content of the active canvas
	private boolean isEditingUnderway=false; // trick to know when to add a new UndoableEvent

	/** for debugging purpose ; run it if you just wanna see how this particular GUI looks like
	 * w/o having to run the whole jPicEdt application */
	public static void main(String[] args){
		
		JFrame f = new JFrame("test");
		f.setDefaultCloseOperation(f.EXIT_ON_CLOSE);
		f.getContentPane().add(new DockableAttributesCustomizer(null));
		f.pack();
		f.setVisible(true);
		
	}
	
	/**
	 * Create a DockableAttributesCustomizer with no specific PECanvas attached to it.
	 * @param progressBar when non-null, increment() is called at some points during the construction process.
	 */
	public DockableAttributesCustomizer(jpicedt.ui.util.PEProgressBar progressBar){
		setLayout(new BorderLayout(5,5));
		box = new Box(BoxLayout.Y_AXIS);
		
		// title
		titleLbl = new JLabel("[]"); // no title by default
		box.add(titleLbl);
		
		if (progressBar != null) progressBar.increment();

		// attribute pane :
		tabbedPane = new JTabbedPane();
		tabbedPane.addChangeListener(this);
		AbstractCustomizer[] customizers = new AbstractCustomizer[8];
		if (progressBar != null) progressBar.increment();
		customizers[0] = new FillPropertiesPanel();
		customizers[1] = hatchPropertiesPanel  =new HatchPropertiesPanel();
		customizers[2] = new StrokePropertiesPanel();
		customizers[3] = new ShadowPropertiesPanel();
		customizers[4] = new ArrowPropertiesPanel();
		customizers[5] = new PolydotsPropertiesPanel();
		customizers[6] = new TextPropertiesPanel();
		customizers[7] = new PstCustomPropertiesPanel();
		for (int i=0; i<customizers.length; i++){
			tabbedPane.addTab(customizers[i].getTitle(),customizers[i].getIcon(),customizers[i],customizers[i].getTooltip());
		}

		box.add(tabbedPane);
		this.add(box,BorderLayout.CENTER);		
	}
	
	/**
	 * Changes the customizer title
	 */
	public void setTitle(String title){
		StringBuffer buf = new StringBuffer();
		buf.append("JPicEdt: ");
		buf.append('[');
		buf.append(title);
		buf.append(']');
		titleLbl.setText(buf.toString());
	}
	
	/**
	 * Called :
	 * - by the colorchooser if the user picks a new colour ;
	 * - or by the tabbed pane when a different pane is made active (ie brought on top)
	 */
	public void stateChanged(ChangeEvent e){
		//System.out.println(e);
		if (e.getSource()==tabbedPane){
			//System.out.println("Inner pane " + ((AbstractCustomizer)tabbedPane.getSelectedComponent()).getTitle() + " was made active...");
			((AbstractCustomizer)tabbedPane.getSelectedComponent()).load(); 
		}
	}
			
		
	
	/**
	 * Implementation of the SelectionListener interface ; 
	 * called when a change occurs in the selection from the target PECanvas
	 */
	 public void selectionUpdate(SelectionEvent e){
		 if (isEditingUnderway) {
			 canvas.endUndoableUpdate(); // mark compound edit as completed before
			 isEditingUnderway = false;
		 }
		 // either switching to a new active board, or starting a new edit.
		 
		// SelectionEvent.EventType type = e.getType(); not used
		canvas = (PECanvas)e.getSource();
		// synchronize state of active pane with newly selected object (or with input attribute):
		AbstractCustomizer c = (AbstractCustomizer)tabbedPane.getSelectedComponent();
		c.load();
		
	 }
	 
	 /**
	  * Update customizer's title from the given event if it's a MDIManager.ACTIVE_BOARD_CHANGE event.
	  */
	 public void propertyChange(PropertyChangeEvent e){ // posted from MDIManager.MDIDesktopManager
		if (e.getPropertyName()==MDIManager.ACTIVE_BOARD_CHANGE){
			if (e.getNewValue()==null) setTitle("");
			else {
				PEDrawingBoard b = (PEDrawingBoard)e.getNewValue();
				setTitle(b.getTitle());
				// make all open editor-kit share the same input attribute set:
				b.getCanvas().getEditorKit().setInputAttributes(inputAttributes); // not a deep copy !
				selectionUpdate(new SelectionEvent(b.getCanvas(), (Element)null, null));
			}
		}
	 }
		 

	 /**
	  * @return the attribute value for the given attribute name ; guaranteed to never return a null pointer.
	  * Depending on the size of the selection in the currently active drawing board, either fetch
	  * attribute value from the selected graphical element, or from the editor-kit's input attribute-set.
	  */
	 private Object getAttribute(PicAttributeName name){
		if (canvas==null || canvas.getSelectionSize()==0) return inputAttributes.getAttribute(name); 
		/*
		if (canvas.getSelectionSize()==0){
			return canvas.getEditorKit().getInputAttributes().getAttribute(name);
		}
		*/
		// or fetch attributes from the first selected element 
		// [pending] we should "gray" the widgets or sth if more than two elements are selected
		else { 
			return ((Element)canvas.selection().next()).getAttribute(name);
		}
	 }

	/**
  	 * Set the attribute value for the given attribute name. Depending on the size of the selection
	 * in the currently active drawing board, either acts upon the attribute set of the selected graphical
	 * element, or upon the input attribute-set of the editor kit attached to this drawing board.
  	 */
	private void setAttribute(PicAttributeName name, Object value){
		//System.out.println("canvas=" + canvas);
		if (canvas==null || canvas.getSelectionSize()==0) {
			this.inputAttributes.setAttribute(name,value); // safe copy
			/*
			// set same input attribute set for all open EditorKit's:
			MDIManager mdiManager = JPicEdt.getMDIManager();
			PEDrawingBoard[] allOpenBoards = mdiManager.getAllDrawingBoards();
			for (int i=0; i<allOpenBoards.length; i++){
				EditorKit kit = allOpenBoards[i].getCanvas().getEditorKit();
				if (kit != null) kit.getInputAttributes().setAttribute(name,value);
			}
			*/
			return;
		}
		else {
			if (!isEditingUnderway) { // start a new compound edit ; end-update is done when de-selecting
				canvas.beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("attributes.Edit"));
				isEditingUnderway=true;
			}
			for(Iterator it = canvas.selection(); it.hasNext(); ){
 				 Element e = (Element)it.next();
				 e.setAttribute(name, value);
				 //System.out.println("setAttribute : name = " + name + ", value = " + value + " for e = " + e);
 			}
		}
	}
	
	/**
 	 * A panel for editing of Element's fill attributes 
	 */
	public class FillPropertiesPanel extends AbstractCustomizer  implements ActionListener, ChangeListener {

		private JComboBox fillStyleCB;
		private JColorChooser colorChooser;

		/**
		 * creates a JPanel for attributes editing
		 */
		public FillPropertiesPanel(){

			super();
			this.add(createFillPanel(), BorderLayout.NORTH);
			// add listeners
			fillStyleCB.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); // fire change event when user picks a colour
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Fill");
		}

		/**
		 * creates a sub-panel for "fill" properties
		 */
		private JComponent createFillPanel(){

			Box box = new Box(BoxLayout.Y_AXIS);
			// 1) fillStyle :
			ImageIcon[] iconArray = new ImageIcon[8];
			iconArray[0] = PEToolKit.createImageIcon("attributes.FillStyleNone");
			iconArray[1] = PEToolKit.createImageIcon("attributes.FillStyleSolid");
			iconArray[2] = PEToolKit.createImageIcon("attributes.FillStyleVlines");
			iconArray[3] = PEToolKit.createImageIcon("attributes.FillStyleVlinesFilled");
			iconArray[4] = PEToolKit.createImageIcon("attributes.FillStyleHlines");
			iconArray[5] = PEToolKit.createImageIcon("attributes.FillStyleHlinesFilled");
			iconArray[6] = PEToolKit.createImageIcon("attributes.FillStyleCrosshatches");
			iconArray[7] = PEToolKit.createImageIcon("attributes.FillStyleCrosshatchesFilled");
			fillStyleCB = createComboBox(iconArray);
			box.add(fillStyleCB);

			// 2) color chooser :
			colorChooser = createColorChooser();
			box.add(colorChooser);
			
			return box;
		}

		/**
		 * update widgets values according to currently active attribute set.
		 */
		public void load(){

			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			fillStyleCB.removeActionListener(this); 
			colorChooser.getSelectionModel().removeChangeListener(this); 
			
			colorChooser.setColor((Color)getAttribute(FILL_COLOR));
			
			if (getAttribute(FILL_STYLE) == NONE) {
				fillStyleCB.setSelectedIndex(0);
				hatchPropertiesPanel.setEnabled(false);
			}
			else if (getAttribute(FILL_STYLE) == SOLID) {
				fillStyleCB.setSelectedIndex(1);
				hatchPropertiesPanel.setEnabled(false);
			}
			else {
				if (getAttribute(FILL_STYLE) == VLINES) fillStyleCB.setSelectedIndex(2);
				else if (getAttribute(FILL_STYLE) == VLINES_FILLED) fillStyleCB.setSelectedIndex(3);
				else if (getAttribute(FILL_STYLE) == HLINES) fillStyleCB.setSelectedIndex(4);
				else if (getAttribute(FILL_STYLE) == HLINES_FILLED) fillStyleCB.setSelectedIndex(5);
				else if (getAttribute(FILL_STYLE) == CROSSHATCH) fillStyleCB.setSelectedIndex(6);
				else if (getAttribute(FILL_STYLE) == CROSSHATCH_FILLED) fillStyleCB.setSelectedIndex(7);
				hatchPropertiesPanel.setEnabled(true);
			}

			fillStyleCB.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); // fire change event when user picks a colour
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){
		}

		/**
		 * Called by the colorchooser when the user picks a new colour, and this customizer lies currently
		 * on top.
		 */
		public void stateChanged(ChangeEvent e){
			setAttribute(FILL_COLOR,colorChooser.getColor());
		}
		
		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){

				int fillStyle = fillStyleCB.getSelectedIndex();
				if (fillStyle==0 || fillStyle==1){ // NONE or SOLID
					switch (fillStyle){
						case 0 : setAttribute(FILL_STYLE, NONE); break;
						case 1 : setAttribute(FILL_STYLE, SOLID);break;
						default:
					}
					hatchPropertiesPanel.setEnabled(false);
				}
				else { // HATCHES
					switch (fillStyle){
						case 2 : setAttribute(FILL_STYLE, VLINES);  break;
						case 3 : setAttribute(FILL_STYLE, VLINES_FILLED); break;
						case 4 : setAttribute(FILL_STYLE, HLINES); break;
						case 5 : setAttribute(FILL_STYLE, HLINES_FILLED); break;
						case 6 : setAttribute(FILL_STYLE, CROSSHATCH); break;
						case 7 : setAttribute(FILL_STYLE, CROSSHATCH_FILLED); break;
						default:
					}
					hatchPropertiesPanel.setEnabled(true);
				}
		}

	} // class


	/**
 	 * A panel for editing of Element's hatch attributes 
	 */
	public class HatchPropertiesPanel extends AbstractCustomizer  implements ActionListener, ChangeListener {

		private IncrementableTextField hatchWidthTF, hatchSepTF, hatchAngleTF;
		private JColorChooser colorChooser; // hatches color

		/**
		 * creates a JPanel for attributes editing
		 */
		public HatchPropertiesPanel(){

			super();
			this.add(createHatchPanel(), BorderLayout.NORTH);
			// add listeners
			hatchWidthTF.addActionListener(this);
			hatchSepTF.addActionListener(this);
			hatchAngleTF.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); // fire change event when user picks a colour
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Hatch");
		}

		public void setEnabled(boolean b){
			hatchWidthTF.setEnabled(b);
			hatchSepTF.setEnabled(b);
			hatchAngleTF.setEnabled(b);
			//colorChooser.setEnabled(b); // ca marche pas !
		}
		
		/**
		 * creates a sub-panel for "fill" properties
		 */
		private JComponent createHatchPanel(){

			Box box = new Box(BoxLayout.Y_AXIS);

			JPanel p = new JPanel(new GridLayout(1,3,5,5));
			// 1) hatch width :
			hatchWidthTF = createIncrementableTextField(0.05,"attributes.HatchWidth","mm");
			hatchWidthTF.setMinimum(0);
			hatchWidthTF.setLowBounded(true);
			p.add(hatchWidthTF);

			// 2) hatch sep :
			hatchSepTF = createIncrementableTextField(0.05,"attributes.HatchSep","mm");
			hatchSepTF.setMinimum(0);
			hatchSepTF.setLowBounded(true);
			p.add(hatchSepTF);

			// 3) hatch angle :
			hatchAngleTF = createIncrementableTextField(5.0,"attributes.HatchAngle","deg");
			p.add(hatchAngleTF);
			box.add(p);
			
			// 4) hatch color :
			box.add(colorChooser = createColorChooser());
			
			return box;
		}

		/**
		 * update widgets values according to currently active attribute set.
		 */
		public void load(){

			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			hatchWidthTF.removeActionListener(this);
			hatchSepTF.removeActionListener(this);
			hatchAngleTF.removeActionListener(this);
			colorChooser.getSelectionModel().removeChangeListener(this); 

			hatchWidthTF.setValue(((Double)getAttribute(HATCH_WIDTH)).doubleValue());
			hatchSepTF.setValue(((Double)getAttribute(HATCH_SEP)).doubleValue());
			hatchAngleTF.setValue(((Double)getAttribute(HATCH_ANGLE)).doubleValue());
			colorChooser.setColor((Color)getAttribute(HATCH_COLOR));

			hatchWidthTF.addActionListener(this);
			hatchSepTF.addActionListener(this);
			hatchAngleTF.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); // fire change event when user picks a colour
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){
		}

		/**
		 * Called by the colorchooser when the user picks a new colour, and this customizer lies currently
		 * on top.
		 */
		public void stateChanged(ChangeEvent e){
			setAttribute(HATCH_COLOR,colorChooser.getColor());
		}
		
		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){

			// HATCH WIDTH
			if (e.getSource() == hatchWidthTF){
				setAttribute(HATCH_WIDTH, new Double(hatchWidthTF.getValue()));
			}
			// HATCH SEP
			else if (e.getSource() == hatchSepTF){
				setAttribute(HATCH_SEP, new Double(hatchSepTF.getValue()));
			}
			// HATCH ANGLE
			else if (e.getSource() == hatchAngleTF){
				setAttribute(HATCH_ANGLE, new Double(hatchAngleTF.getValue()));
			}
		}

	} // class


	// -------------------- Arrows ---------------------------- //

	/**
	 * A panel for editing of Element's attributeSet 
	 *
	 * see GraphicPropertiesToolBar which is very similar (though it has less widget than this Panel)
	 */
	public class ArrowPropertiesPanel extends AbstractCustomizer  implements ActionListener {

		private JComboBox leftArrowCB,rightArrowCB;
		private JTabbedPane tabbedPane; // one pane for each family of arrows
		private ImageIcon[] arrowIconsLeft, arrowIconsRight;
		private IncrementableTextField globalScaleWidthTF, globalScaleLengthTF;
		private IncrementableTextField headWidthMinimumMmTF, headWidthLineWidthScaleTF;
		private IncrementableTextField lengthScaleTF, insetScaleTF;
		private IncrementableTextField tbarWidthMinimumMmTF, tbarWidthLineWidthScaleTF;
		private IncrementableTextField bracketLengthScaleTF, roundedBracketLengthScaleTF;
		private IncrementableTextField dotSizeMinimumMmTF, dotSizeLineWidthScaleTF;
		
		/**
		 * creates a JPanel for attributes editing
		 */
		public ArrowPropertiesPanel(){

			super();
			arrowIconsLeft = Arrow.createArrowIcons(false);
			arrowIconsRight = Arrow.createArrowIcons(true);
			this.add(createArrowPanel(),BorderLayout.NORTH);
			// add listeners...
			leftArrowCB.addActionListener(this);
			rightArrowCB.addActionListener(this);
			globalScaleWidthTF.addActionListener(this);
			globalScaleLengthTF.addActionListener(this);
			headWidthMinimumMmTF.addActionListener(this);
			headWidthLineWidthScaleTF.addActionListener(this);
			lengthScaleTF.addActionListener(this);
			insetScaleTF.addActionListener(this);
			tbarWidthMinimumMmTF.addActionListener(this);
			tbarWidthLineWidthScaleTF.addActionListener(this);
			bracketLengthScaleTF.addActionListener(this);
			roundedBracketLengthScaleTF.addActionListener(this);
			dotSizeMinimumMmTF.addActionListener(this);
			dotSizeLineWidthScaleTF.addActionListener(this);
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Arrow");
		}

		/**
		 * create a sub-panel for arrows
		 */
		private Box createArrowPanel(){
			Box box = new Box(BoxLayout.Y_AXIS);
			JPanel p;
			
			// *) arrow choice panel
			p = new JPanel(new GridLayout(1,2,5,5));
			// left arrow:
			leftArrowCB = createComboBox(arrowIconsLeft);
			p.add(leftArrowCB);
			// right arrow:
			rightArrowCB = createComboBox(arrowIconsRight);
			p.add(rightArrowCB);
			box.add(p);
			
			// *) global scale factors
			p = new JPanel(new GridLayout(1,2,5,5));
			// global scale factors
			p.add(globalScaleWidthTF=new IncrementableTextField(0, 0.05, createLabel("attributes.ArrowGlobalScaleWidth"), null, false));
			globalScaleWidthTF.setMinimum(0);
			globalScaleWidthTF.setLowBounded(true);
			p.add(globalScaleLengthTF=new IncrementableTextField(0, 0.05, createLabel("attributes.ArrowGlobalScaleLength"), null, false));
			globalScaleLengthTF.setMinimum(0);
			globalScaleLengthTF.setLowBounded(true);
			box.add(p);
			
			// *) params specific to each arrow family
			tabbedPane = new JTabbedPane();
			
			// a) arrow heads: width
			JPanel pHeads = new JPanel(new GridLayout(3,1,5,5));
			pHeads.setBorder(BorderFactory.createEtchedBorder());
			p = new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowHeadWidth"));
			p.add(headWidthMinimumMmTF=new IncrementableTextField(0, 0.05, new JLabel("="), new JLabel("mm"), false));
			headWidthMinimumMmTF.setMinimum(0);
			headWidthMinimumMmTF.setLowBounded(true);
			p.add(new JLabel("+ "));
			p.add(createLabel("attributes.LineWidth"));
			p.add(headWidthLineWidthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("* "), null, false));
			headWidthLineWidthScaleTF.setMinimum(0);
			headWidthLineWidthScaleTF.setLowBounded(true);
			pHeads.add(p);
			// *) arrow heads: length
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowHeadLength"));
			p.add(lengthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("="), null, false));
			lengthScaleTF.setMinimum(0);
			lengthScaleTF.setLowBounded(true);
			p.add(new JLabel("* ("));
			p.add(createLabel("attributes.ArrowHeadWidth"));
			p.add(new JLabel(")"));
			pHeads.add(p);
			// *) arrow heads: inset
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowHeadInset"));
			p.add(insetScaleTF=new IncrementableTextField(0, 0.05, new JLabel("="), null, false));
			insetScaleTF.setLowBounded(false);
			p.add(new JLabel("* ("));
			p.add(createLabel("attributes.ArrowHeadWidth"));
			p.add(new JLabel(")"));
			pHeads.add(p);
			tabbedPane.addTab(jpicedt.Localizer.currentLocalizer().get( // JHf
                                  "attributes.Heads"),null,pHeads,null); // JHf
			
			// b) T-bars width
			JPanel pTbars = new JPanel(new GridLayout(3,1,5,5));
			pTbars.setBorder(BorderFactory.createEtchedBorder());
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowTBarWidth"));
			p.add(tbarWidthMinimumMmTF=new IncrementableTextField(0, 0.05, new JLabel("="), new JLabel("mm"), false));
			tbarWidthMinimumMmTF.setMinimum(0);
			tbarWidthMinimumMmTF.setLowBounded(true);
			p.add(new JLabel("+ "));
			p.add(createLabel("attributes.LineWidth"));
			p.add(tbarWidthLineWidthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("* "), null, false));
			tbarWidthLineWidthScaleTF.setMinimum(0);
			tbarWidthLineWidthScaleTF.setLowBounded(true);
			pTbars.add(p);
			// *) bracket length
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowBracketLength"));
			p.add(bracketLengthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("="), null, false));
			bracketLengthScaleTF.setMinimum(0);
			bracketLengthScaleTF.setLowBounded(true);
			p.add(new JLabel("* ("));
			p.add(createLabel("attributes.ArrowTBarWidth"));
			p.add(new JLabel(")"));
			pTbars.add(p);
			// *) R-bracket length
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowRoundedBracketLength"));
			p.add(roundedBracketLengthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("="), null, false));
			roundedBracketLengthScaleTF.setMinimum(0);
			roundedBracketLengthScaleTF.setLowBounded(true);
			p.add(new JLabel("* ("));
			p.add(createLabel("attributes.ArrowTBarWidth"));
			p.add(new JLabel(")"));
			pTbars.add(p);
			tabbedPane.addTab(jpicedt.Localizer.currentLocalizer().get( // JHf
                                  "attributes.BarsBrackets"),null,pTbars,null); // JHf
			
			// c) dot size (circles and disks)
			JPanel pDotSize = new JPanel(new GridLayout(3,1,5,5));
			pDotSize.setBorder(BorderFactory.createEtchedBorder());
			p= new JPanel(new FlowLayout(FlowLayout.LEFT));
			p.add(createLabel("attributes.ArrowDotSize"));
			p.add(dotSizeMinimumMmTF=new IncrementableTextField(0, 0.05, new JLabel("="), new JLabel("mm"), false));
			dotSizeMinimumMmTF.setMinimum(0);
			dotSizeMinimumMmTF.setLowBounded(true);
			p.add(new JLabel("+ "));
			p.add(createLabel("attributes.LineWidth"));
			p.add(dotSizeLineWidthScaleTF=new IncrementableTextField(0, 0.05, new JLabel("* "), null, false));
			dotSizeLineWidthScaleTF.setMinimum(0);
			dotSizeLineWidthScaleTF.setLowBounded(true);
			pDotSize.add(p);
			tabbedPane.addTab(jpicedt.Localizer.currentLocalizer().get( // JHf
                                  "attributes.CirclesDisks"),null,pDotSize,null); // JHf
			
			box.add(tabbedPane);
			return box;
		}

		private JLabel createLabel(String s){
			return new JLabel(jpicedt.Localizer.currentLocalizer().get(s));
		}

		/**
		 * update widgets values according to currently active attribute set.
		 */
		public void load(){
			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			leftArrowCB.removeActionListener(this);
			rightArrowCB.removeActionListener(this);
			globalScaleWidthTF.removeActionListener(this);
			globalScaleLengthTF.removeActionListener(this);
			headWidthMinimumMmTF.removeActionListener(this);
			headWidthLineWidthScaleTF.removeActionListener(this);
			lengthScaleTF.removeActionListener(this);
			insetScaleTF.removeActionListener(this);
			tbarWidthMinimumMmTF.removeActionListener(this);
			tbarWidthLineWidthScaleTF.removeActionListener(this);
			bracketLengthScaleTF.removeActionListener(this);
			roundedBracketLengthScaleTF.removeActionListener(this);
			dotSizeMinimumMmTF.removeActionListener(this);
			dotSizeLineWidthScaleTF.removeActionListener(this);

			leftArrowCB.setSelectedIndex(Arrow.getIndex((Arrow)(getAttribute(LEFT_ARROW))));
			rightArrowCB.setSelectedIndex(Arrow.getIndex((Arrow)(getAttribute(RIGHT_ARROW))));
			globalScaleWidthTF.setValue(((Double)getAttribute(ARROW_GLOBAL_SCALE_WIDTH)).doubleValue());
			globalScaleLengthTF.setValue(((Double)getAttribute(ARROW_GLOBAL_SCALE_LENGTH)).doubleValue());
			headWidthMinimumMmTF.setValue(((Double)getAttribute(ARROW_WIDTH_MINIMUM_MM)).doubleValue());
			headWidthLineWidthScaleTF.setValue(((Double)getAttribute(ARROW_WIDTH_LINEWIDTH_SCALE)).doubleValue());
			lengthScaleTF.setValue(((Double)getAttribute(ARROW_LENGTH_SCALE)).doubleValue());
			insetScaleTF.setValue(((Double)getAttribute(ARROW_INSET_SCALE)).doubleValue());
			tbarWidthMinimumMmTF.setValue(((Double)getAttribute(TBAR_WIDTH_MINIMUM_MM)).doubleValue());
			tbarWidthLineWidthScaleTF.setValue(((Double)getAttribute(TBAR_WIDTH_LINEWIDTH_SCALE)).doubleValue());
			bracketLengthScaleTF.setValue(((Double)getAttribute(BRACKET_LENGTH_SCALE)).doubleValue());
			roundedBracketLengthScaleTF.setValue(((Double)getAttribute(RBRACKET_LENGTH_SCALE)).doubleValue());
			dotSizeMinimumMmTF.setValue(((Double)getAttribute(POLYDOTS_SIZE_MINIMUM_MM)).doubleValue());
			dotSizeLineWidthScaleTF.setValue(((Double)getAttribute(POLYDOTS_SIZE_LINEWIDTH_SCALE)).doubleValue());

			leftArrowCB.addActionListener(this);
			rightArrowCB.addActionListener(this);
			globalScaleWidthTF.addActionListener(this);
			globalScaleLengthTF.addActionListener(this);
			headWidthMinimumMmTF.addActionListener(this);
			headWidthLineWidthScaleTF.addActionListener(this);
			lengthScaleTF.addActionListener(this);
			insetScaleTF.addActionListener(this);
			tbarWidthMinimumMmTF.addActionListener(this);
			tbarWidthLineWidthScaleTF.addActionListener(this);
			bracketLengthScaleTF.addActionListener(this);
			roundedBracketLengthScaleTF.addActionListener(this);
			dotSizeMinimumMmTF.addActionListener(this);
			dotSizeLineWidthScaleTF.addActionListener(this);
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){
		}

		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){
			if (e.getSource()==leftArrowCB) 
				setAttribute(LEFT_ARROW,Arrow.getArrow(leftArrowCB.getSelectedIndex()));
			else if (e.getSource()==rightArrowCB)
				setAttribute(RIGHT_ARROW,Arrow.getArrow(rightArrowCB.getSelectedIndex()));
			else if (e.getSource()==globalScaleWidthTF)
				setAttribute(ARROW_GLOBAL_SCALE_WIDTH, new Double(globalScaleWidthTF.getValue()));
			else if (e.getSource()==globalScaleLengthTF)
				setAttribute(ARROW_GLOBAL_SCALE_LENGTH, new Double(globalScaleLengthTF.getValue()));
			else if (e.getSource()==headWidthMinimumMmTF)
				setAttribute(ARROW_WIDTH_MINIMUM_MM, new Double(headWidthMinimumMmTF.getValue()));
			else if (e.getSource()==headWidthLineWidthScaleTF)
				setAttribute(ARROW_WIDTH_LINEWIDTH_SCALE, new Double(headWidthLineWidthScaleTF.getValue()));
			else if (e.getSource()==lengthScaleTF)
				setAttribute(ARROW_LENGTH_SCALE, new Double(lengthScaleTF.getValue()));
			else if (e.getSource()==insetScaleTF)
				setAttribute(ARROW_INSET_SCALE, new Double(insetScaleTF.getValue()));
			else if (e.getSource()==tbarWidthMinimumMmTF)
				setAttribute(TBAR_WIDTH_MINIMUM_MM, new Double(tbarWidthMinimumMmTF.getValue()));
			else if (e.getSource()==tbarWidthLineWidthScaleTF)
				setAttribute(TBAR_WIDTH_LINEWIDTH_SCALE, new Double(tbarWidthLineWidthScaleTF.getValue()));
			else if (e.getSource()==bracketLengthScaleTF)
				setAttribute(BRACKET_LENGTH_SCALE, new Double(bracketLengthScaleTF.getValue()));
			else if (e.getSource()==roundedBracketLengthScaleTF)
				setAttribute(RBRACKET_LENGTH_SCALE, new Double(roundedBracketLengthScaleTF.getValue()));
			else if (e.getSource()==dotSizeMinimumMmTF)
				setAttribute(POLYDOTS_SIZE_MINIMUM_MM, new Double(dotSizeMinimumMmTF.getValue()));
			else if (e.getSource()==dotSizeLineWidthScaleTF)
				setAttribute(POLYDOTS_SIZE_LINEWIDTH_SCALE, new Double(dotSizeLineWidthScaleTF.getValue()));
		}

	} // class


	// -------------------- Shadow ---------------------------- //

	/**
	 * A customizer dedicated to editing shadow attributes 
	 */
	public class ShadowPropertiesPanel extends AbstractCustomizer  implements ActionListener,ChangeListener {

		private IncrementableTextField shadowSizeTF, shadowAngleTF;
		private JToggleButton shadowB;
		private JColorChooser colorChooser;

		/**
		 * creates a customizer for shadow attributes editing
		 */
		public ShadowPropertiesPanel(){
			super();
			this.add(createShadowPanel(),BorderLayout.NORTH);
			// add listeners
			shadowB.addActionListener(this);
			shadowSizeTF.addActionListener(this);
			shadowAngleTF.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this);
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Shadow"); 
		}

		/**
		 * creates a sub-panel for "shadow" properties : shadow, shadow_size, shadow_angle,shadow_color
		 */
		private JComponent createShadowPanel(){

			Box box = new Box(BoxLayout.Y_AXIS);
			
			JPanel p = new JPanel(new GridLayout(1,3,5,5));
			// 1) shadow (true/false) :
			shadowB = new JToggleButton(PEToolKit.createImageIcon("attributes.Shadow"));
			p.add(shadowB);

			// 2) shadow size :
			shadowSizeTF = createIncrementableTextField(0.05,"attributes.ShadowSize","mm");
			shadowSizeTF.setMinimum(0);
			shadowSizeTF.setLowBounded(true);
			p.add(shadowSizeTF);

			// 3) shadow angle :
			shadowAngleTF = createIncrementableTextField(5,"attributes.ShadowAngle","deg");
			p.add(shadowAngleTF);
			box.add(p);
			
			// 4) shadow color:
			box.add(colorChooser = createColorChooser());
			
			return box;
		}

		/**
		 * update widgets values according to currently active attribute set (ie either an appropriate
		 * input attribute, or the attribute set of a selected element).
		 */
		public void load(){

			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			colorChooser.getSelectionModel().removeChangeListener(this); 
			shadowB.removeActionListener(this);
			shadowSizeTF.removeActionListener(this);
			shadowAngleTF.removeActionListener(this);
			
			shadowB.setSelected(((Boolean)getAttribute(SHADOW)).booleanValue());
			shadowSizeTF.setValue(((Double)getAttribute(SHADOW_SIZE)).doubleValue());
			shadowAngleTF.setValue(((Double)getAttribute(SHADOW_ANGLE)).doubleValue());
			colorChooser.setColor((Color)getAttribute(SHADOW_COLOR));
			shadowSizeTF.setEnabled(shadowB.isSelected());
			shadowAngleTF.setEnabled(shadowB.isSelected());
			
			shadowB.addActionListener(this);
			shadowSizeTF.addActionListener(this);
			shadowAngleTF.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); 
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){ // done in actionPerformed
		}

		/**
		 * Called by the colorchooser when the user picks a new colour, and this customizer lies currently
		 * on top.
		 */
		public void stateChanged(ChangeEvent e){
			Color newColor = colorChooser.getColor();
			setAttribute(SHADOW_COLOR,newColor);
		}

		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){

			if (e.getSource() == shadowB) setAttribute(SHADOW, new Boolean(shadowB.isSelected()));
			else if (e.getSource() == shadowSizeTF) setAttribute(SHADOW_SIZE, new Double(shadowSizeTF.getValue()));
			else if (e.getSource() == shadowAngleTF) setAttribute(SHADOW_ANGLE, new Double(shadowAngleTF.getValue()));
			shadowSizeTF.setEnabled(shadowB.isSelected());
			shadowAngleTF.setEnabled(shadowB.isSelected());
		}

	} // class




	// -------------------- Stroke ---------------------------- //

	/**
	 * A panel for editing Element's stroke attributes (width, colour, dashing, ...) 
	 */
	public class StrokePropertiesPanel extends AbstractCustomizer  implements ActionListener, ChangeListener {

	    private IncrementableTextField lineWidthTF, dashTF, dashOpaqueRatioTF, dotSepTF, overStrikeWidthTF;
	    private JComboBox lineStyleCB;
	    private JCheckBox overStrikeCB;
		private JColorChooser colorChooser;

		/**
		 * creates a JPanel for stroke attributes editing
		 */
		public StrokePropertiesPanel(){
			super();
			this.add(createStrokePanel(),BorderLayout.NORTH);
			// add listeners
			lineStyleCB.addActionListener(this);
			lineWidthTF.addActionListener(this);
			dashTF.addActionListener(this);
			dashOpaqueRatioTF.addActionListener(this);
			dotSepTF.addActionListener(this);
			overStrikeCB.addActionListener(this);
			overStrikeWidthTF.addActionListener(this);
			colorChooser.getSelectionModel().addChangeListener(this); // fire change event when user picks a colour
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Stroke");
		}

		/**
		 * creates sub-panel for stroke properties
		 */
		private JComponent createStrokePanel(){

			Box box = new Box(BoxLayout.Y_AXIS);
			JPanel p = new JPanel(new GridLayout(4,2,5,5));
			
			// 1)lineStyle
			ImageIcon[] iconArray = new ImageIcon[4];
			iconArray[0] = PEToolKit.createImageIcon("attributes.LineStyleNone");
			iconArray[1] = PEToolKit.createImageIcon("attributes.LineStyleSolid");
			iconArray[2] = PEToolKit.createImageIcon("attributes.LineStyleDashed");
			iconArray[3] = PEToolKit.createImageIcon("attributes.LineStyleDotted");
			lineStyleCB = createComboBox(iconArray);
			//String lineStyleConstants[] = {NONE,SOLID, DASHED, DOTTED};lineStyleCB = new JComboBox(lineStyleConstants);
			p.add(lineStyleCB);

			// 2) lineWidth:
			lineWidthTF = createIncrementableTextField(0.05,"attributes.LineWidth"," mm ");
			lineWidthTF.setMinimum(0);
			lineWidthTF.setLowBounded(true);
			p.add(lineWidthTF);

			// 3) dash length :
			dashTF = createIncrementableTextField(1,"attributes.LineDash"," mm ");
			dashTF.setMinimum(0);
			dashTF.setLowBounded(true);
			p.add(dashTF);
			
			// 4) dash opaque ratio (100% = opaque only, 0% = transparent only)
			dashOpaqueRatioTF = createIncrementableTextField(10,"attributes.LineDashOpaque"," % ");
			dashOpaqueRatioTF.setMinimum(0);
			dashOpaqueRatioTF.setLowBounded(true);
			dashOpaqueRatioTF.setMaximum(100);
			dashOpaqueRatioTF.setHighBounded(true);
			p.add(dashOpaqueRatioTF);

			// 5) dot sep :
			dotSepTF = createIncrementableTextField(1,"attributes.LineDotSep"," mm ");
			dotSepTF.setMinimum(0);
			dotSepTF.setLowBounded(true);
			dotSepTF.setAlignmentY(CENTER_ALIGNMENT);
			p.add(dotSepTF);

			// 6) over strike :
			p.add(new JLabel());
			overStrikeCB = new JCheckBox(jpicedt.Localizer.currentLocalizer().get("attributes.LineOverstrike"));
			overStrikeCB.setSelected(false);
			p.add(overStrikeCB);
			overStrikeWidthTF = createIncrementableTextField(0.05,"attributes.LineOverstrikeWidth"," mm ");
			overStrikeWidthTF.setMinimum(0);
			overStrikeWidthTF.setLowBounded(true);
			overStrikeWidthTF.setAlignmentY(CENTER_ALIGNMENT);
			p.add(overStrikeWidthTF);

			box.add(p);
			box.add(colorChooser = createColorChooser());
			return box;

		}

		/**
		 * update widgets values according to currently active attribute set.
		 */
		public void load(){

			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			lineStyleCB.removeActionListener(this);
			lineWidthTF.removeActionListener(this);
			dashTF.removeActionListener(this);
			dashOpaqueRatioTF.removeActionListener(this);
			dotSepTF.removeActionListener(this);
			overStrikeWidthTF.removeActionListener(this);
			overStrikeCB.removeActionListener(this);
			colorChooser.getSelectionModel().removeChangeListener(this); 
			
			lineWidthTF.setValue(((Double)getAttribute(LINE_WIDTH)).doubleValue());
			double dashOpaque = ((Double)getAttribute(DASH_OPAQUE)).doubleValue();
			double dashTransparent = ((Double)getAttribute(DASH_TRANSPARENT)).doubleValue();
			dashTF.setValue(dashOpaque + dashTransparent);
			dotSepTF.setValue(((Double)getAttribute(DOT_SEP)).doubleValue());
			double opaqueRatio = dashOpaque/(dashTransparent+dashOpaque)*100.0;
			dashOpaqueRatioTF.setValue(opaqueRatio);
			overStrikeCB.setSelected(((Boolean)getAttribute(OVER_STRIKE)).booleanValue());
			overStrikeWidthTF.setValue(((Double)getAttribute(OVER_STRIKE_WIDTH)).doubleValue());

			if (getAttribute(LINE_STYLE) == NONE) {
				lineStyleCB.setSelectedIndex(0);
				dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(false);
				lineWidthTF.setEnabled(false);
			}
			else if (getAttribute(LINE_STYLE) == SOLID) {
				lineStyleCB.setSelectedIndex(1);
				dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(false);
				lineWidthTF.setEnabled(true);
			}
			else if (getAttribute(LINE_STYLE) == DASHED) {
				lineStyleCB.setSelectedIndex(2);
				dashTF.setEnabled(true); dashOpaqueRatioTF.setEnabled(true); dotSepTF.setEnabled(false);
				lineWidthTF.setEnabled(true);
			}
			else if (getAttribute(LINE_STYLE) == DOTTED) {
				lineStyleCB.setSelectedIndex(3);
				dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(true);
				lineWidthTF.setEnabled(true);
			}

			colorChooser.setColor((Color)getAttribute(LINE_COLOR));
			
			colorChooser.getSelectionModel().addChangeListener(this); 
			lineStyleCB.addActionListener(this);
			lineWidthTF.addActionListener(this);
			dashTF.addActionListener(this);
			dashOpaqueRatioTF.addActionListener(this);
			dotSepTF.addActionListener(this);
			overStrikeWidthTF.addActionListener(this);
			overStrikeCB.addActionListener(this);

		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){ // done in actionPerformed
		}

		/**
		 * Called by the colorchooser when the user picks a new colour, and this customizer lies currently
		 * on top.
		 */
		public void stateChanged(ChangeEvent e){
			Color newColor = colorChooser.getColor();
			setAttribute(LINE_COLOR,newColor);
		}
		
		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){

			if (e.getSource() == lineStyleCB){
				switch (lineStyleCB.getSelectedIndex()){
				case 0 :
					setAttribute(LINE_STYLE, NONE); 
					dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(false);
					lineWidthTF.setEnabled(false);
					break;
				case 1 :
					setAttribute(LINE_STYLE, SOLID);
					dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(false);
					lineWidthTF.setEnabled(true);
					break;
				case 2 :
					setAttribute(LINE_STYLE, DASHED); 
					dashTF.setEnabled(true); dashOpaqueRatioTF.setEnabled(true); dotSepTF.setEnabled(false);
					lineWidthTF.setEnabled(true);
					break;
				case 3:
					setAttribute(LINE_STYLE, DOTTED); 
					dashTF.setEnabled(false); dashOpaqueRatioTF.setEnabled(false); dotSepTF.setEnabled(true);
					lineWidthTF.setEnabled(true);
					break;
				default:
				}
			}
			else if (e.getSource() == lineWidthTF){
				setAttribute(LINE_WIDTH, new Double(lineWidthTF.getValue()));
			}
			else if (e.getSource() == dashTF || e.getSource()==dashOpaqueRatioTF){
				double totalDash = dashTF.getValue();
				double opaqueRatio = dashOpaqueRatioTF.getValue()/100.0;
				setAttribute(DASH_OPAQUE, new Double(totalDash * opaqueRatio));
				setAttribute(DASH_TRANSPARENT, new Double(totalDash * (1.0-opaqueRatio)));
			}
			else if (e.getSource() == dotSepTF){
				setAttribute(DOT_SEP, new Double(dotSepTF.getValue()));
			}
			else if (e.getSource() == overStrikeCB){
				setAttribute(OVER_STRIKE, new Boolean(overStrikeCB.isSelected()));
			}
			else if (e.getSource() == overStrikeWidthTF){
				setAttribute(OVER_STRIKE_WIDTH, new Double(overStrikeWidthTF.getValue()));
			}

			// else color chooser
		}

	} // class



	// -------------------- PsDots Attributes --------------------------------- //
	/**
	 * A panel for editing AbstractCurve's polydots attributes
	 */
	public class PolydotsPropertiesPanel extends AbstractCustomizer implements ActionListener {

		private JCheckBox isSuperimposedDotsCB, isDottedCB;
		private IncrementableTextField angleDNF, scaleHorDNF, scaleVertDNF, sizePrefactorDNF, sizeRelativeDNF;
		private JComboBox dotStyleCombo; 
		//private boolean isListenersAdded = false;// flag set to true after listener have been registered
		private String[] availableDotStyles = new String[]{
				 POLYDOTS_CIRCLE,
				 POLYDOTS_DISK,
				 POLYDOTS_PLUS, 
				 POLYDOTS_TRIANGLE, 
				 POLYDOTS_TRIANGLE_FILLED, 
				 POLYDOTS_SQUARE, 
				 POLYDOTS_SQUARE_FILLED, 
				 POLYDOTS_PENTAGON, 
				 POLYDOTS_PENTAGON_FILLED};


		/**
		 * Initializes the GUI. 		 
		 */
		public PolydotsPropertiesPanel() {
			super();
			this.add(createPolydotsPanel(), BorderLayout.NORTH);
			isDottedCB.addActionListener(this);
			isSuperimposedDotsCB.addActionListener(this);
			dotStyleCombo.addActionListener(this);
			angleDNF.addActionListener(this);
			sizePrefactorDNF.addActionListener(this);
			sizeRelativeDNF.addActionListener(this);
			scaleHorDNF.addActionListener(this);
			scaleVertDNF.addActionListener(this);
		}
		
		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.PolydotsAttributes");
		}
		
		/**
		 * Create a sub-panel for polydots attributes
		 */
		private JPanel createPolydotsPanel(){
			JPanel p = new JPanel(new GridLayout(5,2,5,5));
			p.add(isDottedCB = new JCheckBox(jpicedt.Localizer.currentLocalizer().get("attributes.PolyDots")));// [pending] i18n
			p.add(isSuperimposedDotsCB = new JCheckBox(jpicedt.Localizer.currentLocalizer().get("attributes.PolydotsSuperimpose")));// [pending] i18n
			// create ImageIcon array for JComboBox :
			ImageIcon[] iconArray = new ImageIcon[9];
			iconArray[0] = PEToolKit.createImageIcon("attributes.PolydotsCircle");
			iconArray[1] = PEToolKit.createImageIcon("attributes.PolydotsDisk");
			iconArray[2] = PEToolKit.createImageIcon("attributes.PolydotsPlus");
			iconArray[3] = PEToolKit.createImageIcon("attributes.PolydotsTriangle");
			iconArray[4] = PEToolKit.createImageIcon("attributes.PolydotsTriangleFilled");
			iconArray[5] = PEToolKit.createImageIcon("attributes.PolydotsSquare");
			iconArray[6] = PEToolKit.createImageIcon("attributes.PolydotsSquareFilled");
			iconArray[7] = PEToolKit.createImageIcon("attributes.PolydotsPentagon");
			iconArray[8] = PEToolKit.createImageIcon("attributes.PolydotsPentagonFilled");
			dotStyleCombo = createComboBox(iconArray);
			p.add(dotStyleCombo);
			p.add(angleDNF=new IncrementableTextField(0, 5, createLabel("attributes.PolydotsAngle"), new JLabel("deg"), false));
			p.add(sizePrefactorDNF=new IncrementableTextField(1, 1, createLabel("attributes.PolydotsSizePrefactor"), null, false));
			p.add(sizeRelativeDNF=new IncrementableTextField(0, .1, createLabel("attributes.PolydotsSizeRelative"), new JLabel("mm"), false));
			p.add(scaleHorDNF=new IncrementableTextField(1, 1, createLabel("attributes.PolydotsScaleHor"), null, false));
			p.add(scaleVertDNF=new IncrementableTextField(1, 1, createLabel("attributes.PolydotsScaleVert"), null, false));
			return p;
		}
		
		private JLabel createLabel(String s){
			return new JLabel(jpicedt.Localizer.currentLocalizer().get(s));
		}


		/**
		 * load widgets with object's properties
		 */
		public void load() {
			// first remove listeners to avoid deadlocks (through recursion) when setting widgets values ;-)
			isDottedCB.removeActionListener(this);
			isSuperimposedDotsCB.removeActionListener(this);
			dotStyleCombo.removeActionListener(this);
			angleDNF.removeActionListener(this);
			sizePrefactorDNF.removeActionListener(this);
			sizeRelativeDNF.removeActionListener(this);
			scaleHorDNF.removeActionListener(this);
			scaleVertDNF.removeActionListener(this);
			
			// then update widgets
			String dotStyle = (String)getAttribute(POLYDOTS_STYLE);
			boolean isDotAttribute = (!dotStyle.equals(NONE));
			
			isDottedCB.setSelected(isDotAttribute);
			
			isSuperimposedDotsCB.setEnabled(isDotAttribute);
			isSuperimposedDotsCB.setSelected(getAttribute(POLYDOTS_SUPERIMPOSE).equals(Boolean.TRUE));
			
			angleDNF.setEnabled(isDotAttribute);
			angleDNF.setValue((Double)getAttribute(POLYDOTS_ANGLE));
			sizePrefactorDNF.setEnabled(isDotAttribute);
			sizePrefactorDNF.setValue((Double)getAttribute(POLYDOTS_SIZE_LINEWIDTH_SCALE));
			sizePrefactorDNF.setBoundValues(0,1000);
			sizePrefactorDNF.setLowBounded(true);
			sizeRelativeDNF.setEnabled(isDotAttribute);
			sizeRelativeDNF.setValue((Double)getAttribute(POLYDOTS_SIZE_MINIMUM_MM));
			scaleHorDNF.setEnabled(isDotAttribute);
			scaleHorDNF.setValue((Double)getAttribute(POLYDOTS_SCALE_H));
			scaleHorDNF.setBoundValues(0,1000);
			scaleHorDNF.setLowBounded(true);
			scaleVertDNF.setEnabled(isDotAttribute);
			scaleVertDNF.setValue((Double)getAttribute(POLYDOTS_SCALE_V));
			scaleVertDNF.setBoundValues(0,1000);
			scaleVertDNF.setLowBounded(true);
			
			dotStyleCombo.setEnabled(isDotAttribute);
			if (dotStyle.equals(NONE)) dotStyleCombo.setSelectedIndex(0);
			else {
				for (int i=0; i<availableDotStyles.length;i++){
					if (dotStyle.equals(availableDotStyles[i])) {
						dotStyleCombo.setSelectedIndex(i);
						break;
					}
				}
			}
			
			// finally re-register listeners :
			isDottedCB.addActionListener(this);
			isSuperimposedDotsCB.addActionListener(this);
			dotStyleCombo.addActionListener(this);
			angleDNF.addActionListener(this);
			sizePrefactorDNF.addActionListener(this);
			sizeRelativeDNF.addActionListener(this);
			scaleHorDNF.addActionListener(this);
			scaleVertDNF.addActionListener(this);
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store() {
		}
		
		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){
		
			if (e.getSource()==isDottedCB){
				if (isDottedCB.isSelected()) setAttribute(POLYDOTS_STYLE, availableDotStyles[dotStyleCombo.getSelectedIndex()]);
				else setAttribute(POLYDOTS_STYLE, NONE);
				isSuperimposedDotsCB.setEnabled(isDottedCB.isSelected());
				angleDNF.setEnabled(isDottedCB.isSelected());
				sizePrefactorDNF.setEnabled(isDottedCB.isSelected());
				sizeRelativeDNF.setEnabled(isDottedCB.isSelected());
				scaleHorDNF.setEnabled(isDottedCB.isSelected());
				scaleVertDNF.setEnabled(isDottedCB.isSelected());
				dotStyleCombo.setEnabled(isDottedCB.isSelected());
			}
			else if (e.getSource()==dotStyleCombo){
				if (isDottedCB.isSelected()) setAttribute(POLYDOTS_STYLE, availableDotStyles[dotStyleCombo.getSelectedIndex()]);
			}
			else if (e.getSource()==isSuperimposedDotsCB){
				setAttribute(POLYDOTS_SUPERIMPOSE, new Boolean(isSuperimposedDotsCB.isSelected()));
			}
			else if (e.getSource()==angleDNF){
				setAttribute(POLYDOTS_ANGLE, new Double(angleDNF.getValue()));
			}
			else if (e.getSource()==sizePrefactorDNF){
				setAttribute(POLYDOTS_SIZE_LINEWIDTH_SCALE, new Double(sizePrefactorDNF.getValue()));
			}
			else if (e.getSource()==sizeRelativeDNF){
				setAttribute(POLYDOTS_SIZE_MINIMUM_MM, new Double(sizeRelativeDNF.getValue()));
			}
			else if (e.getSource()==scaleHorDNF){
				setAttribute(POLYDOTS_SCALE_H, new Double(scaleHorDNF.getValue()));
			}
			else if (e.getSource()==scaleVertDNF){
				setAttribute(POLYDOTS_SCALE_V, new Double(scaleVertDNF.getValue()));
			}
		}
	}

	// -------------------- Text Attributes ----------------------------------- //
	
	/**
	 * A panel for editing text attributes (alignment + frame)
	 */
	public class TextPropertiesPanel extends AbstractCustomizer implements ActionListener {

		private String[] availableVertAlignValues = new String[]{
			TEXT_VALIGN_BOTTOM,
			TEXT_VALIGN_BASELINE,
			TEXT_VALIGN_CENTER,
			TEXT_VALIGN_TOP};
	
		private ImageIcon[] vertAlignementIcons = {
			PEToolKit.createImageIcon("attributes.TextValignBottom"), 
			PEToolKit.createImageIcon("attributes.TextValignBaseline"),
		    PEToolKit.createImageIcon("attributes.TextValignMiddle"),
		    PEToolKit.createImageIcon("attributes.TextValignTop")};

		private String[] availableHorAlignValues = new String[]{
			TEXT_HALIGN_LEFT,
			TEXT_HALIGN_CENTER,
			TEXT_HALIGN_RIGHT};

		private ImageIcon[] horAlignementIcons = {
			PEToolKit.createImageIcon("attributes.TextHalignLeft"), 
		    PEToolKit.createImageIcon("attributes.TextHalignCenter"),
			PEToolKit.createImageIcon("attributes.TextHalignRight")};

		private  String[] availableBoxValues = new String[]{
			TEXT_BOX_NO_FRAME,
			TEXT_BOX_RECTANGLE,
			TEXT_BOX_CIRCLE,
			TEXT_BOX_OVAL};
 
		private ImageIcon[] boxTypeIcons = {
		    PEToolKit.createImageIcon("attributes.TextBoxNone"), 
		    PEToolKit.createImageIcon("attributes.TextBoxFrame"),
			PEToolKit.createImageIcon("attributes.TextBoxCircle"),
			PEToolKit.createImageIcon("attributes.TextBoxOval")};

		/* widgets */
		private JComboBox horAlignementCombo,vertAlignementCombo, boxTypeCombo;
		private IncrementableTextField angleTF; // text rotation

		/** GUI init */
		public TextPropertiesPanel(){

			super();
			this.add(createTextPanel(), BorderLayout.NORTH);
			horAlignementCombo.addActionListener(this);
			vertAlignementCombo.addActionListener(this);
			boxTypeCombo.addActionListener(this);
			angleTF.addActionListener(this);
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("attributes.Text");
		}

		/**
		 * Create a sub-panel for text attributes
		 */
		private JPanel createTextPanel(){
			JPanel p = new JPanel(new GridLayout(2,2,5,5));

			horAlignementCombo = PEToolKit.createComboBox(horAlignementIcons);
			p.add(horAlignementCombo);
			
			vertAlignementCombo = PEToolKit.createComboBox(vertAlignementIcons);
			p.add(vertAlignementCombo);
			
			boxTypeCombo = PEToolKit.createComboBox(boxTypeIcons);
			p.add(boxTypeCombo);
			
			angleTF = createIncrementableTextField(5,"attributes.TextRotation","deg");
			p.add(angleTF);
			
			return p;
		}
			
		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){

			horAlignementCombo.removeActionListener(this);
			vertAlignementCombo.removeActionListener(this);
			boxTypeCombo.removeActionListener(this);
			angleTF.removeActionListener(this);
			
			angleTF.setValue(((Double)getAttribute(TEXT_ROTATION)).doubleValue());
			
			String value = (String)getAttribute(TEXT_VERT_ALIGN);
			for (int i=0; i<availableVertAlignValues.length;i++){
				if (value.equals(availableVertAlignValues[i])) {
					vertAlignementCombo.setSelectedIndex(i);
					break;
				}
			}

			value = (String)getAttribute(TEXT_HOR_ALIGN);
			for (int i=0; i<availableHorAlignValues.length;i++){
				if (value.equals(availableHorAlignValues[i])) {
					horAlignementCombo.setSelectedIndex(i);
					break;
				}
			}

			value = (String)getAttribute(TEXT_FRAME);
			for (int i=0; i<availableBoxValues.length;i++){
				if (value.equals(availableBoxValues[i])) {
					boxTypeCombo.setSelectedIndex(i);
					break;
				}
			}
			
			horAlignementCombo.addActionListener(this);
			vertAlignementCombo.addActionListener(this);
			boxTypeCombo.addActionListener(this);
			angleTF.addActionListener(this);
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){
		}

		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){
			if (e.getSource()==horAlignementCombo){
				setAttribute(TEXT_HOR_ALIGN, availableHorAlignValues[horAlignementCombo.getSelectedIndex()]);
			}
			else if (e.getSource()==vertAlignementCombo){
				setAttribute(TEXT_VERT_ALIGN, availableVertAlignValues[vertAlignementCombo.getSelectedIndex()]);
			}
			else if (e.getSource()==boxTypeCombo){
				setAttribute(TEXT_FRAME, availableBoxValues[boxTypeCombo.getSelectedIndex()]);
			}
			else if (e.getSource() == angleTF) setAttribute(TEXT_ROTATION, new Double(angleTF.getValue()));
		}

	}
		
	// -------------------- Pst Custom Attributes  ---------------------------- //

	/**
	 * A panel for adding custom attributes (pstricks only)
	 * [todo] disable panel when not in Pstricks content-type
	 */
	public class PstCustomPropertiesPanel extends AbstractCustomizer  implements ActionListener {

	    private JTextArea pstCustomTA;
	    private JScrollPane pstCustomSP;
	    private JButton applyButton;



		/**
		 * creates a JPanel for attributes editing
		 */
		public PstCustomPropertiesPanel(){
			super();
			this.add(createPstCustomPanel(),BorderLayout.NORTH);
			// add listeners
			applyButton.addActionListener(this);
		}

		/**
		 * @return the customizer title , used e.g. for Border or Tabpane title.
		 */
		public String getTitle(){
			return "PSTricks";
		}

		/**
		 * creates sub-panel for stroke properties
		 */
		private JComponent createPstCustomPanel(){

			JPanel p = new JPanel(new BorderLayout(5,5));
			p.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
		            Localizer.currentLocalizer().get("attributes.PstCustomParameters")));
			p.setToolTipText(Localizer.currentLocalizer().get("attributes.PstCustomParameters.tooltip"));
			//gbl.setConstraints(l,c);
			//p.add(l);
			pstCustomTA = new JTextArea(10,20);
			pstCustomTA.setEditable(true);
			pstCustomTA.setLineWrap(true);
			pstCustomTA.setWrapStyleWord(true);
			pstCustomTA.setToolTipText(Localizer.currentLocalizer().get("attributes.PstCustomParameters.tooltip"));
			pstCustomSP = new JScrollPane(pstCustomTA);
			pstCustomSP.setToolTipText(Localizer.currentLocalizer().get("attributes.PstCustomParameters.tooltip"));
			p.add(pstCustomSP, BorderLayout.NORTH);

			applyButton=new JButton(Localizer.currentLocalizer().get("button.OK"));
			p.add(applyButton, BorderLayout.SOUTH);
			return p;

		}

		/**
		 * update widgets values according to currently active attribute set.
		 */
		public void load(){

			applyButton.removeActionListener(this); // [SR:???]
			pstCustomTA.setText(((String)getAttribute(PST_CUSTOM)));
			applyButton.addActionListener(this);
		}

		/**
		 * does nothing (handled by actionPerformed)
		 */
		public void store(){ // done in actionPerformed
		}

		/**
		 * synchronize the currently active attribute set with the widget's state.
		 */
		public void actionPerformed(ActionEvent e){

			if (e.getSource() == applyButton)
			    setAttribute(PST_CUSTOM,pstCustomTA.getText());
			// else color chooser
		}

	} // class



	////////////////////////////////////////////////////////////////
	//// COLOURS
	////////////////////////////////////////////////////////////////
	
	private static final Color[] PREDEFINED_COLORS = {
	            Color.black,
	            Color.darkGray,
	            Color.gray,
	            Color.lightGray,
	            Color.white,
	            Color.red,
	            Color.green,
	            Color.blue,
	            Color.cyan,
	            Color.magenta,
	            Color.yellow
	};
	// for tooltip text : (to be i18n through jpicedt.Localizer)
	private static final String[] PREDEFINED_COLORS_STRINGS = {
	            "colour.black",
	            "colour.dark_grey",
	            "colour.grey",
	            "colour.light_grey",
	            "colour.white",
	            "colour.red",
	            "colour.green",
	            "colour.blue",
	            "colour.cyan",
	            "colour.magenta",
	            "colour.yellow"
	};
	private static final ImageIcon[] PREDEFINED_COLORS_ICONS = createColorIcons();

	/**
	 * @return the index of the given colour in array "PREDEFINED_COLORS", which holds the various colours
	 * displayed by the colour comboBoxes (either fill- or stroke- colour)
	 */
	private int getColorIndex(Color colour){
		for(int i=0; i<PREDEFINED_COLORS.length; i++){
			if (PREDEFINED_COLORS[i] == colour) return i;
		}
		return PREDEFINED_COLORS.length; // = user-defined colour
	}

	/**
	 * @return the colour associated to the given index
	 * return null if index is out of range (it's the responsibility of the caller to handle this case)
	 */
	private Color getColor(int index){
		if( index > PREDEFINED_COLORS.length-1) return null;
		else return PREDEFINED_COLORS[index];
	}

	/**
	 * build icons from PREDEFINED_COLORS for color lists or buttons
	 */
	private static ImageIcon[] createColorIcons(){

		BufferedImage im;
		ImageIcon[] colorIcons = new ImageIcon[PREDEFINED_COLORS.length];
		for (int i=0; i<PREDEFINED_COLORS.length; i++){
			im = new BufferedImage(16,16,BufferedImage.TYPE_INT_RGB);
			Graphics2D g = im.createGraphics();
			g.setPaint(Color.white);
			g.fill(new Rectangle(1,1,15,15));
			g.setStroke(new BasicStroke(2.0f));
			g.setPaint(PREDEFINED_COLORS[i]);
			g.fill(new Rectangle(3,3,11,11));
			colorIcons[i] = new ImageIcon(im);
		}
		return colorIcons;
	}
	
	/**
	 * custom color chooser for predefined colours
	 */
	private static class PredefinedColorChooser extends AbstractColorChooserPanel implements ActionListener {
		
		JToggleButton[] buttons = new JToggleButton[PREDEFINED_COLORS.length];
		
		/** create the GUI */
		protected void buildChooser(){
			setLayout(new GridLayout(4,4,5,5));
			ButtonGroup bg = new ButtonGroup();
			for (int i=0; i<buttons.length; i++){
				buttons[i] = new JToggleButton(PREDEFINED_COLORS_ICONS[i]);
				buttons[i].addActionListener(this);
				buttons[i].setToolTipText(jpicedt.Localizer.currentLocalizer().get(PREDEFINED_COLORS_STRINGS[i]));
				bg.add(buttons[i]);
				add(buttons[i]);
			}
		}
		/** called when the chooser panel is displayed */
		public void updateChooser(){
	        Color color = getColorFromModel();
			for (int i=0; i<buttons.length; i++){
				if (PREDEFINED_COLORS[i].equals(color)) {
					buttons[i].setSelected(true);
					return;
				}
			}
		}
        public void actionPerformed(ActionEvent e) {
            JToggleButton source = (JToggleButton)e.getSource();
			for (int i=0; i<buttons.length; i++){
				if (source == buttons[i]){
	            	getColorSelectionModel().setSelectedColor(PREDEFINED_COLORS[i]);
					return;
				}
			}
        }
		
		/** @return the pane title */
		public String getDisplayName(){
			return jpicedt.Localizer.currentLocalizer().get("colour.predefined");
		}
		public Icon getSmallDisplayIcon(){
			return null;
		}
		public Icon getLargeDisplayIcon(){
			return null;
		}
	}
	

	//////////////////////////////////////////////////////////////
	//// GUI Utilities
	//////////////////////////////////////////////////////////////	
	
	/** used by createComboBox */
	private static final Dimension dimButton = (new JLabel("dummy")).getPreferredSize();
	
	/** creates a JComboBox */
	private JComboBox createComboBox(Object[] items){
		JComboBox cb = PEToolKit.createComboBox(items);
		cb.setMaximumSize(new Dimension(dimButton.width * 3,dimButton.height));
		return cb;
	}

	/** 
	 * creates an IncrementableTextField :<br> 
	 * - icon is built from  "/jpicedt/images/"+icon+".gif"<br>
	 * - initial value is 0<br>
	 * - border are never drawn.<br>
	 * - other properties must be set apart.<br>
	 */
	private IncrementableTextField createIncrementableTextField(double initVal,
									double increment, String icon, String postFix){ //, int width){
		IncrementableTextField tf = new IncrementableTextField(
			initVal, // init value
			increment, // increment
			PEToolKit.createImageIcon(icon),
			postFix, // postFix
			false); // draw border
		//tf.setMaximumSize(new Dimension(dimButton.width * width,(int)(2*dimButton.height)));
		tf.setAlignmentY(JPanel.CENTER_ALIGNMENT);
		return tf;
	}
	
	/** 
	 * creates an IncrementableTextField :<br> 
	 * - icon is built from  "/jpicedt/images/"+icon+".gif"<br>
	 * - initial value is 0<br>
	 * - border are never drawn.<br>
	 * - other properties must be set apart.<br>
	 */
	private IncrementableTextField createIncrementableTextField(
									double increment, String icon, String postFix){ //, int width){
		return createIncrementableTextField(0,increment,icon,postFix);
	}
	
	private JColorChooser createColorChooser(){
		JColorChooser colorChooser = new JColorChooser();
		AbstractColorChooserPanel[] panels = colorChooser.getChooserPanels();
		//for (int i=0; i<panels.length; i++) colorChooser.removeChooserPanel(panels[i]);
		colorChooser.addChooserPanel(new PredefinedColorChooser());
		//for (int i=0; i<panels.length; i++) colorChooser.addChooserPanel(panels[i]);
		colorChooser.setPreviewPanel(new JPanel()); // remove preview panel
		return colorChooser;
		/*
		viewColorChooserCB.setAlignmentX(LEFT_ALIGNMENT);
		box.add(viewColorChooserCB);
		viewColorChooserCB.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent evt){
				if (viewColorChooserCB.isSelected()) box.add(colorChooser);
				else box.remove(colorChooser);
				invalidate();
				getRootPane().revalidate();
			}
		});
		*/
	}
		

} // class


