/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/


package jpicedt.ui.dialog;

import jpicedt.ui.*;
import jpicedt.Localizer;
import jpicedt.graphic.ContentType;

import javax.swing.*;
import java.io.*;
import java.util.*;

/**
 * A factory than produces and caches frequently accessed JFileChooser's.<p>
 * JFileChooser's get instanciated on first call, 
 * and are then cached for better efficiency. As a side-effect, the last accessed directory is
 * kept in memory, unless a call to <code>configure(Properties)</code> is made, which reinit the
 * file chooser's.
 */
public class PEFileChooser extends JFileChooser { // [pending] change name to FileChooserFactory and move to jpicedt.ui.util

	/** properties's key for open-file-directory */
	public static final String KEY_FILE_DIRECTORY = "directory.files";
	/** file chooser type for saving files */
	public static final int SAVEFILE = 0; // [pending] change to SAVE_FILE
	/** file chooser type for saving fragments */
	public static final int SAVEFRAGMENT = 1;
	/** file chooser type for opening files */
	public static final int OPENFILE = 2;
	/** file chooser type for opening fragments */
	public static final int OPENFRAGMENT = 3;
	/** file chooser type for browsing the save/open file directory */
	public static final int BROWSEFILEDIR = 4;
	/** the user's home directory */
    public static final String DEFAULT_DIRECTORY = System.getProperty("user.home");

	/////////////////////////////////////////////////////////////////////////////////
	
	private static String fileDirectory = DEFAULT_DIRECTORY; 
    private static String fragmentDirectory = jpicedt.JPicEdt.getUserSettingsDirectory()+File.separator+"fragments";
	private static PEFileChooser saveFile,saveFragment,openFile,openFragment; // cache for frequently accessed choosers
	private PEFileFilter[] filters; // one for each available content-type (see MiscUtilities.getAvailableContentTypes)
	
	/////////////////////////////////////////////////////////////////////////////////

	/**
	 * Opens (and possibly inits on first call) a JFileChooser, the type of which is specified by one of 
	 * the hereabove constants. If this is the first time a particular file chooser is accessed,
	 * it's created using, either the default settings, or the settings set by the last call to
	 * <code>configure(Properties)</code>.
	 * @param contentType allows to set the current file filter
	 * @return the selected file ; null if a cancel action occured at any time.
	 */
	public static String getFileName(int chooserType, String contentTypeClassName){

		switch (chooserType) {

		case SAVEFILE:
			if (saveFile == null)
				saveFile = new PEFileChooser(fileDirectory, Localizer.currentLocalizer().get("action.ui.SaveAs"));
			else saveFile.rescanCurrentDirectory();
			saveFile.setFileFilter(contentTypeClassName);
			return saveFile.saveToFileChooser(saveFile);

		case SAVEFRAGMENT :
			if (saveFragment == null)
				saveFragment = new PEFileChooser(fragmentDirectory, Localizer.currentLocalizer().get("action.ui.SaveSelectionAs"));
			else saveFragment.rescanCurrentDirectory();
			saveFragment.setFileFilter(contentTypeClassName);
			return saveFragment.saveToFileChooser(saveFragment);

		case OPENFILE :
			if (openFile == null)
				openFile = new PEFileChooser(fileDirectory, Localizer.currentLocalizer().get("action.ui.Open"));
			else openFile.rescanCurrentDirectory();
			openFile.setFileFilter(contentTypeClassName);
			if(openFile.showOpenDialog(null) == CANCEL_OPTION) return null;
			return openFile.getSelectedFile().toString();

		case OPENFRAGMENT :
			if (openFragment == null)
				openFragment = new PEFileChooser(fragmentDirectory, Localizer.currentLocalizer().get("action.ui.InsertFragment"));
			else openFragment.rescanCurrentDirectory();
			openFragment.setFileFilter(contentTypeClassName);
			if(openFragment.showOpenDialog(null) == CANCEL_OPTION) return null;
			return openFragment.getSelectedFile().toString();

		case BROWSEFILEDIR :
			PEFileChooser browseFileDir = new PEFileChooser(fileDirectory, Localizer.currentLocalizer().get("preferences.DefaultSaveDirectory"));
			browseFileDir.setFileSelectionMode(DIRECTORIES_ONLY);
			if (browseFileDir.showDialog(null, Localizer.currentLocalizer().get("misc.SelectDirectory")) == CANCEL_OPTION) return null;
			return browseFileDir.getSelectedFile().toString(); // [pending] save to properties file...

		default :
			throw new IndexOutOfBoundsException(Integer.toString(chooserType));
		}
	}

	/**
	 * Private constructor (prevents instanciation from the outside)
	 * Instanciates a new FileChooser with the given title and the given directory,
	 * and attaches it the adequate file filter (e.g. ".pic" & ".tex")
	 */
	private PEFileChooser(String directory, String title){

		super(directory);
		setDialogTitle(title);
		// 	setApproveButtonText("text"); //[pending] (syd) Don't seem to work ! Besides, there's no way to localize the Cancel button properties :-(((
		// 	setApproveButtonToolTipText("tool tip");
		// 	setApproveButtonMnemonic('a');

		String[] contentTypeNames = jpicedt.MiscUtilities.getAvailableContentTypesNames();
		String[][] contentTypeFileExtensions = jpicedt.MiscUtilities.getAvailableContentTypesFileExtensions();
		filters = new PEFileFilter[contentTypeNames.length];
		
		for (int i=0; i<contentTypeNames.length; i++){
			filters[i] = new PEFileFilter();
			String[] extensions = contentTypeFileExtensions[i];
			for (int j=0; j<extensions.length; j++){
				filters[i].addExtension(extensions[j]);
			}
			filters[i].setDescription(contentTypeNames[i]);
			addChoosableFileFilter(filters[i]);
		}
		
		addChoosableFileFilter(getAcceptAllFileFilter());
		//setFileFilter(filter); // now done from inside getFileName() according to given ContentType
	}

	private void setFileFilter(String contentTypeClassName){
		if (contentTypeClassName==null){ // do nothing
		}
		else {
			int idx = jpicedt.MiscUtilities.getContentTypeIndex(contentTypeClassName);
			setFileFilter(filters[idx]);
		}
	}
	
	/**
	 * Open a saveDialog box using the given FileChooser
	 * @return the selected file ; if it already exists, 
	 * ask user for confirmation ; return null if a CANCEL action did occur at any time.
	 */
	private String saveToFileChooser(PEFileChooser fc){
		File file;
		do {
			if (fc.showSaveDialog(null) == CANCEL_OPTION) return null;
			if (!(file = fc.getSelectedFile()).exists()) return file.toString();
			if (JOptionPane.showConfirmDialog(null,
			                                  Localizer.currentLocalizer().get("exception.FileAlreadyExists"),
			                                  getDialogTitle(),
			                                  JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
				return file.toString();
		} while (true);
	}
	
	/**
	 * Configure the default directory for the static JFileChooser's 
	 * from the given Properties object, using the following key/value pairs :<br>
	 * <ul>
	 * <li> key = KEY_FILE_DIRECTORY, value = path-to-a-directory-for-opening-new-files
	 * </ul>
	 * Then reinits the static JFileChooser so that they'll reflect the new preferences
	 * next time they're called.<p>
	 * Should be called on start-up, then each time preferences change.
	 */
	public static void configure(Properties preferences){
		fileDirectory = preferences.getProperty(KEY_FILE_DIRECTORY,DEFAULT_DIRECTORY);
		saveFile=saveFragment=openFile=openFragment=null;
	}
	
	/**
	 * Store the current file and fragment directories to the given Properties object 
	 */
	public static void store(Properties preferences){
		preferences.setProperty(KEY_FILE_DIRECTORY, fileDirectory);
	}
	
	
	
}
