/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.ui.dialog;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;
import java.util.*;
import javax.swing.filechooser.*;

/** 
 * a concrete implementation of FileFilter for LaTeX and rel. file extensions
 *
 * @author Sylvain Reynal
 * @since PicEdt 1.1
 */
public class PEFileFilter extends FileFilter {

    private static String TYPE_UNKNOWN = "Type Unknown";
    private static String HIDDEN_FILE = "Hidden File";

    private Hashtable filters = null;
    private String description = null;
    private String fullDescription = null;
    private boolean useExtensionsInDescription = true;

    /** 
     * Construct a new empty file filter
     */
    public PEFileFilter() { 
		this.filters = new Hashtable(); 
	}

    /**
     * Construct a new file filter with the given extension, e.g. "tex"
     */
    public PEFileFilter(String extension) { 
		this(extension,null);  
	}

    /** 
	 * Construct a new file filter with the given extension (e.g. "pst") and description ("Pstricks files")
     */
    public PEFileFilter(String extension, String description) {

    	this();
    	if(extension!=null) addExtension(extension);
     	if(description!=null) setDescription(description);
    }

    /** 
	 * Construct a new file filter with the given array of extensions, e.g. 
     * new PEFileFilter(String {"gif", "jpg"})
     */
    public PEFileFilter(String[] filters) { 
		this(filters, null);  
	}

    /**
	 * Construct a new file filter with the given array of extensions and the given description.
     * e.g. new PEFileFilter(String {"pic", "tex"}, "Fichiers Picture et Latex")
     */
    public PEFileFilter(String[] filters, String description) {

	this();
    	for (int i = 0; i < filters.length; i++) {
	      // ajoute les filtres un par un
	      addExtension(filters[i]);
    	}
     	if(description!=null) setDescription(description);
    }

    /** 
     * @return "true" if the given file can be displayed in the file filter
     */
    public boolean accept(File f) {
	    if(f != null) {
  	    if(f.isDirectory()) {
    		return true;
  	    }
	    String extension = getExtension(f);
	    //System.out.println("extension="+extension);
	    if(extension != null && filters.get(getExtension(f)) != null) {
	    	return true;
	      };
    	}
    	return false;
    }

    /** 
     * @return the extension associated with the given file
     */
     public String getExtension(File f) {

	 if(f != null) {
	     String filename = f.getName();
	     int i = filename.lastIndexOf('.');
	     if(i>0 && i<filename.length()-1) {
		 return filename.substring(i+1).toLowerCase();
	     };
	 }
	 return null;
     }

    /** 
     * Add a new extension to the file filter
     */
    public void addExtension(String extension) {

    	if(filters == null) {
	      filters = new Hashtable(5);
    	}
    	filters.put(extension.toLowerCase(), this);
    	fullDescription = null;
    }

    /** 
     * @return the file filter description e.g. "Picture files for LaTeX : *.pic, *.tex"
     */
    public String getDescription() {

    	if(fullDescription == null) {
    	  if(description == null || isExtensionListInDescription()) {
        	fullDescription = (description==null) ? "(" : description + " (";
     		  // construit la description à partir de la liste des extensions
      		Enumeration extensions = filters.keys();
      		if(extensions != null) {
    		    fullDescription += "." + (String) extensions.nextElement();
    		    while (extensions.hasMoreElements()) {
        			fullDescription += ", " + (String) extensions.nextElement();
    		    }
      		}
      		fullDescription += ")";
    	    } else {
        		fullDescription = description;
    	    }
      	}
      	return fullDescription;
      }

    /** 
     * Set the description string for this filter
     */
    public void setDescription(String description) {
    	this.description = description;
    	fullDescription = null;
    }

    /** 
	 * Set whether the list of extensions is included in the description string.
     */
    public void setExtensionListInDescription(boolean b) {
    	useExtensionsInDescription = b;
    	fullDescription = null;
    }

    /** 
     * @return whether the list of extensions is included in the description string.
     */
    public boolean isExtensionListInDescription() {
    	return useExtensionsInDescription;
    }

} // PEFileFilter

