/*
 InternalFrameMDIManager.java - January 1, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui.internal;

import jpicedt.*;

import jpicedt.graphic.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.toolkit.*;

import jpicedt.ui.*;
import jpicedt.ui.util.*;
import jpicedt.ui.action.*;
import jpicedt.ui.dialog.*;

import jpicedt.widgets.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.util.*;
import javax.swing.undo.*;
import javax.swing.event.*;

/**
 * An MDI Manager implementation for "childframe" mode (ie using JInternalFrame's).
 * <p>
 * <b>Layer management: </b>
 * <ul>
 * <li> PEDrawingBoard's are added to JLayeredPane.DEFAULT_LAYER (bottom most layer)
 * <li> DockablePanel's are added to JLayeredPane.PALETTE_LAYER
 * </ul>
 * </p>
 * @since jPicEdt 1.2
 * @author Sylvain Reynal
 * @version $Id: InternalFrameMDIManager.java,v 1.20.2.1 2007/09/02 11:56:30 reynal Exp $
 */
public class InternalFrameMDIManager extends MDIManager {

	private JFrame mainFrame;
	private PEMenuBar menubar;
	private PEToolBar mainToolbar; // shared actions, e.g. "Save", "Open", ...
	public JDesktopPane desktopPane;
	private InternalDesktopManager desktopManager;
	private InternalFrameEventHandler internalFrameHandler;
	private PopupMenuFactory popupMenuFactory;

	public static final String KEY_GEOMETRY_X = "ui.geometry.x";
	public static final String KEY_GEOMETRY_Y = "ui.geometry.y";
	public static final String KEY_GEOMETRY_WIDTH = "ui.geometry.width";
	public static final String KEY_GEOMETRY_HEIGHT = "ui.geometry.height";

	public static final String KEY_DESKTOP_COLOR = "ui.desktop-color";
	public static final Color desktopColorDEFAULT = new Color(106,105,207); // KDE2 !!!

	public String getName(){
		return MDI_CHILDFRAMES;
	}
	
	/**
	 * @param progressBar The progress bar to inform of progress in the initialization ; can be null.
	 */
	public InternalFrameMDIManager(PEProgressBar progressBar){

		// init LAF
		jpicedt.ui.LAFManager.updateLaf();

		// add a PropertyChangeListener to the UIManager, so that
		// every change in LAF leads to an immediate repainting of
		// the whole component tree
		UIManager.addPropertyChangeListener(new PropertyChangeListener(){
			                                    public void propertyChange(PropertyChangeEvent e) {
				                                    if (e.getPropertyName().equals("lookAndFeel")) {
					                                    JComponent c = (JComponent)mainFrame.getRootPane();
					                                    SwingUtilities.updateComponentTreeUI(c);
					                                    c.invalidate();	c.validate(); c.repaint();
				                                    }}});

		// init the application main frame
		mainFrame = new JFrame("jPicEdt "+Version.getVersion());
		mainFrame.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		mainFrame.getContentPane().setLayout(new BorderLayout());
		// allows JOptionPane to be instanciated with an implicit root frame
		JOptionPane.setRootFrame(mainFrame);

		try {
			// menubar
			menubar=new PEMenuBar();
			mainFrame.setJMenuBar(menubar);

			// main toolbar
			mainToolbar=new PEToolBar();
			mainFrame.getContentPane().add(mainToolbar, BorderLayout.NORTH);

			// test:
			//JPanel dac = new DockableAttributesCustomizer(progressBar);
			//mainFrame.getContentPane().add(dac, BorderLayout.WEST);


			// desktop
			desktopPane = new JDesktopPane();
			desktopPane.setBackground(desktopColorDEFAULT);
			desktopPane.setBorder(BorderFactory.createLineBorder(Color.black));
			//desktopPane.setDragMode(JDesktopPane.OUTLINE_DRAG_MODE); // JDK1.2 -> desktopPane.putClientProperty("JDesktopPane.dragMode","outline");
			mainFrame.getContentPane().add(desktopPane, BorderLayout.CENTER);
			int desktopPanePreferredWidth = (int)(Toolkit.getDefaultToolkit().getScreenSize().width * 0.9);
			int desktopPanePreferredHeight = (int)(Toolkit.getDefaultToolkit().getScreenSize().height * 0.9);
			desktopPane.setPreferredSize(new Dimension(desktopPanePreferredWidth,desktopPanePreferredHeight));
			desktopManager=new InternalDesktopManager();
			desktopPane.setDesktopManager(desktopManager);
			internalFrameHandler = new InternalFrameEventHandler();
		}
		catch(MissingResourceException e){
			System.err.println("During GUI initialization, I was unable to find internationalization resource: " + e.getKey());
			System.err.println("JVM stack trace dump :");
			e.printStackTrace();
			System.err.println("In any case, you may need to check the integrity of the jpicedt.jar file you downloaded.");
			System.err.println("If it still doesn't work, please send a bug report to : reynal@ensea.fr, with this console dump included.");
			System.err.println(":-((((((((((  Exiting... sorry !");
			SystemOutUtilities.instance().redirect(SystemOutUtilities.STANDARD); // close any open File Stream
			if (progressBar != null)
				progressBar.fatalError(e.getMessage() + "\nSee "+ SystemOutUtilities.getErrorLogFile() + " for details."); //[todo] to localize
			System.exit(0);
		}
		catch (Exception e) {
			System.err.println("OK, there SURE is a bug somewhere... Thank you for sending a bug report including the following stack trace to syd@jpicedt.org");
			System.err.println("Stack trace dump starts here :");
			e.printStackTrace();
			System.err.println(":-((((((((((  Exiting...sorry !");
			SystemOutUtilities.instance().redirect(SystemOutUtilities.STANDARD);// close any open File Stream
			if (progressBar != null)
				progressBar.fatalError(e.getMessage() + "\nSee "+ SystemOutUtilities.getErrorLogFile() + " for details."); //[todo] to localize
			System.exit(0);
		}
		mainFrame.addWindowListener(new WindowAdapter(){
			                            public void windowClosing(WindowEvent e){
				                            new jpicedt.JPicEdt.ExitAction().actionPerformed(null);
			                            }
		                            });
		mainFrame.validate();
		// size and location on the screen :
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		int x = JPicEdt.getProperty(KEY_GEOMETRY_X,(int)(0.05*screenSize.width));
		int y = JPicEdt.getProperty(KEY_GEOMETRY_Y,(int)(0.05*screenSize.height));
		int w = JPicEdt.getProperty(KEY_GEOMETRY_WIDTH,(int)(0.9*screenSize.width));
		int h = JPicEdt.getProperty(KEY_GEOMETRY_HEIGHT,(int)(0.9*screenSize.height));
		mainFrame.setSize(w,h);
		mainFrame.setLocation(x,y);

		// init colours from pref.
		update();
		if (progressBar != null) {
			progressBar.increment("GUI created successfully !");
			//[pending] progressBar.destroy(); // close progress bar
		}

		// setVisible "realizes" the main JFrame (i.e. create new Swing thread)
		mainFrame.setVisible(true);
	}


	/**
	 * Factory method that creates an MDIComponent which acts as a UI delegate to the given pane.
	 * This implementation returns a properly init'd JInternalFrame, with the given pane
	 * added to its content-pane.
	 */
	public MDIComponent createMDIComponent(JPanel pane){
		return new InternalMDIComponent(pane);
	}

	/**
	 * Depending on the inner pane of the given container, 
	 * adds it to the DEFAULT_LAYER or to the PALETTE_LAYER.
	 */
	public void addMDIComponent(MDIComponent c){
		//if (!(c instanceof InternalMDIComponent)) return;
		JPanel innerPane = c.getInnerPane();
		if (innerPane instanceof PEDrawingBoard) {
			desktopPane.add((InternalMDIComponent)c, JLayeredPane.DEFAULT_LAYER);
			((InternalMDIComponent)c).addInternalFrameListener(internalFrameHandler);
		}
		else {
			desktopPane.add((InternalMDIComponent)c, JLayeredPane.PALETTE_LAYER);
		}
		/*
		if (layer==JLayeredPane.DEFAULT_LAYER) desktopPane.add((InternalMDIComponent)c); // DEFAULT_LAYER (side effect : calls "activateFrame()")
		else mainFrame.getLayeredPane().add((InternalMDIComponent)c,layer);
		*/
		
		
	}

	public void closeBoard(PEDrawingBoard board){
		InternalMDIComponent mdiComp = (InternalMDIComponent)getHostingContainer(board);
		if (mdiComp==null) return;
		// simulate a click on the "close" button, it's safe because it's gives us a unified way of closing boards
		try {
			mdiComp.setClosed(true);
		}
		catch (PropertyVetoException e){}
	}

	/**
	 * Returns the dimension of the desktop that hosts the components of this MDIManager.
	 */
	public Dimension getDesktopSize(){
		return desktopPane.getSize();
	}

	/**
	 * Saves UI geometry to JPicEdt's preferences.
	 * @see #InternalFrameMDIManager
	 * @see #addDockablePanel
	 * @see DockablePanel#setVisible
	 */
	public void saveGeometry(){
		// save desktop geometry :
		Properties preferences = JPicEdt.getPreferences();
		preferences.setProperty(KEY_GEOMETRY_X,Integer.toString(mainFrame.getX()));
		preferences.setProperty(KEY_GEOMETRY_Y,Integer.toString(mainFrame.getY()));
		preferences.setProperty(KEY_GEOMETRY_WIDTH,Integer.toString(mainFrame.getWidth()));
		preferences.setProperty(KEY_GEOMETRY_HEIGHT,Integer.toString(mainFrame.getHeight()));
		super.saveGeometry(); // save DockablePanel's geometry
	}

	/**
	 * Updates local properties from the JPicEdt's preferences (e.g. GUI colours, rendering-hints,...)
	 * This implementation calls the superclass's method, then update the desktop-pane colour.
	 */
	public void update(){
		//desktopPane = null; // test error.log
		// colours :
		Color desktopColor = JPicEdt.getProperty(KEY_DESKTOP_COLOR,desktopColorDEFAULT);
		desktopPane.setBackground(desktopColor);
		if (mainFrame.isVisible()){
			desktopPane.repaint();
		}
		super.update(); // update RenderingHints for each PEDrawingBoard
	}



	//////////////////////////////////////////////////////////////////////
	//// GUI COMPONENTS
	//////////////////////////////////////////////////////////////////////

	/**
	 * Update JMenuItem's accelerators from the key/value pairs stored in JPicEdt's preferences 
	 */
	public void updateAccelerators(){
		 menubar.updateAccelerators();
	}

	/**
	 * Update the content (i.e. PEMenu's) of the menu-bar depending on the current state of the MDIManager.
	 */
	 public void updateMenuBar(){
		 menubar.updateMenuBar();
	 }
	
	/** 
	 * Updates the "recent files" sub-menu items after a frame was saved or opened.
	 * @since PicEdt 1.1
	 */
	 public void updateRecentFilesSubmenu(){
		 menubar.updateRecentFilesSubmenu();
	 }
	
	/** 
	 * Updates the "script" menu items 
	 * @since jpicedt 1.3.2
	 */
	 public void updateScriptsMenu(){
		 menubar.updateScriptsMenu();
	 }

	/** 
	 * Updates the "fragments" menu items 
	 * @since jpicedt 1.3.2
	 */
	 public void updateFragmentsMenu(){
		 menubar.updateFragmentsMenu();
	 }

	/**
	 * update "undo" and "redo" text
	 * @since jpicedt 1.3.2
	 */
	 public void updateUndoRedoMenus(String undoName,String redoName){
		 menubar.updateUndoRedoMenus(undoName, redoName);
	 }

	/** 
	 * Updates the Window menu as soon as an internal frame is opened or closed 
	 * (this menu keeps tracks of every open internal frames so that the user can activate 
	 * them quickly by use of a shortcut)
	 * 
	 * @since PicEdt 1.1
	 */
	 public void updateWindowMenu() throws MissingResourceException {
		 menubar.updateWindowMenu();
	 }

	/**
	 * update currently "selected" menuitem in Toolkit menu
	 */
	 public void updateToolkitMenu(PropertyChangeEvent e){
		 menubar.propertyChange(e);
	 }
		 
	/**
	 * Return an array containing all the menu-items components attached to menubars
	 */
	 public JMenuItem[] getMenuItems(){
		 return menubar.getMenuItems();
	 }


	//////////////////////////////////////////////////////////////////////
	//// BOARDS MANAGEMENT
	//////////////////////////////////////////////////////////////////////

	/**
	 * Returns the currently active drawing board ;  
	 * null if there's no internal frame in the desktopPane.<br>
	 * If there's at least one internal frame in the desktop, but it's deactivated 
	 * (this should never happen however), then the policy is to arbitrarly activate one frame 
	 * (namely the topmost frame on the desktop which is also the last one in the framearray)
	 */
	public PEDrawingBoard getActiveDrawingBoard(){

		JInternalFrame jifr = desktopPane.getSelectedFrame(); // returns null if there's no frame or if none is activated
		if (jifr != null){ // there's an active frame
			InternalMDIComponent ifr = (InternalMDIComponent)jifr;
			JPanel p = ifr.getInnerPane();
			if (p instanceof PEDrawingBoard) return (PEDrawingBoard)p;
			else return null;
		}
		// else either there's no frame, or none is active
		JInternalFrame[] allFrames = desktopPane.getAllFrames();
		if (allFrames.length==0) return null;
		// we have at least one frame in the desktop,
		// but none is activated, we activate (arbitrarly) one of them :
		//jpicedt.Log.debug("NO ACTIVE FRAME ! Activating top-most..");
		InternalMDIComponent ifr = (InternalMDIComponent)allFrames[allFrames.length-1]; // top-most frame
		if (ifr.isIcon()) desktopManager.deiconifyFrame(ifr);
		desktopManager.activateFrame(ifr); // same as if user clicked on the title bar
		//ifr.show();
		JPanel p = ifr.getInnerPane();
		if (p instanceof PEDrawingBoard) return (PEDrawingBoard)p;
		else return null;

	}

	/**
	 * selects the given drawing board by bringing it to front.
	 */
	public void selectDrawingBoard(PEDrawingBoard board){

		JInternalFrame iFrame = (JInternalFrame)getHostingContainer(board);
		if (iFrame == null) return;
		if (iFrame.isIcon()) desktopManager.deiconifyFrame(iFrame);
		desktopManager.activateFrame(iFrame); // simulate a user click on the title bar
	}

	///////////////////////////////////////////////////////////////////////
	/////// MODAL DIALOG BOXES
	///////////////////////////////////////////////////////////////////////
	/**
	 * [pending] experimental ! 1.3.2-beta9 ; not used yet.
	 */
	public void showModalDialogBox(String title, JComponent innerPane){
		ModalInternalFrame mif = new ModalInternalFrame(title, mainFrame.getRootPane(), desktopPane, innerPane);
		mif.setVisible(true); // block I/O
	}

	//////////////////////////////////////////////////////////////////////
	//// GEOMETRY
	//////////////////////////////////////////////////////////////////////

	/**
	 * Cascade all open internal frames
	 */
	public void cascadeDrawingBoards(){

		PEDrawingBoard[] allBoards = getAllDrawingBoards();
		ArrayList nonIconFrames = new ArrayList();

		int counter=0;
		for(int i = 0; i < allBoards.length; i++){
			MDIComponent container = getHostingContainer(allBoards[i]);
			if (!(container instanceof JInternalFrame)) continue;
			JInternalFrame ifr = (JInternalFrame)container;
			//if (ifr.isIcon()) desktopManager.deiconifyFrame(ifr);
			if (ifr.isIcon()) continue;
			try {
				ifr.setMaximum(false); // in case this one was Maximized (may fire a PropertyVetoException)
				ifr.pack();
				//ifr.show(); // already visible !
				//int dist = ifr.getHeight() - ifr.getContentPane().getHeight(); // title bar height
				int dist=30;
				int x = counter * dist;
				if (x + ifr.getWidth() > getDesktopSize().width) x = 0; // wrap around at desktop edge
				int y = counter * dist;
				if (y + ifr.getHeight() > getDesktopSize().height) y = 0; // wrap around at desktop edge
				ifr.setLocation(x,y);
				counter++;
			}
		catch (PropertyVetoException ex){}
		}
	}

	/**
	 * Tiles all open internal frames horizontally
	 */
	public void tileDrawingBoardsHorizontally(){

		PEDrawingBoard[] allBoards = getAllDrawingBoards();
		ArrayList nonIconFrames = new ArrayList();
		for(int i = 0; i < allBoards.length; i++){
			MDIComponent container = getHostingContainer(allBoards[i]);
			if (!(container instanceof JInternalFrame)) continue;
			JInternalFrame jif = (JInternalFrame)container;
			if (jif.isIcon()) {
				try {
					jif.setMaximum(false); // (may fire a PropertyVetoException)
					nonIconFrames.add(jif);
				}
				catch (PropertyVetoException ex){}
			}
		}
		int nonIconNb = nonIconFrames.size();
		if (nonIconNb == 0) return;

		// number of columns = int(sqrt(N))
		int colNb = (int)Math.floor(Math.sqrt(nonIconNb));
		int colWidth = getDesktopSize().width / colNb;

		int remaining = nonIconNb;

		for(int col = 0; col < colNb; col++){

			int numberOfFrameInThisColumn = (int)Math.floor(remaining / (colNb - col));
			int rowHeight = getDesktopSize().height / numberOfFrameInThisColumn;

			for (int i = 0; i < numberOfFrameInThisColumn; i++){
				JInternalFrame ifr = (JInternalFrame)nonIconFrames.get(nonIconNb-remaining);
				ifr.reshape(col*colWidth, i * rowHeight , colWidth, rowHeight);
				remaining--;
			}
		}
	}

	/**
	 * Tiles all open internal frames vertically.
	 */
	public void tileDrawingBoardsVertically(){

		PEDrawingBoard[] allBoards = getAllDrawingBoards();
		ArrayList nonIconFrames = new ArrayList();
		for(int i = 0; i < allBoards.length; i++){
			MDIComponent container = getHostingContainer(allBoards[i]);
			if (!(container instanceof JInternalFrame)) continue;
			JInternalFrame jif = (JInternalFrame)container;
			if (!jif.isIcon()) {
				try {
					jif.setMaximum(false); // (may fire a PropertyVetoException)
					nonIconFrames.add(jif);
				}
				catch (PropertyVetoException ex){}
			}
		}
		int nonIconNb = nonIconFrames.size();
		if (nonIconNb == 0) return;

		// number of row = int(sqrt(N)) where N = number of non-iconified frame
		int rowNb = (int)Math.floor(Math.sqrt(nonIconNb));
		int rowHeight = getDesktopSize().height / rowNb;
		int remaining = nonIconNb;

		for(int rowIndex = 0; rowIndex < rowNb; rowIndex++){

			int numberOfFrameInThisRow = (int)Math.floor(remaining / (rowNb - rowIndex));
			int colWidth = getDesktopSize().width / numberOfFrameInThisRow;

			for (int i = 0; i < numberOfFrameInThisRow; i++){
				JInternalFrame ifr = (JInternalFrame)nonIconFrames.get(nonIconNb-remaining);
				ifr.reshape(i*colWidth, rowIndex * rowHeight , colWidth, rowHeight);
				remaining--;
			}
		}
	}


	////////////////////////////////////////////////////
	/// MDIComponent -> INTERNAL FRAME
	/////////////////////////////////////////////////////

	/** simple adapter for JInternalFrame so that they implement the MDIComponent interface */
	private class InternalMDIComponent extends JInternalFrame implements MDIComponent {

		JPanel pane; // the inner pane, set getInnerPane()

		InternalMDIComponent(JPanel pane){
			super(null, true, true, true, true); // no title
			this.getContentPane().add(pane);
			this.pane = pane;
		}

		public void setTitle(String title){
			super.setTitle(title);
			revalidate();
			repaint();
		}

		public JPanel getInnerPane(){
			return pane;
		}

		public void setSelected(boolean b) throws PropertyVetoException {
			//Log.debug("title="+getTitle()+", sel="+b);
			if (!(pane instanceof PEDrawingBoard)) b=false; // never select sth else than a PEDrawingBoard
			super.setSelected(b);
		}

		// all other methods of the MDIComponent interface are already implemented by JInternalFrame...

	}

	////////////////////////////////////////////////////
	/// DESKTOP MANAGER
	/////////////////////////////////////////////////////

	/**
	 * Handler for internal frame events. 
	 * @since jPicEdt 1.4
	 */
	private class InternalFrameEventHandler extends InternalFrameAdapter {
		
		/**
		 * Updates "activeBoard".
		 * (note that this method is invoked AFTER JDesktopManager.activeFrame)
		 */
		public void internalFrameActivated(InternalFrameEvent e){ 
			//Log.debug(e.getInternalFrame().getTitle() + " got activated...");

			InternalMDIComponent f = (InternalMDIComponent)e.getInternalFrame();
			PEDrawingBoard board = (PEDrawingBoard)f.getInnerPane();  
			activeBoard = board; 				    
			//actionRegistry.updateActionState();

			PropertyChangeEvent propevt = new PropertyChangeEvent(this,ACTIVE_BOARD_CHANGE, null, board);
			boardEventHandler.propertyChange(propevt);
			
			/* 
			EditorKit kit = board.getCanvas().getEditorKit();
			propevt = new PropertyChangeEvent(kit,EditorKit.EDIT_MODE_CHANGE, null, kit.getCurrentMouseTool());
			boardEventHandler.propertyChange(propevt);
			*/
			
		}
		
		/**
		 * Try to save the content of the board if it's dirty before it gets closed.
		 * Also updates the widget states (menubar and toolbars).
		 * (note that this method is invoked BEFORE JDesktopManager.closeFrame)
		 */
		public void internalFrameClosing(InternalFrameEvent e){
			//Log.debug(e.getInternalFrame().getTitle() + " is closing...");
			
			InternalMDIComponent f = (InternalMDIComponent)e.getInternalFrame();
			boolean wasSelected = f.isSelected(); // remember its selected state for later
			
			if (removeDrawingBoard((PEDrawingBoard)f.getInnerPane())){ // cancelled ?
				f.dispose(); // indirectly invokes JDesktopManager.closeFrame(f), and sets up a new active board

				// [SR:pending] some day, move this code to MDIManager:
				
				// update GUI
				updateWindowMenu(); // it's fine because board was just removed from the hash

				// update Action's :
				JPicEdt.getActionRegistry().updateActionState(); 
				updateMenuBar();
			}
			
		}
	}
	
	/**
	 * Handler for desktop events. This work in conjunction with InternalFrameEventHandler, since there are 
	 * subtle chronology issues between incoming events ;-)
	 */
	private class InternalDesktopManager extends DefaultDesktopManager {

		/**
		 * Called when a frame gets activated, and before InternalFrameEventHandler.internalFrameActivated() is invoked.
		 * Enforces the selected state of the given frame (this is a work-around to the presence of DockablePanel's
		 * in the same layer).
		 */
		public void activateFrame(JInternalFrame f) {
			//jpicedt.Log.debug("iframe="+f.getTitle());
			super.activateFrame(f);
			if (((InternalMDIComponent)f).getInnerPane() instanceof PEDrawingBoard){
				try {f.setSelected(true);} // simulated a click on the title bar; no risk of recursion, since this has no effect on an already selected frame 
				catch(PropertyVetoException e2) {}
			}
		}

		/**
		 * called when a frame gets closed, and AFTER InternalFrameEventHandler.internalFrameClosing()
		 * Overriden so as to update "activeBoard".
		 * @param f frame getting closed
		 */
		public void closeFrame(JInternalFrame f){
			//jpicedt.Log.debug("iframe="+f.getTitle());
			boolean wasSelected = f.isSelected(); // remember its selected state for later
			super.closeFrame(f); // closeFrame() sometimes tries to activate a DockablePanel (which is not what we want), 
			if (((InternalMDIComponent)f).getInnerPane() instanceof PEDrawingBoard){
				if (wasSelected) {
					activeBoard = activateNextBoard(); // ... so we gotta do the job ourselves...
					PropertyChangeEvent propevt = new PropertyChangeEvent(this,ACTIVE_BOARD_CHANGE, null, activeBoard); 
					boardEventHandler.propertyChange(propevt);
				}
			}
		}


		private PEDrawingBoard activateNextBoard() { // copied from DefaultDesktopManager with minor changes...
			int i;
			InternalMDIComponent nextFrame = null;
			for (i = 0; i < desktopPane.getComponentCount(); i++) {
				InternalMDIComponent comp = (InternalMDIComponent)desktopPane.getComponent(i);
				if (comp.getInnerPane() instanceof PEDrawingBoard){ // skip non drawing boards...
					nextFrame = comp;
					break;
				}
			}
			if (nextFrame != null) {
				try { nextFrame.setSelected(true); }
				catch (PropertyVetoException e2) { }
				nextFrame.moveToFront();
				return (PEDrawingBoard)nextFrame.getInnerPane();
			}
			return null;
		}
		
		/**
		 * used for debugging purpose
		 */
		private String printAllFrames(){

			JInternalFrame[] jif = desktopPane.getAllFrames();
			String s = "          Desktop contains : ";
			for (int i = 0; i < jif.length; i++){
				s += jif[i].getTitle() + "; ";
			}
			return s;
		}

	} // InternalDesktopManager

} // class

