/*
 PEPopupMenuFactory.java - January 4, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui.internal;

import jpicedt.JPicEdt;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.*;
import jpicedt.graphic.view.HitInfo;
import jpicedt.ui.dialog.*;
import jpicedt.ui.action.*;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * A factory to create a PEPopupMenu for an EditorKit.
 * @since jPicEdt
 * @author Sylvain Reynal
 * @version $Id: PEPopupMenuFactory.java,v 1.22.2.1 2007/09/02 11:56:30 reynal Exp $
 */
public class PEPopupMenuFactory implements PopupMenuFactory {

	/**
	 * Create a new PEPopupMenuFactory using the given action-map to retrieve actions. 
	 */
	public PEPopupMenuFactory(){
	}
		
	/**
	 * @param board the popup-menu invoker component
	 * @param element if non-null, the graphic element on which the click occured.
	 */
	public JPopupMenu createPopupMenu(PECanvas board, HitInfo hi){
		PopupMenu pm = new PopupMenu(board, hi);
		return pm;
	}

	/**
	 * <p>This class implements a popup-menu activable from a right-click on the associated PECanvas.</p>
	 * <p>This is a context-sensitive popup-menu, which depends on the object being clicked and its "selected" state.
	 * @since PicEdt 1.2.a
	 */
	public class PopupMenu extends JPopupMenu {

		ActionLocalizer localizer;
		
		public void show(java.awt.Component c, int x, int y){
			super.show(c,x,y);
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("x="+x+",y="+y);
		}

		/**
		 * Construct a new pop-up menu associated with the given PECanvas (the component on which a right mouse click occured that triggered this popup).
		 * @param hi the HitInfo obtained from the mouse-event which triggered this popup-menu, by scanning the whole drawing.
		 * @since PicEdt 1.2.a
		 */
		public PopupMenu(PECanvas canvas, HitInfo hi) {
			
			// hi is not used yet, but this may prove useful to specialize menu-items production 
			
			JMenuItem menuItem;
			ActionDispatcher actionDispatcher = new DefaultActionDispatcher(canvas);
			localizer = jpicedt.Localizer.currentLocalizer().getActionLocalizer();
			ActionRegistry actionMap = JPicEdt.getActionRegistry();

			// compute size of active selection :
			int number_of_selected_obj=canvas.getSelectionSize();
			
			

			// 1°) one selected object :
			if (number_of_selected_obj == 1) {

				Iterator it = canvas.selection();
				Element selObj = (Element)it.next();

				add(new JLabel(" "+selObj.getName(), SwingConstants.RIGHT)); addSeparator();
				
				menuItem=add(actionMap.getAction(EditorKit.EditGeometryAction.KEY));
				menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F2,0));

				// add Element specific actions
				if (selObj instanceof ActionFactory){
					PEAction[] actionArray = ((ActionFactory)selObj).createActions(actionDispatcher,localizer,hi);
					if (actionArray != null){
						for (int i=0; i<actionArray.length; i++) add(actionArray[i]);
					}
				}
				// ibid. for elements which don't support ActionFactory
				if (selObj instanceof PicGroup){
					add(actionMap.getAction(EditorKit.UngroupAction.KEY));
				}
				addSeparator();
				
				// disposition
				BranchElement parent = (BranchElement)selObj.getParent();
				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.TOBACK));
				menuItem.setEnabled(!parent.isToBack(selObj));

				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.BACKWARD));
				menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_MULTIPLY,0));
				menuItem.setEnabled(!parent.isToBack(selObj));

				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.FORWARD));
				menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_SLASH,0));
				menuItem.setEnabled(!parent.isToFront(selObj));

				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.TOFRONT));
				menuItem.setEnabled(!parent.isToFront(selObj));
				addSeparator();
				
			}

			// 2°) more than one selected object :
			else if (number_of_selected_obj > 1) {
				add(new JLabel(" "+jpicedt.Localizer.currentLocalizer().get("misc.Selection"), SwingConstants.RIGHT));
				addSeparator();
				add(actionMap.getAction(EditorKit.GroupAction.KEY));
				menuItem = add(actionMap.getAction(EditorKit.JoinCurvesAction.KEY));
				menuItem.setEnabled(canvas.getEditorKit().getSelectionHandler().containsClass(PicMultiCurveConvertable.class));
				addSeparator();
				// disposition 
				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.TOBACK));
				menuItem.setEnabled(true); // [SR:pending] should work this out, depending on the lower-z element being selected or not
										// yet this is a hard one !
				menuItem=add(actionMap.getAction(EditorKit.EditDispositionAction.TOFRONT));
				menuItem.setEnabled(true);
				addSeparator();
			}

			// get always added :
			add(actionMap.getAction(JPicEdt.ViewLaTeXFileAction.KEY));
			add(actionMap.getAction(EditorKit.EditBoundingBoxAction.KEY));
		}

		/**
		 * <p>Overriden from JPopupMenu so as to take localization into account.</p>
		 * <p>Exemple : suppose we add an action whose name (ie value associated with key Action.NAME) is "GroupAction". Then :<ul>
		 * <li>menuitem's text gets constructed from localized "GroupAction" in the adequate locale resource file (res/i18n_xx.properties)</li>
		 * <li>menuitem's mnemonic gets constructed from "GroupAction_menuitem_mnemonic" in the same resource file, IF IT'S FOUND.</li>
		 * <li>menuitem's accelerator gets constructed from "GroupAction_menuitem_accelerator" in the same resource file, IF IT'S FOUND.</li>
		 * </ul>
		 */
		public JMenuItem add(Action a){

			JMenuItem mi = super.add(a);
			//mi.setText((String)a.getValue(Action.NAME));
			//mi.setMnemonic(Localizer.currentLocalizer().get(name + "_menuitem_mnemonic").charAt(0));
			KeyStroke ks =  (KeyStroke)a.getValue(Action.ACCELERATOR_KEY);
			if (ks != null) mi.setAccelerator(ks);
			mi.setIcon(null);
			return mi;
		}

		/**
		 * Creates a PEMenu from the given String, using JPicEdt's default ActionLocalizer to fetch
		 * the associated label (ie "name") and mnemonic ("name.mnemonic").
		 * @param name the name of the menu, used to determine :
		 * Example : createMenu("File") create a menu with label "File" and mnemonic = first char of "File.mnemonic"
		 *
		 */
		 private PEMenu createMenu(String name) throws MissingResourceException {

			 PEMenu menu = new PEMenu(localizer.getActionName(name));
			 menu.setMnemonic(localizer.getActionMnemonic(name).intValue());
			 return menu;
		 }
		

	} // PopupMenu


} // PEPopupMenuFactory


