/*
 DefaultParser.java - February 21, 2002 - jPicEdt 1.3.3, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input;


import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.*;
import jpicedt.format.input.util.*;
import jpicedt.graphic.io.parser.*;

import java.awt.Color;
import java.io.*;
import java.util.*;

/**
 * Default implementation of the Parser interface.
 * <p> 
 * This expression doesn't contains rules in itself. On the contrary, it simply lumps 
 * sub-trees together, each sub-tree containing grammar rules for a particular format, e.g.
 * PsTricks, eepic,...
 * <p>
 * Basically, this parser hold data using three objects :
 * <ul>
 * <li> an instance of util.Pool : this stores persistent data, and allow them to be shared by 
 * AbstractRegularExpression's across the whole parser tree.
 * <li> an instance of util.Context : this holds the current state of the lexicographic analyser (line number, caret 
 * position, block boundaries, etc...)
 * <li> an instance of AlternateExpression : this holds a reference to the various supported grammars by
 * mixing them into a single tree.
 * </ul>  
 * <p>
 * [Developpers] for those willing to add their own parser to the tree, simply write a new parser-tree whose 
 * root expression inherits from 
 * {@link jpicedt.format.input.util.RootExpression RootExpression}, 
 * then add it to this parser by using the
 * <code>addGrammar</code> method.
 * @since jpicedt 1.2
 * @author Sylvain Reynal
 * @version $Id: DefaultParser.java,v 1.8.2.1 2007/09/02 11:56:00 reynal Exp $
 *
 */
public class DefaultParser extends SequenceExpression implements ExpressionConstants, Parser {

	/** a pool of primitive types and objects that may be shared by sub-expressions */
	protected Pool pool;

	/** holds information about the current parser state */
	protected Context context;
	
	/** holds a reference to each supported grammar */ 
	protected AlternateExpression headExp;

	/**
	 * Creates a new DefaultParser, and attach some grammar rules to it.<br> 
	 * Tree structure = (WhiteSpaceOrEOL & (headExp | notParsable))*<br>
	 * where : headExp = Root1 | Root2 | Root3 ... (each RootX sub-tree representing a particular format).
	 * <p>
	 * Note that with this current structure, sub-trees must NOT be mutually exclusive, ie it's ok with eepic and pstricks
	 * for instance (it wouldn't work with PDF and LaTeX, since once we've started with PDF, it wouldn't make
	 * sense trying to parse LaTeX commands...) -> 
	 */
	public DefaultParser() {

		super(false); // doesn't throw IncompleteSequence exception (wouldn't make sense, seeing that we may have no whitespaces at all, see below)
		pool = new Pool();
		AlternateExpression alt = new AlternateExpression();
		alt.add(headExp = new AlternateExpression());
		// must always be the last-added expression :
		alt.add(new NotParsableExpression(pool));
		// swallow white-spaces and/or EOL between each picObjExp's subexpression:
		add(new RepeatExpression(new SequenceExpression(new WhiteSpacesOrEOL(), alt), 0, AT_LEAST));

		// [SR:pending] load grammar dynamically by using our own classloader
		this.addGrammar(new jpicedt.format.input.latex.LaTeXParser(this.getPool()));
		this.addGrammar(new jpicedt.format.input.eepic.EepicParser(this.getPool()));
		this.addGrammar(new jpicedt.format.input.pstricks.PstricksParser(this.getPool()));
		
	}

	/**
	 * Add a new sub-tree to this parser
	 */
	public void addGrammar(RootExpression root){
		headExp.add(root);
	}

	/**
	 * Return the pool containing variables shared across the grammar tree
	 */
	public Pool getPool(){
		return pool;
	}
	
	/**
	 * kicks off the parsing process
	 */
	public Drawing parse(Reader reader) throws REParserException {
		// [pending] change to "Drawing parse(Reader r, Drawing d)" allowing to pass an already init'd drawing

		
		// reinit Pool, then every RootExpression that was added to headExp :
		pool.reinit(); 
		ArrayList rootExpList = headExp.getExpressionList();
		for (Iterator it = rootExpList.iterator(); it.hasNext();){
			Object o = it.next();
			if (o instanceof RootExpression)  ((RootExpression)o).reinit(); 
		}

		try{
			interpret(context=new Context(reader));
		}
		catch(REParserException.EndOfPicture e){
			// check if every "begin group" has a matching "end group", which is the case
			// only if picGroupStack (the fifo that stores open groups) is empty.
			if (!pool.picGroupStack.empty()) throw new REParserException.BeginGroupMismatch(context, null);
			Drawing dr = new Drawing(pool.currentGroup);
			// [pending] change to dr.addElement(pool.currentGroup)
			// [pending] set ContentType for Drawing
			dr.setNotparsedCommands(pool.notParsed.toString()); // add commands that weren't parsed
			// to the "end" of the current drawing (i.e. these may appear at the end of the text file, though
			// not on the drawing).
			//System.out.println(this);
			return dr;
		}
		catch(REParserException.EOF e){
			if (!pool.picGroupStack.empty()) throw new REParserException.BeginGroupMismatch(context, null);
			Drawing dr = new Drawing(pool.currentGroup);
			dr.setNotparsedCommands(pool.notParsed.toString()); // add commands that weren't parsed
			//System.out.println(this);
			return dr;
		}
		throw new REParserException.EndOfPictureNotFound();
	}


	/**
	 * @return a String representation of this Parser, i.e. global variables contents
	 */
	public String toString(){

		return "DefaultParser :\n"
		       + pool.toString()
			   + "\nGrammar :\n"
			   + headExp.toString();
	}

	/**
	 * parser standalone test
	 * @param arg[0] name of file to parse.
	 */
	public static void main(String arg[]){

		DefaultParser parser = new DefaultParser(); // create a new grammar
		/*
		Pool pool = parser.getPool();
		parser.addGrammar(new jpicedt.format.input.latex.LaTeXParser(pool));
		parser.addGrammar(new jpicedt.format.input.eepic.EepicParser(pool));
		parser.addGrammar(new jpicedt.format.input.pstricks.PstricksParser(pool));
		*/
		
		FileReader reader=null;
		try{
			reader = new FileReader(arg[0]);
		} catch (IOException ioex){ioex.printStackTrace(); System.exit(0);}
		if (reader==null) System.exit(0);
		Drawing dr = null;
		try{
			dr=parser.parse(reader);
		}
		catch (Exception e){e.printStackTrace();}
		System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
		System.out.println("Default parser=" + parser);
		System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
		System.out.println("Parsed Drawing=" + dr);
		try{
			reader.close();
		} catch (IOException ioex){ioex.printStackTrace(); System.exit(0);}
		System.exit(0);

	}
}
