/*
 LaTeXBox.java - November 13, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.latex;

import jpicedt.format.input.pstricks.PsRPutExpression;
import jpicedt.format.input.util.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;

/**
 * Parses LaTeX boxes, i.e. makebox, framebox, frame and dashbox. This is to be used as a child expression of either
 * {@link #LaTeXPutExpression LaTeXPutExpression} or 
 * {@link #jpicedt.format.input.pstricks.PsRPutExpression jpicedt.format.input.pstricks.PsRPutExpression}.
 * <p>
 * Instanciates a PicText element. Argument =<ul> 
 * <li> \\makebox(w,h)[vh]{text or LaTeX command}</li>
 * <li> \\framebox(w,h)[vh]{text or LaTeX command}</li>
 * <li> \\dashbox{dash}(w,h)[vh]{text or LaTeX command}</li>
 * </ul>
 * <p> The parsing code uses several keys/values pairs stored by in the Pool:
 * <ul>
 * <li> KEY_PUT_POINT : PicPoint's argument of the \\put or \\rput command
 * <li> 
 * <p>
 * <code>interpret()</code> return false if none of these have been found, and cursor position is left unchanged<br>
 * @author Sylvain Reynal
 * @since jpicedt 1.4
 * @version $Id: LaTeXBox.java,v 1.3.2.1 2007/09/02 11:56:01 reynal Exp $
 *
 */
public class LaTeXBox extends SequenceExpression  implements PicObjectConstants, ExpressionConstants {

	private double boxWidth, boxHeight, boxDash;
	private boolean isFixedSizeBox=false;
	private String horAlign, vertAlign; // set by TextAlignment
	private Pool pool;
	

	public LaTeXBox(Pool pl) {

		super(true);// possibly throws IncompleteSequence Exception if only \\XXXXbox was found
		this.pool = pl;

		// boxtype :
		AlternateExpression boxType = new AlternateExpression();

		// \\makebox
		PicAttributeSet dummy = new PicAttributeSet();
		boxType.add(new LaTeXInstanciationExpression("\\makebox",new PicText(new PicPoint(), "", dummy),pool));

		// \\framebox
		dummy.setAttribute(TEXT_FRAME, TEXT_BOX_RECTANGLE);
		boxType.add(new LaTeXInstanciationExpression("\\framebox",new PicText(new PicPoint(), "", dummy),pool));

		// \\frame (must appear AFTER framebox !!!)
		boxType.add(new LaTeXInstanciationExpression("\\frame",new PicText(new PicPoint(), "", dummy),pool));

		// \\dashbox{dash}
		boxType.add(new SequenceExpression(
		                new LaTeXInstanciationExpression("\\dashbox",new PicText(new PicPoint(), "", dummy),pool),
		                WHITE_SPACES_OR_EOL,
		                new SequenceExpression(
		                    new LiteralExpression("{"), // {dash}
		                    new NumericalExpression(DOUBLE, POSITIVE, "}", true) {
			                    public void action(ParserEvent e) {
				                    if (DEBUG) System.out.println(e);
				                    boxDash = ((Double) e.getValue()).doubleValue() * pool.getDouble(LaTeXParser.KEY_UNIT_LENGTH);
				                    pool.currentObj.setAttribute(DASH_OPAQUE, new Double(boxDash));
				                    pool.currentObj.setAttribute(DASH_TRANSPARENT, new Double(boxDash));
			                    }
		                    },
		                    true),
		                true));// possibly throws IncompleteSequence Exception once "{" was found
		this.add(boxType);
		this.add(WHITE_SPACES_OR_EOL);
		// (w,h) (optional ; if found, we instanciate a PicRectangle and make the previously instanciated
		//        PicText frameless)
		this.add(new OptionalExpression(new PicPointExpression("(", ",", ")"){
			                                public void action(ParserEvent e){
				                                if (DEBUG) System.out.println(e);
				                                isFixedSizeBox = true; // flag to inform the EnclosingExpression - see below - it must instanciate a PicRectangle
				                                PicPoint pt = (PicPoint)e.getValue();
				                                boxWidth = pt.x; boxHeight = pt.y;
			                                }}));
		this.add(WHITE_SPACES_OR_EOL);
		// optional alignment : [vh] where v = b|t|c|"" and h = r|l|c|"" ("" default to center)
		SequenceExpression alignmentExp = new SequenceExpression(true);// throw IncompleteSequence Exception
		alignmentExp.add(new LiteralExpression("["));
		alignmentExp.add(new TextAlignment()); // swallows chars till "]" is found
		alignmentExp.add(WHITE_SPACES_OR_EOL);
		this.add(new OptionalExpression(alignmentExp)); // set horAlign and vertAlign
		// text or LaTeX Command
		this.add(WHITE_SPACES_OR_EOL);
		this.add(new BoxEnclosedText()); // instanciate a new PicText, and sets its attributes from global vars
	}
	
	/** called when this SequenceExpression was successfully parsed -> reinit locals for next time */
	public void action(ParserEvent e){
		if (DEBUG) System.out.println(e);
		boxWidth = boxHeight = boxDash = 0;
		isFixedSizeBox=false;
		horAlign=TEXT_HALIGN_CENTER;
		vertAlign=TEXT_VALIGN_CENTER;
	}
	

	/**
	 * Handles text inside a box-command (e.g. \\makebox{text}), after replacing CR by white-spaces as TeX does.
	 * This is where a new PicText is added to the current stack
	 */
	class BoxEnclosedText extends EnclosingExpression {

		public BoxEnclosedText(){
			super("{", null, "}");
		}

		public void action(ParserEvent e) {
			if (DEBUG) System.out.println(e);
			// instanciate PicText:
			PicText text = ((PicText)pool.currentObj);
			// fill string:
			String s = getEnclosedString().replace('\n',' ');
			s = Context.removeRedundantWhiteSpaces(s);
			text.setText(s);
			// set alignment: (note that for makebox and the like, \\rput's alignment parameters are meaningless)
			text.setVertAlign(vertAlign);
			text.setHorAlign(horAlign);
			// set put point:
			PicPoint putPoint = (PicPoint)pool.get(PsRPutExpression.KEY_RPUT_POINT);
			if (putPoint==null) putPoint = (PicPoint)pool.get(LaTeXPutExpression.KEY_PUT_POINT);
			if (putPoint==null) putPoint = new PicPoint(); // default to (0,0)
			// set rotation if applicable (i.e. called from PsRPutExpression):
			// rotation:
			Double rotation = (Double)pool.get(PsRPutExpression.KEY_RPUT_ROTATION);
			if (rotation!=null) text.setAttribute(TEXT_ROTATION, rotation);
			// set box size, if applicable:
			if (isFixedSizeBox){
				// change text location  :
				PicPoint ptText = new PicPoint();
				if (vertAlign == TEXT_VALIGN_TOP) ptText.y = putPoint.y + boxHeight;
				else if (vertAlign == TEXT_VALIGN_BOTTOM) ptText.y = putPoint.y;
				else ptText.y = putPoint.y + 0.5*boxHeight; // CENTER_V
				if (horAlign == TEXT_HALIGN_LEFT) ptText.x = putPoint.x;
				else if (horAlign == TEXT_HALIGN_RIGHT) ptText.x = putPoint.x + boxWidth;
				else ptText.x = putPoint.x + 0.5*boxWidth; // CENTER_H
				text.setPoint(PicText.P_ANCHOR, ptText,null);
				// possibly add a box on its own, set its size and location :
				if (text.getFrameType() == TEXT_BOX_RECTANGLE){
					// suppress box from PicText
					text.setFrameType(TEXT_BOX_NO_FRAME);
					PicPoint pt2 = new PicPoint(putPoint.x + boxWidth, putPoint.y + boxHeight);
					// copy attribute set from text to box (dash, ...)
					PicParallelogram box = new PicParallelogram(putPoint,pt2);
					box.setAttributeSet(text.getAttributeSet());
					pool.currentGroup.addChild(box);
					pool.currentObj = box; // only for security reason... (probably useless)
				}
				isFixedSizeBox = false;	// reset flag for next time !
			}
			else {
				text.setPoint(PicText.P_ANCHOR, putPoint,null);
			}
		}
	}
	
	
	/**
	 * handles boxes alignement, e.g. "bc]" ,... First "[" isn't parsed here. 
	 * Modifies global vars "vertAlign" and "horAlign".
	 */
	class TextAlignment extends WordExpression {

		public TextAlignment(){
			super("]", true); // swallows "]" and can return an empty String as in "[]"
		}

		public void action(ParserEvent e) {
			if (DEBUG) System.out.println(e);
			String s = (String) e.getValue();

			if (s.indexOf("t")!=-1) vertAlign = TEXT_VALIGN_TOP;
			else if (s.indexOf("b")!=-1) vertAlign = TEXT_VALIGN_BOTTOM;
			else vertAlign = TEXT_VALIGN_CENTER;

			if (s.indexOf("l")!=-1) horAlign = TEXT_HALIGN_LEFT;
			else if (s.indexOf("r")!=-1) horAlign = TEXT_HALIGN_RIGHT;
			else horAlign = TEXT_HALIGN_CENTER;
		}
	}

}

