/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.model.*;

/**
 * Parses statement involving length units (as set by PsTricks's "runit" register), 
 * e.g. "linewidth=13mm" or "doublesep=5.6" (default to current "runit")<p>
 * Numerical parsed value is pushed in the Pool's PicAttributeSet having the given key, 
 * (after conversion to mm)
 * <p>
 * This expression contains a StatementExpression with parameters "DOUBLE" and "POSITIVE".
 * So far, this expression parses "cm", "mm" and "pt".
 * <p>
 * This expression MUST NOT be used to parse expressions that modify one of "xunit", "yunit" or "runit", since
 * it relies on the value of these registers to compute the actual length value in case no unit is present !!!
 * @author Sylvain Reynal
 * @version $Id: PSTLengthParameter.java,v 1.4.2.1 2007/09/02 11:56:12 reynal Exp $ 
 */
public class PSTLengthParameter extends SequenceExpression implements PicObjectConstants {

	private Pool pool;
	private PicAttributeName attribute;
	private Pool.Key setKey;
	private double val; // temporary stores parsed value

	/**
	 * @param pl parser's pool
	 * @param tag LHS tag (e.g. "linewidth" or "doublesep") for the StatementExpression
	 * @param attribute name of attribute to modify (must be a predefined PicAttributeName of type "Double" or at least "Number")
	 * @param attributeSetKey used to fetch the attribute set in which parsed parameters are stored. 
	 */
	public PSTLengthParameter(Pool pl, String tag, PicAttributeName attributeName, Pool.Key attributeSetKey){

		super(false);
		pool = pl;
		attribute = attributeName;
		this.setKey = attributeSetKey;

		add(new StatementExpression(tag,"=",null,DOUBLE,POSITIVE){ // no postfix, but possibly whitespaces before
			    public void action(ParserEvent e){
				    if(DEBUG) System.out.println(e);
				    val = pool.getDouble(PstricksParser.KEY_R_UNIT) * ((Number)e.getValue()).doubleValue(); // suppose there's no unit
				    pool.setAttribute(setKey,attribute, new Double(val));
		}});
		// now, if there's a specified unit, we rescale to parsed value (i.e. dividing by pstRunit), then we compute the actual length value depending on the unit found
		add(new OptionalExpression(
			new AlternateExpression(
		    	new LiteralExpression("cm"){
					public void action(ParserEvent e){
						if (DEBUG) System.out.println(e);
				        val = val * 10.0 / pool.getDouble(PstricksParser.KEY_R_UNIT);
				        pool.setAttribute(setKey,attribute, new Double(val));
				}},
		        new LiteralExpression("mm"){
					public void action(ParserEvent e){
						if (DEBUG) System.out.println(e);
				        val = val / pool.getDouble(PstricksParser.KEY_R_UNIT);
				        pool.setAttribute(setKey,attribute, new Double(val));
				}},
		        new LiteralExpression("pt"){
					public void action(ParserEvent e){
						if (DEBUG) System.out.println(e);
				        val = val * PS_POINT / pool.getDouble(PstricksParser.KEY_R_UNIT);
				        pool.setAttribute(setKey,attribute, new Double(val));
				}})));
	}
} // class
