/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.PicPoint;

/**
 * Parses \\psarc commands :
 * <ul>
 * <li>\\psarc[parameters]{<->}(x0,y0){rad}{angleA}{angleB} ; where at least "arrows" or (x0,y0) args must be included
 * <li>\\psarc*[parameters]{<->}(x0,y0){rad}{angleA}{angleB} ; idid. but filled
 * </ul>
 * @author Sylvain Reynal
 * @version $Id: PsArcExpression.java,v 1.5.2.1 2007/09/02 11:56:12 reynal Exp $ 
 */
public class PsArcExpression extends SequenceExpression implements PicObjectConstants {

	private Pool pool;
	private double startAngle;
	private String type;
	public static final String ARC = "\\psarc";
	public static final String WEDGE = "\\pswedge";

	/**
	 * @param type ARC, WEDGE
	 * @param tag optional (used only by \\newpsobject commands), can be null
	 */
	public PsArcExpression(Pool pl, String type){
		this(pl,type,null);
	}

	/**
	 * @param type ARC, WEDGE
	 * @param tag optional (used only by \\newpsobject commands), can be null
	 */
	public PsArcExpression(Pool pl, String type, String tag){

		super(true); // throw IncompleteSequence Exception
		this.pool = pl;
		this.type = type;
		if (tag == null) tag = type;

		PicEllipse arc;
		if (type==ARC) arc = new PicEllipse(java.awt.geom.Arc2D.OPEN);
		else arc = new PicEllipse(java.awt.geom.Arc2D.PIE);
		
		add(new PSTInstanciationExpression(tag, arc, pool));
		add(WHITE_SPACES_OR_EOL);
		add(new OptionalExpression(new StarExpression(pool))); // fill solid
		add(WHITE_SPACES_OR_EOL);
		// add optional param here
		add(new OptionalExpression(new EnclosingExpression("[",new PSTParametersExpression(pool,Pool.CURRENT_OBJ_ATTRIBUTES),"]"))); // push in object's attributeSet
		// arrows (optional for arc, forbidden for wedge)
		add(WHITE_SPACES_OR_EOL);
		if (type==ARC) add(new OptionalExpression(new PSTArrowExpression(pool)));
		// arc's center
		add(WHITE_SPACES_OR_EOL);
		add(new PSTPicPointExpression(PicEllipse.P_CENTER,pool));
		add(WHITE_SPACES_OR_EOL);
		// radius :
		add(new LiteralExpression("{"));
		add(WHITE_SPACES_OR_EOL);	
		add(new NumericalExpression(DOUBLE,POSITIVE,"}",true){
			    public void action(ParserEvent e){
				    if (DEBUG) System.out.println(e);
					double radius = ((Double)e.getValue()).doubleValue() * pool.getDouble(PstricksParser.KEY_R_UNIT);
					// recall center in order to build BL and TR corners :
					PicPoint pBL = pool.currentObj.getPoint(PicEllipse.P_CENTER,null);
					PicPoint pTR = new PicPoint(pBL);
					pBL.translate(-radius, -radius);
					pTR.translate(radius, radius);
				    ((PicEllipse)(pool.currentObj)).setPoint(PicEllipse.P_BL, pBL,null);
					((PicEllipse)(pool.currentObj)).setPoint(PicEllipse.P_TR, pTR,null);
			    }});
		add(WHITE_SPACES_OR_EOL);

		// angleA = startAngle :
		add(new LiteralExpression("{"));
		add(WHITE_SPACES_OR_EOL);
		add(new NumericalExpression(DOUBLE,ANY_SIGN,"}",true){
			    public void action(ParserEvent e){
				    if (DEBUG) System.out.println(e);
				    startAngle = ((Double)e.getValue()).doubleValue();
				    ((PicEllipse)(pool.currentObj)).setAngleStart(startAngle);
			    }});
		add(WHITE_SPACES_OR_EOL);
		// angleB = endAngle ; we need angleExtent = angleB - angleA
		add(new LiteralExpression("{"));
		add(WHITE_SPACES_OR_EOL);
		add(new NumericalExpression(DOUBLE,ANY_SIGN,"}",true){
			    public void action(ParserEvent e){
				    if (DEBUG) System.out.println(e);
				    double angleExtent = ((Double)e.getValue()).doubleValue() - startAngle;
				    ((PicEllipse)(pool.currentObj)).setAngleExtent(angleExtent);
			    }});
	}
	public String toString(){
		return "[PsArcExpression."+type+"]";
	}
}
