/*
 PsBox.java - November 13, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.format.input.latex.LaTeXPutExpression;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
 
/**
 * Parses PsTricks boxes, i.e. (depending on the "type" parameter of the constructor)
 * <ul>
 * <li> \\psframebox[param]{text}}
 * <li> \\psovalbox[param]{text}}
 * <li> \\pscirclebox[param]{text}}
 * </ul>
 * This expression is to be used as a child expression of either
 * {@link #jpicedt.format.input.latex.LaTeXPutExpression jpicedt.format.input.latex.LaTeXPutExpression} or 
 * {@link #jpicedt.format.input.pstricks.PsRPutExpression PsRPutExpression}.
 * <p>
 * The parsing code makes use of the following Pool's key/value pairs (as set by the enclosing XXXPutExpression beforehands):
 * <ul>
 * <li> KEY_RPUT_VALIGN: \\rput's [bBt] vertical alignment parameter</li>
 * <li> KEY_RPUT_HALIGN: \\rput's [lr] horizontal alignment parameter </li>
 * <li> KEY_RPUT_POINT: \\rput's (x,y) KEY_RPUT_VALIGNparameter </li>
 * <li> KEY_RPUT_ROTATION: \\rput's {rotation} parameter </li>
 * </ul>
 * For each of these four keys, permitted values are the same as in PicAttributeSet. For instance, the value
 * associated with KEY_RPUT_VALIGN may be TEXT_VALIGN_BOTTOM, TEXT_VALIGN_BASELINE, ...
 * <p>
 * @author Sylvain Reynal
 * @since jpicedt 1.4pre2
 * @version $Id: PsBox.java,v 1.3.2.1 2007/09/02 11:56:12 reynal Exp $ 
 */
public class PsBox extends SequenceExpression implements PicObjectConstants, ExpressionConstants {

	/** expects a \\psframebox macro */
	public final static String RECTANGLE_BOX = "\\psframebox";
	/** expects a \\pscirclebox macro */
	public final static String CIRCLE_BOX = "\\pscirclebox";
	/** expects a \\psovalbox macro */
	public final static String OVAL_BOX = "\\psovalbox";
	
	private String type; // e.g. \\psframebox
	private Pool pool;

	/**
	 * @param type RECTANGLE_BOX, CIRCLE_BOX or OVAL_BOX depending on the expected box type. 
	 */
	public PsBox(Pool pl, String type){
		super(true); // throw IncompleteSequence Exception
		this.pool = pl;
		this.type = type;
		// command :
		this.add(new PSTInstanciationExpression(type, new PicText(),pool));
		this.add(new OptionalExpression(new StarExpression(pool))); // "*"
		this.add(WHITE_SPACES_OR_EOL);
		// optional [param]
		this.add(new OptionalExpression(new EnclosingExpression("[",new PSTParametersExpression(pool,Pool.CURRENT_OBJ_ATTRIBUTES),"]"))); // push in object's attributeSet
		this.add(WHITE_SPACES_OR_EOL);
		// {text}
		this.add(new EnclosedText()); // set PicText's text
	}

	public void action(ParserEvent e){
		if (DEBUG) System.out.println(e);
		// set put point:
		PicPoint putPoint = (PicPoint)pool.get(PsRPutExpression.KEY_RPUT_POINT);
		if (putPoint==null) putPoint = (PicPoint)pool.get(LaTeXPutExpression.KEY_PUT_POINT);
		if (putPoint==null) putPoint = new PicPoint();
		((PicText)(pool.currentObj)).setPoint(PicText.P_ANCHOR,putPoint,null);
		// handle alignment:
		String horAlign = (String)pool.get(PsRPutExpression.KEY_RPUT_HALIGN);
		String vertAlign = (String)pool.get(PsRPutExpression.KEY_RPUT_VALIGN);
		if (horAlign!=null) pool.currentObj.setAttribute(TEXT_HOR_ALIGN, horAlign);
		if (vertAlign!=null) pool.currentObj.setAttribute(TEXT_VERT_ALIGN, vertAlign);
		// set box type:
		if (type==RECTANGLE_BOX) ((PicText)(pool.currentObj)).setFrameType(TEXT_BOX_RECTANGLE);
		else if (type==CIRCLE_BOX) ((PicText)(pool.currentObj)).setFrameType(TEXT_BOX_CIRCLE);
		if (type==OVAL_BOX) ((PicText)(pool.currentObj)).setFrameType(TEXT_BOX_OVAL);
		// rotation:
		Double rotation = (Double)pool.get(PsRPutExpression.KEY_RPUT_ROTATION);
		if (rotation!=null) pool.currentObj.setAttribute(TEXT_ROTATION, rotation);
	}

	/**
	 * handles {text} content (for PsBox) by setting PicText's text content, replacing
	 * linefeeds by whitespaces beforehands (as TeX does...)
	 */
	class EnclosedText extends EnclosingExpression {

		public EnclosedText(){
			super("{", null, "}");
		}
		public void action(ParserEvent e){
			if (DEBUG) System.out.println(e);
			String txt = getEnclosedString().replace('\n',' ');
			txt = Context.removeRedundantWhiteSpaces(txt);
			((PicText)(pool.currentObj)).setText(txt);
		}
	}
}


