/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.PicPoint;

/**
 * Parses \\pspolygon, \\psdots and \\psline commands : <ul>
 * <li>\\pspolygon[param](2,4)(3,15) // NO ARROW ALLOWED HERE !<br>
 * <li>\\pspolygon*[param](5,1)(5,8)...(xN,yN)<br>
 * <li>\\psline[param]{arrows}(5,1)(5,8)...(xN,yN)<br>
 * <li>\\psline*[param]{arrows}(5,1)(5,8)...(xN,yN)<br>
 * <li>\\psdots[param,dotstyle=style,dotscale=a b,dotangle=angle]{arrows}(5,1)(5,8)...(xN,yN)<br>
 * <li>\\psdots*[param]{arrows}(5,1)(5,8)...(xN,yN) (same as above, '*' being unused)<br>
 * </ul> 
 * Note : PsPolygon -> close path ; PsLine -> open path ; PsDots -> dots only<br>
 * @author Sylvain Reynal
 * @since jpicedt 1.3.2
 * @version $Id: PsPolygonExpression.java,v 1.5.2.1 2007/09/02 11:56:12 reynal Exp $
 * <p>
 * 
 */
public class PsPolygonExpression extends SequenceExpression  implements PicObjectConstants {

	private Pool pool;
	private String type;
	/** parses a \\pspolygon command */
	public static final String POLYGON = "\\pspolygon";
	/** parses a \\psline command */
	public static final String LINE = "\\psline";
	/** parses a \\psdots command */
	public static final String DOTS = "\\psdots";

	/**
	 * Uses the given type as the default tag
	 */
	public PsPolygonExpression(Pool pl, String type){
		this(pl,type,null);
	}
	
	/**
	 * @param shape POLYGON, LINE or DOTS
	 * @param tag if null, default to shape
	 */
	public PsPolygonExpression(Pool pl, String shape, String tag){

		super(true); // throw IncompleteSequence Exception
		pool = pl;
		this.type = shape;
		if (tag==null) tag = shape;

		PicMultiCurve prototype = new PicMultiCurve();
		if (type==POLYGON) prototype.setClosed(true);
		else  prototype.setClosed(false);

		add(new PSTInstanciationExpression(tag, prototype,pool));
		add(WHITE_SPACES_OR_EOL);
		add(new OptionalExpression(new StarExpression(pool)));
		add(WHITE_SPACES_OR_EOL);
		// add optional param here
		add(new OptionalExpression(new EnclosingExpression("[",new PSTParametersExpression(pool,Pool.CURRENT_OBJ_ATTRIBUTES),"]"))); // push in object's attributeSet
		// add arrows if type==LINE
		if (type==LINE) add(new OptionalExpression(new PSTArrowExpression(pool)));
		// multi-points :
		add(WHITE_SPACES_OR_EOL);
		int minNbPts = (type==DOTS ? 1 : 2);
		add(new RepeatExpression(new SequenceExpression(
		                             WHITE_SPACES_OR_EOL,
		                             new PicPointExpression("(",",",")"){
			                             public void action(ParserEvent e){
				                             if (DEBUG) System.out.println(e);
				                             PicPoint pt = (PicPoint)e.getValue();
				                             ((PicMultiCurve)(pool.currentObj)).addPoint(pt.toMm(pool.getDouble(PstricksParser.KEY_X_UNIT),pool.getDouble(PstricksParser.KEY_Y_UNIT)));
											 if (type!=DOTS) pool.currentObj.setAttribute(POLYDOTS_STYLE,NONE); // force "no dots"
											 // else dot style attributes have been parsed by PSTParametersExpressioin, and will be used
			                             }
		                             }),minNbPts, AT_LEAST));
	}

	public String toString(){
		return "[PsPolygonExpression."+type+"]";
	}
}
