/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

import java.util.*;

import java.io.*;

/**
 * An expression that can encompass a sub-expression it encloses with markers, e.g. "{" + sub-expression + "}"<br>
 * The interpret() methods work as follows :<br>
 * <ul>
 * <li>look up an endMarker matching beginMarker in Context.getRemainingSubstring (that is, skip enclosed blocks with the same markers type)
 * <li>set this endMarker as the new Context's endMarker
 * <li>save enclosed expression as "value", and interpret it
 * <li>restore old Context's endMarker
 * </ul>
 * @since jpicedt 1.3
 * @author Sylvain Reynal
 * @version $Id: EnclosingExpression.java,v 1.4.2.1 2007/09/02 11:56:13 reynal Exp $
 */
public class EnclosingExpression extends AbstractRegularExpression {

	private String value, openingDelimiter, closingDelimiter;
	private AbstractRegularExpression child;
	private boolean noLineFeed;

	/**
	 * @param openingDelimiter the opening delimiter
	 * @param child the Expression that must be parsed inside the delimiters (can be null)
	 * @param closingDelimiter the closing delimiter
	 * @param noLineFeed if true, closingDelimiter must be on the same line as openingDelimiter
	 *
	 * A typical use would be : EnclosingExpression("{", new WildChar(ANY_CHAR),"}",false)
	 */
	public EnclosingExpression(String openingDelimiter, AbstractRegularExpression child, String closingDelimiter, boolean noLineFeed){

		this.child = child;
		this.openingDelimiter=openingDelimiter;
		this.closingDelimiter=closingDelimiter;
		this.noLineFeed = noLineFeed;
	}

	/**
	 * @param openingDelimiter the opening delimiter
	 * @param child the Expression that must be parse inside the delimiter (can be null)
	 * @param closingDelimiter the closing delimiter
	 * Linefeeds are allowed.
	 */
	public EnclosingExpression(String openingDelimiter, AbstractRegularExpression child, String closingDelimiter){
		this(openingDelimiter, child, closingDelimiter, false);
	}

	/**
	 * Change the Expression that must be parsed inside the delimiters 
	 * to the given expression (can be null)
	 */
	public void setChild(AbstractRegularExpression child){
		this.child = child;
	}
	
	/**
	 * Call action() with value=enclosed string 
	 * @return TRUE if delimiters as well as inward expr have been found, 
	 */
	public boolean interpret(Context context) throws REParserException {

		if (!context.matchAndMove(openingDelimiter)) return false;
		context.mark(); // mark pos. just behin openingDelimiter
		
		// look up closingDelimiter:
		int inwardBlocks = 0;
		while(true){
			if (context.matchAndMove(openingDelimiter)) { // <-- enter inner block
				inwardBlocks++;
				continue;
			}
			if (context.matchAndMove(closingDelimiter)) { // <-- exit inner block
				if (inwardBlocks==0){ // ok, this closingDelimiter matches the FIRST openingDelimiter, 
					// i.e. this is the outermost one : first compute new block boundaries, then enter block.
					int pastClosingDelimiter = context.getCaretPosition();
					int newEndBlockMarker = pastClosingDelimiter - closingDelimiter.length();
					context.reset(); // move caret back to block start, to set newBlockStart
					context.enterBlock(newEndBlockMarker);
					// now parse the content of this new block :
					boolean result;
					if (child != null) result = child.interpret(context);
					else result = true;
					value=context.getBlockContent();
					if (result==true) action(new ParserEvent(this, context, true, value));
					// restore old context: (caret now points to "newEndBlockMarker")
					context.exitBlock();
					// move cursor just after closingDelimiter :
					context.moveCaretTo(pastClosingDelimiter);
					return result;
				}
				else { // we are still "inside" the outermost block : go on until we find the outermost closingDelimiter
					inwardBlocks--;
					continue;
				}
			}
			try {
				Character C = context.read(); // move one step forward
				if (C == null) break; // end-of-block (EOF raising a REParserException.EOF exception)
				if (noLineFeed && C.charValue()=='\n') break;
			}
			catch (REParserException.EOF pe){
				break; // mutate EOF exception to BlockMismatch exception
			}
		}
		throw new REParserException.BlockMismatch(context,this);
	}

	/**
	 * @return the enclosed string (= block content as returned by Context)
	 */
	public String getEnclosedString(){
		return value;
	}
	
	/**
	 *
	 */
	public String toString(){

		return "[EnclosingExpression:\"" + openingDelimiter + "\", " + child + ", \"" + closingDelimiter + "\"]";
	}
}
