/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.format.input.util;

import java.util.*;

/** 
 * An expression containing only digits, possibly preceded by whitespaces ; 
 * a post-delimiters can be specified, as well as the number's type (int or double) and its sign
 * @since jpicedt 1.3
 * @author Sylvain Reynal
 * @version $Id: NumericalExpression.java,v 1.4.2.1 2007/09/02 11:56:13 reynal Exp $
 *
 */
public class NumericalExpression extends AbstractRegularExpression {

    private int numberType;
    private int sign;
    private String postfix;
    private boolean swallowPostFix;

    private Number value;

    /**
     * @param numberType INTEGER or DOUBLE
     * @param sign ANY_SIGN, POSITIVE, NEGATIVE, STRICTLY_POSITIVE, STRICTLY_NEGATIVE or NON_NULL
     * @param postfix a String that signals the end of this numerical expression (if NULL, swallows as many digits as possible)
     * @param swallowPostFix if TRUE, move cursor JUST BEHIND the postfix; else move it TO THE POSTFIX POSITION
	 *        N/A if postFix is null.
     */
    public NumericalExpression(int numberType, int sign, String postfix, boolean swallowPostFix){

		this.numberType=numberType;
		this.sign=sign;
		this.postfix=postfix;
		this.swallowPostFix = swallowPostFix;
    }

    /**
	 * Constructor w/o postfix, i.e. parse as many digits as possible.
     * @param numberType INTEGER or DOUBLE
     * @param sign ANY_SIGN, POSITIVE, NEGATIVE, STRICTLY_POSITIVE, STRICTLY_NEGATIVE or NON_NULL
     */
    public NumericalExpression(int numberType, int sign){
		this(numberType, sign, null, false);
    }

    /**
	 * Constructor w/o postfix, type=DOUBLE and sign=ANY_SIGN.
     */
    public NumericalExpression(){
		this(DOUBLE, ANY_SIGN, null, false);
    }

    /**
     * @return TRUE if expr has been found on the current line
     * Calls action() in case of successful parsing, and send a ParserEvent with value=parsed number.
     */
    public boolean interpret(Context context) throws REParserException {

		String str;
		int postFixIndex;
		// if there's a postfix :
		if (postfix != null){
			postFixIndex = context.indexOfBeforeEOL(postfix); // it'd make no sense searching for occurences on the next 
								// line since parseDouble would raise an exception if there's a "\n" in the string
			if (postFixIndex <= 0) return false; // no postfix found or lentgh=0
			str = context.readTo(postFixIndex);
			if (swallowPostFix) context.read(); // read postfix but don't use it
		}
		// no postfix -> swallow as many digits as possible (and postFixIndex = index of the last digit+1), dropping any leading whitespaces
		else {
			StringBuffer buf = new StringBuffer(); // temp. storage for characters
			postFixIndex=0;
			boolean dotFound = false; // "." can be found only once !
			boolean minusFound = false; // "-" can be found only at the beginning !
			boolean digitFound = false; // true as soon as one digit has been found
			while(true){
				Character cc = context.read();
				//System.out.println("Read char = " + cc);
				if (cc == null) break; // EOF or End-of-block
				char c = cc.charValue();
				if (c==' '){
					if (minusFound || digitFound || dotFound) break; // whitespaces must be leading only !
					buf.append(c);
					continue; // skip leading whitespaces
				}
				if (c=='-' || c=='+') {
					if (minusFound || digitFound || dotFound) break; // "-" must precede any occurence of a digit or the dot sign, so any ensuing "-" signals the end of the number
					minusFound = true;
					buf.append(c);
					continue;
				}
				if (c=='.') { // [pending] there's still a bug if "-.5" is parsed ! (it's accepted where it shouldn't)
					if (dotFound) break; // we can have only one ".", so the second one signals the end of the expression
					dotFound = true;
					buf.append(c);
					continue;
				}
				if (Character.isDigit(c)){
					digitFound = true;
					buf.append(c);
					continue;
				}
				context.moveCaretBy(-1); // push back last char read
				break; // any other char (including "\n") signals the end of the expression
			}
			str = buf.toString();
		}
		
		if (str.length() == 0) return false; // empty string
		
		try{
			switch(numberType){
				case INTEGER:
					value = new Integer(str);
					break;
				default:
				case DOUBLE:
					value = new Double(str);
					break;
			}
		}
		catch(NumberFormatException nfe){
			throw new REParserException.NumberFormat(context,this);
		}
		switch(sign){
			case POSITIVE:
				if (value.doubleValue() < 0) throw new REParserException.NumberSign(context,this);
				break;
			case NEGATIVE:
				if (value.doubleValue() > 0) throw new REParserException.NumberSign(context,this);
				break;
			case STRICTLY_POSITIVE:
				if (value.doubleValue() <= 0) throw new REParserException.NumberSign(context,this);
				break;
			case STRICTLY_NEGATIVE:
				if (value.doubleValue() >= 0) throw new REParserException.NumberSign(context,this);
				break;
			case NON_NULL:
				if (value.doubleValue() == 0) throw new REParserException.NumberSign(context,this);
				break;
			default:
			case ANY_SIGN:
				break;
		}
		
		action(new ParserEvent(this, context, true, value));
		return true;
    }

	/**
	 * Return the parsed value, wrapped in a Number (either Integer or Double)
	 */
	public Number getValue(){
		return value;
	}

    /**
     *
     */
    public String toString(){

		String typeStr;
		switch(numberType){
			case INTEGER:
				typeStr="int";
				break;
			default:
			case DOUBLE:
				typeStr="double";
				break;
		}
		String signStr;
		switch(sign){
			case POSITIVE:
				signStr=">=0";
				break;
			case NEGATIVE:
				signStr="<=0";
				break;
			case STRICTLY_POSITIVE:
				signStr=">0";
				break;
			case STRICTLY_NEGATIVE:
				signStr="<0";
				break;
			case NON_NULL:
				signStr="!=0";
				break;
			default:
			case ANY_SIGN:
				signStr="ANY";
				break;
		}

		return "[Numerical:" + typeStr + signStr + " post=" + postfix + " swallow=" + swallowPostFix + "]";
    }
}
