/*
 Pool.java - July 23, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.util;

import jpicedt.graphic.PicPoint;
import jpicedt.graphic.model.*;

import java.awt.Color;
import java.io.*;
import java.util.*;

/**
 * Offers a means for expressions belonging to the parser-tree to share variables across the tree.
 * In addition to storing persistent data in some predefined public fields, 
 * this class also acts as a hashtable, 
 * and may thus store key/value pairs of
 * objects of any class (these may be PicAttributeSet's, etc...), which may be shared across the whole parser-tree.
 * @since jpicedt 1.3
 * @author Sylvain Reynal
 * @version $Id: Pool.java,v 1.5.2.1 2007/09/02 11:56:14 reynal Exp $
 */
public class Pool {
	
	/** either the main drawing, or a subgroup ; each new parsed Element should be added to this group */
	public PicGroup currentGroup;

	/** 
	 * a fifo-like stack used to store the main drawing and its subgroups ; each time a "begin group" is encountered,
	 * the current PicGroup is pushed onto the stack, and a new PicGroup is instanciated, which then
	 * represents the current PicGroup ; the opposite operations are executed in the reverse order
	 * when a "end group" is found. */
	public Stack picGroupStack;

	/**
	 * Convenience used to share information (e.g. parameters, location, ...) across expressions acting
	 * on the same element. Generally, an InstanciationExpression reinits "currentObj" to an instance of
	 * an Element of the proper type, then ensuing expression modify this element's attributes and/or
	 * geometry.
	 */
	public Element currentObj;

	/** a string that stores commands that couldn't be parsed in the current context (e.g. in the text file to be parsed)
	  * so that we can keep track of these commands (e.g. to be able to include them again when it comes to saving
	  * the text file to disk) */
	public StringBuffer notParsed;
	
	/**
	 * Key used to fetch the attribute set attached to the current element "currentObj"
	 * @see #get(String)
	 */
	public static final Key CURRENT_OBJ_ATTRIBUTES = new Key("current_obj_attributes");

	
	private HashMap map = new HashMap();
	
	/**
	 * Simply call reinit()
	 */
	public Pool(){
		reinit();
	}
	
	/**
	 * Reset public fields values.
	 */
	public void reinit(){
		//System.out.println("Reinit' Pool...");
		currentGroup = new PicGroup();
		picGroupStack = new Stack();
		notParsed = new StringBuffer();
		currentObj=null;
		map = new HashMap();
	}
	
	/**
	 * Associates the specified value with the given key in the pool's hashmap
	 * @return old value for the given key, if any.
	 */
	public Object put(Key key, Object value){
		//jpicedt.Log.debug("key="+key+", value="+value);
		return map.put(key,value);
	}
	
	/**
	 * Return the value associated with the given key
	 */
	public Object get(Key key){
		return map.get(key);
	}
	
	/**
	 * Convenience for retrieving a "double" value from the pool's hashmap
	 */
	public double getDouble(Key key){
		return ((Double)get(key)).doubleValue();
	}
	
	/**
	 * Convenience for storing a "double" value into the pool's hashmap
	 * @return the old value, if any, for convenience
	 */
	public double put(Key key, double d){
		return ((Double)put(key, new Double(d))).doubleValue();
	}
	
	/**
	 * Convenience for retrieving a "PicAttributeSet" value ; can be used in conjunction with
	 * static methods defined in jpicedt.graphic.model.StyleConstants to retrieve 
	 * attributes values with less burden.
	 * @param key if CURRENT_OBJ_ATTRIBUTES, retrieves the attribute set of the current Element in the pool ;
	 *        any other key is acceptable as long as it corresponds to a valid entry in the Pool's hashtable,
	 *        that is, it has been previously pushed in the Pool using <code>put(key, value)</code>.
	 */
	public PicAttributeSet getAttributeSet(Key key){
		if (key == CURRENT_OBJ_ATTRIBUTES){
			if (currentObj != null){
				return currentObj.getAttributeSet();
			}
			else return null;
		}
		return (PicAttributeSet)get(key); // may raise a ClassCastException if wrong key !!! 
	}

	/**
	 * Add a name/value attribute pair to the attribute set with the given key, i.e.
	 * either CURRENT_OBJ_ATTRIBUTES for "currentObj", or any other key refering to
	 * an attribute set previously pushed in the Pool (e.g. PsTricks registers,...)
	 */
	public void setAttribute(Key key, PicAttributeName name, Object value){
		PicAttributeSet set = getAttributeSet(key);
		if (set != null) set.setAttribute(name,value);
	}
	
	/**
	 * @return a text description of this Pool for debugging purpose
	 */
	public String toString(){
		String s = "Pool@" + hashCode() + ":\n";
		s += "* currentGroup=" + currentGroup
			 + "\n* picGroupStack=" + picGroupStack
			 + "\n* currentObj=" + currentObj
			 + "\n* notParsed=[" + notParsed + "]\n";
		for(Iterator it = map.keySet().iterator(); it.hasNext();){
			Object k = it.next();
			s += "* " + k + ":" + map.get(k) + "\n";
		}
		return s;
	}
	
	/**
	 * Enforces use of strong typing for keys being pushed in the map .
	 */
	public static class Key {
		private String str;
		/** @param str may be chosen with debugging in view (not used otherwise) */
		public Key(String str){
			this.str = str;
		}
		public String toString(){
			return str;
		}
	}
	
}

