/*
 PicEllipseFormatter.java - August 30, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.output.eepic;

import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.PicPoint;

import java.awt.*;

/**
 * Formats a PicEllipse using macros of the eepic package.
 * [SR:pending] handle rotated ellipses
 * @author Sylvain Reynal
 * @since jpicedt 1.3.2
 * @version $Id: PicEllipseFormatter.java,v 1.7.2.1 2007/09/02 11:55:13 reynal Exp $
 *
 */
public class PicEllipseFormatter extends jpicedt.format.output.latex.PicEllipseFormatter  {

	// inherited : ellipse, factory
	
	public PicEllipseFormatter(PicEllipse ellipse,EepicFormatter factory){
		super(ellipse,factory); // init width and height
	}

	/**
	 * Returns a String representing this ellipse or arc using epic/eepic macros
	 */
	public String format(){

		StringBuffer buf = new StringBuffer(100); // 100 as initial capacity seems to be a good guess
		PicAttributeSet set = ellipse.getAttributeSet();

		factory.appendThicknessString(buf,ellipse);

		// plain circle/ellipse : 
		if (ellipse.isPlain()){
			if (ellipse.isRotated()){ // emulated lines, not filled
				super.appendEmulatedArcString(buf);
			}
			else { // \\[filltype]\\ellipse{x-diam}{y-diam}
				buf.append("\\put");
				buf.append(ellipse.getPoint(PicEllipse.P_CENTER,null));
				buf.append("{");
				if (StyleConstants.getFillStyle(set)==SOLID){
					Color fillColor = StyleConstants.getFillColor(set);
					if (fillColor.equals(Color.WHITE)) buf.append("\\whiten");
					else if (fillColor.equals(Color.BLACK)) buf.append("\\blacken");
					else buf.append("\\shade"); // SHADE
				}
				buf.append("\\ellipse{");
				double width = ellipse.getPoint(PicEllipse.P_BL,null).distance(ellipse.getPoint(PicEllipse.P_BR,null));
				buf.append(PEToolKit.doubleToString(width));
				buf.append("}{");
				double height = ellipse.getPoint(PicEllipse.P_BR,null).distance(ellipse.getPoint(PicEllipse.P_TR,null));
				buf.append(PEToolKit.doubleToString(height));
				buf.append("}}");buf.append(CR_LF);
			}
		}

		// an arc
		else {
			// circular arc (can be filled) : [\\filltype]\\arc{diameter}{start angle}{end angle}
			// where : 0 < start-angle < 2 PI ; start-angle < end-angle < start-angle + 2 PI
			// and angles are counted CW and in radians.
			// Note that stroke is always painted black
			if (ellipse.isCircular()){
				// we've got to construct specific startAngleStr and endAngleStr strings
				// since eepic requests CLOCKWISE angles, in RADIAN units, 
				// while we store this information COUNTERCLOCKWISE and in DEGREES in PicEllipse
				buf.append("\\put");
				buf.append(ellipse.getPoint(PicEllipse.P_CENTER,null));
				buf.append("{");
				if (StyleConstants.getFillStyle(set)==SOLID){
					Color fillColor = StyleConstants.getFillColor(set);
					if (fillColor.equals(Color.WHITE)) buf.append("\\whiten");
					else if (fillColor.equals(Color.BLACK)) buf.append("\\blacken");
					else buf.append("\\shade"); // SHADE
				}
				buf.append("\\arc{"); // ! OPEN arcs only ! (and aka CHORD is filled)
				buf.append(PEToolKit.doubleToString(ellipse.getGreatAxisLength()));
				buf.append("}{");
				double startAngle, endAngle;
				if (ellipse.getSmallAxisLength()<0){
					startAngle = ellipse.getRotatedAngleStart()%360; // guaranteed to lie b/w -360 and 360
					endAngle = ellipse.getRotatedAngleEnd()%360;
				}
				else {
					startAngle = -ellipse.getRotatedAngleEnd()%360;
					endAngle = -ellipse.getRotatedAngleStart()%360;
				}
				if (startAngle < 0) startAngle += 360;
				while (endAngle < startAngle) endAngle += 360;
				double startAngleRad = Math.toRadians(startAngle);
				double endAngleRad = Math.toRadians(endAngle);
				buf.append(PEToolKit.doubleToString(startAngleRad));
				buf.append("}{");
				buf.append(PEToolKit.doubleToString(endAngleRad));
				buf.append("}}");buf.append(CR_LF);
				
				// since filled \\arc = CHORD, we must add a white polygon to emulate a pie if angle-extend > 180
				// or a properly filled polygon if angle-extent < 180
				if (ellipse.getArcType()==PicEllipse.PIE){
					if (StyleConstants.getFillStyle(set)==SOLID){
						double angleExtent = endAngle - startAngle; // guaranteed to be positive, in DEGREES
						if (angleExtent != 180){ // add a filled path
							if (angleExtent < 180){
								Color fillColor = StyleConstants.getFillColor(set);
								if (fillColor.equals(Color.WHITE)) buf.append("\\whiten");
								else if (fillColor.equals(Color.BLACK)) buf.append("\\blacken");
								else buf.append("\\shade"); // SHADE
							}
							else buf.append("\\whiten"); // assume bgColor=white
						}
						// else do nothing
					}
					buf.append("\\path");
					buf.append(ellipse.getPoint(PicEllipse.P_ANGLE_START,null));
					buf.append(ellipse.getPoint(PicEllipse.P_CENTER,null));
					buf.append(ellipse.getPoint(PicEllipse.P_ANGLE_END,null));
					buf.append(CR_LF);
				}

					
				// append chord or pie line if applicable 
				else if (ellipse.getArcType()==PicEllipse.CHORD){
					buf.append("\\path");
					buf.append(ellipse.getPoint(PicEllipse.P_ANGLE_START,null));
					buf.append(ellipse.getPoint(PicEllipse.P_ANGLE_END,null));
					buf.append(CR_LF);
				}
				// else OPEN: leave it as it is
			}
			// ELLIPTIC arc (CAN'T be filled) and we emulate it
			else super.appendEmulatedArcString(buf);
			// ARROW in the ARC case :
			appendArrows(buf);
		}
		buf.append(CR_LF);
		return buf.toString();
	}
}
