/*
 AbstractCurveFormatter.java - August 29, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package jpicedt.format.output.latex;

import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.PicPoint;

import java.awt.*;

/**
 * Format an AbstractCurve, or any subclass of it, to the LaTeX-picture format.
 * [SR:pending] synchronize LatexViewFactory with new (un)supported attributes.
 * @author Sylvain Reynal
 * @since jpicedt 1.4
 * @version $Id: AbstractCurveFormatter.java,v 1.4.2.1 2007/09/02 11:55:13 reynal Exp $
 * <p>
 *
 */
public class AbstractCurveFormatter implements Formatter, PicObjectConstants, LatexConstants {

	/** the Element this formater acts upon */
	protected AbstractCurve curve;
	protected LatexFormatter factory;

	/** */
	public AbstractCurveFormatter(AbstractCurve curve, LatexFormatter factory){
		this.curve = curve;
		this.factory=factory;;
	}

	/**
	 * Returns a string representing this AbstractCurve using \\line and \\qbezier LaTeX macros only.
	 * Dashed style is allowed if all segments are straight, since \\qbezier does not support this style.
	 */
	public String format(){

		StringBuffer buf = new StringBuffer(100); // 100 char as initial capacity seems a good guess
		PicAttributeSet set = curve.getAttributeSet();

		factory.appendThicknessString(buf,curve);
		
		// check if there are only straight segments, in which case we can afford dashed lines ; else only
		// solid style is supported by the  \\qbezier command :
		double dash=(StyleConstants.getDashOpaque(set)+StyleConstants.getDashTransparent(set))/2.0;
		if (StyleConstants.getLineStyle(set)!=DASHED || !curve.isPolygon()) dash=0;// [pending] add support for black/white dash + dotted
		
		// for each segment :
		// - straight : add \\line command
		// - curved : add \\qbezier command, after converting cubic to two quads (see PEToolKit)
		// If curve is open, first and last segment deserve special processing as for arrows
		for (int segIdx=0; segIdx < curve.getNumberOfSegments(); segIdx++){
			PicPoint p1 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SUBDIVISION_POINT),null);
			PicPoint p2 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx+1, curve.SUBDIVISION_POINT),null);
			if (curve.isStraight(segIdx)){
				Arrow arrowL=Arrow.NONE;
				Arrow arrowR=Arrow.NONE;
				if (!curve.isClosed()){
					if (segIdx==0) arrowL = StyleConstants.getLeftArrow(set);
					if (segIdx==curve.getNumberOfSegments()-1) arrowR = StyleConstants.getRightArrow(set);
				}
				buf.append(factory.lineToLatexString(p1.x,p1.y,p2.x,p2.y,arrowL,arrowR,dash));
				buf.append(CR_LF);
			}
			else {
				PicPoint pCtrl1 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.FIRST_SEGMENT_CONTROL_POINT),null);
				PicPoint pCtrl2 = curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx, curve.SECOND_SEGMENT_CONTROL_POINT),null);
				PicPoint[] quadsPts = PEToolKit.convertCubicBezierToQuad(p1,pCtrl1,pCtrl2,p2); // five points = two joined quads
				buf.append("\\qbezier"); // first quad
				buf.append(quadsPts[0]);buf.append(quadsPts[1]);buf.append(quadsPts[2]);
				buf.append(CR_LF);
				buf.append("\\qbezier"); // second quad
				buf.append(quadsPts[2]);buf.append(quadsPts[3]);buf.append(quadsPts[4]);
				buf.append(CR_LF);
				if (!curve.isClosed()){
					if (segIdx==0 && StyleConstants.getLeftArrow(set) != Arrow.NONE){
						PicVector dir = new PicVector(pCtrl1,p1).normalize();
						buf.append(factory.arrowToLatexString(p1, dir));
						buf.append(CR_LF);
					}
					if (segIdx==curve.getNumberOfSegments()-1 && StyleConstants.getRightArrow(set) != Arrow.NONE){
						PicVector dir = new PicVector(pCtrl2,p2).normalize();
						buf.append(factory.arrowToLatexString(p2, dir));
						buf.append(CR_LF);
					} 
				}
			}
		}
		return buf.toString();
	}
	
}

 
