/*
 *  AbstractCurveFormatter.java - December 25, 2003 - jPicEdt, a picture editor for LaTeX.
 *  Copyright (C) 1999-2007 Sylvain Reynal
 *
 *  Departement de Physique
 *  Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 *  6, avenue du Ponceau
 *  F-95014 CERGY CEDEX
 *
 *  Tel : +33 130 736 245
 *  Fax : +33 130 736 667
 *  e-mail : reynal@ensea.fr
 *  jPicEdt web page : http://www.jpicedt.org
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package jpicedt.format.output.pstricks;

import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.*;
import java.awt.*;

/**
 * Format an AbstractCurve, or any subclass of it, using PsTricks macros. 
 * These are "rendered" to PsTricks using a <code>\\pscustom</code> command that lumps
 * several Bezier curves together in a single Pstricks object.
 *
 * @author    Vincent Guirardel, Sylvain Reynal
 * @since jpicedt 1.4
 * @version   $Id: AbstractCurveFormatter.java,v 1.14.2.1 2007/09/02 11:55:13 reynal Exp $
 */
public class AbstractCurveFormatter implements Formatter, PicObjectConstants, PstricksConstants {

	/**
	 * the Element this formatter acts upon
	 */
	protected AbstractCurve curve;
	/** the producing factory */
	protected PstricksFormatter factory;


	public AbstractCurveFormatter(AbstractCurve curve, PstricksFormatter factory) {
		this.curve = curve;
		this.factory = factory;
	}


	/**
	 * Return a String representing this curve in the PsTricks format.
	 * This delegates to either {@link #formatPsCustom formatPsCustom()}
	 * or {@link #formatPsDots formatPsDots()} depending on the value of POLYDOTS_STYLE attribute.
	 */
	public String format() {
		boolean isPsDots = (StyleConstants.getPolydotsStyle(curve.getAttributeSet())!=NONE);
		boolean isSuperimposed = StyleConstants.isPolydotsSuperimposed(curve.getAttributeSet());
		StringBuffer buf = new StringBuffer(100);
		if (isPsDots) {
			if (isSuperimposed) buf.append(formatCurve());
			buf.append(formatPsDots());
			return buf.toString();
		}
		else return formatCurve();
	}

	/**
	 * this is a convenience call to either formatPsCustom() or formatPsBezier() or
	 * formatPsLine().
	 */

	protected String formatCurve(){
		if (curve.isPolygon()) return formatPsLine(); // one or more straight segments exclusively
		if (curve instanceof PicPsCurve) return formatPsCurve();
		else if (curve.getNumberOfSegments()>1) return formatPsCustom(); // more than one segment
		else return formatPsBezier(); // only one segment, not straight
	}



	/**
	 * If the curve contains a single Bezier segment: \\psbezier... <br>
	 */
	protected String formatPsBezier() {

		StringBuffer buf = new StringBuffer(100);// 100 as initial capacity seems to be a good guess
		PicAttributeSet set = curve.getAttributeSet();

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(curve);
		if (paramStr.isDefinedColourString()) buf.append(paramStr.getUserDefinedColourBuffer());

		buf.append("\\psbezier");
		// parameters
		if (paramStr.getParameterBuffer().length()>0){
			buf.append("[");
			buf.append(paramStr.getParameterBuffer());
			buf.append("]");
		}

		// arrows
		if (!curve.isClosed()) buf.append(PstricksUtilities.createPstricksStringFromArrows(curve));

		// points
		buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(0,curve.SUBDIVISION_POINT), null));
		buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(0,curve.FIRST_SEGMENT_CONTROL_POINT), null));
		buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(0,curve.SECOND_SEGMENT_CONTROL_POINT), null));
		buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(1,curve.SUBDIVISION_POINT), null));
		buf.append(CR_LF);
		return buf.toString();
	}

	/**
	 * Polygon: \\psline...
	 */
	protected String formatPsLine() {

		StringBuffer buf = new StringBuffer(100);// 100 as initial capacity seems to be a good guess
		PicAttributeSet set = curve.getAttributeSet();

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(curve);
		if (paramStr.isDefinedColourString()) buf.append(paramStr.getUserDefinedColourBuffer());

		if (curve.isClosed()) buf.append("\\pspolygon");
		else buf.append("\\psline");

		// parameters
		if (paramStr.getParameterBuffer().length()>0){
			buf.append("[");
			buf.append(paramStr.getParameterBuffer());
			buf.append("]");
		}

		if (!curve.isClosed()) buf.append(PstricksUtilities.createPstricksStringFromArrows(curve));

		for (int i = 0; i < curve.getNumberOfSegments(); i++) {
			buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i,curve.SUBDIVISION_POINT), null));
			// only add second point if this is the last segment (thereby avoiding redundancy, since writing each
			// point twice was also reported to create some mess when in conjunction with a "linearc" PSTricks parameter)
			if ( i == curve.getNumberOfSegments() - 1 ) { // JHf
            			buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i+1,curve.SUBDIVISION_POINT), null)); // JHf
        		} // JHf
			buf.append(CR_LF);
		}
		return buf.toString();
	}


	/**
	*  Format for pscurve
	* @author Vincent
	* @since jPicEdt 1.4pre5
	*/
	protected String formatPsCurve(){

		PicPsCurve pscurve=(PicPsCurve)curve;
		StringBuffer buf = new StringBuffer(100);// 100 as initial capacity seems to be a good guess
		PicAttributeSet set = pscurve.getAttributeSet();

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(pscurve);
		if (paramStr.isDefinedColourString()) buf.append(paramStr.getUserDefinedColourBuffer());

		if (pscurve.isClosed()) buf.append("\\psccurve");
		else buf.append("\\psecurve");

		// parameters
		if (paramStr.getParameterBuffer().length()>0){
			buf.append("[");
			buf.append(paramStr.getParameterBuffer());
			// curvatures: (bug fix 01/08/2006)
			if (paramStr.getParameterBuffer().length()>0){
				buf.append(',');
				buf.append("curvature=");
				double[] curvatures = pscurve.getCurvatures();
				buf.append(curvatures[0]);
				buf.append(' ');
				buf.append(curvatures[1]);
				buf.append(' ');
				buf.append(curvatures[2]);
			}
			buf.append("]");
		}

		if (!pscurve.isClosed()) buf.append(PstricksUtilities.createPstricksStringFromArrows(pscurve));


		for (int i = 0; i <= pscurve.getLastPointIndex(); i++) {
			buf.append(pscurve.getPoint(i,null));
		}
		buf.append(CR_LF);
		return buf.toString();
	}




	/**
	 * If the curve has more than one segment, and it's not a polygon.
	 * \\pscustom[param]{arrows} <br>
	 * \\psline... or  \\psbezier... <br>
	 * \\closepath <br>
	 */
	protected String formatPsCustom() {

		StringBuffer buf = new StringBuffer(100);// 100 as initial capacity seems to be a good guess
		PicAttributeSet set = curve.getAttributeSet();

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(curve);
		if (paramStr.isDefinedColourString()) buf.append(paramStr.getUserDefinedColourBuffer());

		buf.append("\\pscustom");
		// parameters
		buf.append("[");
		buf.append(paramStr.getParameterBuffer()); // [VG:pending] we should strip redundant border-related parameters from this buffer
		buf.append("]{");

		Arrow leftArrow = StyleConstants.getLeftArrow(set);
		Arrow rightArrow = StyleConstants.getRightArrow(set);
		boolean isAtLeastOneArrow = (leftArrow!=Arrow.NONE || rightArrow!=Arrow.NONE);

		for (int i = 0; i < curve.getNumberOfSegments(); i++) {
			if (curve.isStraight(i)) buf.append("\\psline");
			// else buf.append("\\psbezier[liftpen=2]"); // SR: 08/22/2007 added [liftpen=2] (BUG 1769858)
			else buf.append("\\psbezier"); // SR: 10/21/2007 liftpen=2 does not work flawlessly, there are still pbs when filling a complexe shape.
			// -> better workaround is to remove every leading point except for the first bezier

			// arrows if at least one arrow is non-NONE, curve is open, and this is either the first or the last segment...
			boolean allowArrow = (isAtLeastOneArrow && !curve.isClosed() && curve.getNumberOfSegments() >= 1 && (i == 0 || i == curve.getNumberOfSegments() - 1));
			if (allowArrow) {
				buf.append("{");
				if (i == 0) buf.append(PstricksUtilities.toPstricksString(leftArrow, false));// e.g. "<" for ARROW_HEAD
				buf.append("-");
				if (i == curve.getNumberOfSegments() - 1)
					buf.append(PstricksUtilities.toPstricksString(rightArrow, true));// e.g. ">" for ARROW_HEAD (this way, we really get "<->" instead of ">->" which was wrong !)
				buf.append("}");
			}

			// points
			if (curve.isStraight(i)) { // line
				buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i,curve.SUBDIVISION_POINT), null));
				buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i+1,curve.SUBDIVISION_POINT), null));
			}
			else { // cubic
				if (i == 0) // SR: 10/21/2007 (BUG 1769858)
					buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i,curve.SUBDIVISION_POINT), null));
				buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i,curve.FIRST_SEGMENT_CONTROL_POINT), null));
				buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i,curve.SECOND_SEGMENT_CONTROL_POINT), null));
				buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(i+1,curve.SUBDIVISION_POINT), null));
			}
			buf.append(CR_LF);
		}// end of ``for'' loop
		if (curve.isClosed()) buf.append("\\closepath");

		// pscustom does not draw the shadow by itself...
		if (StyleConstants.isShadow(set)){ // if we have to draw a shadow
			if (curve.isClosed()) buf.append("\\closedshadow ");
			else buf.append("\\openshadow ");
		}

		if (StyleConstants.isOverStrike(set)){ // if we have to draw a border
			//first, get parameters
			PstricksFormatter.ParameterString paramStrForBorder = createParameterStringForBorderInPsCustom(curve);
			buf.append("\\stroke[");
			buf.append(paramStrForBorder.getParameterBuffer());
			buf.append("]");
		}// end of border

		buf.append("}"); // end of pscustom
		buf.append(CR_LF);
		return buf.toString();
	}

	/**
	 * Helper method for formatPSCustom().
	 * Returns PsTricks's special parameters for handling the border in a \\pscustom command
	 * @return the total width of the border to be struck underneath the real path.
	 * Some day, should also return a color. So far color=white. 
	 * @since jpicedt 1.4pre3
	 * @author V.G. (moved from PsTricksFormatter by S.R. 01-08-06)
	 */
	private PstricksFormatter.ParameterString createParameterStringForBorderInPsCustom(Element obj){

		PicAttributeSet attributes = obj.getAttributeSet();

		// only useful if border > 0
		if (! StyleConstants.isOverStrike(attributes)){
			return null;
		}

		StringBuffer paramBuf = new StringBuffer(100);
		StringBuffer userDefinedColorBuffer=null; //  [VG:pending] unused so far

		// linewidth equals  the linewidth plus twice the border.
		double lineWidth = StyleConstants.getLineWidth(attributes);
		double border = StyleConstants.getOverStrikeWidth(attributes);
		double totalWidth = lineWidth + 2*border;

		//if (paramBuf.length()>0) paramBuf.append(','); // not necessary since we're building the buffer from scratch
		paramBuf.append("linewidth=");
		paramBuf.append(PEToolKit.doubleToString(totalWidth));
		paramBuf.append(",linecolor=white");

		PstricksFormatter.ParameterString ps = factory.createParameterString(paramBuf, userDefinedColorBuffer);
		return ps;
	}


	/**
	 * \\psdots(x1,y1)...(xn,yn)
	 */
	protected String formatPsDots() {

		StringBuffer buf = new StringBuffer(100);// 100 as initial capacity seems to be a good guess
		PicAttributeSet set = curve.getAttributeSet();
		PicAttributeSet defaultAttributes = new PicAttributeSet();

		// first handle possibly user-defined colours
		PstricksFormatter.ParameterString paramStr = factory.createParameterString(curve);
		if (paramStr.isDefinedColourString())
			buf.append(paramStr.getUserDefinedColourBuffer());

		buf.append("\\psdots");

		// parameters
		buf.append("[");
		buf.append(paramStr.getParameterBuffer());

		// add psdots specific parameters : we only add non-default values to save some space (and possibly some memory shortage on some TeX installs)
		StringBuffer psdotsBuffer = new StringBuffer();

		double x =  StyleConstants.getPolydotsSizeMinimumMm(set);
		double y = StyleConstants.getPolydotsSizeLineWidthScale(set);
		if (x!=StyleConstants.getPolydotsSizeMinimumMm(defaultAttributes) || y!=StyleConstants.getPolydotsSizeLineWidthScale(defaultAttributes)){
			psdotsBuffer.append("dotsize=");
			psdotsBuffer.append(PEToolKit.doubleToString(x));
			psdotsBuffer.append(' ');
			psdotsBuffer.append(PEToolKit.doubleToString(y));
		}

		String dotStyle = StyleConstants.getPolydotsStyle(set);
		if (!dotStyle.equals(POLYDOTS_DISK)){
			if (psdotsBuffer.length()>0) psdotsBuffer.append(',');
			psdotsBuffer.append("dotstyle=");
			if (dotStyle==POLYDOTS_CIRCLE) psdotsBuffer.append('o');
			else if (dotStyle==POLYDOTS_PLUS) psdotsBuffer.append('+');
			else if (dotStyle==POLYDOTS_TRIANGLE) psdotsBuffer.append("triangle");
			else if (dotStyle==POLYDOTS_TRIANGLE_FILLED) psdotsBuffer.append("triangle*");
			else if (dotStyle==POLYDOTS_SQUARE) psdotsBuffer.append("square");
			else if (dotStyle==POLYDOTS_SQUARE_FILLED) psdotsBuffer.append("square*");
			else if (dotStyle==POLYDOTS_PENTAGON) psdotsBuffer.append("pentagon");
			else if (dotStyle==POLYDOTS_PENTAGON_FILLED) psdotsBuffer.append("pentagon*");
			else psdotsBuffer.append('*'); // default, should never happen
		}

		x = StyleConstants.getPolydotsHorizontalScale(set);
		y = StyleConstants.getPolydotsVerticalScale(set);
		if (x!=StyleConstants.getPolydotsHorizontalScale(defaultAttributes) || y!=StyleConstants.getPolydotsVerticalScale(defaultAttributes)){
			if (psdotsBuffer.length()>0) psdotsBuffer.append(',');
			psdotsBuffer.append("dotscale=");
			psdotsBuffer.append(PEToolKit.doubleToString(x));
			psdotsBuffer.append(' ');
			psdotsBuffer.append(PEToolKit.doubleToString(y));
		}

		x = StyleConstants.getPolydotsAngle(set);
		if (x!=StyleConstants.getPolydotsAngle(defaultAttributes)){
			if (psdotsBuffer.length()>0) psdotsBuffer.append(',');
			psdotsBuffer.append("dotangle=");
			psdotsBuffer.append(PEToolKit.doubleToString(x));
		}

		if (psdotsBuffer.length()>0){
			if (paramStr.getParameterBuffer().length()>0) buf.append(',');
			buf.append(psdotsBuffer);
		}

		// close parameters string :
		buf.append(']');

		// no arrow for \\psdots

		// append coordinates :
		int segIdx;
		for (segIdx = 0; segIdx < curve.getNumberOfSegments(); segIdx++) {
			buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx,curve.SUBDIVISION_POINT), null));
			if (segIdx%10==0) buf.append(CR_LF); // hard wrap
		}
		if (!curve.isClosed())
			buf.append(curve.getSpecificationPoint(curve.segmentToPointIndex(segIdx,curve.SUBDIVISION_POINT), null));
		buf.append(CR_LF);
		return buf.toString();
	}
}

