/*
 PECanvas.java - 1999 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic;

import jpicedt.graphic.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.grid.Grid;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.io.parser.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;
import java.beans.*;
import java.awt.datatransfer.*;
import javax.swing.undo.*;

/**
 * This is a JComponent on which graphic elements are drawn. It's has an underlying model (a Drawing) to
 * represent the content, an EditorKit to manipulate the content, and a View responsible for rendering
 * the content. EditorKit, Drawing, and View's are pluggable : the EditorKit is responsible for creating
 * 1) a default Drawing and 2) a ViewFactory that will populate the View tree associated with the Drawing, 
 * by attaching a View to each element in the model.<p>
 * Depending on the content type this Component is loaded with, it may plug a new EditorKit on-the-fly 
 * that's suited for the given content type (e.g. LaTeX, Postscript, SVG-XML, etc...).
 *  
 *
 * <h1>Model to View/View to Model</h1>
 * <p>
 * Graphic objects are stored (e.g. in a Drawing) 
 * in natural coordinates, ie LaTeX/Postscript/... coordinates, using e.g. a "1 mm" unitlength : 
 * that's what we call "model coordinate". Obviously, we've to translate these coordinates to
 * screen coordinates (e.g. JViewport-coordinate or JPanel-coordinate) before rendering, 
 * and this is done very simply by using an AffineTransform and adding it to the
 * current Graphic2D context in the body of the "paintComponent" method. The following picture
 * sums up the translation process that takes place b/w model- and view-coordinate.
 * </p>
 * <p><img src="doc-files/model2view.png"></p>
 * <p>The benefits of such an approach is to make it easy for someone willing to develop a parser/formater
 * to handle objects coordinates with the fewest possible overhead, since objects coordinates 
 * are "natively" available in natural (ie from left to right and from bottom to top) coordinates, and
 * this is perfectly suited for formatting language like LaTeX, Postscript or SVG-XML. Besides,
 * this makes sense with the grid ticks marks. 
 * </p>
 * <p>Margins are encapsulated in a PageFormat (an inner class) object. Contrary to previous jpicedt releases,
 * it's no longer necessary, as of jpicedt 1.3.2, to provide formater methods with the ptOrg parameter.
 * <p>In addition to the standard behaviour inherited from JPanel, PropertyChangeEvent's are triggered when :
 * <ul>
 * <li>changing the page format ;
 * <li>changing the zoom factor ;
 * <li>changing the editor kit ;
 * </ul>
 * </p>
 * @since jpicedt 1.0
 * @author Sylvain Reynal
 * @version $Id: PECanvas.java,v 1.24.2.2 2007/09/02 11:56:16 reynal Exp $ <p>
 */

public class PECanvas extends JPanel implements Scrollable {

	//////////////////////////////////// PUBLIC FIELDS /////////////////////////////////

	public static final String[] PREDEFINED_ZOOM_STRINGS = {"100%", "200%", "400%", "800%"}; // [pending] use java.text.NumberFormat.getPercentInstance()
	public static final double[] PREDEFINED_ZOOMS =     { 1.0,   2.0,  4.0,   8.0};
	public static final double ZOOM_DEFAULT = 1.0;

	/** key for Properties's zoom value */
	public static final String KEY_ZOOM = "canvas.zoom";
	/** key for Properties's content-type value */
	public static final String KEY_CONTENT_TYPE = "canvas.content-type";
	/** key for Properties's nb of undoable steps value */
	public static final String KEY_UNDOABLE_STEPS = "canvas.max-undoable-steps";
	/** default undoable events to remember */
	public static final int MAX_UNDOABLE_STEPS_DEFAULT = 100; // same as UndoManager
	

	/** property name for drawing-change events */
	public static final String DRAWING_CHANGE = "drawing-change";

	/** property name for editorkit-change events */
	public static final String EDITOR_KIT_CHANGE = "editor-kit-change";
	
	/** property name for content-type's change events */
	public static final String CONTENT_TYPE_CHANGE = "content-type-change";

	//////////////////////////////////// PROTECTED FIELDS /////////////////////////////////
	
	/** the model for this canvas */
	protected Drawing drawing;

	/** pageFormat encapsulates board size and margin data */
	protected PageFormat pageFormat;
	
	/** the current content-type for this PECanvas (determines the EditorKit behaviour) */
	protected ContentType contentType;

	/** the AffineTransform used to translate from model-coordinates to view-coordinates ;
	 *  gets updated each time either the zoom factor or the page format changes */
	protected AffineTransform model2ViewTransform;

	/** the AffineTransform used to translate from mouse-coordinates to model-coordinates ;
	 *  gets updated each time model2ViewTransform changes */
	protected AffineTransform view2ModelTransform;

	/** the grid attached to this canvas */
	protected Grid grid;

	/** the current editor kit for this component */
	protected EditorKit kit;
	
	/** the UndoableEditSupport delegate for UndoableEditEvent firing */
	protected UndoableEditSupport undoableEditSupport;
	
	/** the UndoManager delegate for undo/redo operation */
	protected UndoManager undoManager;
	
	/** the UndoableEdit in progress */
	protected StateEdit stateEdit;
	
	/** a Map storing RenderingHints to be applied to the graphic context when rendering the drawing */
	protected RenderingHints renderingHints = new RenderingHints(null);


	//////////////////////////////////// PRIVATE FIELDS /////////////////////////////////

	/** the zoom factor ; this determines the zoom factor to be applied to Graphics2D through an AffineTransform */
	private double zoom;
	
	/** the total scale factor, including DPMM and zoom ; updated each time zoom changes */
	private double scale;

	/** chained list of PEMouseInputListener's for this component */
	private PEMouseInputListener mouseInputListener;

	/** tmp. buffer used by processMouseEvent */
	private PicPoint peMousePoint = new PicPoint();
	
	/** tmp. buffer used by repaintFromModelRect */
	private double[] tmpCoords = new double[2];
	private Rectangle repaintRectangle = new Rectangle();

	
	
	//////////////////////////////////// CONSTRUCTORS /////////////////////////////////

	
	/**
	 * Construct a new PECanvas initialized with default values.
	 */
	public PECanvas(){
		this(ZOOM_DEFAULT, new PageFormat(), new Grid(), null);
	}

	/**
	 * Construct a new PECanvas with the default editor-kit and drawing as content storage.
	 * @param zoom initial zoom factor
	 * @param initial page format (page size + page margins)
	 * @param content-type (e.g. LaTeX, PsTricks,...) ; this will determine the EditorKit for
	 *        editing the Drawing, and indirectly the ViewFactory that produces View's for the drawing,
	 *        (since the ViewFactory is obtained through the currently installed EditorKit) and the
	 *        FormatterFactory used to write the drawing to a writer.
	 *        If null, the default editor-kit/content-type is used.
	 * @author Sylvain Reynal
	 * @since PicEdt 1.0
	 */
	public PECanvas(double zoom, PageFormat pageFormat, Grid grid, ContentType contentType){

		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("start");
		this.zoom = zoom;
		setPageFormat(pageFormat); // need zoom to be non-null
		setZoomFactor(zoom);
		// create a default drawing, and create View's for it using 
		// the ViewFactory for the given content-type
		setContentType(contentType);
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Installing grid");
		this.grid = grid;
		// miscellaneous...
		// [pending] setBorder(BorderFactory.createEtchedBorder());
		setBackground(Color.white);
		// undo/redo
		this.undoableEditSupport = new UndoableEditSupport(this); // source for event = PECanvas 
		this.undoManager = new UndoManager();
		setUndoLimit(MAX_UNDOABLE_STEPS_DEFAULT);
		this.undoableEditSupport.addUndoableEditListener(this.undoManager);
		
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("completed !");

		// [pending] debug (press F3 to write a comment line for lisibility) 
		if (jpicedt.Log.DEBUG)
			registerKeyboardAction(new ActionListener(){
				public void actionPerformed(ActionEvent e){
					System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
			}},KeyStroke.getKeyStroke(KeyEvent.VK_F3,0),WHEN_FOCUSED);
		
	}

	
	
	///////////////////////////////////////////////////////////////////////////////
	//// PAINT 
	//////////////////////////////////////////////////////////////////////////////
	
	/**
	 * paintComponent(Graphics g) is called by AWTEventDispatchThread via "paint(g)"
	 * it's absolutely necessary to call super.paintComponent(g) so that the background gets properly painted (PECanvas is opaque)
	 *<p>
	 * This implementation first applies an AffineTransform to the graphic context : this transform maps in effect
	 * model-coordinates to screen-coordinates. Then invokes the paint() method in that order :
	 * <ul>
	 * <li> on the Grid object attached to this PECanvas;
	 * <li> on the root-view associated with Drawing.RootElement.
	 * <li> on the EditorKit in which this PECanvas is installed.
	 * </ul>
	 *
	 */
	public void paintComponent(Graphics g){

		super.paintComponent(g); // paint background
		Graphics2D g2 = (Graphics2D)g;
		g2.setRenderingHints(renderingHints);
		g2.transform(model2ViewTransform); // Note : g2.setTransform(model2ViewTransform) -> bug when partial repaint !!! use g2.transform instead
		Rectangle2D allocation = g2.getClip().getBounds2D();

		if (jpicedt.Log.DEBUG) { 
			jpicedt.Log.debug("clip=" + allocation);
			g2.setPaint(Color.green);
			g2.draw(allocation);
		}
		if (grid != null) grid.paint(g2,allocation,scale);
		if (drawing != null && drawing.getRootView()!=null) 
			drawing.getRootView().paint(g2,allocation); // security : if we are just installing a new drawing and its view-tree hasn't been set yet
		if (kit != null) kit.paint(g2,allocation,scale);
	}

	/**
	 * add the given rectangle, given in model-coordinates, to the list of dirty regions.
	 */
	public void repaintFromModelRect(Rectangle2D rect){

		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("repaintFromModelRect");
		if (jpicedt.Log.DEBUG) jpicedt.Log.debugAppendLn("rect2D="+rect);
		// fetch topleft corner :
		tmpCoords[0] = rect.getX();
		tmpCoords[1] = rect.getMaxY();
		// translate it to view coord system :
		model2ViewTransform.transform(tmpCoords,0,tmpCoords,0,1);
		// set topleft corner for destination rectangle
		repaintRectangle.x = (int)tmpCoords[0];
		repaintRectangle.y = (int)tmpCoords[1];
		// fetch double-precision width and height
		tmpCoords[0] = rect.getWidth();
		tmpCoords[1] = rect.getHeight();
		// delta-translate it to view coord system :
		model2ViewTransform.deltaTransform(tmpCoords,0,tmpCoords,0,1);
		// set dimension for destination rectangle
		repaintRectangle.width = (int)tmpCoords[0];
		repaintRectangle.height = -(int)tmpCoords[1];
		// add repaintRectangle to dirty region
		if (jpicedt.Log.DEBUG) jpicedt.Log.debugAppendLn("repaintRect="+repaintRectangle);
		repaint(repaintRectangle);	
	}

	/**
	 * Return the RenderingHints applied to the graphic context when rendering this component
	 */
	public RenderingHints getRenderingHints(){
		return renderingHints;
	}




	////////////////////////////////////////////////////
	//// CONTENT HANDLING
	////////////////////////////////////////////////////

	/**
	 * @return the model, ie a Drawing containing only non-selected objects
	 */
	public Drawing getDrawing(){
		return drawing;
	}

	/**
	 * set the Drawing model for this component. The currently registered EditorKit is 
	 * used to build a viewtree for the drawing. A PropertyChange event (DRAWING_CHANGE) is sent
	 * to each listener.
	 */
	public void setDrawing(Drawing dr){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("drawing="+dr);
		Drawing old = this.drawing;
		this.drawing = dr;
		this.drawing.setViewTree(getEditorKit().getViewFactory());
		firePropertyChange(DRAWING_CHANGE, old, this.drawing); // [pending] doesn't seem to be fired !
	}

	/**
	 * Fetches the currently installed kit for handling content.
	 */
	public EditorKit getEditorKit(){
		return kit;
	}

	/**
	 * Creates a default editor kit (<code>EditorKit</code>) whose factory delegates are initialized
	 * from the given content-type. If an EditorKit was already installed in this PECanvas, 
	 * we try to reuse its properties in the new EditorKit, if applicable.
	 * @param contentType null if a default content-type should be used
	 * @return the editor kit
	 */
	protected EditorKit createDefaultEditorKit(ContentType contentType) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("createDefaultEditorKit");
		return new EditorKit(contentType, this.kit); 
	}

	/**
	 * Sets the currently installed kit for handling content.  This is the bound property that
	 * establishes the content type of the editor. Any old kit is first deinstalled, then if kit is
	 * non-<code>null</code>, the new kit is installed. <p>
	 * A default drawing is created from it if there was no drawing set in this canvas before, 
	 * otherwise the old drawing is reused : in both cases, <code>setDrawing</code> is called, but
	 * this allows the caller to change the ContentType w/o changing the Drawing if it deems it unnecessary
	 * (otherwise, it may call setDrawing() afterwards).
	 * A <code>PropertyChange</code> event (EDITOR_KIT_CHANGE) is always fired when
	 * <code>setEditorKit</code> is called.
	 * @param kit the desired editor behavior
	 * @see #getEditorKit
	 */
	public void setEditorKit(EditorKit kit) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("kit="+kit);
		EditorKit old = this.kit;
		if (old != null) {
			old.deinstall(this);
		}
		this.kit = kit;
		if (this.kit != null) {
			this.kit.install(this); // add mouse/keybrd event listeners to PECanvas
			if (drawing == null) setDrawing(new Drawing());
			else setDrawing(drawing); 
		}
		firePropertyChange(EDITOR_KIT_CHANGE, old, kit);
	}




	///////////////////////////////////
	///// CONTENT TYPE
	///////////////////////////////////
	/**
	 * @return the current content-type
	 */
	public ContentType getContentType(){
		return contentType;
	}

	/**
	 * change the current content-type: this implies plugging a new ViewFactory and a new FormatterFactory
	 * to the currently installed EditorKit, as obtained from the given newContentType argument.
	 * @param newContentType if null, the DefaultContentType is installed.
	 */
	public void setContentType(ContentType newContentType){
		ContentType old = this.contentType;
		if (old!=null && old==newContentType) return;
		this.contentType = newContentType;
		// update editor's factories:
		if (this.kit == null){ // first install a default kit
			setEditorKit(createDefaultEditorKit(newContentType));
		}
		else {
			this.kit.setFactoriesFromContentType(newContentType);
			this.drawing.setViewTree(getEditorKit().getViewFactory()); // else this was done by invoking setEditorKit()
		}
		firePropertyChange(CONTENT_TYPE_CHANGE, old, this.contentType);
	}


	///////////////////////////////////
	///// DRAWING BOARD SIZE
	///////////////////////////////////

	/**
	 * Set the size of the drawing board. Length are given in mm (this should approximately
	 * represent true mm on the screen, however this might slightly depend on the underlying platform).
	 * <br>
	 * This in turn sets the preferred size of the component. .
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void setPageFormat(PageFormat pageFormat){

		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("page-format="+pageFormat);
		PageFormat oldFormat = this.pageFormat;
		this.pageFormat = pageFormat;
		model2ViewTransform = pageFormat.getModel2ViewTransform(zoom);
		view2ModelTransform = pageFormat.getView2ModelTransform(zoom);
		setPreferredSize(pageFormat.getSizePx(zoom));
		//firePropertyChange(PAGE_FORMAT_CHANGE, oldFormat, this.pageFormat); //[pending] bug : not fired ! 
		// give a chance
		// to listeners to update their layout (e.g. ScrollPane)
		invalidate();
		fireZoomUpdate(this.zoom, this.zoom, null); // force e.g. a scrollpane to update rulers
		//validate();
		repaint();
	}

	/**
		 * @return the page format for this drawing board
		 * @author Sylvain Reynal
		 * @since jPicEdt
		 */
	public PageFormat getPageFormat(){
		return pageFormat;
	}

	/**
	 * @return the pixel-coordinates of the (0,0) model origin when zoom = 1.0
	 */
	public PicPoint getSheetOrigin(){
		return pageFormat.getOrgPx(1.0);
	}

	/**
	 * @return the grid attached to this canvas
	 */
	public Grid getGrid() {
		return grid;
	}

	///////////////////////////////////
	///// Model <-> View
	///////////////////////////////////
	

	/**
	 * @return an AffineTransform that represents the maping b/w the model-coordinate system
	 *         and the pixel coordinate system. Guaranteed not to change over time.
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public AffineTransform getModelToViewTransform(){
		return (AffineTransform)model2ViewTransform.clone();
	}
	
	/**
	 * @return an AffineTransform that represents the maping b/w the pixel-coordinate system
	 *         and the model-coordinate system. Guaranteed not to change over time.
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public AffineTransform getViewToModelTransform(){
		return (AffineTransform)view2ModelTransform.clone();
	}

	/**
	 * Converts a point from the model-coordinate system to the pixel-coordinate system.
	 * @param src the source point in model-coordinate
	 * @param dst the destination point ; if null, a new point is allocated, and returned for convenience.
	 * @return the result (same as dst if non-null)
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public PicPoint modelToView(PicPoint src, PicPoint dst){
		if (dst==null) dst = new PicPoint();
		// fetch topleft corner :
		tmpCoords[0] = src.x;
		tmpCoords[1] = src.y;
		// translate it to view coord system :
		model2ViewTransform.transform(tmpCoords,0,tmpCoords,0,1); // transform 1 point
		dst.x = tmpCoords[0];
		dst.y = tmpCoords[1];
		return dst;
	}
		
	/**
	 * Converts a point from the pixel-coordinate system to the model-coordinate system.
	 * @param src the source point in pixel-coordinate
	 * @param dst the destination point ; if null, a new point is allocated, and returned for convenience.
	 * @return the result (same as dst if non-null)
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public PicPoint view2Model(PicPoint src, PicPoint dst){
		if (dst==null) dst = new PicPoint();
		// fetch topleft corner :
		tmpCoords[0] = src.x;
		tmpCoords[1] = src.y;
		// translate it to view coord system :
		view2ModelTransform.transform(tmpCoords,0,tmpCoords,0,1); // transform 1 point
		dst.x = tmpCoords[0];
		dst.y = tmpCoords[1];
		return dst;
	}

	/**
	 * Converts a Shape from the model-coordinate system to the pixel-coordinate system.
	 * @return a new Shape corresponding to the given Shape once transformed to the pixel-coordinate system.
	 * @param src a Shape in the model-coordinate system
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public Shape modelToView(Shape src){
		return model2ViewTransform.createTransformedShape(src);
	}

	/**
	 * Converts a Shape from the pixel-coordinate system to the model-coordinate system.
	 * @return a new Shape corresponding to the given Shape once transformed to the model-coordinate system.
	 * @param src a Shape in the pixel-coordinate system
	 * @author Sylvain Reynal
	 * @since jPicEdt
	 */
	public Shape viewToModel(Shape src){
		return view2ModelTransform.createTransformedShape(src);
	}
	
	///////////////////////////////////
	///// I/O
	///////////////////////////////////

	/**
	 * read drawing content from a reader and erase old one. Listener's interested in
	 * DrawingEvent's should register their listener anew (this can be done systematically by
	 * registering a PropertyCHangeListener to this canvas, and waiting for DRAWING_CHANGE events).<br> 
	 * @param reader the reader to read content from
	 */
 	public void read(Reader reader, Parser parser) throws jpicedt.graphic.io.parser.ParserException {
		getEditorKit().getSelectionHandler().unSelectAll(); // don't fire selection event
		Drawing parsed = parser.parse(reader); // takes some time...
		setDrawing(parsed);
		Rectangle2D bb = parsed.getBoundingBox();
		if (!this.pageFormat.isFitInto(bb)){
			//PageFormat newPageFormat = new PageFormat(); // [pending] clone current format and enlarge it
			//newPageFormat.fitTo(bb);
			this.pageFormat.enlargeTo(bb);
			//setPageFormat(newPageFormat);
			setPageFormat(pageFormat);
		}
		repaint();
	}
 
	/**
	 * insert content from a reader into the current drawing<br>
	 * @param reader the reader to insert content from
	 */
 	public void insert(Reader reader, Parser parser) throws jpicedt.graphic.io.parser.ParserException {

		//jpicedt.graphic.io.parser.LaTeXParser latexParser = new jpicedt.graphic.io.parser.LaTeXParser(); // adapt parser to the type of content this EditorKit handles
		Drawing parsed = parser.parse(reader);
		getEditorKit().getSelectionHandler().unSelectAll(); // don't fire selection event
		// add parsed content to the current drawing
		for (Iterator it=parsed.getRootElement().children(); it.hasNext(); ){
			Element o = (Element)it.next();
			o = (Element)o.clone();
			drawing.addElement(o);
			getEditorKit().getSelectionHandler().addToSelection(o);
		}
		fireSelectionUpdate(kit.getSelectionHandler().asArray(), SelectionEvent.EventType.SELECT);
		// add more non-parsed command to current drawing, if applicable :
		if (parsed.getNotparsedCommands().length()==0) return;
		String s = drawing.getNotparsedCommands();
		if (s==null || s.length()==0) drawing.setNotparsedCommands(s);
		else {
			s += "\n";
			s += parsed.getNotparsedCommands();
			drawing.setNotparsedCommands(s);
		}
	}

	/**
	 * Write drawing content to the given stream
	 * @param writer The writer to write to
	 * @param writeSelectionOnly if true, only write selection content
	 * @exception IOException on any I/O error
	 * @since jPicEdt
	 */
 	public void write(Writer writer, boolean writeSelectionOnly) throws IOException {

		String buffer;
		if (writeSelectionOnly){
			Drawing dr = new Drawing(getEditorKit().getSelectionHandler().asCollection()); // deep copy
			// translate fragment to (0,0)
			Rectangle2D bb = dr.getBoundingBox();
			dr.getRootElement().translate(- bb.getX(), -bb.getY());
			// first write JPIC-XML text :
			buffer = (new jpicedt.graphic.io.formatter.JPICFormatter()).createFormatter(dr,null).format();
			// then write content-type specific text, except if the currently installed content-type is the default		
			// since the latter already handles the JPIC-XML format... and we would get the JPIC-XML text twice...
			if (!(this.contentType instanceof DefaultContentType))			
				buffer += getEditorKit().getFormatterFactory().createFormatter(dr,null).format();
		}
		else { 
			// first write JPIC-XML text :
			buffer = (new jpicedt.graphic.io.formatter.JPICFormatter()).createFormatter(drawing,null).format();
			// then write content-type specific text, except if the currently installed content-type is the default
			// since the latter already handles the JPIC-XML format... and we would get the JPIC-XML text twice...
			if (!(this.contentType instanceof DefaultContentType))
				buffer += getEditorKit().getFormatterFactory().createFormatter(drawing,null).format();
		} 
		writer.write(buffer);
	}






	///////////////////////////////////
	///// ZOOM
	///////////////////////////////////

	/**
	 * Convenience call to setZoomFactor(zoom,null)
	 */
	public void setZoomFactor(double zoom){
		setZoomFactor(zoom,null);
	}
	
	/**
	 * sets the current zoom factor to the given double, 
	 * then updates various properties (model <-> view transforms, dimension, preferredSize...), finally,
	 * sources a ZoomEvent to give a chance to receiver to update their state accordingly (this may
	 * be used e.g. by a parent scrollpane to update its view port location, or by a GUI widget
	 * to reflect the new zoom value).
	 * @param zoom the new zoom factor
	 * @param ptClick this only makes sense if the parent of this component is aka ScrollPane ; 
	 *        <br> Coordinates for this point are in the model-coordinate system.
	 */
	public void setZoomFactor(double zoom, PicPoint ptClick){

		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("zoom="+zoom);
		double oldZoom = this.zoom; // store old value for further use by zoom event 
		this.zoom = zoom; // [pending] test if zoom really changed !
		
		// update global variables according to new zoom value :
 		model2ViewTransform = pageFormat.getModel2ViewTransform(zoom);
 		view2ModelTransform = pageFormat.getView2ModelTransform(zoom);
 		scale = model2ViewTransform.getScaleX(); // assumption : scaleX = scaleY !
		
		// update preferred size according to new zoom :
		setPreferredSize(pageFormat.getSizePx(zoom));
		invalidate(); // JDK's documentation says : "revalidate()"
		// give a change to a scrollpane to update the view location in the viewport and ruler sizes :
		fireZoomUpdate(oldZoom, this.zoom, ptClick);
		repaint();
	}

	/**
	 * @return the current zoom factor
	 */
	public double getZoomFactor(){
		return zoom;
	}

	/**
	 * @return the current scale factor between model- and view-coordinates, as given by the current
	 *         model2ViewTransform. This is usually the product of the current zoom factor, and the 
	 *         DotPerMilliMeter screen factor.
	 */
	public double getScaleFactor(){
		return scale;
	}

	/**
	 * Notify all listeners that have registered interest for notification on this event type.
	 * @param oldZoom previous zoom value
	 * @param newZoom new zoom value
	 * @param ptClick the point (in model-coordinates) that is expected to be at the center of the view-port ;
	 *        can be null
	 */
	protected void fireZoomUpdate(double oldZoom, double newZoom, PicPoint ptClick){
		Object[] listeners = listenerList.getListenerList();
		ZoomEvent e = null;
		for (int i = listeners.length-2; i>=0; i-=2) {
			// lazily create the event :
			if (e==null) e = new ZoomEvent(this,oldZoom,newZoom,ptClick);
			if (listeners[i]==ZoomListener.class) {
				((ZoomListener)listeners[i+1]).zoomUpdate(e);
			}
		}
	}

	/**
	 * adds a ZoomListener to the Canvas
	 */
	public void addZoomListener(ZoomListener l){
		listenerList.add(ZoomListener.class, l);
	}
	
	/**
	 * removes a ZoomListener from the Canvas
	 */
	public void removeZoomListener(ZoomListener l){
		listenerList.remove(ZoomListener.class, l);
	}
	
	/**
	 * utilities to retrieve the index of the given zoom in PREDEFINED_ZOOMS ; this may be used by GUI widgets,
	 * e.g. JComboBox,...
	 * @return index of the given zoom in array "PREDEFINED_ZOOMS" ; returns -1 if not found.
	 */
	public static int getZoomIndex(double zoom){
		for(int i=0; i<PREDEFINED_ZOOMS.length; i++){
			if (PREDEFINED_ZOOMS[i] == zoom) return i;
		}
		return -1; // not found
	}
	

	/////////////////////////////
	///// UNDO/REDO
	/////////////////////////////

	/**
	 * set the number of undoable events to remember 
	 */
	public void setUndoLimit(int limit){
		this.undoManager.setLimit(limit);
	}
	
	/**
	 * Undo last change
	 * @since PicEdt 1.1.3
	 */
	public void undo() throws CannotUndoException {
		selectAll(false);
		this.undoManager.undo();
	}

	/**
	 * Redo last change
	 * @since PicEdt 1.1.3
	 */
	public void redo() throws CannotRedoException {
		selectAll(false);
		this.undoManager.redo();
	}

	/**
	 * Register an UndoableEditListener for the Drawing hosted by this canvas.
	 */
	public void addUndoableEditListener(UndoableEditListener l){
		this.undoableEditSupport.addUndoableEditListener(l);
	}

	/**
	 * Unregister an UndoableEditListener for the Drawing hosted by this canvas.
	 */
	public void removeUndoableEditListener(UndoableEditListener l){
		this.undoableEditSupport.removeUndoableEditListener(l);
	}

	/**
	 * Create a new UndoableEdit that holds the current state of the Drawing. 
	 */
	public void beginUndoableUpdate(String presentationName){
		if (this.stateEdit != null) {
			endUndoableUpdate(); // if end() was not called !
		}
		this.stateEdit = new PEStateEdit(getDrawing(),presentationName);
	}
	
	/**
	 * Ends the current UndoableEdit and fire an event to registered listeners.
	 */
	public void endUndoableUpdate(){
		if (this.stateEdit==null) {
			return; // refuse two calls to end()
		}
		this.stateEdit.end();
		this.undoableEditSupport.postEdit(this.stateEdit); // notify undoableEditListener's of a new undoable-edit (this is the way UndoMgr gets informed...)
		this.stateEdit = null; // flag for begin() to know we've ended !
	}
	
	/**
	 * @return true if a "redo" operation would be successfull
	 */
	public boolean canRedo(){
		return this.undoManager.canRedo();
	}

	/**
	 * @return true if a "undo" operation would be successfull
	 */
	public boolean canUndo(){
		return this.undoManager.canUndo();
	}

	/**
	 * @return the presentation name of the next edit that can be redone
	 */
	public String getRedoPresentationName(){
		if (!canRedo()) return "";
		return this.undoManager.getRedoPresentationName();
	}

	/**
	 * @return the presentation name of the last edit that can be undone
	 */
	public String getUndoPresentationName(){
		if (!canUndo()) return "";
		return this.undoManager.getUndoPresentationName();
	}

	/** overriden so as to display our own undo- and redo- presentation names */
	private class PEStateEdit extends StateEdit {
		
		PEStateEdit(StateEditable anObject, String name){
			super(anObject, name);
		}
		
	    public String getUndoPresentationName() {
			return getPresentationName();
		}

	    public String getRedoPresentationName() {
			return getPresentationName();
		}
	}

	//////////////////////////////////
	////// SELECTION MANAGEMENT
	//////////////////////////////////

	/**
	 * @return an Iterator over selected graphic elements
	 */
	public Iterator selection(){ // [pending] change to  "getSelection"
		return getEditorKit().getSelectionHandler().elements();
	}
	
	/**
	 * @return the number of selected elements in the current selection
	 */
	public int getSelectionSize(){ 
		return getEditorKit().getSelectionHandler().size();
	}
	
	/**
	 * @return whether the given element is selected or not
	 */
	public boolean isSelected(Element e){
		return getEditorKit().getSelectionHandler().isSelected(e, true);
	}
	
	/**
	 * select or unselect every object in this drawing
	 * @param state if true, selectAll, otherwise unselect all.
	 */
	public void selectAll(boolean state){
		if (state == false){ // unselect all
			if (getEditorKit().getSelectionHandler().size()==0) return; // already empty
			Element[] deselected = getEditorKit().getSelectionHandler().asArray();
			getEditorKit().getSelectionHandler().unSelectAll();
			fireSelectionUpdate(deselected, SelectionEvent.EventType.UNSELECT);
		}
		else { // select all
			kit.getSelectionHandler().selectAll(drawing); 
			fireSelectionUpdate(kit.getSelectionHandler().asArray(), SelectionEvent.EventType.SELECT);
		}
	}

	/**
	 * select the elements in the given collection (if they belong to the drawing)
	 * @param incremental if true, add to the existing selection ; replace otherwise.
	 */
	public void select(Collection c, boolean incremental){
		if (incremental == false) getEditorKit().getSelectionHandler().unSelectAll();
		ArrayList selected = new ArrayList();
		for (Iterator it = c.iterator(); it.hasNext();){
			Element obj = (Element)it.next();
			if (isSelected(obj)) {
				if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Already selected:elem="+obj);
			}
			else {
				getEditorKit().getSelectionHandler().addToSelection(obj);
				selected.add(obj);
			}
		}
		Element[] selectedArray = (Element[])selected.toArray(new Element[0]); 
		fireSelectionUpdate(selectedArray, SelectionEvent.EventType.SELECT);
	}
	
	/**
	 * select the given element
	 * @param incremental if true, add to the existing selection ; replace otherwise.
	 */
	public void select(Element obj, boolean incremental){
		if (incremental == true) {
			if (isSelected(obj)) {
				if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Already selected:elem="+obj);
				return; // already selected
			}
			getEditorKit().getSelectionHandler().addToSelection(obj);
			fireSelectionUpdate(obj, SelectionEvent.EventType.SELECT);
		}
		else {
			getEditorKit().getSelectionHandler().replaceSelection(obj);
			fireSelectionUpdate(obj, SelectionEvent.EventType.SELECT); // don't fire "unselect"
		}
	}

	/**
	 * unselect the given object
	 */
	public void unSelect(Element obj){
		if (!isSelected(obj)) {
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Already unselected:elem="+obj);
			return; // already unselected
		}
		getEditorKit().getSelectionHandler().unSelect(obj);
		fireSelectionUpdate(obj, SelectionEvent.EventType.UNSELECT);
	}

	/**
	 * remove all selected objects from the drawing
	 */
	public void deleteSelection(){
		getEditorKit().getSelectionHandler().delete(drawing);
	}

	/**
	 * Add the content of the given ClipBoard to the current drawing, then select it.
	 * If only DataFlavor.stringFlavor is provided by the Transferable, 
	 * we try to parse the string and insert the parsed content. Otherwise, it's taken for granted
	 * that the ClipBoard content support the jpicedt.graphic.toolkit.TransferableGraphic.JPICEDT_DATA_FLAVOR
	 * data flavor.
	 * @param translate if true, translate the pasted content by a grid step so that it doesn't hide old one
	 */
	public void paste(Clipboard clipbrd, boolean translate) throws ParserException, IOException, UnsupportedFlavorException {
		getEditorKit().getSelectionHandler().unSelectAll(); // don't fire selection event
		Transferable transferable = clipbrd.getContents(this); // according to doc. requestor is not used.
		if (transferable==null) return; // clipboard is empty !
		
		// first check if it's a local clipboard supporting JPICEDT_DATA_FLAVOR :
		if (transferable.isDataFlavorSupported(TransferableGraphic.JPICEDT_DATA_FLAVOR)){
			Element[] content = (Element[])transferable.getTransferData(TransferableGraphic.JPICEDT_DATA_FLAVOR);
			for (int i=0; i<content.length; i++){
				// translate clipboard so that new content doesn't hide old one 
				// Warning : this translates the source !!!
				if (translate) content[i].translate(grid.getSnapStep(),-grid.getSnapStep());
				// add (a copy of the) content to the current drawing
				Element element = (Element)content[i].clone();
				drawing.addElement(element);
				getEditorKit().getSelectionHandler().addToSelection(element);
			}
			fireSelectionUpdate(kit.getSelectionHandler().asArray(), SelectionEvent.EventType.SELECT);
		}
		
		// otherwise that may be the System Clipboard : only text is supported :
		else {
			StringReader reader = new StringReader(jpicedt.MiscUtilities.getClipboardStringContent(clipbrd));
			Parser parser = jpicedt.MiscUtilities.createParser();
			insert(reader,parser);
		}
		
		// else does nothing
	}

	/**
	 * Add the content of the System's ClipBoard to the current drawing, then select it.
	 * More specifically, we try to parse the string and insert the parsed content. 
	 * @param translate if true, translate the pasted content by a grid step so that it doesn't hide old one
	 */
	public void paste(boolean translate) throws ParserException, IOException, UnsupportedFlavorException {
		paste(Toolkit.getDefaultToolkit().getSystemClipboard(),translate);
	}
	
	/**
	 * Copy the content of the current selection (through a GraphicTransferable) to the
	 * System's clipboard (after a formatting to text), AND to the given clipboard if non-null
	 * (the latter can be a local clipboard supporting more data-flavors than the system clipboard)
	 * @param clipbrd the target clipboard ; can be null, in which case only the System clipboard
	 *        is modified.
	 */
	public void copy(Clipboard clipbrd) {
		// if selection buffer is empty, don't modify clipboard's content
		if (getSelectionSize() == 0) return;
		// create array of selected Elements for GraphicTransferable
		Element[] elements = new Element[getSelectionSize()];
		int counter = 0;
		for (Iterator it = selection(); it.hasNext();){
			Element e = (Element)it.next();
			elements[counter++] = e;
		}
		// create formatted string using "write" with selectionOnly = true :
		StringWriter writer = new StringWriter();
		try {
			write(writer,true); 
			// create GraphicTransferable :
			TransferableGraphic transferable = new TransferableGraphic(elements,writer.toString());
			writer.close();
			if (clipbrd!=null) clipbrd.setContents(transferable,transferable);
			Clipboard systemClipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
			StringSelection ss = new StringSelection(writer.toString());
			systemClipboard.setContents(ss,ss);
		} catch (IOException ioEx){ioEx.printStackTrace();}

	}

	/**
	 * Copy the content of the current selection to the
	 * System's clipboard (after a formatting to text)
	 */
	public void copy() {
		copy(null);
	}

	/**
	 * Cut the content of the current selection (through a GraphicTransferable) to the 
	 * System clipboard, AND to the given ClipBoard if non-null.
	 * @param clipbrd the target clipboard ; can be null, in which case only the System clipboard
	 *        is modified.
	 */
	public void cut(Clipboard clipbrd) {
		// if selection buffer is empty, don't modify clipboard's content
		if (getSelectionSize() == 0) return;
		copy(clipbrd);
		deleteSelection();
	}

	/**
	 * Cut the content of the current selection to the System clipboard, after formatting to text.
	 */
	public void cut() {
		cut(null);
	}

	/**
	 * group all selected objects into a new PicGroup and add it to the drawing.
	 * @since jPicEdt 1.2.a
	 */
	public void groupSelection(){
		PicGroup group = new PicGroup(getEditorKit().getSelectionHandler().asCollection());
		getEditorKit().getSelectionHandler().delete(drawing); // delete old selected elements
		drawing.addElement(group); 
		getEditorKit().getSelectionHandler().addToSelection(group); // select group
		fireSelectionUpdate(group, SelectionEvent.EventType.SELECT);
	}

	/**
	 * fetch all Element's belonging to the given PicGroup and add them to 
	 * its parent, removing the given PicGroup from its parent afterward.
	 */
	public void unGroup(PicGroup g){
		getEditorKit().getSelectionHandler().unSelectAll(); // otherwise the selectionHandler just gets confused
		BranchElement p = (BranchElement)g.getParent();
		p.removeChild(g); 
		int max = g.getChildCount();
		for (int i=0; i<max; i++){
			Element o = (Element)g.getChildAt(0); // always fetch at position 0 
			p.addChild(o); // hence removed from group (was former parent)
			select(o,true); // incremental
		}
	}
	
	/**
	 * join all selected objects of type PicMultiCurveConvertable into a new PicMulticurve and add it to the drawing.
	 * @since jPicEdt 1.4.pre5
	 */
	public void joinSelection(){
		// fetch PicMultiCurveConvertable's in selection: (note that they don't lose their parent in this process)
		ArrayList l = getEditorKit().getSelectionHandler().createFilteredCollection(PicMultiCurveConvertable.class);
		ArrayList lBackup = new ArrayList(l);
		if (l.size()<=1) return;
		// join along the shortest-distance path:
		PicMultiCurve c0 = ((PicMultiCurveConvertable)l.get(0)).convertToMultiCurve();
		l.remove(0);
		lBackup.remove(0); // this is the one to which other were joined
		unSelect((Element)l.get(0));
		drawing.replaceElement((Element)l.get(0),c0);
		select(c0,true);
		while (l.isEmpty()==false){
			PicMultiCurveConvertable c1 = c0.fetchClosestCurve(l);
			c0.join(c1); 
			l.remove(c1); 
		}
		// unselect then delete old PicMultiCurveConvertable's from drawing:
		for (Iterator it = lBackup.iterator(); it.hasNext();){
			Element e = (Element)it.next();
			getEditorKit().getSelectionHandler().unSelect(e);
			drawing.removeElement(e);
		}
	}
	
	/**
	 * Notify all listeners that have registered interest for notification on this event type.
	 * @param element the Element that was (un)selected
	 * @param type the event type
	 */
	protected void fireSelectionUpdate(Element element, SelectionEvent.EventType type){
		Object[] listeners = listenerList.getListenerList();
		SelectionEvent e = null;
		for (int i = listeners.length-2; i>=0; i-=2) {
			// lazily create the event :
			if (e==null) e = new SelectionEvent(this,element,type);
			if (listeners[i]==SelectionListener.class) {
				((SelectionListener)listeners[i+1]).selectionUpdate(e);
			}
		}
	}
	
	/**
	 * Notify all listeners that have registered interest for notification on this event type.
	 * @param elements the Element's that were (un)selected
	 * @param type the event type
	 */
	protected void fireSelectionUpdate(Element[] elements, SelectionEvent.EventType type){
		Object[] listeners = listenerList.getListenerList();
		SelectionEvent e = null;
		for (int i = listeners.length-2; i>=0; i-=2) {
			// lazily create the event :
			if (e==null) e = new SelectionEvent(this,elements,type);
			if (listeners[i]==SelectionListener.class) {
				((SelectionListener)listeners[i+1]).selectionUpdate(e);
			}
		}
	}

	/**
	 * adds a SelectionListener to the Canvas
	 */
	public void addSelectionListener(SelectionListener l){
		listenerList.add(SelectionListener.class, l);
	}
	
	/**
	 * removes a SelectionListener from the Canvas
	 */
	public void removeSelectionListener(SelectionListener l){
		listenerList.remove(SelectionListener.class, l);
	}
	

	
	
	
	
	
	
	////////////////////////////////////
	///// SCROLLABLE INTERFACE and rel.
	////////////////////////////////////

	/**
	 * @return the drawing board size (aka ScrollPane's View size, as opposed to ViewPort's size)
	 */
	public Dimension getPreferredScrollableViewportSize() {

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension componentSize = getPageFormat().getSizePx(zoom);
		// return half screen size at maximum
		return(new Dimension((int)Math.min(componentSize.width,screenSize.width/2),
		                     (int)Math.min(componentSize.height,screenSize.height/2)));
	}

	/**
	 * @return a grid step
	 */
	public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {

		return (int)(grid.getSnapStep() * zoom);
	}

	/**
	 * @return the viewport size minus a grid step
	 */
	public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {

		if (orientation == SwingConstants.HORIZONTAL)
			return (int)(visibleRect.width - grid.getSnapStep() * zoom);
		else
			return (int)(visibleRect.height - grid.getSnapStep() * zoom);
	}

	/**
	 * @return false for this implementation
	 */
	public boolean getScrollableTracksViewportWidth() {
		return false;
	}

	/**
	 * @return false for this implementation
	 */
	public boolean getScrollableTracksViewportHeight() {
		return false;
	}








	//////////////////////////
	//// MISC
	//////////////////////////


	/**
	 * Overriden from JComponent
	 * Signals that this component can receive focus (useful for handling keyevents)
	 */
	public boolean isRequestFocusEnabled(){
		return true;
	}





	//////////////////////////
	//// PEMOUSE LISTENER
	//////////////////////////

	/**
	 * Adds the specified mouse listener to receive mouse events from this component.
	 * If l is null, no exception is thrown and no action is performed.
	 * @param l the mouse listener.
	 */
	public synchronized void addPEMouseInputListener(PEMouseInputListener l) {
		if (l == null) {
			return;
		}
		mouseInputListener = PEEventMulticaster.add(mouseInputListener,l);
		this.enableEvents(AWTEvent.MOUSE_EVENT_MASK);
		this.enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);		
	}

	/**
	 * Removes the specified mouse listener so that it no longer
	 * receives mouse events from this component. This method performs 
	 * no function, nor does it throw an exception, if the listener 
	 * specified by the argument was not previously added to this component.
	 * If l is null, no exception is thrown and no action is performed.
	 *
	 * @param l the mouse listener.
	 */
	public synchronized void removePEMouseInputListener(PEMouseInputListener l) {
		if (l == null) {
			return;
		}
		mouseInputListener = PEEventMulticaster.remove(mouseInputListener, l);
	}

	/**
	 * Processes mouse events occurring on this component by
	 * dispatching them to any registered
	 * <code>PEMouseListener</code> objects.
	 * @param e the mouse event.
	 */
	protected void processMouseEvent(MouseEvent e) {
		super.processMouseEvent(e); // process "standar" mouse-events BEFORE (see pb with JPopupMenu for instance)
		if (mouseInputListener != null) {
			view2ModelTransform.transform(e.getPoint(),peMousePoint);
			PEMouseEvent me = new PEMouseEvent(e, this, peMousePoint);
			switch(e.getID()) {
			case MouseEvent.MOUSE_PRESSED:
				mouseInputListener.mousePressed(me);
				break;
			case MouseEvent.MOUSE_RELEASED:
				mouseInputListener.mouseReleased(me);
				break;
			case MouseEvent.MOUSE_CLICKED:
				mouseInputListener.mouseClicked(me);
				break;
			case MouseEvent.MOUSE_EXITED:
				mouseInputListener.mouseExited(me);
				break;
			case MouseEvent.MOUSE_ENTERED:
				mouseInputListener.mouseEntered(me);
				break;
			}
		}
		
	}

	/**
	 * Processes mouse motion events occurring on this component by
	 * dispatching them to any registered
	 * <code>PEMouseInputListener</code> objects.
	 * @param e the mouse motion event.
	 */
	protected void processMouseMotionEvent(MouseEvent e) {
		super.processMouseMotionEvent(e); // process "standar" mouse-events
		if (mouseInputListener != null) {
			view2ModelTransform.transform(e.getPoint(),peMousePoint);
			PEMouseEvent me = new PEMouseEvent(e, this, peMousePoint);
			switch(e.getID()) {
			case MouseEvent.MOUSE_MOVED:
				mouseInputListener.mouseMoved(me);
				break;
			case MouseEvent.MOUSE_DRAGGED:
				mouseInputListener.mouseDragged(me);
				break;
			}
		}
		
	}
} // class PECanvas


