/*
 PicPoint.java - 1999 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic;

import java.awt.*;
import java.awt.geom.*;
import java.util.StringTokenizer;
import java.util.Comparator;

/**
 * Enhancement of Point2D.Double with convenient formatting methods and constructors.
 * @since jpicedt 1.0
 * @author Sylvain Reynal
 * @version $Id: PicPoint.java,v 1.12.2.1 2007/09/02 11:56:17 reynal Exp $
 */

public class PicPoint extends Point2D.Double implements Cloneable {

	/**
	 * Construct a (0,0) point.
	 */
	public PicPoint() {
		super();
	}

	/**
	 * Clone the given point.
	 */
	public PicPoint(Point2D p){
		super(p.getX(), p.getY());
	}

	/**
	 * Creates and returns a deep copy of this PicPoint. 
	 */
	public Object clone(){
		return new PicPoint(this);
	}
	
	/**
	 * Construct a new PicPoint "p" located on the line joining p1 with p2, so that 
	 * (p1,p) = ratio * (p1,p2)
	 * @param ratio any double, positive or not.
	 */
	public PicPoint(Point2D p1, Point2D p2, double ratio){
		x = p1.getX() + ratio * (p2.getX() - p1.getX());
		y = p1.getY() + ratio * (p2.getY() - p1.getY());
	}
	
	/**
	 * Construct (x,y)
	 */
	public PicPoint(double x, double y){
		super(x,y);
	}

	/**
	 * Construct a point from the given pair of Number (using their double value).
	 */
	public PicPoint(Number x, Number y){
		super(x.doubleValue(), y.doubleValue());
	}

	/**
	 * Construct a point from the first two elements of the given array.
	 */
	public PicPoint(float[] f){
		super(f[0], f[1]);
	}

	/**
	 * Construct a point from the first two elements of the given array.
	 */
	public PicPoint(double[] f){
		super(f[0], f[1]);
	}

	/**
	 * Construct a point by parsing a String similar to that given by the
	 * {@link #toString() toString()} method, ie <code>(x,y)</code>.
	 * @since jpicedt 1.3.3
	 * @author Sylvain Reynal
	 */
	public PicPoint(String str) throws NumberFormatException {
		StringTokenizer tokenizer = new StringTokenizer(str," (,)",false); // doesn't return delimiters
		// if everything went fine, should have 2 tokens, namely x and y :
		if (tokenizer.countTokens() != 2) throw new NumberFormatException(str+": wrong PicPoint formatting, should be \"(double,double)\".");
		else {
			this.x = java.lang.Double.parseDouble(tokenizer.nextToken()); // explicit package specification needed
			this.y = java.lang.Double.parseDouble(tokenizer.nextToken()); // to lift ambiguity with Point2D.Double
		}
	}
	

	/**
	 * Set the coordinates of this point from the given point.
	 * @return this for convenience
	 */
	public PicPoint setCoordinates(PicPoint pt){
		this.x = pt.x;
		this.y = pt.y;
		return this;
	}
	
	/**
	 * Set the coordinates of this point from the given pair
	 */
	public void setCoordinates(double x, double y){
		this.x = x;
		this.y = y;
	}
	
	// ================= translators ==============================
	
	
	
	/**
	 * Return a two-element array filled with x and y ; if f is null, a new array is allocated ; otherwise,
	 *         the given array is directly modified and returned for convenience.
	 */
	public float[] toFloatArray(float[] f){
		if (f==null) f = new float[2];
		f[0] = (float)x;
		f[1] = (float)y;
		return f;
	}

	/**
	 * Return a two-element array filled with x and y ; if f is null, a new array is allocated ; otherwise,
	 *         the given array is directly modified and returned for convenience.
	 */
	public double[] toDoubleArray(double[] f){
		if (f==null) f = new double[2];
		f[0] = x;
		f[1] = y;
		return f;
	}

	/**
	 * Return a "(x,y)" string representing this point. 
	 * The returned String in turn may be fed to the PicPoint(String) constructor. 
	 */
	public String toString(){
		StringBuffer buf = new StringBuffer(15);
		buf.append("(");
		buf.append(PEToolKit.doubleToString(x));
		buf.append(",");
		buf.append(PEToolKit.doubleToString(y));
		buf.append(")");
		return buf.toString();
	}

	/**
	 * Convert a PicPoint with coordinates expressed in the given unitlenth (expressed in mm), 
	 * to a new PicPoint in mm coordinates.
	 * @param unitLength In mm
	 */
	public PicPoint toMm(double unitLength){
		return new PicPoint(x * unitLength, y * unitLength);
	}

	/**
	 * Convert a PicPoint with coordinate expressed in the given unitlenths along X- and Y-axis (the latter being expressed in mm) 
	 * to a new PicPoint in mm coordinate.
	 * @param xUnit unilength along the X-axis expressed in mm
	 * @param yUnit unilength along the Y-axis expressed in mm
	 */
	public PicPoint toMm(double xUnit, double yUnit){
		return new PicPoint(xUnit * x, yUnit * y);
	}


	// ====================== GEOM. UTILITIES ===========================
	
	/**
	 * translates this point by (dx,dy)
	 * @return this for convenience
	 */
	public PicPoint translate(double dx, double dy){
		x += dx;
		y += dy;
		return this;
	}
	
	/**
	 * translates this point by (p.x, p.y), ie the given point is considered as a translation vector.
	 * @return this for convenience
	 */
	public PicPoint translate(PicPoint p){
		x += p.x;
		y += p.y;
		return this;
	}
	
	/**
	 * translates this point by (p2.x-p1.x, p2.y-p2.y), ie the given point is considered as a translation vector
	 * build from the two given points.
	 * @return this for convenience
	 */
	public PicPoint translate(PicPoint p1, PicPoint p2){
		x += p2.x-p1.x;
		y += p2.y-p1.y;
		return this;
	}

	/**
	 * translates this point by a*(p.x, p.y), ie the given point is considered as a translation vector scaled
	 * by the given double.<br>
	 * This method proves a useful when one wants to minimize object creation, since it avoids cloning a given PicPoint,
	 * scaling it by "a", then passing it to the translate(PicPoint) method.
	 * @return this for convenience
	 */
	public PicPoint translate(PicPoint p, double a){
		x += a*p.x;
		y += a*p.y;
		return this;
	}

	/**
	 * translates this point by a*(p2.x-p1.x, p2.y-p2.y), ie the given point is considered as a translation vector
	 * build from the two given points, then scaled by the given double.
	 * This method proves a useful when one wants to minimize object creation.
	 * @return this for convenience
	 */
	public PicPoint translate(PicPoint p1, PicPoint p2, double a){
		x += a*(p2.x-p1.x);
		y += a*(p2.y-p1.y);
		return this;
	}
	
	/**
	 * Apply a central-symmetry wrt the given point
	 * @return this for convenience	 
	 */
	public PicPoint symmetry(PicPoint center){
		x = 2 * center.x - x;
		y = 2 * center.y - y;
		return this;
	}
	
	/**
	 * Return a new PicPoint obtained by applying a central-symmetry with the given center to
	 * the given src point. If src==null, it is allocated and returned for convenience.
	 */
	public static PicPoint symmetry(PicPoint center, PicPoint src){
		PicPoint dest = new PicPoint(src);
		dest.symmetry(center);
		return dest;
	}
	
	/**
	 * Apply a scaling transform to this point. 
	 * @param sx Scaling factors along the X-axis
	 * @param sy Scaling factors along the Y-axis
	 * @param ptOrg transformation centre
	 * @return this for convenience
	 */
	public PicPoint scale(PicPoint ptOrg, double sx, double sy){
		x = ptOrg.x + sx * (x-ptOrg.x);
		y = ptOrg.y + sy * (y-ptOrg.y);
		return this;
	}	
	
	/**
	 * Apply a scaling transform to this point. 
	 * @param s Scaling factors along the X- and Y- axis
	 * @param ptOrg transformation centre
	 * @return this for convenience	 
	 */
	public PicPoint scale(PicPoint ptOrg, double s){
		x = ptOrg.x + s * (x-ptOrg.x);
		y = ptOrg.y + s * (y-ptOrg.y);
		return this;
	}	

	/**
	 * Apply a scaling transform to this point. 
	 * @param s Scaling factors along the X- and Y- axis
	 * @param ptOrgX X-coord of transformation centre
	 * @param ptOrgY Y-coord of transformation centre
	 * @return this for convenience
	 */
	public PicPoint scale(double ptOrgX, double ptOrgY, double s){
		x = ptOrgX + s * (x-ptOrgX);
		y = ptOrgY + s * (y-ptOrgY);
		return this;
	}	

	/**
	 * Apply a scaling transform to this point. 
	 * @param sx Scaling factors along the X-axis
	 * @param sy Scaling factors along the Y-axis
	 * @param ptOrgX X-coord of transformation centre
	 * @param ptOrgY Y-coord of transformation centre
	 * @return this for convenience
	 */
	public PicPoint scale(double ptOrgX, double ptOrgY, double sx, double sy){
		x = ptOrgX + sx * (x-ptOrgX);
		y = ptOrgY + sy * (y-ptOrgY);
		return this;
	}	

	/**
	 * Apply a rotation of center ptOrg and the given angle in radians to this PicPoint
	 * Current implementation arranges for a very fast code if angle is Math.PI, Math.PI/2 or
	 * -Math.PI/2.
	 * @return this for convenience
	 */
	public PicPoint rotate(PicPoint ptOrg, double angle){
		if (angle == Math.PI){
			x = ptOrg.x + ptOrg.x - x;
			y = ptOrg.y + ptOrg.y - y;
		}
		else if (angle == Math.PI/2){
			double _x =  - this.y + ptOrg.y + ptOrg.x;
			this.y = this.x-ptOrg.x+ ptOrg.y;
			this.x = _x;
		}
		else if (angle == -Math.PI/2){
			double _x =  this.y-ptOrg.y + ptOrg.x;
			this.y = - this.x+ptOrg.x+ ptOrg.y;
			this.x = _x;
		}
		else {
			double cosTheta = Math.cos(angle);
			double sinTheta = Math.sin(angle);
			double _x = cosTheta * (this.x-ptOrg.x) - sinTheta * (this.y-ptOrg.y) + ptOrg.x;
			this.y = sinTheta * (this.x-ptOrg.x) + cosTheta * (this.y-ptOrg.y) + ptOrg.y;
			this.x = _x;
		}
		return this;
	}

	/**
	 * Apply a shearing transform of given parameters wrt to the given origin, to this PicPoint
	 * @return this for convenience
	 */
	public PicPoint shear(PicPoint ptOrg, double shx, double shy){
		double _x = this.x + shx * (this.y-ptOrg.y);
		this.y = shy * (this.x-ptOrg.x) + this.y;
		this.x = _x;
		return this;
	}
	
	/**
	 * Translate this point to the middle of the segment made of [this,other].
	 * @return this for convenience
	 * [SR:pending] refactor method name to "midpoint"
	 */
	public PicPoint middle(PicPoint other){
		return scale(other, 1./2.);
	}
	
	/**
	 * Project this point onto the line joining p1 and p2. Projectiong is orthogonal. Does nothing it p1==p2.
	 * @return this for convenience.
	 */
	public PicPoint project(PicPoint p1, PicPoint p2){
		if (p1.equals(p2)) return this;
		// algorithm : let H be the orthogonal projection of THIS on v=(p1,p2), let w=(p1,THIS), then 
		// (P1,H) = v * (v.w)/||v||^2, which yields H.
		// Note that i don't make use of PicPoint's as intermediate buffer in order to reduce object creation ; code
		// obviously looks weirder, but "voor niets kom de zon op !" (dutch proverb)
		double vDotW = (p2.x-p1.x) * (x-p1.x) + (p2.y-p1.y) * (y-p1.y);
		double normV2 = (p2.x-p1.x) * (p2.x-p1.x) + (p2.y-p1.y) * (p2.y-p1.y); 
		x = p1.x + (p2.x - p1.x) * vDotW / normV2;
		y = p1.y + (p2.y - p1.y) * vDotW / normV2;
		return this;
	}
		
	/**
	 * Project this point onto the line joining p1 and p2. Projection is along the given vector. Does nothing it p1==p2
	 * or if "dir" is null, or if dir is parallel to the (p1,p2) line.
	 * @param dir a vector indicating the projection axis
	 * @return this for convenience.
	 */
	public PicPoint project(PicPoint p1, PicPoint p2, PicPoint dir){
		// algorithm : let H be the projection along "dir" of THIS point on v=(p1,p2), let w=(p1,THIS), and
		// e3 the vertical unit vector, then 
		// (P1,H) =  (w x dir).e3 / || v x dir || * v, which yields H.
		// Note that i don't make use of PicPoint's as intermediate buffer in order to reduce object creation ; code
		// obviously looks weirder, but "voor niets kom de zon op !" (dutch proverb)
		double vDotDirvec = (p2.x-p1.x) * dir.y - (p2.y-p1.y) * dir.x;
		if (vDotDirvec==0) return this;
		double wDotDirvec = (x-p1.x) * dir.y - (y-p1.y) * dir.x;  
		x = p1.x + (p2.x - p1.x) * wDotDirvec / vDotDirvec ;
		y = p1.y + (p2.y - p1.y) * wDotDirvec / vDotDirvec ;
		return this;
	}
	
	/** validation tests */
	public static void main(String[] args){
		PicPoint pt = new PicPoint(2,2);
		System.out.println(pt);
		pt.project(new PicPoint(4,2), new PicPoint(0,0), new PicPoint(1,2));
		System.out.println(pt);
	}
	
	/**
	 * Apply the given AffineTransform to the coordinates of this point
	 */
	public void apply(AffineTransform at){
		double _x;
		_x = at.getScaleX()*x + at.getShearX()*y + at.getTranslateX();
		y =  at.getShearY()*x + at.getScaleY()*y + at.getTranslateY();
		x = _x;
	}

	
	////////////////////// Collections specific //////////////////////////
	
	// note to developpers : since PicPoint extends java.awt.Point2D, it inherits the field-to-field
	// equal() method implemented therein.
	
	public static final XComparator X_COMPARATOR = new XComparator();
	public static final YComparator Y_COMPARATOR = new YComparator();
	
	/** a comparator b/w PicPoint for X-axis ordering */
	public static class XComparator implements Comparator {
		
		public int compare(Object o1, Object o2){
			return (int)(((PicPoint)o2).x-((PicPoint)o1).x);
		}
		
		public boolean equals(Object otherComp){
			return otherComp instanceof XComparator;
		}
	}
	
	/** a comparator b/w PicPoint for Y-axis ordering */
	public static class YComparator implements Comparator {
		
		public int compare(Object o1, Object o2){
			return (int)(((PicPoint)o2).y-((PicPoint)o1).y);
		}
		
		public boolean equals(Object otherComp){
			return otherComp instanceof YComparator;
		}
	}
	
} // PicPoint
