/*
 PicVector.java - August 8, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

/**
 * This class encapsulates a geometrical vector (ie it has absolutely no relation with java.util.Vector !).
 * It is based on PicPoint, from where it conveniently inherits formatting methods, yet add some useful
 * vector-calculus related methods, e.g. dot product, norm, etc... Being a subclass of PicPoint is a major
 * benefit where polymorphism is concerned, since instances
 * of this class can be fed as argument to most getters and setters in jpicedt.graphic.model.Element.
 * @since jPicEdt 1.3.3
 * @author Sylvain Reynal
 * @version $Id: PicVector.java,v 1.7.2.1 2007/09/02 11:56:17 reynal Exp $
 */
public class PicVector extends PicPoint  {

	/** the (1,0) unit vector ; this may be used to compute e.g. angles, projections, ... */
	public static PicVector X_AXIS = new PicVector(1,0);
	
	/** the (0,1) unit vector ; this may be used to compute e.g. angles, projections, ...  */
	public static PicVector Y_AXIS = new PicVector(0,1);

	// inherited from Point2D.Double via PicPoint : double x,y
	
	/**
	 * Construct a null vector.
	 */
	public PicVector() {
		super();
	}

	/**
	 * Clone the given vector.
	 */
	public PicVector(PicVector picVec){
		this(picVec.x, picVec.y);
	}

	/**
	 * Construct a vector with coordinates (x,y)
	 */
	public PicVector(double x, double y){
		super(x,y);
	}

	/**
	 * Construct a vector from the given pair of Number (using their double value).
	 */
	public PicVector(Number x, Number y){
		super(x,y);
	}

	/**
	 * Construct a vector from the first two elements of the given array.
	 */
	public PicVector(float[] f){
		super(f);
	}

	/**
	 * Construct a vector from the first two elements of the given array.
	 */
	public PicVector(double[] f){
		super(f);
	}

	/**
	 * Construct a point by parsing a String similar to that given by the
	 * {@link #toString() toString()} method, ie <code>(x,y)</code>.
	 * @since jpicedt 1.3.3
	 * @author Sylvain Reynal
	 */
	public PicVector(String str) throws NumberFormatException {
		super(str);
	}
	
	/**
	 * Construct a new PicVector from the two given points
	 */
	public PicVector(Point2D pt1, Point2D pt2){
		setCoordinates(pt1,pt2);
	}
	
	/**
	 * Construct a new PicVector pointing from (0,0) to the given point
	 */
	public PicVector(Point2D pt){
		super(pt);
	}
	
	/**
	 * Construct a new PicVector from the two end-points of the given Line2D object. 
	 */
	public PicVector(Line2D line){
		this(line.getP1(), line.getP2());
	}
	
	/**
	 * Set the coordinates of this vector from the two given points.
	 * @return this for convenience
	 */
	public PicVector setCoordinates(Point2D pt1, Point2D pt2){
		this.x = pt2.getX() - pt1.getX();
		this.y = pt2.getY() - pt1.getY();
		return this;
	}
		
	/**
	 * Set the coordinates of this vector from the given Line2D object
	 * @return this for convenience
	 */
	public PicVector setCoordinates(Line2D line){
		setCoordinates(line.getP1(), line.getP2());
		return this;
	}
	
	// ========================== GEOM. UTILITIES ====================
	
	/**
	 * Add the given vector to this vector, and return this for convenience.
	 */
	public PicVector add(PicVector other){
		x += other.x;
		y += other.y;
		return this;
	}
	
	/**
	 * Add the given vector scaled by "a" to this vector, and return this for convenience.
	 */
	public PicVector add(PicVector other, double a){
		x += a*other.x;
		y += a*other.y;
		return this;
	}

	/**
	 * Subtract the given vector from this vector, and return this for convenience.
	 */
	public PicVector subtract(PicVector other){
		x -= other.x;
		y -= other.y;
		return this;
	}

	/**
	 * Return the norm of this vector
	 */
	public double norm(){
		return Math.sqrt(x*x+y*y);
	}
	
	/**
	 * Return the squared of the norm of this vector
	 */
	public double norm2(){
		return x*x+y*y;
	}
	
	/**
	 * Normalize this vector so that this becomes a unitary vector
	 * Does nothing if this vector is a null-vector.
	 * @return this for convenience
	 */
	public PicVector normalize(){
		double norm = norm();
		if (norm==0) return this;
		this.x /= norm;
		this.y /= norm;
		return this;
	}
	
	/**
	 * Return the dot product of this vector with the given vector
	 */
	public double dot(PicVector other){
		return x * other.x + y * other.y;
	}
	
	/**
	 * Return whether this vector is orthogonal to the given vector
	 */
	public boolean isOrthogonal(PicVector other){
		return this.dot(other)==0;
	}
	
	/**
	 * Return whether this vector is proportional to the given vector
	 */
	public boolean isColinear(PicVector other){
		return this.det(other)==0;
	}
	
	/**
	 * Multiply each coordinate of this vector by the given double
	 * @return this for convenience
	 */
	public PicVector scale(double a){
		x *= a;
		y *= a;
		return this;
	}
	
	/**
	 * Multiply each coordinate of this vector by the given pair of double
	 * @param ax scale factor for the X-coord
	 * @param ay scale factor for the Y-coord
	 * @return this for convenience
	 */
	public PicVector scale(double ax, double ay){
		x *= ax;
		y *= ay;
		return this;
	}

	/**
	 * Change the sign of each coordinate of this vector
	 * @return this for convenience
	 */
	public PicVector inverse(){
		x = -x;
		y = -y;
		return this;
	}
	
	/**
	 * Return true if this vector has a null-norm
	 */
	public boolean isNull(){
		return (x==0 && y==0);
	}
	
	/**
	 * Return the determinant of {{x,y},{other.x,other.y}} ; this is aka vector product, where only the z-coordinate
	 * gets returned.
	 */
	public double det(PicVector other){
		return x * other.y - y * other.x;
	}
	
	/**
	 * Return the (CCW oriented) angle between this vector and the given vector, in radians.
	 * @return an angle in the range (-PI, PI]
	 */
	public double angle(PicVector other){
		double thisNorm = this.norm();
		if (thisNorm==0) return 0;
		double otherNorm = other.norm();
		if (otherNorm==0) return 0;
		double thisDetOther = this.det(other);
		double thisDotOther = this.dot(other);
		if (thisDetOther==0) {
			if (thisDotOther < 0) return Math.PI;
			else return 0;
		}
		double theta = Math.acos(thisDotOther / thisNorm / otherNorm); 
		if (thisDetOther < 0) theta = - theta;
		return theta;
	}
	
	/**
	 * Return the (CCW oriented) angle between this vector and the given vector, in degrees.
	 */
	public double angleDegrees(PicVector other){
		return Math.toDegrees(this.angle(other));
	}
	
	/**
	 * Rotate this vector by the given (CCW-oriented) angle in radians.<br>
	 * Current implementation arranges for a very fast code if angle is Math.PI, Math.PI/2 or
	 * -Math.PI/2.
	 * @return this for convenience
	 */
	public PicVector rotate(double angle){
		if (angle == Math.PI){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "angle=PI (exact)");
			return inverse();
		}
		else if (angle == Math.PI/2){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "angle=PI/2 (exact)");
			double _x = -this.y;
			this.y = this.x;
			this.x = _x;
		}
		else if (angle == -Math.PI/2){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "angle=-PI/2 (exact)");
			double _x = this.y;
			this.y = -this.x;
			this.x = _x;
		}
		else {
			double cosTheta = Math.cos(angle);
			double sinTheta = Math.sin(angle);
			double _x = cosTheta * this.x - sinTheta * this.y;
			this.y = sinTheta * this.x + cosTheta * this.y;
			this.x = _x;
		}
		return this;
	}

	// ========================== STATIC UTILITIES =======================

	/** validation tests */
	public static void main(String[] args){
		PicVector v = new PicVector(1,1);
		PicVector w = new PicVector(v);
		w.rotate(0);
		System.out.println("dot="+v.dot(w));
		System.out.println("det="+v.det(w));
		System.out.println("angle="+v.angle(w));
		System.out.println("angleDeg="+v.angleDegrees(w));
	}
	
	/**
	 * Return a normalized copy of the given vector.
	 * If dest==null, it is allocated and returned for convenience.
	 */
	public static PicVector normalize(PicVector src, PicVector dest){
		if (dest==null) dest = new PicVector();
		dest.setCoordinates(src);
		dest.normalize();
		return dest;
	}
	
	/**
	 * Return a rotated copy of the given vector. If dest==null, it is allocated and returned for convenience.. 
	 * @see #rotate(double)
	 */
	public static PicVector rotate(PicVector src, PicVector dest, double angle){
		if (dest==null) dest = new PicVector();
		dest.setCoordinates(src);
		dest.rotate(angle);
		return dest;
	}

	/**
	 * Return a scaled copy of the given vector. If dest==null, it is allocated and returned for convenience.. 
	 * @see #rotate(double)
	 */
	public static PicVector scale(PicVector src, PicVector dest, double scale){
		if (dest==null) dest = new PicVector();
		dest.setCoordinates(src);
		dest.scale(scale);
		return dest;
	}

	/**
	 * Return a unitary vector pointing from startPt to endPt
	 */
	public static PicVector getDirector(PicPoint startPt, PicPoint endPt){
		return getDirector(startPt, endPt, null);
	}

	/**
	 * Return a unitary vector pointing from startPt to endPt ; if dest is non-null, it's filled with the result
	 * and returned for convenience ; otherwise, a new PicPoint gets allocated.
	 */
	public static PicVector getDirector(PicPoint startPt, PicPoint endPt, PicVector dest){
		if (dest == null) dest = new PicVector();
		dest.setCoordinates(startPt, endPt);
		dest.normalize();
		return dest;
	}

	/**
	 * Return a unitary vector pointing from startPt to endPt ; if dest is non-null, it's filled with the result
	 * and returned for convenience ; otherwise, a new PicPoint gets allocated.
	 */
	public static PicVector getDirector(double startPtX, double startPtY, double endPtX, double endPtY, PicVector dest){
		if (dest == null) dest = new PicVector();
		dest.x = endPtX-startPtX;
		dest.y = endPtY-startPtY;
		dest.normalize();
		return dest;
	}

	/**
	 * Return a UNITARY vector orthogonal to the vector pointing from startPt to endPt and
	 * built by rotating this vector CCW.
	 */
	public static PicVector getOrthogonal(PicPoint startPt, PicPoint endPt){

		PicVector orthogonal = new PicVector(startPt, endPt);
		orthogonal.rotate(Math.PI/2);
		orthogonal.normalize();
		return orthogonal;

	}
	
}  

