/*
 EditorKitEvent.java - January 2, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.event;

import java.util.EventListener;
import javax.swing.event.*;

import jpicedt.graphic.model.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.view.*;
import jpicedt.graphic.*;

/**
 * A class that allows an EditorKit's mousetool to send high-level mouse-events, i.e. somehow pre-processed
 * by the EditorKit machinery. This is inspired from 
 * {@link javax.swing.event.HyperlinkListener javax.swing.event.HyperlinkListener}.<br>
 * Application cover :
 * <ul>
 * <li> UI implementation outside the jpicedt.graphic package can react to mouse-event w/o the burden
 *      of processing the event.
 * <li> Internal use by other EditorKit's mousetools.
 * <p> [pending] underway ; compile ok but not used yet.
 * @author Sylvain Reynal
 * @version $Id: EditorKitEvent.java,v 1.5.2.1 2007/09/02 11:56:18 reynal Exp $
 * @since jpicedt 1.3.2
 */
public class EditorKitEvent  {

    private EventType type;
    private EditorKit source;
	private HitInfo hitInfo;

	/**
	 * contruct a new EditorKitEvent.
	 * @param source the editorkit that sourced the event
	 * @param type the event type
	 * @param hitInfo the HitInfo associated with the mouse event
	 */
    public EditorKitEvent(EditorKit source, EventType type, HitInfo hitInfo){
    
		this.source=source;
		this.type=type;
		this.hitInfo = hitInfo;
    }
	
	/**
 	 * Return the editor kit that sourced this event 
 	 */
    public EditorKit getSource(){
		return source;
	}

	/**
	 * Return the type of this event
	 */
    public EventType getType(){
		return type;
	}
	
	/**
	 * Return the HitInfo associated with the mouse-event
	 */
	public HitInfo getHitInfo(){
		return hitInfo;
	}

	/**
	 * Return a String representation of this event for debugging purpose.
	 */
    public String toString(){

		return "EditorKitEvent@" + Integer.toHexString(hashCode()) 
		    + ", type=" + type
	    	+ ", value=" + hitInfo
	    	+ ", source=" + source;
    }
	
	/**
	 * typesafe enumeration of allowed event types
	 */
	public static class EventType {
		
		private String name;
		private EventType(String s){name = s;}

		/** 
		 * signal that the mouse-cursor entered the sensitive area of a graphic element
		 * [SR:pending] change name to ON_MOUVE_OVER to conform to W3C's DOM for XML-SVG ?
		 */
		public final static EventType ELEMENT_ENTERED = new EventType("element-entered");

		/** 
		 * signals that the mouse-cursor exited the sensitive area of a graphic element
		 */
		public final static EventType ELEMENT_EXITED = new EventType("element-exited");
		
		/** 
		 * signals that the user clicked on a graphic element
		 * [SR:pending] change name to ON_MOUSE_CLICK ?
		 */
		public final static EventType ELEMENT_CLICKED = new EventType("element-clicked");

		/** 
		 * Return the name of this event type ; this can be used by a GUI, but it's in english, so
		 *  it needs to be localized.
		 */
		public String toString(){
			return name;
		}
	}
}
