/*
 PEEventMulticaster.java - January 3, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.event;

import java.awt.AWTEventMulticaster;
import java.awt.event.*;
import java.util.*;

/**
 * An enhanced version of {@link java.awt.AWTEventMulticaster java.awt.AWTEventMulticaster}
 * that supports chaining {@link jpicedt.graphic.event.PEMouseInputListener PEMouseInputListener's}.
 * @author Sylvain Reynal
 * @version $Id: PEEventMulticaster.java,v 1.6.2.1 2007/09/02 11:56:18 reynal Exp $
 * @since jpicedt 1.3.2
 */
public class PEEventMulticaster extends AWTEventMulticaster implements PEMouseInputListener {

	/**
	 * Creates a PEEventMulticaster instance which chains listener-a with listener-b 
	 */
	protected PEEventMulticaster(EventListener a, EventListener b){
		super(a,b);
	}
	
    /**
     * Adds PEMouseInputListener-a with PEMouseInputListener-b and
     * returns the resulting multicast listener.
     * @param a PEMouseInputListener-a
     * @param b PEMouseInputListener-b
     */
    public static PEMouseInputListener add(PEMouseInputListener a, PEMouseInputListener b) {
        return (PEMouseInputListener)addInternal(a, b);
    }

    /**
     * Removes the old PEMouseInputListener from PEMouseInputListener-l and
     * returns the resulting multicast listener.
     * @param l PEMouseInputListener-l
     * @param oldl the PEMouseInputListener being removed
     */
    public static PEMouseInputListener remove(PEMouseInputListener l, PEMouseInputListener oldl) {
		return (PEMouseInputListener) removeInternal(l, oldl);
    }

    /**
     * Removes a listener from this multicaster and returns the
     * resulting multicast listener.
     * @param oldl the listener to be removed
     */
    protected EventListener remove(EventListener oldl) {
		if (oldl == a)  return b;
		if (oldl == b)  return a;
		EventListener a2 = removeInternal(a, oldl);
		EventListener b2 = removeInternal(b, oldl);
		if (a2 == a && b2 == b) {
			return this;	// it's not here
		}
		return addInternal(a2, b2);
    }


    /** 
     * Returns the multicast listener resulting from the addition of listener-a
     * and listener-b.<br>  
     * If listener-a is null, it returns listener-b;  
     * If listener-b is null, it returns listener-a
     * If neither are null, then it creates and returns
     * a new PEEventMulticaster instance which chains a with b.
     * @param a event listener-a
     * @param b event listener-b
     */
    protected static EventListener addInternal(EventListener a, EventListener b) {
		if (a == null)  return b;
		if (b == null)  return a;
		return new PEEventMulticaster(a, b);
    }

    /** 
     * Returns the multicast listener obtained after removing the
     * old listener from listener-l.<br>
     * If listener-l equals the old listener OR listener-l is null, 
     * returns null.<br>
     * Else if listener-l is an instance of AWTEventMulticaster, 
     * then removes the old listener from it.<br>
     * Else, returns listener l.
     * @param l the listener being removed from
     * @param oldl the listener being removed
     */
    protected static EventListener removeInternal(EventListener l, EventListener oldl) {
		if (l == oldl || l == null) {
		    return null;
		} else if (l instanceof PEEventMulticaster) {
	    	return ((PEEventMulticaster)l).remove(oldl);
		} else {
	    	return l;		// it's not here
		}
    }
      
	
    /**
     * Handles a mouseClicked event by invoking the
     * mouseClicked methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseClicked(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseClicked(e);
        ((PEMouseInputListener)b).mouseClicked(e);
    }

    /**
     * Handles the mousePressed event by invoking the
     * mousePressed methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mousePressed(PEMouseEvent e) {
        ((PEMouseInputListener)a).mousePressed(e);
        ((PEMouseInputListener)b).mousePressed(e);
    }

    /**
     * Handles the mouseReleased event by invoking the
     * mouseReleased methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseReleased(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseReleased(e);
        ((PEMouseInputListener)b).mouseReleased(e);
    }

    /**
     * Handles the mouseEntered event by invoking the
     * mouseEntered methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseEntered(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseEntered(e);
        ((PEMouseInputListener)b).mouseEntered(e);
    }

    /**
     * Handles the mouseExited event by invoking the
     * mouseExited methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseExited(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseExited(e);
        ((PEMouseInputListener)b).mouseExited(e);
    }

    /**
     * Handles the mouseDragged event by invoking the
     * mouseDragged methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseDragged(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseDragged(e);
        ((PEMouseInputListener)b).mouseDragged(e);
    }

    /**
     * Handles the mouseMoved event by invoking the
     * mouseMoved methods on listener-a and listener-b.
     * @param e the mouse event
     */
    public void mouseMoved(PEMouseEvent e) {
        ((PEMouseInputListener)a).mouseMoved(e);
        ((PEMouseInputListener)b).mouseMoved(e);
    }
	
	public String toString(){
		return super.toString() + "[a=" + a + ", b="+ b+"]";
	}
}

