/*
 JPICFormatter.java - July 29, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.io.formatter;

import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.PicPoint;

import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.Arrow;
import jpicedt.graphic.model.PicAttributeSet;
import jpicedt.graphic.model.PicCircleFrom3Points;
import jpicedt.graphic.model.PicEllipse;
import jpicedt.graphic.model.PicGroup;
import jpicedt.graphic.model.PicMultiCurve;
import jpicedt.graphic.model.PicObjectConstants;
import jpicedt.graphic.model.PicParallelogram;
import jpicedt.graphic.model.PicSmoothPolygon;
import jpicedt.graphic.model.PicPsCurve;
import jpicedt.graphic.model.PicText;
import jpicedt.graphic.model.StyleConstants;

import java.awt.Color;
import java.awt.geom.Rectangle2D;

import java.util.Iterator;
import java.util.Date;
import java.util.Properties;
import java.util.HashMap;
import java.util.Set;
import java.util.Map;


import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;

/**
 * a FormatterFactory that produces formatters for the JPIC-XML native format
 * <p>
 * Syntax for common attributes, that is, attributes not specific to a given subclass of Element, is 
 * strongly inspired from the XML-SVG markup language, although it is restricted to a subset of SVG attributes 
 * supported by jPicEdt, i.e. mainly by pstricks. 
 * Hence attributes names and values strongly mimic those of PsTricks, except for some keyword changes.
 * So far, default values are hardcoded in {@link jpicedt.graphic.model.DefaultAttributeSet DefaultAttributeSet}.
 * This may be changed in the near future, once the language has stabilized and the DtD is ready for use.
 * <p>
 * Here is a list of supported attributes as of release 1.4-pre :
 * <ul>
 * <li> stroke-style=none|dashed|dotted|solid (default : solid)
 * <li> stroke-color=#a_color_in_HEX_radix (default #000000)
 * <li> stroke-width=length_in_mm
 * <li> stroke-dasharray=a_dash_array (comma separated values)
 * <li> stroke-dotsep=length_in_mm
 * <li> fill-style=none|solid|any_PsTricks_valid_hatches_style (default : solid)
 * <li> fill-color=#a_color_in_HEX_radix (default #000000)
 * <li> shadow=true|false (default : false)
 * <li> shadow-size=length_in_mm
 * <li> shadow-angle=angle_in_degrees
 * <li> shadow-color=#a_color_in_HEX_radix (default #000000)
 * <li> hatch-width=length_in_mm
 * <li> hatch-sep=length_in_mm
 * <li> hatch-angle=angle_in_degrees
 * <li> hatch-color=#a_color_in_HEX_radix (default #000000)
 * <li> doubleline=true|false (default : false)
 * <li> doubleline-sep=length_in_mm
 * <li> doubleline-color=length_in_mm (default #000000)
 * <li> left-arrow=arrow_name_as_in_Arrow_class (default Arrow.NONE)
 * <li> right-arrow=arrow_name_as_in_Arrow_class (default Arrow.NONE)
 * <li> arrow-global-scale-width=scale_factor_for_width
 * <li> arrow-global-scale-length=scale_factor_for_length
 * <li> arrow-width-minimum=minimum_width_in_mm
 * <li> arrow-width-linewidth-scale=width_scale_factor_wrt_linewidth
 * <li> arrow-length-scale=scale_factor_for_length
 * <li> arrow-inset-scale=scale_factor_for_inset
 * <li> tbar-width-minimum=minimum_width_for_tbars_in_mm
 * <li> tbar-width-linewidth-scale=width_scale_factor_for_tbars_wrt_linewidth
 * <li> bracket-length-scale=length_scale_factor_for_brackets
 * <li> rbracket-length-scale=length_scale_factor_for_rounded_brackets
 * <li> [todo] style=name_of_a_user_defined_style (see future &lt;style&gt; tag). Any attribute defined "in place" overrides
 *      the same attribute when defined in the corresponding &lt;style&gt; tag. 
 * </ul>
 * These syntax rules are only briefly laid out here as a reminder to end-users. 
 * The corresponding DtD, once it is set up, will be the only valid specification for that matter. 
 *
 * @since jpicedt 1.4
 * @author Sylvain Reynal
 * @version $Id: JPICFormatter.java,v 1.18.2.1 2007/09/02 11:56:19 reynal Exp $ 
 */
public class JPICFormatter implements FormatterFactory, PicObjectConstants, JPICConstants {

	private final PicAttributeSet defaultAttributes = new PicAttributeSet(); // [pending] still needed?
	public static final String STROKE_DASHARRAY="stroke-dasharray";
	
	/**
	 * Return a Formatter able to format the given Element in the JPIC-XML format
	 */
	public Formatter createFormatter(Element e){
		// put daughters first !
		if (e instanceof PicCircleFrom3Points) return new JPICCircleFormatter((PicCircleFrom3Points)e,this);
		if (e instanceof PicEllipse) return new JPICEllipseFormatter((PicEllipse)e,this);
		if (e instanceof PicParallelogram) return new JPICParallelogramFormatter((PicParallelogram)e,this);
		if (e instanceof PicSmoothPolygon) return new JPICSmoothPolygonFormatter((PicSmoothPolygon)e,this);
		if (e instanceof PicMultiCurve) return new JPICMultiCurveFormatter((PicMultiCurve)e,this);
		if (e instanceof PicPsCurve) return new JPICPsCurveFormatter((PicPsCurve)e,this);
		if (e instanceof PicText) return new JPICTextFormatter((PicText)e,this);
		if (e instanceof PicGroup) return new JPICGroupFormatter((PicGroup)e,this);
		return new NonSupportedFormatter(e);
	}

	/**
	 * a formatter for Element's which are not supported yet by the JPIC-XML language ;
	 * this a reminder dedicated to "scutterbrain" developpers (like me...).
	 */
	class NonSupportedFormatter implements Formatter {
		Element element;
		NonSupportedFormatter(Element e){
			this.element = e;
		}
		public String format(){
			return "<notsupported>" + element.getName() + "</notsupported>" + CR_LF;
		}
	}

	/**
	 * Configure static fields using the given Properties object
	 * This implementation does nothing so far.
	 * @param properties used to read shared parameters. If null, default values are used.
	 */
	public static void configure(Properties preferences){
	}

	/**
	 * Return a Formatter able to format the given Drawing in the JPIC-XML language ;
	 * @param outputConstraint constraint used by this factory to create a specific Formatter on-the-fly, with
	 *        e.g. dynamic constraints on the output format. There are no constraints so far for this factory,
	 *        and this parameter may safely be set to null.
	 */
	public Formatter createFormatter(Drawing d, Object outputConstraint){
		return new DrawingFormatter(d,outputConstraint);
	}

	////////////////////////////////////////////////////////////////////
	//// Toolkit
	////////////////////////////////////////////////////////////////////

	/*
	 * Appends a string containing JPIC-XML common attributes for the given Element to the
	 * given buffer, that is,
	 * attributes specific to a given subclass of Element are not included here, but should
	 * be handled directly by the appropriate formatter.<p>
	 * Attributes handled in this method include stoke-, fill-, shadow- and double-line attributes, as well as arrows,
	 * and are appended only if they differ from the default value in PicAttributeSet.
	 * See documentation on top of this class for details on the JPIC-XML syntax.
	 * <p>
	 * Not supported yet : "border" and "bordercolor"
	 * @deprecated we now use a HashMap to avoid redundant definition, which XML forbids
	 * @since jpicedt 1.4.pre5
	 */
	 /*
	protected void appendAttributesString(StringBuffer buf, Element obj){ // moved to XMLAttributeSet

		PicAttributeSet attributes = obj.getAttributeSet();
		Color col;
		double x;
		
		buf.append(CR_LF);
		buf.append('\t'); // pretty formatting...
		// linewidth [SR:pending] add only if not the default value, this will save some space
		appendNameValuePair(buf,LINE_WIDTH,StyleConstants.getLineWidth(attributes),StyleConstants.getLineWidth(defaultAttributes));

		// linecolor
		appendNameValuePair(buf,LINE_COLOR,StyleConstants.getLineColor(attributes),StyleConstants.getLineColor(defaultAttributes));

		// linestyle
		if (StyleConstants.getLineStyle(attributes)==DASHED){
			appendNameValuePair(buf,LINE_STYLE,DASHED); // it's too complicated to distinguish b/w default and non-default values
			double[] dashArray = {StyleConstants.getDashOpaque(attributes),StyleConstants.getDashTransparent(attributes)};
			appendNameValuePair(buf,STROKE_DASHARRAY,dashArray);
		}
		else if (StyleConstants.getLineStyle(attributes)==DOTTED){
			appendNameValuePair(buf,LINE_STYLE, DOTTED);
			appendNameValuePair(buf,DOT_SEP,StyleConstants.getDotSep(attributes),StyleConstants.getDotSep(defaultAttributes));
			
		}
		else if (StyleConstants.getLineStyle(attributes)==NONE){
			appendNameValuePair(buf,LINE_STYLE, NONE);
		}
		//else buf.append(" stroke-style=\"solid\""); // default !

		// double line
		if (StyleConstants.isDoubleLine(attributes)) {
			appendNameValuePair(buf,DOUBLE_LINE,true);
			appendNameValuePair(buf,DOUBLE_SEP,StyleConstants.getDoubleLineSep(attributes),StyleConstants.getDoubleLineSep(defaultAttributes));
			appendNameValuePair(buf,DOUBLE_COLOR,StyleConstants.getDoubleLineColor(attributes),StyleConstants.getDoubleLineColor(defaultAttributes));
		}

		// fill style and fill colour
		buf.append(CR_LF);buf.append('\t');
		String fillStyle = StyleConstants.getFillStyle(attributes);
		if (fillStyle==NONE){
			appendNameValuePair(buf,FILL_STYLE,NONE);
		}
		else { // solid or hatches -> add fill colour (even if it's redundant for hatches that are not "starred"=
			appendNameValuePair(buf,FILL_COLOR,StyleConstants.getFillColor(attributes),StyleConstants.getFillColor(defaultAttributes));
			if (fillStyle==SOLID) appendNameValuePair(buf,FILL_STYLE,SOLID);
			else { // hatches
				if (fillStyle==VLINES) appendNameValuePair(buf,FILL_STYLE,VLINES);
				else if (fillStyle==VLINES_FILLED) appendNameValuePair(buf,FILL_STYLE,VLINES_FILLED);
				else if (fillStyle==HLINES) appendNameValuePair(buf,FILL_STYLE,HLINES);
				else if (fillStyle==HLINES_FILLED) appendNameValuePair(buf,FILL_STYLE,HLINES_FILLED);
				else if (fillStyle==CROSSHATCH) appendNameValuePair(buf,FILL_STYLE,CROSSHATCH);
				else if (fillStyle==CROSSHATCH_FILLED) appendNameValuePair(buf,FILL_STYLE,CROSSHATCH_FILLED);

				appendNameValuePair(buf,HATCH_WIDTH,StyleConstants.getHatchWidth(attributes),StyleConstants.getHatchWidth(defaultAttributes));
				appendNameValuePair(buf,HATCH_SEP,StyleConstants.getHatchSep(attributes),StyleConstants.getHatchSep(defaultAttributes));
				appendNameValuePair(buf,HATCH_ANGLE,StyleConstants.getHatchAngle(attributes),StyleConstants.getHatchAngle(defaultAttributes));
				appendNameValuePair(buf,HATCH_COLOR,StyleConstants.getHatchColor(attributes),StyleConstants.getHatchColor(defaultAttributes));
			}
		}
		// shadow
		if (StyleConstants.isShadow(attributes)){
			buf.append(CR_LF);buf.append('\t');
			appendNameValuePair(buf,SHADOW,true);
			appendNameValuePair(buf,SHADOW_SIZE,StyleConstants.getShadowSize(attributes),StyleConstants.getShadowSize(defaultAttributes));
			appendNameValuePair(buf,SHADOW_ANGLE,StyleConstants.getShadowAngle(attributes),StyleConstants.getShadowAngle(defaultAttributes));
			appendNameValuePair(buf,SHADOW_COLOR,StyleConstants.getShadowColor(attributes),StyleConstants.getShadowColor(defaultAttributes));
		}

		// arrows
		buf.append(CR_LF);buf.append('\t');
		Arrow leftArrow = (Arrow)obj.getAttribute(LEFT_ARROW);
		Arrow rightArrow = (Arrow)obj.getAttribute(RIGHT_ARROW);
		if (leftArrow != Arrow.NONE) appendNameValuePair(buf,LEFT_ARROW,leftArrow.getName());
		if (rightArrow != Arrow.NONE) appendNameValuePair(buf,RIGHT_ARROW,rightArrow.getName());
		if (leftArrow != Arrow.NONE || rightArrow != Arrow.NONE){
			appendNameValuePair(buf,ARROW_GLOBAL_SCALE_WIDTH,StyleConstants.getArrowGlobalScaleWidth(attributes),StyleConstants.getArrowGlobalScaleWidth(defaultAttributes));
			appendNameValuePair(buf,ARROW_GLOBAL_SCALE_LENGTH,StyleConstants.getArrowGlobalScaleLength(attributes),StyleConstants.getArrowGlobalScaleLength(defaultAttributes));
			appendNameValuePair(buf,ARROW_WIDTH_MINIMUM_MM,StyleConstants.getArrowWidthMinimumMm(attributes),StyleConstants.getArrowWidthMinimumMm(defaultAttributes));
			appendNameValuePair(buf,ARROW_WIDTH_LINEWIDTH_SCALE,StyleConstants.getArrowWidthLineWidthScale(attributes),StyleConstants.getArrowWidthLineWidthScale(defaultAttributes));
			appendNameValuePair(buf,ARROW_LENGTH_SCALE,StyleConstants.getArrowLengthScale(attributes),StyleConstants.getArrowLengthScale(defaultAttributes));
			appendNameValuePair(buf,ARROW_INSET_SCALE,StyleConstants.getArrowInsetScale(attributes),StyleConstants.getArrowInsetScale(defaultAttributes));
			appendNameValuePair(buf,TBAR_WIDTH_MINIMUM_MM,StyleConstants.getTBarWidthMinimumMm(attributes),StyleConstants.getTBarWidthMinimumMm(defaultAttributes));
			appendNameValuePair(buf,TBAR_WIDTH_LINEWIDTH_SCALE,StyleConstants.getTBarWidthLineWidthScale(attributes),StyleConstants.getTBarWidthLineWidthScale(defaultAttributes));
			appendNameValuePair(buf,BRACKET_LENGTH_SCALE,StyleConstants.getBracketLengthScale(attributes),StyleConstants.getBracketLengthScale(defaultAttributes));
			appendNameValuePair(buf,RBRACKET_LENGTH_SCALE,StyleConstants.getRoundedBracketLengthScale(attributes),StyleConstants.getRoundedBracketLengthScale(defaultAttributes));
			appendNameValuePair(buf,POLYDOTS_SIZE_MINIMUM_MM,StyleConstants.getPolydotsSizeMinimumMm(attributes),StyleConstants.getPolydotsSizeMinimumMm(defaultAttributes));
			appendNameValuePair(buf,POLYDOTS_SIZE_LINEWIDTH_SCALE,StyleConstants.getPolydotsSizeLineWidthScale(attributes),StyleConstants.getPolydotsSizeLineWidthScale(defaultAttributes));
		}
		
		// overstrike :
		if (StyleConstants.isOverStrike(attributes)){
			appendNameValuePair(buf, OVER_STRIKE, true);
			appendNameValuePair(buf, OVER_STRIKE_WIDTH, StyleConstants.getOverStrikeWidth(attributes), StyleConstants.getOverStrikeWidth(defaultAttributes));
			appendNameValuePair(buf, OVER_STRIKE_COLOR, StyleConstants.getOverStrikeColor(attributes),StyleConstants.getOverStrikeColor(defaultAttributes));
		}

		// custom pstricks parameters
		if (StyleConstants.getPstCustom(attributes).length()>0){
			appendNameValuePair(buf,PST_CUSTOM,StyleConstants.getPstCustom(attributes));
		}
		
		// misc params
		if (StyleConstants.getDimen(attributes)==INNER) appendNameValuePair(buf,DIMEN,INNER);
		else if (StyleConstants.getDimen(attributes)==MIDDLE) appendNameValuePair(buf,DIMEN,MIDDLE);
		// else default OUTER
	}*/

	/*
	 * Given attribute name and value, append a "name=\"value\"" string to the given buffer.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, String value){
		appendName(buf,name);
		appendValue(buf,value);
	}*/

	/*
	 * Append the given attribute name to the given buffer
	 */
	/*protected void appendName(StringBuffer buf, Object name){
		buf.append(' ');
		buf.append(name.toString());
		buf.append('=');
	}*/

	/*
	 * Append the given attribute value to the given buffer
	 */
	/*protected void appendValue(StringBuffer buf, String value){
		buf.append('\"');
		buf.append(value);
		buf.append('\"');
	}*/

	/*
	 * Given attribute name and value, append a "name=\"value\"" string to the given buffer if value differs from the given default value.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, Color col, Color def){
		if (def!=null && def.equals(col)) return;
		appendNameValuePair(buf, name, colorToHex(col));
	}*/

	/*
	 * Given attribute name and value, append a "name=\"value\"" string to the given buffer.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, boolean bol){
		appendNameValuePair(buf, name, (bol==true ? "true" : "false"));
	}*/

	/*
	 * Append a name="(x,y)" string to the given buffer
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, PicPoint pt){
		appendNameValuePair(buf, name, pt.toString());
	}*/

	/*
	 * Given attribute name and value, append a "name=\"value\"" string to the given buffer.
	 * double are formatted with two-digit precision.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, double value){
		appendNameValuePair(buf, name, PEToolKit.doubleToString(value));
	}*/

	/*
	 * Given attribute name and value, append a "name=\"value\"" string to the given buffer if value differs from the given default value.
	 * double are formatted with two-digit precision.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, double value, double def){
		if (def==value) return;
		appendNameValuePair(buf, name, PEToolKit.doubleToString(value));
	}*/

	/*
	 * Given attribute name and an array of doubles, append a "name=\"value\"" string to the given buffer.
	 * double are formatted with two-digit precision, and are separated by semicolons.
	 */
	/*protected void appendNameValuePair(StringBuffer buf, Object name, double[] values){
		buf.append(' ');
		buf.append(name);
		buf.append('=');
		buf.append('\"');
		for (int i=0; i<values.length-1; i++){
			buf.append(PEToolKit.doubleToString(values[i]));
			buf.append(";");
		}
		buf.append(PEToolKit.doubleToString(values[values.length-1]));
		buf.append('\"');
	}*/

	/*
	 * Convert the given Color to a XML-like RGB string, each colour being expressed in the hexadecimal radix 
	 */
	/*protected String _colorToHex(Color c){
		String red,green,blue;
		if (c.getRed()==0) red="00"; else red = Integer.toHexString(c.getRed());
		if (c.getGreen()==0) green="00"; else green = Integer.toHexString(c.getGreen());
		if (c.getBlue()==0) blue="00"; else blue = Integer.toHexString(c.getBlue());
		StringBuffer buf = new StringBuffer("#");
		buf.append(red);
		buf.append(green);
		buf.append(blue);
		return buf.toString();
	}*/

	// --- same with HashMap (allows us to avoid double definition of same parameter, which is forbidden by XML) ---
	
	
	////////////////////////////////////////////////////////////////////////////////////
	//// DRAWING
	////////////////////////////////////////////////////////////////////////////////////

	/**
	 *  A Drawing's formatter for the JPIC-XML language. 
	 */
	protected class DrawingFormatter implements Formatter {

		/** the Drawing this formatter acts upon */
		private Drawing drawing;
		private Object outputConstraint;

		/**
		 * init this formatter for the given drawing
		 * @param outputConstraint if MAKE_STANDALONE_FILE, prepend/append prolog and epilog.
		 */
		public DrawingFormatter(Drawing drawing,Object outputConstraint){
			this.drawing = drawing;
			this.outputConstraint = outputConstraint;
		}

		/**
		 * @return  a String representing this Drawing in the PsTricks format
		 * !!! PsTricks's default unit = 1cm !!!
		 */
		public String format(){

			StringBuffer buf = new StringBuffer(200);

			buf.append("%Created by jPicEdt "+jpicedt.Version.getVersion()+": mixed JPIC-XML/LaTeX format");
			buf.append(CR_LF);
			buf.append("%");
			buf.append(new Date());
			buf.append(CR_LF);
			buf.append("%Begin JPIC-XML");
			buf.append(CR_LF);
			buf.append("<?xml version=\"1.0\" standalone=\"yes\"?>");
			buf.append(CR_LF);
			// wait DtD is ok to uncomment :
			// buf.append("<!DOCTYPE jpic PUBLIC \"-JPIC 1.0//EN\" \"http://www.jpicedt.org/jpic.dtd\">");


			// start document with <jpic> tag :
			buf.append("<jpic");

			/* first we compute the picture size from the drawing's bounding box */

			Rectangle2D box = drawing.getBoundingBox(); // recursively calls getBounds2D() on each Element [pending] annoying bug with PicText objects !
			//jpicedt.Log.debug("bbox = "+box);
			buf.append(" x-min=\"");
			buf.append(PEToolKit.doubleToString(box.getMinX()));
			buf.append("\" x-max=\"");
			buf.append(PEToolKit.doubleToString(box.getMaxX()));
			buf.append("\" y-min=\"");
			buf.append(PEToolKit.doubleToString(box.getMinY()));
			buf.append("\" y-max=\"");
			buf.append(PEToolKit.doubleToString(box.getMaxY()));
			buf.append("\" auto-bounding=\"");
			buf.append(drawing.isAutoComputeBoundingBox() ? "true" : "false");
			buf.append("\">");
			buf.append(CR_LF);

			/*  then for each Element in the Drawing, we call "toFormatedString" and append the returned String to our buffer */
			for (Iterator it=drawing.elements(); it.hasNext(); ){
				Element e = (Element)it.next();
				buf.append(createFormatter(e).format());
			}
			//epilogue
			buf.append("</jpic>");
			buf.append(CR_LF);
			buf.append("%End JPIC-XML");
			buf.append(CR_LF);
			
			// latexify...
			return latexifyText(buf.toString());
		}

		/**
		 * Prepend a "%" sign at the beginning of each line so that LaTeX does not get confused.
		 */
		private String latexifyText(String str) {
			StringBuffer output = new StringBuffer(100);
			BufferedReader input = new BufferedReader(new StringReader(str)); // gives access to a readLine() method which handles CR-LF properly
			String line=null;
			try {
				while ((line=input.readLine())!=null){
					// [SR:pending] hard wrap to 80 columns
					output.append('%');
					output.append(line);
					output.append(CR_LF);
				}
			}
			catch (IOException ioEx){}
			return output.toString();
		}
	}

}
