/*
 JPICParser.java - August 1, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.io.parser;

import jpicedt.graphic.model.Drawing;
import java.io.Reader;
import java.io.StringReader;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;

/**
 * An implementation of the Parser interface dedicated to parsing JPIC-XML files.<p>
 * <b>How it works :</b>
 * We first try to know if we're parsing a JPIC-XML formatted file, i.e. starting with a "%%Begin JPIC-XML" marker.
 * If this is the case :
 * <ul>
 * <li> look up matching "%%End JPIC-XML" marker
 * <li> save text between both markers, and drop leading "%"
 * <li> pass corresponding text to our JPIC-XML parser (implementation based on a Java-SAX parser).
 * <li> return parsed Drawing (parsing process then stops here)
 * </ul>
 * Otherwise, go ahead using the fallback Parser, if any.
 * @author Sylvain Reynal
 * @since jpicedt 1.3.3
 * @version $Id: JPICParser.java,v 1.6.2.1 2007/09/02 11:56:20 reynal Exp $
 */
public class JPICParser implements Parser {
	
	// the fallback Parser that gets called whenever the JPIC-XML parsing process fails.
	private Parser fallback;
	
	/**
	 * kicks off the parsing process
	 * @return a new instance of a Drawing populated from the content of the given reader
	 * @throws ParserException when a parsing error was encountered, e.g. syntax error, block mismatch,...
	 */
	public Drawing parse(Reader reader) throws ParserException {
		// first save stream content to a StringBuffer, for later use by fallback :
		String inputString = readerToString(reader); // now reader is UNUSABLE because it's been totally read
		// then try to extract a JPIC-XML text from the input string ; if the returned String is null,
		// this is not a JPIC-XML formatted content :
		String xmlStr = extractXMLText(inputString);
		if (xmlStr == null){
			if (fallback != null) {
				return fallback.parse(new StringReader(inputString));
			}
			else throw new ParserException.UnrecognizedFileFormat();
		}
		else {
			try {
				// ok, we have got an XML-ready String ; we must pass it to a SAX-parser as an InputSource object...
				InputSource inputSrc = new InputSource(new StringReader(xmlStr));
				
				// instanciate SAX parser (following hints given in the "Java Tutorial on XML") :
				SAXParserFactory saxFactory = SAXParserFactory.newInstance(); // returns an org.apache.xerces.jaxp.SAXParserFactoryImpl when using SUN or IBM JVM's
					// possibly configure factory here to adapt parsing behaviours :
					//saxFactory.setNamespaceAware(true/false); // support XML namespacing ?
					//saxFactory.setValidating(true/false); // use DtD ?
				SAXParser xmlParser = saxFactory.newSAXParser(); // we don't use separate handlers, hence don't rely on the encapsulated XMLReader.
				
				// pass it our own content-handler :
				JPICXmlHandler xmlHandler = new JPICXmlHandler();
				// initiate the parsing process :
				xmlParser.parse(inputSrc, xmlHandler);
				// fetch parsed Drawing from our Handler
				Drawing parsedDrawing = xmlHandler.fetchParsedDrawing();
				return parsedDrawing;
			}
			catch (ParserConfigurationException pce){ // thrown by SAXParserFactory methods
				throw new ParserException(pce);
			}
			catch (SAXParseException se){ // [SR:pending] use getLineNumber() and getColumnNumber()
				throw new ParserException(se);
			}
			catch (SAXException se){
				throw new ParserException(se);
			}
			catch (IOException ioEx){
				throw new ParserException(ioEx);
			}
			catch (NumberFormatException nfe){
				throw new ParserException(nfe);
			}
		}
	}
	
	/**
	 * for quick debugging purpose
	 */
	public static void main(String[] args){
		if (args.length == 0){
			System.out.println("Usage : java jpicedt.graphic.io.parser.JPICParser filename");
			System.exit(0);
		}
		try {
			BufferedReader reader = new BufferedReader(new FileReader(args[0]));
			JPICParser parser = new JPICParser();
			parser.setFallbackParser(new jpicedt.format.input.DefaultParser());
			Drawing dr = parser.parse(reader);
			System.out.println(dr);
		}
		catch (Exception ex){
			System.out.println(ex);
		}
	}
	
	/**
	 * set the Parser that'll be called when JPIC-XML parsing fails. This may be, e.g. a PsTricks parser, or
	 * whatsoever.
	 */
	public void setFallbackParser(Parser fallback){
		this.fallback = fallback;
	}
	
	/**
	 * Returns the Parser that'll be called when JPIC-XML parsing fails. This may be, e.g. a PsTricks parser, or
	 * whatsoever.
	 */
	public Parser getFallbackParser(){
		return this.fallback;
	}

	/**
	 * Returns a String initialized from the given Reader
	 */
	private String readerToString(Reader reader){
		StringBuffer output = new StringBuffer(100);
		BufferedReader input = new BufferedReader(reader); // gives access to a readLine() method which handles CR-LF properly
		String line=null;
		try {
			while ((line=input.readLine())!=null){
				output.append(line);
				output.append('\n'); // convert CR to Unix-like CR
			}
		}
		catch (IOException ioEx){}
		return output.toString();
	}
	
	/**
	 * Transform the given input String by removing any leading "%"
	 * <br>
	 * Used so far by extractXMLText only.
	 */
	private String removeLeadingCommentPrefix(String str){
		if (str.startsWith("%")) return str.substring(1);
		else return str;
	}
	
	/**
	 * Try to extract a JPIC-XML text from the given reader, by looking up markers of the
	 * form "%%Begin JPIC-XML" and "%%End JPIC-XML", and removing any leading "%" at the 
	 * beginning of each line. The returned String may then be fed directly to a SAX or DOM parser.
	 * @return null if the extraction process failed, which probably means JPIC-XML starting or ending
	 *         markers were not found. This signals the receiver it may have to resort to the fallback parser, if any.
	 */
	private String extractXMLText(String inputString) {
		StringBuffer output = new StringBuffer(100);
		BufferedReader input = new BufferedReader(new StringReader(inputString)); // gives access to a readLine() method which handles CR-LF properly
		String line=null;
		try {
			while ((line=input.readLine())!=null){
				if (line.startsWith("%%Begin JPIC-XML")){
					while ((line=input.readLine())!=null){
						if (line.startsWith("%%End JPIC-XML")){
							return output.toString();
						}
						else {
							output.append(removeLeadingCommentPrefix(line));
							output.append('\n'); // append Unix-like CR-LF = no problemo : 
							// we don't care if it's not the current platform's CR-LF, cause 
							// this StringBuffer will be used by our SAX parser only, and 
							// this one knows how to handle any type of CR-LF properly
						}
					}
				}
			}
			return null; // means extraction process failed
		}
		catch (IOException ioEx){
			return null;
		}
	}
	 
}


