/*
 AbstractElement.java - February 18, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.view.*;
import jpicedt.graphic.event.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;


/**
 * An abstract class that can serve as the basis class for elements of a Drawing.<p>
 * This implementation provides some useful behaviour where the integration of Element's in a tree-model
 * is concerned.  
 * @author Sylvain Reynal
 * @since PicEdt 1.0
 */
public abstract class AbstractElement implements Element {

	////////////////////////////
	//// PROTECTED MEMBERS
	///////////////////////////
	/** the parent of this element */
	protected Element parent;
	/** the attribute set */
	protected PicAttributeSet attributeSet;
	/** the view that render this element */
	protected View view;
	/**
	 * construct an AbstractElement with no parent and a default PicAttributeSet
	 */
	public AbstractElement(){
		attributeSet = new PicAttributeSet();
	}

	/**
	 * construct an AbstractElement with no parent and the given PicAttributeSet.
	 */
	public AbstractElement(PicAttributeSet attributeSet){
		this.attributeSet = new PicAttributeSet(attributeSet);
	}

	/**
	 * cloning constructor ; attribute set is deeply copied.
	 * The parent and the view are set to null. 
	 */
	public AbstractElement(AbstractElement obj){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("cloning " + obj);
		this.attributeSet = new PicAttributeSet(obj.getAttributeSet());
	}

	/**
	 * Returns a clone of this Element
	 */
	public abstract Object clone();
	

	///////////////////////////////////////
	//// tree structure
	///////////////////////////////////////
	
    /**
     * Retrieves the underlying drawing
     * @return the drawing ; null if this AbstractElement doesn't belong to any drawing yet.
     */
	public Drawing getDrawing() {
		Element p = getParent();
		return (p!=null) ? p.getDrawing() : null;
	}

    /**
     * Gets the parent of the element.
     * @return the parent
     */
	public Element getParent() {
	    return parent;
	}

    /**
     * Sets the parent of this element to be the given element.
     */
	public void setParent(Element p) {
		if (p!=null && !p.getAllowsChildren()) throw new IllegalArgumentException("Can't set a parent that doesn't allow children !");
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("parent=" + p + "\n\tthis="+this);
		this.parent = p;
	}

	/////////////////////////////////////
	/// EVENTS
	////////////////////////////////////
	
	/**
	 * Must be called each time this AbstractElement changes. 
	 * This default implementation updates the associated View if any,
	 * then propagates upward along the tree by calling 
	 * <code>forwardChangedUpdate(this)</code> on the Element's parent, if any.
	 * <br>
	 * Tree's root-element should override so as to fire the change to the hosting drawing.
	 * <br>
	 * If subclasser are willing to override this method, they should call <code>super.fireChangeUpdate</code>.
	 * @param type the event type
	 */
	 protected void fireChangedUpdate(DrawingEvent.EventType eventType){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType+", this="+toString());
		if (view != null) view.changedUpdate(eventType);
		Element p = getParent();
		if (p!=null) p.forwardChangedUpdate(this,eventType); // forward to parent , package restricted access.
	}
	

		
	////////////////////////////////////
	//// GEOM.
	///////////////////////////////////

	/**
	 * Create an Iterator over points that can serve as anchor points for grid alignment. 
	 *         This implementation returns a DefaultPointIndexIterator which simply iterates through all
	 *         user-controlled points in ascending order.
	 */
	public PointIndexIterator anchorPointsIterator(){
		return new DefaultPointIndexIterator(this);
	}


	/**
	 * Scale this object by (sx,sy) using ptOrg as origin
	 * sx and sy can be negative. This is a convenience call to 
	 * <code>scale(double,double,double,double)</code>.
	 */
	public void scale(PicPoint ptOrg, double sx, double sy){
		scale(ptOrg.x, ptOrg.y, sx, sy);
	}
	
	
	
	/////////////////////////////////
	/// VIEW
	/////////////////////////////////

	/**
	 * Returns the View that's responsible for rendering this AbstractElement
	 */
	public View getView(){
		return view;
	}
	
	/**
	 * set the view for this AbstractElement from the given view factory
	 * or remove the view if f is null.
	 */
	public void setViewFromFactory(ViewFactory f){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("factory="+f);
		removeView(); // remove cross-reference
		if (f!=null) view = f.createView(this); 
	}

	/**
	 * remove the view that render this element ; this may be used to remove any reference to the view,
	 * and render it eligible for garbage collection ; if no View, does nothing.
	 */
	public void removeView(){
		if (view == null) return;
		view.setHighlighter(null); // remove highlighter to avoid GC deadlocks
		view.setElement(null);
		view = null;
	}


	/////////////////////////////////
	/// ATTRIBUTE SET
	/////////////////////////////////

	/**
	 * Returns the AttributeSet bound to this AbstractElement.
	 * <p>
	 * Be careful that modifying attributes using this method does not fire any DrawingEvent's, hence
	 * does not keep the view synchronized with the state of the model.
	 */
	public PicAttributeSet getAttributeSet(){ 
		return attributeSet;
	}

	/**
	 * Bind the given attributes set to this Element. This implementation actually makes a deep copy of
	 * the given set beforehands.
	 */
	public void setAttributeSet(PicAttributeSet attributeSet){ 
		this.attributeSet = new PicAttributeSet(attributeSet);
		fireChangedUpdate(DrawingEvent.EventType.ATTRIBUTE_CHANGE);
	}

	/**
	 * Set the given attribute name to the given value for this AbstractElement
	 */
	public void setAttribute(PicAttributeName name, Object value){
		attributeSet.setAttribute(name,value);
		fireChangedUpdate(DrawingEvent.EventType.ATTRIBUTE_CHANGE);
	}

	/**
	 * Returns the value for the given attribute name
	 */
	public Object getAttribute(PicAttributeName name){
		return attributeSet.getAttribute(name);
	}

	///////////////////////////////
	//// STRING FORMATING
	///////////////////////////////

	/**
	 * Returns a String representation of the attribute set for this AbstractElement 
	 */
	public String toString(){ 
		String s = "[" + getName() + "@" + Integer.toHexString(hashCode()) + ":\n\t" + attributeSet.toString(); 
		return s;
	}
} 

