/*
 DefaultLeafElement.java - January 19, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.view.*;
import jpicedt.graphic.event.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;


/**
 * This is a default implementation of AbstractElement targetting leaf Element's, i.e. Element's that
 * can NOT have children.<br> 
 * The <b>element's geometry</b> is backed by an array of coordinates (labelled <b>specification points</b>)
 * containing the minimum number of pair of
 * coordinates required to fully specify the shape of the element. For convenience, this array 
 * has been made accessible by means of the <code>getSpecificationPointX/Y()</code> methods : this may be used
 * e.g. by an associated View to create the appropriate java.awt.Shape if required.<p> 
 * Even if, as of this implementation, specification points act as user-controlled point, 
 * there may be more control points however (e.g. for a rectangle, there are two "specification" points, 
 * yet five control points, including the center). 
 * @author Sylvain Reynal
 * @since jpicedt 1.3.3
 */
public class DefaultLeafElement extends AbstractElement {

	////////////////////////////
	//// PROTECTED MEMBERS
	///////////////////////////
	/** an ArrayList containing PicPoint's representing the geometry of this element (specification points) */ 
	protected ArrayList pts=new ArrayList();

	// --- experimental: pst-nodes ---
    /** name of node for pstricks export */
    protected String nodeName ;
	/** a list of node-connection this element is attached to */
	protected ArrayList nodeConnections; // list of PicNodeConnection's

	/**
	 * construct an Element with no parent and a default PicAttributeSet, and no specification points.
	 */
	public DefaultLeafElement(){
		this(0);
	}

	/**
	 * construct a Element with no parent and a default PicAttributeSet, and the given
	 * number of specification points.
	 */
	public DefaultLeafElement(int nPts){
		super();
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init> : npts="+nPts);
		for (int i=0; i<nPts; i++) pts.add(new PicPoint());
		this.nodeConnections = new ArrayList();
	}

	/**
	 * construct a DefaultElement with no parent and the given PicAttributeSet,
	 * and the given number of specification points.
	 */
	public DefaultLeafElement(int nPts, PicAttributeSet attributeSet){
		this(nPts);
		this.attributeSet = new PicAttributeSet(attributeSet);
	}

	/**
	 * cloning constructor ; specification points and attribute set are deeply copied.
	 * The parent and the view are set to null. 
	 */
	public DefaultLeafElement(DefaultLeafElement obj){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init> : cloning " + obj);
		for (Iterator it = obj.pts.iterator(); it.hasNext();) 
			pts.add(new PicPoint((PicPoint)it.next()));
		this.attributeSet = new PicAttributeSet(obj.getAttributeSet());
		this.nodeConnections = new ArrayList();
	}

	/**
	 * Returns a non-localised string representing this object's name
	 * This default implementation returns the class name.
	 */
	public String getName(){
		return getClass().getName();
	}
	
	/**
	 * Returns a clone of this element
	 */
	public Object clone(){
		return new DefaultLeafElement(this);
	}
	
	///////////////////////////// 
	// node connection management 
	/////////////////////////////
	
	/**
	 * Returns an iterator over all the node connections
	 */
	public Iterator iterator(){
		return this.nodeConnections.iterator();
	}

    /**
	 * Are we a Node ?(ie do we have connections ?)
	 */
	 public boolean isNode(){
	    return !this.nodeConnections.isEmpty();
	}

    /** sets node name. If argument <code>name</code> is null, create a new name.*/

    public void setNodeName(String name){
	if (name !=null) nodeName=name;
	else {
	    nodeName="Node"+Math.floor(Math.random()*100000);
		// [pending:SR] while (nodeDB.exist(nodeName)) etc. etc.
	    //[pending:VG] si t'as un moyen simple de faire des noms de node + petits et dont on soit sur qu'ils se recoupent pas, ca serait mieux...
	}
    }

    /** gets node name. If nodeName is null, create a new name.*/

    public String getNodeName(){
	if (nodeName == null) setNodeName(null);
	return nodeName;
    }


        /**
	 *  Can we be made into a node ?
	 *  to be overriden by subclass
	 */
    
        public boolean isNodeable(){
	    return false;
	}


	public void addConnection(PicNodeConnection edge){
		this.nodeConnections.add(edge);
	}
	
	public void removeConnection(PicNodeConnection edge){
		this.nodeConnections.remove(edge);
	}
	
	public void removeAllConnections(){
		this.nodeConnections.clear();
	}

    // dummy node methods
    //[pending] is it the right way ?
    public PicPoint nodeReferencePoint(){
	return null;
    }
    public double nodeReferencePointX(){
	return Double.NaN;
    }
    public double nodeReferencePointY(){
	return Double.NaN;
    }
    public PicPoint nodeConnectionOrigin(double dx,double dy){
	return null;
    }

	///////////////////////////////////////
	//// tree structure
	///////////////////////////////////////

	/**
	 * Returns true if the receiver allows children. This implementation returns false.
	 */
	public boolean getAllowsChildren(){
		return false;
	}

	/**
	 * Called by a child of this element to inform its parent of some change that occured to it or one of its children.
	 * This gives a chance to the receiver to update its layout, then to propagate the change-event upward.<p>
	 * This implementation does nothing, since this method can never be called.
	 * @param eventType the event type
	 * @param child the child that sent the change-event. 
	 */
	public void forwardChangedUpdate(Element child,DrawingEvent.EventType eventType){
	}
	
	/**
	 * Overriden in order to support forwarding to node connections.
	 * @since jpicedt 1.4pre5
	 */
	 protected void fireChangedUpdate(DrawingEvent.EventType eventType){
		 super.fireChangedUpdate(eventType);
		 //System.out.println("######leaf element changeupdate. connections="+nodeConnections);//debug
		 if (nodeConnections==null) return;
		 for (Iterator it=nodeConnections.iterator(); it.hasNext();){
			 PicNodeConnection edge = (PicNodeConnection)it.next();
			 edge.forwardChangedUpdate(this,eventType);
		 }
	}
	

	////////////////////////////////////
	//// GEOM.
	///////////////////////////////////

	/**
	 * Return the X-coordinate of the specification point with the given index, by calling
	 * {@link #getSpecificationPoint(int). getSpecificationPoint(index)}.
	 * @param index the point index, should be greater or equal to zero, and lower or equal
	 *        to the index returned by <code>getNumberOfSpecificationPoints()</code>.
	 * @since jpicedt1.3.3
	 */
	public double getSpecificationPointX(int index){
		return getSpecificationPoint(index).x;
	}
	
	/**
	 * Return the Y-coordinate of the specification point with the given index, by calling
	 * {@link #getSpecificationPoint(int). getSpecificationPoint(index)}.
	 * @param index the point index, should be greater or equal to zero, and lower or equal
	 *        to the index returned by <code>getNumberOfSpecificationPoints()</code>.
	 * @since jpicedt1.3.3
	 */
	public double getSpecificationPointY(int index){
		return getSpecificationPoint(index).y;
	}

	/**
	 * Return the <b>specification-point</b> having the given index. This implementation 
	 * exactly mimics the general contract for
	 * <code>getPoint()</code> in interface Element, and simply clones a PicPoint obtained by a call to 
	 * {@link #getSpecificationPoint(int). getSpecificationPoint(index)}.
	 * @return the specification-point indexed by <code>index</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param index the point index, should be greater or equal to zero, and lower or equal
	 *        to the index returned by <code>getNumberOfSpecificationPoints()</code>.
	 * @since jpicedt1.3.3
	 */
	public PicPoint getSpecificationPoint(int index, PicPoint src){
		if (src==null) src = new PicPoint();
		src.setCoordinates(getSpecificationPoint(index));
		return src;
	}

	/**
	 * Return a reference on the specification point with the given index.
	 * This may be used when subclassers wants to apply a transformation to the given point (e.g. PicPoint.translate())
	 * w/o having to resort to an intermediate buffer.
	 * <p>
	 * It is enough to override this method if one wants to change
	 * the behaviour of all other getSpecificationPoint() methods at once (typical applications include :
	 * carrying out security checks on the given index, applying a geometric transformation on the PicPoint to
	 * be returned, creating index aliases, ...).
	 * <p>
	 * [developpers] protecting access is MAN-DA-TO-RY ! I don't trust ourselves enough to let us play with references from outta-hood ;-)
	 */
	protected PicPoint getSpecificationPoint(int index){
		return (PicPoint)pts.get(index);
	}
	
	/**
	 * Return the number of specification points backing the geometry of this element.
	 */
	public int getNumberOfSpecificationPoints(){
		return pts.size();
	}

	/**
	 * Returns the index of the first user-controlled point that can be retrieved by <code>getPoint()</code>.
	 * This default implementation returns 0.
	 */
	public int getFirstPointIndex(){
		return 0;
	}
	
	/**
	 * Returns the index of the last user-controlled point that can be retrieved by <code>getPoint()</code>.
	 * This default implementation returns the number of specification points minus one.
	 */
	public int getLastPointIndex(){
		return pts.size()-1;
	}

	/**
	 * Return the <b>user-controlled point</b> having the given index. The general contract in Element
	 * is to return an IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method.<br>
	 * This implementation simply call getSpecificationPoint() with the same arguments.
	 * @return the point indexed by <code>numPoint</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 */
	public PicPoint getPoint(int numPoint, PicPoint src){
		return getSpecificationPoint(numPoint, src);
	}
	
	/**
	 * Convenience call to getPoint(numPoint,null), i.e., a new Point is allocated on the fly.
	 */
	public PicPoint getPoint(int numPoint){
		return getPoint(numPoint,null);
	}
	
	/**
	 * This default implementation returns the X-coord of the specification-point with the same index. 
	 * This might be a valid implementation as long as subclasses don't have other user-controlled points.
	 * @return the X-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointX(int index){
		return getSpecificationPointX(index);
	}

	/**
	 * This default implementation returns the Y-coord of the specification-point with the same index. 
	 * This might be a valid implementation as long as subclasses don't have other user-controlled points.
	 * @return the Y-coord of the point indexed by <code>index</code>. 
	 * @param index the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since PicEdt 1.0
	 */
	public double getPointY(int index){
		return getSpecificationPointY(index);
	}
	
	/**
	 * Set the user-controlled point with the given index to the given value.
	 * This default implementation simply call setSpecificationPoint, then fires a change-update of
	 * type GEOMETRY_CHANGE
	 * @deprecated use setPoint(index,pt,constraint) instead
	 */
	public void setPoint(int index, PicPoint pt){
		setPoint(index,pt,null);
	}
	
	/**
	 * Set the user-controlled point indexed by "index" to the given value, using the specified geometrical constraint.
	 * Constraints depend on the particular concrete implementation, and may involve restricting movement along
	 * a particular direction, moving several points at once to preserve parallelism,... The default constraint 
	 * in this implementation reduces to simply calling setSpecificationPoint, then firing a GEOMETRY_CHANGE change-update. 
	 * @param constraint a geometry constraint, or null if no particular constraint is being imposed (aka default).
	 */
	public void setPoint(int index, PicPoint pt, EditPointConstraint constraint){
		setSpecificationPoint(index,pt);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	
	/**
	 * set the specification point with the given index to the given location.
	 * This method does NOT fire any DrawingEvent ! It's up to receivers to do so, e.g. setPoint().
	 * <br>
	 * [developpers] this method is protected because i don't want to break the general contract in Element, which
	 * says that the getters'n setters for the geometry are throught getPoint/setPoint methods, possibly using
	 * appropriate constraints. This method is thus just a convenience to you.
	 */
	protected void setSpecificationPoint(int index, PicPoint pt){
		getSpecificationPoint(index).setCoordinates(pt);
	}
	
	/**
	 * Translate this Element by (dx,dy) ; this implementation translates the specification-points,
	 * then fires a changed-update event.
	 * @param dx The X coordinate of translation vector
	 * @param dy The Y coordinate of translation vector
	 * @since PicEdt 1.0
	 */
	public void translate(double dx, double dy){
		for (Iterator it = pts.iterator(); it.hasNext();) 
			((PicPoint)it.next()).translate(dx,dy);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Scale this object by (sx,sy) using (ptOrgX,ptOrgY) as the origin. This implementation
	 * simply apply a scaling transform to all specification-points.
	 * Note that <code>sx</code> and <code>sy</code> may be negative. 
	 * This method eventually fires a changed-update event.
	 */
	public void scale(double ptOrgX, double ptOrgY, double sx, double sy){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).scale(ptOrgX,ptOrgY,sx,sy);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}	
	
	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).rotate(ptOrg,angle);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
		
	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).shear(ptOrg, shx, shy);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	
	//////////////////////////////////////////////////////////
	/// BOUNDING BOX
	//////////////////////////////////////////////////////////
	
	/**
	 * Returns the bounding box (i.e. the surrounding rectangle) in double precision
	 * Used e.g. to determine the arguments of a \\begin{picture} command.<p>
	 * This implementation compute the bb from the smallest rectangle that encompasses 
	 * all specification-points.
	 * @since PicEdt 1.0
	 */
	public Rectangle2D getBoundingBox(Rectangle2D r){
		if (r==null) r = new Rectangle2D.Double();
		if (pts.isEmpty()) {
			r.setFrameFromDiagonal(0,0,0,0);
			return r;
		}
		r.setFrameFromDiagonal((Point2D)pts.get(0), (Point2D)pts.get(0)); // init bb from 1st spec-point
		for (int i=1; i<pts.size(); i++) r.add((Point2D)pts.get(i)); // then enlarge bb by adding subsequent points
		return r;
	}

	
	///////////////////////////////
	//// Debugging
	///////////////////////////////
	/**
	 * Returns a String representation of this DefaultElement 
	 */
	public String toString(){ 
		String s = super.toString();
		s += "\n\t";
		int i=0;
		for (Iterator it = pts.iterator(); it.hasNext();) s += (i++) + ":"+it.next().toString()+";";
		return s;
	}
	
} 

