/*
 Drawing.java - January 2, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.view.*;
import jpicedt.widgets.*;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.undo.*;

/**
 * An extensible array used to store Elements<p> 
 * This is the MODEL part of the View-Model-Controller paradigm (aka javax.swing.text.Document), aimed
 * at being plugged into an existing PECanvas. It has the structure of a <b>tree-model</b>, with the root element
 * of the tree being Drawing.RootElement : Element's added to this Drawing are actually added to its RootElement ;
 * each Element in turn may or may not have children (see e.g. classes DefaultLeafElement and BranchElement).
 * <br>
 * Element's should be added to this Drawing using the Drawing's API (i.e. shouldn't be directly added to RootElement),
 * by using the <code>addElement()</code> method. 
 * so that DrawingEvent's dispatching works properly.
 * <p>
 * Objects are stored in ascending order according to their z-value, i.e. from back to front.
 * Note that this class does NOT include any selection-handling mechanism : for a variety of reasons, this mechanism has been
 * moved to the associated EditorKit (aka "Controller" in the VMC paradigm), which may itself delegate this behaviour
 * to an appropriate SelectionHandler. This makes the Drawing class a versatile document model for handling a variety of 
 * graphics elements without bothering about the UI interaction. 
 * @see jpicedt.graphic.toolkit.EditorKit
 * @see jpicedt.graphic.PECanvas
 *
 * @since jPicEdt 1.3
 * @author Sylvain Reynal
 * @version $Id: Drawing.java,v 1.14.2.1 2007/09/02 11:56:26 reynal Exp $
 */
public class Drawing implements StateEditable {

	////////////////////////////
	//// PUBLIC CONSTANT FIELDS
	////////////////////////////

	///////////////////////////////
	//// PRIVATE & PROTECTED FIELDS
	///////////////////////////////

	/** the root-element of the drawing */
	protected RootElement root;

	/** list of listener's (e.g. DrawingListener's) that get notified events from this model */
	protected EventListenerList listenerList = new EventListenerList();

	/** a string containing commands read from file but not parsed */
	protected String notParsedCommands="";
	
	/** bounding box used when formatting to text ; null means that it'll be computed automatically */
	protected Rectangle2D boundingBox;
	
	// latest ViewFactory used to build root-view (used for the StateEditable interface implementation)
	private ViewFactory viewFactory;
	
	/////////////////////////
	// CONSTRUCTORS
	/////////////////////////

	/**
	 * Construct a new empty Drawing
	 */
	public Drawing(){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init>");
		root = new RootElement();
	}

	/**
	 * Construct a new Drawing from the given Collection, which is supposed to contain Element's.
	 * Children are cloned beforehands.
	 */
	public Drawing(Collection c){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init> from collection "+c);
		root = new RootElement(c);
	}

	/**
	 * construct a new Drawing whose content is initialized from the content of the given BranchElement.
	 * Children are cloned beforehands, so it's perfectly safe to use this constructor if one doesn't
	 * want to modify the initial content of the given BranchElement.
	 */
	public Drawing(BranchElement e){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("<init> from BranchElement : " + e);
		root = new RootElement(e);
	}

	/**
	 * cloning constructor
	 */
	public Drawing(Drawing dr){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("cloning : "+dr);
		root = new RootElement(dr.root);
		notParsedCommands = new String(dr.notParsedCommands);
	}

	/**
	  * @return a deep copy of this Drawing
	  */
	public Object clone(){
		return new Drawing(this);
	}


	///////////////////////////////////////////////////////////////////////////
	//// CONTENT HANDLING
	///////////////////////////////////////////////////////////////////////////

	/**
	 * @return the root-element for the tree hierarchy that stores element in this drawing
	 */
	public RootElement getRootElement(){
		return root;
	}

	/**
	 * @return the number of graphic objects in this drawing.
	 */
	public int size(){
		return root.getChildCount();
	}

	/**
	 * Remove the given graphic object from this Drawing, overriden so as to 
	 * update selection if this element was selected.
	 */
	public void removeElement(Element element){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeElement : "+element);
		root.removeChild(element); // fire a changed update to the drawing
	}

	/**
	 * Add the given element to this Drawing
	 */
	public void addElement(Element element){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("addElement : "+element);
		root.addChild(element); // fire a changed update to the drawing (INSERT)
	}

	/**
	 * Replace the element at the given index with the given graphic element
	 * No effet if "element" already belongs to the drawing.
	 */
	public void replaceElement(int index, Element element){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("OLD element = "+getElement(index)+"\n\tNEW element = "+element);
		if (root.contains(element)) return;
		root.replaceChild(index, element); // fire a changed update to the drawing (REPLACE)
	}

	/**
	 * Replace the given "src" element with the given "dest" element, if "src" belongs to this drawing.
	 * No effet if "dest" already belongs to the drawing.
	 */
	public void replaceElement(Element src, Element dest){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("OLD element = "+src+"\n\tNEW element = "+dest);
		if (root.contains(dest)) return;
		int index = root.getChildIndex(src);
		root.replaceChild(index, dest); // fire a changed update to the drawing (REPLACE)
	}
	
	/**
	 * @return an iterator over elements
	 */
	public Iterator elements(){
		return root.children();
	}

	/**
	 * @return the element at position i
	 */
	public Element getElement(int i){
		return root.getChildAt(i);
	}

	/**
	 * @return the bounding-box of the drawing ; this is a convenience call to 
	 *         <code>rootElement.getBoundingBox</code> except if setBoundingBox was called 
	 *         with a non-null argument before.
	 */
	public Rectangle2D getBoundingBox(){
		if (this.boundingBox != null) return this.boundingBox; // manual mode
		root.updateBoundingBox(); // because implementation of forwardChangeUpdate in RootElement doesn't call updateBB
		return root.getBoundingBox(null);
	}

	/**
	 * set bounding box manually ; set to null to switch back to automatic behaviour
	 */
	public void setBoundingBox(Rectangle2D bb){
		if (bb==null) this.boundingBox = null; // switch to automatic mode
		else this.boundingBox = new Rectangle2D.Double(bb.getX(),bb.getY(),bb.getWidth(),bb.getHeight());
	}
	
	/**
	 * Return true if the bb is computed automatically on-the-fly
	 */
	public boolean isAutoComputeBoundingBox(){
		return this.boundingBox == null;
	}
	
	/////////////////////////
	//// PAINT
	/////////////////////////

	/**
	 * @return the root-view that renders this Drawing to a graphic context ; this is actually
	 *         the View associated with the root-element.
	 */
	public View getRootView(){
		return root.getView();
	}

	/**
	 * set the view tree that renders this drawing to screen, starting from the root-element.
	 * The root-view produced by the factory (to be associated with the root-element) 
	 * must know how to find its container (usually a PECanvas), otherwise
	 * calls to <code>repaint</code>, etc... will fail.
	 * @param f the ViewFactory that produces View's for the Element's of the tree
	 */
	public void setViewTree(ViewFactory f){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("using factory :"+f);
		viewFactory = f;
		root.setViewFromFactory(f);
	}




	/////////////////////////
	//// LISTENERS
	/////////////////////////

	/**
	 * adds a DrawingListener to the Drawing
	 */
	public void addDrawingListener(DrawingListener l) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("addDrawingListener : "+l);
		listenerList.add(DrawingListener.class, l);
	}

	/**
	 * removes a DrawingListener from the Drawing
	 */
	public void removeDrawingListener(DrawingListener l) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("removeDrawingListener : "+l);
		listenerList.remove(DrawingListener.class, l);
	}

	/**
	 * Notify all listeners that have registered interest for notification on this event type.
	 * @param changed an array containing the elements that changed
	 */
	protected void fireChangedUpdate(Element changed, DrawingEvent.EventType type) {
		Object[] listeners = listenerList.getListenerList();
		DefaultDrawingEvent e = null;
		for (int i = listeners.length-2; i>=0; i-=2) {
			// lazily create the event :
			if (e==null) e = new DefaultDrawingEvent(changed, type);
			if (listeners[i]==DrawingListener.class) {
				((DrawingListener)listeners[i+1]).changedUpdate(e);
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////
	//// StateEditable interface
	////////////////////////////////////////////////////////////////////////////
	/**
	 * Store a clone of RootElement in the given Hashtable, with key = "state"
	 */
	public void storeState(Hashtable map){
		RootElement clone = (RootElement)root.clone();
		//clone.setViewFromFactory(null); // remove views
		map.put("state", clone);
	}
	
	/**
	 * Restore RootElement from the given Hashtable, using key="state"
	 */
	public void restoreState(Hashtable map){
		RootElement restored = (RootElement)map.get("state");
		root = (RootElement)restored.clone(); // if we don't clone, then any editing made to the drawing
		// (i.e. to "root") from now on, will modify, not only "root" itself (which is what we want!)
		// but also the state stored in the hashtable. PB is that the hash MUST remain immutable 
		// if we want undo (or redo) to recall the states that were stored in the exact configuration
		// they had when they were stored !
		if (viewFactory != null) setViewTree(viewFactory); // ensure root is displayed using current viewfactory,
		// not the one it was stored with.
		root.fireChangedUpdate(DrawingEvent.EventType.INSERT);
	}
	
	////////////////////////////////////////////////////////////////////////////
	//// EVENTS HANDLING
	////////////////////////////////////////////////////////////////////////////

	/**
	 * an implementation of DrawingEvent
	 */
	public class DefaultDrawingEvent implements DrawingEvent {

		private DrawingEvent.EventType eventType;
		private Element element;

		/**
		 * @param element the changed Element
		 * @param type the event type
		 */
		public DefaultDrawingEvent(Element element, DrawingEvent.EventType type){

			this.element=element;
			this.eventType=type;
		}

		/**
		 * @return the source of this event
		 */
		public Drawing getDrawing(){
			return Drawing.this;
		}

		/**
		 * @return the changed Element
		 */
		public Element getElement(){
			return element;
		}

		/**
		 * @return the event type
		 */
		public EventType getType(){
			return eventType;
		}

		/**
		 * @return a string describing this event 
		 */
		public String toString(){

			String s = "[DrawingEvent@" + Integer.toHexString(this.hashCode())
			       + ", type=" + getType()
				   + ", changed-element=" + getElement()
				   + ", source=" + getDrawing() + "]";
			return s;
		}
	} // DefaultDrawingEvent


	////////////////////////////////////////////////////////////////////////////
	//// ROOT ELEMENT
	////////////////////////////////////////////////////////////////////////////

	/**
	 * the Element that is the root of the tree of Element's in this Drawing.
	 * Children can be directly added to this RootElement (i.e. instead of using the Drawing API),
	 * since changed-update are always forwarded to the hosting drawing.
	 */
	public class RootElement extends BranchElement {

		/**
		 * Construct a new empty RootElement
		 */
		public RootElement(){
			super();
		}

		/**
		 * construct a new RootElement whose content is initialized from the content of the given BranchElement
		 * Children are cloned beforehands, so it's perfectly safe to use this constructor when one
		 * doesn't want to modify the initial content of the given BranchElement.
		 * @param The Drawing object to clone
		 */
		public RootElement(BranchElement e){
			super(e);
		}

		/**
		 * create a new BranchElement from the content of the given Collection of Element's.
		 * Children are cloned as well (i.e. this isa deep copy).
		 */
		public RootElement(Collection c){
			super(c);
		}
		
		/**
		 * @return the name of this element
		 */
		public String getName(){
			return "drawing.root-element";
		}
		
		/**
		  * @return a deep copy of this Drawing
		  */
		public Object clone(){
			return new RootElement(this);
		}

		/**
		 * Retrieves the underlying drawing
		 * Overriden to return this.
		 */
		public Drawing getDrawing() {
			return Drawing.this;
		}

		/**
		 * Gets the parent of the element. Overriden to return null (root element has no parent)
		 */
		public Element getParent() {
			return null;
		}

		/**
		 * Sets the parent of the element. This implementation does nothing, since this root-element
		 * has no parent.
		 */
		public void setParent(Element p) {
		}

		/**
		 * Called each time the content of this RootElement has changed. This fires the change to the hosting drawing.<br>
		 * Ex : called addChild() on this RootElement -> fires an INSERT event.
		 */
		protected void fireChangedUpdate(DrawingEvent.EventType eventType){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType);
			if (Drawing.this==null) return; // this may be the cased during instanciation
			// of a new Drawing using a cloning constructor ; in this case, Drawing.this
			// is null as long as <init> isn't through.
			Drawing.this.fireChangedUpdate(this, eventType);
			if (view != null) view.changedUpdate(eventType); // updates rootView AFTER
			// listener have had a chance to update their state, so that a call to repaint
			// (triggered by view.changedUpdate) doesn't raise an Exception (this especially useful for
			// selection-handlers which must keep their state synchronized with the Drawing)
		}

		/**
		 * Called by a child to inform of some change that occured to it or one of its children.
		 * This fires a DrawingEvent with the given <code>child</code> as the element that changed.
		 * This mainly occurs when the geometry of an element is modified.
		 */
		public void forwardChangedUpdate(Element child,DrawingEvent.EventType eventType){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType);
			if (Drawing.this==null) return; // see note above
			if (!changeLock) { // if a change is not underway here
				Drawing.this.fireChangedUpdate(child, eventType);
			}
		}
		
		/**
		 * return null 
		 */
		public PicPoint getPoint(int numPoint, PicPoint src){
			return null;
		}

		/**
		 * does nothing
		 * @deprecated use setPoint(int, PicPoint, EditPointConstraint) instead
		 */
		public void setPoint(int numPoint, PicPoint src){
		}
		
		/**
		 * does nothing
		 */
		public void setPoint(int numPoint, PicPoint src, EditPointConstraint c){
		}

		/**
		 * Return 0
		 */
		public int getFirstPointIndex(){
			return 0;
		}

		/**
		 * Return 0
		 */
		public int getLastPointIndex(){
			return 0;
		}

		/**
		 * Update the bounding box by "unioning" the children's bounding box.
		 * This implementation does nothing. 
		 */
		//protected void updateBoundingBox(){}

	} // RootElement



	///////////////////////////////////
	//// STRING FORMATING
	///////////////////////////////////

	/**
	 * Returns a String representing the drawing for debugging use only.
	 */
	public String toString(){

		String s = "[Drawing@" + Integer.toHexString(hashCode());
		s += ", size=" + root.getChildCount();
		s += " {";
		int j=0;
		for(Iterator i = root.iterator(); i.hasNext(); j++){
			StringBuffer buf = new StringBuffer(((Element)i.next()).toString());
			int pos=0;
			while(true){
				pos = buf.toString().indexOf('\n',pos);
				if (pos == -1) break;
				buf.insert(pos+1,'\t');
				pos += 2;
			}
			s += "\n\t" + j + ":" + buf;
		}
		s += "\n}]";
		return s;
	}

	/**
	 * @param a string containing commands read from file or inserted by user, but not parsable.
	 */
	public void setNotparsedCommands(String str){
		notParsedCommands = str;
	}

	/**
	 * @return a string containing (e.g. LaTeX) commands read from file or inserted by user, but not parsable.
	 */
	public String getNotparsedCommands(){
		return notParsedCommands;
	}

	////////////////////////////// CUSTOMIZER /////////////////////////////
	
	/**
	 * Return a bounding box customizer for this Drawing
	 */
	public Customizer getCustomizer(){
		return new Customizer();
	}
	
	/**
	 * a customizer for editing the bounding box
	 */
	public class Customizer extends AbstractCustomizer implements ChangeListener {
		
		private DecimalNumberField xMinTF, yMinTF, xMaxTF, yMaxTF;
		private JCheckBox automaticCB;
		
		/**
		 * Construct a customizer for editing the bounding box
		 */
		public Customizer(){
			initGui();
		}
		
		private void initGui(){
			JPanel p = new JPanel(new GridLayout(5,2,5,5));
			p.setBorder(BorderFactory.createEtchedBorder());
			p.add(new JLabel(" x-min (mm) :"));
			p.add(xMinTF = new DecimalNumberField(0,10)); 
			p.add(new JLabel(" x-max (mm) :"));
			p.add(xMaxTF = new DecimalNumberField(0,10)); 
			p.add(new JLabel(" y-min (mm) :"));
			p.add(yMinTF = new DecimalNumberField(0,10));
			p.add(new JLabel(" y-max (mm) :"));
			p.add(yMaxTF = new DecimalNumberField(0,10));
			p.add(new JLabel(" "+jpicedt.Localizer.currentLocalizer().get("misc.Automatic")));
			p.add(automaticCB = new JCheckBox());
			automaticCB.addChangeListener(this);
			add(p, BorderLayout.NORTH);
		}
		
		/** called by a click on "automaticCB" */
		public void stateChanged(ChangeEvent evt){
			boolean sel = automaticCB.isSelected();
			xMinTF.setEnabled(!sel);
			yMinTF.setEnabled(!sel);
			xMaxTF.setEnabled(!sel);
			yMaxTF.setEnabled(!sel);
		}
		
		/**
		 * @return the panel title, used e.g. for Border or Tabpane title. 
		 */
		public String getTitle(){
			return jpicedt.Localizer.currentLocalizer().get("misc.BoundingBox");
		}


		/**
		 * Load widgets display content with a default value, 
		 * presumably from a "default preferences" file or a dedicated storage class. 
		 */
		public void loadDefault(){
			automaticCB.setSelected(true);
		}

		/**
		 * Load widgets value, presumably from a "preferences" file or a dedicated storage class
		 */
		public void load(){
			automaticCB.setSelected(boundingBox==null);
			Rectangle2D bb = getBoundingBox();
			xMinTF.setValue(bb.getMinX());
			yMinTF.setValue(bb.getMinY());
			xMaxTF.setValue(bb.getMaxX());
			yMaxTF.setValue(bb.getMaxY());
		}

		/**
		 * Store current widgets value, presumably to a file or to a dedicated storage class
		 */
		public void store(){
			if (automaticCB.isSelected()) setBoundingBox(null);
			else {
				Rectangle2D.Double bb = new Rectangle2D.Double();
				bb.x=xMinTF.getValue();
				bb.y=yMinTF.getValue();
				bb.width=xMaxTF.getValue()-bb.x;
				bb.height=yMaxTF.getValue()-bb.y;
				setBoundingBox(bb);
			}
		}
	}
		
	
} // Drawing
