/*  PicCirlceFrom3Points.java - March 2003 - jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.graphic.model;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.*;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.view.View;
import jpicedt.graphic.view.HitInfo;
import jpicedt.widgets.*;

/**
 * A class implementing either a circle or an arc, whose geometry is specified by a PicEllipse,
 * yet is further controllable by three additional points :
 * this may either be a plain circle going through these three points,
 * or an arc going from P_1 to P_3 through P_2, in which case a pie or a chord may be added as well.<br>
 * Superclass control-points are still available here, yet geometry is constrained to a circle by 
 * invoking e.g. {@link #setPoint super.setPoint} using a 
 * {@link #SQUARE PicParallelogram.SQUARE} EditPointConstraint.
 * @author Vincent Guirardel
 * @since jPicEdt 1.4
 * @version $Id: PicCircleFrom3Points.java,v 1.19.2.1 2007/09/02 11:56:26 reynal Exp $
 */
public class PicCircleFrom3Points extends PicEllipse implements CustomizerFactory {

	////////////////////////////
	//// PUBLIC CONSTANT FIELDS
	////////////////////////////

	// inherited : P_BL, P_BR, P_TR (specification and control points), 
	//             P_TL, P_ANGLE_START, P_ANGLE_END (control points only)
	// FIRST_PT left unchanged
	
	/** index of first end-point */
	public static final int P_1 = PicEllipse.LAST_PT+1 ; // these are our three additional control-points
	/** index of intermediate point */
	public static final int P_2 = P_1+1;
	/** index of last end-point */
	public static final int P_3 = P_1+2;

	public static final int LAST_PT = P_3;

	//////////////////////////////
	/// PRIVATE/PROTECTED VARIABLES
	//////////////////////////////
	/**
	 * A list holding the three control-points this circle goes through, and serving
	 * as a convenient caching mechanism.
	 * These are controlled by the {@link #setPoint setPoint} method, and
	 * are used to update superclass geometry when {@link #updateEllipse updateEllipse} is invoked.
	 * Besides, protected getter/setter {@link #getControlPoint} and {@link #setControlPoint} allow to
	 * edit them easily without cumbersome class-cast.
	 */
	protected ArrayList controlPts = new ArrayList();
	
    // the following indices are the indices of P_1, P_2, P_3 in the controlPts array.
	/** index of first control-point, aka P_1, in {@link #controlPts controlPts} list */
    protected static final int CTRL_P_1 = 0 ;
	/** index of second control-point, aka P_2, in {@link #controlPts controlPts} list */
    protected static final int CTRL_P_2 = 1 ;
	/** index of last control-point, aka P_3, in {@link #controlPts controlPts} list */
    protected static final int CTRL_P_3 = 2 ;
	
    private static final int INDEX_SHIFT = P_1-CTRL_P_1 ;// equals the difference between CTRL_P_i and P_i. Used only in setPoint and getPoint.

    /** 
     * true if we want to draw a full circle, false if we just want an arc.
     */
    private boolean isFullCircle; 

	private PicVector ptBuffer1 = new PicVector(); // used by setPoint()
	
	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Creates a new circle reduced to a single point located at (0,0). 
	 */
	public PicCircleFrom3Points(){
		this(true,Arc2D.OPEN);
	}

	/**
	 * Creates a new arc or circle reduced to a single point located at (0,0), with a default attribute set 
	 * and the given closure type. 
	 * @param plain if true, we draw a plain circle and ignore the closure parameter.
	 * @param closure closure type as defined in java.awt.geom.Arc2D
	 */
	public PicCircleFrom3Points(boolean plain,int closure){
		super(plain?Arc2D.OPEN:closure);
		initDefaultPoints();
		isFullCircle=plain;
	}

	/**
	 * Creates a new (full) circle reduced to a single point located at (0,0), with the given attribute set.
	 */
	public PicCircleFrom3Points(PicAttributeSet set){
	    this(true,Arc2D.OPEN,set);
	    initDefaultPoints();
	}

	/**
	 * Creates a new arc or circle reduced to a single point located at (0,0), 
	 * with the given closure and attribute set.
	 * @param plain if true, we draw a plain circle and ignore the closure parameter.
	 * @param closure closure type as defined in java.awt.geom.Arc2D
	 */
	public PicCircleFrom3Points(boolean plain, int closure, PicAttributeSet set){
		super(plain?Arc2D.OPEN:closure,set);
		initDefaultPoints();
		isFullCircle=plain;
	}

	/**
	 * Create a new circle whose geometry is specificed by the the 3 given points, and
	 * with the given attribute set.
	 */
	public PicCircleFrom3Points(PicPoint pt1, PicPoint pt2, PicPoint pt3, PicAttributeSet set){
	    this(pt1,pt2,pt3,true,Arc2D.OPEN,set);
	}

	/**
	 * Create a new arc or circle whose geometry is specified by the 3 given points, and 
	 * with the given attribute set.
	 * @param plain whether to draw a plain circle or an arc
	 * @param closure closure type as defined in java.awt.geom.Arc2D
	 */
	public PicCircleFrom3Points(PicPoint pt1, PicPoint pt2, PicPoint pt3, boolean plain, int closure, PicAttributeSet set){
		super(plain?Arc2D.OPEN:closure,set);
		controlPts.add(new PicPoint(pt1));
		controlPts.add(new PicPoint(pt2)); 
		controlPts.add(new PicPoint(pt3));
		isFullCircle=plain;
		updateEllipse();
	}

	/**
	 * "cloning" constructor (to be used by clone())
	 */
	public PicCircleFrom3Points(PicCircleFrom3Points circle){
	    super(circle);
		controlPts.add(new PicPoint(circle.getControlPoint(CTRL_P_1)));
		controlPts.add(new PicPoint(circle.getControlPoint(CTRL_P_2)));
		controlPts.add(new PicPoint(circle.getControlPoint(CTRL_P_3)));
		isFullCircle = circle.isFullCircle;
	}

	/**
	 * Overrides Object.clone() method
	 */
	public Object clone(){
		return new PicCircleFrom3Points(this);
	}


	/**
	 * Called by contructors to init fields to their default value
	 */
    private void initDefaultPoints(){
	    controlPts.add(new PicPoint(0,0));
	    controlPts.add(new PicPoint(0,0));
	    controlPts.add(new PicPoint(0,0));
	}


	/**
	 * Returns the index of the first user-controlled point that can be retrieved by <code>getPoint()</code>.
	 */
	public int getFirstPointIndex(){
		return FIRST_PT;
	}
	
	/**
	 * Returns the index of the last user-controlled point that can be retrieved by <code>getPoint()</code>.
	 */
	public int getLastPointIndex(){
		return LAST_PT;
	}


	/**
	 * @return a string that represents this object's name ; should be a key-entry to i18n files.
	 * @since jpicedt 1.3.3
	 */
	public String getName(){
		return jpicedt.Localizer.currentLocalizer().get("model.Circle");
	}

	/////////////////////////////////////
    // OTHER FIELDS ACCESSORS
    ////////////////////////////////////////

	/**
	 * Convert this circle to a new PicEllipse object.
	 */
	public PicEllipse convertToEllipse(){
		return new PicEllipse(this);
	}


	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////
        
	/**
	 * Returns a reference to the ith control-point in the
	 * {@link #controlPts controlPts} list.
	 * @param i one of CTRL_P_1, CTRL_P_2 or CTRL_P_3.
	 */
	protected PicPoint getControlPoint(int i){
		return (PicPoint)controlPts.get(i);
	}
	
	/**
	 * Set the coordinates of ith control-point from the given point.
	 * @param i one of P_1, P_2 or P_3.
	 */
	protected void setControlPoint(int i, PicPoint pt){
		((PicPoint)controlPts.get(i)).setCoordinates(pt);
	}

	/**
 	 * This method update the geometry of the superclass, i.e. specification points P_BL, P_BR and P_TL,
	 * from the current value of the three control points P_1, P_2, P_3.<br>
	 * Subclasser may want to adapt updating scheme to their own need here.
	 */
    protected void updateEllipse(){

        // we compute the following vectors:
	    // u and v are the vectors joining P_1 to P_2 and P_3 respectively
        PicVector u=new PicVector(getControlPoint(CTRL_P_1),getControlPoint(CTRL_P_2));
        PicVector v=new PicVector(getControlPoint(CTRL_P_1),getControlPoint(CTRL_P_3));
		// m and n are the midpoints of the segments P_1P_2 and P_1P_3
		PicPoint  m=new PicPoint(getControlPoint(CTRL_P_1));
		m.middle(getControlPoint(CTRL_P_2));
		PicPoint  n=new PicPoint(getControlPoint(CTRL_P_1));
		n.middle(getControlPoint(CTRL_P_3));
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("points m,n,u,v"+m+n+u+v);
		double det=u.det(v);
		if (det==0){//flat case
		    PicPoint extremity1;
		    PicPoint extremity2;
		    if (u.dot(v) <=0) {//P_1 is between P_2 and P_3
			extremity1=getControlPoint(CTRL_P_2);
			extremity2=getControlPoint(CTRL_P_3);
		    }
		    else { 
			PicVector w=new PicVector(getControlPoint(CTRL_P_2),getControlPoint(CTRL_P_3));
			if (u.dot(w)>=0){//if P_2 is between P_1 and P_3
			    extremity1=getControlPoint(CTRL_P_1);
			    extremity2=getControlPoint(CTRL_P_3);
			}
			else {
			    extremity1=getControlPoint(CTRL_P_1);
			    extremity2=getControlPoint(CTRL_P_2);
			}
		    }
		    setSpecificationPoint(P_BL,extremity1);
		    setSpecificationPoint(P_BR,extremity2);
		    setSpecificationPoint(P_TR,extremity2);
		    setAngleStart(-180); 
		    setAngleEnd(180); 
		}
		else {//non-flat case
			//First compute center (just a big formula that should do the job)
		    PicPoint c= new PicPoint((-v.x*n.x*u.y+v.y*u.x*m.x+v.y*u.y*m.y-v.y*n.y*u.y)/det,
					     (u.x*v.x*n.x+u.x*v.y*n.y-u.x*m.x*v.x-u.y*m.y*v.x)/det);
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Centre: "+c.x+","+c.y);
			// Now the radius
			double R=(new PicVector(c,getControlPoint(CTRL_P_1))).norm();
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("radius"+R);
			PicPoint p= new PicPoint(c.x-R,c.y-R);
			// Set the controling paralleogram (which is a square !)
			setSpecificationPoint(PicEllipse.P_BL,p);
			p.x=c.x+R;
			p.y=c.y-R;
			setSpecificationPoint(PicEllipse.P_BR,p);
			p.x=c.x+R;
			p.y=c.y+R;
			setSpecificationPoint(PicEllipse.P_TR,p);
			if (isFullCircle){
				setAngleStart(-180); 
				setAngleEnd(180); 
			}
			else {
				// find the correct order between angles...
			    PicVector v1=new PicVector(c,getControlPoint(CTRL_P_1));
			    PicVector v2=new PicVector(c,getControlPoint(CTRL_P_2));
			    PicVector v3=new PicVector(c,getControlPoint(CTRL_P_3));
			    
			    double a1=PicVector.X_AXIS.angleDegrees(v1);
			    double a2=PicVector.X_AXIS.angleDegrees(v2);
			    double a3=PicVector.X_AXIS.angleDegrees(v3);

				if (a1<=a2 && a2<=a3) {//do nothing 
				}
				else if (a3<=a2 && a2<=a1) {
					double b=a1;
					a1=a3;
					a3=b;
				}
				else if (a1<=a3 && a3<=a2 ){
				    double b=a1;
					a1=a3-360;
					a2=a2-360;
					a3=b;
				}
				else if (a3<=a1 && a1<=a2){
					a1=a1-360;
					a2=a2-360;
				}
				else if (a2<=a3 && a3<=a1){
					a1=a1-360;
				}
				else if (a2<=a1 && a1<=a3){
					double b=a1;
					a1=a3-360;
					a3=b;
				}
				setAngleStart(a1); 
				setAngleEnd(a3); 
			}
		}
		updateAxis();
    }
    
	/**
	 * Update the values of the three control-points from the geometry of the superclass.
	 * This aka updateEllipse(), yet the other way around.
	 * @param numPoint index of the PicParallelogram's control-point that was moved
	 */
	protected void updateControlPoints(int numPoint){
		
		// [SR:todo] => rotate circle's control-points along opposite corner of pts[numPoint].
		jpicedt.Log.warning("Not implemented yet !");
	}


	/**
	 * Set the coordinate of the point indexed by "numPoint" to the given value.
	 */
	public void setPoint(int numPoint, PicPoint pt, EditPointConstraint c){
		if (numPoint == P_CENTER) {
			getPoint(P_CENTER,ptBuffer1); // save old CENTER in ptBuffer1
			getControlPoint(CTRL_P_1).translate(ptBuffer1,pt);
			getControlPoint(CTRL_P_2).translate(ptBuffer1,pt);
			getControlPoint(CTRL_P_3).translate(ptBuffer1,pt);
			super.setPoint(P_CENTER, pt, null);
		}
	    else if (numPoint < P_1 ){
			// [SR:underway]
			super.setPoint(numPoint,pt,SQUARE); // parallelogram/ellipse control-points, including check for index-out-of-bounds
			// Reminder : SQUARE constraint ensure parallelogram is a square (hence ellipse is a circle), 
			// then "rotate" parallelogram's specification-points along the opposite corner of pts[numPoint]
			updateControlPoints(numPoint); // update control-points P_1,... so that they reflect changes made in the geometry of the superclass
		}
	    else {
			setControlPoint(numPoint-INDEX_SHIFT,pt); // e.g. setPoint(P_1) => changes CTRL_P_1
			updateEllipse();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	    }
	}

	/**
	 * Get the coordinate of the point indexed by "numPoint" to the given value.
	 * @param numPoint one of P_1, P_2 or P_3.
	 * @param src a PicPoint to update and create if necessary
	 */
	public PicPoint getPoint(int numPoint,PicPoint src){
	    if (numPoint < P_1 ){
			return super.getPoint(numPoint,src);// parallelogram/ellipse control-points, including check for index-out-of-bounds
		}
	    else {
			if (src==null) src = new PicPoint();
			src.setCoordinates(getControlPoint(numPoint-INDEX_SHIFT));
			return src;
	    }
	}

	/**
	 * Translate this circle by (dx,dy)
	 */
	public void translate(double dx, double dy){
		getControlPoint(CTRL_P_1).translate(dx,dy);
		getControlPoint(CTRL_P_2).translate(dx,dy);
		getControlPoint(CTRL_P_3).translate(dx,dy);
		super.translate(dx,dy);
	}

	/**
	 * Scale this object by (sx,sy) using (ptOrgX,ptOrgY) as the origin. This implementation
	 * simply apply a scaling transform to all specification-points.
	 * Note that <code>sx</code> and <code>sy</code> may be negative. 
	 * This method eventually fires a changed-update event.
	 */
	public void scale(double ptOrgX, double ptOrgY, double sx, double sy){
		getControlPoint(CTRL_P_1).scale(ptOrgX,ptOrgY,sx,sy);
		getControlPoint(CTRL_P_2).scale(ptOrgX,ptOrgY,sx,sy);
		getControlPoint(CTRL_P_3).scale(ptOrgX,ptOrgY,sx,sy);
		super.scale(ptOrgX, ptOrgY, sx, sy);
	}	
	
	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		getControlPoint(CTRL_P_1).rotate(ptOrg,angle);
		getControlPoint(CTRL_P_2).rotate(ptOrg,angle);
		getControlPoint(CTRL_P_3).rotate(ptOrg,angle);
		super.rotate(ptOrg, angle);
	}

	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		getControlPoint(CTRL_P_1).shear(ptOrg,shx,shy);
		getControlPoint(CTRL_P_2).shear(ptOrg,shx,shy);
		getControlPoint(CTRL_P_3).shear(ptOrg,shx,shy);
		super.shear(ptOrg,shx,shy);
	}

	// doc inherited
	public boolean isPlain(){
		return isFullCircle;
	}
	
	// doc inherited
	public void setPlain(boolean b){
		this.isFullCircle=b;
		updateEllipse();
	}
	
	/**
	 * Implementation of the Object.toString() method, used for debugging purpose
	 * @since PicEdt 1.1.4
	 */
	public String toString() {

		String s = super.toString()
		           + "\n\t controlPts="
		           + getControlPoint(CTRL_P_1)
				   + " "
				   + getControlPoint(CTRL_P_2)
				   + " "
				   + getControlPoint(CTRL_P_3);

		return s;
	}

	/////////////////////////////////
	//// Action's
	/////////////////////////////////
	
	/**
	 * Create an array of Action's related to this object
	 *
	 * @param actionDispatcher  dispatches events to the proper PECanvas
	 * @param localizer         i18n localizer for PEAction's
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {
		PEAction[] actionArray = new PEAction[1];
		actionArray[0] = new ConvertToEllipseAction(actionDispatcher, localizer);
		return actionArray;
	}

	// ---- parallelo to multiCurve ----
	
	/**
	 * Convert this circle to an ellipse, selecting it if applicable.
	 * @author    Sylvain Reynal
	 * @since     jpicedt 1.4
	 */
	class ConvertToEllipseAction extends PEAction {

		public static final String KEY = "action.editorkit.ConvertCircleToEllipse";
		
		public ConvertToEllipseAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void undoableActionPerformed(ActionEvent e){
			PicEllipse curve = convertToEllipse();
			Drawing dr = getDrawing();
			if (dr != null){
				int index = dr.getRootElement().getChildIndex(PicCircleFrom3Points.this);
				dr.replaceElement(index, curve);
				View view = curve.getView();
				if (view != null){
					PECanvas canvas = view.getContainer();
					if (canvas != null)   canvas.select(curve, true);
				}
			}
		}
	}
	
	////////////////////////////////
	//// GUI
	////////////////////////////////


	/**
	 * Return a Customizer for geometry editing
	 * @since jpicedt 1.3.3
	 */
	public AbstractCustomizer createCustomizer(){
		return new Customizer();
	}
	
	/**
	 * Geometry customizer
	 */
	class Customizer extends AbstractCustomizer implements ActionListener {

	    private DecimalNumberField pt1XTF,pt1YTF,pt2XTF,pt2YTF,pt3XTF,pt3YTF;
		private JCheckBox fullCircleCB;
		private JComboBox arcTypeList;
		private boolean isListenersAdded = false; // flag

		private ImageIcon[] arcTypeIcons = {
			PEToolKit.createImageIcon("action.draw.ArcChord"), 
		    PEToolKit.createImageIcon("action.draw.ArcOpen"),
			PEToolKit.createImageIcon("action.draw.ArcPie")};
		
		public Customizer(){

			super();
			JPanel p = new JPanel(new GridLayout(6,3,5,5));

			p.add(PEToolKit.createJLabel("action.draw.CircleFrom3Pts"));
			p.add(new JLabel("x"));
			p.add(new JLabel("y"));

			p.add(new JLabel("1"));
			// x-coordinate of pt1
			p.add(pt1XTF = new DecimalNumberField(4));
			// y-coordinate of pt1
			p.add(pt1YTF = new DecimalNumberField(4));

			p.add(new JLabel("2"));
			// x-coordinate of pt2
			p.add(pt2XTF = new DecimalNumberField(4));
			// y-coordinate of pt2
			p.add(pt2YTF = new DecimalNumberField(4));

			p.add(new JLabel("3"));
			// x-coordinate of pt3
			p.add(pt3XTF = new DecimalNumberField(4));
			// y-coordinate of pt3
			p.add(pt3YTF = new DecimalNumberField(4));

			p.add(new JLabel("Plain :"));
			p.add(fullCircleCB=new JCheckBox());
			p.add(new JLabel(" "));
			
			// arc type (chord, pie or open)
			arcTypeList = PEToolKit.createComboBox(arcTypeIcons);
			p.add(new JLabel("Arc :"));
			p.add(arcTypeList);
			p.add(new JLabel(" "));

			add(p, BorderLayout.NORTH);
		}
		
		/** add action listeners to widgets to reflect changes immediately */ 
		private void addActionListeners(){
			if (isListenersAdded) return; // already done
			pt1XTF.addActionListener(this);
			pt1YTF.addActionListener(this);
			pt2XTF.addActionListener(this);
			pt2YTF.addActionListener(this);
			pt3XTF.addActionListener(this);
			pt3YTF.addActionListener(this);
			fullCircleCB.addActionListener(this);
			arcTypeList.addActionListener(this);
			isListenersAdded = true;
		}

		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){
			pt1XTF.setValue(getControlPoint(CTRL_P_1).x);
			pt1YTF.setValue(getControlPoint(CTRL_P_1).y);
			pt2XTF.setValue(getControlPoint(CTRL_P_2).x);
			pt2YTF.setValue(getControlPoint(CTRL_P_2).y);
			pt3XTF.setValue(getControlPoint(CTRL_P_3).x);
			pt3YTF.setValue(getControlPoint(CTRL_P_3).y);
			fullCircleCB.setSelected(isPlain());
			if (getArcType()==CHORD) arcTypeList.setSelectedIndex(0);
			else if (getArcType()==OPEN) arcTypeList.setSelectedIndex(1);
			else if (getArcType()==PIE) arcTypeList.setSelectedIndex(2);

			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded 
			addActionListeners(); // done the first time load is called
		}

		/**
		 * update Element's properties
		 */
		public void store(){
		    PicPoint p = new PicPoint();
		    p.x = pt1XTF.getValue();
		    p.y = pt1YTF.getValue();
		    setControlPoint(CTRL_P_1,p);
		    p.x = pt2XTF.getValue();
		    p.y = pt2YTF.getValue();
		    setControlPoint(CTRL_P_2,p);
		    p.x = pt3XTF.getValue();
		    p.y = pt3YTF.getValue();
		    setControlPoint(CTRL_P_3,p);

			setPlain(fullCircleCB.isSelected());
			switch (arcTypeList.getSelectedIndex()){
				case 0 : setArcType(Arc2D.CHORD); break;
				case 1 : setArcType(Arc2D.OPEN); break;
				case 2 : setArcType(Arc2D.PIE); break;
				default:
			}

		    updateEllipse();
		    fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}

		public void actionPerformed(ActionEvent e){
			store();
		}

		/**
		 * @return the panel title, used e.g. for Border or Tabpane title. 
		 */
		public String getTitle(){ 
			return PicCircleFrom3Points.this.getName();
		}
		
	}

} // PicCircleFrom3Points
