/*
 PicEllipse.java - August 29, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.event.*;
import jpicedt.widgets.*;
import java.awt.geom.PathIterator;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;


/**
 * Ellipse or arc, based on parallelogram 
 * <br>
 * The geometrical model of this ellipse (or its arc counterpart) is based on a parallelogram, yet
 * is equivalent to an ellipse rotated around its center.
 * <p>
 * In the following code documentation, we will be using two coordinate-systems, namely : 
 * <ul>
 * <li> a so-called <b>parallelogram basis</b>, whose vectors are defined by the sides of the parallelogram.
 *      These correspond to fields {@link #l2rVec l2rVec} and {@link #b2tVec b2tVec} inherited from superclass.
 * <li> an <b>ellipse basis</b> (also called canonical basis), whose vectors are defined by the great and small axes
 *      of the ellipse respectively.
 * </ul>
 * Arc angles may be expressed in either system, and this is indicated in code documentation when applicable.
 * Prefix <code>skew</code> refers to the parallelogram basis, whereas <code>rotated</code> denotes
 * angles measured in the ellipse basis.
 * Since ellipse axes are generally not parallel to parallelogram sides, except in the event this parallelogram is
 * a rectangle, both sets of angle DO obviously differ.
 * <p>
 * In the parallelogram basis, arc is modelled by the following parametrized curve :<br>
 * <b>Center</b> + cos(t)*<b>U</b>+sin(t)*<b>V</b>,<br>
 * for t ranging from <code>skewAngleStart</code> to <code>skewAngleEnd</code>.
 * <p>
 * Rotated angles are computed so that the arc of ellipse is obtained simply by rotating by "rotationAngle" (CCW)
 * the arc of ellipse whose parametrization is Center+greatAxis*cos(t)*I+smallAxis*sin(t)*J
 * where I,J is the ellipse canonical basis, and t varies between "rotatedAngleStart" and "rotatedAngleEnd".
 * Note that this means that when small axis is negative, this parametrization goes CW and not CCW.
 * <p>
 * [TODO] changes ALL internal angles to radian, and move conversion to setters/getters is user may want to enter degrees.
 * [TODO] add a picture (in GIF or PNG format) to this documentation (as in AbstractCurve) to help end-users
 *        understand all these tricky things !
 * @author Vincent Guirardel, Sylvain Reynal.
 * @since jPicEdt 1.4
 * @version $Id: PicEllipse.java,v 1.26.2.1 2007/09/02 11:56:26 reynal Exp $
 */
public class PicEllipse extends PicParallelogram  {


	/** point marking start-of-arc  */
	public static final int P_ANGLE_START = 9;

	/** point marking end-of-arc  */
	public static final int P_ANGLE_END = 10;
	
	public static final int LAST_PT = 10;
	
	/** prefined closure type for arcs */
	public static final int CHORD = Arc2D.CHORD;
	/** prefined closure type for arcs */
	public static final int OPEN = Arc2D.OPEN;
	/** prefined closure type for arcs */
	public static final int PIE = Arc2D.PIE;

	//////////////////////////
	//// PRIVATE/PROTECTED FIELDS
	//////////////////////////

	/** maximal angle (in radians) between axes of the parallelogram
	 * below which we consider the ellipse to be flat
	 * also used to avoid arcs whose angle extent is almost 0
	 */
	private final static double MAX_FLAT_ANGLE=0.0001;
	
	// buffers used in some computation-intensive methods to minimize object creation, while still maintaining
	// code readibility by use of the PicVector/PicPoint geom. utilities :
	private PicVector ptBuffer1 = new PicVector();
	private PicVector ptBuffer2 = new PicVector();

	/** 
	 * Arc start- and end- angles in degrees as measured with respect to parallolegram's basis 
	 * (see {@link #toParalleloBasisCoordinates toParalleloBasisCoordinates()} for
	 * details on how this basis is defined).
	 * <p>
	 * To wind up shortly, let <b>U</b> and <b>V</b> denote unit-vectors defined by the sides of the parallelogram
	 * (these are actually given by "l2rVec" and "b2tVec" protected fields, yet immediately after 
	 * {@link #computeParalleloBasis computeParalleloBasis} has been called),
	 * then the starting point of the arc is located at : 
	 * <b>Center</b>+cos(skewAngleStart)*<b>U</b>+sin(skewAngleStart)*<b>V</b>.
	 * <p>
	 * This value is closer to the user than rotatedAngleStart. [SR:je ne comprend pas cette phrase !!!]
	 * <p>
	 * Moreover, arc is modelled by the following parametrized curve :<br>
	 * <b>Center</b> + cos(t)*<b>U</b>+sin(t)*<b>V</b>,<br>
	 * for t ranging from skewAngleStart to skewAngleEnd.
	 * <p>
	 * We set skewAngleStart in (-180,180], and skewAngleEnd in (skewAngleStart,skewAngleStart+360]. 
	 * This means that if one makes a mirror symmetry of the parallelogram and keeps the values of these "skew" angles unchanged,
	 * then we get the mirror image of the arc.
	 * <br>
	 * We use degrees because user may enter a value.
	 * [SR:pending] On peut faire la conversion dans les setters correspondants (setAngleXXX).
	 * L'ideal est qd meme d'avoir la meme unite partout (en interne) pour ne pas s'emmeler les pedales.
	 * @see #l2rVec
	 * @see #b2tVec
	 */
	protected double skewAngleStart;
	
	/** 
	 * @see #skewAngleStart 
	 */
	protected double  skewAngleEnd;

	/**
	 * Angle in radians (CCW) between the great axis of the ellipse and the horizontal axis.
	 */
	protected double rotationAngle;

	/**
	 * Length of the small axis of the ellipse.
	 * May be negative, as 
	 * this gives a more coherent behaviour under negative scaling in one direction.
	 * The sign is determined by the orientation of the parallelogram.
	 */
	protected double smallAxis;
	/**
	 * Lengths of the great axis of the ellipse.
	 */
	protected double greatAxis;

	/**
	 * Arc angles in degrees, as measured in the <b>ellipse basis</b>. These are computed from their "skew" counterparts,
	 * through a call to {@link #updateRotatedAngles updateRotatedAngles}, and are obviously identical with them
	 * if surrounding parallelogram is a rectangle.  
	 * <p>
	 * These angles may be used to build an appropriate {@link java.awt.Shape Shape} for this Element, 
	 * by first creating an {@link java.awt.geom.Arc2D Arc2D} from these angles, then rotating this shape
	 * by {@link #rotationAngle rotationAngle}.
	 * <p>
	 * We restrict values of rotatedAngleStart to (-180,180], and those of 
	 * rotatedAngleEnd to (rotatedAngleStart,rotatedAngleStart+360].
	 */
	protected double rotatedAngleStart,rotatedAngleEnd;

	/** closure type */
	protected int closure;


	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Create a new PicEllipse, centered at (0,0), with a null radius.
	 * CLOSED is the default closure
	 */
	public PicEllipse(){
		super();
		initDefault();
	}

	/**
	 * Create a new PicEllipse, centered at (0,0), with a null radius, and the given attribute set.
	 */
	public PicEllipse(PicAttributeSet set){
		super(set);
		initDefault();
	}

	/**
	 * Create a new PicEllipse, centered at (0,0), with a null radius, the given closure type.
	 * @param closure one of CHORD, PIE or OPEN.
	 */
	public PicEllipse(int closure){
		super();
		initDefault();
		this.closure = closure;
	}

	/**
	 * Create a new PicEllipse, centered at (0,0), with a null radius, 
	 * the given closure type, and the given attribute set.
	 * @param closure one of CHORD, PIE or OPEN.	 
	 */
	public PicEllipse(int closure, PicAttributeSet set){
		super(set);
		initDefault();
		this.closure = closure;
	}

	/**
	 * Create a new PicEllipse object using the 3 given points as 3 consecutive points of the 
	 * surrounding parallelogram, and the given attribute set.
	 * @param closure one of CHORD, PIE or OPEN.	 
	 */
	public PicEllipse(PicPoint p1, PicPoint p2, PicPoint p3, int closure, PicAttributeSet set){
		super(p1,p2,p3,set);
		initDefault();
		this.closure = closure; // override default value
		updateAxis();
	}
	
	/**
	 * Create a new PicEllipse object using the 3 given points as 3 consecutive points of the 
	 * surrounding parallelogram, and a default attribute set.
	 * @param closure one of CHORD, PIE or OPEN.	 
	 */
	public PicEllipse(PicPoint p1, PicPoint p2, PicPoint p3, int closure){
		super(p1,p2,p3);
		initDefault();
		this.closure = closure; // override default value
		updateAxis();
	}

	/**
	 * "cloning" constructor (to be used by clone())
	 * @param The PicEllipse object to clone
	 */
	public PicEllipse(PicEllipse ellipse){
		super(ellipse); // parallelogram cloner
		greatAxis = ellipse.greatAxis;
		smallAxis = ellipse.smallAxis;
		rotationAngle = ellipse.rotationAngle;
		closure = ellipse.closure;
		rotatedAngleStart=ellipse.rotatedAngleStart;
		rotatedAngleEnd=ellipse.rotatedAngleEnd;
		skewAngleStart=ellipse.skewAngleStart;
		skewAngleEnd=ellipse.skewAngleEnd;
	}

	/**
	 * Called by contructors to init fields to their default value
	 */
	private void initDefault(){
		greatAxis=0;
		smallAxis=0;
		rotationAngle=0;
		closure = Arc2D.OPEN;
		rotatedAngleStart=0;
		rotatedAngleEnd=360;
		skewAngleStart=0;
		skewAngleEnd=360;
	}

	/**
	 * Overload Object.clone() method
	 */
	public Object clone(){
		return new PicEllipse(this);
	}

	/**
	 * Return a localised string that represents this object's name
	 */
	public String getName(){
		return jpicedt.Localizer.currentLocalizer().get("model.Ellipse");
	}

	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////

	/**
	 * Set the coordinate of the point indexed by "numPoint" to the given value.
	 * If point controls the parallelogram, calls the paralleogram setPoint method, and updates axis.
	 */
	public void setPoint(int numPoint, PicPoint pt, EditPointConstraint constraint){
		if (numPoint== P_ANGLE_END){
			if (isFlat()) return;
			updateParalleloBasis(); // [SR:pending] terrible hack to ensure coordinates (l2r,b2t) are reliable
			// [delete] skewAngleEnd=Math.toDegrees(computeAngleWithHorizontal(
			// 				l2rCoordinate(getPointX(P_CENTER),getPointY(P_CENTER),pt.x,pt.y),
			//				b2tCoordinate(getPointX(P_CENTER),getPointY(P_CENTER),pt.x,pt.y))) ;
			toParalleloBasisCoordinates(getPoint(P_CENTER,ptBuffer1), pt, ptBuffer2); // org/src/dest
			setAngleEnd(Math.toDegrees(PicVector.X_AXIS.angle(ptBuffer2)));
		}
		else if (numPoint== P_ANGLE_START){
			if (isFlat()) return;
			updateParalleloBasis();
			// [delete] skewAngleStart=Math.toDegrees(computeAngleWithHorizontal(
				// l2rCoordinate(getPointX(P_CENTER),getPointY(P_CENTER),pt.x,pt.y),
				// b2tCoordinate(getPointX(P_CENTER),getPointY(P_CENTER),pt.x,pt.y))) ;
			toParalleloBasisCoordinates(getPoint(P_CENTER,ptBuffer1), pt, ptBuffer2); // org/src/dest
			setAngleStart(Math.toDegrees(PicVector.X_AXIS.angle(ptBuffer2)));
		}
		else {
			super.setPoint(numPoint,pt,constraint); 
			updateAxis();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE); 
		}
	}

	/**
	 * Return the <b>user-controlled point</b> having the given index. The general contract in Element
	 * is to return an IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method.<br>
	 * @return the point indexed by <code>numPoint</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 */
	public PicPoint getPoint(int numPoint, PicPoint src) {
		if (numPoint==P_ANGLE_START){
			return getPoint(P_CENTER,src)
				.translate(getSpecificationPoint(P_BL),getSpecificationPoint(P_BR), 0.5*Math.cos(Math.toRadians(skewAngleStart)))
				.translate(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR), 0.5*Math.sin(Math.toRadians(skewAngleStart)));
				// [SR:pending] can't use l2rVec/b2tVec here, seeing they barely have the expected length... hence P_ANGLE_START can't
				// shows up in the right place. We really need to do some clean-up in superclass's setPoint() method.
		}
		else if (numPoint==P_ANGLE_END){
			return getPoint(P_CENTER,src)
				.translate(getSpecificationPoint(P_BL),getSpecificationPoint(P_BR), 0.5*Math.cos(Math.toRadians(skewAngleEnd)))
				.translate(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR), 0.5*Math.sin(Math.toRadians(skewAngleEnd)));
		}
		else return super.getPoint(numPoint,src);
	}

	/**
	 * Scale this object by (sx,sy) using (ptOrgX,ptOrgY) as the origin. This implementation
	 * simply apply a scaling transform to all specification-points.
	 * Note that <code>sx</code> and <code>sy</code> may be negative. 
	 * This method eventually fires a changed-update event.
	 */
	public void scale(double ptOrgX, double ptOrgY, double sx, double sy){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).scale(ptOrgX,ptOrgY,sx,sy);
		updateAxis();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}	
	
	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).rotate(ptOrg,angle);
		updateParalleloBasis();
		updateAxis();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).shear(ptOrg,shx,shy);
		updateParalleloBasis();
		updateAxis();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Return a Bezier curve created from this ellipse
	 *
	 * @since jpicedt 1.4
	 */
	public PicMultiCurve convertToMultiCurve(){
		
		// 1) create Ellipse2D or Arc2D Shape from spec' points and arc angles:
		Arc2D shapeEllipse = new Arc2D.Double(getArcType());
		shapeEllipse.setAngleExtent(-getRotatedAngleExtent());
		shapeEllipse.setAngleStart(-getRotatedAngleStart());
		shapeEllipse.setFrameFromCenter(0,0,getGreatAxisLength()/2,Math.abs(getSmallAxisLength()/2));
		AffineTransform rot=new AffineTransform();
		rot.setToIdentity();
		rot.translate(getPointX(P_CENTER),getPointY(P_CENTER));
		rot.rotate(getRotationAngle()); 
		// rot = translate o rotate
		if (getSmallAxisLength()<0) rot.scale(1,-1);
		Shape shape=rot.createTransformedShape(shapeEllipse);
		
		// 2) flatten path in terms of Cubic segments and build curve on-the-fly
		PicMultiCurve curve = new PicMultiCurve(false, this.attributeSet); // open
		PathIterator pi = shape.getPathIterator(null);
		double[] coords = new double[6]; // (x0,y0,x1,y1,x2,y2)
		while(!pi.isDone()){
			int type = pi.currentSegment(coords); // fill coords array with current segment
			switch (type){
			case PathIterator.SEG_MOVETO: // reinit "new"
				curve.addPoint(new PicPoint(coords[0], coords[1]));
				//System.out.println(" SEG_MOVETO "+new PicPoint(coords[0], coords[1]));
				break;
			case PathIterator.SEG_QUADTO: // reinit "new"
				//System.out.println(" SEG_QUADTO "+new PicPoint(coords[0], coords[1])+", "+new PicPoint(coords[2], coords[3]));
				break;
			case PathIterator.SEG_CUBICTO: // reinit "new"
				curve.curveTo(new PicPoint(coords[0],coords[1]), new PicPoint(coords[2],coords[3]), new PicPoint(coords[4], coords[5]));
				//System.out.println(" SEG_CUBICTO "+new PicPoint(coords[0], coords[1])+", "+new PicPoint(coords[2], coords[3])+", "+new PicPoint(coords[4], coords[5]));
				break;
			case PathIterator.SEG_LINETO: // update "new"
				curve.lineTo(new PicPoint(coords[0],coords[1]));
				//System.out.println(" SEG_LINETO "+new PicPoint(coords[0], coords[1]));
				break;
			case PathIterator.SEG_CLOSE:
				//System.out.println(" SEG_CLOSE "+new PicPoint(coords[0], coords[1]));
				curve.setClosed(true);
				// ensure "closing" segment is straight, i.e. move control-points to the same position
				// as their associated subdivision point (notice that using setStraight() does NOT work here, because
				// this would also alter the position of other control points, which would break the job made by the
				// path iterator).
				int Npts = curve.getNumberOfSpecificationPoints();
				curve.setSpecificationPoint(Npts-1, curve.getSpecificationPoint(0));
				curve.setSpecificationPoint(Npts-2, curve.getSpecificationPoint(Npts-3));
				curve.fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
				break;
			default:
			}
			pi.next();
		}

		return curve;
	}

	
	//////////////////////////////////
	/// OTHER FIELDS ACCESSORS
	//////////////////////////////////

	/**
	 * Return the rotation angle, ie the angle between the great axis of the ellipse and the horizontal axis.<br>
	 * @return the rotation angle in radians (CCW)
	 * @see #rotationAngle
	 */
	public double getRotationAngle(){
		return rotationAngle;
	}


	/**
	 * Return the length of the great-axis of this ellipse/arc.
	 */
	public double getGreatAxisLength(){
		return greatAxis;
	}

	/**
	 * Return the length of the small-axis of this ellipse/arc.
	 */
	public double getSmallAxisLength(){
		return smallAxis;
	}


	/**
	 * Set the angle start, as measured in the parallelogram basis, then fire a changed-update.
	 * Arc always extends CCW from angle-start to angle-end, ie angle-extent is always positive.
	 * @param The starting angle of the arc, in degrees,
	 *        counted COUNTERCLOCKWISE (aka trigonometric), 
	 *        and measured in the parallelogram basis, ie {@link #skewAngleStart skewAngleStart}.
	 */
	public void setAngleStart(double angleStart){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "angleStart="+angleStart);
		this.skewAngleStart = angleStart % 360;// in [-360,360] !!! (java is weird)
		if (skewAngleStart > 180) skewAngleStart -= 360;
		if (skewAngleStart <= -180) skewAngleStart += 360; // [SR:pending] was "<=180" ; did i think right ?
		// make skewAngleEnd always .gt. skewAngleStart, but never .gt. skewAngleStart+360 :
		if (skewAngleEnd <= skewAngleStart+MAX_FLAT_ANGLE) skewAngleEnd += 360;
		else if (skewAngleEnd > skewAngleStart-MAX_FLAT_ANGLE+360) skewAngleEnd -= 360;
		updateRotatedAngles();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Returns {@link #skewAngleStart skewAngleStart}, ie the starting angle of the arc in degrees,
	 * measured in the frame defined by the parallelogram.
	 */
	public double getAngleStart(){
		return skewAngleStart;
	}

	/**
	 * Set the angle end, then fire a changed-update.
	 * Arc always extends CCW from angle-start to angle-end, ie angle-extent is always positive.
	 * @param The angle end of the arc, in degrees,
	 *        measured in the frame defined by the parallelogram, ie {@link #skewAngleEnd skewAngleEnd}.
	 */
	public void setAngleEnd(double angleEnd){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "angleEnd="+angleEnd);
		this.skewAngleEnd = angleEnd % 360;
		if (skewAngleEnd <= skewAngleStart+MAX_FLAT_ANGLE) skewAngleEnd += 360;
		if (skewAngleEnd <= skewAngleStart+MAX_FLAT_ANGLE) skewAngleEnd += 360;// there may be up to 360+180 degrees of difference so we need to do it twice
		if (skewAngleEnd > skewAngleStart-MAX_FLAT_ANGLE+360) skewAngleEnd -= 360;
		if (skewAngleEnd > skewAngleStart-MAX_FLAT_ANGLE+360) skewAngleEnd -= 360;
		updateRotatedAngles();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Returns {@link #skewAngleEnd skewAngleEnd}, ie angle end of the arc in degrees,
	 * measured in the frame defined by the parallelogram. (value of skewAngleStart)
	 */
	public double getAngleEnd(){
		return skewAngleEnd;
	}


	/**
	 * Sets the angle extent, keeping the angle start fixed, then fire a changed update.
	 * If not positive, 360 is added to it.
	 * @param Angular extent of the arc, in degrees, counted COUNTERCLOCKWISE
	 *        as measured in parallelogram basis.
	 */
	public void setAngleExtent(double angleExtent){
		setAngleEnd(skewAngleStart + angleExtent);
	}


	/**
	 * Returns the extent of the arc in degrees (CCW) as measured in the parallelogram basis.
	 * Note that this method returns 360 for a plain ellipse, not 0.
	 * @return an angle in (0,360]
	 */
	public double getAngleExtent(){
		return   skewAngleEnd - skewAngleStart; // >0
	}



	/**
	 * Return {@link #rotatedAngleStart rotatedAngleStart}, ie the angle start of the arc in degrees (CCW),
	 * as measured in the frame defined by the axes of the ellipse.
	 */
	public double getRotatedAngleStart(){
		return rotatedAngleStart;
	}

	/**
	 * Returns {@link #rotatedAngleEnd rotatedAngleEnd}, ie the angle end of the arc in degrees
	 *        as measured in the frame defined by the axes of the ellipse.
	 */
	public double getRotatedAngleEnd(){
		return rotatedAngleEnd;
	}

	/**
	 * Returns the angle extent of the arc in degrees
	 * as measured in the frame defined by the axes of the ellipse.
	 */
	public double getRotatedAngleExtent(){
		return rotatedAngleEnd-rotatedAngleStart; // >0
	}

	/**
	 * Returns the closure type, ie one of CHORD, PIE or OPEN predefined constant fields.
	 */
	public int getArcType(){
		return closure;
	}

	/**
	 * Returns the closure type as a String, ie one of "chord", "pie" or "open"
	 */
	public String getArcTypeAsString(){
		switch (closure){
			case OPEN : return "open";
			case PIE : return "pie";
			case CHORD : return "chord";
			default : return null;
		}
	}

	/**
	 * Sets the closure type to one of CHORD, PIE or OPEN.
	 */
	public void setArcType(int closure){
		this.closure=closure;
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Return the index of the last point that can be retrieved by 
	 * {@link #getPoint getPoint}.
	 */
	public int getLastPointIndex(){
		return LAST_PT;
	}

	///////////////////////
	/// BOUNDING BOX
	///////////////////////

	// inherited

	///////////////////////
	/// ELLIPSE SPECIFIC
	///////////////////////

	/**
	 * Return true if this ellipse is flat.
	 */
	public boolean isFlat(){
		return ((smallAxis==0) || (greatAxis==0));
	}

	/**
	 * Returns TRUE if this ellipse is circular, ie if smallAxis==greatAxis.
	 * @see #getSmallAxisLength
	 * @see #getGreateAxisLength.
	 */
	public boolean isCircular(){
		return Math.abs(greatAxis-Math.abs(smallAxis)) < 0.0000001 * greatAxis;
	}
	
	/**
	 * @deprecated use isCircular() instead
	 */
	public boolean isCircle(){
		return Math.abs(greatAxis-Math.abs(smallAxis)) < 0.0000001 * greatAxis;
	}

	/**
	 * Return true if this PicEllipse is a plain ellipse, ie if angle extent equals 360.
	 */
	public boolean isPlain(){
		if (Math.abs(getAngleExtent()) < MAX_FLAT_ANGLE) return true; // [SR:pending] better use another specific constant field ?
		if (Math.abs(getAngleExtent()-360) < MAX_FLAT_ANGLE) return true; // [SR:pending] i'm getting totally confused with this f*** angle-extent ;-( maybe this line is useless...
		return false;
	}
	
	/**
	 * Returns true if this PicEllipse is an arc, ie if angle extent does not equl 360
	 */
	public boolean isArc(){
		return !isPlain();
	}
	
	/**
	 * Return true if this PicEllipse has a null rotation angle along the x-axis
	 */
	public boolean isRotated(){
		if (Math.abs(rotationAngle) < MAX_FLAT_ANGLE) return false;
		return true;
	}
	
	/**
	 * Set angle parameters so that this ellipse is a plain ellipse.
	 */
	public void setPlain(){
		setAngleStart(0);
		setAngleEnd(0);
	}

	/**
	 * Returns whether this arc or ellipse is closed or not. This depends on the closure type, and this PicEllipse being plain or not. 
	 */
	public boolean isClosed(){
		return (isPlain() || getArcType()==CHORD || getArcType()==PIE);
	}
	
	/**
	 * Compute the axes of the ellipse (smallAxis, greatAxis), and the rotated angles in terms of the skew angles.
	 */
	// [SR:pending] new version using recent enhancements in class PicPoint. Still buggy... 
	protected void _updateAxis(){

		double uX = l2rVec.x; // buggy ! these vectors don't have the expected length now and often. See my comment on top of updateAxis()
		double vX = b2tVec.x;
		double uY = l2rVec.y;
		double vY = b2tVec.y;
		double det=l2rVec.det(b2tVec);
		double lu=l2rVec.norm2();
		double lv=b2tVec.norm2();
		
		//if parallelogram is flat :
		if (det*det <= MAX_FLAT_ANGLE*lu*lv) {
			this.smallAxis=0;
			this.greatAxis=Math.sqrt(lu+lv);
			if (lu>lv) { rotationAngle=PicVector.X_AXIS.angle(l2rVec);}
			else { rotationAngle=PicVector.X_AXIS.angle(b2tVec);}
			updateRotatedAngles();
			return;
		}//end of case of flat parallelogram

		//l'equation de l'ellipse est donnee par Q(x,y)=det^2 avec Q= [[vY^2+uY^2, -vY*vX-uY*uX], [-vY*vX-uY*uX, vX^2+uX^2]]
		//le det de Q est det^2, la trace de Q est lu+lv
		double delta=(lu+lv)*(lu+lv)-4*det*det;
		if (delta <=0){//circle [delta should never be <0]
			this.greatAxis=Math.abs(det)/Math.sqrt(uX*uX+vX*vX);//the denominator is non-zero since otherwise, we are flat.
			this.smallAxis=det/greatAxis;//==+- greatAxis
			rotationAngle=0;
			updateRotatedAngles();
			return;
		}
		double lambda=((lu+lv)-Math.sqrt(delta))/2;//small eigenval=eigenval of great axis
		
		PicVector greatAxisDir = ptBuffer1; // use class buffer
		//choose the bigger coef to compute eigenvector
		if (Math.abs(vY*vY+uY*uY-lambda)>Math.abs(vX*vX+uX*uX-lambda)){
			greatAxisDir.x = vX*vY+uX*uY;
			greatAxisDir.y = vY*vY+uY*uY-lambda;
		}
		else {
			greatAxisDir.x = vX*vX+uX*uX-lambda;
			greatAxisDir.y = vX*vY+uX*uY;
		}
		rotationAngle=PicVector.X_AXIS.angle(greatAxisDir);
		greatAxis=Math.abs(det)/Math.sqrt(lambda); //lambda always > 0
		smallAxis=det/greatAxis;
		updateRotatedAngles();
	}//updateAxis

	/**
	 * Compute the axes of the ellipse (smallAxis, greatAxis), and the rotated angles in terms of the skew angles.
	 */
	 // [SR:pending] vincent, this is your old version. Remove it if you think mine is ok (but i don't think it is ok thus far ;-))).
	protected void updateAxis(){
		double uX=getPointX(P_BR)-getPointX(P_BL); // [SR:pending] we could use b2tVec and l2rVec, but this would
		double vX=getPointX(P_TR)-getPointX(P_BR); // require calling computeParalleloBasis() beforehands, and there
		double uY=getPointY(P_BR)-getPointY(P_BL); // again flat parallelo needs special treatment if we don't want
		double vY=getPointY(P_TR)-getPointY(P_BR); // to lose track of parallelo axis orientation... 
		double det=uX*vY-uY*vX;
		double lu=uX*uX+uY*uY;
		double lv=vX*vX+vY*vY;
		
		//if parallelogram is flat :
		if (det*det <= MAX_FLAT_ANGLE*lu*lv) {
			this.smallAxis=0;
			this.greatAxis=Math.sqrt(lu+lv);
			if (lu>lv) { rotationAngle=computeAngleWithHorizontal(uX,uY); }
			else { rotationAngle=computeAngleWithHorizontal(vX,vY); }
			updateRotatedAngles();
			return;
		}//end of case of flat parallelogram

		//l'equation de l'ellipse est donnee par Q(x,y)=det^2 avec Q= [[vY^2+uY^2, -vY*vX-uY*uX], [-vY*vX-uY*uX, vX^2+uX^2]]
		//le det de Q est det^2, la trace de Q est lu+lv
		double delta=(lu+lv)*(lu+lv)-4*det*det;
		if (delta <=0){//circle [delta should never be <0]
			this.greatAxis=Math.abs(det)/Math.sqrt(uX*uX+vX*vX);//the denominator is non-zero since otherwise, we are flat.
			this.smallAxis=det/greatAxis;//==+- greatAxis
			rotationAngle=0;
			updateRotatedAngles();
			return;
		}
		double lambda=((lu+lv)-Math.sqrt(delta))/2;//small eigenval=eigenval of great axis
		double greatAxisX,greatAxisY;
		//choose the bigger coef to compute eigenvector
		if (Math.abs(vY*vY+uY*uY-lambda)>Math.abs(vX*vX+uX*uX-lambda)){
			greatAxisX=vX*vY+uX*uY;
			greatAxisY=vY*vY+uY*uY-lambda;
		}
		else {
			greatAxisX=vX*vX+uX*uX-lambda;
			greatAxisY=vX*vY+uX*uY;
		}
		rotationAngle=computeAngleWithHorizontal(greatAxisX,greatAxisY);
		greatAxis=Math.abs(det)/Math.sqrt(lambda); //lambda always > 0
		smallAxis=det/greatAxis;
		updateRotatedAngles();
	}//updateAxis


	/**
	 * Updates values of rotatedAngleStart and  rotatedAngleEnd. 
	 * Axes must be updated beforehands for this computation to be valid.
	 * if greatAxis==0 (implies smallAxis==0), we keep the old values [case of ellipse reduced to a point]. 
	 * Flat case is handled by fromSkewToRotated.
	 */
	protected void updateRotatedAngles(){
		if (skewAngleEnd-skewAngleStart == 360){
			rotatedAngleStart=0;
			rotatedAngleEnd=360;
			return;
		}
		if (greatAxis==0) return;
		rotatedAngleStart=fromSkewToRotated(skewAngleStart);
		rotatedAngleEnd=fromSkewToRotated(skewAngleEnd);
		if (rotatedAngleEnd <= rotatedAngleStart +MAX_FLAT_ANGLE) rotatedAngleEnd += 360;

	}


	/**
	 * Converts an angle measured wrt the parallelogram basis to an 
	 * angle as measured in the ellipse basis.
	 * <p>
	 * More precisely, if U and V are the vectors defined by the sides of the parallelogram, 
	 * and U' V' are the vectors corresponding
	 * to the small and great axes of the ellipse, and if W=cos(theta)*U+sin(theta)*V ,
	 * then computes theta' such that W=cos(theta')*U'+sin(theta')*V'.
	 * <p>
	 *  Axes must be updated beforehands for this computation to be valid. 
	 * This method returns 0 whenever greatAxis==0 
	 * [but one should rather keep old values of rotatedAngles in this case.]
	 * @param angle an angle in degrees
	 */
	private double _fromSkewToRotated(double angle){
		// [SR:en_cours] un chti coup de refactoring ici et c'est fini. 'tain la lerga. Allez, au dod'. 
		if (greatAxis==0) return 0;
		//x,y= coordinates of the vector in the ellipse-basis.
		// buggy : same shortcoming over again => l2rVec and b2tVec don't have the expected length ;-(
		double x= 0.5 * l2rVec.x * Math.cos(Math.toRadians(angle)) + 0.5 * b2tVec.x * Math.sin(Math.toRadians(angle));
		double y= 0.5 * l2rVec.y * Math.cos(Math.toRadians(angle)) + 0.5 * b2tVec.y * Math.sin(Math.toRadians(angle));
		// rotate by -rotationAngle
		double u= x*Math.cos(rotationAngle)+y*Math.sin(rotationAngle);
		double v= -x*Math.sin(rotationAngle)+y*Math.cos(rotationAngle);
		if (smallAxis==0) return Math.toDegrees(Math.acos(u/greatAxis));
		return Math.toDegrees(computeAngleWithHorizontal(u/greatAxis,v/smallAxis)); //no division by zero
	}

    private double fromSkewToRotated(double angle){
		// [code de vincent]
		if (greatAxis==0) return 0;
		//x,y= coordinates of the vector in the canonical basis.
		double x= 0.5*(super.getPointX(P_BR)-super.getPointX(P_BL))* Math.cos(Math.toRadians(angle))
	          + 0.5*(super.getPointX(P_TR)-super.getPointX(P_BR))* Math.sin(Math.toRadians(angle));
		double y= 0.5*(super.getPointY(P_BR)-super.getPointY(P_BL))* Math.cos(Math.toRadians(angle))
	          + 0.5*(super.getPointY(P_TR)-super.getPointY(P_BR))* Math.sin(Math.toRadians(angle));
		// rotate by -rotationAngle
		double u= x*Math.cos(rotationAngle)+y*Math.sin(rotationAngle);
		double v= -x*Math.sin(rotationAngle)+y*Math.cos(rotationAngle);
		if (smallAxis==0) return Math.toDegrees(Math.acos(u/greatAxis));
		return Math.toDegrees(computeAngleWithHorizontal(u/greatAxis,v/smallAxis)); //no division by zero 
    }

	
	/**
	 * Compute the angle (in Radians) of a vector (x,y) with the horizontal in -Pi,Pi
	 */
	private double computeAngleWithHorizontal(double x,double y){
		double l=Math.sqrt(x*x+y*y);
		if (l==0){return 0;}
		double X=x/l; double Y=y/l;
		if (x>=0) { return 2*Math.atan(Y/(1+X)); }
		else {
			if (y>=0) return Math.PI-2*Math.atan(Y/(1-X));
			else return -Math.PI-2*Math.atan(Y/(1-X));
		}
	}

	/**
	 * Compute the angle (in Radians) of a vector (x,y) with the horizontal in -Pi,Pi
	 */
	private double computeAngleWithHorizontal(PicVector p){
		return PicVector.X_AXIS.angle(p);
	}





	/**
	 * Utility for computing arrow direction.
	 * @return a unit-1 vector tangent to the arc, at phi = start angle, pointing "outwardly"
	 * @param pt a preallocated PicPoint that get filled with the result ; a new one is allocated if pt==null
	 */
	public PicPoint getTangentAtAngleStart(PicVector pt){
		if (pt==null) pt = new PicVector();
		
		pt.setCoordinates(getSpecificationPoint(P_BL),getSpecificationPoint(P_BR));
		pt.scale(Math.sin(Math.toRadians(skewAngleStart)));
		
		ptBuffer1.setCoordinates(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR));
		ptBuffer1.scale(-Math.cos(Math.toRadians(skewAngleStart)));
		
		pt.add(ptBuffer1);
		pt.normalize();
		return pt;
	}


	/**
	 * Utility for computing arrow direction.
	 * @return a unit-1 vector tangent to the arc, at phi = end angle, pointing "outwardly"
	 * @param pt a preallocated PicPoint that get filled with the result ; a new one is allocated if pt==null
	 */
	public PicPoint getTangentAtAngleEnd(PicVector pt){
		if (pt==null) pt = new PicVector();
		
		pt.setCoordinates(getSpecificationPoint(P_BL),getSpecificationPoint(P_BR)); // aka l2rVec with correct length
		pt.scale(Math.sin(Math.toRadians(skewAngleEnd)));
		
		ptBuffer1.setCoordinates(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR)); // aka b2tVec
		ptBuffer1.scale(-Math.cos(Math.toRadians(skewAngleEnd)));
		
		pt.add(ptBuffer1);  // pt = pt+ptBuffer1
		pt.normalize();
		pt.inverse(); // angle end !!!
		return pt;
	}


	/**
	 * Implementation of the Object.toString() method, used for debugging purpose
	 * @author Sylvain Reynal
	 * @since PicEdt 1.1.4
	 */
	public String toString() {

		String s = super.toString()
		           + "\n\t greatAxis=" + greatAxis
		           + ", smallAxis=" + smallAxis
		           + ", rotationAngle="+ rotationAngle
		           + ", skewAngleStart=" + skewAngleStart
		           + ", skewAngleEnd=" + skewAngleEnd
		           + ", rotatedAngleStart=" + rotatedAngleStart
		           + ", rotatedAngleEnd=" + rotatedAngleEnd
				   + ", isPlain=" + (isPlain() ? "true":"false")
				   + ", closure=" + getArcTypeAsString();

		return s;
	}

	////////////////////////////////
	//// GUI
	////////////////////////////////

	/**
	 * @return a Customizer for geometry editing
	 * @since jpicedt 1.3.3
	 */
	public AbstractCustomizer createCustomizer(){
		return new Customizer();
	}

	/**
	 * geometry customizer
	 */
	class Customizer extends PicParallelogram.Customizer implements ActionListener {

		private DecimalNumberField ellipseAngleStartTF, ellipseAngleEndTF;
		private JComboBox arcTypeList;
		private boolean isListenersAdded = false; // flag

		private ImageIcon[] arcTypeIcons = {
			PEToolKit.createImageIcon("action.draw.ArcChord"), 
		    PEToolKit.createImageIcon("action.draw.ArcOpen"),
			PEToolKit.createImageIcon("action.draw.ArcPie")};
			
		/** create a new Customizer for editing the geometry and attributes of this PicEllipse */ 
		public Customizer(){

			super(); // add parallelogram related widgets
			
			// following panel = angles only + arc type
			JPanel p = new JPanel(new GridLayout(3,3,5,5));

			// line 1: arc start angle
			p.add(PEToolKit.createJLabel("attributes.EllipseStartAngle"));
			p.add(ellipseAngleStartTF = new DecimalNumberField(5));
			p.add(new JLabel("deg"));

			// line 2: arc extent angle
			p.add(PEToolKit.createJLabel("attributes.EllipseEndAngle"));
			p.add(ellipseAngleEndTF = new DecimalNumberField(5));
			p.add(new JLabel("deg"));
			
			// line 3 : arc type (chord, pie or open)
			arcTypeList = PEToolKit.createComboBox(arcTypeIcons);
			p.add(new JLabel(" "));
			p.add(arcTypeList);
			p.add(new JLabel(" "));
			

			add(p, BorderLayout.CENTER);
		}
		
		/** add action listeners to widgets to reflect changes immediately */ 
		private void addActionListeners(){
			if (isListenersAdded) return; // already done
			ellipseAngleStartTF.addActionListener(this);
			ellipseAngleEndTF.addActionListener(this);
			arcTypeList.addActionListener(this);
			isListenersAdded = true;
		}

		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){

			super.load();
			ellipseAngleStartTF.setValue(getAngleStart());
			ellipseAngleEndTF.setValue(getAngleEnd());
			if (getArcType()==CHORD) arcTypeList.setSelectedIndex(0);
			else if (getArcType()==OPEN) arcTypeList.setSelectedIndex(1);
			else if (getArcType()==PIE) arcTypeList.setSelectedIndex(2);
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded 
			addActionListeners(); // done the first time load is called
		}

		/**
		 * update Element's properties
		 */
		public void store(){

			super.store();
			setAngleStart(ellipseAngleStartTF.getValue());
			ellipseAngleStartTF.setValue(getAngleStart());//redisplay after 2 PI-modulo
			setAngleEnd(ellipseAngleEndTF.getValue());
			ellipseAngleEndTF.setValue(getAngleEnd()); //redisplay after 2 PI-modulo
			switch (arcTypeList.getSelectedIndex()){
				case 0 : setArcType(Arc2D.CHORD); break;
				case 1 : setArcType(Arc2D.OPEN); break;
				case 2 : setArcType(Arc2D.PIE); break;
				default:
			}
			updateAxis();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}

		public void actionPerformed(ActionEvent e){
			store();
		}
		
		public String getTitle(){ 
			return PicEllipse.this.getName();
		}
	}

	/////////////////////////////////
	/// Trashed code
	/////////////////////////////////
	
	// [SR:pending] dunno what to do with this one. Remove it ? (in any case, refactor method name to something more self-explanatory)
	//    /**
	//    * Compute, in degrees, the ``angle'' of a vector with the great axis (for the scalar product making the ellipse a circle)
	// * ie this is a way to compute startAngle, or endAngle from a vector. Values in (-180,180]
	// * Of course, the axes are supposed to be updated.
	// */

	//    private double getRelativeAngle(double x, double y){
	//	if (isFlat()) return 0;
	//	//first rotate by -rotationAngle
	//	double xx=Math.cos(rotationAngle)*x+Math.sin(rotationAngle)*y;
	//	double yy=-Math.sin(rotationAngle)*x+Math.cos(rotationAngle)*y;
	//	//rescale by great and small axis
	//	xx=xx/greatAxis;//non zero
	//	yy=yy/smallAxis;
	//	return Math.toDegrees(computeAngleWithHorizontal(xx,yy));
	//}

	
} 
