/*
 PicMultiCurve.java - December 25, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import jpicedt.graphic.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.widgets.*;
import jpicedt.graphic.view.HitInfo;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.table.*;
import java.util.*;

/**
 * PicMultiCurve is basically an AbstractCurve where each subdivision point has
 * additionnal features regarding 2nd-order smoothness and symmetry. These
 * features, however, are dynamic in the sense that they may be enforced by passing
 * geometric constraints to the {@link #setPoint setPoint()} method instead of
 * having them tightly bound to the class. Besides, 
 * {@link #setSmooth setSmooth()} and {@link #setSymmetric setSymmetric()} methods may
 * be used to dynamically enforce these features w/o having to move control-points.
 * <p>
 * @author    Vincent Guirardel, Sylvain Reynal
 * @since     jpicedt 1.3.3
 * @version $Id: PicMultiCurve.java,v 1.28.2.1 2007/09/02 11:56:27 reynal Exp $ 
 */
public class PicMultiCurve extends AbstractCurve implements CustomizerFactory,PicMultiCurveConvertable {

	/** predefined constraint for setPoint() */
	public static final BasicEditPointConstraint SYMMETRY = new BasicEditPointConstraint("symmetry");
	/** predefined constraint for setPoint() */
	public static final BasicEditPointConstraint SMOOTHNESS = new BasicEditPointConstraint("smoothness");
	/** predefined constraint for setPoint() */
	public static final BasicEditPointConstraint FREELY = new BasicEditPointConstraint("free-control");

	//////////////////////////
	//// PRIVATE FIELDS
	//////////////////////////

	private PicVector ptBuffer1 = new PicVector(); // buffer used e.g. by set/getPoint to minimize object creation
	private PicVector ptBuffer2 = new PicVector(); // Please indicate which buffer you use in the first line of each method that use buffers.
					// also feel free to add new buffers here if necessary, especially if you're writing methods which are
					// critical in terms of UI performance. The most secure policy is to create one new buffer per method
					// unless you know EXACTLY what you're doing ;-) In any case, remember: garbage collecting always is 
					// exceedingly time-consuming...
	
	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Create a new empty open multi curve, with a default attribute set.
	 */
	public PicMultiCurve() {
		this(false);
	}

	/**
	 * Create a new empty multi curve, with a default attribute set.
	 * @param closed  whether this curve is closed or not
	 */
	public PicMultiCurve(boolean closed) {
		super(closed);
	}

	/**
	 * Create a new empty multi curve with the given set of attribute
	 * @param closed  whether this curve is closed or not
	 * @param set     attribute set to be bound to this element
	 */
	public PicMultiCurve(boolean closed, PicAttributeSet set) {
		this(closed);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * Create a new open multi curve reduced to the given point, and a default attribute set.
	 */
	public PicMultiCurve(PicPoint pt1) {
		super(0, false); // reminder : 0 segment (hence 1pt), open
		getSpecificationPoint(0).setCoordinates(pt1);
	}

	/**
	 * Create a new open multi curve reduced to the given point, and the given attribute set
	 */
	public PicMultiCurve(PicPoint pt1, PicAttributeSet set) {
		super(0, false); // reminder : 0 segment (hence 1pt), open
		getSpecificationPoint(0).setCoordinates(pt1);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * Create a new line, i.e. a new straight open curve using the two given points, and a default attribute set.
	 */
	public PicMultiCurve(PicPoint pt1, PicPoint pt2) {
		this(pt1);
		lineTo(pt2);
	}

	/**
	 * Create a new straight open multi curve with the two given points, and the given attribute set.
	 */
	public PicMultiCurve(PicPoint pt1, PicPoint pt2, PicAttributeSet set) {
		this(pt1,pt2);
		this.attributeSet = new PicAttributeSet(set);
	}

	/**
	 * Create an open single cubic bezier curve from the four given points, and a default attribute set
	 */
	public PicMultiCurve(PicPoint pt1, PicPoint ctrl1, PicPoint ctrl2, PicPoint pt2) {
		this(pt1);
		curveTo(ctrl1,ctrl2,pt2);
	}

	/**
	 * Create a new multi-curve filled with the given array of PicPoint's, each
	 * segment being a curved Bezier segment.
	 * <ul>
	 * <li> if the number of points mod 3 equals 0, then it will be closed. 
	 * <li> if the number of points mod 3 equals 1 then the curve will be open, 
	 * <li> if the number of points mod 3 equals 2, curve is open and last segment is a straight one. 
	 * </ul>
	 * Smoothness and symmetry properties are automatically asserted
	 * by checking if the given points do satisfy them.
	 */
	public PicMultiCurve(PicPoint[] pts, PicAttributeSet set) {
		isClosed = (pts.length%3==0);
		if (pts.length%3 == 2){ // open curve, last segment is straight
			for (int i = 0; i < pts.length-1; i++) this.pts.add(pts[i].clone());
			lineTo(new PicPoint(pts[pts.length-1]));
		}
		else for (int i = 0; i < pts.length; i++) this.pts.add(pts[i].clone());
		
		this.attributeSet = new PicAttributeSet(set);
	}


	/**
	 *  constructor for conversion of smoothPoly into multiCurve
	 * @param poly  the smoothPoly to be converted
	 */
	public PicMultiCurve(PicSmoothPolygon poly) {
		super(poly);
	}

	/**
	 *  constructor for conversion of pscurve into multiCurve
	 * @param poly  the PsCurve to be converted
	 */
	public PicMultiCurve(PicPsCurve curve) {
		super(curve);
	}


	/**
	 * "cloning" constructor (to be used by clone())
	 *
	 * @param curve  the curve to be cloned
	 */
	public PicMultiCurve(PicMultiCurve curve) {
		super(curve);
	}

	/**
	 * Override Object.clone() method
	 */
	public Object clone() {
		return new PicMultiCurve(this);
	}

	/**
	 * @return   a localised string that represents this object's name
	 */
	public String getName() {
		return jpicedt.Localizer.currentLocalizer().get("model.MultiCurve");
	}

	/**
	 * converts this ELement to a PicMultiCurve, i.e., itself (no copy).
	 */
	 public PicMultiCurve convertToMultiCurve(){
		 return this;
	 }

	//////////////////////////////////
	/// GLOBAL GEOM. OPERATIONS
	//////////////////////////////////

	/**
	 * Returns the shortest distance b/w the end-points of this curve and those of the given one.
	 * @since jpicedt 1.4pre5
	 */
	public double distance(PicMultiCurve other){
		 double distance = this.getSpecificationPoint(0).distance(other.getSpecificationPoint(0));
		 distance = Math.min(distance, this.getSpecificationPoint(0).distance(other.getSpecificationPoint(other.getLastPointIndex())));
		 distance = Math.min(distance, this.getSpecificationPoint(this.getLastPointIndex()).distance(other.getSpecificationPoint(0)));
		 distance = Math.min(distance, this.getSpecificationPoint(this.getLastPointIndex()).distance(other.getSpecificationPoint(other.getLastPointIndex())));
		 return distance;
	}
	
	/**
	 * Returns the curve in the given collection that is closest to this one in the sense of the distance() method.
	 * In case of two or more curves being equally close to this one, the first occurence in the collection is
	 * returned.
	 * @param l a list of PicMultiCurveConvertable's ; l may contain this curve, in which case it's simply skipped.
	 * @return null if l doesn't contain any PicMultiCurveConvertable.
	 * @since jpicedt 1.4pre5
	 */
	public PicMultiCurveConvertable fetchClosestCurve(Collection l){
		PicMultiCurveConvertable closest = null;
		double d = Double.MAX_VALUE;
		for (Iterator it = l.iterator(); it.hasNext();){
			Object o = it.next();
			if (o==this) continue;
			if (o instanceof PicMultiCurveConvertable){
				PicMultiCurve c = ((PicMultiCurveConvertable)o).convertToMultiCurve();
				if (this.distance(c) < d) {
					closest = (PicMultiCurveConvertable)o;
					d = this.distance(c);
				}
			}
		}
		return closest;
	}
	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////

	/**
	 * Reverse the order in which points are stored in the protected <b>pts</b> array.
	 * If the curve is open, this results in the origin being swapped with the last point, etc...
	 * May be useful when, e.g., joining two curves.
	 * @since jpicedt 1.4pre5
	 */
	public void reverseIndexing(){
		ArrayList _pts=new ArrayList(); // tmp
		for (int i=pts.size(); --i>=0; ){
			_pts.add(pts.get(i));
		}
		this.pts = _pts;
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	/**
	 * Join this curve and the given Element after proper conversion of the latter 
	 * to a PicMultiCurve, by appending the control points of the given curve to this one.
	 * The order in which points are appended, i.e., wheter an index-reverse takes place before appending
	 * (see reverseIndexing), is
	 * such that both curves get joined along their shortest path. If both curves are not in contact, a segment is
	 * added between them (through a call to lineto()).
	 * @since jpicedt 1.4pre5
	 */
	public void join(PicMultiCurveConvertable c){
		if (c==null) return;
		PicMultiCurve other = c.convertToMultiCurve(); // if c is a multicurve, simply returns itself.
		//System.out.println("Joining to " + other);
		// 1) let A=this curve and B=other curve, _0 and _N the first and last point indices,
		//    check which distance is the shortest amidst:
		//    d(A_0,B_0): reverse A, append B
		//	  d(A_0,B_N): reverse A and B, append B
		//    d(A_N,B_0): append B
		//    d(A_N,B_N): reverse B, append B
		double dA0B0 = this.getSpecificationPoint(0).distanceSq(other.getSpecificationPoint(0));
		double dA0BN = this.getSpecificationPoint(0).distanceSq(other.getSpecificationPoint(other.getLastPointIndex()));
		double dANB0 = this.getSpecificationPoint(this.getLastPointIndex()).distanceSq(other.getSpecificationPoint(0));
		double dANBN = this.getSpecificationPoint(this.getLastPointIndex()).distanceSq(other.getSpecificationPoint(other.getLastPointIndex()));
		double[] distances = new double[]{dA0B0,dA0BN,dANB0,dANBN};
		Arrays.sort(distances);
		double distMin = distances[0];
		if (distMin==dA0B0){
			this.reverseIndexing();
		}
		else if (distMin==dA0BN){
			this.reverseIndexing();
			other.reverseIndexing();
		}
		else if (distMin==dANBN){
			other.reverseIndexing();
		}
		// 2) append B:
		// check if we must add a segment:
		boolean isSegmentAdded = false;
		final int joinSegIdx = this.getNumberOfSegments(); // "join" segment if any
		if (this.getSpecificationPoint(this.getLastPointIndex()).distance(other.getSpecificationPoint(0))>0){
			//this.addPoint(other.getSpecificationPoint(0));
			this.lineTo(other.getSpecificationPoint(0));
			isSegmentAdded=true;
		}
		// then simply add other's "pts" to this's "pts":
		// BUT: skip first point to avoid duplicates (get(0), either was already added through lineto(), 
		// or curves are already in contact)
		for (int i = 1; i<other.pts.size(); i++){ 
			PicPoint pt = (PicPoint)other.pts.get(i);
			this.pts.add(new PicPoint(pt));
		}
		// ensure smoothness: [pending] not fully functional yet
		/*
		if (!isSegmentAdded){ // "direct" contact
			if (isStraight(joinSegIdx)==false && isStraight(joinSegIdx-1)==false){
				setSmooth(joinSegIdx);
				setSymmetric(joinSegIdx);
			}
		}
		else if (this.getNumberOfSegments() > joinSegIdx+1){ // make sure enough points were added
			// move to first subdiv point of second curve, which we want to smoothen 
			int subdiv = segmentToPointIndex(joinSegIdx+1, SUBDIVISION_POINT);
			int srcControl = getPBCBezierIndex(subdiv+1); 
			if (isValidBezierIndex(srcControl)){
				int destControl = getPBCBezierIndex(subdiv-1); // that point we wanna move so that smoothness is ensured
				PicPoint pt = getPoint(srcControl);
				pt.symmetry(getPoint(subdiv));
				getSpecificationPoint(destControl).setCoordinates(pt);
			}
		}*/
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	
	/**
	 * Set the coordinates of a Bezier point having the given index to the given location,
	 * possibly moving other control-points according to the given constraint.
	 * Available constraints are :
	 * <ul> 
	 * <li> SMOOTHNESS : enforces smoothness at nearby subdivision-point when moving a control-point ; N/A otherwise
	 * <li> SYMMETRY : enforces symmetry at nearby subdivision-point when moving a control-point ; N/A otherwise
	 * <li> SMOOTHNESS.and(SYMMETRY) : a combination of both.
	 * <li> FREELY : control-points are freely moved w/o any constraint being put on alternate control-points
	 * <li> null (default) : only the given point index gets moved ; this is aka editing a single specification point. This
	 *      is applicable for both subdivision- and control-points.
	 * </ul>
	 * Except in the <b>default case</b>, subdivision-points generally "drag" their control-points with them, 
	 * with the exception of straight-segment end-points which need a somewhat more complex handling.
	 * @see jpicedt.graphic.model.EditPointConstraintCollator
	 */
	public void setPoint(int index, PicPoint pt, EditPointConstraint constraint) {
		// this method use ptBuffer1
		
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("idx="+index);
		int ptType = getPointType(index);
		if (ptType==INVALID_POINT_INDEX) throw new IndexOutOfBoundsException(new Integer(index).toString());

		// aka setSpecificationPoint :
		if (constraint==null){
			getSpecificationPoint(index).setCoordinates(pt);
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
			return;
		}
		
		if (ptType==SUBDIVISION_POINT || ptType==FIRST_CURVE_END_POINT || ptType==LAST_CURVE_END_POINT) {
			// if we are at a subdivision point, don't have to care about Smoothness/Symmetry :
			// if neighbouring segments exist and are not straight , we translate neighbouring control points accordingly.
			// otherwise, we just have to retain smoothness at straight-segment's end-points : this is carried out by
			// projecting appropriate control-points onto the line directed by the straight-segment (a tough one indeed ;-)
			int thisSeg = pointToSegmentIndex(index);
			int prevSeg = getPBCSegmentIndex(thisSeg-1); // we don't take for granted these are valid indices !
			int nextSeg = getPBCSegmentIndex(thisSeg+1);
			PicVector mouseTranslation = ptBuffer1.setCoordinates(getSpecificationPoint(index),pt); // store for later use
			// the following are used for handling straight-segment smoothness only : it's vital to store these booleans
			// BEFORE moving anything !
			boolean smoothAtThisSubdiv, smoothAtPrevSubdiv=false, smoothAtNextSubdiv=false;
			boolean isThisSegStraight=false, isPrevSegStraight=false; // putain c'est bourrin ;-(
			smoothAtThisSubdiv = isSmooth(thisSeg);
			if (isValidBezierIndex(index+1)) {
				smoothAtNextSubdiv = isSmooth(nextSeg);
				isThisSegStraight = isStraight(thisSeg);
			}
			if (isValidBezierIndex(index-1)) {
				smoothAtPrevSubdiv = isSmooth(prevSeg);
				isPrevSegStraight = isStraight(prevSeg);
			}
			
			// and now we are entitled to move points...
			getSpecificationPoint(index).translate(mouseTranslation); // move subdiv point
			if (isValidBezierIndex(index+1)) getSpecificationPoint(index+1).translate(mouseTranslation); // (ptBuffer1)
			if (isValidBezierIndex(index-1)) getSpecificationPoint(index-1).translate(mouseTranslation);
			
			// and enforce smoothness at both end-points if applicable...
			if (isThisSegStraight) enforceSmoothnessOfStraightSegment(thisSeg, smoothAtThisSubdiv, smoothAtNextSubdiv);
			if (isPrevSegStraight) enforceSmoothnessOfStraightSegment(prevSeg, smoothAtPrevSubdiv, smoothAtThisSubdiv);
		}
		
		else { // CONTROL_POINT :
			int subdiv = getNearestSubdivisionPoint(index); // nearest subdivision-point
			int altControl = getAlternateControlPoint(index); // alternate control-point

			// then check if we have to modify altControl
			if (isValidBezierIndex(altControl)) {  
				int thisSeg = pointToSegmentIndex(index); // segment containing this control-point
				int altSeg = pointToSegmentIndex(altControl); // alternate segment = segment containing alternate control-point
				if (!isStraight(altSeg)){
					getSpecificationPoint(index).setCoordinates(pt); //[SR:pending] change to setSpecPoint after "fireEvent" gets removed from it
					if (constraint.imposes(SMOOTHNESS.and(SYMMETRY))){ // ah que c'est Beau la POO ;-)
						getSpecificationPoint(altControl).setCoordinates(pt).symmetry(getSpecificationPoint(subdiv));
					}
					else if (constraint.imposes(SMOOTHNESS)){
						double length = getSpecificationPoint(subdiv).distance(pt);
						if (length != 0) {
							double altlength = getSpecificationPoint(subdiv).distance(getSpecificationPoint(altControl));
							getSpecificationPoint(altControl).setCoordinates(pt).symmetry(getSpecificationPoint(subdiv));
							getSpecificationPoint(altControl).scale(getSpecificationPoint(subdiv), altlength/length);
						}
					}
					else if (constraint.imposes(SYMMETRY)){
						double altlength = getSpecificationPoint(subdiv).distance(getSpecificationPoint(altControl));
						if (altlength != 0) {
							double length = getSpecificationPoint(subdiv).distance(pt);
							getSpecificationPoint(altControl).scale(getSpecificationPoint(subdiv),length/altlength);
						}
						else {// if altlength ==0, we set altControl point to the symmetric of original control point.
							getSpecificationPoint(altControl).setCoordinates(pt);
							getSpecificationPoint(altControl).symmetry(getSpecificationPoint(subdiv));
						}
					}
					else if (constraint.imposes(FREELY)){
						// don't modify altControl
					}
				}
				// else if altSeg is straight, possibly enforce smoothness by restricting the movement of THIS control-point
				// to the line directed by the straight-segment.
				// Since for a straight segment, altControl is identical with the corresponding subdiv-point,
				// we leave it as is (i.e. we don't modify segment straightness)
				else { 
					getSpecificationPoint(index).setCoordinates(pt);
					if (constraint.imposes(SMOOTHNESS)){ // make correction so that control-point is aligned with straight-segment
						// by projecting onto the line directed by the straight-segment
						getSpecificationPoint(index).project(getSpecificationPoint(altSeg*3),getSpecificationPoint(altSeg*3+3));
					}
				}
			}
			else getSpecificationPoint(index).setCoordinates(pt); // no altControl => nothing else to do 
			//end of modification of altControl.
		}
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * Utility method that enforces smoothness at each end-point of the given straight segment 
	 * by moving alternate control-points
	 * (provided alternate segments aren't straight themselves).
	 * <br>
	 * The policy is to move each alternate control-point to the same position as setPoint() would do 
	 * when called with the index of this alternate control-point, that is,
	 * to project this control-point onto the line directed by the straight-segment (geez, this is an obscure sentence indeed ;-)
	 * <br>
	 * This is used e.g. by <code>setPoint()</code> when moving a subdivision-point belonging to a straight segment,
	 * because the usual "translate-all-points-at-once" scheme does not work anymore for straight segments (this would obviously
	 * break smoothness except for very exceptional cases).
	 * <br>
	 * @param segIndex index of a straight segment, following the segment numbering scheme. The given segment is
	 *        supposed to be straight (no check is done), but this will work with non-straight segment as well, yet with probably
	 *        amazing results....
	 * @param atFirstEndPoint whether to enforce smoothness at the first end-point
	 * @param atSecondEndPoint whether to enforce smoothness at the second end-point
	 */
	private void enforceSmoothnessOfStraightSegment(int segIndex, boolean atFirstEndPoint, boolean atSecondEndPoint){
		// Note : this private method uses "ptBuffer1/2".
		segIndex = getPBCSegmentIndex(segIndex);
		if (!isValidSegmentIndex(segIndex)) return ;
		
		int leftSubdiv = segmentToPointIndex(segIndex, SUBDIVISION_POINT);
		int rightSubdiv = getPBCBezierIndex(leftSubdiv+3);
		if (atFirstEndPoint){
			int altControl = getPBCBezierIndex(leftSubdiv-1);
			if (isValidBezierIndex(altControl)){
				double length = getSpecificationPoint(leftSubdiv).distance(getSpecificationPoint(altControl));
				// check if we must change sign of length if altControl belongs to the segment [leftSubdiv, leftSubdiv+3]
				// this is carried out by computing a dot product :
				ptBuffer1.setCoordinates(getSpecificationPoint(leftSubdiv),getSpecificationPoint(altControl));
				ptBuffer2.setCoordinates(getSpecificationPoint(leftSubdiv),getSpecificationPoint(leftSubdiv+3)).normalize();
				double dotProduct = ptBuffer1.dot(ptBuffer2);
				if (dotProduct < 0) length = -length;
				getSpecificationPoint(altControl).setCoordinates(getSpecificationPoint(leftSubdiv)).translate(ptBuffer2, length);
			}
		}
		if (atSecondEndPoint){
			int altControl = getPBCBezierIndex(leftSubdiv+4);
			if (isValidBezierIndex(altControl)){
				// see comment above
				double length = getSpecificationPoint(rightSubdiv).distance(getSpecificationPoint(altControl));
				ptBuffer1.setCoordinates(getSpecificationPoint(rightSubdiv),getSpecificationPoint(altControl));
				ptBuffer2.setCoordinates(getSpecificationPoint(rightSubdiv),getSpecificationPoint(leftSubdiv)).normalize();
				double dotProduct = ptBuffer1.dot(ptBuffer2);
				if (dotProduct < 0) length = -length;
				getSpecificationPoint(altControl).setCoordinates(getSpecificationPoint(rightSubdiv)).translate(ptBuffer2, length);
			}
		}
	}


	/**
	 * Adds a new subdivision point to the end of this curve. As far as control-points location goes,
	 * a SMOOTHNESS and SYMMETRY constraint is imposed (which as a nice side-effect results in the
	 * added segment automagically inheriting its "straightness" property from the last segment of the original curve).
	 * <br>
	 * This implementation relies on {@link #lineTo lineTo()}
	 * and {@link #curveTo curveTo()} methods.
	 */
	public void addPoint(PicPoint pt) {
		
		if (!hasValidSize()) throw new RuntimeException("wrong nb of specification points [error]"); // [SR:pending] use assert keyword instead
		
		if (getNumberOfSegments() < 1){
			if (getNumberOfSpecificationPoints()==0) {
				if (isClosed()){
					pts.add(pt.clone());pts.add(pt.clone());pts.add(pt.clone()); // because nb pts = 0,3,6,... so now we've got ONE segment
				}
				else pts.add(pt.clone()); // because nb pts = 0,1,4,...
			}
			else if (getNumberOfSpecificationPoints()==1) { // this SHOULD be an open curve (otherwise there's a bug somewhere ;-)
				// assert !isClosed();
				lineTo(pt);
			}
			// else should never happen (pending : use "assert")
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}
		else {
			if (isClosed()) splitSegment(getNumberOfSegments() - 1, pt); // split last segment
			else {
				PicPoint C1,C2; // 1st and 2nd control points
				int p = getNumberOfSpecificationPoints() - 1; // index of old end-point
				C1 = getSpecificationPoint(p-1,null); // safely copy last control-point
				C1.symmetry(getSpecificationPoint(p)); // C1 = symmetric of last control point wrt last subdivision point
			                     // note that this handles the case with straight segments as well ;-)
				
				// the following is too complicated and does not give nice results.
				// Take C2= so that the Bezier segment is symmetrical (if v1=is the control vector defined by C1,
				// and \vec PQ=vector betwen endpoints, we take control vector v2=symmetric of v1 wrt the line perpendicular to PQ
				// that is v2=v1-2*(v1.\vec PQ / ||PQ||^2 )\vec PQ (the dot is a dot product)
				// so C2=Q+(C1-P)-2*(v1.\vec PQ / ||PQ||^2 )\vec PQ
				// special case if P=Q
				//		double PQx=pt.x-ptsX[p];
				//		double PQy=pt.y-ptsY[p];
				//		double v1x=ptsX[p-1]-ptsX[p];
				//		double v1y=ptsY[p-1]-ptsY[p];
				//		double l=PQx*PQx+PQy*PQy;
				//		if (l!= 0){
				//		    C2.x=pt.x+v1x-2*(v1x*PQx+v1y*PQy)/l*PQx;
				//		    C2.y=pt.y+v1y-2*(v1x*PQx+v1y*PQy)/l*PQy;
				//			}
				//		else {
				//		    C2.x=C1.x;
				//		    C2.y=C1.y;
				//		};
				// this is MUCH simpler, and better: set 2nd control point = endpoint
				C2 = new PicPoint(pt);
				// now C1 and C2 are defined
				curveTo(C1, C2, pt);
			}
		}
	}

	/**
	 * If this curve if OPEN and NON-EMPTY, 
	 * adds the given points (2 control points and an endpoint) to the end of the curve, then
	 * fire a GEOMETRY_CHANGE event.<br>
	 * If the curve is closed,  you should use <code>splitSegment</code> instead, since this method does nothing 
	 * in this case ;-)
	 * @param ptCtrl1  first control point of the new Bezier segment
	 * @param ptCtrl2  second control point of the new Bezier segment
	 * @param ptEnd second end-point of the new Bezier segment
	 */
	public void curveTo(PicPoint ptCtrl1, PicPoint ptCtrl2, PicPoint ptEnd) {
		super.curveTo(ptCtrl1, ptCtrl2, ptEnd);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	// [developpers] lineTo doesn't need overriding where firing DrawingEvent is concerned, since it calls curveTo anyway.


	/**
	 * Adds a new point to a (maybe curved) segment then fires a changed-update.
	 * Control-points are computed automagically.
	 * @param seg_idx  index of the segment to be split
	 * @param pt point at which segment is to be split
	 */
	public int splitSegment(int seg_idx, PicPoint pt) {
		if (seg_idx < 0 || seg_idx >= getNumberOfSegments()) throw new IndexOutOfBoundsException(new Integer(seg_idx).toString());

		PicPoint ptL,ptR; // these will be the control points to insert together with pt.
		if (isStraight(seg_idx)) {
			ptL = ptR = pt; // means new segments are straight as well.
		}
		else { // compute ptL and ptR so as to retain tangent at split point
			// this is so far the best idea i've had...
			CubicCurve2D.Double cubic = new CubicCurve2D.Double();
			cubic.x1 = getSpecificationPoint(segmentToPointIndex(seg_idx,SUBDIVISION_POINT)).x;
			cubic.y1 = getSpecificationPoint(segmentToPointIndex(seg_idx,SUBDIVISION_POINT)).y;
			cubic.ctrlx1 = getSpecificationPoint(segmentToPointIndex(seg_idx,FIRST_SEGMENT_CONTROL_POINT)).x;
			cubic.ctrly1 = getSpecificationPoint(segmentToPointIndex(seg_idx,FIRST_SEGMENT_CONTROL_POINT)).y;
			cubic.ctrlx2 = getSpecificationPoint(segmentToPointIndex(seg_idx,SECOND_SEGMENT_CONTROL_POINT)).x;
			cubic.ctrly2 = getSpecificationPoint(segmentToPointIndex(seg_idx,SECOND_SEGMENT_CONTROL_POINT)).y;
			cubic.x2 = getSpecificationPoint(segmentToPointIndex(seg_idx+1,SUBDIVISION_POINT)).x;
			cubic.y2 = getSpecificationPoint(segmentToPointIndex(seg_idx+1,SUBDIVISION_POINT)).y;
			PicVector tangent = PEToolKit.computeTangentToPath(cubic, pt, Double.POSITIVE_INFINITY);
			// tangent is oriented in ascending order of the "t" bezier-spline parameter, i.e. in ascending order of point indices
			
			// now set new control-points along this tangent, at a some arbitrary distance (ok, and this is where i've
			// got to improve the deal somehow ;-( The only cool thing is that SMOOTHNESS & SYMMETRY are enforced
			// [SR:pending] add EditPointConstraint's as argument to this method (as in setPoint()) in order to adapt behaviour
			ptL = new PicPoint(pt);
			double dist1 = getSpecificationPoint(segmentToPointIndex(seg_idx,SUBDIVISION_POINT)).distance(getSpecificationPoint(segmentToPointIndex(seg_idx,FIRST_SEGMENT_CONTROL_POINT)));
			double dist2 = getSpecificationPoint(segmentToPointIndex(seg_idx+1,SUBDIVISION_POINT)).distance(getSpecificationPoint(segmentToPointIndex(seg_idx,SECOND_SEGMENT_CONTROL_POINT)));
			ptL.translate(tangent, -(dist1+dist2)/4.0); // best comprise is to take half the mean value ; feel free to improve if you can
			ptR = new PicPoint(pt);
			ptR.translate(tangent, (dist1+dist2)/4.0);
		}
		int idx = super.splitSegment(seg_idx, ptL, pt, ptR); // save index of new subdivision point
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		return idx;
	}


	/**
	 * Removes a subdivision point from this curve, together with its two neighbouring control points.
	 * This shifts any ensuing points to the "left", and creates a new segment by concatenating 
	 * dangling control-points.
	 * Nothing is done if this curve is reduced to a single point. <br>
	 * If both neighbouring segments
	 * had the same straightness value beforehands, the latter property
	 * is simply inherited by the new segment. Otherwise, we set it to false as this is the more flexible approach.<p>
	 * This method fires a DrawingEvent of type GEOMETRY_CHANGE.. 
	 * @param index index of the subdivision point to be removed with respect to the SUBDIVISION point numbering scheme, 
	 *        e.g. 0 for the first subdivision point (= first curve end-point).
	 */
	public void removeSubdivisionPoint(int subdivIndex) {
		if (getNumberOfSegments() <= 1) return; // we won't remove the only remaining segment
		super.removeSubdivisionPoint(subdivIndex);

		// [SR:pending] the following code snipet is probably useless now that straightness depends on control-point location...
		//           i'm leaving it until we make up our mind over it.
		//int pos = segmentToPointIndex(subdivIndex, SUBDIVISION_POINT);
		// pos-3 = index of subdivision-point of "new" segment
		//if (isValidBezierIndex(pos-3) && getPointType(pos-3)!=LAST_CURVE_END_POINT) 
		//	enforceSmoothnessOfStraightSegment(pointToSegmentIndex(pos-3)); // possibly move alternate controls to enforce smoothness
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * Removes the point with the given index from this curve. 
	 * If this is a subdivision point, we simply delegate to {@link #removeSubdivisionPoint removeSubdivisionPoint}.
	 * Else, this means removing a control point, and we just "straighten" the corresponding segment end-point 
	 *  by moving the control-point to the location of the nearest subdivision-point (hence this has no effect on
	 * straight segments).
	 * <br>
	 * This method fires a DrawingEvent of type GEOMETRY_CHANGE.
	 * @param index any valid control- or subdivision- index to be removed
	 */
	public void removePoint(int index) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "index="+index);
		if (getPointType(index)==INVALID_POINT_INDEX) throw new IndexOutOfBoundsException(new Integer(index).toString());
		if (getPointType(index)==SUBDIVISION_POINT || getPointType(index)==FIRST_CURVE_END_POINT || getPointType(index)==LAST_CURVE_END_POINT) 
			removeSubdivisionPoint(pointToSegmentIndex(index)); // already fires drawing event, don't need to do it twice ;-)
		else {
			getSpecificationPoint(index).setCoordinates(getSpecificationPoint(getNearestSubdivisionPoint(index)));
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}
	}
	

	////////////////////////////
	//// OTHER FIELDS ACCESSORS
	////////////////////////////


	/**
	 * close or open this curve, that is, remove or add an end-point.<br>
	 * This methods fires a DrawingEvent of type GEOMETRY_CHANGE .
	 * @param state  The new close value
	 */
	public void setClosed(boolean state) {
		super.setClosed(state);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Forces the smoothness feature of the given subdivision point by moving alternate control-points.<br>
	 * If one neightbouring segment is straight, it imposes the location of the alternate control.<br>
	 * If both segments are straight, this method can't do anything unfortunately.<br>
	 * If both segments are curved, the resulting tangent is computed from the external bissector of both control vectors,
	 * yet retaining norm of control-vectors.<br>
	 * This method fires a drawing-event of type GEOMETRY_CHANGE.
	 * @param subdivIndex index of the subdivision point, e.g. 0,1,2... for the 1st,2nd,3rd,... subdivision point.
	 */
	public void setSmooth(int subdivIndex) {
		if (isSmooth(subdivIndex)) return; // also checks for index validity ;-)
		
		if (getPointType(subdivIndex*3)!=SUBDIVISION_POINT) return; // N/A for closed curve only!

		int thisSeg = subdivIndex;
		int prevSeg = getPBCSegmentIndex(thisSeg-1);
		
		if (isStraight(thisSeg) && isStraight(prevSeg)) return;
		
		if (isStraight(thisSeg)){ // move alternate control "subdivIndex-1" so as to enforce smoothness at first end-point
			enforceSmoothnessOfStraightSegment(thisSeg, true, false);
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
			return;
		}
		
		if (isStraight(prevSeg)){ // move alternate control "subdivIndex+1" so as to enforce smoothness at second end-point of "prevSeg"
			enforceSmoothnessOfStraightSegment(prevSeg, false, true);
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
			return;
		}

		// else both segment are curved :
		//PicVector v = new PicVector(getSpecificationPoint(subdivIndex*3),getSpecificationPoint(subdivIndex*3-1));
		//PicVector w = new PicVector(getSpecificationPoint(subdivIndex*3),getSpecificationPoint(subdivIndex*3+1));
		PicVector v = getIncomingTangent(subdivIndex); // non-normalized
		PicVector w = getOutgoingTangent(subdivIndex);
		
		PicVector biss = new PicVector(v); // don't modify v directly, cause we'll use it later on
		biss.inverse(); // get rid of dangling PI in final rotation angle
		double angle = biss.angle(w); // positive if (-v) -> w CCW !
		biss.rotate(angle/2.0); // now biss is the external bissector of (subdiv,subdiv+1) and (subdiv, subdiv-1),  
		biss.normalize();       // it has a unit norm, and point "toward" w (ok, just sketch a drawing, and you'll get what i mean)
		// move both control point onto the line directed by biss, 
		// retaining the norm of both control vectors : (see also enforceSmoothOf..., same trick used there, by the way)
		// a) "previous" control-point :
		double length = v.norm();
		getSpecificationPoint(subdivIndex*3-1).setCoordinates(getSpecificationPoint(subdivIndex*3)).translate(biss, -length);
		// b) "next" control-point :
		length = w.norm();
		getSpecificationPoint(subdivIndex*3+1).setCoordinates(getSpecificationPoint(subdivIndex*3)).translate(biss, length);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * changes the smoothness feature of all the subdivision points at once.
	 * This method indirectly calls setSmooth(int), hence fires drawing-events of type GEOMETRY_CHANGE.
	 */
	public void setSmooth() {
		for (int i=0; i<getNumberOfSubdivisionPoints(); i++) setSmooth(i);
	}


	/**
	 * Force the SYMMETRY feature of the given subdivision point.
	 * This is straightforwardly carried out by setting the length of both control-vectors to their "mean value".
	 * This method does nothing if at least one neighbouring segment is straight.
	 * @param subdivIndex index of the subdivision point, e.g. 0,1,2... for the 1st,2nd,3rd,... subdivision point.
	 */
	public void setSymmetric(int subdivIndex) {
		if (isSymmetric(subdivIndex)) return;
		if (getPointType(subdivIndex*3)!=SUBDIVISION_POINT) return; // N/A for closed curve only!

		int thisSeg = subdivIndex;
		int prevSeg = getPBCSegmentIndex(thisSeg-1);
		
		if (isStraight(thisSeg) || isStraight(prevSeg)) return;
		
		// else both segment are curved :
		//PicVector v = new PicVector(getSpecificationPoint(subdivIndex*3),getSpecificationPoint(subdivIndex*3-1));
		//PicVector w = new PicVector(getSpecificationPoint(subdivIndex*3),getSpecificationPoint(subdivIndex*3+1));
		PicVector v = getIncomingTangent(subdivIndex); // non-normalized
		PicVector w = getOutgoingTangent(subdivIndex);
		double meanLength = (v.norm() + w.norm())/2.0;
		v.normalize();
		w.normalize();
		// a) "previous" control-point :
		getSpecificationPoint(subdivIndex*3-1).setCoordinates(getSpecificationPoint(subdivIndex*3)).translate(v, meanLength);
		// b) "next" control-point :
		getSpecificationPoint(subdivIndex*3+1).setCoordinates(getSpecificationPoint(subdivIndex*3)).translate(w, meanLength);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}


	/**
	 * force the symmetry feature of all the subdivision points at once.
	 * This method indirectly calls setSymmetric(int), hence fires drawing-events of type GEOMETRY_CHANGE.
	 */
	public void setSymmetric() {
		for (int i=0; i<getNumberOfSubdivisionPoints(); i++) setSymmetric(i);
	}

	/**
	 * changes the straightness feature for the given segment by making its control- and end-points identical.
	 * We then force smoothness on each end-point depending on the "old" smoothness value, that is, before straightening
	 * this segment.
	 */
	public void setStraight(int segIndex) {
		if (isStraight(segIndex)) return;
		int leftIdx = segmentToPointIndex(segIndex, SUBDIVISION_POINT);
		int ctrl1Idx = segmentToPointIndex(segIndex, FIRST_SEGMENT_CONTROL_POINT);
		int ctrl2Idx = segmentToPointIndex(segIndex, SECOND_SEGMENT_CONTROL_POINT);
		int rightIdx = segmentToPointIndex(segIndex+1, SUBDIVISION_POINT);
		// store old smoothness value at both end-points :
		boolean smoothAtFirstEndPoint = isSmooth(segIndex);
		boolean smoothAtSecondEndPoint = isSmooth(getPBCSegmentIndex(segIndex));
		setSpecificationPoint(ctrl1Idx, getSpecificationPoint(leftIdx));
		setSpecificationPoint(ctrl2Idx, getSpecificationPoint(rightIdx));
		enforceSmoothnessOfStraightSegment(segIndex, smoothAtFirstEndPoint, smoothAtSecondEndPoint);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Straighten ALL segments (ie make this curve a polygon).
	 * This method indirectly calls setStraight(int), hence fires drawing-events of type GEOMETRY_CHANGE.
	 */
	public void setStraight() {
		for (int i=0; i<getNumberOfSegments(); i++) setStraight(i);
	}

	/**
	 * Return a string for debugging purpose.
	 */
	public String toString() {
		String s = super.toString();
		s += "\n\tisSmooth=";
		for (int i = 0; i < getNumberOfSubdivisionPoints(); i++) s+= " " + isSmooth(i);
		s += "\n\tisSym=";
		for (int i = 0; i < getNumberOfSubdivisionPoints(); i++) s+= " " + isSymmetric(i);
		return s;
	}

	////////////////////////////////
	//// GUI
	////////////////////////////////

	/**
	 * Create an array of Action's related to this object
	 *
	 * @param actionDispatcher  dispatches events to the proper PECanvas
	 * @param localizer         i18n localizer for PEAction's
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {

		PEAction[] _actionArray=super.createActions(actionDispatcher, localizer, hi);
		if (getNumberOfSubdivisionPoints()==0) return _actionArray;
		// we add some actions to make all subdivision points smooth or symmetric
		// meaningless actions are disabled (they will show up in grey colour)
		boolean existNonSmooth=false;  // will be true if isSmooth(i)==false for some i
		boolean existNonSymmetric=false; // will be true if isSymmetric(i)==false for some i
		boolean existNonStraight=false; // will be true if isStraight(i)==false for some i
		for (int i=0;i< getNumberOfSubdivisionPoints();i++) {
			if (!isSmooth(i)) existNonSmooth=true; 
			if (!isSymmetric(i)) existNonSymmetric=true;
		}
		for (int i=0;i< getNumberOfSegments();i++) {
			if (!isStraight(i)) existNonStraight=true;
		}
		int nbActions= _actionArray.length + 3;
		if (hi instanceof HitInfo.Point) nbActions += 2;
		else if (hi instanceof HitInfo.Stroke && !(hi instanceof HitInfo.HighlighterStroke)) nbActions += 1;
		PEAction[] actionArray = new PEAction[nbActions];
		System.arraycopy(_actionArray,0,actionArray,0,_actionArray.length);
		int i=_actionArray.length;
		
		actionArray[i] = new SetAllSegmentsPropertyAction(actionDispatcher, SetAllSegmentsPropertyAction.SET_SMOOTH, localizer);
		actionArray[i].setEnabled(existNonSmooth);
		i++;
		
		actionArray[i] = new SetAllSegmentsPropertyAction(actionDispatcher, SetAllSegmentsPropertyAction.SET_SYMMETRIC, localizer);
		actionArray[i].setEnabled(existNonSymmetric);
		i++;
		
		actionArray[i] = new SetAllSegmentsPropertyAction(actionDispatcher, SetAllSegmentsPropertyAction.SET_STRAIGHT, localizer);
		actionArray[i].setEnabled(existNonStraight);
		i++;

		if (hi instanceof HitInfo.Point){
			HitInfo.Point hip = (HitInfo.Point)hi;
			int ptIndex=-1;// point index, with priority given to subdivision points
			for (int j=0; j<hip.getNbHitPoints(); j++){ // look-up first subdiv-point in the set of clicked points
				if (!isControlPoint(hip.getIndex(j))){  // ok, we got one !
						ptIndex = hip.getIndex(j);
						break;
				}
			}			
			if (ptIndex<0) ptIndex = hip.getIndex(); // if no subdiv point undex cursor
			//jpicedt.Log.debug("ptIndex="+ptIndex);
			// compute nearest subdivision point:
			int subdivIdx = getNearestSubdivisionPoint(ptIndex)/3; // nearest subdiv point
			//jpicedt.Log.debug("subdivIdx="+subdivIdx);
			actionArray[i] = new SetSegmentPropertyAction(actionDispatcher, SetSegmentPropertyAction.SET_SMOOTH, localizer, subdivIdx);
			actionArray[i].setEnabled(!this.isSmooth(subdivIdx));
			i++;
		
			actionArray[i] = new SetSegmentPropertyAction(actionDispatcher, SetSegmentPropertyAction.SET_SYMMETRIC, localizer, subdivIdx);
			actionArray[i].setEnabled(!this.isSymmetric(subdivIdx));
			i++;
		}
			
		else if (hi instanceof HitInfo.Stroke && !(hi instanceof HitInfo.HighlighterStroke)){
			HitInfo.Stroke his = (HitInfo.Stroke)hi;
			int segIdx = his.getClickedSegment();
			actionArray[i] = new SetSegmentPropertyAction(actionDispatcher, SetSegmentPropertyAction.SET_STRAIGHT, localizer, segIdx);
			actionArray[i].setEnabled(!this.isStraight(segIdx));
			i++;
		}

		return actionArray ;
	}



	/**
	 * Force a given segment's properties, eg symmetry, smoothness or straightness.
	 * @author    Sylvain Reynal
	 */
	class SetSegmentPropertyAction extends PEAction {
		final static String SET_SMOOTH = "action.editorkit.SetSmooth";
		final static String SET_SYMMETRIC = "action.editorkit.SetSymmetric";
		final static String SET_STRAIGHT = "action.editorkit.SetStraight";
		String type;
		int index; // depending on type, either a segment or a subdivision point index

		/**
		 * @param type SET_SMOOTH, SET_STRAIGHT or SET_SYMMETRIC
		 */
		public SetSegmentPropertyAction(ActionDispatcher actionDispatcher, String type, ActionLocalizer localizer, int index) {
			super(actionDispatcher, type, localizer);
			this.type = type;
			this.index = index;
		}

		public void undoableActionPerformed(ActionEvent e) {
				if (type== SET_SYMMETRIC) setSymmetric(index);
				else if (type == SET_SMOOTH) setSmooth(index);
				else if (type == SET_STRAIGHT) setStraight(index);
		}
	}

	/**
	 * Force ALL segments' properties, eg symmetry, smoothness or straightness.
	 * @author    Vincent Guirardel
	 */
	class SetAllSegmentsPropertyAction extends PEAction {
		final static String SET_SMOOTH = "action.editorkit.SetSmoothAll";
		final static String SET_SYMMETRIC = "action.editorkit.SetSymmetricAll";
		final static String SET_STRAIGHT = "action.editorkit.SetStraightAll";
		String type;

		/**
		 * @param type SET_SMOOTH, SET_STRAIGHT or SET_SYMMETRIC
		 */
		public SetAllSegmentsPropertyAction(ActionDispatcher actionDispatcher, String type, ActionLocalizer localizer) {
			super(actionDispatcher, type, localizer);
			this.type = type;
		}

		public void undoableActionPerformed(ActionEvent e) {
				if (type== SET_SYMMETRIC) setSymmetric();
				else if (type == SET_SMOOTH) setSmooth();
				else if (type == SET_STRAIGHT) setStraight();
		}
	}

	/**
	 * Returns a Customizer for geometry editing
	 */
	public AbstractCustomizer createCustomizer() {
		return new Customizer();
	}


	/**
	 * geometry customizer
	 */
	class Customizer extends AbstractCurve.Customizer {

		private DefaultCellEditor doubleEditor;
		private boolean isListenersAdded = false;// flag
		private MultiCurveJTable table;


		/**
		 * Constructor for the Customizer object
		 */
		public Customizer() {

			Box p = new Box(BoxLayout.Y_AXIS);
			// polygon points:
			table = new MultiCurveJTable();
			table.setPreferredScrollableViewportSize(new Dimension(200, 300));

			// wrap scrollpane around:
			JScrollPane scrollPane = new JScrollPane(table);
			p.add(scrollPane);
			add(p, BorderLayout.NORTH);
			add(createPanel(), BorderLayout.CENTER);
		}

		/**
		 * update Element's properties
		 */
		public void store() {
			super.store();
			table.repaint();
		}

		// public void actionPerformed(ActionEvent e) => inherited

		/**
		 * @return   the panel title, used e.g. for Border or Tabpane title.
		 */
		public String getTitle() {
			return PicMultiCurve.this.getName();
		}
	}

	///////////////////////////////////////////////////
	//// INNER CLASSES for Customizer
	///////////////////////////////////////////////////

	/**
	 * an implementation of JTable for PicMultiCurve's. This makes use of the underlying model MultiCurveTableModel.
	 * @author guirardel
	 */
	class MultiCurveJTable extends JTable {

		private final CoordinateCellEditor coordCellEditor = new CoordinateCellEditor(new DecimalNumberField(0, 5));
		private final MultiCurveTableModel model;
		TableCellRenderer invisibleRenderer = new InvisibleRenderer();
		TableCellRenderer numberRenderer = new NumberRenderer();
		TableCellRenderer booleanRenderer = new BooleanRenderer();

		/**
		 * Constructor for the MultiCurveJTable object
		 */
		MultiCurveJTable() {
			super();
			setModel(model = new MultiCurveTableModel());
		}


		/**
		 * Returns an appropriate editor for the cell specified by row and column.
		 */
		public TableCellEditor getCellEditor(int row, int col) {
			if ((col > 0) && (col < 3))
				return coordCellEditor; // i.e. DecimalNumberField (=enhanced JTextField)
			else
				return super.getCellEditor(row, col); // JTable's native cell editor
		}


		// MultiCurveJTable inner class
		class InvisibleRenderer extends JLabel  implements TableCellRenderer {

			public InvisibleRenderer(){
				super();//empty JLabel
			}
			public Component getTableCellRendererComponent(JTable table, Object color,boolean isSelected, boolean hasFocus,int row, int column) {
				JLabel label=new JLabel();
				if ((row % 3) !=0) label.setBackground(Color.lightGray);//???ca marche pas, je sais pas pourquoi...
				else label.setBackground(Color.white);
				return label;// returns colored empty JLabel
			}
		}

		// MultiCurveJTable inner class
		class BooleanRenderer extends Component implements TableCellRenderer {

			public BooleanRenderer(){
				super();// empty component
			}
			public Component getTableCellRendererComponent(JTable table, Object color,boolean isSelected, boolean hasFocus,int row, int column) {
				JCheckBox comp=new JCheckBox();
				comp.setSelected(((Boolean)getValueAt(row,column)).booleanValue());
				if ((column ==0) || model.isCellEditable(row,column)){
					comp.setForeground(Color.black);
				}
				else {
					comp.setForeground(Color.lightGray);
				}
				if ((row % 3) ==0) comp.setBackground(Color.lightGray);
				else comp.setBackground(Color.white);

				return comp;
			}
		}

		// MultiCurveJTable inner class"
		class NumberRenderer extends  DefaultTableCellRenderer implements TableCellRenderer {

			public NumberRenderer(){
				super();// default renderer
			}
			public Component getTableCellRendererComponent(JTable table, Object color,boolean isSelected, boolean hasFocus,int row, int column) {
				Component comp=super.getTableCellRendererComponent(table,color,isSelected,hasFocus,row,column);
				if ((column ==0) || model.isCellEditable(row,column)){
					comp.setForeground(Color.black);
				}
				else {
					comp.setForeground(Color.lightGray);
				}
				if ((row % 3) ==0) comp.setBackground(Color.lightGray);
				else comp.setBackground(Color.white);

				return comp;
			}
		}

		public TableCellRenderer getCellRenderer(int row, int col) {
			// 6 columns : "pt", "X", "Y", smooth, symmetric, straight
			if (col == 5) { // straight
				if (model.getValueAt(row,col)==null) return invisibleRenderer; // if not a control-point
				else return booleanRenderer;
			}
			else if (col >= 3) {
				if (model.getValueAt(row,col)==null)  return invisibleRenderer;
				else return booleanRenderer;
			}
			//else
			return numberRenderer;
		}
	}


	/**
	 * MultiCurveJTable's coordinate cell editor is based on a DecimalNumberField
	 * @author    reynal
	 */
	class CoordinateCellEditor extends DefaultCellEditor {
		private DecimalNumberField dnf;

		/**
		 * construct a new cell editor from the given DecimalNumberField
		 */
		CoordinateCellEditor(DecimalNumberField dnf) {
			super(dnf);
			this.dnf = dnf;
		}

		/**
		 * Returns the value contained in the editor - overriden from
		 * DefaultCellEditor so as to return a Double instead of a String.
		 */
		public Object getCellEditorValue() {
			return new Double(dnf.getValue());
		}
	}

	/**
	 * a class that specifies the methods the JTable will use to interrogate a
	 * tabular data model.
	 * @author    reynal
	 */
	class MultiCurveTableModel extends AbstractTableModel {

		/**
		 * Returns the name of the column at columnIndex.
		 */
		public String getColumnName(int colIndex) {
			switch (colIndex) {
			case 0:
				return jpicedt.Localizer.currentLocalizer().get("misc.Point");// polygon point index
			case 1:
				return "X";// x-coordinate
			case 2:
				return "Y";// y-coordinate
			case 3:
				return jpicedt.Localizer.currentLocalizer().get("attributes.Smooth");
			case 4:
				return jpicedt.Localizer.currentLocalizer().get("attributes.Symmetric");
			case 5:
				return jpicedt.Localizer.currentLocalizer().get("attributes.Straight");
			default:
				return "";
			}
		}

		/**
		 * Returns the number of columns in the model.
		 */
		public int getColumnCount() {
			return 6;// 6 columns : "pt", "X", "Y", smooth, symmetric, straight
		}

		/**
		 * Returns the number of rows in the model.
		 */
		public int getRowCount() {
			return getNumberOfSpecificationPoints();// there are as many rows as points in the associated PicMultiCurve
		}

		/**
		 * Returns true if the cell at rowIndex and columnIndex is editable.
		 * @param rowIndex this is equivalent to a bezier point index
		 */
		public boolean isCellEditable(int rowIndex, int colIndex) {
			if (colIndex == 0)
				return false;// first column is not editable since it containts point indexes
			else if (colIndex <= 2) // "X" and "Y"
				//return ( (rowIndex % 3 == 0)|| !isStraight(rowIndex /3) ); //can't edit control points when straight
				return true;
			else if (colIndex <= 4) // "smooth" "symmetric"
				return (getPointType(rowIndex) == SUBDIVISION_POINT); // smoothness and symmetry make sense only for subdivision points
			else
				return (getPointType(rowIndex) == FIRST_SEGMENT_CONTROL_POINT ); // "straightness" button will show up
			                               // beneath a control-point, so that it's clearly identified with 
										   // a segment (putting it in front of a subdiv-point would certainly not
										   // make the deal.
		}

		/**
		 * Called when the associated JTable wants to know what to display at
		 * columnIndex and rowIndex. Depending on colIndex, we return, either the
		 * point index, the x-coord, or the y-coord of the corresponding curve point, and when applicable,
		 * smoothness, symmetry and straighness properties.
		 */
		public Object getValueAt(int rowIndex, int colIndex) {

			switch (colIndex) {
			case 0:
				return new Integer(rowIndex);// polygon point index
			case 1:
				return new Double(PEToolKit.doubleToString(getSpecificationPointX(rowIndex)));// x-coord for the row-th point
			case 2:
				return new Double(PEToolKit.doubleToString(getSpecificationPointY(rowIndex)));// y-coord for the row-th point
			case 3:
				if (getPointType(rowIndex)==SUBDIVISION_POINT)
					return new Boolean(isSmooth(rowIndex/3));// smoothness of the subdivision point
				else
					return null;
			case 4:
				if (getPointType(rowIndex)==SUBDIVISION_POINT)
					return new Boolean(isSymmetric(rowIndex/3));// symmetry of the subdivision point
				else
					return null;
			case 5:
				if (getPointType(rowIndex)==FIRST_SEGMENT_CONTROL_POINT)
					return new Boolean(isStraight((rowIndex-1)/3));// straightness of the segment
				else
					return null;
			default:
				return null;
			}
		}

		/**
		 * Returns the most specific superclass for all the cell values in the column. 
		 * This is used by the JTable to set up a default renderer and editor for the column.
		 */
		public Class getColumnClass(int c) {
			switch (c) {
			case 0: return Integer.class;
			case 1: return Double.class;
			case 2: return Double.class;
			case 3: return Boolean.class;
			case 4: return Boolean.class;
			case 5: return Boolean.class;
			default: return null;
			}
		}


		/**
		 * Invoked by the UI (aka event-handler) when a user entered a new value in the cell at columnIndex and
		 * rowIndex. We update PicMultiCurve geometry according to the cell
		 * value.<br>
		 * This method does nearly the same thing as actionPerformed in other
		 * xxxxCustomizer's.
		 *
		 * @param value     entered value
		 * @param rowIndex  equivalent to a bezier-point index
		 */
		public void setValueAt(Object value, int rowIndex, int colIndex) {

			switch (colIndex) {// do we modify x or y or what ?
			case 1:
				PicPoint pt = getPoint(rowIndex, null);
				pt.x = ((Double)value).doubleValue();
				setPoint(rowIndex, pt, SMOOTHNESS.and(SYMMETRY)); // [SR:pending] change to AUTO as soon as it's supported by setPoint
				break;
			case 2:
				pt = getPoint(rowIndex, null);
				pt.y = ((Double)value).doubleValue();
				setPoint(rowIndex, pt, SMOOTHNESS.and(SYMMETRY)); // [SR:pending] change to AUTO as soon as it's supported by setPoint
				break;
			case 3:
				if (rowIndex % 3==0)
					setSmooth(rowIndex/3);// smoothness of the subdivision point
				else
					return ;
				break;
			case 4:
				if (rowIndex % 3==0)
					setSymmetric(rowIndex/3);// symmetry of the subdivision point
				else
					return ;
				break;
			case 5:
				if (rowIndex % 3==1)
					setStraight((rowIndex-1)/3);// straightness of the segment

				else
					return ;
				break;
			default:
				return;
			}
			fireTableDataChanged();
		}
	}// inner class
}

