/*
 PicParallelogram.java - March 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.model;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.*;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.view.View;
import jpicedt.graphic.view.HitInfo;
import jpicedt.widgets.*;

/**
 * A class implementing a parallelogram. Geometry is fully specified by three corner points 
 * (namely, the bottom-left, bottom-right and top-right corners), which thus
 * act as specification-points (see DefaultLeafElement for details), and can be retrieved using
 * <code>getSpecificationPointX/Y</code>. These are obviously stored in the <code>pts</code>
 * ArrayList, as defined in DefaultLeafElement.<p>
 * [SR:TODO] add rounded-corners feature.
 * @author Vincent Guirardel.
 * @since jPicEdt 1.4
 * @version $Id: PicParallelogram.java,v 1.28.2.1 2007/09/02 11:56:27 reynal Exp $
 */
public class PicParallelogram extends DefaultLeafElement implements CustomizerFactory, ActionFactory,PicMultiCurveConvertable {

	////////////////////////////
	//// PUBLIC CONSTANT FIELDS
	////////////////////////////

	// [pending:SR] there's just one pb with the following naming strategy : BR and TR may as well be upside down
	//              after a call to setPoint() (see similar remark in PicGroup). Either advise developper "publicly" here,
	//              or change names to, e.g. X1Y1, ... as in (now deprecated) PicRectangle.
	/** Bottom-Left corner (also first specification point) */
	public static final int P_BL = 0;
	/** Bottom-Right corner (also second specification point) */
	public static final int P_BR = 1;
	/** Top-Right corner  (also third specification point) */
	public static final int P_TR = 2;
	/** Top-Left corner */
	public static final int P_TL = 3;
	/** Bottom side mid-point */
	public static final int SIDE_B = 4;
	/** Right side mid-point */
	public static final int SIDE_R = 5;
	/** Top side mid-point */
	public static final int SIDE_T = 6;
	/** Left side mid-pint */
	public static final int SIDE_L = 7;
	/** Parallelogram's centre */
	public static final int P_CENTER = 8;
	/** Index of first control-point */
	public static final int FIRST_PT = 0;
	/** Index of last control-point */
	public static final int LAST_PT = 8;

	/** predefined constraint for setPoint()  */
	public static final BasicEditPointConstraint SQUARE = new BasicEditPointConstraint("square");

	/** predefined constraint for setPoint() */
	public static final BasicEditPointConstraint DEFAULT = new BasicEditPointConstraint("default");

	/** predefined constraint for setPoint()  */
	public static final BasicEditPointConstraint CENTER_FIXED = new BasicEditPointConstraint("center-fixed");

	// [SR:pending] use EditPointConstraint instead
	/** 1st specification point : Bottom-Left corner */
	public static final int P_SPEC_BL = -1;
	/** 2nd specification point : Bottom-Right corner */
	public static final int P_SPEC_BR = -2;
	/** 3rd specification point : Top-Right corner */
	public static final int P_SPEC_TR = -3; 

	/**
	 * "l2rVec" is a vector pointing from P_BL to P_BR (or equally, from P_TL to P_TR)
	 * "b2tVec" is a vector pointing from P_BR to P_TR (or equally, from P_BL to P_TL)
	 * These two vectors simply define a convenient parallelogram's basis, and
	 * are updated by a call to updateParalleloBasis().
	 * They are used e.g. when moving a corner-point.<p>
	 * An essential function of these vectors is to "retain" the initial parallelogram basis even
	 * if the element becomes flat at some point during an edit operation.
	 */
	protected PicVector l2rVec = new PicVector(1,0);
	/** see documentation for {@link #l2rVec l2rVec} */
	protected PicVector b2tVec = new PicVector(0,1);

	//////////////////////////////
	/// PRIVATE VARIABLES
	//////////////////////////////

	private PicVector ptBuffer1 = new PicVector(); // convenience buffer for get/setPoint methods (being a PicVector, this can obsiously serve as a PicPoint as well ;-)	
	private PicVector ptBuffer2 = new PicVector(); 
	//////////////////////////////
	/// CONSTRUCTORS
	//////////////////////////////

	/**
	 * Creates a new PicParallelogram reduced to (0,0)
	 * @since jpicedt 1.3.3
	 */
	public PicParallelogram(){
		super(3);
		l2rVec.setCoordinates(1,0);
		b2tVec.setCoordinates(0,1);
	}

	/**
	 * Creates a new PicParallelogram reduced to (0,0), with the given attribute set
	 * @since jpicedt 1.3.3
	 */
	public PicParallelogram(PicAttributeSet set){
		super(3,set);
		l2rVec.setCoordinates(1,0);
		b2tVec.setCoordinates(0,1);
	}

	/**
	 * Create a new PicParallelogram object using the 3 given points as 3 consecutive points of the parallelogram, 
	 * and a default attribute set.
	 * @since jpicedt 1.4
	 */
	public PicParallelogram(PicPoint pt1, PicPoint pt2, PicPoint pt3){
		super(3);
		setSpecificationPoint(0,pt1);
		setSpecificationPoint(1,pt2);
		setSpecificationPoint(2,pt3);
		updateParalleloBasis();
	}

	/**
	 * Create a new PicParallelogram object using the 3 given points as 3 consecutive points of the parallelogram. 
	 * @since jpicedt 1.3.3
	 */
	public PicParallelogram(PicPoint pt1, PicPoint pt2, PicPoint pt3, PicAttributeSet set){
		super(3,set);
		setSpecificationPoint(0,pt1);
		setSpecificationPoint(1,pt2);
		setSpecificationPoint(2,pt3);
		updateParalleloBasis();
	}

	/**
	 * Create a new rectangle with axes parallel to X- and Y-axes, using the 2 given points to build the diagonal, 
	 * and a default attribute set.
	 * @since jpicedt 1.4
	 */
	public PicParallelogram(PicPoint pt1, PicPoint pt3){
		// [SR:pending] this method needs to be tested
		this(pt1, new PicPoint(pt3.x, pt1.y), pt3);
	}

	/**
	 * Create a new PicParallelogram reduced to a point located at the given position
	 * @since jpicedt 1.3.3
	 */
	public PicParallelogram(PicPoint pt, PicAttributeSet set) {
		this(pt,pt,pt,set);
		l2rVec.setCoordinates(1,0); // security
		b2tVec.setCoordinates(0,1);
	}

	/**
	 * "cloning" constructor (to be used by clone())
	 * @param The PicParallelogram object to clone
	 * @since jpicedt 1.3.3
	 */
	public PicParallelogram(PicParallelogram parallelo){
		super(parallelo);
		updateParalleloBasis();
	}

	/**
	 * Overrides Object.clone() method
	 * @since jpicedt 1.3.3
	 */
	public Object clone(){
		return new PicParallelogram(this);
	}

	/**
	 * @return a localised string that represents this object's name
	 * @since jpicedt 1.3.3
	 */
	public String getName(){
		return jpicedt.Localizer.currentLocalizer().get("model.Parallelogram");
	}

	//////////////////////////////////
	/// OPERATIONS ON CONTROL POINTS
	//////////////////////////////////

	/**
	 * Return the X-coord of the control-point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since jpicedt 1.3.3
	 */
	 // [SR:pending] move to superclass
	public double getPointX(int numPoint) {
		return getPoint(numPoint, ptBuffer1).x;
	}

	/**
	 * Return the Y-coord of the control-point indexed by <code>numPoint</code>. 
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 * @since jpicedt 1.3.3
	 */
	 // [SR:pending] move to superclass
	public double getPointY(int numPoint) {
		return getPoint(numPoint, ptBuffer1).y;
	}

	/**
	 * Return the <b>user-controlled point</b> having the given index. The general contract in Element
	 * is to return an IMMUTABLE instance of PicPoint, so that the only way to alter the geometry of this element is
	 * by calling the <code>setPoint</code> method.<br>
	 * @return the point indexed by <code>numPoint</code> ; 
	 *         if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @param numPoint the point index, should be greater or equal to the value returned by
	 *        <code>getFirstPointIndex</code>, and lower or equal to <code>getLastPointIndex</code>.
	 */
	public PicPoint getPoint(int numPoint, PicPoint src){
		// [developpers] warning, using "ptBuffer1/2" is strongly discouraged here, as this would interfere with
		//               "setPoint", whic also uses it ; if for any reason you really need to use a PicPoint buffer here, 
		//               just create a new one (e.g. "ptBuffer3") as a private class field.
		if (src==null) src = new PicPoint();
		
		//if (numPoint < 0) return getPoint(-numPoint -1,src); 
		if (numPoint <= 2)  // spec-points
			return getSpecificationPoint(numPoint,src); //P_BL,P_BR, or P_TR correspond to real indexes in ptsY
		if (numPoint == P_TL) 
			return getSpecificationPoint(P_BL,src).translate(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR));
		if (numPoint == SIDE_B)
			return getSpecificationPoint(P_BR,src).middle(getSpecificationPoint(P_BL)); 
		if (numPoint == SIDE_R) 
			return getSpecificationPoint(P_BR,src).middle(getSpecificationPoint(P_TR)); 
		if (numPoint == SIDE_T) 
			return getPoint(SIDE_B,src).translate(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR));
		if (numPoint == SIDE_L) 
			return getPoint(SIDE_R,src).translate(getSpecificationPoint(P_BR),getSpecificationPoint(P_BL));
		if (numPoint==P_CENTER) 
			return getSpecificationPoint(P_BL,src).middle(getSpecificationPoint(P_TR));
		throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}
	
	/**
	 * Return the index of the first point that can be retrieved by getPoint
	 */
	public int getFirstPointIndex(){
		return FIRST_PT;
	}

	/**
	 * Return the index of the last point that can be retrieved by getPoint
	 */
	public int getLastPointIndex(){
		return LAST_PT;
	}

	/**
	 * Compute the "l2r" and "b2t" vectors of the parallelogram basis from the location of the three
	 * specification-points of this parallelogram.
	 * @since jpicedt 1.3.3
	 */
	protected void updateParalleloBasis(){ // was "computeVectors"
		l2rVec.setCoordinates(getSpecificationPoint(P_BL),getSpecificationPoint(P_BR)); // left-to-right
		b2tVec.setCoordinates(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR)); // bottom-to-top
		
		// check if l2r and b2t are parallels or null (in which case our parallelogram is flat either way),
		// or if parallelogram is reduced to a single-point :
		if (l2rVec.det(b2tVec) == 0){
			if (!l2rVec.isNull()){ // if l2r is a non-null vector, use it to generate basis
				b2tVec.setCoordinates(l2rVec);
				b2tVec.rotate(Math.PI/2);
			}
			else if (!b2tVec.isNull()){ // else this is just the other way around ;-)
				l2rVec.setCoordinates(b2tVec);
				l2rVec.rotate(-Math.PI/2);
			}
			else { // our parallelogram is reduced to a single point => make it a rectangle parallel to cartesian axes next time setPoint() is called
				l2rVec.setCoordinates(1,0);
				b2tVec.setCoordinates(0,1);
			}
			if (l2rVec.det(b2tVec) == 0){  //should almost never happen (merely in the event there is a very small but non-zero vector...)
				l2rVec.setCoordinates(1,0);
				b2tVec.setCoordinates(0,1);
			}
		}
		
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("l2rVec="+l2rVec+", b2tVec="+b2tVec);
	}

	/**
	 * Compute the contravariant coordinate of the given point "ptSrc" along the "l2r" and "b2t" vectors of
	 * the parallelogram's basis, the origin of the basis being set at "ptOrg"
	 * This simply reduces to a contravariant change of coordinates from the cartesian basis to the 
	 * (non-orthonormalized) parallelogram basis.<br>
	 * To wind up : denoting OM = vec(ptOrg -> ptSrc), (e1,e2) the cartesian basis (ie std model-coordinates),
	 * and (x',y') the new coords in the (l2r,b2t) basis, yields : <br>
	 * <b>OM</b> = x <b>e1</b> + y <b>e2</b> = x' <b>l2r</b> + y' <b>b2t</b>.
	 * <br>
	 * Change of coordinates is carried out by using elementary geometry. This yields :<br>
	 * <ul>
	 * <li> x' = (OM x b2t).e3/||l2r x b2t||
	 * <li> y' = -(OM x l2r).e3/||l2r x b2t||
	 * </ul>
	 * where e3 denotes the z-axis unit vector of the (x,y,z) direct diedra.
	 * <p>
	 * Note that {@link #updateParalleloBasis updateParalleloBasis} must be called beforehands for this method to return
	 * a valid results [SR:pending] need to be cleaned up.
	 * @param ptOrg origin of parallelogram's basis
	 * @param ptSrc source point, coordinates of which have to be converted 	 
	 * @param ptDest if null, gets allocated and returned for convenience ; besides, it is perfectly safe to call
	 *        this method with ptDest and one of ptOrg and ptSrc referencing the same PicPoint.
	 * @return the new coordinates (x',y') along (l2r,b2t), encapsulated in a PicVector 
	 * @since jpicedt 1.3.3
	 */
	protected PicVector toParalleloBasisCoordinates(PicPoint ptOrg, PicPoint ptSrc, PicVector ptDest){
		if (ptDest==null) ptDest = new PicVector();
		double det = l2rVec.det(b2tVec); // updateParalleloBasis guarantees det != 0 
		ptDest.setCoordinates(ptOrg, ptSrc); // OM vector
		double x = ptDest.det(b2tVec)/det;
		double y = -ptDest.det(l2rVec)/det;
		ptDest.setCoordinates(x, y);
		if (jpicedt.Log.DEBUG) 
			jpicedt.Log.debug( "l2rVec="+l2rVec+", b2tVec="+b2tVec+"\n\tl2rCoord="+ptDest.x+", b2tCoord="+ptDest.y);
		return ptDest;
	}


	/**
	 * Set the coordinate of the control-point indexed by "numPoint" to the given value.
	 * The control-policy for the default (null or DEFAULT) constraint is as follows :
	 * <ul>
	 * <li> when moving a corner, the opposite CORNER is kept fixed (ie segments experience a parallel transport) 
	 * <li> when moving a SIDE_X point, the opposite SEGMENT is kept fixed, so that this operation allows one to change the 
	 *      parallelogram skewness.
	 * </ul>
	 * <p>
	 * The SQUARE constraint imposes that the parallelogram be a square. The square diagonal is computed from the current
	 * control-point and its opposite corner. Other spec' points are set accordingly.
	 * <p>
	 * The CENTER_FIXED constraint imposes that the parallelogram center be kept fixed. Spec' points are set accordingly.
	 * <p>
	 * @param numPoint one of P_TL, P_TR, P_BL, P_BR, SIDE_T, SIDE_B, SIDE_L, SIDE_R or P_CENTER.
	 * @param constraint either null, DEFAULT, CENTER_FIXED or SQUARE
	 * @since jpicedt 1.3.3
	 */
	public void setPoint(int numPoint, PicPoint pt, EditPointConstraint constraint){
		if (jpicedt.Log.DEBUG) {
			jpicedt.Log.debug("numPoint="+numPoint+",pt="+pt);
			jpicedt.Log.debugAppendLn("l2rVec="+l2rVec+", b2tVec="+b2tVec);
		}
		if (constraint==null || constraint.imposes(DEFAULT)) setPointDefaultConstraint(numPoint,pt);
		else if (constraint.imposes(SQUARE)) setPointSquareConstraint(numPoint, pt);
		else if (constraint.imposes(CENTER_FIXED)) setPointCenterFixedConstraint(numPoint, pt);
		else setPointDefaultConstraint(numPoint,pt);
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
		
	/**
 	 * Helper for setPoint() invoked with the DEFAULT constraint
	 * Remark: when moving SIDE_XX points, vectors defining the parallelogram basis change their direction, so that it's
	 *         mandatory to call updateParalleloBasis() afterwards.
	 */
	private void setPointDefaultConstraint(int numPoint, PicPoint pt){
		//if (numPoint <0) setSpecificationPoint(-numPoint-1,pt); return; 
		//double tX,tY; //translation ...
		// the following code keeps the center fixed, while moving the corner
		//  		if (numPoint==P_BL) {
		//  		    ptsX[P_BL]=pt.x;ptsY[P_BL]=pt.y;
		//  		    ptsX[P_TR]=2*cX-pt.x;ptsY[P_TR]=2*cY-pt.y;
		//  		}
		//  		else if (numPoint==P_BR){
		//  		    ptsX[P_BR]=pt.x;ptsY[P_BR]=pt.y;
		//  		}
		//  		else if (numPoint==P_TR){
		//  		    ptsX[P_TR]=pt.x;ptsY[P_TR]=pt.y;
		//  		    ptsX[P_BL]=2*cX-pt.x;ptsY[P_BL]=2*cY-pt.y;
		//  		}
		//  		else if (numPoint==P_TL){
		//  		    ptsX[P_BR]=2*cX-pt.x;ptsY[P_BR]=2*cY-pt.y;
		//  		}
		
		// Reminder : P_BL, P_BR and P_TR are the three specification points
		if (numPoint==P_BL) { // also moves BR so that TR is kept fixed
			setSpecificationPoint(P_BL, pt);
			getSpecificationPoint(P_TR, ptBuffer1).translate(b2tVec); // rebuild BR->TR line in case these points were identical
			setSpecificationPoint(P_BR, pt);
			getSpecificationPoint(P_BR).project(getSpecificationPoint(P_TR),ptBuffer1,l2rVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==P_TR){ // also moves BR so that BL is kept fixed
			setSpecificationPoint(P_TR, pt);
			getSpecificationPoint(P_BL, ptBuffer1).translate(l2rVec); // rebuild BL->BR line in case these points were identical 
			setSpecificationPoint(P_BR, pt);
			getSpecificationPoint(P_BR).project(getSpecificationPoint(P_BL),ptBuffer1,b2tVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==P_BR){ // also moves BL and TR so that TL is kept fixed
			// warning : P_TL it's NOT a specification point, hence moving specification points may "move" it as well ; 
			getPoint(P_TL, ptBuffer1);   // => it vital to save its value beforehands
			getPoint(P_TL, ptBuffer2);
			setSpecificationPoint(P_BR, pt); // side-effect : this (indirectly) "moves" P_TL hence the need to save P_TL beforehands
			ptBuffer2.translate(b2tVec); // rebuild TL->BL line
			getSpecificationPoint(P_BL).setCoordinates(pt).project(ptBuffer1,ptBuffer2,l2rVec); // (p1,p2,proj-axis)
			ptBuffer2.setCoordinates(ptBuffer1).translate(l2rVec); // rebuild TL->TR line
			getSpecificationPoint(P_TR).setCoordinates(pt).project(ptBuffer1,ptBuffer2,b2tVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==P_TL){ // also moves BL and TR so that BR is kept fixed
			getSpecificationPoint(P_BR, ptBuffer1).translate(l2rVec); // rebuild BR->BL line
			getSpecificationPoint(P_BL).setCoordinates(pt).project(getSpecificationPoint(P_BR),ptBuffer1,b2tVec); // (p1,p2,proj-axis)
			getSpecificationPoint(P_BR, ptBuffer1).translate(b2tVec); // rebuild BR->TR line
			getSpecificationPoint(P_TR).setCoordinates(pt).project(getSpecificationPoint(P_BR),ptBuffer1,l2rVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==SIDE_L){ // keep BR and TR fixed, translate BL
			getSpecificationPoint(P_BL).translate(getPoint(SIDE_L,ptBuffer1),pt);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_B){ // keep TR and TL(*) fixed, translate BL and BR (*:TL is not a spec-point)
			getPoint(SIDE_B,ptBuffer1); // save old SIDE_B in ptBuffer1
			getSpecificationPoint(P_BL).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BR).translate(ptBuffer1,pt);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_R){ // keep BL and TL(*) fixed, translate TR and BR
			getPoint(SIDE_R,ptBuffer1); // save old SIDE_R in ptBuffer1
			getSpecificationPoint(P_TR).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BR).translate(ptBuffer1,pt);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_T){ // keep BL and BR fixed, translate TR
			getSpecificationPoint(P_TR).translate(getPoint(SIDE_T,ptBuffer1),pt);
			updateParalleloBasis();
		}
		else if (numPoint==P_CENTER) { // translate BL, BR and TR
			getPoint(P_CENTER,ptBuffer1); // save old CENTER in ptBuffer1
			getSpecificationPoint(P_TR).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BL).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BR).translate(ptBuffer1,pt);
		}
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}

	/**
 	 * Helper for setPoint() invoked with the CENTER_FIXED constraint.
	 * Policy : center is kept fixed, spec' points are updated accordingly, otherwise this works exactly as with
	 * the DEFAULT constraint.
	 */
	private void setPointCenterFixedConstraint(int numPoint, PicPoint pt){
		// Reminder : P_BL, P_BR and P_TR are the three specification points
		getPoint(P_CENTER, ptBuffer1); // save center into p1
		if (numPoint==P_BL) {
			setSpecificationPoint(P_BL, pt); // BL=pt
			setSpecificationPoint(P_TR, pt); // TR = sym. of BL wrt CENTER
			getSpecificationPoint(P_TR).symmetry(ptBuffer1);
			getSpecificationPoint(P_TR, ptBuffer2).translate(b2tVec); // rebuild BR->TR line in case these points were identical
			setSpecificationPoint(P_BR, pt);
			getSpecificationPoint(P_BR).project(getSpecificationPoint(P_TR),ptBuffer2,l2rVec); // and project (p1,p2,proj-axis)
		}
		else if (numPoint==P_TR){ 
			setSpecificationPoint(P_TR, pt); // TR=pt
			setSpecificationPoint(P_BL, pt); // BL = sym. of TR wrt CENTER
			getSpecificationPoint(P_BL).symmetry(ptBuffer1); // ptBuffer1=center
			getSpecificationPoint(P_BL, ptBuffer2).translate(l2rVec); // rebuild BL->BR line in case these points were identical 
			setSpecificationPoint(P_BR, pt);
			getSpecificationPoint(P_BR).project(getSpecificationPoint(P_BL),ptBuffer2,b2tVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==P_BR){ 
			setSpecificationPoint(P_BR, pt); // BR=pt
			getSpecificationPoint(P_BR, ptBuffer2);
			ptBuffer2.symmetry(ptBuffer1); // p2 = TL (yet TL is NOT a spec' point)
			ptBuffer1.setCoordinates(ptBuffer2).translate(l2rVec); // rebuild TL -> TR line for the projection (p1 as CENTER no longer used)
			setSpecificationPoint(P_TR, pt);
			getSpecificationPoint(P_TR).project(ptBuffer1,ptBuffer2,b2tVec); // (p1,p2,proj-axis)
			ptBuffer1.setCoordinates(ptBuffer2).translate(b2tVec); // rebuild BL -> TL line for the projection
			setSpecificationPoint(P_BL, pt);
			getSpecificationPoint(P_BL).project(ptBuffer1,ptBuffer2,l2rVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==P_TL){ 
			setSpecificationPoint(P_BR, pt); // BR=sym of pt wrt center
			getSpecificationPoint(P_BR).symmetry(ptBuffer1);
			ptBuffer2.setCoordinates(pt).translate(l2rVec); // rebuild TL -> TR line for the projection
			setSpecificationPoint(P_TR, getSpecificationPoint(P_BR));
			getSpecificationPoint(P_TR).project(pt,ptBuffer2,b2tVec); // (p1,p2,proj-axis)
			ptBuffer2.setCoordinates(pt).translate(b2tVec); // rebuild BL -> TL line for the projection
			setSpecificationPoint(P_BL, getSpecificationPoint(P_BR));
			getSpecificationPoint(P_BL).project(pt,ptBuffer2,l2rVec); // (p1,p2,proj-axis)
		}
		else if (numPoint==SIDE_L){ 
			ptBuffer2.setCoordinates(getPoint(SIDE_L,ptBuffer2),pt); // translation vector SIDE_L -> pt
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			getSpecificationPoint(P_BR).translate(ptBuffer2.inverse());
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_B){ 
			ptBuffer2.setCoordinates(getPoint(SIDE_B,ptBuffer2),pt); // translation vector SIDE_B -> pt
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			getSpecificationPoint(P_BR).translate(ptBuffer2);
			getSpecificationPoint(P_TR).translate(ptBuffer2.inverse());
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_R){ 
			ptBuffer2.setCoordinates(getPoint(SIDE_R,ptBuffer2),pt); // translation vector SIDE_R -> pt
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			getSpecificationPoint(P_BR).translate(ptBuffer2);
			getSpecificationPoint(P_BL).translate(ptBuffer2.inverse());
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_T){ 
			ptBuffer2.setCoordinates(getPoint(SIDE_T,ptBuffer2),pt); // translation vector SIDE_T -> pt
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			getSpecificationPoint(P_BL).translate(ptBuffer2.inverse());
			getSpecificationPoint(P_BR).translate(ptBuffer2);
			updateParalleloBasis();
		}
		else if (numPoint==P_CENTER) { // does nothing (center fixed !)
		}
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}
	
	/**
 	 * Helper for setPoint() invoked with the SQUARE constraint.
	 * Policy : opposite corner is kept fixed. Third specification point is computed on-the-fly so that
	 * the parallelogram becomes or stays square.
	 */
	private void setPointSquareConstraint(int numPoint, PicPoint pt){
		// Reminder : P_BL, P_BR and P_TR are the three specification points
		
		if (numPoint==P_BL) { // leave P_TR fixed, update P_BR
			setSpecificationPoint(P_BL, pt);
			// compute P_BR = P_BL + p1 + p2, where p1+p2 = square's side properly oriented
			ptBuffer1.setCoordinates(getSpecificationPoint(P_BL),getSpecificationPoint(P_TR));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			setSpecificationPoint(P_BR, getSpecificationPoint(P_BL));
			getSpecificationPoint(P_BR).translate(ptBuffer1.add(ptBuffer2)); // BR = BL + p1 + p2, where p1+p2 = square's side
			updateParalleloBasis();
		}
		
		else if (numPoint==P_TR){ // leave P_BL fixed, update P_BR
			setSpecificationPoint(P_TR, pt);
			// compute P_BR = P_BL + p1 + p2, where p1+p2 = square's side
			ptBuffer1.setCoordinates(getSpecificationPoint(P_BL),getSpecificationPoint(P_TR));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			setSpecificationPoint(P_BR, getSpecificationPoint(P_BL));
			getSpecificationPoint(P_BR).translate(ptBuffer1.add(ptBuffer2)); 
			updateParalleloBasis();
		}
		else if (numPoint==P_BR){ // leave P_TL fixed, update P_TR and P_BL
			// save P_TL into p1:
			getPoint(P_TL,ptBuffer1);
			// compute P_BL = P_BR + p1 + p2, where p1+p2 = square's side
			// and P_TR = P_BR + p1 - p2
			ptBuffer1.setCoordinates(pt,ptBuffer1); // i.e. p1 = (P_BR -> P_TL)
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			setSpecificationPoint(P_BL, pt); // pt = P_BR
			getSpecificationPoint(P_BL).translate(ptBuffer1);
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			setSpecificationPoint(P_TR, pt); // pt = P_BR
			getSpecificationPoint(P_TR).translate(ptBuffer1);
			getSpecificationPoint(P_TR).translate(ptBuffer2.inverse());
			// move P_BR to pt:
			setSpecificationPoint(P_BR, pt);
			updateParalleloBasis();
		}
		else if (numPoint==P_TL){ // leave P_BR fixed, update P_TR and P_BL
			// compute P_BL = P_BR + p1 + p2, where p1+p2 = square's side
			// and P_TR = P_BR + p1 - p2
			ptBuffer1.setCoordinates(getSpecificationPoint(P_BR),pt); // p1 = P_BR -> P_TL
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			setSpecificationPoint(P_BL, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_BL).translate(ptBuffer1);
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			setSpecificationPoint(P_TR, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_TR).translate(ptBuffer1);
			getSpecificationPoint(P_TR).translate(ptBuffer2.inverse());
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_L){ // leave SIDE_R fixed, update all spec points
			ptBuffer1.setCoordinates(pt,getPoint(SIDE_R,ptBuffer1));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(-Math.PI/2);
			// set BL
			setSpecificationPoint(P_BL, pt); 
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			// set BR
			setSpecificationPoint(P_BR, getSpecificationPoint(P_BL)); 
			getSpecificationPoint(P_BR).translate(ptBuffer1);
			getSpecificationPoint(P_BR).translate(ptBuffer1);
			// set TR
			setSpecificationPoint(P_TR, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_TR).translate(ptBuffer2.inverse());
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_B){ // leave SIDE_T fixed, update all spec points
			ptBuffer1.setCoordinates(pt,getPoint(SIDE_T,ptBuffer1));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			// set BL
			setSpecificationPoint(P_BL, pt); 
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			// set BR
			setSpecificationPoint(P_BR, pt); 
			getSpecificationPoint(P_BR).translate(ptBuffer2.inverse());
			// set TR
			setSpecificationPoint(P_TR, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_TR).translate(ptBuffer1);
			getSpecificationPoint(P_TR).translate(ptBuffer1);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_R){ // leave SIDE_L fixed, update all spec points
			ptBuffer1.setCoordinates(pt,getPoint(SIDE_L,ptBuffer1));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(-Math.PI/2);
			// set TR
			setSpecificationPoint(P_TR, pt); 
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			// set BR
			setSpecificationPoint(P_BR, pt); 
			getSpecificationPoint(P_BR).translate(ptBuffer2.inverse());
			// set BL
			setSpecificationPoint(P_BL, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_BL).translate(ptBuffer1);
			getSpecificationPoint(P_BL).translate(ptBuffer1);
			updateParalleloBasis();
		}
		else if (numPoint==SIDE_T){ // leave SIDE_B fixed, update all spec points
			ptBuffer1.setCoordinates(pt,getPoint(SIDE_B,ptBuffer1));
			ptBuffer1.scale(1/2.);
			ptBuffer2.setCoordinates(ptBuffer1);
			ptBuffer2.rotate(Math.PI/2);
			// set TR
			setSpecificationPoint(P_TR, pt); 
			getSpecificationPoint(P_TR).translate(ptBuffer2);
			// set BR
			setSpecificationPoint(P_BR, getSpecificationPoint(P_TR)); 
			getSpecificationPoint(P_BR).translate(ptBuffer1);
			getSpecificationPoint(P_BR).translate(ptBuffer1);
			// set BL
			setSpecificationPoint(P_BL, getSpecificationPoint(P_BR)); 
			getSpecificationPoint(P_BL).translate(ptBuffer2.inverse());
			getSpecificationPoint(P_BL).translate(ptBuffer2);
			updateParalleloBasis();
		}
		else if (numPoint==P_CENTER) { // translate BL, BR and TR
			getPoint(P_CENTER,ptBuffer1); // save old CENTER in ptBuffer1
			getSpecificationPoint(P_TR).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BL).translate(ptBuffer1,pt);
			getSpecificationPoint(P_BR).translate(ptBuffer1,pt);
		}
		else throw new IndexOutOfBoundsException(new Integer(numPoint).toString());
	}

	/**
	 * Return the X-coord of the parallelogram's center
	 * @since jpicedt 1.3.3
	 */
	public double getCenterX(){
		return getCenter(ptBuffer1).x;
	}

	/**
	 * Return the Y-coord of the parallelogram's center
	 * @since jpicedt 1.3.3
	 */
	public double getCenterY(){
		return getCenter(ptBuffer1).y;
	}

	/**
	 * Return the parallelogram's centre
	 * @return if <code>src</code> is null, allocates a new PicPoint and return it, 
	 *         otherwise directly modifies <code>src</code> and returns it as well for convenience.
	 * @since jpicedt 1.3.3
	 */
	public PicPoint getCenter(PicPoint src){
		if (src==null) src=new PicPoint();
		src.setCoordinates(getSpecificationPoint(P_BL));
		src.middle(getSpecificationPoint(P_TR));
		return src;
	}

	/**
	 * Rotate this Element by the given angle along the given point
	 * @param angle rotation angle in radians
	 */
	public void rotate(PicPoint ptOrg, double angle){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).rotate(ptOrg,angle);
		updateParalleloBasis();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}
	
	/**
	 * Shear this Element by the given params wrt to the given origin
	 */
	public void shear(PicPoint ptOrg, double shx, double shy){
		for (Iterator it = pts.iterator(); it.hasNext();)
			((PicPoint)it.next()).shear(ptOrg,shx,shy);
		updateParalleloBasis();
		fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
	}

	/**
	 * Return a polygon created from the sides of this parallelogram.
	 *
	 * @since jpicedt 1.3.3
	 */
	public PicMultiCurve convertToMultiCurve(){

		PicMultiCurve curve = new PicMultiCurve(true, this.attributeSet); // closed
		curve.addPoint(getPoint(P_BL,null));
		PicPoint pt = getPoint(P_BR,null);
		curve.splitSegment(0, pt, pt, pt);
		pt = getPoint(P_TR,null);
		curve.splitSegment(1, pt, pt, pt);
		pt = getPoint(P_TL,null);
		curve.splitSegment(2, pt, pt, pt);
		return curve;
	}

	/**
	 * Returns the bounding box (ie the surrounding rectangle) in double precision
	 * Used for instance to determine the arguments of a \\begin{picture} command.<p>
	 * This implementation compute the bounding-box from the smallest rectangle that encompasses 
	 * all the specification-points.
	 * @since jpicedt 1.3.3
	 */
	public Rectangle2D getBoundingBox(Rectangle2D r){
		if (r==null) r = new Rectangle2D.Double();
		r.setFrameFromDiagonal(getSpecificationPoint(P_BL),getSpecificationPoint(P_TR));
		r.add(getSpecificationPoint(P_BR));
		r.add(getPoint(P_TL, ptBuffer1));
		return r;
	}
	
	/**
	 * Returns true if this parallelogram is a rectangle
	 */
	public boolean isRectangle(){
		PicVector p1 = new PicVector(getSpecificationPoint(P_BR),getSpecificationPoint(P_BL));
		PicVector p2 = new PicVector(getSpecificationPoint(P_BR),getSpecificationPoint(P_TR));
		return (p1.dot(p2)==0);
	}

	/**
	 * Implementation of the Object.toString() method, used for debugging purpose
	 */
	public String toString() {

		String s = super.toString()
		           + "\n\t l2rVec=" + l2rVec
		           + ", b2tVec=" + b2tVec;

		return s;
	}
	
	/////////////////////////////////
	//// Action's
	/////////////////////////////////
	
	/**
	 * Create an array of Action's related to this object
	 *
	 * @param actionDispatcher  dispatches events to the proper PECanvas
	 * @param localizer         i18n localizer for PEAction's
	 */
	public PEAction[] createActions(ActionDispatcher actionDispatcher, ActionLocalizer localizer, HitInfo hi) {
		PEAction[] actionArray = new PEAction[1];
		actionArray[0] = new ConvertToCurveAction(actionDispatcher, localizer);
		return actionArray;
	}

	// ---- parallelo to multiCurve ----
	
	/**
	 * Convert this parallelogram to a multicurve, selecting it if applicable.
	 * @author    Sylvain Reynal
	 * @since     jpicedt 1.4
	 */
	class ConvertToCurveAction extends PEAction {

		public static final String KEY = "action.editorkit.ConvertParallelogramToMulticurve";
		
		public ConvertToCurveAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void undoableActionPerformed(ActionEvent e){
			PicMultiCurve curve = convertToMultiCurve();
			Drawing dr = getDrawing();
			if (dr != null){
				dr.replaceElement(PicParallelogram.this, curve);
				View view = curve.getView();
				if (view != null){
					PECanvas canvas = view.getContainer();
					if (canvas != null)   canvas.select(curve, true);
				}
			}
		}
	}
	
	////////////////////////////////
	//// GUI
	////////////////////////////////


	/**
	 * @return a Customizer for geometry editing
	 * @since jpicedt 1.4
	 */
	public AbstractCustomizer createCustomizer(){
		return new Customizer();
	}

	/**
	 * Geometry customizer
	 * @since jpicedt 1.4
	 */
	class Customizer extends AbstractCustomizer implements ActionListener {

		private DecimalNumberField paraBLCornerXTF,paraBLCornerYTF, paraBRCornerXTF,paraBRCornerYTF,paraTRCornerXTF,paraTRCornerYTF;
		private boolean isListenersAdded = false; // flag

		public Customizer(){

			super();
			JPanel p = new JPanel(new GridLayout(4,3,5,5));

			p.add(PEToolKit.createJLabel("action.draw.Parallelogram"));
			p.add(new JLabel("x"));
			p.add(new JLabel("y"));
			// line 1: bottom left corner label
			p.add(new JLabel("1"));
			// x-coordinate of bl-corner
			p.add(paraBLCornerXTF = new DecimalNumberField(4));
			// y-coordinate of bl-corner
			p.add(paraBLCornerYTF = new DecimalNumberField(4));


			// line 2: bottom right corner label
			p.add(new JLabel("2"));
			// x-coordinate of br-corner
			p.add(paraBRCornerXTF = new DecimalNumberField(4));
			// y-coordinate of br-corner
			p.add(paraBRCornerYTF = new DecimalNumberField(4));

			// line 3: top right corner label
			p.add(new JLabel("3"));
			// x-coordinate of TR-corner
			p.add(paraTRCornerXTF = new DecimalNumberField(4));
			// y-coordinate of TR-corner
			p.add(paraTRCornerYTF = new DecimalNumberField(4));

			add(p, BorderLayout.NORTH);
		}

		/** add action listeners to widgets to reflect changes immediately */
		private void addActionListeners(){
			if (isListenersAdded) return; // already done
			paraBLCornerXTF.addActionListener(this);
			paraBLCornerYTF.addActionListener(this);
			paraBRCornerXTF.addActionListener(this);
			paraBRCornerYTF.addActionListener(this);
			paraTRCornerXTF.addActionListener(this);
			paraTRCornerYTF.addActionListener(this);
			isListenersAdded = true;
		}

		/**
		 * (re)init widgets with Element's properties
		 */
		public void load(){
			paraBLCornerXTF.setValue(getPointX(PicParallelogram.P_BL));
			paraBLCornerYTF.setValue(getPointY(PicParallelogram.P_BL));
			paraBRCornerXTF.setValue(getPointX(PicParallelogram.P_BR));
			paraBRCornerYTF.setValue(getPointY(PicParallelogram.P_BR));
			paraTRCornerXTF.setValue(getPointX(PicParallelogram.P_TR));
			paraTRCornerYTF.setValue(getPointY(PicParallelogram.P_TR));
			// add listeners AFTERWARDS ! otherwise loading widgets initial value has a painful side-effet...
			// since it call "store" before everything has been loaded
			addActionListeners(); // done the first time load is called
		}

		/**
		 * update Element's properties
		 */
		public void store(){
			PicPoint bl= new PicPoint(paraBLCornerXTF.getValue(),paraBLCornerYTF.getValue());
			PicPoint br= new PicPoint(paraBRCornerXTF.getValue(),paraBRCornerYTF.getValue());
			PicPoint tr= new PicPoint(paraTRCornerXTF.getValue(),paraTRCornerYTF.getValue());
			setSpecificationPoint(P_BL,bl);
			setSpecificationPoint(P_BR,br);
			setSpecificationPoint(P_TR,tr);
			updateParalleloBasis();
			fireChangedUpdate(DrawingEvent.EventType.GEOMETRY_CHANGE);
		}

		public void actionPerformed(ActionEvent e){
			store();
		}

		public String getTitle(){
			return PicParallelogram.this.getName();
		}

	}
} // PicParallelogram
