/*
 AbstractCustomizer.java - December 31, 2001 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import javax.swing.Icon;
import javax.swing.JPanel;
import java.awt.BorderLayout;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
 * An abstract superclass that alleviates the burden of creating a customizer panel, 
 * by providing a common set of behaviours where storing/retrieving value 
 * to/from a Properties object (or any storage class) is concerned. In particular,
 * the hosting panel should provide UI commands to load default values, reload previously stored values,
 * and store value to the approriate stream (choice of which is being left to concrete implementation
 * of related abstract methods).
 * <p>
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @see jpicedt.graphic.toolkit.CustomizerDialog
 * @version $Id: AbstractCustomizer.java,v 1.6.2.1 2007/09/02 11:56:27 reynal Exp $
 */
public abstract class AbstractCustomizer extends JPanel implements ComponentListener {

	/**
	 * Construct a new preferences-panel with BorderLayout as the default layout.
	 */
	public AbstractCustomizer(){
		super(new BorderLayout(5,5));
		addComponentListener(this);
	}
	
	
	/**
	 * Loads widgets display content with a default value, 
	 * presumably from a "default preferences" file or a dedicated storage class. 
	 */
	public void loadDefault(){}
	
	/**
	 * Loads widgets display content, presumably from a "preferences" file or a dedicated storage class
	 */
	abstract public void load();
	
	/**
	 * Stores current widgets value, presumably to a file or to a dedicated storage class
	 */
	abstract public void store();
	
	/**
	 * Returns the panel title, used e.g. for Border or Tabpane title.<br>Default implementation returns this class name. 
	 */
	public String getTitle(){ 
		return getClass().getName();	
	}
	
	/**
	 * Returns the Icon associated with this panel, used e.g. for TabbedPane decoration ; default implementation returns null. 
	 */
	public Icon getIcon(){ 
		return null;	
	}
	
	/**
	 * Returns the tooltip string associated with this panel ; default implementation returns an empty string. 
	 */
	public String getTooltip(){ 
		return null;
	}
	
	/** 
	 * called when this component gets activated ; may be used to request focus 
	 * this implementation does nothing 
	 */
	protected void activated(){
	}
	
	/** public as an implementation side effect */
	public void componentHidden(ComponentEvent e){}

	/** public as an implementation side effect */
	public void componentMoved(ComponentEvent e){}

	/** public as an implementation side effect */
	public void componentResized(ComponentEvent e){}

	/** calls "activated" */
	public void componentShown(ComponentEvent e){
		activated();
	}
	
} // class 

