/*
 EditorKit.java - January 2, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.PECanvas;
import jpicedt.graphic.SelectionHandler;
import jpicedt.graphic.ContentType;
import jpicedt.graphic.DefaultContentType;

import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.BranchElement;
import jpicedt.graphic.model.PicAttributeSet;
import jpicedt.graphic.model.PicCircleFrom3Points;
import jpicedt.graphic.model.PicEllipse;
import jpicedt.graphic.model.PicGroup;
import jpicedt.graphic.model.PicMultiCurve;
import jpicedt.graphic.model.PicParallelogram;
import jpicedt.graphic.model.PicSmoothPolygon;
import jpicedt.graphic.model.TextEditable;

import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.event.DrawingListener;
import jpicedt.graphic.event.HelpMessageListener;
import jpicedt.graphic.event.HelpMessageEvent;

import jpicedt.graphic.view.View;
import jpicedt.graphic.view.AbstractView;
import jpicedt.graphic.view.ViewFactory;
import jpicedt.graphic.view.ViewConstants;
import jpicedt.graphic.view.Highlighter;
import jpicedt.graphic.view.HighlighterFactory;
import jpicedt.graphic.view.DefaultHighlighterFactory;
import jpicedt.graphic.view.DefaultViewFactory;
import jpicedt.graphic.view.HitInfo;

import jpicedt.graphic.io.formatter.FormatterFactory;

import javax.swing.event.EventListenerList;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.KeyStroke;

import java.awt.Toolkit;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.geom.Rectangle2D;

import java.io.IOException;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.UnsupportedFlavorException;

/**
 * A pluggable EditorKit for a PECanvas, that mainly provides mouse-tool management,
 * e.g. zooming, selecting and drawing capabilities, together with a set of Action's.
 * <br>
 * Action's that make sense for this EditorKit are implemented as static inner classes to reduce
 * name space. They can be shared across multiple PECanvas through the use of an ActionDispatcher
 * given to the Action constructor.
 * <br>New mouseTool's can be registered via the registerMouseTool method ; they're backed by a hashMap,
 * and can be activated (ie set as the current mouse-tool) by using setCurrentMouseTool with the tool name as
 * the parameter string.
 * <p>
 * This EditorKit provides also a ViewFactory, which delegates most of its behaviour to an underlying
 * factory (except for the View associated with the RootElement of the Drawing), and 
 * whose type depends on the current content-type, as set by setContentType. This allows the
 * "root-view" to maintain a reference to the container (which this EditorKit can easily provide), 
 * while still allowing underlying ViewFactory's to be really independant of the exact structure
 * of the Drawing (especially the existence of a RootElement). Besides, the EditorKit can plug
 * a new ViewFactory on-the-fly when the content-type changed, w/o other objects having to be informed.
 * <p>
 * [todo] 
 * <ul>
 * <li> provide an EditorKit abstract class + a minimal implementation for developpers who don't need
 *        all the editing stuff (e.g. setting attributes, undoing,...) when just implementing a viewer
 * </ul>
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: EditorKit.java,v 1.41.2.1 2007/09/02 11:56:28 reynal Exp $ 
 * <p> 
 *
 */
public class EditorKit {

	////////////////////////////////////////////////////////////////////////////////////
	//// PUBLIC FIELDS
	////////////////////////////////////////////////////////////////////////////////////

	/** zoom mode */
	public static final String ZOOM = "action.editorkit.Zoom";
	/** selection */
	public static final String SELECT = "action.editorkit.Select";
	/** point edition tool */
	public static final String EDIT_POINT = "action.editorkit.EditBezierPoint";	
	/** move */
	public static final String MOVE = "action.editorkit.Translate";
	/** scale */
	public static final String SCALE = "action.editorkit.Scale";
	/** rotate */
	public static final String ROTATE = "action.editorkit.Rotate";
	/** shear */
	public static final String SHEAR = "action.editorkit.Shear";

	/** use by getAvailableToolNames ; add your own names here */
	private static final String[] ALL_EDIT_TOOL_NAMES=new String[]{ZOOM, SELECT, EDIT_POINT, MOVE, SCALE, ROTATE};
	
	/** property name for edit mode change : "edit-mode" */
	public static final String EDIT_MODE_CHANGE = "edit-mode";

	////////////////////////////////////////////////////////////////////////////////////
	//// PROTECTED FIELDS
	////////////////////////////////////////////////////////////////////////////////////

	/** the PECanvas this EditorKit is installed in. null if not installed */
	protected PECanvas board;
	/** list of HelpMessageListener's and PropertyChangeListener's for this EditorKit */
	protected EventListenerList listenerList = new EventListenerList();

	////////////////////////////////////////////////////////////////////////////////////
	//// PRIVATE FIELDS
	////////////////////////////////////////////////////////////////////////////////////
	/** a string that stored the current edit mode */
	private String editMode;
	/** the PropertyChangeListener that gets notified of property changes in the canvas */
	private PropertyChangeHandler propertyChangeHandler = new PropertyChangeHandler();
	/** the DrawingListener that gets notified of drawing changes in the drawing */
	// [pending] not used anymore ! private DrawingHandler drawingHandler = new DrawingHandler();
	/** the factory that creates popup-menu on-the-fly */
	private PopupMenuFactory popupMenuFactory;
	/** selection handler for this kit */
	//private EditorKit.SelectionHandler selectionHandler;
	private DefaultSelectionHandler selectionHandler;
	/** the current installed mousetool */
	private MouseTool currentMouseTool;
	/** a map that stores registered mousetools */
	private HashMap toolMap = new HashMap();
	/** the attribute set used when instanciating a new graphic element */
	private PicAttributeSet inputAttributes;
	/** the view factory used to produce view for the drawing */
	private ViewFactory viewFactory;
	/** the FormatterFactory used to produce formatter that can write the drawing content to a writer */
	private FormatterFactory formatterFactory;
	/** the factory that produces View's Highlighters */
	private HighlighterFactory highlighterFactory;

	////////////////////////////////////////////////////////////////////////////////////
	//// PRIVATE STATIC FIELDS
	////////////////////////////////////////////////////////////////////////////////////
	// to be shared across every instance of EditorKit :
	private static Clipboard clipboard = new Clipboard("jPicEdt local clipboard");
	
	////////////////////////////////////////////////////////////////////////////////////
	//// CONSTRUCTORS
	////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Construct a new EditorKit with SELECT being the default initial edit-mode, and a default viewfactory.
	 */
	public EditorKit(){
		this(null,null);
	}

	/**
	 * Construct a new EditorKit with SELECT being the default initial edit-mode.
	 * @param contentType if non-null, this set the view- and formatter- factories to use for rendering the associated content ; 
	 *        otherwise, default factories are created using <code>createDefaultContentType()</code>.
	 * @param oldKit if an old instance already existed, and one wishes to reuse its properties, e.g.
	 *        listener-list, popupmenu-factory, input-attributes,... except obviously for the ViewFactory
	 *        and the FormatterFactory given as arguments. May be safely be set to null.
	 */
	public EditorKit(ContentType contentType, EditorKit oldKit){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("view factory delegate="+viewFactory);
		
		setFactoriesFromContentType(contentType);
		
		if (oldKit==null) this.highlighterFactory = new DefaultHighlighterFactory();
		else this.highlighterFactory = oldKit.highlighterFactory;
		
		this.selectionHandler = new DefaultSelectionHandler(this);
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("installing selection handler:"+this.selectionHandler);
		
		if (oldKit==null) this.inputAttributes = new PicAttributeSet(); // default
		else this.inputAttributes = oldKit.inputAttributes;
		
		if (oldKit != null) this.listenerList = oldKit.listenerList;
		if (oldKit != null) setPopupMenuFactory(oldKit.getPopupMenuFactory());
		
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("init completed !");
	}



	////////////////////////////////////////////////////////////////////////////
	//// PLUGGABLE BEHAVIOUR
	////////////////////////////////////////////////////////////////////////////

	/**
	 * Called when the kit is being installed into a PECanvas. 
	 * This registers default mouse-tools, ie :
	 * <ul>
	 * <li> tools corresponding to constant-fields SELECT, MOVE,... defined in this class ;
	 * <li> all tools available from DrawToolFactory, and which may be accessed using constant-fields defined therein,
	 *      e.g. DrawToolFactory.RECTANGLE to access the mousetool dedicated to drawing rectangles.
	 * </ul>
	 * Then registers events-listener from the attached canvas,
	 * and finally set SELECT as the default mouse-tool.
	 * @param board the PECanvas
	 */
	public void install(PECanvas board) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Kit is being installed in "+board);
		this.board = board;
		if (jpicedt.Log.DEBUG) jpicedt.Log.debugAppendLn("Registering default mouse-tools");
		registerMouseTool(ZOOM,new ZoomTool(this));
		registerMouseTool(SELECT,new SelectionTool(this, new EditElementMouseTransformFactory(this))); // (un)select, move elements or their end-points
		registerMouseTool(EDIT_POINT,new SelectionTool(this, new EditPointMouseTransformFactory(this))); // add/remove/edit control-points features
		registerMouseTool(MOVE,new MoveTool(this,selectionHandler));
		registerMouseTool(SCALE,new ScaleTool(this,selectionHandler));
		registerMouseTool(ROTATE,new RotateTool(this,selectionHandler));
		//registerMouseTool(SHEAR,new ShearTool(this,selectionHandler));
		
		DrawToolFactory dtf = new DrawToolFactory(this);
		String[][] allDrawToolNames = dtf.getAvailableToolNames();
		for (int i=0; i<allDrawToolNames.length; i++){
			for (int j=0; j<allDrawToolNames[i].length; j++){ // scan each group of related tools, e.g. DrawArcXXXX
				String toolName = allDrawToolNames[i][j];
				MouseTool mt = dtf.createDrawTool(toolName);
				registerMouseTool(toolName,mt);
			}
		}
		
		setCurrentMouseTool(SELECT);
		board.addPropertyChangeListener(propertyChangeHandler); // respond to DRAWING_CHANGE events
		// install keybd bindings :
		board.setActionMap(createActionMap(new DefaultActionDispatcher(board),jpicedt.Localizer.currentLocalizer().getActionLocalizer()));
		board.setInputMap(JComponent.WHEN_FOCUSED,createInputMap());
	}

	/**
	 * Called when the kit is being removed from a PECanvas
	 * This is used to unregister any previously registered mouse-tool.
	 * @param board the PECanvas
	 */
	public void deinstall(PECanvas board) {
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Kit is being deinstalled from "+board);
		board.removePEMouseInputListener(currentMouseTool);
		board.removeKeyListener(currentMouseTool);
		board.removePropertyChangeListener(propertyChangeHandler);
		// [pending] unregister mouse tools ?
		String[] registered = getRegisteredMouseToolsByName();
		for (int i=0; i<registered.length; i++){
			unRegisterMouseTool(registered[i]);
		}
		this.board = null;
	}

	/**
	 * Return the PECanvas into which this kit is installed, or null if the kit has not been installed yet.
	 */
	public PECanvas getCanvas(){
		return board;
	}
	
	////////////////////////////////
	/// MODEL & ATTRIBUTES
	////////////////////////////////

	/**
	 * plugs new FormatterFactory and ViewFactory into this EditorKit, using the given ContentType to create them on-the-fly.
	 * @param ct if null, a default content-type is used, as returned by <code>createDefaultContentType</code>
	 */
	public void setFactoriesFromContentType(ContentType ct){
		if (ct==null) ct = createDefaultContentType();
		this.viewFactory = new ViewFactoryWrapper(ct.createViewFactory());
		this.formatterFactory = ct.createFormatter();
	}

	/**
	 * Creates a default content-type suited for this editor 
	 */
	public ContentType createDefaultContentType(){
		return new DefaultContentType();
	}
	
	/**
	 * get the input attributes for the PECanvas attached to this EditorKit
	 */
	public PicAttributeSet getInputAttributes(){
		return inputAttributes;
	}

	/**
	 * set the input attributes for the PECanvas attached to this EditorKit
	 */
	public void setInputAttributes(PicAttributeSet set){
		inputAttributes = set;
	}

	/**
	 * @return the FormatterFactory used to write the content of a Drawing to a stream, according
	 *         to the type of content this kit can edit
	 */
	public FormatterFactory getFormatterFactory(){
		return formatterFactory;
	}


	///////////////////////////////////////////
	/// MOUSE TOOL MANAGEMENT
	//////////////////////////////////////////

	/**
	 * Returns all mousetool-names registered by default with this EditorKit.
	 * @return an array of arrays of String, each sub-array containing a group of tightly related tools  
	 */
	public static String[][] getAvailableToolNames(){
		String[][] allDrawToolNames = DrawToolFactory.getAvailableToolNames();
		String[][] allToolNames = new String[allDrawToolNames.length + 1][]; // one group for editing tools + as many groups as necessary for drawing tools
		allToolNames[0] = ALL_EDIT_TOOL_NAMES;
		for (int i=1; i<allToolNames.length; i++){
			allToolNames[i] = allDrawToolNames[i-1];
		}
		return allToolNames;
	}
		

	/**
	 * Registers the given mouse-tool by associating it with the given name in the hashtable of
	 * available mouse-tools. This name can then be fed as a parameter to 
	 * {@link #setCurrentMouseTool setCurrentMouseTool()}.
	 */
	public void registerMouseTool(String name, MouseTool mt){
		if (mt==null || name==null) return; // allows the underlying factory to return null objects, if any.
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Registering mouse tool " + mt + " with name " + name);
		toolMap.put(name,mt);
	}

	/**
	 * Unregisters the mouse-tool having the given name from the table of available mouse-tools.
	 */
	public void unRegisterMouseTool(String name){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Unregistering mouse tool " + name);
		toolMap.remove(name);
	}

	/**
	 * @return an array containing the currently registered mouse-tools names
	 */
	public String[] getRegisteredMouseToolsByName(){
		Set keySet = toolMap.keySet();
		return (String[]) keySet.toArray(new String[0]);
	}

	/**
	 * @return an array containing the currently registered mouse-tools
	 */
	public MouseTool[] getRegisteredMouseTools(){
		Collection values = toolMap.values();
		return (MouseTool[]) values.toArray(new MouseTool[0]);
	}

	/**
	 * set the current mouse tool for this editor kit
	 * <p>
	 * When setting a new edit-mode, the editor kit fire a PropertyChange Event with key
	 * "EDIT_MODE_CHANGE" and the given edit-mode as the new value.
	 * <br>
	 * This may be used by the UI to get informed that a change occured (since some
	 * mouse-tools in this kit may change the edit-mode from the inside).
	 * @param editMode a previously registered mouseTool's name
	 */
	public void setCurrentMouseTool(String editMode) {

		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Changing edit mode to " + editMode);
		if (board == null) return; // not installed
		String oldMode = this.editMode;
		// [pending] commented because startTextEditor doesn't currently set a new edit-mode
		// but instead stores the "old" edit mode when its starts, and tries to restore it
		// when it stop (hence, nothing would happend here because the old edit mode and the
		// current one would be the same.
		if (this.editMode == editMode) return; // no change
		this.editMode = editMode;
		// [pending] currentMouseTool.deinstall() ? or stop() or...
		MouseTool newMouseTool = (MouseTool)toolMap.get(this.editMode);
		if (newMouseTool ==null) return; // [pending] throws NotRegisteredMouseToolException
		board.removePEMouseInputListener(currentMouseTool);
		board.removeKeyListener(currentMouseTool);
		if (currentMouseTool!=null) currentMouseTool.flush();
		currentMouseTool=newMouseTool;
		currentMouseTool.init(); // give it a chance to do some init work before a mouse-event occurs
		board.addPEMouseInputListener(currentMouseTool);
		board.addKeyListener(currentMouseTool);
		//board.setCursor(currentMouseTool.getCursor());
		// !!! board.unSelectAll(); // for draw_xxx only
		firePropertyChange(EDIT_MODE_CHANGE, oldMode, this.editMode);
		board.repaint(); // clean-up after mousetool
	}


	/**
	 * @return the current edit mode for this EditorKit
	 */
	public String getCurrentMouseTool() {
		return editMode;
	}

	/**
	 * Start a TextEditor for the given TextEditable, using the given
	 * PEMouseEvent as the starting caret location.
	 * [pending] should be implemented as an editor kit on its own, with an associated Action
	 * The point is : we must provide two ways of starting the TextEditor, one with a known
	 * textEditable and mouseEvent (current implementation), the other W/O these information
	 * (then, the TextEditor should first fetch a TextEditable in the drawing from mouse input,
	 * and start the TextEditor on it)
	 */
	/* 	protected void startTextEditor(TextEditable textEditable, PEMouseEvent me){ [pending]
			board.setCursor(new Cursor(Cursor.TEXT_CURSOR));
			currentMouseTool = new TextEditor(textEditable, me, this);
			// [pending] 09/02/2002 set canvas's mouse-listener to it otherwise it won't work !
		} */











	/////////////////////////////////////////////////////////////////////////////
	//// MOUSE UTILITIES (package access because they're mainly accessed from MouseTool's
	/////////////////////////////////////////////////////////////////////////////

	/**
	 * Convenience for retrieving a HitInfo from a given mouse-event, by calling <code>hitTest</code>
	 * on either the selection-handler, or the drawing.
	 * [SR:pending] in EditGroup mode, this should be adapted so as to look up the PicGroup instead of the
	 * drawing.
	 * @return the HitInfo for the given mouse event
	 * @param selectionOnly if true, lookup selection-handler only ; otherwise, look up the whole drawing.
	 */
	/* package */ HitInfo hitTest(PEMouseEvent e, boolean selectionOnly){
		// if (!EditGroupMode){
		if (selectionOnly) {
			HitInfo hi = selectionHandler.hitTest(e); // include highlighter
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "hi (in selection) = "+hi);
			return hi;
		}
		else {
			HitInfo hi = e.getCanvas().getDrawing().getRootView().hitTest(e,false); // don't include highlighter
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug( "hi (in drawing) = "+hi);
			return hi;
		}
		// }
		// else {
		// if (selectionOnly) return selectionHandler.hitTest(e,true);
		// takes for granted that selection-handler
		// reflects changes in the group being edited
		// else return group.getView().hitTest(e,false);
	}



	///////////////////////////////////////////////
	//// EVENTS MANAGEMENT
	//////////////////////////////////////////////

	/**
	 * adds a PropertyChangeListener to the listener list
	 */
	public void addPropertyChangeListener(PropertyChangeListener l) {
		listenerList.add(PropertyChangeListener.class, l);
	}

	/**
	 * removes an PropertyChangeListener from the EditorKit
	 */
	public void removePropertyChangeListener(PropertyChangeListener l) {
		listenerList.remove(PropertyChangeListener.class, l);
	}

	/**
	 * Notify all listeners that have registered interest for notification on this event type.  
	 * The event instance is lazily created using the parameters passed into the fire method.
	 */
	protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		PropertyChangeEvent e = null;
		// Process the listeners last to first, notifying those that are interested in this event
		for (int i = listeners.length-2; i>=0; i-=2) {
			if (listeners[i]==PropertyChangeListener.class) {
				// Lazily create the event:
				if (e == null) 	e = new PropertyChangeEvent(this,propertyName, oldValue, newValue);
				((PropertyChangeListener)listeners[i+1]).propertyChange(e);
			}
		}
	}

	/**
	 * adds a HelpMessageListener to the listener list
	 */
	public void addHelpMessageListener(HelpMessageListener l) {
		listenerList.add(HelpMessageListener.class, l);
	}

	/**
	 * removes an HelpMessageListener from the EditorKit
	 */
	public void removeHelpMessageListener(HelpMessageListener l) {
		listenerList.remove(HelpMessageListener.class, l);
	}

	/**
	 * Notify all listeners that have registered interest for notification on this event type.  
	 * The event instance is lazily created using the parameters passed into the fire method.
	 */
	protected void postHelpMessage(String message){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("posted msg="+message);
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		HelpMessageEvent e = null;
		// Process the listeners last to first, notifying those that are interested in this event
		for (int i = listeners.length-2; i>=0; i-=2) {
			if (listeners[i]==HelpMessageListener.class) {
				// Lazily create the event:
				if (e == null) 	e = new HelpMessageEvent(this,message);
				((HelpMessageListener)listeners[i+1]).helpMessagePosted(e);
			}
		}
	}



	///////////////////////////////////////////
	/// POPUP MENU
	///////////////////////////////////////////
	/**
	 * Set the popup-menu factory that produces a JPopupMenu when the EditorKit asks for it.
	 * <p>
	 * In this implementation, a popup-menu gets raised if :
	 * <ul>
	 * <li> a PopupTrigger mouse-event occured, (see MouseEvent.isPopupTrigger())
	 * <li> the current edit mode is "SELECT"
	 * <li> the PopupMenuFactory is not null
	 * <li> the PopupMenuFactory.createPopupMenu() method returned a non-null JPopupMenu
	 * </ul>
	 */
	public void setPopupMenuFactory(PopupMenuFactory popupMenuFactory){
		this.popupMenuFactory = popupMenuFactory;
	}

	/**
	 * @return the popup-menu factory that produces JPopupMenu's for this EditorKit
	 */
	public PopupMenuFactory getPopupMenuFactory(){
		return popupMenuFactory ;
	}



	//////////////////////////////////////////
	//// VIEW and VIEW FACTORY
	//////////////////////////////////////////

	/**
	 * Allow this EditorKit to paint shapes specific to the current tool, either directly,
	 * or through the paint method of its tools.
	 * @param scale the current scale factor b/w model-coordinates and view-coordinates ;
	 *        this depends on the AffineTransform being currently active in PECanvas's paintComponent,
	 *        and may be used, e.g. to scale down lines thickness so that they're displayed with a
	 *        constant thickness whatever the scale factor attached to Graphics2D (this is faster than
	 *        retrieving the AffineTransform attached to the graphic context given as parameter, and
	 *        computing the scale factor by using getScaleX).
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug();
		// paint highlighter :
		selectionHandler.paint(g, allocation, scale);
		// paint current tool :
		if (currentMouseTool != null) currentMouseTool.paint(g, allocation, scale);
		// add other paint rendering here :
	}

	/**
	 * @return the ViewFactory that produces View's for the kind of content-type this editor-kit
	 *         can edit or create.
	 */
	public ViewFactory getViewFactory(){
		return viewFactory;
	}

	/**
	 * Return the factory delegate for producing View's Highlighters
	 */
	public HighlighterFactory getHighlighterFactory(){
		return highlighterFactory;
	}
	
	/**
	 * Sets the factory delegate for producing View's Highlighters
	 */
	public void setHighlighterFactory(HighlighterFactory hf){
		this.highlighterFactory=hf;
		//[SR:pending] rebuild View tree !!!
	}
	
	/**
	 * This is the core ViewFactory that produces View's for the associated Drawing.
	 * It actually delegates most of its job to an underlying ViewFactory as set by 
	 * <code>setContentType</code>, except for producing :
	 * <ul>
	 * <li> the root-view for the Drawing's root-element ;
	 * <li> View's for Element's that implement the ViewFactory interface themselves, in which
	 *      case the Element itself acts as the delegate ; this allow plug-ins designers willing
	 *      to implement new Element's, to provide their own ViewFactory for their element, w/o having
	 *      to rebuild a whole factory from scratch, especially if they want to base their plug-in
	 *      on the currently installed EditorKit (ie this editorkit). 
	 * </ul>
	 */
	class ViewFactoryWrapper implements ViewFactory {

		private ViewFactory delegate;

		/**
		 * construct a new view factory for this editor-kit, with the given view-factory
		 * as the delegates for elements that aren't handled here.
		 */
		public ViewFactoryWrapper(ViewFactory delegate){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("delegate="+delegate);
			this.delegate = delegate;
		}

		/**
		 * Return a View for the given Element.
		 * If <code>element</code> is a Drawing.RootElement, returns a RootView,
		 * else if <code>element</code> implements the ViewFactory interface, it's asked to return a view for itself,
		 * otherwise delegates to the factory given as a parameter to the constructor.
		 */
		public View createView(Element element){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("creating view for element : " + element);
			if (element instanceof Drawing.RootElement){
				return new RootView((Drawing.RootElement)element);
			}
			else if (element instanceof ViewFactory){
				if (jpicedt.Log.DEBUG) jpicedt.Log.debug("delegating View creation to the element itself");
				return ((ViewFactory)element).createView(element);
			}
			else {
				if (jpicedt.Log.DEBUG) jpicedt.Log.debug("delegating View creation to " + delegate);
				View v = delegate.createView(element);
				v.setHighlighter(highlighterFactory.createHighlighter(element));
				return v;
			}
		}
	}

	/**
	 * the view associated with the root-element of the drawing
	 */
	class RootView extends AbstractView {

		public RootView(Drawing.RootElement e){
			super(e);
		}

		/**
		 * Notify this View of some change in the model. This is usually called by the associated 
		 * Drawing (through its RootElement) when its content changes (e.g. child added). Conversely,
		 * it's not called when the geometry of a child changed. [pending] quite awkward...
		 * We simply call <code>repaint</code> on the associated canvas, if this kit is installed in it.
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType);
			if (board != null) board.repaint();
		}

		/**
		 * Return a HitInfo corresponding to the given click-point in model-coordinate.
		 * This simply provides a way of testing a mouse-hit on every element of the associated drawing.
		 * @param isHighlightVisible N/A here.
		 */
		public HitInfo hitTest(PEMouseEvent me, boolean isHighlightVisible){
			return hitTest(me);
		}
		
		/**
		 * Returns a HitInfo corresponding to the given mouse-event.
		 */
		protected HitInfo hitTest(PEMouseEvent me){

			Drawing.RootElement e = (Drawing.RootElement)element;
			for(int i = e.getChildCount()-1; i>=0; i--){ // from top to bottom
				Element o = e.getChildAt(i);
				HitInfo hi = o.getView().hitTest(me,false); // don't include highlighter
				if (hi != null) return hi;
			}
			return null;
		}

		/**
		 * Fetches the container hosting the view.  
		 * @return the container, null if this editor-kit hasn't been installed yet.
		 */
		public PECanvas getContainer() {
			return EditorKit.this.board;
		}

		/**
			 * Fetches the ViewFactory implementation that is feeding the view hierarchy.
		 * @return the factory associated with this kit, according to the current content-type.
		 */
		public ViewFactory getViewFactory() {
			return viewFactory;
		}

		/**
		 * Render the View for the graphic element to the given graphic context. This calls "paint"
		 * on each child's view.
		 * @param a the current graphic clip
		 */
		public void paint(Graphics2D g, Rectangle2D a){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug();
			//if (!allocation.intersects(getBounds())) return; // not reliable, because getBounds is not always up-to-date ?
			Drawing.RootElement be = (Drawing.RootElement)element;
			for(Iterator it=be.children(); it.hasNext();){
				View v  = ((Element)it.next()).getView();
				if (v !=null) v.paint(g,a);
			}
		}

		/** does nothing ; there's no hightlighter to be painted for Drawing.RootElement */
		public void paintHighlighter(Graphics2D g, Rectangle2D allocation, double scale){}
	}


	////////////////////////////////////////////////////////////////////////////////
	//// SELECTION HANDLER
	////////////////////////////////////////////////////////////////////////////////

	/**
	 * @return the current selection-handler that suited for the kind of content this editorkit must edit
	 */
	public SelectionHandler getSelectionHandler(){
		return selectionHandler;
	}


	//////////////////////////////////////////////////////////////
	/// EVENTS HANDLER
	//////////////////////////////////////////////////////////////

	/**
	 * Listens to changes from PECanvas, and dispatches to various inner classes interested in these
	 * changes. 
	 */
	private class PropertyChangeHandler implements PropertyChangeListener {

		/**
		 * called when a bound property changed in the canvas. If this is a DRAWING_CHANGE,
		 * we remove the old DrawingHandler, and register a new one for the new drawing.
		 * Currently, this only affects event dispatching to the selection-handler.
		 */
		public void propertyChange(PropertyChangeEvent e){
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("event:name="
				        +e.getPropertyName()+",old="+e.getOldValue()+",new="+e.getNewValue());
			if (e.getPropertyName()==PECanvas.DRAWING_CHANGE){
				// this is called e.g. when canvas.setDrawing() is invoked, as a result for instance of a read() operation.
				Drawing oldDr = (Drawing)e.getOldValue();
				Drawing newDr = (Drawing)e.getNewValue();
				selectionHandler.unSelectAll(); // security
				if (oldDr != null) oldDr.removeDrawingListener(selectionHandler);
				if (newDr != null) newDr.addDrawingListener(selectionHandler);
			}
		}
	}





	///////////////////////////////////////////
	//// ACTIONS (static inner classes)
	///////////////////////////////////////////

	// ---- select edit mode ----

	/**
	 * Sets the current mouse-tool for the editor-kit. Action commands are the same
	 * as the predefined mouse-tool's names for this kit.
	 */
	public static class SelectMouseToolAction extends PEAction {
		private String mouseToolName;
		/**
		 * Construct a new SelectMouseToolAction for the given mouse-tool name.
		 */
		public SelectMouseToolAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer,String mouseToolName){
			super(actionDispatcher, mouseToolName, localizer);
			this.mouseToolName = mouseToolName;
		}
		public void actionPerformed(ActionEvent e){
			EditorKit kit = getEditorKit();
			if (kit != null) kit.setCurrentMouseTool(mouseToolName);
		}
	}

	// ---- toggle edit-points-mode ----

	/**
	 * Toggles the EditorKit's edit-points-mode b/w LOCAL_MODE and GLOBAL_MODE.
	 */
	public static class ToggleEditPointsModeAction extends PEAction {

		public static final String KEY = "action.editorkit.EditPointsMode";
		
		/**
		 * Construct a new ToggleEditPointsModeAction.
		 */
		public ToggleEditPointsModeAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void actionPerformed(ActionEvent e){
			EditorKit kit = getEditorKit();
			if (kit != null && kit.getSelectionHandler() instanceof DefaultSelectionHandler){
				DefaultSelectionHandler h = (DefaultSelectionHandler)kit.getSelectionHandler();
				h.toggleHighlightingMode();
			}
		}
	}

	// ---- edit disposition ----

	/**
	 * Bring one or more selected objets to back, to front, forward or backward.
	 * Latter actions work only if a single Element is selected (otherwise it 's really too tedious to work out)
	 */
	public static class EditDispositionAction extends PEAction {
		/** Action command for bringing elements to front */
		public static final String TOFRONT = "action.editorkit.BringToFront";
		/** Action command for bringing elements to back */
		public static final String TOBACK = "action.editorkit.BringToBack";
		/** Action command for moving element forward */
		public static final String FORWARD = "action.editorkit.BringForward";
		/** Action command for moving element backward */
		public static final String BACKWARD = "action.editorkit.BringBackward";
		private String type;
		/**
		 * Change the layer-disposition of one or more selected objects if there's one (and only ONE ! ) 
		 * at the time the action is performed.<br>
		 * @param type TOFRONT, TOBACK, FORWARD or BACKWARD
		 */
		public EditDispositionAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer,String type){
			super(actionDispatcher, type, localizer);
			this.type = type;
		}

		public void actionPerformed(ActionEvent e){

			if (getCanvas()==null) return;
			if (getCanvas().getSelectionSize()==0) return;
			if (getCanvas().getSelectionSize()>1 && (type==FORWARD||type==BACKWARD)) return;
			
			getCanvas().beginUndoableUpdate((String)getValue(NAME)); // i18n'd if there was a localizer provided in the contructor
			// first build a separate list of element to reorder, in order to prevent concurrent reordering of the selection:
			ArrayList selElements = new ArrayList();
			for (Iterator it = getCanvas().selection(); it.hasNext();){
				selElements.add(it.next());
			}
			int selElementIdx, increment;
			if (type==TOFRONT || type==FORWARD) {
				selElementIdx=0;
				increment=1;
			}
			else {
				selElementIdx=selElements.size()-1;
				increment=-1;
			}
			for (int i=0; i<selElements.size(); i++){
				Element obj1 = (Element)selElements.get(selElementIdx);
				BranchElement p = (BranchElement)obj1.getParent(); // this will modify the disposition of this element
																	// amongst its siblings.
				// for each operation, we check if it was successful ; if it's the case, we fire an undoable event.
				if (type == TOFRONT) p.bringToFront(obj1); // warning : this in turn trigger a re-ordering of the selection
				else if (type == TOBACK) p.bringToBack(obj1);
				else if (type == FORWARD) p.bringForward(obj1);
				else if (type == BACKWARD) p.bringBackward(obj1);
				// getCanvas().repaint(); // done from inside BranchElement through a fireChangedUpdate
				selElementIdx += increment;
			}
			getCanvas().endUndoableUpdate();
		}
	}

	// ---- group selection ----

	/**
	 * Group all selected objects into a new PicGroup.
	 */
	public static class GroupAction extends PEAction {

		public static final String KEY = "action.editorkit.GroupSelection";
		
		/** set "GroupAction" as the action's name */
		public GroupAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}
		public void undoableActionPerformed(ActionEvent e){
			getCanvas().groupSelection();
		}
	}
	
	// ---- ungroup selected group ----
	
	/**
	 * Ungroup the current selection if this makes sense.
	 */
	public static class UngroupAction extends PEAction {

		public static final String KEY = "action.editorkit.UngroupSelection";
		
		public UngroupAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void undoableActionPerformed(ActionEvent e){
			Element o = getSelectedObject();
			if (o instanceof PicGroup){
				getCanvas().unGroup((PicGroup)o);
			}
		}
	}
	
	// ---- join PicMultiCurveConvertable's ----

	/**
	 * Join all selected objects of type PicMultiCurveConvertable into a new PicMultiCurve.
	 * @since jpicedt 1.4pre5
	 */
	public static class JoinCurvesAction extends PEAction {

		public static final String KEY = "action.editorkit.JoinSelection";
		
		public JoinCurvesAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}
		public void undoableActionPerformed(ActionEvent e){
			getCanvas().joinSelection();
		}
	}
	

	// ---- grid/snap on/off ----

	/**
	 * Toggles grid's visible state
	 */
	public static class ShowGridAction extends PEToggleAction {

		public static final String KEY = "action.editorkit.ShowGrid";
		
		/** set "GridVisible" as the action's name */
		public ShowGridAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			if (e.getSource() instanceof AbstractButton){// simply toggle
				AbstractButton source = (AbstractButton)e.getSource();
				getCanvas().getGrid().setVisible(source.isSelected());
				getCanvas().repaint();
			}
			else {
				boolean b = getCanvas().getGrid().isVisible();
				getCanvas().getGrid().setVisible(!b);
				getCanvas().repaint();
			}
		}
	}
	
	/**
	 * Toggles grid-snap activation
	 */
	public static class ActivateSnapAction extends PEToggleAction {

		public static final String KEY = "action.editorkit.ActivateSnap";
		
		public ActivateSnapAction(ActionDispatcher actionDispatcher, ActionLocalizer localizer){
			super(actionDispatcher, KEY, localizer);
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			if (e.getSource() instanceof AbstractButton){// simply toggle
				AbstractButton source = (AbstractButton)e.getSource();
				getCanvas().getGrid().setSnapOn(source.isSelected());
			}
			else {
				boolean b = getCanvas().getGrid().isSnapOn();
				getCanvas().getGrid().setSnapOn(!b);
			}
		}
	}
	
	// ---- edit geometry ----
	
	
	/**
	 * Open a dialog that allows the user to edit the geometry of the currently selected element
	 * from a dialog box.
	 * (N/A if the selection contains more than one element)
	 */
	public static class EditGeometryAction extends PEAction {

		public static final String KEY = "action.editorkit.EditGeometry";
		
		/**
		 * target=currently selected object if there's one, no effect otherwise.
		 */
		public EditGeometryAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void actionPerformed(ActionEvent e){
			Element elem = getSelectedObject();
			if (elem != null && elem instanceof CustomizerFactory){
				getCanvas().beginUndoableUpdate((String)getValue(NAME)); // i18n'd if there was a localizer provided in the contructor
				AbstractCustomizer pane = ((CustomizerFactory)elem).createCustomizer();
				int buttonsMask = CustomizerDialog.OK_CANCEL_BUTTONS | CustomizerDialog.RELOAD_BUTTON; 
				CustomizerDialog dlg = new CustomizerDialog(pane, buttonsMask);
				if (elem instanceof TextEditable) dlg.setOkButtonClosesDialog(true);
				else dlg.setOkButtonClosesDialog(false);
				dlg.setVisible(true);
				boolean isCancelled = dlg.isCancelled();
				getCanvas().endUndoableUpdate();
				if (isCancelled) {
					try {
						getCanvas().undo();
					}
					catch (javax.swing.undo.CannotUndoException ex){
						//if (jpicedt.Log.DEBUG) 
						ex.printStackTrace();
					}
				}
			}
		}
	}

	// ---- paste ----
	
	/**
	 * Paste the content of the given ClipBoard into the target canvas 
	 */
	public static class PasteAction extends PEAction {
		private Clipboard clipboard;
		public static final String KEY_PASTE ="action.editorkit.Paste";
		public static final String KEY_PASTE_SPECIAL="action.editorkit.PasteSpecial";

		/** paste from the given clipboard */
		public PasteAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer,Clipboard clipboard){
			super(actionDispatcher,KEY_PASTE,localizer);
			this.clipboard = clipboard;
		}

		/** paste from System clipboard ; action name = "PasteSpecial" */
		public PasteAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY_PASTE_SPECIAL,localizer);
			this.clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		}

		public void undoableActionPerformed(ActionEvent e){
			try {
				getCanvas().paste(clipboard,true); // translate clipboard before pasting
			}
			catch (jpicedt.graphic.io.parser.ParserException pEx){
				JOptionPane.showMessageDialog(
				    null,
				    jpicedt.Localizer.currentLocalizer().get("exception.ParserWarning") + " :\n" + pEx,
				    jpicedt.Localizer.currentLocalizer().get("Paste"),
				    JOptionPane.ERROR_MESSAGE);
			}
			catch (IOException ioEx){
				JOptionPane.showMessageDialog(
				    null,
				    jpicedt.Localizer.currentLocalizer().get("exception.IOError") + "\n" + ioEx.getMessage(),
				    jpicedt.Localizer.currentLocalizer().get("Paste"),
				    JOptionPane.ERROR_MESSAGE);
			}
			catch (UnsupportedFlavorException ufEx){
				JOptionPane.showMessageDialog(
				    null,
				    jpicedt.Localizer.currentLocalizer().get("UnsupportedDataFlavor") + "\n" + ufEx.getMessage(),
				    jpicedt.Localizer.currentLocalizer().get("Paste"),
				    JOptionPane.ERROR_MESSAGE);
			}
		}
	}

	// ---- copy ----
	
	/**
	 * Copy target's content to clipboard 
	 */
	public static class CopyAction extends PEAction {
		public static final String KEY = "action.editorkit.Copy";
		private Clipboard clipboard;
		public CopyAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer,Clipboard clipboard){
			super(actionDispatcher,KEY,localizer);
			this.clipboard = clipboard;
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			getCanvas().copy(clipboard);
		}
	}

	// ---- cut ----

	/**
	 * Cut target's content to clipboard
	 */
	public static class CutAction extends PEAction {
		public static final String KEY = "action.editorkit.Cut";
		private Clipboard clipboard;
		public CutAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer,Clipboard clipboard){
			super(actionDispatcher,KEY,localizer);
			this.clipboard = clipboard;
		}

		public void undoableActionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			getCanvas().cut(clipboard);
		}
	}

	// ---- delete ----
	
	/**
	 * Delete the current selection 
	 */
	public static class DeleteAction extends PEAction {
		public static final String KEY = "action.editorkit.Delete";
		public DeleteAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void undoableActionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			getCanvas().deleteSelection();
		}
	}

	// ---- select all ----
	
	/**
	 * selects all objects in the current active internal frame
	 */
	public static class SelectAllAction extends PEAction {
		public static final String KEY = "action.editorkit.SelectAll";
		public SelectAllAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			getCanvas().selectAll(true);
		}
	}

	// ---- zoom in ----
	
	/**
	 * zoom in 
	 */
	public static class ZoomInAction extends PEAction {
		public static final String KEY = "action.editorkit.ZoomIn";
		public ZoomInAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			double zoom = getCanvas().getZoomFactor();
			zoom *= 2.0;
			getCanvas().setZoomFactor(zoom);
		}
	}

	// ---- zoom out ----
	
	/**
	 * zoom out 
	 */
	public static class ZoomOutAction extends PEAction {
		public static final String KEY = "action.editorkit.ZoomOut";
		public ZoomOutAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void actionPerformed(ActionEvent e){
			double zoom = getCanvas().getZoomFactor();
			zoom /= 2.0;
			getCanvas().setZoomFactor(zoom);
		}
	}

	// ---- edit drawing's bounding box ----
	
	/**
	 * edit Drawing's bounding box
	 */
	public static class EditBoundingBoxAction extends PEAction {
		public static final String KEY = "action.editorkit.EditBoundingBox";
		public EditBoundingBoxAction(ActionDispatcher actionDispatcher,ActionLocalizer localizer){
			super(actionDispatcher,KEY,localizer);
		}

		public void actionPerformed(ActionEvent e){
			if (getCanvas()==null) return;
			new CustomizerDialog(getCanvas().getDrawing().getCustomizer()).setVisible(true);
		}
	}

	// ---- Action Map ----
	
	/**
	 * @return an ActionMap that can be used on a PECanvas that is using a model, view and
	 *         formatter produced by this kit. ActionCommand key is created from key
	 *         ACTION_COMMAND_KEY.
	 */
	public static ActionMap createActionMap(ActionDispatcher dispatcher, ActionLocalizer localizer){
		
		ActionMap map = new ActionMap();
		Action a;

		// copy, cut, delete, paste, pasteSpecial, selectAll :		
		a = new CopyAction(dispatcher,localizer,clipboard); // clipboard is a static variable !
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
 		a = new CutAction(dispatcher,localizer,clipboard);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
 		a = new DeleteAction(dispatcher,localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
 		a = new PasteAction(dispatcher,localizer,clipboard);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new PasteAction(dispatcher,localizer); // paste special
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
 		a = new SelectAllAction(dispatcher,localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		// edit disposition :
		a = new EditDispositionAction(dispatcher, localizer, EditDispositionAction.TOBACK);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new EditDispositionAction(dispatcher, localizer, EditDispositionAction.BACKWARD);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new EditDispositionAction(dispatcher, localizer, EditDispositionAction.FORWARD);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new EditDispositionAction(dispatcher, localizer, EditDispositionAction.TOFRONT);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		// toggle edit points mode :
		a = new ToggleEditPointsModeAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		// registered mouse tool :
		String[][] allToolNames = getAvailableToolNames();
		for (int i=0; i<allToolNames.length; i++){
			for (int j=0; j<allToolNames[i].length; j++){
				a = new SelectMouseToolAction(dispatcher, localizer,allToolNames[i][j]);
				map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
			}
		}
		
		// edit attributes :
		a = new EditGeometryAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		// group, ungroup and related :
		a = new GroupAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new UngroupAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new JoinCurvesAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		// zoom in, zoom out, grid :
		a = new ZoomInAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new ZoomOutAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		a = new ShowGridAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		
		// bounding box :
		a = new EditBoundingBoxAction(dispatcher, localizer);
		map.put(a.getValue(Action.ACTION_COMMAND_KEY), a);
		
		return map;
	}

	/////////////////////////////////////////
	//// KEYBRD EVENTS MANAGEMENT
	///////////////////////////////////////////
	
	/**
	 * @return an ActionMap that can be used on a PECanvas that is using a model, view and
	 *         formatter produced by this kit. ActionCommand key is created from key
	 *         ACTION_COMMAND_KEY.
	 * [todo] provide a Properties object for key-bindings
	 */
	public InputMap createInputMap(){
		
		InputMap map = new InputMap();
		map.put(KeyStroke.getKeyStroke('/'),EditDispositionAction.FORWARD);
		map.put(KeyStroke.getKeyStroke('*'),EditDispositionAction.BACKWARD);
		map.put(KeyStroke.getKeyStroke('+'),ZoomInAction.KEY);
		map.put(KeyStroke.getKeyStroke('-'),ZoomOutAction.KEY);
		map.put(KeyStroke.getKeyStroke(KeyEvent.VK_F2,0),EditGeometryAction.KEY);
		// add your own key/action bindings here ...
		return map;
	}

	/////////////////////////////////////////
	//// MISC STATIC METHODS
	///////////////////////////////////////////
	/**
	 * Return a static reference to the local ClipBoard shared across all instance of this EditorKit
	 */
	public static Clipboard getClipboard(){
		return clipboard;
	}
	 
	
}

