/*
 RotateTool.java - September 25, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.event.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.Element;

import java.awt.*;
import java.awt.geom.*;


/**
 * A MouseTool used to rotate a target Element using a origin point and a mobile point used to
 * compute the rotation angle..
 * <p>
 * Sequence of mouse operations :
 * <ul>
 * <li>mousePressed -> set origin
 * <li>mouseDragged -> set initial position of mobile point ; draw corresponding vector
 * <li>mouseReleased -> switch to "move mobile point" task
 * <li>mouseMoved -> move mobile point, and rotate target element accordingly
 * <li>mousePressed -> complete operation, then reset.
 *
 * </ul>
 * @since jPicEdt 1.4
 * @author Sylvain Reynal
 * @version $Id: RotateTool.java,v 1.6.2.1 2007/09/02 11:56:28 reynal Exp $
 *
 */
public class RotateTool extends MouseTool {

	private EditorKit kit;
	
	// paint related :
	private Line2D.Double lineSrc, lineDst; // initial and mobile vectors used to compute the rotation angle
	private Stroke stroke;
	private CursorFactory cursorFactory = new CursorFactory();

	// geometry related :
	private Element target;
	private PicPoint ptOrg = new PicPoint(); // rotation center
	private PicPoint lastDragPoint = new PicPoint(); // last mobile point position
	private PicVector v1=new PicVector(), v2=new PicVector(); // buffers used to compute rotation angle from lineSrc and lineDst
	
	// task related :
	private int currentTask;
	private final static int SET_SRC_VECT = 0;
	private final static int SET_MOBILE_VECT = 1;
	private final static int COMPLETED = 2;

	/**
	 * @param kit the EditorKit this mouse-tool is to registered with.
	 */
	public RotateTool(EditorKit kit, Element target){
		this.kit = kit;
		this.target = target;
		currentTask = SET_SRC_VECT; // first mouse pressed will set the rotation center (ptOrg)
	}
	
	/** set the source rect, then the destination rect */
	public void mousePressed(PEMouseEvent e){
		super.mousePressed(e);
		if (e.isPopupTrigger())  {
			kit.setCurrentMouseTool(EditorKit.SELECT); // callbacks flush()
			return;
		}
		PicPoint pt = e.getPicPoint();
		e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
		switch (currentTask){
			case SET_SRC_VECT : 
				e.getCanvas().beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("action.editorkit.Rotate.tooltip"));
				lineSrc = new Line2D.Double(pt.x,pt.y,pt.x,pt.y);
				lineDst = new Line2D.Double(pt.x,pt.y,pt.x,pt.y);
				ptOrg.x = pt.x;
				ptOrg.y = pt.y;
				break;
			case SET_MOBILE_VECT : // after a mouseMoved
				currentTask = COMPLETED;
				break;
			default:
		}
	}

	/**  */
	public void mouseDragged(PEMouseEvent e){
		super.mouseDragged(e);
		if (e.isPopupTrigger()) return;
		switch (currentTask){
			case SET_SRC_VECT : // set initial position of mobile point
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				if (lineSrc.x2 == pt.x && lineSrc.y2 == pt.y) return;
				lineSrc.x2 = pt.x;
				lineSrc.y2 = pt.y;
				e.getCanvas().repaint(); // [pending]
				break;
			case SET_MOBILE_VECT :
				break;
			default:
		}
	}

	/** */	
	public void mouseReleased(PEMouseEvent e){
		super.mouseReleased(e);
		if (e.isPopupTrigger())  return;
		switch (currentTask){
			case SET_SRC_VECT : // complete drawing of source vector, then switch to mobile vector.
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				lastDragPoint.setCoordinates(pt);
				currentTask = SET_MOBILE_VECT;
				break;
			case COMPLETED :
				flush();
				e.getCanvas().repaint();
				e.getCanvas().endUndoableUpdate();
				break;
			default:
		}
	}

	/**  */
	public void mouseMoved(PEMouseEvent e){
		super.mouseMoved(e);
		e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.SELECT));// [SR:pending] create specific ROTATE cursor
		kit.postHelpMessage("help-message.MoveEndPointTransform");
		switch (currentTask){
			case SET_SRC_VECT :
				break;
			case SET_MOBILE_VECT :
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				if (lineDst.x2 == pt.x && lineDst.y2 == pt.y) return;
				lineDst.x2 = pt.x;
				lineDst.y2 = pt.y;
				e.getCanvas().repaint(); // [pending]
				// compute rotation angle :
				v1.setCoordinates(ptOrg, lastDragPoint);
				v2.setCoordinates(ptOrg, pt);
				double angle = v1.angle(v2); // no worry, this handles null vectors by returning a null angle
				//jpicedt.Log.debug("v1="+v1+", v2="+v2+", angle="+Math.toDegrees(angle)+" deg");
				if (angle == 0) return;
				target.rotate(ptOrg,angle); // rotate element
				lastDragPoint.setCoordinates(pt);
				break;
			default:
		}
	}


	/**
	 * Called when this tool is being replaced by another tool in the hosting kit
	 */
	public void flush(){
		lineSrc = lineDst = null;
		currentTask = SET_SRC_VECT; // prepare next mousePressed
	}
	
	/**
	 * Allow the tool to paint shapes that are specific to this tool. 
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		if (lineSrc == null || lineDst == null) return;
		if (stroke==null) {
			float[] dash = {1.0f,1.0f};
			stroke = new BasicStroke((float)(1.0/scale),BasicStroke.CAP_ROUND,BasicStroke.JOIN_ROUND,10.0f,dash,0.5f);
		}
		g.setStroke(stroke);
		g.setPaint(Color.blue);
		g.draw(lineSrc);
		g.setPaint(Color.red);
		g.draw(lineDst);
	}
	
} // class 

