/*
 ScaleTool.java - February 28, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.toolkit;

import jpicedt.graphic.event.*;
import jpicedt.graphic.*;
import jpicedt.graphic.model.Element;

import java.awt.*;
import java.awt.geom.*;


/**
 * A MouseTool used to scale a target Element using a source and a destination vector.
 * <p>
 * Sequence of mouse operations :
 * <ul>
 * <li>mousePressed -> set first corner of source rectangle
 * <li>mouseDragged -> drag second corner source rectangle
 * <li>mouseReleased -> switch to destination rectangle
 * <li>mouseMoved -> set second corner of destination rectangle, and rescale target element accordingly
 * <li>mousePressed -> complete operation, then reset.
 *
 * </ul>
 * @since jPicEdt 1.3
 * @author Sylvain Reynal
 * @version $Id: ScaleTool.java,v 1.7.2.1 2007/09/02 11:56:29 reynal Exp $
 *
 */
public class ScaleTool extends MouseTool {

	private EditorKit kit;
	private Line2D.Double vecSrc, vecDst;
	private PicPoint ptOrg = new PicPoint();
	private CursorFactory cursorFactory = new CursorFactory();
	private Element target;
	private PicPoint lastDragPoint = new PicPoint();
	private int currentTask;
	private final static int SET_SRC_RECT = 0;
	private final static int SET_DST_RECT = 1;
	private final static int COMPLETED = 2;

	/**
	 * @param kit the EditorKit this mouse-tool is to registered with.
	 */
	public ScaleTool(EditorKit kit, Element target){
		this.kit = kit;
		this.target = target;
		currentTask = SET_SRC_RECT; // first mouse pressed will set the first corner of the source rectangle
	}
	
	/** set the source rect, then the destination rect */
	public void mousePressed(PEMouseEvent e){
		super.mousePressed(e);
		if (e.isPopupTrigger())  {
			kit.setCurrentMouseTool(EditorKit.SELECT); // callbacks flush()
			return;
		}
		PicPoint pt = e.getPicPoint();
		e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
		switch (currentTask){
			case SET_SRC_RECT : 
				e.getCanvas().beginUndoableUpdate(jpicedt.Localizer.currentLocalizer().get("action.editorkit.Scale.tooltip"));
				vecSrc = new Line2D.Double(pt.x,pt.y,pt.x,pt.y);
				vecDst = new Line2D.Double(pt.x,pt.y,pt.x,pt.y);
				ptOrg.x = pt.x;
				ptOrg.y = pt.y;
				break;
			case SET_DST_RECT : // after a mouseMoved
				currentTask = COMPLETED;
				break;
			default:
		}
	}

	/** */	
	public void mouseReleased(PEMouseEvent e){
		super.mouseReleased(e);
		if (e.isPopupTrigger())  return;
		switch (currentTask){
			case SET_SRC_RECT : // complete drawing of source rectangle, then switch to destination rect.
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				lastDragPoint.x = pt.x;
				lastDragPoint.y = pt.y;
				currentTask = SET_DST_RECT;
				break;
			case COMPLETED :
				flush();
				e.getCanvas().repaint();
				e.getCanvas().endUndoableUpdate();
				break;
			default:
		}
	}

	/**  */
	public void mouseMoved(PEMouseEvent e){
		super.mouseMoved(e);
		e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.NE_RESIZE));
		kit.postHelpMessage("help-message.MoveEndPointTransform");
		switch (currentTask){
			case SET_SRC_RECT :
				break;
			case SET_DST_RECT :
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				if (vecDst.x2 == pt.x && vecDst.y2 == pt.y) return;
				vecDst.x2 = pt.x;
				vecDst.y2 = pt.y;
				e.getCanvas().repaint(); // [pending]
				if (lastDragPoint.x==ptOrg.x || lastDragPoint.y==ptOrg.y) return; // div by 0
				double sx = (vecDst.x2-ptOrg.x)/(lastDragPoint.x-ptOrg.x);
				double sy = (vecDst.y2-ptOrg.y)/(lastDragPoint.y-ptOrg.y);
				if (sx == 0 || sy ==0) return;
				target.scale(ptOrg,sx,sy); // scale element
				lastDragPoint.x = pt.x;
				lastDragPoint.y = pt.y;
				break;
			default:
		}
	}

	/**  */
	public void mouseDragged(PEMouseEvent e){
		super.mouseDragged(e);
		if (e.isPopupTrigger()) return;
		switch (currentTask){
			case SET_SRC_RECT :
				PicPoint pt = e.getPicPoint();
				e.getCanvas().getGrid().nearestNeighbour(pt,pt); // pt = nn(pt)
				if (vecSrc.x2 == pt.x && vecSrc.y2 == pt.y) return;
				vecSrc.x2 = pt.x;
				vecSrc.y2 = pt.y;
				e.getCanvas().repaint(); // [pending]
				break;
			case SET_DST_RECT :
				break;
			default:
		}
	}

	/**
	 * Called when this tool is being replaced by another tool in the hosting kit
	 */
	public void flush(){
		vecSrc = vecDst = null;
		currentTask = SET_SRC_RECT; // prepare next mousePressed
	}
	
	/**
	 * Allow the tool to paint shapes that are specific to this tool. 
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		if (vecSrc == null || vecDst == null) return;
		float[] dash = {1.0f,1.0f};
		g.setStroke(new BasicStroke((float)(1.0/scale),BasicStroke.CAP_ROUND,
		BasicStroke.JOIN_ROUND,10.0f,dash,0.5f));
		g.setPaint(Color.blue);
		g.draw(vecSrc.getBounds2D());
		g.setPaint(Color.red);
		g.draw(vecDst.getBounds2D());
	}
	
} // class 

