/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.graphic.toolkit;

import jpicedt.graphic.*;
import jpicedt.graphic.grid.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.view.HitInfo;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.*;

/**
 * This generic mouse-tool is dedicated to applying MouseTransform's to elements of a Drawing, following a uniform sequence of
 * mouse-events. This class works hand-in-hand with a pluggable MouseTransformFactory which produces MouseTransform's
 * on the fly from incoming mouse-events. Then any ensuing mouse-event invokes a specific method from the produced
 * MouseTransform.
 * <p>
 * Basically, the sequence of mouse-events comprises :
 * <ul>
 * <li> an initial mouse-pressed event, which asks the factory to create an appropriate MouseTransform
 * <li> a sequence of mouse-motion and mouse-button-changed events, where :
 * <ol>
 * <li> mouse-motion events effectively apply the transform to the element ;
 * <li> mouse-button-changed events asks the transform to jump to its next available task, if any.
 * </ol>
 * </ul>
 * Besides, this tool opens a context-sensitive popup menu when a right-click occurs, by delegating to
 * the hosting EditorKit to obtain a suitable PopupMenuFactory. It also sets the cursor shape according to
 * the currently installed MouseTransform.
 * <p>
 * [SR:pending] refactor class name to sth like "GenericTool" or "MouseTransformTool"...
 * @author Sylvain Reynal
 * @since jPicEdt 1.3.2
 * @version $Id: SelectionTool.java,v 1.13.2.1 2007/09/02 11:56:29 reynal Exp $
 * <p>
 *
 */
public class SelectionTool extends MouseTool {

	private MouseTransform currentMouseTransform;
	private EditorKit editorKit;
	private CursorFactory cursorFactory = new CursorFactory();
	private PicPoint ptBuffer = new PicPoint();
	private MouseTransformFactory mtFactory;

	/**
	 * @param editorKit the editorKit that owns this mouseTool
	 * @param mtFactory the factory that produces MouseTransform's in response to a mousePressed
	 */
	SelectionTool(EditorKit editorKit, MouseTransformFactory mtFactory){
		this.editorKit = editorKit;
		this.mtFactory = mtFactory;

	}

	/**
	 * Do tool specific painting. Merely delegates to the currently active mouse transform, as well
	 * as to the installed MouseTransformFactory.
	 */
	public void paint(Graphics2D g, Rectangle2D allocation, double scale){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("paint");
		mtFactory.paint(g,allocation,scale);
		if (currentMouseTransform != null) currentMouseTransform.paint(g,allocation,scale);
	}

	/** called when this tool is being activated in the hosting
	 * editor kit ; this delegates to the underlying MouseTransform factory
	 */
	public void init(){
		editorKit.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.SELECT));
		mtFactory.init();
	}

	/** called when this tool is being replaced by another mouse-tool in the hosting
	 * editor kit ; this delegates to the underlying MouseTransform factory 
	 */
	public void flush(){
		mtFactory.flush();
	}
	
	/////////////////////////////////////////////////
	// MOUSE EVENTS
	/////////////////////////////////////////////////
	/**
	 * - right button click (popup trigger) -> raise popup menu.<br>
	 * - left button ->  createMouseTransform using the mtFactory given in the constructor.<br>
	 */
	public void mousePressed(PEMouseEvent e) {
		super.mousePressed(e);
		if (e.getClickCount() == 2) return;

		// mouse RIGHT button pressed : we open a JPopupMenu if applicable
		if (e.isPopupTrigger()) {
			// if we're currently not in SELECT mode (see, e.g., EDIT_POINT), force this mode before proceeding:
			if (editorKit.getCurrentMouseTool() != EditorKit.SELECT){
				editorKit.setCurrentMouseTool(EditorKit.SELECT); 
				return;
			}
			
			// Note : 23/03/2002 Reynal "Struggling to make the popup-menu visible !"
			// Explanation : in PECanvas.processMouseEvent, the call to super.processMouseEvent was done
			// after processing PEMouseEvent's, whereas it'd to be
			// done BEFORE, otherwise, given that JPopupMenu listens to PECanvas *standard* mouse-events
			// (I didn't know it, actually it's used to close the menu on a 2nd mouse-pressed that occurs outside
			// the menu-items, this is a standard Swing behaviour), we never see the menu (Swing closes
			// it immediately after we open it !) To sum up (with the old implementation of 
			// PECanvas.processMouseEvent), we had :
			// 1) mousePressed -> process PEMouseEvent -> call popup.show in SelectionTool (i.e. here)
			// 2) super.processMouseEvent -> ok, Swing catches the event, and test if the menu is visible,
			//    if it is, it closes it, because the click occured OUTSIDE (or more exactly, on the border of)
			//    the JPopupMenu. Yeah, yeah, this SURE was tricky !
			// We the new implementation, it's just the other way around, so it just works perfect :-)
			PopupMenuFactory factory = editorKit.getPopupMenuFactory();
			if (factory != null){
				HitInfo hi = this.editorKit.hitTest(e, true); // selection first
				if (hi==null) hi = this.editorKit.hitTest(e, false); // whole drawing (and it's up to the receiver to determine if element is selected or not)
				JPopupMenu menu = factory.createPopupMenu(e.getCanvas(),hi);
				if (menu != null){
					menu.show(e.getCanvas(), e.getAwtMouseEvent().getX(), e.getAwtMouseEvent().getY());
				}
			}
			return;
		}

		// mouse LEFT button pressed (single click):

		// a) if a mouse-transform is currently active, try to jump to next task...
		if (currentMouseTransform != null){
			if (!currentMouseTransform.next(e)) currentMouseTransform = null; // finish now
			e.getCanvas().repaint(); // merely a security, since MouseTransform should've done the job anyway.
			return;
		}
		
		// b) else reinit a new mouse-transform...
		currentMouseTransform = mtFactory.createMouseTransform(e);
		if (currentMouseTransform != null) {// there's a selected object under the cursor (and this object can be the selection buffer itself)
			e.getCanvas().setCursor(currentMouseTransform.getCursor());
			currentMouseTransform.start(e);
			editorKit.postHelpMessage(currentMouseTransform.getHelpMessage());
		}
	}

	/**
	 * Invoke "process" on the current MouseTransform (if non-null).
	 */
	public void mouseDragged(PEMouseEvent e) {
		super.mouseDragged(e);
		if (currentMouseTransform != null) currentMouseTransform.process(e);
	}


	/**
	 * Invokes "end" on the current MouseTransform if non-null.
	 */
	public void mouseReleased(PEMouseEvent e){
		super.mouseReleased(e);
		if (e.isPopupTrigger()) return;
					
		// left button :
		if (currentMouseTransform != null){
			if (!currentMouseTransform.next(e)) currentMouseTransform = null; // finish now
			e.getCanvas().repaint(); // merely a security, since MouseTransform should've done the job anyway.
			return;
		}
	}

	/**
	 * Change mouse cursor according to the MouseTransform returned by the factory for the
	 * given PEMouseEvent.
	 */
	public void mouseMoved(PEMouseEvent e) {
		super.mouseMoved(e);
		// if a mouse-transform is currently active, process it...
		if (currentMouseTransform != null) {
			currentMouseTransform.process(e);
			return;
		}
		// else update cursor shape...
		MouseTransform mt = mtFactory.createMouseTransform(e);
		if (jpicedt.Log.DEBUG) 
			jpicedt.Log.debug("returned mt="+mt);
		if (mt != null){ 
			e.getCanvas().setCursor(mt.getCursor());
			editorKit.postHelpMessage(mt.getHelpMessage());
			// [pending] MDIManager.getCurrentManager().getStatusBar().showMessage(mt.getClickedObject().toMessageString(canvas.getSheetOrigin()));
		}
		else e.getCanvas().setCursor(cursorFactory.getPECursor(CursorFactory.SELECT));
	}
	

} // class SelectionTool




