/*
 AbstractView.java - February 14, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.PointIndexIterator;
import jpicedt.graphic.model.DefaultPointIndexIterator;
import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.event.PEMouseEvent;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.util.ArrayList;

/**
 * Abstract implementation that provide some basic common behaviours for View's.
 * @since jpicedt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: AbstractView.java,v 1.11.2.1 2007/09/02 11:56:29 reynal Exp $
 *
 */
public abstract class AbstractView implements View, ViewConstants {

	/**
	 * the graphic element that this View renders 
	 */
	protected Element element;
	
	/**
	 * the bounds rectangle used for clipping (as returned by getBounds) ; this should be updated
	 * by changedUpdate.
	 */
	protected Rectangle2D bounds = new Rectangle2D.Double();
	
	/**
	 * the highlighter delegate for this view
	 */
	protected Highlighter highlighter;
	
	
	/**
	 * construct a new View for the given Element
	 */
	public AbstractView(Element element){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".<init> for element " + element);
		this.element = element;
	}

	/**
	 * construct a new View for the given Element, delegating highlighting
	 * to the given Highlighter.
	 */
	public AbstractView(Element element, Highlighter h){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".<init> for element " + element);
		this.element = element;
		this.highlighter = h;
	}

	/**
 	 * @return the element the View is responsible for rendering
	 */
	public Element getElement(){
		return element;
	}
	
	/**
 	 * set the element the View is responsible for rendering
	 */
	public void setElement(Element e){
		this.element = e;
	}

	///////////////////////////////////////////////////////////////////
	//// tree-structure
	///////////////////////////////////////////////////////////////////
	
	/**
	 * Returns the parent of the view, as given by the tree-structure the associated graphic element belongs to.
	 * @return the parent view, null if none (either because the element has no parent, or
	 *         because the parent has no View).
	 */
	public View getParentView() {
		Element parent = element.getParent();
		if (parent == null) return null;
		else return parent.getView();
	}


	/**
	 * Fetches the container hosting the view.  This is useful for
	 * things like scheduling a repaint, finding out the host 
	 * components font, etc.  The default implementation
	 * of this is to forward the query to the parent view, if any... 
	 * (that means that for this method to return non-null, the element attached to this view
	 * must have a parent, and this parent, a view, etc...)
	 *
	 * @return the container, null if none
	 */
	public PECanvas getContainer() {
		View v = getParentView();
		return (v != null) ? v.getContainer() : null;
	}

	/**
	 * Fetches the ViewFactory implementation that is feeding the view hierarchy.  
	 * @return the factory, null if none
	 */
	public ViewFactory getViewFactory() {
		View v = getParentView();
		return (v != null) ? v.getViewFactory() : null;
	}

	/**
	 * Fetch a Graphics for rendering from the hosting container (if not null). 
	 * This can be used to determine
	 * font characteristics, and will be different for a print view
	 * than a component view.
	 *
	 * @since 1.3
	 */
	public Graphics getGraphics() {
		Component c = getContainer();
		if (c==null) return null;
		return c.getGraphics();
	}

	/**
	 * Fetches the model associated with the view.
	 * @return the view model, null if none
	 */
	public Drawing getDrawing() {
		return element.getDrawing();
	}

	
	////////////////////////////////////////////////////////
	//// PAINT
	////////////////////////////////////////////////////////
	
	/**
	 * ask the hosting container (if it's non-null) to repaint itself
	 * @param clip the clip rectangle in model-coordinate ; if null, simply call repaint() on the 
	 *        hosting container.
	 */
	public void repaint(Rectangle2D clip){
		PECanvas c = getContainer();
		if (c==null) return; // if this view is not installed in a tree with a container
		if (clip != null) c.repaintFromModelRect(clip);
		else c.repaint();
	}


	/**
	 * @return the bounds of this View<br> 
	 *         This will determine the clipping rectangle passed as a parameter to repaint, and may
	 *         include the highlighter's bounds as well.
	 *         <br>
	 *         Default implementation return <code>bounds</code>.
	 */
	public Rectangle2D getBounds(){
		return bounds;
	}

	//////////////////////////////////////////////////////
	//// HIGHLIGHTER
	//////////////////////////////////////////////////////
	
	/**
	 * Returns the Highlighter responsible for rendering the highlighted part of this view.
	 * @return null if this view cannot be highlighted
	 */
	public Highlighter getHighlighter(){
		return highlighter;
	}
	
	/**
	 * Sets the Highlighter responsible for rendering the highlighted part of this view.
	 */
	public void setHighlighter(Highlighter h){
		this.highlighter = h;
		changedUpdate(null);
	}

	/**
	 * Render the Highlighter to the given graphic context.<br>
	 * @param allocation current clipping
	 * @param scale The current scale factor from model to screen for the Graphics2D context ;
	 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
	 *        same lenght on the screen whatever the scale factor that's set to the graphic context.
	 */
	public void paintHighlighter(Graphics2D g, Rectangle2D allocation, double scale){
		if (highlighter!=null) highlighter.paint(g,allocation,scale);
	}

	/////////////////////////////////////////
	/////// MOUSE
	/////////////////////////////////////////
	
	/**
	 * Returns a HitInfo corresponding to the given mouse-event.
	 * Only the view's shape should be included in this test.
	 * 
	 */
	abstract protected HitInfo hitTest(PEMouseEvent e);

	/**
	 * Depending on the value of isHighlighVisible, current implementation first delegates
	 * to the hosted Highlighter's hitTest() method, then calls 
	 * {@link #hitTest(PEMouseEvent) hitTest(PEMouseEvent)}.
	 * @return a HitInfo corresponding to the given mouse-event
	 * @param isHighlightVisible whether the receiver should include the highlighter shapes (e.g. end-points)
	 *        in the click-sensitive area.
	 */
	public HitInfo hitTest(PEMouseEvent e, boolean isHighlightVisible){

		HitInfo hi=null;
		if (isHighlightVisible && highlighter!=null) hi=highlighter.hitTest(e);
		if (hi==null) hi = hitTest(e);
		return hi;
	}
}

