/*
 CompositeView.java - September 24, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.*; // all elements + drawing
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PicVector;
import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;

import java.awt.*;
import java.awt.font.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.util.*;

/**
 * a CompositeView is a graphic representation of a BranchElement
 * @author Sylvain Reynal
 * @since jpicedt 1.4
 * @version $Id: CompositeView.java,v 1.4.2.1 2007/09/02 11:56:29 reynal Exp $
 */
public class CompositeView extends AbstractView {

	private PicPoint ptUL = new PicPoint(); // buffer
	private PicPoint ptLR = new PicPoint(); // buffer

	/**
	 * construct a new View for the given PicGroup
	 */
	public CompositeView(BranchElement e){
		super(e);
		changedUpdate(null);
	}

	/**
	 * Give notification from the model that a change occured for an element this view is responsible
	 * for rendering. This implementation update the view's bounds from the element's bounding box,
	 * delegating highlighter's bounds to the current Highlighter delegate..
	 */
	public void changedUpdate(DrawingEvent.EventType eventType){
		Rectangle2D oldClip = bounds;
		// update bounds
		bounds = element.getBoundingBox(null); // => null if branchElement is empty !
		if (highlighter!=null) {
			double s = 1.0;
			PECanvas canvas = getContainer();
			if (canvas != null) s=canvas.getScaleFactor();   
			highlighter.changedUpdate(eventType,s);
			if (bounds!=null) bounds.add(highlighter.getBounds());
		}
		// if event == ADD/REMOVE/REPLACE child, don't bother, just repaint it all :
		if (eventType==null || eventType != DrawingEvent.EventType.GEOMETRY_CHANGE) repaint(null);
		else { 
			if (oldClip!=null) {
				if (bounds!=null) oldClip.add(bounds);
				repaint(oldClip);
			}
			else {
				if (bounds!=null) repaint(bounds);
				// else branchElement was empty, and is still empty : do nothing
			}
		}
	}

	/**
	 * Render the View for the graphic element to the given graphic context. This called "paint"
	 * on each child's view if its bounds rectangle intersects the clip.
	 * @param allocation the current graphic clip
	 */
	public void paint(Graphics2D g, Rectangle2D a){
		if (!a.intersects(getBounds())) return;
		BranchElement be = (BranchElement)element;
		for(Iterator it=be.children(); it.hasNext();){
			View v = ((Element)it.next()).getView();
			if (v!=null) v.paint(g,a);
		}
	}

	/**
	 * @return a HitInfo corresponding to the given click-point in model-coordinate. 
	 */
	public HitInfo hitTest(PEMouseEvent e){

		PicPoint ptClick = e.getPicPoint();
		// test hit on children, highlighted or not, from top to bottom :
		BranchElement be = (BranchElement)element;
		for(int i = be.getChildCount()-1; i>=0; i--){
			Element o = be.getChildAt(i);
			if (o.getView().hitTest(e,false)!=null) return new HitInfo.Composite(be, i, e);
		}
		return null;
	}
} // CompositeView

