/*
 DefaultHighlighterFactory.java - September 20, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.*; // all elements + drawing
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PicVector;
import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;

import java.awt.*;
import java.awt.font.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.util.*;

/**
 * This is the default factory used to create a Highlighter for a given Element.
 * @since jPicEdt 1.4
 * @author Sylvain Reynal
 * @version $Id: DefaultHighlighterFactory.java,v 1.10.2.1 2007/09/02 11:56:29 reynal Exp $
 */
public class DefaultHighlighterFactory implements HighlighterFactory, ViewConstants, PicObjectConstants {

	public static final Color LOCAL_HIGHLIGHTING_COLOR= Color.green;
	public static final Color GLOBAL_HIGHLIGHTING_COLOR= Color.red;

	public DefaultHighlighterFactory(){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Initializing");
	}

	/**
	 * @return a Highlighter for the given Element
	 * Current implementation returns a View appropriate for the following elements :
	 * <ul>
	 * <li> PicEllipse -> EllipseHighlighter ;
	 * <li> PicCircleFrom3Points -> CircleHighlighter ;
	 * <li> BranchElement -> CompositeHighlighter ;
	 * <li> PicParallelogram -> ParallelogramHighlighter ;
	 * <li> PicText -> TextHighlighter ;
	 * <li> other -> DefaultHighlighter
	 * </ul>
	 */
	public Highlighter createHighlighter(Element element){
		
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("element=" + element);
		// always place daughters BEFORE superclass :
		if (element instanceof PicCircleFrom3Points) return new CircleHighlighter((((PicCircleFrom3Points)element)));
		else if (element instanceof PicEllipse) return new EllipseHighlighter((PicEllipse)element);
		else if (element instanceof PicParallelogram) return new DefaultHighlighter((PicParallelogram)element);
		else if (element instanceof PicSmoothPolygon) return new SmoothPolygonHighlighter((PicSmoothPolygon)element);
		else if (element instanceof PicPsCurve) return new PsCurveHighlighter((PicPsCurve)element);
		else if (element instanceof AbstractCurve) return new AbstractCurveHighlighter((AbstractCurve)element);
		else if (element instanceof PicText) return new DefaultHighlighter((PicText)element);
		else if (element instanceof PicNodeConnection) return new DefaultHighlighter((PicNodeConnection)element);
		else if (element instanceof BranchElement) return new CompositeHighlighter((BranchElement)element);
		else return new DefaultHighlighter(element);
	}

	/**
	 * Returns a Stroke object suited for painting hightlighter's stroke, e.g. tangents.
	 * @param scale The current scale factor from model to screen for the Graphics2D context ;
	 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
	 *        same length on the screen whatever the scale factor that's set to the graphic context.
	 */
	protected Stroke createStroke(double scale){
		//return new BasicStroke((float)(1.0/scale));
		final float[] dotArray = {(float)(5.0/scale),(float)(5.0/scale)};
		return new BasicStroke((float)(1.0/scale),BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, dotArray, 0);
	}

	/////////////////////////////////////////////////////////////////////////
	//// Default Highlighter
	//////////////////////////////////////////////////////////////////////////
	/**
	 * Default implementation of the Highlighter interface.
	 */
	public class DefaultHighlighter implements Highlighter {

		/** highlighted Element */
		protected Element element;

		/** iterator over visible control-points */
		protected PointIndexIterator pointIndexIterator;
		
		/** highlighter shape ; should be null if N/A in subclass */
		protected Shape shape;
		
		/** cache for highlighter bounds */
		protected Rectangle2D.Double bounds = new Rectangle2D.Double();
		
		// ====== protected buffers used by performance critical methods ========
		/** buffer used by paint only ; may be safely used by subclasses as well */
		protected Rectangle2D rectBuffer = new Rectangle2D.Double();
		/** buffer used by paint only ; may be safely used by subclasses as well */
		protected PicPoint ptBuffer = new PicPoint();

		private ArrayList hitIndices = new ArrayList(); // used by hitTest (avoid re-creating arrays at each invokation)


		/**
		 * @param element the Element to be highlighted
		 */
		public DefaultHighlighter(Element element){
			this.element = element;
		}
		
		/** returns the higlighted Element */
		public Element getElement(){
			return element;
		}
		
		/**
		 * Fetches the HighlighterFactory implementation that is feeding the view hierarchy. 
		 * @return the factory, null if none
		 */
		public HighlighterFactory getHighlighterFactory(){
			return DefaultHighlighterFactory.this;
		}
		
		//////////////////////////////////////////////
		/// EVENT HANDLING
		/////////////////////////////////////////////

		/**
		 * Give notification from the model that a change occured for an element's highlighting
		 * this highlighter is responsible for rendering. 
		 * To reduce the burden for subclasses, this implemention dispatches to the following methods,
		 * in that order, depending on the value of "eventType" :
		 * <ul>
		 * <li> <code>syncShape()</code> if it's a DrawingEvent.GEOMETRY_CHANGE
		 * <li> <code>syncBounds()</code> if it's a DrawingEvent.GEOMETRY_CHANGE
		 * </ul>
		 */
		public void changedUpdate(DrawingEvent.EventType eventType, double scale){
			if (eventType==null || eventType!=DrawingEvent.EventType.ATTRIBUTE_CHANGE) {
				syncShape(scale);
				syncBounds(scale);
			}
		}

		/**
		 * Synchronizes the highlighter's bounding rectangle (aka clip) with the model ; 
		 * "bounds" is first computed from the set of visible control-points,
		 * then its size gets increased by BARBELL_SIZE. Finally, is shape is not null,
		 * its gets added to the bounding rectangle.
		 * @param scale The current scale factor from-model-to-screen for the Graphics2D context ;
		 *        this may be used to scale down line thickess, etc... so that e.g. barbells appear with the
		 *        same size on the screen whatever the scale factor being set to the graphic context.
		 */
		protected void syncBounds(double scale){
			// 1) compute rectangle containing all visible control-points :
			boolean inited = false;
			PointIndexIterator it = getControlPointsIterator();
			for(it.reset(); it.hasNext();){
				int ptIndex = it.next();
				PicPoint pt = element.getPoint(ptIndex, ptBuffer); // store in ptBuffer
				if (!inited) {
					bounds.setRect(pt.x, pt.y, 0,0);
					inited=true;
				}
				else {
					bounds.add(pt); 
				}
			}

			// 3) add shape if applicable :
			if (shape !=null) {
				if (!inited) {
					bounds.setRect(shape.getBounds2D()); // ie if no control-points
					inited=true;
				}
				else bounds.add(shape.getBounds2D());
			}

			// 2) enlarge by BARBELL_SIZE :
			if (inited){
				double barbell_size = 2.0*BARBELL_SIZE/scale; // x 1.2 : good guess so that small smudges  
				bounds.setFrame(
			   		bounds.getX()-barbell_size, bounds.getY()-barbell_size,
					bounds.getWidth()+2*barbell_size, bounds.getHeight()+2*barbell_size);
			}
		}
		
		/**
		 * Synchronize the shape needed to paint this highlighter, with the model ; 
		 * This implementation does nothing.
		 * @param scale The current scale factor from-model-to-screen for the Graphics2D context ;
		 *        this may be used to scale down line thickess, etc... so that e.g. barbells appear with the
		 *        same size on the screen whatever the scale factor being set to the graphic context.
		 */
		protected void syncShape(double scale){
		}
		
		////////////////////////////////////////////////////////
		//// PAINT
		////////////////////////////////////////////////////////

		/**
		 * Render the Highlighter to the given graphic context.<br>
		 * Current implementation paints end-points by delegating to a 
		 * {@link jpicedt.graphic.model.DefaultPointIndexIterator DefaultPointIndexIterator}, 
		 * using the highlighter color if the given
		 * allocation intersects the bounds of this view,
		 * then iterates over the <code>shapes</code> array, and paints any non-null element.
		 * @param scale The current scale factor from-model-to-screen for the Graphics2D context ;
		 *        this may be used to scale down line thickess, etc... so that e.g. barbells appear with the
		 *        same size on the screen whatever the scale factor being set to the graphic context.
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			//jpicedt.Log.debug("allocation="+allocation);
			if (!allocation.intersects(getBounds())) return;
			
			// compute barbell size for the current zoom factor [pending] cache ?
			double barbellSize = BARBELL_SIZE/scale;

			// 1) paint highlighter shape if applicable :
			g.setStroke(createStroke(scale));
			if (shape!=null) g.draw(shape);

			// 2) paint end-points for the given object
			PointIndexIterator it = getControlPointsIterator();
			for(it.reset(); it.hasNext();){
				int ptIndex = it.next();
				//jpicedt.Log.debug("ptIndex="+ptIndex);
				PicPoint pt = element.getPoint(ptIndex, ptBuffer);
				rectBuffer.setRect(pt.x-barbellSize,pt.y-barbellSize,2*barbellSize,2*barbellSize);
				g.fill(rectBuffer);
				if (jpicedt.Log.DEBUG){
					Paint oldCol = g.getPaint();
					Font font = new Font("SansSerif", Font.PLAIN, 6).deriveFont(AffineTransform.getScaleInstance(0.5,-0.5));
					FontRenderContext frc = g.getFontRenderContext();
					TextLayout tl = new TextLayout(new Integer(ptIndex).toString(), font, frc);
					g.setPaint(Color.black);
					tl.draw(g,(float)pt.x,(float)pt.y);
					g.setPaint(oldCol);
				}
			}
		}

		/**
		 * Returns an iterator over control-points that should be displayed by the hightligher.
		 * This default implementation iterates over all control-points.
		 */
		protected PointIndexIterator getControlPointsIterator(){
			// lasily allocate new Iterator :
			if (this.pointIndexIterator==null) this.pointIndexIterator = new DefaultPointIndexIterator(element);
			return this.pointIndexIterator;
		}

		/**
		 * @return the bounds of this Highlighter<br> 
		 *         This will determine the clipping rectangle passed as a parameter to repaint
		 *         in the hosting view
		 */
		public Rectangle2D getBounds(){
			return bounds;
		}




		/////////////////////////////////////////////////////
		//// MOUSE
		/////////////////////////////////////////////////////

		/**
		 * Current implementation returns a HitInfo.Point if a click
		 * occured on one of the controlled-point indices returned by 
		 * {@link #getControlPointsIterator getControlPointsIterator()} ; return null otherwise. 
		 * @return a HitInfo corresponding to the given mouse-event
		 */
		public HitInfo hitTest(PEMouseEvent e){
			
			PicPoint ptClick = e.getPicPoint(); // in model coords

			// 1) test if click occured outside the view bounds, which can speed up a lot...
			if (!getBounds().contains(ptClick)) return null;

			// 2) test click on control-points :
			// the true maximum distance for a click to be considered as a hit must be scaled down
			// by the current zoom factor, so that zooming in allows for a greater precision in the
			// selection/deselection of graphical elements :
			double click_distance = CLICK_DISTANCE / e.getCanvas().getZoomFactor();
			hitIndices.clear();
			PointIndexIterator it = getControlPointsIterator();
			for(it.reset(); it.hasNext();){
				int ptIndex = it.next();
				//jpicedt.Log.debug("ptIndex="+ptIndex);
				element.getPoint(ptIndex, ptBuffer); // dest = ptBuffer
				if (ptBuffer.distance(ptClick) <= click_distance) hitIndices.add(new Integer(ptIndex));
			}
			if (!hitIndices.isEmpty()) return new HitInfo.Point(element, hitIndices, e);
			
			// 3) test click on hightlighter's shape, if applicable :
			if (shape != null){
				int segmentIndex = PEToolKit.testDistanceToPath(shape, ptClick, click_distance);
				if (segmentIndex >= 0) return new HitInfo.HighlighterStroke(element,segmentIndex,e);
			}

			// 4) no match
			return null;

		}
	}


	//////////////////////////////////////////////////////////////////////////
	//// ELLIPSE HIGHLIGHTER
	//////////////////////////////////////////////////////////////////////////
	
	/**
	 * a Highlighter for PicEllipse's. This comprises control-points and the
	 * surrounding parallelogram.
	 */
	public class EllipseHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {

		/**
		 * construct a new Highlighter for the given ellipse
		 */
		public EllipseHighlighter(PicEllipse ellipse){
			super(ellipse);
		}

		/**
		 * Synchronize the "shape" variable with the model.
		 */
		protected void syncShape(double scale){

			if (shape==null) shape = new GeneralPath();
			GeneralPath parallelo = (GeneralPath)shape;
			PicEllipse ellipse = (PicEllipse)element;
			parallelo.reset(); 
			parallelo.moveTo((float)ellipse.getPointX(0), (float)ellipse.getPointY(0)); // set first point
			parallelo.lineTo((float)ellipse.getPointX(1), (float)ellipse.getPointY(1));
			parallelo.lineTo((float)ellipse.getPointX(2), (float)ellipse.getPointY(2));
			parallelo.lineTo((float)ellipse.getPointX(3), (float)ellipse.getPointY(3));
			parallelo.closePath();

		}

		/**
		 * Render the Highlighter to the given graphic context.<br>
		 * Current implementation paints end-points using the highlighter color if the given
		 * allocation intersects the bounds of this view, then paint the surrounding parallelogram.
		 * @param scale The current scale factor from model to screen for the Graphics2D context ;
		 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
		 *        same length on the screen whatever the scale factor that's set to the graphic context.
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			super.paint(g, allocation, scale);
			// paint SIDE_R control-point to help user know where angle starts
			double barbellSize = BARBELL_SIZE/scale/2.0;
			PicPoint pt = element.getPoint(PicEllipse.SIDE_R,ptBuffer); // use dedicated buffers from AbstractHighlighter
			rectBuffer.setRect(pt.x-barbellSize,pt.y-barbellSize,2*barbellSize,2*barbellSize);
			g.setPaint(Color.black);
			g.fill(rectBuffer);
		}


	} // EllipseHighlighter

	//////////////////////////////////////////////////////////////////////////
	//// CIRCLE HIGHLIGHTER
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a Highlighter for rendering circles. This comprises the three control-points
	 * that make sense with PicCircleFrom3Points.
	 */
	public class CircleHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {

		/**
		 * construct a new Highlighter for the given ellipse
		 */
		public CircleHighlighter(PicCircleFrom3Points circle){
			super(circle);
		}
		/**
		 * Returns an iterator over control-points that should be displayed by the hightligher.
		 * This default implementation iterates over all control-points.
		 */
		protected PointIndexIterator getControlPointsIterator(){
			// lasily allocate new Iterator :
			if (this.pointIndexIterator==null) this.pointIndexIterator = new CircleControlPointsIterator();
			return this.pointIndexIterator;
		}

		/** iterates over a subset of PicCircleFrom3Points control-points so as to reduce the number of visible control-points */
		class CircleControlPointsIterator implements PointIndexIterator {

			protected int counter;
			protected int[] idx = {PicCircleFrom3Points.P_CENTER,PicCircleFrom3Points.P_1,PicCircleFrom3Points.P_2,PicCircleFrom3Points.P_3};

			public CircleControlPointsIterator(){
				counter = 0;
			}

			public boolean hasNext(){
				if (counter >= idx.length) return false;
				return true;
			}

			public int next(){
				int i = idx[counter];
				counter++;
				return i;
			}

			public void reset(){
				counter = 0;
			}
		}
	} // CircleHighlighter

	//////////////////////////////////////////////
	/// ABSTRACT CURVE HIGHLIGHTER
	/////////////////////////////////////////////

	/**
	 * a Highlighter for any concrete implementation of jpicedt.graphic.model.AbstractCurve.
	 * This comprises tangents for non-straight segments, and is wrapped into a GeneralPath.
	 */
	public class AbstractCurveHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {

		/**
		 * construct a new Highlighter for the given curve
		 */
		public AbstractCurveHighlighter(AbstractCurve curve){
			super(curve);
		}

		/**
		 * Synchronizes the highlighter's tangents with the model. 
		 */
		protected void syncShape(double scale){
			
			if (shape==null) shape = new GeneralPath();	
			GeneralPath path = (GeneralPath)shape;			// compute number of non-straight segments
			path.reset();
			
			AbstractCurve curve = (AbstractCurve)element;
			
			for (int i=0; i<curve.getNumberOfSegments();i=i+1){
				if (!curve.isStraight(i)){
					path.moveTo((float)curve.getSpecificationPointX(3*i), (float)curve.getSpecificationPointY(3*i));
					path.lineTo((float)curve.getSpecificationPointX(3*i+1), (float)curve.getSpecificationPointY(3*i+1));
					path.moveTo((float)curve.getSpecificationPointX(3*i+2), (float)curve.getSpecificationPointY(3*i+2));
					path.lineTo((float)curve.getSpecificationPointX(3*i+3), (float)curve.getSpecificationPointY(3*i+3));
				}
			}
		}


		/**
		 * Render the Highlighter to the given graphic context.<br>
		 * Current implementation paints end-points using the highlighter color if the given
		 * allocation intersects the bounds of this view, then renders tangents,
		 * finally paint first subdivision-point in black to help user know where closed-curve
		 * start exactly.
		 * @param scale The current scale factor from model to screen for the Graphics2D context ;
		 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
		 *        same length on the screen whatever the scale factor that's set to the graphic context.
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			super.paint(g, allocation, scale);
			
			// paint first control-point to help user know where curve start (useful for closed curves)
			double barbellSize = BARBELL_SIZE/scale/2.0;
			PicPoint pt = element.getPoint(0,ptBuffer);
			rectBuffer.setRect(pt.x-barbellSize,pt.y-barbellSize,2*barbellSize,2*barbellSize);
			g.setPaint(Color.black);
			g.fill(rectBuffer);
		}

	} // AbstractCurveHighlighter

	//////////////////////////////////////////////
	/// SMOOTH POLYGON VIEW
	/////////////////////////////////////////////
	/**
	 * a Highlighter for SmoothPolygon. So far, it is based on AbstractCurveHighlighter, except where computing
	 * tangents is concerned, since for SmoothPolygon's, tangents are specified by polygon-point's rather than
	 * Bezier-points.
	 */
	public class SmoothPolygonHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {
		
		// shapes[0]=GeneralPath (polygon) + closing tangent if applicable

		/**
		 * construct a new Highlighter for the given smooth-polygon
		 */
		public SmoothPolygonHighlighter(PicSmoothPolygon curve){
			super(curve);
		}

		/**
		 * Synchronizes the highlighter's tangents with the model. This method is called from inside
		 * <code>syncShape()</code> only, and is overriden so as to compute tangents based on polygon's points
		 * rather than Bezier-points. Also compute the ``envelope'' polygon which is used for adding a new point
		 * (hitTest method).
		 */
		protected void syncShape(double scale){
			PicSmoothPolygon poly = (PicSmoothPolygon)element;

			if (shape==null) shape=new GeneralPath();
			GeneralPath envelope = (GeneralPath)shape;
			envelope.reset();
			envelope.moveTo((float)poly.getPointX(poly.getFirstPointIndex()),(float)poly.getPointY(poly.getFirstPointIndex()));

			for (int i=poly.getFirstPointIndex(); i<poly.getLastPointIndex();i=i+1){
				envelope.lineTo((float)poly.getPointX(i+1),(float)poly.getPointY(i+1));
			}
			if (poly.isClosed()){
				envelope.closePath();
			}
		}
	} // SmoothPolygonHighlighter

	//////////////////////////////////////////////
	/// PS CURVE highlighter
	/////////////////////////////////////////////
	/**
	 * a Highlighter for PsCurve. So far, it is based on AbstractCurveHighlighter, except where computing
	 * tangents is concerned: we draw 2 tangents at the endpoints of an open pscurve, no tangent otherwise.
	 */
	public class PsCurveHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {
		
		// shapes[0]=GeneralPath (polygon) + closing tangent if applicable

		/**
		 * construct a new Highlighter for the given pscurve
		 */
		public PsCurveHighlighter(PicPsCurve curve){
			super(curve);
		}

		/**
		 * Synchronizes the highlighter's tangents with the model. 
		 */
		protected void syncShape(double scale){
			
			if (shape==null) shape = new GeneralPath();	
			GeneralPath path = (GeneralPath)shape;			// compute number of non-straight segments
			path.reset();
			
			PicPsCurve curve = (PicPsCurve)element;
			
			if (curve.isClosed()) return;
			PicPoint[] pts=curve.getInitialControlCurve();// 4 points of control curve
			if (pts.length !=0 ){
			    path.moveTo((float)pts[0].getX(), (float)pts[0].getY());
			    path.curveTo((float)pts[1].getX(), (float)pts[1].getY(),
					 (float)pts[2].getX(), (float)pts[2].getY(),
					 (float)pts[3].getX(), (float)pts[3].getY());
			}
			pts=curve.getFinalControlCurve();// 4 points of control curve
			if (pts.length !=0 ){
			    path.moveTo((float)pts[0].getX(), (float)pts[0].getY());
			    path.curveTo((float)pts[1].getX(), (float)pts[1].getY(),
					 (float)pts[2].getX(), (float)pts[2].getY(),
					 (float)pts[3].getX(), (float)pts[3].getY());
			}

		}
			

	} // PsCurveHighlighter


	//////////////////////////////////////////////////////////////////////////
	//// TEXTHIGHLIGHTER
	//////////////////////////////////////////////////////////////////////////

	
	
	/**
	 * a Highlighter for TextEditable's. [SR:underway] use TextLayout highlighting facilities ; this Highlighter is not used yet. 
	 */
	public class TextHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {

		/**
		 * construct a new Highlighter for the given TextEditable
		 */
		public TextHighlighter(TextEditable text){
			super(text);
		}
		
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			//jpicedt.Log.debug("allocation="+allocation);
			super.paint(g,allocation,scale);
		}
		
	}
	////////////////////////////////////////////////////////
	//// COMPOSITE VIEW
	////////////////////////////////////////////////////////
	/**
	 * an Highlighter for a BranchElement
	 */
	public class CompositeHighlighter extends DefaultHighlighterFactory.DefaultHighlighter {

		/** 
		 * Highlighting mode corresponding to each Element belonging to the selection is highlighted 
		 */
		public static final String LOCAL_MODE = "Local";
		/** 
		 * Highlighting mode corresponding to
		 * the selection-handler behaving as if it were a PicGroup, 
		 * hence its 8 control-points only are highlighted 
		 */
		public static final String GLOBAL_MODE = "Global";
	
		/** highlight mode */
		protected String highlightingMode=GLOBAL_MODE;
		
		/**
		 * construct a new Highlighter for the given PicGroup
		 */
		public CompositeHighlighter(BranchElement e){
			super(e);
		}

		/**
		 * set the current highlighting mode ; 
		 * this influences the way the Highligther is painted, but also the result returned by HitTest.
		 * @param mode or LOCAL_MODE or GLOBAL_MODE
		 */
		public void setHighlightingMode(String mode){
			 if (highlightingMode == mode) return;
			 highlightingMode = mode;
			 View v = element.getView();
			 if (v!=null) v.changedUpdate(null);
		 }

		/**
		 * Return the current highlighting mode
		 */
		public String getHighlightingMode(){
			return highlightingMode;
		}

		/**
		 * Toggle the current highlighting mode
		 */
		public void toggleHighlightingMode(){
			if (highlightingMode.equals(LOCAL_MODE)) setHighlightingMode(GLOBAL_MODE);
			else setHighlightingMode(LOCAL_MODE);
		}
		
		/**
		 * Synchronize the array of "shapes" needed to paint this highlighter, with the model ; 
		 */
		protected void syncShape(double scale){
			/*if (shape==null) shape = new Rectangle2D.Double();
			Rectangle2D r = (Rectangle2D)shape;
			r = element.getBoundingBox(r);
			*/
			if (shape==null) shape = element.getBoundingBox(null);
			else shape = element.getBoundingBox((Rectangle2D)shape);
		}

		/**
		 * Render the Highlighter to the given graphic context.<br>
		 */
		public void paint(Graphics2D g, Rectangle2D allocation, double scale){
			if (!allocation.intersects(getBounds())) return;
			if (highlightingMode==GLOBAL_MODE){ // paint only group's end-points
				BranchElement be = (BranchElement)element;
				g.setPaint(GLOBAL_HIGHLIGHTING_COLOR);
				if (!be.isEmpty()) super.paint(g,allocation,scale);
			}
			else if (highlightingMode==LOCAL_MODE){ // paint children's highlighter.
				BranchElement be = (BranchElement)element;
				for(Iterator it=be.children(); it.hasNext();){
					Element e = (Element)it.next();
					View v = e.getView();
					g.setPaint(LOCAL_HIGHLIGHTING_COLOR);
					if (v!=null) v.paintHighlighter(g, allocation, scale);
				}
			}
			
		}

		/**
		 * @return a HitInfo corresponding to the given mouse-event
		 */
		public HitInfo hitTest(PEMouseEvent e){

			if (highlightingMode==GLOBAL_MODE){ 
				return super.hitTest(e);// test hit on group's end-points + highlighter shape
			}
			
			else if (highlightingMode==LOCAL_MODE) { // test hit on children, highlighted, from top to bottom :
				BranchElement be = (BranchElement)element;
				for(int i = be.getChildCount()-1; i>=0; i--){
					Element o = be.getChildAt(i);
					HitInfo hit = o.getView().hitTest(e,true); // [SR:pending] change to o.getView().getHighlighter().hitTest(e) ?
					if (hit != null) return hit;
				}
				return null;
			}
			return null;
		}
		
	} // CompositeHighlighter
	
} // DefaultHighlighterFactory

