/*
 DefaultViewFactory.java - February 9, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Departement de Physique
 Ecole Nationale Superieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.*; // all elements + drawing
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.PicPoint;
import jpicedt.graphic.PicVector;
import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;
import jpicedt.graphic.ContentType;
import jpicedt.ui.util.RunExternalCommand;

import java.awt.*;
import java.awt.font.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.util.*;
import java.util.regex.*;
import java.io.*;
import javax.swing.*;
import javax.imageio.*;

/**
 * This is the default factory used to create a View for a given Element ; should be used when
 * one wants to render all Element's attributes ; otherwise subclass with your own ViewFactory
 * and override :
 * <ul>
 * <li> createView : simply return your own View for some or all Element's (then call super.createView)
 * <li> factory methods, e.g. createStroke, createPaintForInterior, createPaintForOutline, ... : 
 *     these methods are called by various
 *     View's to extract rendering information from PicAttributeSet's when common behaviour 
 *     is expected, whatever the particular Element being actually rendered.
 * </ul> 
 * @since jPicEdt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: DefaultViewFactory.java,v 1.47.2.2 2007/09/02 11:56:29 reynal Exp $
 */
public class DefaultViewFactory implements ViewFactory, PicObjectConstants {

	// --- experimental support for bitmap images of PicText's ---
    //private static final boolean isExperimentalSupportForBitmap=true; // set to false to get the old behaviour
    private static final boolean isExperimentalSupportForBitmap=false; // set to false to get the old behaviour

	/** default font used to display text with TextView */
	public static Font TEXT_FONT_DEFAULT = new Font("SansSerif", Font.PLAIN, 10);
	/** key used to fetch the font family value from a Properties object */
	public static String KEY_TEXT_FONT = "view.text-font";
	/** scaling factor to be applied to the default font before rendering, so as to compensate for zoom factor */
	protected static final float TEXT_FONT_SCALE=4.0f;
	/** font used to display text strings with TextView */
	protected static Font textFont = TEXT_FONT_DEFAULT.deriveFont((float)TEXT_FONT_DEFAULT.getSize()/TEXT_FONT_SCALE);
        /** pattern used for parsing log file */
    	private static final Pattern LogFilePattern = Pattern.compile("JPICEDT INFO:\\s*([0-9.]*)pt,\\s*([0-9.]*)pt,\\s*([0-9.]*)pt");

    
    private static final String CR_LF = System.getProperty("line.separator");


	/**
	 * for debugging purpose
	 */
	public static void main(String[] arg){
		DefaultViewFactory f = new DefaultViewFactory();
		PicAttributeSet set = new PicAttributeSet();
		//PicRectangle elem = new PicRectangle(new PicPoint(0.0,10.0),new PicPoint(20.0,40.0),set);
		//View v = f.createView(elem);
	}

	// [pending] use constructor to init global parameters (text rendering,...)
	public DefaultViewFactory(){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("Initializing");
	}

	/**
	 * Configure static fields using the given Properties object
	 * @param properties used to read shared parameters
	 *        If null, default values are used.
	 */
	public static void configure(Properties preferences){
		String val = preferences.getProperty(KEY_TEXT_FONT);
		if (val==null) textFont = TEXT_FONT_DEFAULT;
		else textFont = Font.decode(val);
		textFont = textFont.deriveFont((float)textFont.getSize()/TEXT_FONT_SCALE);
	}


	/**
	 * @return a View for the given Element
	 * Current implementation returns a View appropriate for the following elements :
	 * <ul>
	 * <li> PicEllipse and any subclass -> EllipseView ;
	 * <li> BranchElement -> CompositeView ;
	 * <li> PicParallelogram -> ParallelogramView ;
	 * <li> PicEllipse / PicCircleFrom3Points -> EllipseView ;
	 * <li> PicText -> TextView ;
	 * <li> other -> null (this give a change for the receiver to provide its own View)
	 * </ul>
	 */
	public View createView(Element element){
		if (jpicedt.Log.DEBUG) jpicedt.Log.debug("element=" + element);
		// always place daughters BEFORE superclass :
		if (element instanceof PicEllipse) return new DefaultViewFactory.EllipseView((PicEllipse)element);
		else if (element instanceof PicParallelogram) return new DefaultViewFactory.ParallelogramView((PicParallelogram)element);
		else if (element instanceof AbstractCurve) return new DefaultViewFactory.AbstractCurveView((AbstractCurve)element);
		else if (element instanceof PicText) {
			if (isExperimentalSupportForBitmap) return new DefaultViewFactory.TextViewHybrid((PicText)element);
			return new DefaultViewFactory.TextView((PicText)element);
		}
		else if (element instanceof PicNodeConnection) return new DefaultViewFactory.PicNodeConnectionView((PicNodeConnection)element);
		else if (element instanceof BranchElement) return new CompositeView((BranchElement)element);
		else {
			if (jpicedt.Log.DEBUG) jpicedt.Log.debug("No view available for the given element:"+element);
			return null;
		}
	}

	/**
	 * @return a Stroke built from the given attributes ; null if LINE_STYLE=NONE. 
	 *         All currently supported attributes for stroke are rendered.
	 */
	protected Stroke createStroke(PicAttributeSet set){
		//if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".createStroke");

		if(set.getAttribute(LINE_STYLE)==SOLID)
			return new BasicStroke((float)StyleConstants.getLineWidth(set));

		if(set.getAttribute(LINE_STYLE)==DASHED){
			// make sure dashes > 0
			float dashOpaque = (float)StyleConstants.getDashOpaque(set);
			if (dashOpaque>0){
				float dashTransparent = (float)StyleConstants.getDashTransparent(set);
				if (dashTransparent>0){
					return new BasicStroke((float)StyleConstants.getLineWidth(set),
					                       BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f,
					                       new float[]{dashOpaque, dashTransparent}, 0);
				}
			}
			// otherwise, paint as SOLID :
			return new BasicStroke((float)StyleConstants.getLineWidth(set));
		}

		if(set.getAttribute(LINE_STYLE)==DOTTED){
			float dotSep = (float)StyleConstants.getDotSep(set);
			// make sur dotsep > 0
			if (dotSep>0){
				float[] dotArray = {
				                       (float)(dotSep * 0.1),  // opaque 10%
				                       (float)(dotSep * 0.9)}; // transparent 90%
				return new BasicStroke((float)StyleConstants.getLineWidth(set),
				                       BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, dotArray, 0);
			}
			// otherwise paint as SOLID
			else {
				return new BasicStroke((float)StyleConstants.getLineWidth(set));
			}
		}
		else { // LINE_STYLE=NONE
			return null;
		}
	}

	/**
	 * Returns a Paint object from the given attributes, that is suited for painting an outline,
	 * or null if LINE_STYLE is NONE.
	 * <p>
	 * All currently supported attributes for painting an outline are rendered. It's up to subclassers
	 * to adapt this behavior to their need (probably by restricting it).
	 */
	protected Paint createPaintForOutline(PicAttributeSet set){
		//if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".createPaintForOutline");
		if (StyleConstants.getLineStyle(set)!=NONE) return StyleConstants.getLineColor(set);
		else return null;
	}

	/**
	 * Returns a Paint object from the given attributes, that is suited for painting the interior of a shape.
	 * <p>
	 * The returned object depends on the FILL_STYLE and FILL_COLOR attributes of the given set:
	 * <p>
	 * <ul>
	 * <li> FILL_STYLE = NONE: returns null
	 * <li> FILL_STYLE = SOLID, VLINES, VLINES_FILLED, HLINES, HLINES_FILLED, CROSSHATCH and CROSSHATCH_FILLED:
	 *      returns a Color object created from FILL_COLOR.
	 * </ul>
	 * To sum up, all currently supported attributes (as defined in 
	 * <code>jpicedt.graphic.model.PicObjectConstants</code>) for painting the interior of shapes 
	 * are rendered. [pending] gradient (need to add some attributes).
	 */
	protected Paint createPaintForInterior(PicAttributeSet set){
		//if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".createPaintForInterior");
		String fillStyle = StyleConstants.getFillStyle(set);
		if (fillStyle==NONE) return null;
		else if (fillStyle==SOLID || fillStyle==HLINES_FILLED || fillStyle==VLINES_FILLED || fillStyle==CROSSHATCH_FILLED)
			return StyleConstants.getFillColor(set);
		else if (StyleConstants.isShadow(set)) return StyleConstants.getFillColor(set); // if shadowing is on, fill the shape whatever the "fillstyle" parameter.
		// [SR:pending] pending patch by Herbert Voss, see the mail from Vincent 17/04/2004.
		else return null; // [pending] gradient
		// rem : hatches are handled by "paintHatches". Using texturePaint just barfs totally because
		// the BufferedImage used for the texture is scaled afterwards by the current Graphics2D scaler.
	}

	/**
	 * paint hatches (vlines,...) depending on the FILL_STYLE attributes of the given set.
	 * Factories that don't paint hatches should override this method to do nothing.
	 * @param shape used to clip the hatch (ie the shape the calling view must render)
	 */
	protected void paintHatches(Graphics2D g, PicAttributeSet set, Shape shape){

		// [pending] bug when angle > 0 (lines are not parallel)
		String fillStyle = StyleConstants.getFillStyle(set);
		double angle=0;
		if (fillStyle == VLINES || fillStyle == VLINES_FILLED) {
			angle = StyleConstants.getHatchAngle(set);
			Shape oldClip = g.getClip(); // save old clip before we add our own clip
			g.clip(shape); // [pending] apparently, clipping dramatically decreases performances
			_paintHatches(g,set,angle,shape.getBounds2D());
			g.setClip(oldClip); // restore old clip
		}
		else if (fillStyle == HLINES || fillStyle == HLINES_FILLED) {
			angle = StyleConstants.getHatchAngle(set);
			if (angle>0) angle -= 90;
			else angle += 90;
			Shape oldClip = g.getClip(); // save old clip before we add our own clip
			g.clip(shape);
			_paintHatches(g,set,angle,shape.getBounds2D());
			g.setClip(oldClip); // restore old clip
		}
		else if (fillStyle == CROSSHATCH || fillStyle == CROSSHATCH_FILLED) {
			angle = StyleConstants.getHatchAngle(set);
			Shape oldClip = g.getClip(); // save old clip before we add our own clip
			g.clip(shape);
			_paintHatches(g,set,angle,shape.getBounds2D());
			if (angle>0) angle -= 90;
			else angle += 90;
			_paintHatches(g,set,angle,shape.getBounds2D());
			g.setClip(oldClip); // restore old clip
		}
		else return;

	}

	/**
	 * Helper method for paintHatches().
	 * @param angle angle (in degrees) b/w lines and the vertical axis, as in VLINES (positive CCW)
	 * @param clip the bounding rectangle of the shape
	 */
	private void _paintHatches(Graphics2D g, PicAttributeSet set, double angle, Rectangle2D clip){

		g.setStroke(new BasicStroke((float)StyleConstants.getHatchWidth(set)));
		g.setPaint(StyleConstants.getHatchColor(set));
		double sep = StyleConstants.getHatchSep(set);
		Line2D.Double hatch = new Line2D.Double();

		if (angle==0){ // vertical
			hatch.y1 = clip.getMinY();
			hatch.y2 = clip.getMaxY();
			final double maxX=clip.getMaxX();
			for (double x = clip.getMinX(); x<maxX;x += sep){
				hatch.x1 = hatch.x2 = x;
				g.draw(hatch);
			}
		}
		else if (angle==90){ // horizontal
			hatch.x1 = clip.getMinX();
			hatch.x2 = clip.getMaxX();
			final double maxY=clip.getMaxY();
			for (double y = clip.getMinY(); y<maxY;y += sep){
				hatch.y1 = hatch.y2 = y;
				g.draw(hatch);
			}
		}
		else { // others
			double incX = sep/Math.cos(Math.toRadians(angle)); // >0
			double incY = sep/Math.sin(Math.toRadians(angle)); //
			final double maxX;
			if (angle>0) {
				hatch.x1 = clip.getMinX();
				hatch.y1 = clip.getMinY();
				maxX = clip.getMaxX() + clip.getMaxY() * Math.tan(Math.toRadians(angle));
			}
			else {
				hatch.x1 = clip.getMinX();
				hatch.y1 = clip.getMaxY();
				maxX = clip.getMaxX() - clip.getMaxY() * Math.tan(Math.toRadians(angle));
			}
			hatch.x2 = hatch.x1;
			hatch.y2 = hatch.y1;
			while (hatch.x2 < maxX){
				hatch.x2 += incX;
				hatch.y1 += incY;
				//System.out.println("x1="+hatch.x1+"\ty1="+hatch.y1+"\tx2="+hatch.x2+"\ty2="+hatch.y2);
				g.draw(hatch);
			}
		}
	}

	/**
	 * paint shadow depending on the SHADOW attribute and rel.
	 * Factory that don't paint shadow should override to do nothing.
	 * @param shape the Shape under which to drop a shadow
	 * @param set used to fetch the shadow attributes
	 */
	protected void paintShadow(Graphics2D g, PicAttributeSet set, Shape shape){

		if (!StyleConstants.isShadow(set)) return;
		// create shadow shape
		double angle = Math.toRadians(StyleConstants.getShadowAngle(set));
		double sep = StyleConstants.getShadowSize(set);
		AffineTransform at = AffineTransform.getTranslateInstance(sep * Math.cos(angle), sep * Math.sin(angle));
		Shape shadow = at.createTransformedShape(shape);
		g.setPaint(StyleConstants.getShadowColor(set));
		// if shape is not filled, the shadow isn't filled as well !
		if (StyleConstants.getFillStyle(set) == PicObjectConstants.NONE) g.draw(shadow);
		else g.fill(shadow);
	}

	/**
	 * paint overstrike depending on the OVER_STRIKE attribute and rel.
	 * Factory that don't overstrike should override to do nothing.
	 * @param shape the Shape to overstrike
	 * @param set used to fetch the overstrike attributes
	 */
	protected void paintOverStrike(Graphics2D g, PicAttributeSet set, Shape shape){

		if (!StyleConstants.isOverStrike(set)) return;
		// create overStrike
		g.setPaint(Color.white);
		BasicStroke overStrikeStroke = new BasicStroke((float)(StyleConstants.getLineWidth(set)+2*StyleConstants.getOverStrikeWidth(set)),
		                               BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER);
		g.setStroke(overStrikeStroke);
		g.draw(shape);
	}

	/**
	 * @return an Arrow from the given attribute set, or null if Arrow.NONE.
	 * @param direction LEFT_ARROW or RIGHT_ARROW
	 */
	protected Arrow createArrow(PicAttributeSet set, PicAttributeName direction){
		//if (jpicedt.Log.DEBUG) jpicedt.Log.debug(".createArrow");
		Arrow a = (Arrow)set.getAttribute(direction);
		if (a==Arrow.NONE) return null;
		a = a.copy();
		a.syncAttributes(set); // whatever the direction (update arrow parameters from the attribute set)
		return a;
	}

	/////////////////////////////////////////////////////////////////////////
	//// LEAF ELEMENT VIEW (abstract)
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering leaf-elements ; implements attributes caching, and shape rendering.
	 * Subclasses might just override the <code>changedUpdate</coe> method, 
	 * and update the <code>shape</code> variable accordingly.  
	 *
	 */
	protected abstract class LeafElementView extends AbstractView {

		/** shape to be drawn ; subclass must update it properly */
		protected Shape shape;
		/** line width to be used to compute the view bounds (cached) */
		protected double lineWidth = 0.1;
		/** overstrike width */
		protected double overStrikeWidth = 0;
		/** stroke for outline (cached) */
		protected Stroke outlineStroke;
		/** paint for interior (cached) */
		protected Paint interiorPaint;
		/** paint for outline (cached) */
		protected Paint outlinePaint;
		/** buffer */
		protected Rectangle2D.Double shadowBounds = new Rectangle2D.Double();

		/**
		 * construct a new View for the given Element 
		 */
		public LeafElementView(Element e){
			super(e);
		}

		/**
		 * Fetches the ViewFactory implementation that is feeding the view hierarchy.
		 * This implementation first try to fetch a ViewFactory from the parent view,
		 * in case the root-view was produced from another ViewFactory (which for
		 * instance partially delegates to this factory),
		 * then, if there's no parent view, return the enclosing class.
		 * @return the factory, null if none
		 */
		public ViewFactory getViewFactory() {
			ViewFactory f = super.getViewFactory(); // try to fetch the factory from the parent view
			if (f!=null) return f;
			else return DefaultViewFactory.this;
		}

		/**
		 * Give notification from the model that a change occured for an element this view is responsible
		 * for rendering.<p>
		 * To reduce the burden for subclasses, this implemention dispatches to the following methods,
		 * in that order, depending on the value of "eventType" :
		 * <ul>
		 * <li> <code>syncAttributes()</code> if it's an ATTRIBUTE_CHANGE event type or during initialization
		 * <li> <code>syncShape()</code> in any case
		 * <li> <code>syncBounds()</code> in any case.
		 * <li> <code>changedUpdate</code> on the highlighter delegate, if any.
		 * </ul>
		 * then call repaint with the union of the old and the new bounds as the argument.<br>
		 * Subclass might simply want to update <code>shape</code> before calling <code>super.changedUpdate()</code>,
		 * or override <code>syncAttributes()</code> and <code>syncBounds()</code> if they must implement
		 * more specific behaviour.
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
			//if (jpicedt.Log.DEBUG) jpicedt.Log.debug("eventType="+eventType);
			Rectangle2D oldClip = bounds;
			// cache attributes
			if (eventType==null || eventType==DrawingEvent.EventType.ATTRIBUTE_CHANGE) {
				syncAttributes();
				syncShape();
				syncBounds();
				if (highlighter!=null) {
					double s = 1.0;
					PECanvas canvas = getContainer();
					if (canvas != null) s=canvas.getScaleFactor();
					highlighter.changedUpdate(eventType,s);
					bounds.add(highlighter.getBounds());
				}
				repaint(null); // repaint all in case arrow type/shadow/... changed.
			}
			else { // else other events may result from a translation/scaling/setPoint/...
				syncShape();
				// always update bounds (in model-coordinate)
				syncBounds();
				if (highlighter!=null) {
					double s = 1.0;
					PECanvas canvas = getContainer();
					if (canvas != null) s=canvas.getScaleFactor();
					highlighter.changedUpdate(eventType,s);
					bounds.add(highlighter.getBounds());
				}
				// repaint
				oldClip.add(bounds);
				repaint(oldClip);
			}
		}

		/**
		 * Synchronizes cached attributes values with the model ; 
		 * <ul>
		 * <li>lineWidth, from the LINE_WIDTH attribute
		 * <li>outlineStroke, from the ViewFactory, using <code>createStroke</code>
		 * <li>outinePaint, from the ViewFactory, using <code>createPaintForOutline</code>
		 * <li>interiorPaint, from the ViewFactory, using <code>createPaintForInterior</code>
		 * <li>overStrikeWidth, from the OVER_STRIKE and OVER_STRIKE_WIDTH attributes
		 * </ul>
		 */
		protected void syncAttributes(){
			//if (jpicedt.Log.DEBUG) jpicedt.Log.debug("syncAttributes");
			lineWidth = ((Double)element.getAttribute(LINE_WIDTH)).doubleValue();
			outlineStroke = createStroke(element.getAttributeSet());
			outlinePaint = createPaintForOutline(element.getAttributeSet());
			interiorPaint = createPaintForInterior(element.getAttributeSet());
			// [SR:pending] should be moved to factory preambule in order for subclasser to easily
			// specialize overstrike behaviour :
			if (element.getAttribute(OVER_STRIKE)==Boolean.TRUE){
				overStrikeWidth = ((Double)element.getAttribute(OVER_STRIKE_WIDTH)).doubleValue();
			}
			else {
				overStrikeWidth = 0;
			}
		}

		/**
		 * Synchronizes the bounding box with the model ; 
		 * "bounds" is first computed from the current shape's bound, if any,
		 * then its size gets increased by BARBELL_SIZE, line-thickness, overstrike width,
		 * and possibly the shadow area. 
		 */
		protected void syncBounds(){

			//bounds = element.getBoundingBox(null);
			if (shape==null) return;
			bounds = shape.getBounds2D();
			PicAttributeSet set = element.getAttributeSet();
			if (StyleConstants.isShadow(set)) {
				// create shadow shape
				double angle = Math.toRadians(StyleConstants.getShadowAngle(set));
				double sep = StyleConstants.getShadowSize(set);
				shadowBounds.setRect(bounds.getX()+sep * Math.cos(angle),bounds.getY()+sep * Math.sin(angle),bounds.getWidth(),bounds.getHeight());
				bounds.add(shadowBounds);
			}
			double delta = lineWidth+overStrikeWidth+1;
			bounds.setFrame(
			    bounds.getX()-delta,
			    bounds.getY()-delta,
			    bounds.getWidth()+2*delta,
			    bounds.getHeight()+2*delta);
		}

		/**
		 * Synchronize the "shape" variable with the model ; obviously, this implementation does nothing !
		 */
		protected void syncShape(){
		}

		/**
		 * Render the View to the given graphic context.
		 * This implementation relies on a bottom-to-top z-ordering policy, ie 
		 * first renders the shadow, then the interior and hatches, finally the outline (=stroke).
		 * @param a the current graphic clip
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (shape==null) return;
			// possibly overstrike beforehands
			paintOverStrike(g,element.getAttributeSet(),shape);
			// possibly paint shadow beforehands
			paintShadow(g,element.getAttributeSet(),shape);
			// paint interior :
			if (interiorPaint != null) {
				g.setPaint(interiorPaint);
				g.fill(shape);
			}
			// possibly paint hatches
			paintHatches(g,element.getAttributeSet(),shape);
			// paint outline :
			if (outlinePaint != null && outlineStroke != null){
				g.setStroke(outlineStroke);
				g.setPaint(outlinePaint);
				g.draw(shape);
			}
		}

		/**
		 * This implementation returns a HitInfo.Interior if the view is filled and a click occured on the
		 * interior, or a HitInfo.Stroke if a click occured on the stroke path (this use 
		 * a FlatteningPathIterator built from the current<code>shape</code>).
		 * 
		 * @return a HitInfo corresponding to the given mouse-event
		 */
		public HitInfo hitTest(PEMouseEvent e){

			if (shape == null) return null;
			// first, test if click occured outside the view bounds, which can speed up a lot...
			if (!getBounds().contains(e.getPicPoint())) return null;

			// if it's painted filled, test hit on interior :
			if (interiorPaint != null){
				if (shape.contains(e.getPicPoint())) return new HitInfo.Interior(element,e);
			}
			// otherwise, test hit on stroke, first scaling down CLICK_DISTANCE by the current zoom factor
			// so as to remain coherent with the policy used for end-points's hit test (see AbstractView) :
			double click_distance = CLICK_DISTANCE / e.getCanvas().getZoomFactor();
			int segmentIndex = PEToolKit.testDistanceToPath(shape, e.getPicPoint(), click_distance);
			if (segmentIndex >= 0) return new HitInfo.Stroke(element,segmentIndex,e);
			return null;
		}

	} // LeafElementView


	//////////////////////////////////////////////////////////////////////////
	//// PARALLELOGRAM VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering parallelograms
	 */
	public class ParallelogramView extends DefaultViewFactory.LeafElementView {

		/**
		 * construct a new View for the given PicParallelogram
		 */
		public ParallelogramView(PicParallelogram para){
			super(para);
			changedUpdate(null);
		}

		/**
		 * Synchronize the "shape" variable with the model
		 */
		protected void syncShape(){
			PicParallelogram para = (PicParallelogram)element;
			// update shape
			if (shape==null) shape = new GeneralPath();
			GeneralPath path = (GeneralPath)shape;
			// update shape
			path.reset();
			path.moveTo((float)para.getPointX(0), (float)para.getPointY(0)); // set first point
			path.lineTo((float)para.getPointX(1), (float)para.getPointY(1));
			path.lineTo((float)para.getPointX(2), (float)para.getPointY(2));
			path.lineTo((float)para.getPointX(3), (float)para.getPointY(3));
			path.closePath();
		}

	} // ParallelogramView

	//////////////////////////////////////////////////////////////////////////
	//// ELLIPSE VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering ellipses defined by a parallelogram
	 */
	public class EllipseView extends DefaultViewFactory.LeafElementView {

		protected Stroke arrowStroke;
		protected Arrow arrowStart, arrowEnd;
		protected PicPoint arrowStartDir, arrowEndDir;
		protected PicPoint ptAngleEnd, ptAngleStart;

		AffineTransform rot;
		Arc2D shapeEllipse ; //used only for recycling purposes because shape itself is not an Ellipse2D, since it is rotated.

		/**
		 * construct a new View for the given ellipse
		 */
		public EllipseView(PicEllipse ellipse){
			super(ellipse);
			changedUpdate(null);
		}

		/**
		 * Synchronize the "shape" variable with the model
		 */
		protected void syncShape(){

			PicEllipse ellipse = (PicEllipse)element;
			// update shape
			if (shapeEllipse == null || shapeEllipse.getArcType()!= ellipse.getArcType()) {
				shapeEllipse = new Arc2D.Double(ellipse.getArcType());
			}
			shapeEllipse.setAngleExtent(-ellipse.getRotatedAngleExtent());
			shapeEllipse.setAngleStart(-ellipse.getRotatedAngleStart());
			shapeEllipse.setArcType(ellipse.getArcType());

			// arc : add arrows only if arc is OPEN
			if (!ellipse.isPlain()){
				if (ellipse.getArcType()!=PicEllipse.OPEN) arrowStart = arrowEnd = null;
				else {
					arrowStart = createArrow(element.getAttributeSet(),LEFT_ARROW);
					if (arrowStart != null) {
						arrowStartDir = ellipse.getTangentAtAngleStart(null);
						ptAngleStart = element.getPoint(PicEllipse.P_ANGLE_START,ptAngleStart);
					}
					arrowEnd = createArrow(element.getAttributeSet(),RIGHT_ARROW);
					if (arrowEnd != null) {
						arrowEndDir = ellipse.getTangentAtAngleEnd(null);
						ptAngleEnd = element.getPoint(PicEllipse.P_ANGLE_END,ptAngleEnd);
					}
				}
			}
			//ellipse:
			else {
				arrowStart = arrowEnd = null;
				shapeEllipse.setArcType(Arc2D.OPEN); // if closure was PIE, we get a better visual effect then
			}
			arrowStroke = new BasicStroke((float)lineWidth);

			shapeEllipse.setFrameFromCenter(0,0,ellipse.getGreatAxisLength()/2,Math.abs(ellipse.getSmallAxisLength()/2));
			if (rot==null) rot=new AffineTransform();
			rot.setToIdentity();
			rot.translate(ellipse.getPointX(ellipse.P_CENTER),ellipse.getPointY(ellipse.P_CENTER));
			rot.rotate(ellipse.getRotationAngle()); // [SR:pending] use AffineTransform.rotate(angle, x, y) which rotates around an anchor point
			// rot = translate o rotate
			if (ellipse.getSmallAxisLength()<0) rot.scale(1,-1);
			shape=rot.createTransformedShape(shapeEllipse);
		}

		/**
		 * Render the View to the given graphic context.
		 * This implementation render the interior first, then the outline.
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (!a.intersects(getBounds())) return;
			super.paint(g,a);
			g.setStroke(arrowStroke);
			if (arrowStart!=null) arrowStart.paint(g, lineWidth,ptAngleStart, arrowStartDir);
			if (arrowEnd!=null) arrowEnd.paint(g, lineWidth,ptAngleEnd, arrowEndDir);
		}

	} // EllipseView

	//////////////////////////////////////////////
	/// ABSTRACT CURVE VIEW
	/////////////////////////////////////////////
	/**
	 * a View for rendering any concrete implementation of jpicedt.graphic.model.AbstractCurve
	 */
	public class AbstractCurveView extends LeafElementView {

		protected Stroke arrowStroke;
		protected Arrow arrow1, arrow2;
		protected PicVector arrow1Dir=new PicVector(), arrow2Dir=new PicVector();
		protected PicPoint arrow1Loc, arrow2Loc;
		protected boolean hasSegments=false; // true if number of segments >=1 ;
		// The following are set by syncDotsShape(), and used to draw polygon's dots, if applicable:
		protected Shape dotShape=null; // null if no dots to be painted
		protected boolean isDotFilled=false;  // "square" vs. "square*"
		protected Stroke dotStroke=null; // different from outlineStroke !
		protected Paint dotFillPaint=null, dotDrawPaint=null; // may be different from outlinePaint or interiorPaint

		/**
		 * construct a new View for the given curve
		 */
		public AbstractCurveView(AbstractCurve curve){
			super(curve);
			changedUpdate(null);
		}

		/**
		 * Synchronize the "shape" variable with the model. To reduce the burden for subclasses, this 
		 * method calls <code>syncTangents</code> after having updated the <code>shape</code> field. It is
		 * up to subclasses to adapt their behaviour according to the underlying model, if only computing tangents
		 * is different.
		 */
		protected void syncShape(){
			AbstractCurve curve = (AbstractCurve)element;
			boolean hasSegments = (curve.getNumberOfSegments()>=1); // check if curve has at least one segment
			if (!hasSegments ) { // call at <init> time
				shape=null;
				return; // curve is reduced to a point -> don't paint !
			}

			if (shape==null) shape=new GeneralPath(); // avoid over-creating objects !
			// assume Npts > 2 (ie a "true" curve) since curve has a least one segment
			GeneralPath path = (GeneralPath)shape;
			// update shape
			path.reset();
			path.moveTo((float)curve.getSpecificationPointX(0), (float)curve.getSpecificationPointY(0)); // set first point
			for (int i=0; i<curve.getNumberOfSegments(); i=i+1){
				if (curve.isStraight(i)){
					path.lineTo((float)curve.getSpecificationPointX(3*i+3), (float)curve.getSpecificationPointY(3*i+3));
				}
				else {
					path.curveTo((float)curve.getSpecificationPointX(3*i+1), (float)curve.getSpecificationPointY(3*i+1),
					             (float)curve.getSpecificationPointX(3*i+2), (float)curve.getSpecificationPointY(3*i+2),
					             (float)curve.getSpecificationPointX(3*i+3), (float)curve.getSpecificationPointY(3*i+3));
				}
			}
			// either close path
			if (curve.isClosed()){
				path.closePath();
				arrow1 = arrow2 = null; // this is a flag for paint() to signal we mustn't paint arrows
			}
			// or (possibly) paint arrows
			else {
				arrow1 = createArrow(element.getAttributeSet(),LEFT_ARROW);
				if (arrow1!=null){
					arrow1Loc=curve.getSpecificationPoint(0,arrow1Loc);
					if (curve.isStraight(0)) curve.getSpecificationPoint(3,arrow1Dir); // use arrow1Dir as buffer
					else curve.getSpecificationPoint(1,arrow1Dir);
					arrow1Dir.setCoordinates(arrow1Dir, arrow1Loc);
					arrow1Dir.normalize();
				}
				arrow2 = createArrow(element.getAttributeSet(),RIGHT_ARROW);
				if (arrow2!=null){
					arrow2Loc=curve.getSpecificationPoint(curve.getNumberOfSpecificationPoints()-1,arrow2Loc);
					if (curve.isStraight(curve.getNumberOfSegments()-1)) // if last seg is straight...
						curve.getSpecificationPoint(curve.getNumberOfSpecificationPoints()-4,arrow2Dir); // use arrow2Dir as buffer
					else
						curve.getSpecificationPoint(curve.getNumberOfSpecificationPoints()-2,arrow2Dir);
					arrow2Dir.setCoordinates(arrow2Dir, arrow2Loc);
					arrow2Dir.normalize();
				}
			}
			arrowStroke = new BasicStroke((float)lineWidth);
		}

		/**
		 * Inherits from superclass, except where painting dots is concerned, where this method delegates
		 * to {@link #syncDotShape syncDotShape()} to create the appropriate Shape.
		 */
		protected void syncAttributes(){
			super.syncAttributes();
			syncDotShape();
		}

		/**
		 * Computes the Shape (centered at 0,0) used for drawing polygon's dots, built from the given attributes ; 
		 * set it to null if POLYDOTS_STYLE==NONE or if dotting is not supported by the factory.<p>
		 * This method gets invoked by {@link #syncAttributes syncAttributes} as part of the delegation mechanism, 
		 * when style of dots has changed.
		 */
		protected void syncDotShape(){
			PicAttributeSet set = element.getAttributeSet();
			String dotStyle = StyleConstants.getPolydotsStyle(set);
			float dotWidth = (float)((StyleConstants.getLineWidth(set) * StyleConstants.getPolydotsSizeLineWidthScale(set)
			                          + StyleConstants.getPolydotsSizeMinimumMm(set))/2.);
			dotStroke = new BasicStroke(dotWidth/5.0f); // [SR:pending] ask D.Guirou for exact data
			float dotHeight = dotWidth * (float)(StyleConstants.getPolydotsVerticalScale(set));
			dotWidth *= StyleConstants.getPolydotsHorizontalScale(set);
			if (dotWidth<0) dotWidth=0;
			if (dotHeight<0) dotHeight=0;
			AffineTransform rotAt = AffineTransform.getRotateInstance(StyleConstants.getPolydotsAngle(set)*Math.PI/180.0);
			if (dotStyle==POLYDOTS_DISK){
				dotShape = new Ellipse2D.Double(-dotWidth,-dotHeight,2.*dotWidth,2.*dotHeight);
				isDotFilled=true;
				// rotating is useless here ;-)
			}
			else if (dotStyle == POLYDOTS_CIRCLE){
				dotShape = new Ellipse2D.Double(-dotWidth,-dotHeight,2.*dotWidth,2.*dotHeight);
				isDotFilled=false;
				// ibid.
			}
			else if (dotStyle == POLYDOTS_PLUS){
				GeneralPath pth = new GeneralPath();
				pth.moveTo(0,dotHeight);
				pth.lineTo(0,-dotHeight);
				pth.moveTo(-dotWidth,0);
				pth.lineTo(dotWidth,0);
				dotShape = rotAt.createTransformedShape(pth);
				isDotFilled=false;
			}
			else if (dotStyle == POLYDOTS_TRIANGLE || dotStyle == POLYDOTS_TRIANGLE_FILLED){
				GeneralPath pth = new GeneralPath();
				pth.moveTo(-dotWidth,-dotHeight);
				pth.lineTo(dotWidth,-dotHeight);
				pth.lineTo(0,dotHeight);
				pth.closePath();
				dotShape = rotAt.createTransformedShape(pth);
				isDotFilled=(dotStyle==POLYDOTS_TRIANGLE_FILLED);
			}
			else if (dotStyle == POLYDOTS_SQUARE || dotStyle == POLYDOTS_SQUARE_FILLED){
				GeneralPath pth = new GeneralPath();
				pth.moveTo(-dotWidth,-dotHeight);
				pth.lineTo(dotWidth,-dotHeight);
				pth.lineTo(dotWidth,dotHeight);
				pth.lineTo(-dotWidth,dotHeight);
				pth.closePath();
				dotShape = rotAt.createTransformedShape(pth);
				isDotFilled=(dotStyle==POLYDOTS_SQUARE_FILLED);
			}
			else if (dotStyle == POLYDOTS_PENTAGON || dotStyle == POLYDOTS_PENTAGON_FILLED){
				GeneralPath pth = new GeneralPath();
				pth.moveTo(0,dotHeight);
				for (int i=1; i<5; i++){
					double angle = (0.4*i+.5)*Math.PI;
					double x = dotWidth * Math.cos(angle); // don't worry about performances here, syncDotShape is never called
					double y = dotHeight * Math.sin(angle); // when updating an Element's geometry
					pth.lineTo((float)x,(float)y);
				}
				pth.closePath();
				dotShape = rotAt.createTransformedShape(pth);
				isDotFilled=(dotStyle==POLYDOTS_PENTAGON_FILLED);
			}
			// [SR:pending] Tbar dots ?
			else dotShape=null;

			// dot draw/fill attributes:
			if (isDotFilled){ // e.g. "square*" => stroked and filled with "linecolor", and "fillcolor" is not used
				if (outlinePaint==null) dotDrawPaint = dotFillPaint = Color.black; // don't paint multicurve, yet paint dots !
				else dotDrawPaint = dotFillPaint = outlinePaint;
			}
			else {// e.g. "triangle" => filled with "fillcolor", stroked with "linecolor"
				dotDrawPaint = (outlinePaint==null ? Color.black : outlinePaint);
				dotFillPaint = (interiorPaint==null ? Color.white : interiorPaint);
			}
		}

		/**
		 * Render the View to the given graphic context.
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (shape == null && dotShape==null) return;
			if (!a.intersects(getBounds())) return;
			boolean superImposeDots = StyleConstants.isPolydotsSuperimposed(element.getAttributeSet());
			if (shape != null && (dotShape==null || (dotShape!=null && superImposeDots))){
				super.paint(g,a); //paint "this.shape"
				// paint arrows with solid stroke :
				g.setStroke(arrowStroke);
				if (arrow1 != null) arrow1.paint(g, lineWidth, arrow1Loc, arrow1Dir);
				if (arrow2 != null) arrow2.paint(g, lineWidth, arrow2Loc, arrow2Dir);
			}
			if (dotShape != null) { // paint dots
				g.setStroke(dotStroke);
				AbstractCurve curve = (AbstractCurve)element;
				for (int subDivIdx = 0; subDivIdx < curve.getNumberOfSubdivisionPoints(); subDivIdx++){
					int i = curve.segmentToPointIndex(subDivIdx, curve.SUBDIVISION_POINT);
					AffineTransform trans = AffineTransform.getTranslateInstance(curve.getSpecificationPointX(i),curve.getSpecificationPointY(i));
					Shape translated = trans.createTransformedShape(dotShape);
					g.setPaint(dotFillPaint);
					g.fill(translated);
					g.setPaint(dotDrawPaint);
					g.draw(translated);
				}
			}
		}

	} // AbstractCurveView


	//////////////////////////////////////////////
	/// PicEDGE VIEW
	/////////////////////////////////////////////
	/**
	 * a View for rendering any concrete implementation of jpicedt.graphic.model.PicNodeConnection
	 * This view is responsible for painting the edge only! Nodes are rendered with their own view (provided
	 * they were added to the drawing at some early point...)
	 */
	public class PicNodeConnectionView extends LeafElementView {

		protected Stroke arrowStroke;
		protected Arrow arrow1, arrow2; // the connection may have arrows
		protected PicVector arrow1Dir=new PicVector(), arrow2Dir=new PicVector();
		protected PicPoint arrow1Loc, arrow2Loc;

		/**
		 * construct a new View for the given node connection
		 */
		public PicNodeConnectionView(PicNodeConnection edge){
			super(edge);
			changedUpdate(null);
		}

		/**
		 * Synchronize the "shape" variable with the model. To reduce the burden for subclasses, this 
		 * method calls <code>syncTangents</code> after having updated the <code>shape</code> field. It is
		 * up to subclasses to adapt their behaviour according to the underlying model, if only computing tangents
		 * is different.
		 */
		protected void syncShape(){
			PicNodeConnection edge = (PicNodeConnection)element;
			this.shape = edge.fetchEdgeShape();

			// possibly paint arrows
			arrow1 = createArrow(element.getAttributeSet(),LEFT_ARROW);
			if (arrow1!=null){
				arrow1Loc=edge.getEndPointA(arrow1Loc);
				arrow1Dir=edge.getTangentA(arrow1Dir);
			}
			arrow2 = createArrow(element.getAttributeSet(),RIGHT_ARROW);
			if (arrow2!=null){
				arrow2Loc=edge.getEndPointB(arrow2Loc);
				arrow2Dir=edge.getTangentB(arrow2Dir);
			}
			arrowStroke = new BasicStroke((float)lineWidth);
		}

		/**
		 * Render the View to the given graphic context.
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (shape == null) return;
			if (!a.intersects(getBounds())) return;
			super.paint(g,a); //paint "this.shape"
			// paint arrows with solid stroke :
			g.setStroke(arrowStroke);
			if (arrow1 != null) arrow1.paint(g, lineWidth, arrow1Loc, arrow1Dir);
			if (arrow2 != null) arrow2.paint(g, lineWidth, arrow2Loc, arrow2Dir);
		}

	} // PicNodeConnectionView

	//////////////////////////////////////////////////////////////////////////
	//// TEXTEDITABLE VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering Text's from TextLayout's
	 */
	public class TextView extends DefaultViewFactory.LeafElementView {

		/** the TextLayout that renders the text string of this TextEditable */
		protected TextLayout textLayout;
		/** TextLayout location with respect to PicText's anchor point */
		protected double strx, stry;
		// shape = frame box !!! (and not textLayout !)
		protected AffineTransform tl2ModelTr=new AffineTransform();; // maps TextLayout coordinates to Model coordinates


		/**
		 * construct a new View for the given PicRectangle
		 */
		public TextView(PicText te){
			super(te);
			changedUpdate(null);
		}

		/**
		 * Give notification from the model that a change occured to the text this view is responsible
		 * for rendering.<p>
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
			if (eventType==DrawingEvent.EventType.TEXT_CHANGE) super.changedUpdate(null); // update all
			else super.changedUpdate(eventType); 
		}

		/**
		 * Returns the text rotation in radians : subclassers that don't support rotating text may return 0 here.
		 */
		protected double getRotation(){
			PicAttributeSet set = element.getAttributeSet();
			return Math.toRadians(((Double)set.getAttribute(TEXT_ROTATION)).doubleValue());
		}

		/**
		 * Synchronize the textLayout and the shape (=frame box, by calling syncFrame) with the model
		 * This delegates to getRotation() where computing rotation angle is concerned, and
		 * updates the AffineTransform returned by getTextToModelTransform().
		 */
		protected void syncShape(){
			PicText te = (PicText)element;
			// update shape
			//textLayout = new TextLayout(te.getText().length()==0 ? " " : te.getText(), textFont, g.getFontRenderContext());
			textLayout = new TextLayout(te.getText().length()==0 ? " " : te.getText(),
			                            textFont,
			                            new FontRenderContext(null,false,false));
			syncStringLocationX();
			syncStringLocationY();
			// terrible hack ! textLayout.getBounds2D() gives wrong results if text is to be drawn
			// upside-down, hence we reverse it before computing the bounding rectangle :
			tl2ModelTr.setToIdentity(); // reset
			tl2ModelTr.translate(te.getPointX(TextEditable.P_ANCHOR),te.getPointY(TextEditable.P_ANCHOR));
			tl2ModelTr.rotate(getRotation()); // rotate along P_ANCHOR !
			tl2ModelTr.translate(strx,stry);
			tl2ModelTr.scale(1.0,-1.0);
			syncFrame();
		}

		/**
		 * Synchronizes bounding box with the model ; 
		 */
		protected void syncBounds(){
			if (shape!=null) { // if there's a frame around...
				super.syncBounds();
			}
			else {
				Rectangle2D tb = textLayout.getBounds();
				bounds = tl2ModelTr.createTransformedShape(tb).getBounds2D(); // at (0,0)
				double delta = 1;
				bounds.setFrame(
				    bounds.getX()-delta,
				    bounds.getY()-delta,
				    bounds.getWidth()+2*delta,
				    bounds.getHeight()+2*delta);
			}
		}

		/** synchronize frame shape and location */
		protected void syncFrame(){
			PicText te = (PicText)element;
			if (te.getFrameType()==TEXT_BOX_RECTANGLE){
				Rectangle2D tb = textLayout.getBounds();
				shape = tl2ModelTr.createTransformedShape(tb);
			}
			else if (te.getFrameType()==TEXT_BOX_OVAL){
				Rectangle2D tb = textLayout.getBounds();
				Ellipse2D frame = new Ellipse2D.Double();
				frame.setFrame(tb);
				shape = tl2ModelTr.createTransformedShape(frame);
			}
			else if (te.getFrameType()==TEXT_BOX_CIRCLE){
				Rectangle2D tb = textLayout.getBounds();
				Ellipse2D.Double frame = new Ellipse2D.Double();
				frame.setFrameFromCenter(tb.getCenterX(),tb.getCenterY(),tb.getCenterX()+tb.getWidth()*0.5,tb.getCenterY()+tb.getWidth()*0.5);
				shape = tl2ModelTr.createTransformedShape(frame);
			}
			else shape=null; // TEXT_BOX_NO_FRAME
		}

		/** update strx = x-location of TextLayout's bottom-Left corner with respect to PicText's anchor-point */
		protected void syncStringLocationX(){
			PicText te = (PicText)element;
			double textWidth = textLayout.getBounds().getWidth();
			if (te.getHorAlign() == TEXT_HALIGN_LEFT) strx = 0;
			else if (te.getHorAlign() == TEXT_HALIGN_RIGHT) strx = - textWidth;
			else // = if (te.getHorAlign() == TEXT_HALIGN_CENTER)
				strx = - textWidth/2;
		}

		/** update strx = y-location of TextLayout's bottom-Left corner with respect to PicText's anchor-point */
		protected void syncStringLocationY(){
			PicText te = (PicText)element;
			//double ascent = textLayout.getAscent();
			// e.g. return : y = -7 ; h = 8 (ie upside-down)
			double ascent = -textLayout.getBounds().getMinY();
			//double descent = textLayout.getDescent();
			if (te.getVertAlign()== TEXT_VALIGN_BASELINE) {
				stry = 0;
			}
			else if (te.getVertAlign()== TEXT_VALIGN_TOP) {
				if (te.getFrameType()==TEXT_BOX_CIRCLE){
					Rectangle2D tlb = textLayout.getBounds();
					stry =  tlb.getCenterY() - tlb.getWidth()*0.5;
				}
				else stry = textLayout.getBounds().getMinY(); // should be positive (reverse Y-axis !)
			}
			else if (te.getVertAlign()== TEXT_VALIGN_BOTTOM) {
				if (te.getFrameType()==TEXT_BOX_CIRCLE){
					Rectangle2D tlb = textLayout.getBounds();
					stry =  tlb.getCenterY() + tlb.getWidth()*0.5;
				}
				else stry = textLayout.getBounds().getMaxY(); // should be negative  (reverse Y-axis !)
			}
			else { // = if (te.getVertAlign()== TEXT_VALIGN_CENTER) {
				Rectangle2D tlb = textLayout.getBounds();
				double d = tlb.getHeight()*0.5 - tlb.getMaxY(); // e.g. 8 * 0.5 - 1 = 3
				stry =  - d;
			}
		}

		/**
		 * Render the View to the given graphic context.
		 * This implementation render the interior first, then the outline.
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (!a.intersects(getBounds())) return;
			super.paint(g,a); // possibly paint framebox if non-null

			AffineTransform oldAT = g.getTransform();
			// paint text in black
			g.setPaint(Color.black);
			// from now on, we work in Y-direct (<0) coordinates to avoid inextricable problems with font being mirrored...
			g.transform(tl2ModelTr); // also include rotation
			textLayout.draw(g, 0.0f,0.0f);
			// get back to previous transform
			g.setTransform(oldAT);
			if (jpicedt.Log.DEBUG) {
				g.setPaint(Color.red);
				g.draw(bounds);
			}
		}

		/**
		 * This implementation calls <code>super.hitTest</code> and returns the result if non-null
		 * (this should be a HitInfo.Point), 
		 * then returns a HitInfo.Interior if the mouse-click occured inside the text bound (as defined
		 * by text layout)
		 * 
		 * @return a HitInfo corresponding to the given mouse-event
		 */
		public HitInfo hitTest(PEMouseEvent e){

			if (!getBounds().contains(e.getPicPoint())) return null;

			PicText te = (PicText)element;
			// recompute textlayout b-box, but store it in a temporary field !
			Rectangle2D tb = textLayout.getBounds();
			Shape text_bounds = tl2ModelTr.createTransformedShape(tb);
			if (text_bounds.contains(e.getPicPoint())) {
				// [SR:pending] for the hitInfo to be reliable, getPicPoint() should first be transformed by
				//              inverse tl2ModelTr ! (especially when rotationAngle != 0)
				TextHitInfo thi = textLayout.hitTestChar((float)(e.getPicPoint().x - strx), (float)(e.getPicPoint().y - stry)); // guaranteed to return a non-null thi
				return new HitInfo.Text((PicText)element, thi, e);
			}
			// test hit on textlayout's bounding rectangle :
			//else if (bounds.contains(e.getPicPoint())) return new HitInfo.Interior(element,e);
			return null;


		}

		/**
		 * [SR:pending] make this view implement aka TextEditableView interface (or something like it), where
		 * TextEditableView is a subinterface of View with text-editing specific capabilities.
		 * 
		 * Returns the TextLayout which is responsible for painting the textual content of this element
		 */
		public TextLayout getTextLayout(){
			return textLayout;
		}

		/** Return an affine transform which translat b/w the TextLayout coordinate system and the
		 *  jpicedt.graphic.model coordinate system.
		 * [SR:pending] refactor method name to something more explanatory
		 */
		public AffineTransform getTextToModelTransform(){
			return tl2ModelTr;
		}

	} // TextView

	//////////////////////////////////////////////////////////////////////////
	//// TEXTEDITABLE VIEW
	//////////////////////////////////////////////////////////////////////////
	/**
	 * a View for rendering Text's based on bitmaps (when available) or TextLayout (when image not available)
	 */
	public class TextViewHybrid extends DefaultViewFactory.LeafElementView implements Runnable {


	    boolean wantToComputeLatexDimensions=true;// from preferences: do we want to run LateX to compute preferences
	    boolean wantToGetBitMap=true;// from preferences: do we want to run LateX+pstoimg to get a bitmap. 
	    //   wantToGetBitMap=true should Imply wantToComputeLatexDimensions=true  

		protected double strx, stry; // TextLayout/Image location with respect to PicText's anchor point
	    //[pending] pas joli. A enlever sous peu (utilise juste dans HitInfo que je ne comprends pas, je laisse donc en attendant)
	    
		protected TextLayout textLayout; // the TextLayout that renders the text string of this TextEditable
		// [inherited] shape; But here it's the frame box !!! (not textLayout !)
		protected AffineTransform text2ModelTr=new AffineTransform();; // maps text coordinates to Model coordinates
		protected BufferedImage image; // bitmap (if null, we rely on TextLayout)
	        protected boolean areDimensionsComputed;  // has the real dimensions been read from the log file ?
		protected int fileDPI = 300 ; // Dot Per Inch of the file (for image resizing) [pending] gerer les preferences de ce parametre
	    // il faut passer la valeur en DPI en argument a create_bitmap.sh

		/**
		 * construct a new View for the given PicRectangle
		 */
		public TextViewHybrid(PicText te){
			super(te);
			areDimensionsComputed=false;
			changedUpdate(null);
		}

		/**
		 * Returns the text rotation in radians : subclassers that don't support rotating text may return 0 here.
		 * Used by TextLayout only.
		 */
		protected double getRotation(){
			PicAttributeSet set = element.getAttributeSet();
			//			jpicedt.Log.debug(set.getAttribute(TEXT_ROTATION).toString());
			return Math.toRadians(((Double)set.getAttribute(TEXT_ROTATION)).doubleValue());
		}

		/**
		 * Give notification from the model that a change occured to the text this view is responsible
		 * for rendering.<p>
		 */
		public void changedUpdate(DrawingEvent.EventType eventType){
		    PicText text = (PicText)element;
		    if (textLayout==null){
			textLayout = new TextLayout(text.getText().length()==0 ? " " : text.getText(),
						    textFont,
						    new FontRenderContext(null,false,false));
		    }
		    if (eventType==DrawingEvent.EventType.TEXT_CHANGE) {
			textLayout = new TextLayout(text.getText().length()==0 ? " " : text.getText(),
						    textFont,
						    new FontRenderContext(null,false,false));
			
			// first try to create a bitmap
			image=null; // aka "reset" image => we might temporarily resort to TextLayout until the image is ready
			areDimensionsComputed=false;
			//reset dimensions to the textlayout dimensions
			text.setDimensions(textLayout.getBounds().getWidth(),textLayout.getAscent(),textLayout.getDescent());
			
			if ( wantToComputeLatexDimensions && text.getText().length()>0){ 
			    // don't produce a bitmap for an empty string (LaTeX might not like it)
			    // [pending] this should be made dependent on a preference's option
			    new Thread(this).start();
			}
			if (image==null) super.changedUpdate(null); // ie resort to TextLayout (update all)
		    }
		    else {
			text.updateFrame();
			super.changedUpdate(eventType); 
		    }
		}

		// Thread's run method aimed at creating a bitmap asynchronously
		public void run(){
			Drawing drawing = new Drawing();
			PicText rawPicText=new PicText();
			String s=((PicText)element).getText();
			rawPicText.setText(s);
			drawing.addElement(rawPicText); // bug fix: we must add a CLONE of the PicText, otherwise it loses it former parent... (then pb with the view )			
			drawing.setNotparsedCommands("\\newlength{\\jpicwidth}\\settowidth{\\jpicwidth}{"+s+"}"+CR_LF
						     + "\\newlength{\\jpicheight}\\settoheight{\\jpicheight}{" +s + "}"+CR_LF
						     + "\\newlength{\\jpicdepth}\\settodepth{\\jpicdepth}{" +s + "}"+CR_LF
						     + "\\typeout{JPICEDT INFO: \\the\\jpicwidth, \\the\\jpicheight,  \\the\\jpicdepth }"+CR_LF);
			String commandToRun = RunExternalCommand.BITMAP_CREATION;
			//RunExternalCommand command = new RunExternalCommand(drawing, contentType,commandToRun);
			boolean isWriteTmpTeXfile=true;
			String bitmapExt = "png"; // [pending] preferences
			String cmdLine = "{i}/unix/tetex/create_bitmap.sh {p} {f} "+bitmapExt+" "+fileDPI; // [pending] preferences
			ContentType contentType = getContainer().getContentType();
			RunExternalCommand.isGUI=false; // System.out, no dialog box // [pending] debug
			RunExternalCommand command = new RunExternalCommand(drawing, contentType,cmdLine, isWriteTmpTeXfile);
			command.run(); // synchronous in an async. thread => it's ok (anyway, we must way until the LaTeX process has completed)


			if (wantToComputeLatexDimensions){
			    // load size of text:
			    try {
				File logFile = new File(command.getTmpPath(), command.getTmpFilePrefix()+".log");
				BufferedReader reader=null;
				try{
				    reader = new BufferedReader(new FileReader(logFile));
				}
				catch (FileNotFoundException fnfe) {
				    System.out.println("Cannot find log file! "+fnfe.getMessage());
				    System.out.println(logFile);
				}
				catch (IOException ioex){System.out.println("Log file IO exception"); ioex.printStackTrace(); }//utile ?
				System.out.println("Log file created! file="+logFile);
				getDimensionsFromLogFile(reader,(PicText)element);
				syncStringLocation(); // update dimensions
				syncBounds();
				syncFrame();
				SwingUtilities.invokeLater(new Thread(){
					                           public void run(){
						                           repaint(null);
					                           }
				                           });
				//repaint(null); // now that dimensions are available, we force a repaint() [pending] smart-repaint ?
			    }
			    catch (Exception e){
				e.printStackTrace();
			    }
			}

			if (wantToGetBitMap){
			    // load image:
			    try {
				File bitmapFile = new File(command.getTmpPath(), command.getTmpFilePrefix()+"."+bitmapExt);
				this.image = ImageIO.read(bitmapFile);
				System.out.println("Bitmap created! file="+bitmapFile+", width="+image.getWidth()+"pixels, height="+image.getHeight()+"pixels");
				if (image==null) return;
				syncStringLocation(); //sets strx, stry, and dimensions of text
				syncBounds(); 
				// update the AffineTransform that will be applied to the bitmap before displaying on screen
				PicText te = (PicText)element;
				text2ModelTr.setToIdentity(); // reset
				text2ModelTr.rotate(getRotation(),te.getPointX(TextEditable.P_ANCHOR),te.getPointY(TextEditable.P_ANCHOR)); // rotate along P_ANCHOR !
				text2ModelTr.translate(te.getLeftX(),te.getTopY()); 
			      	text2ModelTr.scale(te.getWidth()/image.getWidth(),-(te.getHeight()+te.getDepth())/image.getHeight());
				//[pending]  should do something special to avoid dividing by 0 or setting a rescaling factor to 0 [non invertible matrix] (java will throw an exception)
				syncFrame();
				SwingUtilities.invokeLater(new Thread(){
					                           public void run(){
						                           repaint(null);
					                           }
				                           });
				//repaint(null); // now that bitmap is available, we force a repaint() [pending] smart-repaint ?
			    }
			    catch (Exception e){
				e.printStackTrace();
			    }
			}
		}


		/** update strx stry = location of TextLayout's bottom-Left corner with respect to PicText's anchor-point */
	    protected void syncStringLocation(){
		PicText te = (PicText)element;
		if (image == null){
		    if (! areDimensionsComputed){
			te.setDimensions(textLayout.getBounds().getWidth(),textLayout.getAscent(),textLayout.getDescent());
		    }
		    strx = te.getLeftX()-te.getPointX(TextEditable.P_ANCHOR);
		    stry = te.getBaseLineY()-te.getPointY(TextEditable.P_ANCHOR);
		}
		else {//image not null
		    strx = te.getLeftX()-te.getPointX(TextEditable.P_ANCHOR);
		    stry = te.getBottomY()-te.getPointY(TextEditable.P_ANCHOR);
		}
	    }
	    

		/**
		 * Synchronize the textLayout and the shape (=frame box, by calling syncFrame) with the model
		 * When TextLayout is used, this delegates to getRotation() where computing rotation angle is concerned, and
		 * updates the AffineTransform returned by getTextToModelTransform().
		 */
		protected void syncShape(){
			PicText te = (PicText)element;

			//			textLayout = new TextLayout(te.getText().length()==0 ? " " : te.getText(),
			//			    textFont,
			//			    new FontRenderContext(null,false,false));

			text2ModelTr.setToIdentity(); // reset
			text2ModelTr.rotate(getRotation(),te.getPointX(TextEditable.P_ANCHOR),te.getPointY(TextEditable.P_ANCHOR)); // rotate along P_ANCHOR !
			// the reference point of an image is the top-left one, but the refpoint of a text layout is on the baseline
			if (image!=null){
			    text2ModelTr.translate(te.getLeftX(),te.getTopY()); 
			    if (te.getWidth()!=0 && image.getWidth()!=0 && (te.getDepth()+te.getHeight())!=0 && image.getHeight()!=0)
				text2ModelTr.scale(te.getWidth()/image.getWidth(),-(te.getHeight()+te.getDepth())/image.getHeight());
			}
			else {
			    //Hack ? Just cheating a little bit ? Ou juste ruse ?
			    //we want here to use the dimensions of the textLayout instead of latex dimensions if areDimensionsComputed
			    // sinon on va aligner le textlayout en fonction des parametres latex, et l'Utilisateur (qui est bien bete) ne va rien comprendre.
			    double latexH=0; double latexD=0; double latexW=0;
			    if (areDimensionsComputed){//store latex dimensions, and setDimensions to textLayout ones
				latexH=te.getHeight();
				latexD=te.getDepth();
				latexW=te.getWidth();
				te.setDimensions(textLayout.getBounds().getWidth(),textLayout.getAscent(),textLayout.getDescent());
			    }
			    text2ModelTr.translate(te.getLeftX(),te.getBaseLineY()); 
			    if (areDimensionsComputed){//restore latex dimensions
				te.setDimensions(latexW,latexH,latexD);
			    }
			    //Autre possibilite= comprimer le texte pour qu'il rentre dans la boite (evite le hack ci-dessus):
			    //text2ModelTr.scale(te.getWidth()/textLayout.getWidth(),-(te.getHeight()+te.getDepth())/textLayout.getHeight());
			    text2ModelTr.scale(1.0,-1.0);
			}
			syncFrame();
		}

		/**
		 * synchronize frame shape and location (TextLayout only) ; this is called by syncShape(),
		 * so that subclasser might override easily when only rectangular shapes are availables.
		 */
		protected void syncFrame(){
			PicText te = (PicText)element;
			if (! te.isFramed()){return;}
			AffineTransform tr=new AffineTransform(); // maps Image coordinates to Model coordinates (see paint)
			tr.setToIdentity(); // reset
			tr.rotate(getRotation(),te.getPointX(TextEditable.P_ANCHOR),te.getPointY(TextEditable.P_ANCHOR)); // rotate along P_ANCHOR !
		      	shape = tr.createTransformedShape(te.getShapeOfFrame());
		}

	    
	    protected void getDimensionsFromLogFile(BufferedReader reader,PicText text){
		if (reader == null){return;}
		String line="" ;//il rale si j'initialise pas ...
		boolean finished=false;
		while(!finished){
		    try{
			line=reader.readLine();
		    }
		    catch (IOException ioex){
			ioex.printStackTrace(); 
			return;
		    }
		    if (line == null){
			System.out.println("Size of text not found in log file...");
			return;
		    }

		    System.out.println(line);
		    Matcher matcher=LogFilePattern.matcher(line);
		    
		    if (line != null && matcher.find())
			{
			    System.out.println("FOUND :"+line);
			    finished=true;
			    try{
				text.setDimensions(0.3515*Double.parseDouble(matcher.group(1)),//height, pt->mm (1pt=0.3515 mm)
						   0.3515*Double.parseDouble(matcher.group(2)),//width
						   0.3515*Double.parseDouble(matcher.group(3)));//depth
				areDimensionsComputed=true;
			    }
			    catch(NumberFormatException e){
				System.out.println("Logfile number format problem: $line"+e.getMessage());
			    }
			    catch(IndexOutOfBoundsException e){
				System.out.println("Logfile regexp problem: $line"+e.getMessage());
			    }
			}
		}
		return;
	    }
	    
		/**
		 * Synchronizes bounding box with the model ; 
		 */
		protected void syncBounds(){
			PicText te = (PicText)element;			
			    //[pending] Il faut tenir compte de la rotation !

			Rectangle2D latexBB=null; // BB relative to latex dimensions (including rotation) (without frame)
			Rectangle2D textLayoutBB=null; // BB relative to textLayout dimensions (including rotation) (without frame)
			Rectangle2D textBB=null; // BB of the text (including rotation), without frame

			if (areDimensionsComputed){			//compute latexBB
			    Rectangle2D nonRotated = new Rectangle2D.Double(te.getLeftX(),te.getBottomY(),te.getWidth(),te.getHeight()+te.getDepth());
			    AffineTransform tr=new AffineTransform(); // maps Image coordinates to Model coordinates (see paint)
			    tr.setToIdentity(); // reset
			    tr.rotate(getRotation(),te.getPointX(TextEditable.P_ANCHOR),te.getPointY(TextEditable.P_ANCHOR)); // rotate along P_ANCHOR !
			    latexBB = tr.createTransformedShape(nonRotated).getBounds2D();
			}

			if (image==null) { 			//compute textLayoutBB
			    Rectangle2D nonRotated = textLayout.getBounds();
			    textLayoutBB = text2ModelTr.createTransformedShape(nonRotated).getBounds2D();
			}
			
			//use textLayoutBB or latexBB or their union
			if (image!=null) textBB=latexBB;
			else {
			    if (! areDimensionsComputed) textBB=textLayoutBB;
			    else {
				textBB=latexBB.createUnion(textLayoutBB);
			    }
			}

			//union with frame BB
			if (te.isFramed()) {
			    super.syncBounds();//update bounds of the frame if necessary
			    Rectangle2D.union(super.bounds,textBB,this.bounds);
			}
			else this.bounds=textBB;
		}

		/**
		 * Render the View to the given graphic context.
		 * This implementation render the interior first, then the outline.
		 */
		public void paint(Graphics2D g,Rectangle2D a){
			if (!a.intersects(getBounds())) return;
			if (image!=null){ // paint bitmap
				g.drawImage(image, text2ModelTr, null);
				// debug:
				g.setPaint(Color.red); g.draw(this.bounds);
				super.paint(g,a); // possibly paint framebox if non-null
			}
			else { // paint textlayout
				super.paint(g,a); // possibly paint framebox if non-null

				AffineTransform oldAT = g.getTransform();
				// paint text in black
				g.setPaint(Color.black);
				// from now on, we work in Y-direct (<0) coordinates to avoid inextricable problems with font being mirrored...
				g.transform(text2ModelTr); // also include rotation
				textLayout.draw(g, 0.0f,0.0f);
				//[pending] ajouter un cadre si areDimensionsComputed (wysiwyg du pauvre)
				// get back to previous transform
				g.setTransform(oldAT);
				if (jpicedt.Log.DEBUG) {
					g.setPaint(Color.red);
					g.draw(bounds);
				}
			}
		}

		/**
		 * This implementation calls <code>super.hitTest</code> and returns the result if non-null
		 * (this should be a HitInfo.Point), 
		 * then returns a HitInfo.Interior if the mouse-click occured inside the text bound (as defined
		 * by text layout)
		 * 
		 * @return a HitInfo corresponding to the given mouse-event
		 */
		public HitInfo hitTest(PEMouseEvent e){

			// from Bitmap:
			if (image!=null){
				if (getBounds().contains(e.getPicPoint())) {
					return new HitInfo.Interior((PicText)element, e);
				}
				return null;
			}

			// from TextLayout:
			if (!getBounds().contains(e.getPicPoint())) return null;

			PicText te = (PicText)element;
			// recompute textlayout b-box, but store it in a temporary field !
			Rectangle2D tb = textLayout.getBounds();
			Shape text_bounds = text2ModelTr.createTransformedShape(tb);
			if (text_bounds.contains(e.getPicPoint())) {
				// [SR:pending] for the hitInfo to be reliable, getPicPoint() should first be transformed by
				//              inverse text2ModelTr ! (especially when rotationAngle != 0)
				TextHitInfo thi = textLayout.hitTestChar((float)(e.getPicPoint().x - strx), (float)(e.getPicPoint().y - stry)); // guaranteed to return a non-null thi
				return new HitInfo.Text((PicText)element, thi, e);
			}
			// test hit on textlayout's bounding rectangle :
			//else if (bounds.contains(e.getPicPoint())) return new HitInfo.Interior(element,e);
			return null;


		}

		/**
		 * [SR:pending] make this view implement aka TextEditableView interface (or something like it), where
		 * TextEditableView is a subinterface of View with text-editing specific capabilities.
		 * 
		 * Returns the TextLayout which is responsible for painting the textual content of this element
		 */
		public TextLayout getTextLayout(){
			return textLayout;
		}

		/** Return an affine transform which translat b/w the TextLayout coordinate system and the
		 *  jpicedt.graphic.model coordinate system.
		 * [SR:pending] refactor method name to something more explanatory
		 */
		public AffineTransform getTextToModelTransform(){
			return text2ModelTr;
		}

	} // TextView


} // DefaultViewFactory

