/*
 Highlighter.java - September 20, 2003 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.*;
import jpicedt.graphic.*;
import jpicedt.graphic.event.*;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.util.*;

/**
 * a view dedicated to painting an Element's highlighter, e.g. when the Element
 * is selected. This is a delegate for a hosting HighlightableView, and is created by a dedicated
 * HighlighterFactory.<br>  
 * This highlighter should provide the following capabilities :
 * <ul>
 * <li> Rendering highlighted parts of the given Element to a Graphics2D context ;
 * <li> Mouse-hit testing on the highlighted part only. 
 * </ul>
 * @since jpicedt 1.4
 * @author Sylvain Reynal
 * @version $Id: Highlighter.java,v 1.4.2.1 2007/09/02 11:56:29 reynal Exp $
 */
public interface Highlighter {

	/**
	 * Fetches the HighlighterFactory implementation that is feeding the view hierarchy. 
	 * @return the factory, null if none
	 */
	HighlighterFactory getHighlighterFactory();
	
	//////////////////////////////////////////////
	/// EVENT HANDLING
	/////////////////////////////////////////////
	
	/**
	 * Give notification from the model that a change occured for an element's highlighting
	 * this highlighter is responsible for rendering. 
	 */
	void changedUpdate(DrawingEvent.EventType eventType, double scale); 

	////////////////////////////////////////////////////////
	//// PAINT
	////////////////////////////////////////////////////////
	
	/**
	 * Render the Highlighter to the given graphic context.<br>
	 * @param allocation current clipping
	 * @param scale The current scale factor from model to screen for the Graphics2D context ;
	 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
	 *        same lenght on the screen whatever the scale factor that's set to the graphic context.
	 */
	void paint(Graphics2D g, Rectangle2D allocation, double scale);

	/**
	 * @return the bounds of this Highlighter<br> 
	 *         This will determine the clipping rectangle passed as a parameter to repaint
	 *         in the hosting view
	 */
	Rectangle2D getBounds();

	
	
	
	/////////////////////////////////////////////////////
	//// MOUSE
	/////////////////////////////////////////////////////
	
	/**
	 * @return a HitInfo corresponding to the given mouse-event on this Highlighter.
	 */
	HitInfo hitTest(PEMouseEvent e);
}

