/*
 HitInfo.java - February 9, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.TextEditable;
import jpicedt.graphic.model.BranchElement;
import jpicedt.graphic.event.PEMouseEvent;

import java.util.Collections;
import java.util.ArrayList;
import java.awt.font.TextHitInfo;

/**
 * A HitInfo encapsulates information related to a mouse click that occured on a particular area of an Element.
 * @author Sylvain Reynal
 * @since jpicedt 1.3.2
 * @version $Id: HitInfo.java,v 1.11.2.1 2007/09/02 11:56:29 reynal Exp $
 */
public class HitInfo {

	private Element clickedObject;
	private PEMouseEvent mouseEvent;
	
	/**
	 * Construct a new HitInfo when a click occured on an object's end-point
	 * @param clickedObject the Element on which the hit occured
	 * @param type the hit type
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public HitInfo(Element clickedObject, PEMouseEvent mouseEvent) {
		this.clickedObject = clickedObject;	
		this.mouseEvent = mouseEvent;
	}
		
	/**
	 * @return the Element on which the hit occured  
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public Element getTarget(){
		return clickedObject;
	}

	/**
	 * @return the mouse-event that triggered this HitInfo
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public PEMouseEvent getMouseEvent(){
		return mouseEvent;
	}

	///////////////////////////////// INTERIOR ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on the interior of a Element
	 */
	public static class Interior extends HitInfo {
		
		/**
		 * construct a new HitInfo.Interior that occured on the given Element
		 */
		public Interior(Element clickedObject, PEMouseEvent mouseEvent){
			super(clickedObject, mouseEvent);
		}

		public String toString(){
			return "HitInfo.Interior : target="+getTarget();
		}
	}	

	///////////////////////////////// POINT ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on one or several Element's control-points, the latter case
	 * occuring necessarily iff some control-points are identical.
	 */
	public static class Point extends HitInfo {
		
		private ArrayList pointIndexArray=new ArrayList(); // array of Integer's
		
		/**
		 * construct a new HitInfo.Point that occured on the given Element and on the given point index
		 */
		public Point(Element clickedObject, int pointIndex, PEMouseEvent mouseEvent){
			super(clickedObject, mouseEvent);
			pointIndexArray.add(new Integer(pointIndex));
		}

		/**
		 * construct a new HitInfo.Point that occured on the given Element and on the given points simultaneously.
		 * Note that the ArrayList argument is not deeply copied.
		 */
		public Point(Element clickedObject, ArrayList pointIndices, PEMouseEvent mouseEvent){
			super(clickedObject, mouseEvent);
			this.pointIndexArray = pointIndices;
		}

		/**
	 	 * Return the point index on which the mouse hit occured. If several control-points were hit simultaneously
		 * because they were located at the same place, return the first one in ascending order according to
		 * the natural ordering of control-point indices.
	 	 */
		public int getIndex(){
			return getIndex(0);
		}
		
		/**
		 * Return the ith point index on which the mouse hit occured. If only one control-point was hit, this is
		 * similar as <code>getIndex()</code>.
		 */
		public int getIndex(int i){
			return ((Integer)pointIndexArray.get(i)).intValue();
		}
		
		/**
		 * Return the number of points that were hit simultaneously.
		 */
		public int getNbHitPoints(){
			return pointIndexArray.size();
		}
		
		public String toString(){
			return "HitInfo.Point : target="+getTarget()+", idx="+getIndex();
		}
	}

	///////////////////////////////// STROKE ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on an Element. It includes
	 * information as of which stroke segment was hit (this may be used by the UI to know
	 * where to add new points).
	 */
	public static class Stroke extends HitInfo {
		
		private int clickedSegment;

		
		/**
		 * construct a new HitInfo.Stroke that occured on the stroke of the given Element
		 * @param clickedPoint the index (starting from 0) of the segment of the flattened path
		 * on which the hit occured.
		 */
		public Stroke(Element clicked, int clickedSegment, PEMouseEvent mouseEvent){
			super(clicked, mouseEvent);
			this.clickedSegment = clickedSegment;
		}

		/**
	 	 * @return return the child on which the mouse-click occured
	 	 * @since jPicEdt
	 	 * @author Sylvain Reynal
	 	 */
		public int getClickedSegment(){
			return clickedSegment;
		}

		public String toString(){
			return "HitInfo.Stroke : target="+getTarget()+", clickedSegment="+clickedSegment;
		}
		
	}


	///////////////////////////////// HIGHLIGHTER_STROKE ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on the stroke of a highlighter (eg tangents of a PicSmoothPolygon). It includes
	 * information as of which stroke segment was hit (this may be used by the UI to know
	 * where to add new points).
	 */
	public static class HighlighterStroke extends Stroke {
		
		public HighlighterStroke(Element clicked, int clickedSegment, PEMouseEvent mouseEvent){
			super(clicked, clickedSegment, mouseEvent);
		}
		public String toString(){
			return "HitInfo.HighlighterStroke : target="+getTarget()+", clickedSegment="+getClickedSegment();
		}
		
	}

	///////////////////////////////// ENVELOPE_STROKE ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on the envelope of an element (aka tangents of a PicSmoothPolygon actually). It includes
	 * information as of which stroke segment was hit (this may be used by the UI to know
	 * where to add new points).
	 * @deprecated use HighlighterStroke instead
	 */
	public static class EnvelopeStroke extends Stroke {
		
		public EnvelopeStroke(Element clicked, int clickedSegment, PEMouseEvent mouseEvent){
			super(clicked, clickedSegment, mouseEvent);
		}
		public String toString(){
			return "HitInfo.EnvelopeStroke : target="+getTarget()+", clickedSegment="+getClickedSegment();
		}
		
	}

	///////////////////////////////// COMPOSITE ////////////////////////////////////////
	/**
	 * Represent a particular type of Hit that occured on a child of a branch-element, when
	 * one whants to specify both the element and its child. This can be considered an extended
	 * version of HitInfo.Point where the the "clicked" child play the role of the 
	 * "clicked" control-point.
	 */
	public static class Composite extends HitInfo {
		
		private int clickedChildIndex;
		
		/**
		 * construct a new HitInfo.Point that occured in the given BranchElement, on
		 * the child with the given index.
		 */
		public Composite(BranchElement clicked, int clickedChildIndex, PEMouseEvent mouseEvent){
			super(clicked, mouseEvent);
			this.clickedChildIndex = clickedChildIndex;
		}

		/**
	 	 * @return return the child on which the mouse-click occured
	 	 * @since jPicEdt
	 	 * @author Sylvain Reynal
	 	 */
		public int getClickedChildIndex(){
			return clickedChildIndex;
		}

		/**
	 	 * @return return the child on which the mouse-click occured
	 	 * @since jPicEdt
	 	 * @author Sylvain Reynal
	 	 */
		public Element getClickedChild(){
			return ((BranchElement)getTarget()).getChildAt(clickedChildIndex);
		}
		public String toString(){
			return "HitInfo.Composite : target="+getTarget()+", clickedChild="+getClickedChild();
		}
	}
	
	///////////////////////////////// TEXT ////////////////////////////////////////
	/**
	 * Represent a Hit that occured on some interior area of a TextEditable. This is simply a convenient wrapper for 
	 * {@link java.awt.font.TextHitInfo java.awt.font.TextHitInfo}.
	 */
	public static class Text extends HitInfo.Interior {
		
		private TextHitInfo thi;
		
		/**
		 * construct a new HitInfo.Text that occured on the given TextEditable, fetching useful information
		 * from the given TextHitInfo.
		 */
		public Text(TextEditable clickedObject, TextHitInfo thi, PEMouseEvent mouseEvent){
			super(clickedObject, mouseEvent);
			this.thi = thi;
		}
		
		/**
		 * Returns the {@link java.awt.font.TextHitInfo java.awt.font.TextHitInfo} wrapped into this HitInfo.
		 */
		public TextHitInfo getTextHitInfo(){
			return thi;
		}

		public String toString(){
			return "HitInfo.Text : target="+getTarget()+", charIdx="+thi.getCharIndex()+", insertIdx="+thi.getInsertionIndex();
		}
	}	
}
 
