/*
 View.java - February 9, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.graphic.view;

import jpicedt.graphic.model.Element;
import jpicedt.graphic.model.Drawing;
import jpicedt.graphic.PECanvas;
import jpicedt.graphic.event.PEMouseEvent;
import jpicedt.graphic.event.DrawingEvent;

import java.awt.*;
import java.awt.geom.*;

/**
 * a View is a graphic representation of a Element. It can be rendered using its paint method.<p>
 * It provides the following capabilities :
 * <ul>
 * <li> Rendering to a Graphics2D context ;
 * <li> Highlighting ;
 * <li> Mouse-hit testing (this is here because a View knows better than anyone how a graphic element
 *      looks on the screen, and thus where a user should click to select/whatever/... the element). As
 *      an example, consider a rectangle element that has a "fill-color" attribute set to "green", and
 *      is indeed paint in green in some View implementation, but is transparent in some other implementation :
 *      then when the user clicks inside the rectangle, the first implementation should return a "hit-ok" while
 *      the second should return "hit-fail", even thought the element has the same "fill-color" in both cases.
 *      <br>
 *      View implementation that don't implement editing capabilities may have their hitTest method
 *      simply return null to signal that every mouse-click just fails, and their paintHighlighter do
 *      nothing.
 *      <br>
 *      [pending] Another approach would be to define aka EditableView interface inheriting from View and adding
 *      the afore-mentioned methods ?
 * </ul>
 * View's belong to a tree that has the same structure as the associated Drawing, since each View is
 * attached to an element in the tree.
 * @since jpicedt 1.3.2
 * @author Sylvain Reynal
 * @version $Id: View.java,v 1.5.2.1 2007/09/02 11:56:29 reynal Exp $
 *
 */
public interface View {

	/**
 	 * @return the element the View is responsible for rendering
	 */
	Element getElement();
	
	/**
 	 * set the element the View is responsible for rendering
	 */
	void setElement(Element e);

	///////////////////////////////////////////////////////////////////
	//// tree-structure
	///////////////////////////////////////////////////////////////////
	
	/**
	 * Returns the parent of the view, as given by the tree-structure the associated graphic element belongs to.
	 * @return the parent, null if none
	 */
	View getParentView();


	/**
	 * Fetches the container hosting the view.  This is useful for
	 * things like scheduling a repaint, finding out the host 
	 * components font, etc.  
	 *
	 * @return the container, null if none
	 */
	PECanvas getContainer();

	/**
	 * Fetches the ViewFactory implementation that is feeding the view hierarchy. 
	 * @return the factory, null if none
	 */
	ViewFactory getViewFactory();

	/**
	 * Fetch a Graphics for rendering.  This can be used to determine
	 * font characteristics, and will be different for a print view
	 * than a component view.
	 *
	 * @since 1.3
	 */
	Graphics getGraphics();

	/**
	 * Fetches the document associated with the view.
	 * @return the drawing, null if none
	 */
	Drawing getDrawing();

	
	//////////////////////////////////////////////
	/// EVENT HANDLING
	/////////////////////////////////////////////
	
	/**
	 * Give notification from the model that a change occured for an element this view is responsible
	 * for rendering. 
	 */
	void changedUpdate(DrawingEvent.EventType eventType); 

	////////////////////////////////////////////////////////
	//// PAINT
	////////////////////////////////////////////////////////
	
	/**
	 * ask the hosting container to repaint itself
	 * @param clip the clip rectangle in model-coordinate
	 */
	void repaint(Rectangle2D clip);
	
	/**
	 * Render the View of the underlying Element to the given graphic context.
	 * @param allocation the graphic clip
	 */
	void paint(Graphics2D g, Rectangle2D allocation);

	/**
	 * @return the bounds of this View<br> 
	 *         This will determine the clipping rectangle passed as a parameter to repaint.
	 */
	Rectangle2D getBounds();

	//////////////////////////////////////////////////////
	//// HIGHLIGHTER
	//////////////////////////////////////////////////////
	
	/**
	 * Returns the Highlighter responsible for rendering the highlighted part of this view.
	 * @return null if this view cannot be highlighted
	 */
	Highlighter getHighlighter();

	/**
	 * Sets the Highlighter responsible for rendering the highlighted part of this view.
	 * @param h the delegate ; null if this View mustn't support highlighting
	 */
	void setHighlighter(Highlighter h);
	
	/**
	 * Render the Highlighter to the given graphic context.<br>
	 * @param allocation current clipping
	 * @param scale The current scale factor from model to screen for the Graphics2D context ;
	 *        this may be used to scale down line thickess, etc... so that lines/rectangle/... appear with the
	 *        same lenght on the screen whatever the scale factor that's set to the graphic context.
	 */
	void paintHighlighter(Graphics2D g, Rectangle2D allocation, double scale);

	/////////////////////////////////////////////////////
	//// MOUSE
	/////////////////////////////////////////////////////
	
	/**
	 * @return a HitInfo corresponding to the given mouse-event
	 * @param isHighlightVisible whether the receiver should include the highlighter shapes (e.g. end-points)
	 *        in the click-sensitive area.
	 */
	HitInfo hitTest(PEMouseEvent e, boolean isHighlightVisible);
}

