/*
 MDIManager.java - January 1, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui;

import jpicedt.ui.internal.*;
import jpicedt.*;
import jpicedt.widgets.*;
import jpicedt.graphic.*;
import jpicedt.graphic.event.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.ui.*;
import jpicedt.ui.util.*;
import jpicedt.ui.action.*;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.util.*;
import java.beans.*;

/**
 * An abstract class that serves as the superclass for jPicEdt's Multiple Document Interface desktop managers.
 * <p>
 * Concrete implementation know how to add or remove a document (e.g. a PEDrawingBoard or a DockablePanel) to
 * the MDI interface. Each document is actually contained in a MDIComponent (i.e. either a JFrame or a JInternalFrame), 
 * which is created by the factory method createMDIComponent(). The list of MDIComponent's 
 * which currently belong to the desktop is hold in separate hash maps, one for each family of documents.
 * There are currently two kind of documents supported:
 * <ul>
 * <li> PEDrawingBoard's: they're directly wrapped into a MDIComponent. The "boardMap" hash map holds the list
 * 	    of boards belonging to the desktop, together with their UI delegate. The addDrawingBoard() and
 *      removeDrawingBoard() methods allows one to add or remove boards from the desktop. These methods wrap
 *      board into their UI delegate on-the-fly. Finally, the "activeBoard" variable maintains a reference on
 *      the currently active drawing board (or null if none is active and/or open).
 * <li> DockablePanel's: there's one more layer in this case, as 
 *      	(i) we wish to remember UI properties (location, size, title,...)
 *      		associated with each panel 
 *			(ii) this allows us to implement most methods related to DockablePanel management
 *      		directly in this class, instead of relying on concrete subclasses to do the job. 
 *      Hence: 
 *			(i) there's a "dockablePanelsMap" which holds the list of DockablePanel's
 *      		belonging to the desktop, together with a key which allows us to retrieve these 
 *				UI properties from jPicEdt's properties. 
 *			(ii) each DockablePanel, instead of being directly wrapped into an MDIComponent, rather holds
 *      a reference to it; this in particular allows us to create/hide/show/etc... DockablePanel's directly in this class,
 *      since we do not need to rely on a particular implementation of MDIComponent.  
 * </ul>
 *
 * @since jPicEdt 1.3
 * @author Sylvain Reynal
 * @version $Id: MDIManager.java,v 1.15.2.1 2007/09/02 11:56:30 reynal Exp $
 */
public abstract class MDIManager {
	
	/** key for persistent storage */
	public static final String KEY_MDIMANAGER = "ui.mdimgr";
	/** value for persistent storage */
	public static final String MDI_CHILDFRAMES="child-frames";
	/** value for persistent storage */
	public static final String MDI_STANDALONEFRAMES="Gimp-like";


	/** PropertyChangeEvent's name corresponding to a new board being activated */
	public static final String ACTIVE_BOARD_CHANGE = "active-board-change";

	/**
	 * a hash table used to remember dockable panels added to the desktop, with key=panel's title 
	 */
	protected HashMap dockablePanelsMap = new HashMap();
	
	/** 
	 * A hash table used to bind a PEDrawingBoard (the key) and its MDIComponent container (the value)
	 * This map plays pretty much the same role as JDesktopPane, but it's able to manage any desktop component
	 * implementing the MDIComponent interface, e.g. JFrame's as well as JInternalFrame's.
	 * In particular, since this map holds the list of PEDrawingBoard's together with their hosting MDIComponent, 
	 * it's at the core of the desktop management, e.g. for adding or removing boards properly.
	 */
	protected HashMap boardMap = new HashMap(); // [SR:pending] use WeakHashMap since there are cross-references ?
	
	/** 
	 * delegate which notifies dockable panels when a an event is triggered by a PEDrawingBoard 
	 */
	protected BoardEventHandler boardEventHandler = new BoardEventHandler();
	
	/** 
	 * Reference on the currently active board, or null if no board is currently open.
	 * This reference is constantly being updated by MDIDesktopManager when a MDIComponent
	 * gets (de)activated, and it contains a PEDrawingBoard.
	 */
	protected PEDrawingBoard activeBoard=null;
	
	/////////////////////// utilities ///////////////////////
	
	/**
	 * Create an MDIManager according to the given value.
	 * @param value one of MDI_CHILDFRAMES or MDI_STANDALONEFRAMES
	 */
	public static MDIManager createMDIManager(String value, PEProgressBar progress){
		if (value.equals(MDI_CHILDFRAMES)) return new InternalFrameMDIManager(progress);
		else return new JFrameMDIManager(progress);
	}
	
	public static String[] getInstalledMDIManagers(){
		return new String[]{MDI_CHILDFRAMES,MDI_STANDALONEFRAMES};
	}
		
	public static String getDefaultManagerName(){
		return MDI_CHILDFRAMES;
	}
	
	public abstract String getName();
	
	//////////////////////////////////////////////////////////////////////
	//// MDI Components interface and rel.
	//////////////////////////////////////////////////////////////////////
	
	/**
	 * A common interface for UI components being hosted by this MDIManager. 
	 * Concrete implementation may rely on a JFrame or a JInternalFrame.
	 * The need for this interface stems from the fact that the only superclass common to
	 * JFrame and JInternalFrame is java.awt.Container, which doesn't contain some very important methods
	 * for our purpose, e.g. dispose() or setDefaultCloseOperation(). Yet these method are implemented in both
	 * JFrame and JInternalFrame... (maybe this should be a useful interface added to javax.swing ???)
	 */
	public interface MDIComponent {
		
		/** set the visible state of the component */
		void setVisible(boolean v); // also in Component
		
		/** returns whether the panel is currently visible or not */
		boolean isVisible(); // also in Component
		
		/** gets the bounds of the component wrapped by this DockablePanel */
		Rectangle getBounds(); // also in Component

		/** makes this container unselected, unvisible or closed */
		void dispose(); // both in JInternalFrame and in JFrame
		
		/** set the title of this container */
		void setTitle(String title); // both in JInternalFrame and in JFrame
		
		/** 
		 * control the window-closing operation
		 * @param i see javax.swing.WindowConstants
		 */
		void setDefaultCloseOperation(int i); // both in JInternalFrame and in JFrame 
		
		void pack(); // both in JInternalFrame and in JFrame
		
		void setLocation(int x, int y); // also in Component
		
		Dimension getSize(); // also in Component
		
		void reshape(int x, int y, int w, int h); // in Component
		
		/**
		 * Returns the JPanel wrapped into this MDIComponent. In particular, this may be a PEDrawingBoard.
		 */
		JPanel getInnerPane(); // specific to MDIManager

	}
	
	/**
	 * Factory method that creates a MDIComponent that acts as a wrapper for the given pane.
	 */
	public abstract MDIComponent createMDIComponent(JPanel pane);
		
	/**
 	 * Adds the given component to the desktop of this MDIManager.
	 */
	public abstract void addMDIComponent(MDIComponent c);
	
	// note: removing MDI components is handled directly by WindowListener/InternalFrameListener's.
	
	/**
	 * Returns the dimension of the desktop that hosts the components of this MDIManager.
	 * Depending on the implementation, this may be either the whole screen, or the size of a JFrame
	 * containing all the documents.
	 */
	public abstract Dimension getDesktopSize();

	////////////////////////////////////////////////////////////////////////////
	//// PREFERENCES
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 * Updates local properties from the JPicEdt's preferences (e.g. GUI colours, rendering-hints,...)
	 * This implementation update the rendering hints of each currently opened PEDrawingBoard
	 */
	public void update(){
		// rendering hints :
		RenderingHints rh = MiscUtilities.parseRenderingHints(JPicEdt.getPreferences());
		PEDrawingBoard[] boards = getAllDrawingBoards();
		for (int i=0; i<boards.length; i++){
			boards[i].getCanvas().getRenderingHints().add(rh);
			boards[i].getCanvas().repaint();
		}
		// [underway]
		// updates colors (they might've been modified from inside Preference tabpane) :
		//BoardInternalFrame activeFrame = getActiveDrawingBoard();
		//if (activeFrame != null) activeFrame.getCanvas().repaint(); // Grid, text and bg colours may have changed
	}

	/**
	 * Saves UI geometry to JPicEdt's preferences.<br>This implementation saves geometrical properties related
	 * to DockablePanel only.
	 * @see #addDockablePanel
	 * @see DockablePanel#setVisible
	 */
	public void saveGeometry(){
		// save dockable toolbars geometry :
		Properties preferences = JPicEdt.getPreferences();
		Set keys = dockablePanelsMap.keySet();
		for (Iterator it = keys.iterator(); it.hasNext();){
			String key = (String)it.next();
			DockablePanel tif = (DockablePanel)dockablePanelsMap.get(key);
			if (tif == null) continue;
			Rectangle bounds = tif.getMDIDelegate().getBounds();
			preferences.setProperty("ui."+key+".x",Integer.toString(bounds.x));
			preferences.setProperty("ui."+key+".y",Integer.toString(bounds.y));
			preferences.setProperty("ui."+key+".width",Integer.toString(bounds.width));
			preferences.setProperty("ui."+key+".height",Integer.toString(bounds.height));
			preferences.setProperty("ui."+key+".visible",tif.getMDIDelegate().isVisible()?"true":"false");
		}
		
	}

	////////////////////////////////////////////////////////////////////////
	//// GUI components: status bar, menu bar, popup menu,...
	////////////////////////////////////////////////////////////////////////
	
	/**
	 * Requests that the argument string  be displayed in the status bar
	 */
	public void showStatus(String str){
		PEDrawingBoard b = getActiveDrawingBoard();
		if (b!=null) b.getStatusBar().showMessage(str);
	}
	
	/**
	 * Factory method to create a PopupMenuFactory for newly created board
	 */
	public PopupMenuFactory createPopupMenuFactory(){
		PopupMenuFactory p = new PEPopupMenuFactory();
		return p;
	}
	
	/**
	 * Update JMenuItem's accelerators from the key/value pairs stored in JPicEdt's preferences 
	 */
	public abstract void updateAccelerators();

	/**
	 * Update the content (i.e. PEMenu's) of the menu-bar depending on the current state of the MDIManager.
	 */
	public abstract void updateMenuBar();
	
	/** 
	 * Updates the "recent files" sub-menu items after a frame was saved or opened.
	 * @since PicEdt 1.1
	 */
	public abstract void updateRecentFilesSubmenu();
	
	/** 
	 * Updates the "script" menu items 
	 * @since jpicedt 1.3.2
	 */
	public abstract void updateScriptsMenu();

	/** 
	 * Updates the "fragments" menu items 
	 * @since jpicedt 1.3.2
	 */
	public abstract void updateFragmentsMenu();

	/**
	 * update "undo" and "redo" text
	 * @since jpicedt 1.3.2
	 */
	public abstract void updateUndoRedoMenus(String undoName,String redoName);

	/** 
	 * Updates the Window menu as soon as an internal frame is opened or closed 
	 * (this menu keeps tracks of every open internal frames so that the user can activate 
	 * them quickly by use of a shortcut)
	 * 
	 * @since PicEdt 1.1
	 */
	public abstract void updateWindowMenu() throws MissingResourceException ;

	/**
	 * update currently "selected" menuitem in Toolkit menu
	 */
	public abstract void updateToolkitMenu(PropertyChangeEvent e);
	
	/**
	 * Return an array containing all the menu-items components attached to menubars
	 */
	public abstract JMenuItem[] getMenuItems();

	//////////////////////////////////////////////////////////////////////
	//// BOARDS MANAGEMENT
	//////////////////////////////////////////////////////////////////////

	/**
	 * Returns the MDIComponent that contains the given board
	 */
	public MDIComponent getHostingContainer(PEDrawingBoard board){
		return (MDIComponent)boardMap.get(board);
	}

	/**
	 * Sets the title of the frame/internal frame hosting the given board
	 * to the given string.<p>
	 */
	public void setHostingContainerTitle(String title, PEDrawingBoard board){

		MDIComponent c = getHostingContainer(board);
		if (c==null) return; // no frame hosting this board
		c.setTitle(title);
	}
	
	/**
	 * Adds a new DrawingBoard component to the list of drawing boards. The given board gets wrapped
	 * into a MDIComponent beforehands
	 */
	public void addDrawingBoard(PEDrawingBoard board){
		// !!! registering the selection-listener BEFORE adding the board to the DesktopPane allows dockable
		// toolbars (e.g. the attribute editor) which are instances of SelectionListener to be notified that a new board has been activated,
		// thus giving'em a chance to synchronize their state with that of the new active board.
		board.getCanvas().addSelectionListener(boardEventHandler);
		board.getCanvas().addUndoableEditListener(boardEventHandler);
		board.getCanvas().addPropertyChangeListener(boardEventHandler);
		board.getCanvas().getEditorKit().addPropertyChangeListener(boardEventHandler);
		board.setPopupMenuFactory(createPopupMenuFactory());

		MDIComponent mdiComponent = createMDIComponent(board); // wrap "board" into a MDIComponent (e.g. a JInternalFrame)
		mdiComponent.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE); // the only way to close it is through a call to closeFrame()
		boardMap.put(board, mdiComponent); // remember which MDIComponent is associated with each board
		
		mdiComponent.setTitle(board.getTitle());
		mdiComponent.pack();
		addMDIComponent(mdiComponent);
		int nbFrame = getDrawingBoardCount();
		//int dist = f.getHeight() - f.getContentPane().getHeight(); // title bar height DOESN'T WORK !
		int dist=30;
		mdiComponent.setLocation(nbFrame * dist, nbFrame * dist); 
		mdiComponent.setVisible(true);
		JPicEdt.getActionRegistry().updateActionState(); 
		updateMenuBar();
	}

	/**
	 * Ask the given board to close itself
	 */
	public abstract void closeBoard(PEDrawingBoard board);
	
	/**
	 * Remove the given drawing board from the list of open drawing boards, possibly attempting to
	 * save its content if necessary, then unregisters any previously registered listener.
	 * Closing the associated UI (i.e. the MDIComponent) is the responsibility of the caller.
	 * <p>
	 * @return false if the operation failed because the action was cancelled by the user
	 */
	protected boolean removeDrawingBoard(PEDrawingBoard board){
		// need saving ?
		if (board.isDirty()){
			int answer = JOptionPane.showConfirmDialog(null,
		    	         Localizer.currentLocalizer().get("msg.SaveChanges"),
						 Localizer.currentLocalizer().get("action.ui.Close"),
						 JOptionPane.YES_NO_CANCEL_OPTION);

			if (answer == JOptionPane.YES_OPTION) JPicEdt.saveBoard(false); // doesn't prompt user for file name if not necessary
			else if (answer == JOptionPane.CANCEL_OPTION) {
				//Log.debug("answer=cancel");
				return false;
			}
		}
			
		// remove listeners
		board.getCanvas().removeSelectionListener(boardEventHandler);
		board.getCanvas().removeUndoableEditListener(boardEventHandler);
		board.getCanvas().removePropertyChangeListener(boardEventHandler);
		board.getCanvas().getEditorKit().removePropertyChangeListener(boardEventHandler);
		
		// remove from hash map
		boardMap.remove(board); // remove cross-references (at least i hope so ;-)
		
		return true;
	}

	/**
	 * Returns all opened drawing boards
	 */
	public PEDrawingBoard[] getAllDrawingBoards(){
		Set keys = boardMap.keySet();
		PEDrawingBoard[] boardArray = new PEDrawingBoard[keys.size()];
		return (PEDrawingBoard[])keys.toArray(boardArray);
	}

	/**
	 * Selects the given drawing board ; concrete implementation should call show() or any similar method
	 * on the MDIComponent (aka container) hosting the given board.
	 */
	public abstract void selectDrawingBoard(PEDrawingBoard board);

	/**
	 * Returns the number of open boards
	 */
	public int getDrawingBoardCount(){
		return boardMap.size();
	}

	//////////////////////////////////////////////////////////////////////
	//// DESKTOP MANAGER 
	//////////////////////////////////////////////////////////////////////

	/**
	 * Returns the currently active drawing board 
	 */
	public abstract PEDrawingBoard getActiveDrawingBoard(); 

	//////////////////////////////////////////////////////////////////////
	//// DOCKABLE PANES
	//////////////////////////////////////////////////////////////////////
	
	/**
	 * Adds the given pane to the hashtable of dockable panels, then shows it.
	 * If "pane" is an instance of SelectionListener (resp. PropertyChangeListener), 
	 * it will be notified selection (resp. propery-change) events from any
	 * currently opened drawing board.
	 * @param key used to retrieve geometrical properties from the JPicEdt's preferences, i.e.
	 *        "ui." + key + ".visible" for the visible state at init time.
	 */
	public void addDockablePanel(String key, JPanel pane){
		//jpicedt.Log.debug("key="+key+",pane="+pane+",map="+dockablePanelsMap);
		if (dockablePanelsMap.containsKey(key)) return; // never add the same pane twice
		dockablePanelsMap.put(key,new DockablePanel(pane,key));
		// retrieve things saved last time by saveGeometry() :
		boolean vis = JPicEdt.getProperty("ui."+key+".visible",true);
		if (vis) showDockablePanel(key,true);
	}
	
	/**
	 * Toggle the visible state of the dockable pane with the given key.
	 */
	public void toggleDockablePanel(String key){
		//jpicedt.Log.debug("key="+key+",vis="+visible+",map="+dockablePanelsMap);
		if (!dockablePanelsMap.containsKey(key)) return;
		DockablePanel dp = (DockablePanel)dockablePanelsMap.get(key);
		dp.toggleVisible();
	}

	/**
	 * Set the visible state of the dockable pane with the given key.
	 */
	public void showDockablePanel(String key,boolean visible){
		//jpicedt.Log.debug("key="+key+",vis="+visible+",map="+dockablePanelsMap);
		if (!dockablePanelsMap.containsKey(key)) return;
		DockablePanel dp = (DockablePanel)dockablePanelsMap.get(key);
		dp.setVisible(visible);
	}

	/**
	 * Returns a set containing all keys (of class "String") that map to a DockablePanel
	 */
	public Set getDockablePanelKeys(){
		return dockablePanelsMap.keySet();
	}
	
	/**
	 * Return the inner pane contained in the dockable panel 
	 * associated with the given key, or null if none match.
	 */
	public JPanel getDockablePanelFromKey(String key){
		return ((DockablePanel)dockablePanelsMap.get(key)).getMDIDelegate().getInnerPane();
	}

	/**
	 * A wrapper for dockable panels, e.g. the attribute editor or the toolkit palette. 
	 * The UI is rendered by a MDIComponent delegate  (see {@link #getMDIDelegate getMDIDelegate}) 
	 * that is created at init time by relying on the factory method {@link #createMDIComponent createMDIComponent}.
	 */
	public class DockablePanel {
		
		//private JPanel content;
		private boolean packed=false;
		private String key;
		private MDIComponent mdiComponent; // the UI delegate which renders this panel

		/**
		 * @param pane the content of this DockablePanel
		 * @param key used to set the title (after proper i18n'ing) of the MDIComponent which renders the UI, 
		 *        and for building keys used to retrieve the geometry from JPicEdt's preferences.
		 */
		public DockablePanel(JPanel pane,String key){
			mdiComponent = createMDIComponent(pane);
			mdiComponent.setTitle(jpicedt.Localizer.currentLocalizer().get(key)); 
			//this.content = pane;
			this.key = key;
			mdiComponent.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE); // the only way to close it is through a call to toggleDockablePanel()
			addMDIComponent(mdiComponent);
		}
		
		/** 
		 * returns the MDIComponent that renders the UI of this DockablePanel 
		 */
		public MDIComponent getMDIDelegate(){
			return mdiComponent;
		}
		
		/** returns true if this DockablePanel is currently visible 
		public boolean isVisible(){
			return mdiComponent.isVisible();
		}*/
		
		public void toggleVisible(){
			setVisible(!mdiComponent.isVisible());
		}

		/** 
		 * set the visible state of this DockablePanel, first trying to fetch the geometrical aspect from
		 * the Properties object given in the constructor. This is done on first call only (see "packed" flag).
		 */
		public void setVisible(boolean v){
			if (v==true && !packed && MDIManager.this!=null) {
				getMDIDelegate().pack();
				packed=true;
				// set default location :
				Dimension mainFrameSize = getDesktopSize();
				Dimension iframeSize = getMDIDelegate().getSize();
				// crop panel size to main frame size if necessary :
				if (iframeSize.height > mainFrameSize.height) iframeSize.height = mainFrameSize.height;
				if (iframeSize.width > mainFrameSize.width) iframeSize.width = mainFrameSize.width;
				int x = mainFrameSize.width - iframeSize.width;
				int y = (mainFrameSize.height - iframeSize.height) / 2;
				//jpicedt.Log.debug(y="+y);
				getMDIDelegate().setLocation(x,y);
				// let's try to override it with preferences :
				x = JPicEdt.getProperty("ui."+key+".x",x);
				y = JPicEdt.getProperty("ui."+key+".y",y);
				int w = JPicEdt.getProperty("ui."+key+".width",getMDIDelegate().getSize().width);
				int h = JPicEdt.getProperty("ui."+key+".height",getMDIDelegate().getSize().height);
				// work-around for null dimension inadvertently saved in the preference file :
				if (w==0) w = getMDIDelegate().getSize().width; // i.e. as computed from pack() 
				if (h==0) h = getMDIDelegate().getSize().height;
				// avoid panel sticking out of the main frame area :
				if (x+w > mainFrameSize.width) x = mainFrameSize.width - w;
				if (y+h > mainFrameSize.height) y = mainFrameSize.height - h;
				getMDIDelegate().reshape(x,y,w,h);
			}
			getMDIDelegate().setVisible(v);
		}
		
	}
	

	//////////////////////////////////////////////////////////////////////
	//// EVENTS
	//////////////////////////////////////////////////////////////////////
	
	/**
	 * A handler for events triggered from any PEDrawingBoard, its PECanvas or its EditorKit.
	 */
	protected class BoardEventHandler implements SelectionListener, UndoableEditListener, PropertyChangeListener {

		/**
		 * Handler for "Selection" events triggered e.g. by a click on a graphical object (from a given PEDrawingBoard)
		 * This takes the following actions:
		 * <ul>
		 * <li> marks the active drawing board as dirty and possibly adds a trailing "*" to the frame's title, 
		 * <li> updates widgets states (ie menu items + toolbar buttons),
		 * <li> eventually notifies interested dockable panels (through the "dockablePanelsMap" hash table)
		 * 		that the state of the Drawing has changed (for instance, say there's a 
		 *      graphical object filled in red and we just selected it -> changes 
		 * 		the current active fill colour to red in the attributes editor palette).
		 * </ul>
		 */
		public void selectionUpdate(SelectionEvent e){
			//jpicedt.Log.debug("evt="+e);
			getActiveDrawingBoard().setDirty(true); // moved from PEDrawingBoard ; see note in "PEDrawingBoard.init()"
			JPicEdt.getActionRegistry().updateActionState();
			// update hosting frame title
			String title  = getActiveDrawingBoard().getTitle() + "* [" + getActiveDrawingBoard().getCanvas().getContentType().getPresentationName() + "]";
			setHostingContainerTitle(title,getActiveDrawingBoard());
			// notifies dockable panels (those which are instances of SelectionListener) 
			// of possible changes in the active Drawing:
			Collection c = dockablePanelsMap.values();
			for (Iterator it = c.iterator(); it.hasNext();){
				JPanel p = ((DockablePanel)it.next()).getMDIDelegate().getInnerPane();
				if (p instanceof SelectionListener){
					((SelectionListener)p).selectionUpdate(e);
				}
			}
		}
	
		/**
		 * Handler for "UndoableEdit" events sourced by PECanvas. Updates undo/redo menu items and
		 * corresponding toolbar buttons.
		 */
		public void undoableEditHappened(UndoableEditEvent e){
			PEDrawingBoard board = getActiveDrawingBoard();
			if (board==null) return;
			String redoName = board.getCanvas().getRedoPresentationName();
			String undoName = board.getCanvas().getUndoPresentationName();
			updateUndoRedoMenus(undoName,redoName);
			JPicEdt.getActionRegistry().updateActionState();
		}
		
		/**
		 * Handler for "PropertyChange" events sources from either PECanvas or its EditorKit.
		 * Simply forwards these events to all DockablePanel's which implement the PropertyChangeListener
		 * interface, as well as to the attached PEMenuBar.
		 */
		public void propertyChange(PropertyChangeEvent e){
			//jpicedt.Log.debug("evt="+e.getPropertyName()+":"+e.getOldValue()+"->"+e.getNewValue());
			Collection c = dockablePanelsMap.values();
			for (Iterator it = c.iterator(); it.hasNext();){
				JPanel p = ((DockablePanel)it.next()).getMDIDelegate().getInnerPane();
				if (p instanceof PropertyChangeListener){
					((PropertyChangeListener)p).propertyChange(e);
				}
			}
			updateToolkitMenu(e); // update currently "selected" menuitem in Toolkit menu
		}
	}
	
	//////////////////////////////////////////////////////////////////////
	//// GEOMETRY
	//////////////////////////////////////////////////////////////////////

	/**
	 * Rearranges all open drawing-boards by cascading them
	 */
	public abstract void cascadeDrawingBoards();

	/**
	 * Tiles all open drawing-boards horizontally 
	 */
	public abstract void tileDrawingBoardsHorizontally();

	/**
	 * Tiles all open drawing-boards vertically
	 */
	public abstract void tileDrawingBoardsVertically();

} // class

