/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C)   Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.ui.action;

import jpicedt.ui.*;
import jpicedt.ui.util.*;
import jpicedt.Localizer;
import jpicedt.JPicEdt;
import jpicedt.Log;
import jpicedt.graphic.toolkit.ActionLocalizer;
import jpicedt.graphic.toolkit.ActionDispatcher;
import jpicedt.graphic.toolkit.EditorKit;
import jpicedt.graphic.toolkit.DefaultSelectionHandler;
import jpicedt.graphic.toolkit.PEAction;
import jpicedt.graphic.toolkit.PEToggleAction;
import jpicedt.graphic.event.SelectionListener;
import jpicedt.graphic.event.SelectionEvent;

import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.awt.datatransfer.*;

/**
 * A registry that retains all JPicEdt's actions in a map, 
 * supplies them on-demand through a String key, and knows how to update actions state using
 * the currently installed MDIManager.
 * @since jPicEdt 1.3
 * @author Sylvain Reynal
 * @version $Id: ActionRegistry.java,v 1.14 2006/06/17 22:34:47 reynal Exp $
 *
 */
public class ActionRegistry extends ActionMap {

	/**
	 * construct an ActionRegistry with default Actions
	 */
	public ActionRegistry(){
		ActionLocalizer localizer = jpicedt.Localizer.currentLocalizer().getActionLocalizer();
		ActionDispatcher dispatcher = JPicEdt.getActionDispatcher();

		// set EditorKit's action-map as the parent :
		ActionMap kitMap = EditorKit.createActionMap(dispatcher,localizer);
		setParent(kitMap);

		addAction(new JPicEdt.ExitAction());

		// file I/O :
		addAction(new JPicEdt.FragmentInsertAction());
		addAction(new JPicEdt.FragmentSaveAction());
		addAction(new JPicEdt.FileNewAction());
		addAction(new JPicEdt.FileOpenAction());
		addAction(new JPicEdt.FileReloadAction());
		addAction(new JPicEdt.FileSaveAction());
		addAction(new JPicEdt.FileSaveAllAction());
		addAction(new JPicEdt.FileSaveAsAction());
		addAction(new JPicEdt.FileCloseAction());

		// Editing :
		addAction(new JPicEdt.RedrawAction());
		addAction(new JPicEdt.PageFormatAction());
		addAction(new JPicEdt.ViewLaTeXFileAction());
		addAction(new JPicEdt.UndoAction());
		addAction(new JPicEdt.RedoAction());

		// BeanShell :
		addAction(new JPicEdt.OpenBSHConsoleAction());

		// Content-type :
		for (int i=0; i<jpicedt.MiscUtilities.getAvailableContentTypes().length; i++){
			addAction(new JPicEdt.ContentTypeAction(i)); // key = content-type NAME !!!
		}
		// Dockable panels :
		addAction(new JPicEdt.ToggleAttributesPanelAction());
		addAction(new JPicEdt.ToggleToolkitPanelAction());

		// Preferences :
		addAction(new JPicEdt.PreferencesAction());

		// Help and rel.
		addAction(new JPicEdt.AboutAction());
		addAction(new JPicEdt.ShowHelpAction());
		addAction(new JPicEdt.ShowLicenseAction());
		addAction(new JPicEdt.MemoryMonitorAction());
		addAction(new JPicEdt.ShowErrorLogAction());

		// Window placement :
		addAction(new JPicEdt.WindowCascadeAction());
		addAction(new JPicEdt.WindowTileHorAction());
		addAction(new JPicEdt.WindowTileVertAction());

		//External commands :
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.LATEX));
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.DVIVIEWER));
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.DVIPS));
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.GHOSTVIEW));
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.USER1));
		addAction(new JPicEdt.RunExternalCommandAction(RunExternalCommand.USER2));

		// override EditorKit.SelectMouseToolAction (which dispatch to a single PECanvas) with those in jpicedt.ui.action:
		// (reminder: map resolves to parent only if the key is not found here)
		String[][] allToolNames = EditorKit.getAvailableToolNames();
		for (int i=0; i<allToolNames.length; i++){
			for (int j=0; j<allToolNames[i].length; j++){
				addAction(new SelectEditorkitMouseToolAction(localizer,allToolNames[i][j]));
			}
		}
		// ibid. for local/global mode toggling:
		addAction(new ToggleEditorkitEditPointsModeAction(localizer));
	}


	/**
	 * @param the Action to add 
	 */
	public void addAction(Action a){
		put(a.getValue(Action.ACTION_COMMAND_KEY), a);
	}
	/**
	* @return the action whose name maps "name"
	*/
	public Action getAction(String name){
		Action a = get(name);
		if (a==null) {
			if (Log.DEBUG) Log.debug(name + " was not found!");
			return new DummyAction(name);
		}
		else return a;
	}

	/**
	 * update Action's enabled-state according to the current state of the installed MDIManager. 
	 * This may be called when a frame gets activated, or when a user action occured that 
	 * implies updating action states, e.g. "Save", "Open", edit active board, ...
	 */
	public void updateActionState(){

		// test if no more frame in the desktop
		// => we must disable all actions but those valid even if there are no frames in the desktop !
		if (jpicedt.JPicEdt.getMDIManager().getDrawingBoardCount() == 0){
			switchToNoFrameConfiguration();
			return;
		}

		// else fetch active board :
		PEDrawingBoard board = jpicedt.JPicEdt.getMDIManager().getActiveDrawingBoard();
		if (board == null) {
			switchToNoFrameConfiguration(); // should never happen
			return;
		}

		// ok, now there's at least one board in the desktop :
		switchToMinimalConfiguration();

		// update "Save" Action's state : if the frame's already been saved to disk
		// and not modified yet => disable action
		getAction(JPicEdt.FileSaveAction.KEY).setEnabled(board.isDirty() || !board.isSaved());

		// updating "Reload" Action's state : disable if the frame's hasn't yet been saved to disk,
		getAction(JPicEdt.FileReloadAction.KEY).setEnabled(board.isSaved());

		// update "SaveAll" action (very very tricky, boy !)
		updateSaveAllAction();

		// update "Undo/Redo" actions, depending on this internal frame's  UndoableBuffer state :
		if (board.getCanvas().canRedo()) getAction(JPicEdt.RedoAction.KEY).setEnabled(true);
		else getAction(JPicEdt.RedoAction.KEY).setEnabled(false);
		if (board.getCanvas().canUndo()) getAction(JPicEdt.UndoAction.KEY).setEnabled(true);
		else getAction(JPicEdt.UndoAction.KEY).setEnabled(false);

		// update Cut, Copy, Delete and SaveSelectionAs actions, depending on the size of the active selection:
		boolean b = !(board.getCanvas().getSelectionSize()==0);
		getAction(JPicEdt.FragmentSaveAction.KEY).setEnabled(b); // disabled when no active selection
		getAction(EditorKit.CopyAction.KEY).setEnabled(b);
		getAction(EditorKit.CutAction.KEY).setEnabled(b);
		getAction(EditorKit.DeleteAction.KEY).setEnabled(b);

		// content type
		jpicedt.graphic.ContentType contentType = board.getCanvas().getContentType();
		String contentTypeClassName = contentType.getClass().getName();
		for (int i=0; i<jpicedt.MiscUtilities.getAvailableContentTypes().length; i++){
			String availableClassName = jpicedt.MiscUtilities.getAvailableContentTypes()[i];
			boolean match = availableClassName.equals(contentTypeClassName);
			getAction(jpicedt.MiscUtilities.getAvailableContentTypesNames()[i]).setEnabled(!match);
		}

	}

	/**
	 * update "SaveAll" action depending on the "isDirty" flag of every currently open board
	 */
	private void updateSaveAllAction(){
		if (jpicedt.Log.DEBUG) Log.error("NOT IMPLEMENTED !!! [underway]");
		/* [pending] UNDERWAY
		JInternalFrame[] allFrames = desktopPane.getAllFrames();
		for (int i = 0; i < allFrames.length; i++){
			InternalFrame ifr = (InternalFrame)allFrames[i];
			PEDrawingBoard board = ifr.getBoard();
			if (board.isDirty() || !board.isSaved()) {
				actionRegistry.getAction(JPicEdt.FileSaveAllAction.KEY).setEnabled(true);
				return;
			}
	}
		actionRegistry.getAction("SaveAll").setEnabled(false);
		*/
	}

	/**
	 * disable actions that make no sense when no frame is present in the desktop
	 */
	private void switchToNoFrameConfiguration(){

		getAction(JPicEdt.FragmentInsertAction.KEY).setEnabled(false);
		getAction(JPicEdt.FileReloadAction.KEY).setEnabled(false);
		getAction(JPicEdt.FileSaveAction.KEY).setEnabled(false);
		getAction(JPicEdt.FragmentSaveAction.KEY).setEnabled(false);
		getAction(JPicEdt.FileSaveAllAction.KEY).setEnabled(false);
		getAction(JPicEdt.FileSaveAsAction.KEY).setEnabled(false);
		getAction(JPicEdt.FileCloseAction.KEY).setEnabled(false);
		getAction(EditorKit.CopyAction.KEY).setEnabled(false);
		getAction(EditorKit.CutAction.KEY).setEnabled(false);
		getAction(EditorKit.DeleteAction.KEY).setEnabled(false);
		getAction(EditorKit.PasteAction.KEY_PASTE).setEnabled(false);
		getAction(EditorKit.SelectAllAction.KEY).setEnabled(false);
		getAction(JPicEdt.WindowCascadeAction.KEY).setEnabled(false);
		getAction(JPicEdt.WindowTileHorAction.KEY).setEnabled(false);
		getAction(JPicEdt.WindowTileVertAction.KEY).setEnabled(false);
		getAction(JPicEdt.UndoAction.KEY).setEnabled(false);
		getAction(JPicEdt.RedoAction.KEY).setEnabled(false);
		getAction(JPicEdt.RedrawAction.KEY).setEnabled(false);
		getAction("LaTeX").setEnabled(false);
		getAction("DVI").setEnabled(false);
		getAction("Dvips").setEnabled(false);
		getAction("Ghostview").setEnabled(false);
		getAction("UserProgram1").setEnabled(false);
		getAction("UserProgram2").setEnabled(false);
		// content type
		for (int i=0; i<jpicedt.MiscUtilities.getAvailableContentTypes().length; i++){
			getAction(jpicedt.MiscUtilities.getAvailableContentTypesNames()[i]).setEnabled(false);
		}
	}

	/**
	 * enable a minimal set of actions, when a new frame is created in an initially empty desktop.
	 */
	private void switchToMinimalConfiguration(){

		getAction(JPicEdt.FragmentInsertAction.KEY).setEnabled(true);
		getAction(JPicEdt.FileSaveAsAction.KEY).setEnabled(true);
		getAction(JPicEdt.FileSaveAllAction.KEY).setEnabled(true); // [pending] delete once we fix "updateSaveAllAction"
		getAction(JPicEdt.FileCloseAction.KEY).setEnabled(true);
		getAction(EditorKit.SelectAllAction.KEY).setEnabled(true);
		getAction(JPicEdt.WindowCascadeAction.KEY).setEnabled(true);
		getAction(EditorKit.PasteAction.KEY_PASTE).setEnabled(true);
		getAction(JPicEdt.WindowTileHorAction.KEY).setEnabled(true);
		getAction(JPicEdt.WindowTileVertAction.KEY).setEnabled(true);
		getAction("Redraw").setEnabled(true);
		getAction("LaTeX").setEnabled(true);
		getAction("DVI").setEnabled(true);
		getAction("Dvips").setEnabled(true);
		getAction("Ghostview").setEnabled(true);
		getAction("UserProgram1").setEnabled(true);
		getAction("UserProgram2").setEnabled(true);
	}

	// DEBUGGING : work-around for action that don't exist
	private class DummyAction extends AbstractAction {

		public DummyAction(String s){
			super(s+":NotFound!");
		}
		public void actionPerformed(ActionEvent e){}
	}

	/////////////////////////////////////////////////////////////
	/// Some actions overriden from EditorKit
	/////////////////////////////////////////////////////////////

	/**
	* Sets the current mouse-tool for every EditorKit attached to an open PEDrawingBoard. 
	* Action commands are the same as the predefined mouse-tool's names for EditorKit.
	* This action dispatches to all currently active EditorKit's, by relying on the current MDIManager.
	*/
	static class SelectEditorkitMouseToolAction extends PEToggleAction {

		private String mouseToolName;

		/**
		 * Construct a new SelectEditorkitMouseToolAction for the given mouse-tool name.
		 * @param mouseToolName one of EditorKit's predefined mouse-tool constant, e.g. EditorKit.ROTATE.
		 */
		public SelectEditorkitMouseToolAction(ActionLocalizer localizer,String mouseToolName){
			super(null, mouseToolName, localizer); // no dispatcher
			this.mouseToolName = mouseToolName;
		}

		/** overriden so as to dispatch to all currently open EditorKit's */
		public void actionPerformed(ActionEvent e){
			MDIManager mdiManager = JPicEdt.getMDIManager();
			//mdiManager.getToolBar().updateCurrentTool(mouseToolName); [pending]
			PEDrawingBoard[] allOpenBoards = mdiManager.getAllDrawingBoards();
			for (int i=0; i<allOpenBoards.length; i++){
				EditorKit kit = allOpenBoards[i].getCanvas().getEditorKit();
				if (kit != null) kit.setCurrentMouseTool(mouseToolName);
			}
		}
	}


	/**
	 * Toggles the EditorKit's edit-points-mode b/w LOCAL_MODE and GLOBAL_MODE, for every
	 * EditorKit attached to an open PEDrawingBoard. 
	 * This action dispatches to all currently active EditorKit's, by relying on the current MDIManager.
	 * <p>
	 * If this action is attached to an AbstractButton, then the selection state of the button is used to
	 * toggle the edit-mode between LOCAL (not-selected) and GLOBAL (selected). Otherwise, a simple toggling occurs.
	 */
	static class ToggleEditorkitEditPointsModeAction extends PEToggleAction {

		/* same KEY as the corresponding action in EditorKit */
		public final static String KEY = EditorKit.ToggleEditPointsModeAction.KEY;

		/**
		 * Construct a new SelectEditorkitMouseToolAction for the given mouse-tool name.
		 * @param mouseToolName one of EditorKit's predefined mouse-tool constant, e.g. EditorKit.ROTATE.
		 */
		public ToggleEditorkitEditPointsModeAction(ActionLocalizer localizer){
			super(null, KEY, localizer); // default, if no localizer is provided
		}

		/** overriden so as to dispatch to all currently open EditorKit's */
		public void actionPerformed(ActionEvent e){

			MDIManager mdiManager = JPicEdt.getMDIManager();
			PEDrawingBoard[] allOpenBoards = mdiManager.getAllDrawingBoards();

			if (e.getSource() instanceof AbstractButton){ // on=GLOBAL, off=LOCAL
				AbstractButton source = (AbstractButton)e.getSource();
				for (int i=0; i<allOpenBoards.length; i++){
					EditorKit kit = allOpenBoards[i].getCanvas().getEditorKit();
					if (kit != null && kit.getSelectionHandler() instanceof DefaultSelectionHandler){
						DefaultSelectionHandler h = (DefaultSelectionHandler)kit.getSelectionHandler();
						if (source.isSelected()) h.setHighlightingMode(DefaultSelectionHandler.GLOBAL_MODE);
						else h.setHighlightingMode(DefaultSelectionHandler.LOCAL_MODE);
					}
				}
			}
			else {
				for (int i=0; i<allOpenBoards.length; i++){
					EditorKit kit = allOpenBoards[i].getCanvas().getEditorKit();
					if (kit != null && kit.getSelectionHandler() instanceof DefaultSelectionHandler){
						DefaultSelectionHandler h = (DefaultSelectionHandler)kit.getSelectionHandler();
						h.toggleHighlightingMode();
					}
				}
			}

		}
	}


} // ActionRegistry

