/*
 GeneralCustomizer.java - December 31, 2001 - jPicEdt, a picture editor for LaTeX.
 Copyright (C) 1999-2007 Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.ui.dialog;

import jpicedt.Localizer;
import jpicedt.MiscUtilities;
import jpicedt.graphic.toolkit.AbstractCustomizer;
import jpicedt.graphic.PEToolKit;
import jpicedt.graphic.PECanvas;
import jpicedt.ui.*;
import jpicedt.ui.util.*;
import jpicedt.widgets.*;
import jpicedt.graphic.view.DefaultViewFactory;

import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.util.*;

/**
 * a panel for general preferences editing (LAF, language,...) 
 *
 * @author Sylvain Reynal
 * @since jPicEdt 1.3
 */
public class GeneralCustomizer extends AbstractCustomizer {

	private Properties preferences;
	private JLabel l;
	private JComboBox lookAndFeelList, mdiManagerList;
	private JComboBox languageList, contentTypeList;
	private JSlider maxUndoStepJS;
	private JLabel maxUndoStepL;
	//private JCheckBox showTextAsTCB;
	private FontSelector fontSelector;


	/**
	 * creates a new panel for general preferences. The given Properties is used to load widgets values,
	 * then to store new values.
	 */
	public GeneralCustomizer(Properties preferences){

		this.preferences = preferences;
		JPanel p;

		p = new JPanel(new GridLayout(6,2,5,5));
		p.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Localizer.currentLocalizer().get("preferences.General")));

		// Sub-Panel 1 : MDI Manager
		p.add(new JLabel(" "+Localizer.currentLocalizer().get("preferences.MDIManager") + " :"));
		mdiManagerList = new JComboBox(MDIManager.getInstalledMDIManagers());
		p.add(mdiManagerList);

		// Sub-Panel 2 : look'n feel
		p.add(new JLabel(" "+Localizer.currentLocalizer().get("preferences.LookAndFeel") + " :"));
		lookAndFeelList = new JComboBox(LAFManager.getInstalledLookAndFeelsNames());
		p.add(lookAndFeelList);

		// Sub-Panel 3 : language choice : fr, es, en...
		p.add(new JLabel(" "+Localizer.currentLocalizer().get("preferences.Language") + " :"));
		languageList =  new JComboBox(Localizer.currentLocalizer().getSupportedDisplayLanguages());
		// this is an array of language names localized using the current locale for jpicedt.Localizer
		p.add(languageList);

		// Sub-Panel 4 : undo/redo max step
		maxUndoStepL = new JLabel(" "+Localizer.currentLocalizer().get("preferences.UndoMaxStep") + " : ?");
		p.add(maxUndoStepL);
		maxUndoStepJS = new JSlider(0, 200, 100); // MaxValue = 200
		//maxUndoStepJS.setLabelTable(maxUndoStepJS.createStandardLabels(10));
		//maxUndoStepJS.setPaintLabels(true);
		maxUndoStepJS.setMajorTickSpacing(10);
		maxUndoStepJS.setPaintTicks(true);
		maxUndoStepJS.addChangeListener(new ChangeListener(){
			                                public void stateChanged(ChangeEvent e){
				                                maxUndoStepL.setText(" "+Localizer.currentLocalizer().get("preferences.UndoMaxStep") + " : " + Integer.toString(maxUndoStepJS.getValue()));
			                                }
		                                });
		p.add(maxUndoStepJS);

		// Sub-Panel 5 : font name
		p.add(new JLabel(" "+Localizer.currentLocalizer().get("preferences.Font")));
		p.add(fontSelector=new FontSelector());

		// Sub-Panel ? : display text as "T" or plain text
// 		l = new JLabel(" "+Localizer.currentLocalizer().get("preferences.ShowTextAsT.label")+" :");
// 		l.setToolTipText(Localizer.currentLocalizer().get("preferences.ShowTextAsT.tooltip"));
// 		p.add(l);
// 		showTextAsTCB = new JCheckBox(null, null, Options.showTextAsString);
// 		p.add(showTextAsTCB);

		// Sub-panel 6 : default content-type
		l = new JLabel(" "+Localizer.currentLocalizer().get("preferences.DefaultContentType.label")+" :");
		l.setToolTipText(Localizer.currentLocalizer().get("preferences.DefaultContentType.tooltip"));
		p.add(l);
		contentTypeList = new JComboBox(MiscUtilities.getAvailableContentTypesNames());
		p.add(contentTypeList);
		
		// ---
		add(p, BorderLayout.NORTH);
	}

	/**
	 * @return the panel title, used e.g. for Border or Tabpane title. 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public String getTitle(){ 
		return Localizer.currentLocalizer().get("preferences.General");	
	}
	
	
	/**
	 * @return the Icon associated with this panel, used e.g. for TabbedPane decoration 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public Icon getIcon(){ 
		return null;	
	}
	
	
	/**
	 * @return the tooltip string associated with this panel 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public String getTooltip(){ return Localizer.currentLocalizer().get("preferences.GeneralConfiguration");	}

	/**
	 * Load widgets display content with a default value retrieved from the Options class. 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void loadDefault() {

		lookAndFeelList.setSelectedItem(LAFManager.getDefaultLafName());
		mdiManagerList.setSelectedItem(MDIManager.getDefaultManagerName());
		languageList.setSelectedItem(Localizer.currentLocalizer().getDefaultLanguage());
		contentTypeList.setSelectedIndex(MiscUtilities.DEFAULT_CONTENT_TYPE_INDEX);
		maxUndoStepJS.setValue(PECanvas.MAX_UNDOABLE_STEPS_DEFAULT);
		fontSelector.setFont(DefaultViewFactory.TEXT_FONT_DEFAULT);
		//showTextAsTCB.setSelected(Options.showTextAsStringDEFAULT);
	}

	/**
	 * Load widgets value from the Options class. 
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	public void load() {
		lookAndFeelList.setSelectedItem(LAFManager.getCurrentLafName());
		mdiManagerList.setSelectedItem(jpicedt.JPicEdt.getMDIManager().getName());
		languageList.setSelectedItem(Localizer.currentLocalizer().getCurrentDisplayLanguage());
		contentTypeList.setSelectedIndex(MiscUtilities.getContentTypeIndex(preferences.getProperty(PECanvas.KEY_CONTENT_TYPE)));
		maxUndoStepJS.setValue(MiscUtilities.parseProperty(preferences,PECanvas.KEY_UNDOABLE_STEPS,PECanvas.MAX_UNDOABLE_STEPS_DEFAULT));

		// font :
		String val = preferences.getProperty(DefaultViewFactory.KEY_TEXT_FONT);
		Font textFont;
		if (val==null) textFont = DefaultViewFactory.TEXT_FONT_DEFAULT;
		else textFont = Font.decode(val);
		fontSelector.setFont(textFont);
	}

	/**
	 * Store current widgets value to the Properties object given in the constructor.
	 * @since jPicEdt
	 * @author Sylvain Reynal
	 */
	 public void store(){
		
		preferences.setProperty(LAFManager.KEY_LAF,(String)lookAndFeelList.getSelectedItem());
		preferences.setProperty(MDIManager.KEY_MDIMANAGER,(String)mdiManagerList.getSelectedItem());
		
		Locale[] supportedLocales = Localizer.currentLocalizer().getSupportedLocales();
		String languageISO = supportedLocales[languageList.getSelectedIndex()].getLanguage(); // e.g. "en" or "fr"
		preferences.setProperty(Localizer.KEY_LANGUAGE,languageISO); // ISO Code !
		
		preferences.setProperty(PECanvas.KEY_CONTENT_TYPE, MiscUtilities.getAvailableContentTypes()[contentTypeList.getSelectedIndex()]);
		preferences.setProperty(PECanvas.KEY_UNDOABLE_STEPS, Integer.toString(maxUndoStepJS.getValue()));

		//Options.showTextAsString = showTextAsTCB.isSelected();
		Font textFont = fontSelector.getFont();
		String textProp = MiscUtilities.formatFontAsProperties(textFont);
		preferences.setProperty(DefaultViewFactory.KEY_TEXT_FONT,textProp);
	}

} // class


