/*  jPicEdt, a picture editor for LaTeX.
    Copyright (C) 1999-2007  Sylvain Reynal

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Sylvain Reynal
    Département de Physique
    Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
    6, avenue du Ponceau
    95014 CERGY CEDEX
    FRANCE

    Tel : 00 +33 130 736 245
    Fax : 00 +33 130 736 667
    e-mail : reynal@ensea.fr
    jPicEdt web page : http://www.jpicedt.org
*/

package jpicedt.ui.util;

import jpicedt.*;
import jpicedt.graphic.io.formatter.*;
import jpicedt.graphic.toolkit.*;
import jpicedt.graphic.model.*;
import jpicedt.graphic.*;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;

/**
 * A class implementing methods for calling external processes like latex, (x)dvi, ghostscript...
 * on the current picture file (e.g. the active internal frame)<p>
 * <p>
 * A "tmp" file is generated on-the-fly in the "tmp" directory of the underlying platform, 
 * This tmp file contains an adequate preamble (e.g. usepackage{epic}) depending on current options, and
 * LaTeX commands (or PsTricks or eepic commands) corresponding to each object found in the given picture.
 *<p>
 * These command lines may include the following special symbols :<br> 
 * - {f} : tmp file name w/o extension (the user need to specify explicitely any extension required by the external program)<br>
 * - {p} : tmp file path including trailing separator<br>
 * - {i} : jPicEdt add-ons dir, that is : "jpicedt.home/add-ons" (trailing separtor excluded)
 * <p>
 * Ex : kghostview {p}{f}.ps or {i}run_latex.bat {p}{f}
 * <p>
 * Since we don't want to resort on platform dependent tricks (working dirs, PATH variables,
 * root directory names differing on UNIX and Win32 systems, latex command line which might also differ, aso...),
 * the better thing the user could do is to write one batch/shell script
 * per process he wants to launch, then provide the names of these scripts to PicEdt.
 * <p>
 * Scripts are easy to write on UNIX/Window platforms,
 * and very likely on MacOS platforms by using AppleScript.
 *
 * @author Sylvain Reynal
 * @since PicEdt 1.2
 * @version $Id: RunExternalCommand.java,v 1.10.2.1 2007/09/02 11:56:30 reynal Exp $
 *
 */
public class RunExternalCommand extends Thread {

	/** static field defining an external "latex" command, and also used by ActionLocalizer */
	public final static String LATEX = "action.command.LaTeX";
	/** static field defining an external "xdvi" (or similar) command, and also used by ActionLocalizer */
	public final static String DVIVIEWER = "action.command.DVI";
	/** static field defining an external "dvips" command, and also used by ActionLocalizer */
	public final static String DVIPS = "action.command.Dvips";
	/** static field defining an external "gv" (or similar) command, and also used by ActionLocalizer */
	public final static String GHOSTVIEW = "action.command.Ghostview";
	/** static field defining an external user command, and also used by ActionLocalizer */
	public final static String USER1 = "action.command.UserProgram1";
	/** static field defining an external user command, and also used by ActionLocalizer */
	public final static String USER2 = "action.command.UserProgram2";
	/** static field defining an external bitmap creation command, and also used by ActionLocalizer */
	public final static String BITMAP_CREATION = "action.command.BitmapCreation";

	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_LATEX = "command.latex";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_DVIVIEWER = "command.dvi";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_DVIPS = "command.dvips";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_GHOSTVIEW = "command.ghostview";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_USER1 = "command.user1";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_USER2 = "command.user2";
	/** key used to fetch command parameters from a Properties object */
	public final static String KEY_BITMAP_CREATION = "command.bitmap";

	/** default commands parameters */
	public static final String DEFAULT_LATEX_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_DVI_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_DVIPS_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_GS_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_USER1_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_USER2_COMMAND = "";
	/** default commands parameters */
	public static final String DEFAULT_BITMAP_CREATION_COMMAND = "";

	////////////////////////////////// PRIVATE FIELDS /////////////////////////////////

	private Drawing drawing;
	private ContentType contentType;
	private Process proc; // the process being executed (null if none)
	private PrintStream printStream; // command's input stream (shared by inner classes)
	private String commandLine; // command line created on-the-fly from the Properties and commandToRun
	private static File tmpFile; // temp file created by the JVM the first time a command is run
	private static String tmpPath; // path for tmp file (e.g. "/tmp/"), this is {%p}
	private static String tmpFilePrefix; // file prefix, e.g. "jpicedt4875" (w/o extension)
	private static String addonDir; // jPicEdt's add-ons directory, trailing "/" included

	private static DlgBoxDisplayProcessIOStream dlgBox;
	// --- for debugging purpose ---
	public static boolean isGUI=true; // if true, redirect process output stream to a dialog box, otherwise to System.out 

	/////////////////////////////////// METHODS /////////////////////////////////////

	/**
	 * @param drawing The target document upon which the command will acts
	 * @param contentType the content-type used to format the document
	 * @param commandToRun one of the predefined constant fields (e.g. LATEX, DVIVIEWER, DVIPS or GHOSTVIEW)
	 * @since jPicEdt 1.2
	 */
	public RunExternalCommand(Drawing drawing, ContentType contentType,String commandToRun) {

		super("RunExtCom");
		this.contentType = contentType;
		this.drawing = drawing;

		try {

			createTmpFile();

			// create command line :
			if (commandToRun == LATEX){
				writeTmpTeXFile();
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_LATEX,DEFAULT_LATEX_COMMAND));
			}
			else if (commandToRun==BITMAP_CREATION){
				writeTmpTeXFile();
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_BITMAP_CREATION,DEFAULT_BITMAP_CREATION_COMMAND));
			}
			else if (commandToRun == DVIVIEWER){
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_DVIVIEWER,DEFAULT_DVI_COMMAND));
			}
			else if (commandToRun == DVIPS){
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_DVIPS,DEFAULT_DVIPS_COMMAND));
			}
			else if (commandToRun == GHOSTVIEW){
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_GHOSTVIEW,DEFAULT_GS_COMMAND));
			}
			else if (commandToRun == USER1){
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_USER1,DEFAULT_USER1_COMMAND));
			}
			else if (commandToRun == USER2){
				commandLine=buildCommandLine(JPicEdt.getProperty(KEY_USER2,DEFAULT_USER2_COMMAND));
			}
		}
		catch(IOException ioEx){
			JOptionPane.showMessageDialog(null,
			                              Localizer.currentLocalizer().get("exception.IOError") + " : " + ioEx.getMessage(),
			                              commandToRun,
			                              JOptionPane.ERROR_MESSAGE);
			commandLine = null;
			// now command line is null, and run() takes care of that.
		}

	}
	
	/**
	 * Mostly for debugging purpose, or when used outside the jpicedt.ui package (since then JPicEdt.preferences
	 * are no longer available).
	 * @param drawing The target document upon which the command will acts
	 * @param contentType the content-type used to format the document
	 * @param cmdLine a string containing the command to be executed (e.g. "ext_prod.sh {p} latex {f}")
	 * @since jPicEdt 1.4pre5
	 */
	public RunExternalCommand(Drawing drawing, ContentType contentType,String cmdLine, boolean isWriteTmpTeXfile) {

		super("RunExtCom");
		this.contentType = contentType;
		this.drawing = drawing;

		try {
			createTmpFile();
			if (isWriteTmpTeXfile) writeTmpTeXFile();
			this.commandLine=buildCommandLine(cmdLine);
		}
		catch(IOException ioEx){
			JOptionPane.showMessageDialog(null,
			                              Localizer.currentLocalizer().get("exception.IOError") + " : " + ioEx.getMessage(),
			                              cmdLine,
			                              JOptionPane.ERROR_MESSAGE);
			commandLine = null;
			// now command line is null, and run() takes care of that.
		}

	}
	
	public static File getTmpFile(){
		return tmpFile;
	}
	
	public static String getTmpPath(){
		return tmpPath;
	}
	
	public static String getTmpFilePrefix(){
		return tmpFilePrefix;
	}

	private void writeTmpTeXFile() throws IOException {
		// create "/tmp/jpicedt.tex" :
		BufferedWriter bw = new BufferedWriter(new FileWriter(tmpFile));
		bw.write(contentType.createFormatter().createFormatter(drawing,FormatterFactory.MAKE_STANDALONE_FILE).format());
		bw.newLine();
		bw.close();
	}

	/**
	 * Build the real command line by replacing occurences of special symbols ({f}, {p},...)  in the
	 * given string by their actual value
	 *
	 * @param command command line with {f} and{p}, as specified by the user in the Preferences tabpane
	 * @return The real command line
	 * @since PicEdt 1.2
	 */
	private String buildCommandLine(String command){

		StringTokenizer tokenizer = new StringTokenizer(command,"{}",true); // we don't remove spaces ; tokens are returned
		String realCommand = new String();
		String token;

		while (tokenizer.hasMoreTokens()){
			token = tokenizer.nextToken();
			if (token.equals("{")) {
				token = tokenizer.nextToken(); // should be 'p' or 'f' or 'i'
				if (token.equals("p")) {realCommand += tmpPath; continue;}
				if (token.equals("f")) {realCommand += tmpFilePrefix; continue;}
				if (token.equals("i")) {realCommand += getAddonDir(); continue;}
				continue; // everything else get rejected
			}
			if (token.equals("}")) continue;
			realCommand += token;
		}
		realCommand = realCommand.replace('/',File.separatorChar);
		return realCommand;
	}

	/**
	 * run an external process as a thread (overloading Thread.run() method)
	 * we then pipe the process std output to a JTextArea
	 *
	 * @since PicEdt 1.2
	 */
	public void run() {

		if (commandLine == null) return;

		if (isGUI && dlgBox==null) dlgBox = new DlgBoxDisplayProcessIOStream();

		// run process
		try{
			proc = Runtime.getRuntime().exec(commandLine);
		}
		catch(IOException e){
			if (dlgBox!=null) dlgBox.append(e.toString());
			else e.printStackTrace();
			return;
		}

		// fetch process's output stream (an input stream from the JVM's point of view) and plug a buffered reader input stream on it
		BufferedReader  bufferedReader= new BufferedReader(new InputStreamReader(proc.getInputStream()));

		// fetch process's input stream (used by DlgBox)
		printStream = new PrintStream(proc.getOutputStream(),true);

		// read output chars from command (we can't use readLine() since it blocks until it encounters a CR, which
		// can be annoying in case the running process waits for an input, e.g. with a  "?" or "*" prompt as for LaTeX)
		try{
			int c;
			while((c=bufferedReader.read()) != -1){
				if (dlgBox != null) dlgBox.append(new Character((char)c).toString());
				else System.out.print((char)c); 
			}
			//System.out.println();
		}
		catch(IOException e){
			if (dlgBox != null) dlgBox.append(e.toString());
			else e.printStackTrace();
		}

		// wait for command to terminate
		try {
			proc.waitFor();
		}
		catch (InterruptedException e) {
			if (dlgBox != null) dlgBox.append("\n"+Localizer.currentLocalizer().get("exception.RunExtCmdProcessInterrupted"));
			else e.printStackTrace();
		}

		// check its exit value

		if (proc.exitValue() != 0){
			if (dlgBox != null) dlgBox.append("\n"+Localizer.currentLocalizer().get("exception.RunExtCmdExitValueNonNil"));
			else System.out.println("Process terminated with non-null value");
		}
			

		// close stream
		try{
			bufferedReader.close();
		}
		catch(IOException e){
			if (dlgBox != null) dlgBox.append(e.toString());
			else e.printStackTrace();
		}

		if (dlgBox != null) {
			dlgBox.append("\njPicEdt external process: process terminated (OK).");
			//dlgBox.dispose();
		}
		else System.out.println("\njPicEdt external process: process terminated (OK).");

		return;
	}

	/**
	 * Delete files created by JPicEdt or programms called by JPicEdt (LaTeX, dvips,...).
	 */
	public static void cleanTmpDir(){
		if (tmpPath==null) return; // ok, nothing dirty yet !
		String[] list = new File(tmpPath).list(new FilenameFilter(){
			public boolean accept(File dir, String name){
				return name.startsWith(tmpFilePrefix);
		}});
		// delete them :
		for (int i=0; i<list.length; i++){
			//System.out.println("deleting " + new File(tmpPath,list[i]));
			new File(tmpPath,list[i]).delete();
		}
		tmpFile = null; // force re-creation of tmp file on next call to constructor
	}
		
	/**
	 * Return the add-ons directory w/o trailing "/"
	 */
	public static String getAddonDir(){
		if (addonDir==null) {
			addonDir = MiscUtilities.getJPicEdtHome() + File.separator + "add-ons";
			//System.out.println("addOnDir = " + addonDir);
		}
		return addonDir;
	}

	/**
 	 * Create a new tmp file if it doesn't exit yet
	 */
	public static void createTmpFile() throws IOException {
		if (tmpFile != null) return;
		tmpFile = File.createTempFile("jpicedt",".tex", JPicEdt.getTmpDir());
		tmpFile.deleteOnExit();
		tmpPath = tmpFile.getParent() + File.separator; // "/tmp/"
		String name =  tmpFile.getName(); // "jpicedt1432.tex"
		tmpFilePrefix = name.substring(0,name.indexOf(".")); // "jpicedt1432"
	}
		
	/**
	 * @return an AbstractCustomizer suited for editing Properties related to external commands
	 */
	public static AbstractCustomizer createCustomizer(Properties preferences){
		return new Customizer(preferences);
	}

	/////////////////////////////////////////////////////////////////////////
	/**
	 * An inner class implementing a JDialog to display external process
	 * inputStream and send keyboard events to process outputStream
	 */
	public class DlgBoxDisplayProcessIOStream extends JDialog implements ActionListener,KeyListener, Runnable{

		/** panelStream contains an editable JTextArea and a JScrollPane */
		JPanel panelStream = new JPanel();
		JTextArea streamTA = new JTextArea(25,50);
		JScrollPane scrollStream = new JScrollPane(streamTA);
		StringBuffer areaBuffer = new StringBuffer(); // buffer for async. filling of streamTA

		/** panel containing the buttons */
		JPanel panelButtons = new JPanel();;
		JButton buttonKill = new JButton(Localizer.currentLocalizer().get("commands.Kill"));

		/** buffer for typed chars (see KeyTyped) */
		StringBuffer strBuf = new StringBuffer();
		
		boolean visible = false; // force show on first call to append

		/**
		 * @since PicEdt 1.2
		 */
		DlgBoxDisplayProcessIOStream() {

			super();
			this.setTitle(Localizer.currentLocalizer().get("commands.boxtitle"));
			this.setResizable(true);
			this.setModal(false);
			this.getContentPane().setLayout(new BorderLayout());
			//this.setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
			streamTA.setEditable(false);
			panelStream.add(scrollStream);
			panelStream.setBorder(BorderFactory.createEtchedBorder());
			buttonKill.addActionListener(this);
			panelButtons.add(buttonKill);
			panelButtons.setBorder(BorderFactory.createEtchedBorder());
			this.getContentPane().add(panelStream,BorderLayout.NORTH);
			this.getContentPane().add(panelButtons,BorderLayout.SOUTH);
			//Dimension dlgSize = this.getPreferredSize();
			//this.setSize(getPreferredSize());
			streamTA.requestFocus(); // so that entering command after a prompt doesn't need a mouse click on the JTextArea
			streamTA.addKeyListener(this);
		}

		/** append text ; thread safe */
		public void append(String text){
			areaBuffer.append(text);
			SwingUtilities.invokeLater(new Thread(this)); 
			
		}
		
		/** handle asynchronous widgets updating */
		public void run(){
			if (!visible) {
				this.pack();
				this.setVisible(true);
				visible=true;
			}
			if (streamTA.getLineCount() > 100) streamTA.setText("");
			streamTA.append(areaBuffer.toString());
			//System.out.println(areaBuffer.toString());
			areaBuffer = new StringBuffer();
		}

		public void keyPressed(KeyEvent e){

			if (e.getKeyCode() == KeyEvent.VK_ENTER && printStream != null) {

				printStream.println(strBuf.toString()); // flush buffer, sending all typed keys + CR to external process
				strBuf = new StringBuffer();
				streamTA.append("\n");
			}
		}

		public void keyReleased(KeyEvent e){}

		public void keyTyped(KeyEvent e){

			strBuf.append(e.getKeyChar()); // accumulate types keys in buffer
			streamTA.append(new Character(e.getKeyChar()).toString());

		}

		// called by KILL :
		public void actionPerformed(ActionEvent e) {
			proc.destroy();
			dispose();
		}

	} // inner class DlgBoxDisplayProcessIOStream

	///////////////////////////////////////////////////////////////////////////

	/**
	 * A Customizer for the configuration of external commands preferences.
	 * @author Sylvain Reynal
	 * @since PicEdt 1.3
	 */
	private static class Customizer extends AbstractCustomizer implements ActionListener {

		private JTextField latexCommandTF, dviviewerCommandTF, dvipsCommandTF, ghostviewCommandTF, user1CommandTF, user2CommandTF;
		private Properties preferences;
		private JButton loadConfigBT;

		/**
		 * Creates a new panel for the configuration of external commands
		 */
		public Customizer(Properties preferences){

			this.preferences  = preferences;

			JPanel p;

			p = new JPanel(new GridLayout(7,2,5,5));
			p.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),Localizer.currentLocalizer().get("commands.title")));

			p.add(new JLabel(" "+"LaTeX"+" :"));
			latexCommandTF = new JTextField(20);
			p.add(latexCommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("action.command.DVI")+" :"));
			dviviewerCommandTF = new JTextField(20);
			p.add(dviviewerCommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("action.command.Dvips")+" :"));
			dvipsCommandTF = new JTextField(20);
			p.add(dvipsCommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("action.command.Ghostview")+" :"));
			ghostviewCommandTF = new JTextField(20);
			p.add(ghostviewCommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("action.command.UserProgram1") + " :"));
			user1CommandTF = new JTextField(20);
			p.add(user1CommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("action.command.UserProgram2") + " :"));
			user2CommandTF = new JTextField(20);
			p.add(user2CommandTF);

			p.add(new JLabel(" "+Localizer.currentLocalizer().get("commands.LoadPredefinedConfig")));
			loadConfigBT = new JButton(Localizer.currentLocalizer().get("button.Load"));
			loadConfigBT.addActionListener(this);
			p.add(loadConfigBT);
			
			add(p,BorderLayout.NORTH);
			
			// reminder for {f}, {p}, ... shortcuts :
			Box helpPanel = new Box(BoxLayout.Y_AXIS);
			//helpPanel.add(new JLabel(" "+Localizer.currentLocalizer().get("Commands.helpstring")));
			try {createTmpFile();} // security to avoid "null" !
			catch (IOException ex){}
			helpPanel.add(new JLabel("Tmp. dir : "));
			helpPanel.add(new JLabel(" {p} -> \"" + tmpPath + "\""));
			helpPanel.add(new JLabel("Tmp. TeX file : "));
			helpPanel.add(new JLabel(" {f} -> \"" + tmpFilePrefix + "\""));
			helpPanel.add(new JLabel("Add-ons dir. : "));
			helpPanel.add(new JLabel(" {i} -> \"" + getAddonDir() + "\""));
			JPanel borderingPanel = new JPanel();
			borderingPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),Localizer.currentLocalizer().get("menu.Help")));
			borderingPanel.add(helpPanel);
			add(borderingPanel, BorderLayout.CENTER);
		}

		/**
		* @return the panel title, used e.g. for Border or Tabpane title. 
		* @since jPicEdt
		* @author Sylvain Reynal
		*/
		public String getTitle(){
			return Localizer.currentLocalizer().get("commands.title");
		}


		/**
		 * @return the Icon associated with this panel, used e.g. for TabbedPane decoration 
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		public Icon getIcon(){
			return null;
		}


		/**
		 * @return the tooltip string associated with this panel 
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		public String getTooltip(){
			return Localizer.currentLocalizer().get("commands.tooltip");
		}

		/**
		 * Load widgets display content with a default value retrieved from the Properties object
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		public void loadDefault() {
			latexCommandTF.setText(DEFAULT_LATEX_COMMAND);
			dviviewerCommandTF.setText(DEFAULT_DVI_COMMAND);
			dvipsCommandTF.setText(DEFAULT_DVIPS_COMMAND);
			ghostviewCommandTF.setText(DEFAULT_GS_COMMAND);
			user1CommandTF.setText(DEFAULT_USER1_COMMAND);
			user2CommandTF.setText(DEFAULT_USER2_COMMAND);
		}

		/**
		 * Load widgets value from the Properties object given in the constructor
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		private void load(Properties pref) {
			latexCommandTF.setText(pref.getProperty(KEY_LATEX,DEFAULT_LATEX_COMMAND));
			dviviewerCommandTF.setText(pref.getProperty(KEY_DVIVIEWER,DEFAULT_DVI_COMMAND));
			dvipsCommandTF.setText(pref.getProperty(KEY_DVIPS,DEFAULT_DVIPS_COMMAND));
			ghostviewCommandTF.setText(pref.getProperty(KEY_GHOSTVIEW,DEFAULT_GS_COMMAND));
			user1CommandTF.setText(pref.getProperty(KEY_USER1,DEFAULT_USER1_COMMAND));
			user2CommandTF.setText(pref.getProperty(KEY_USER2,DEFAULT_USER2_COMMAND));
		}

		public void load(){
			load(JPicEdt.getPreferences());
		}
		
		/**
		 * ActionHandler for "loadConfigBT"
		 */
		public void actionPerformed(ActionEvent e){
			JFileChooser fileChooser = new JFileChooser(getAddonDir());
			fileChooser.setDialogTitle(Localizer.currentLocalizer().get("commands.LoadPredefinedConfig"));
			fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter(){
				public boolean accept(File f){return f.isDirectory() || f.getPath().endsWith(".properties");}
				public String getDescription(){return "External scripts configuration files";}
			});
			if (fileChooser.showOpenDialog(null) == JFileChooser.CANCEL_OPTION) return;
			File file =  fileChooser.getSelectedFile();
			if (file!= null) {
				Properties prop = new Properties();
				try {
					InputStream is = new BufferedInputStream(new FileInputStream(file));
					prop.load(is);
					//prop.list(System.out);
					load(prop);
					is.close();
				}
				catch (IOException ioEx){
					JOptionPane.showMessageDialog(
						JOptionPane.getRootFrame(),
						Localizer.currentLocalizer().get("exception.IOError") + "\n" + ioEx.getMessage(),
						Localizer.currentLocalizer().get("commands.LoadPredefinedConfig"),
						JOptionPane.ERROR_MESSAGE);
				}
			}
		}

		/**
		 * Store current widgets value to the Properties object given in the constructor
		 * @since jPicEdt
		 * @author Sylvain Reynal
		 */
		public void store(){
			preferences.setProperty(KEY_LATEX,latexCommandTF.getText());
			preferences.setProperty(KEY_DVIVIEWER,dviviewerCommandTF.getText());
			preferences.setProperty(KEY_DVIPS,dvipsCommandTF.getText());
			preferences.setProperty(KEY_GHOSTVIEW,ghostviewCommandTF.getText());
			preferences.setProperty(KEY_USER1,user1CommandTF.getText());
			preferences.setProperty(KEY_USER2,user2CommandTF.getText());
		}


	} // PanelExternalCommand


} // RunExternalCommand
