/*
 PsObjectExpression.java - August 4, 2002 - jPicEdt, a picture editor for LaTeX.
 Copyright (C)  Sylvain Reynal

 Département de Physique
 Ecole Nationale Supérieure de l'Electronique et de ses Applications (ENSEA)
 6, avenue du Ponceau
 F-95014 CERGY CEDEX

 Tel : +33 130 736 245
 Fax : +33 130 736 667
 e-mail : reynal@ensea.fr
 jPicEdt web page : http://www.jpicedt.org/
  
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or any later version.
  
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
  
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package jpicedt.format.input.pstricks;

import jpicedt.format.input.util.*;
import jpicedt.graphic.model.*;

import java.util.*;

/**
 * Handles <b>\\newpsobject{myobj}{ps_object}{par1=val1,...}</B>, for instance
 * <b>\\newpsobject{myline}{psline}{linecolor=green}</B> (param is optional, though in this case, it'd be pretty useless!)
 * @author Sylvain Reynal
 * @version $Id: PsObjectExpression.java,v 1.4.2.1 2007/09/02 11:56:12 reynal Exp $ 
 */
public class PsObjectExpression extends AbstractRegularExpression {

	private Pool pool;
	private AbstractRegularExpression exp1,exp2,exp3;
	private String macroName; // e.g. "\\myline"
	private String shape; // e.g. "psline"
	private String param; // e.g. "linecolor=green,fillstyle=solid"
	private PstricksParser parser; // used to add new rules
	private Element prototype;
	
	public PsObjectExpression(PstricksParser p, Pool pl){
		this.pool = pl;	
		this.parser = p;
		// myobj}
		exp1 = new WordExpression("}", true, true, true){ // no line-feed + postfix.
			public void action(ParserEvent e){
				if (DEBUG) System.out.println(e); 
				macroName = "\\" + getValue(); // e.g. "\\PST@Border"
		}};
		// ps_object}
		exp2 = new WordExpression("}", true, true, true){ // no line-feed
			public void action(ParserEvent e){
				if (DEBUG) System.out.println(e); 
				shape = getValue(); // e.g. "psline" or "psbezier"
		}};
		// params :
		exp3 = new OptionalExpression(new EnclosingExpression("{",null,"}"){
			public void action(ParserEvent e){
				if (DEBUG) System.out.println(e);
				param = getEnclosedString(); // e.g. "linecolor=green,fillstyle=solid"
		}});
		
		
	}
	
	public boolean interpret(Context context) throws REParserException {
		
		int oldCaretPos = context.getCaretPosition(); // if fails in the end
		
		if (!context.matchAndMove("\\newpsobject{")) return false;
		WHITE_SPACES_OR_EOL.interpret(context);

		// retrieve "macroName" :
		if (!exp1.interpret(context)) throw new REParserException.IncompleteSequence(context,this);
		WHITE_SPACES_OR_EOL.interpret(context);

		// retrieve "shape" :
		if (!context.matchAndMove("{")) throw new REParserException.IncompleteSequence(context,this);
		if (!exp2.interpret(context)) throw new REParserException.IncompleteSequence(context,this);
		
		// (optionally) fills attribute set assoc. with PstricksParser.KEY_ATTRIBUTES_PSOBJECTS :
		if (exp3.interpret(context)){

			// associates macroName with paramString for later use at instanciation time, 
			// e.g. associates "\\psline" with "linecolor=green" :
			HashMap map = (HashMap)pool.get(PstricksParser.KEY_NEWPSOBJECTS);
			map.put(macroName,param);
		}

		// add new rule to parser :
		if (shape.equals("psbezier")) {
			parser.add(new PsBezierExpression(pool,macroName));
			return true;
		}
		else if (shape.equals("pscircle")) {
			parser.add(new PsCircleExpression(pool,macroName));
			return true;
		}
		else if (shape.equals("psellipse")) {
			parser.add(new PsEllipseExpression(pool,macroName));
			return true;
		}
		else if (shape.equals("psframe")) {
			parser.add(new PsFrameExpression(pool,macroName));
			return true;
		} 
		else if (shape.equals("pspolygon")) {
			parser.add(new PsPolygonExpression(pool,PsPolygonExpression.POLYGON,macroName));
			return true;
		} 
		else if (shape.equals("psline")) {
			parser.add(new PsPolygonExpression(pool,PsPolygonExpression.LINE,macroName));
			return true;
		} 
		else if (shape.equals("psdots")) {
			parser.add(new PsPolygonExpression(pool,PsPolygonExpression.DOTS,macroName));
			return true;
		} 
		else if (shape.equals("psarc")) {
			parser.add(new PsArcExpression(pool,PsArcExpression.ARC,macroName));
			return true;
		} 
		else if (shape.equals("pswedge")) {
			parser.add(new PsArcExpression(pool,PsArcExpression.WEDGE,macroName));
			return true;
		} 
		else { // shape not supported
			context.moveCaretTo(oldCaretPos);
			return false; // skip, and go to next expression in the parser tree
		}
		
		
	}

	public String toString(){
		return "[PsObjectExpression]";
	}
	
}
